(* ::Package:: *)

(* ::Title::RGBColor[0., 0., 0.]:: *)
(*MSSM Pole Mass Using EFT*)


(* ::Chapter::RGBColor[0., 0., 1.]::Closed:: *)
(*---*)


(* ::Subchapter::Closed:: *)
(*Usage*)


BeginPackage["PoleMass`"];



Print["
---------------------
PoleMass C00
Oct 2017

A. Daniel Ochoa F.
ochoa@physik.rwth-aachen.de
RWTH Aachen
Institute for Theoretical Particle Physics and Cosmology
---------------------
Calculation of mass expansions in the MSSM
---------------------

Loading...
"];



M::usage="

M[Subindex,LimitingCase][LoopOrder,AplhaTOrder] calculates the MSSM pole mass expansion in the DR scheme.
This functions expresses the mass expansion in terms of:

k,mt,at,a3, e.g.

Mh^2 = MZ^2 Cos[2 beta]^2 
+ mt^2 ( at(k^(1/2)Deltamh2_at + k a3 Deltamh2_atas + k^(3/2) a3^2 Deltamh2_atas2 + ... ) +  at^2(k Deltamh2_at2 + ... )).

It can take 2 subindices: 

't', and
'h'.

Available limiting cases are: 

'GeneralSolution', which returns the general mass expansion with symbolic threshold corrections,
'Degenerate', which sets every soft-SUSY mass parameter equal to MS and Xt & Tan[beta] are left free,
'NonDegenerate', which is the complete solution.

The mass expansion contains the following soft-SUSY parameters: 

mu, mA, m3, mQ3, mU3, msq, Xt, Yt and beta,

and in the 'GeneralSolution' case the DR threshold corrections, 'Dxat(n)as(m)', appear.

The Higgs boson DR mass expansion (subindex 'h') can be calculated up to O(mt^2 at a3^3) X O(Log[MR^2/mt^2]^(n-2)), and partial O(mt^2 at^2 a3) & O(mt^2 at^3).
alphatOrder is set to 1 by defualt and can be ommited if no higher powers are desired to be computed.

";

(*
'CompletelyDegenerate', which sets every soft-SUSY mass parameter equal to MR and Xt & Tan[beta] are left free, and
'Harlander', which is similar to the degenerate limit but leaves the squark masses free, and sets Xt=0.
'Custom', where one can choose the desired value of the SUSY parameters by evaluating the following association rule:

Custom=<|
'mu'->mu,
'mA'->mA,
'm3'->m3,
'mQ3'->mQ3,
'mU3'->mU3,
'msq'->msq,
'Xt'->Xt,
'MR'->MR
|>,

but using quotation marks instead of apostrophes in the lhs of the rule, and the desired value of the parameter on the rhs.
Note: certain values of the parameters render the threshold corrections unstable and lead to divergent results in this program.
*)

Mx::usage="

Mx[Subindex,LimitingCase,Scheme][LoopOrder,AplhaTOrder] calculates the MSSM pole mass expansion in the DR scheme.
This functions expresses the mass expansion in terms of:

k,vu,yt,g3, e.g.

Mh^2 = MZ^2 Cos[2 beta]^2 
+ yt^2 vu^2( yt^2 Sin[beta]^2(k Deltamh2_at + k^2 g3^2 Deltamh2_atas + k^3 g3^4 Deltamh2_atas2 + ... ) +  yt^4 Sin[beta]^4(k^2 Deltamh2_at2 + ... )).
 
It can take 2 subindices: 

't', and
'h'.

Available limiting cases are: 

'FromDRToMS', which converts the DR expansion back to MS and must agree with the original MS expansion. Must be used with 'DR' scheme (see below).
'GeneralSolution', which returns the general mass expansion with symbolic threshold corrections,
'Degenerate', which sets every soft-SUSY mass parameter equal to MS and Xt & Tan[beta] are left free,
'NonDegenerate', which is the complete solution.

The mass expansion contains the following soft-SUSY parameters: 

mu, mA, m3, mQ3, mU3, msq, Xt, Yt and beta,

and in the 'GeneralSolution' case the MS threshold corrections, 'dxat(n)as(m)', appear even when the expansion is in the DR scheme.

Available schemes are:

'SM', returns the SM mass expansion in the MS scheme,
'MS', returns the MSSM mass expansion in the MS scheme,
'DR', returns the MSSM mass expansion in the DR Scheme.

The Higgs boson DR mass expansion (subindex 'h') can be calculated up to O(mt^2 at a3^3) X O(Log[MR^2/mt^2]^(n-2)), and partial O(mt^2 at^2 a3) & O(mt^2 at^3).
alphatOrder is set to 1 by defualt and can be ommited if no higher powers are desired to be computed.

All other functions in this program that compute Higgs mass expansions are based on this function.

";

Mh2::usage="

Mh2[_] is a variable containing the complete pre-loaded 3-loop order MSSM DR Higgs boson mass expansionof the form, e.g.

Mh^2 = MZ^2 Cos[2 beta]^2 
+ mt^2 ( at(k^(1/2)Deltamh2_at + k a3 Deltamh2_atas + k^(3/2) a3^2 Deltamh2_atas2 + ... ) +  at^2(k Deltamh2_at2 + ... )).

The blank, _ , can be filled with:

'EFT', which contains the mass expasion expressed in terms of mt,at,a3 and, as an expansion computed using the EFT, this expansion is organized in powers of Log[MR^2/mt^2] as follows:
This expansion is organized as follows:

{Log[MR^2/mt^2],0,3},\[IndentingNewLine]{Log[m3^2/MR^2],0,3},\[IndentingNewLine]{Log[mQ3^2/MR^2],0,3},\[IndentingNewLine]{Log[mU3^2/MR^2],0,3},\[IndentingNewLine]{Log[msq^2/MR^2],0,3},\[IndentingNewLine]{PolyLog[2,1-m3^2/mQ3^2],0,2},\[IndentingNewLine]{PolyLog[2,1-m3^2/mU3^2],0,2},\[IndentingNewLine]{PolyLog[2,1-mU3^2/mQ3^2],0,2},\[IndentingNewLine]{PolyLog[2,1-msq^2/m3^2],0,2},\[IndentingNewLine]{PolyLog[2,1-msq^2/mQ3^2],0,2},\[IndentingNewLine]{PolyLog[2,1-msq^2/mU3^2],0,2},\[IndentingNewLine]{Xt,0,8},

and the coefficient of these variables is a rational function of the Soft-SUSY parameters.

'DR', which contains the mass expansion expressed in terms of vu,yt,g3 and, as a proper expansion of a high scale theory in the DR scheme, this expansion is organized in powers of Log[MR^2/mQ3^2] as follows:

{Log[MR^2/mQ3^2],0,3},\[IndentingNewLine]{Log[mt^2/mQ3^2],0,3},\[IndentingNewLine]{Log[m3^2/mQ3^2],0,3},\[IndentingNewLine]{Log[mU3^2/mQ3^2],0,3},\[IndentingNewLine]{Log[msq^2/mQ3^2],0,3},\[IndentingNewLine]{PolyLog[2,1-m3^2/mQ3^2],0,2},\[IndentingNewLine]{PolyLog[2,1-m3^2/mU3^2],0,2},\[IndentingNewLine]{PolyLog[2,1-mU3^2/mQ3^2],0,2},\[IndentingNewLine]{PolyLog[2,1-msq^2/m3^2],0,2},\[IndentingNewLine]{PolyLog[2,1-msq^2/mQ3^2],0,2},\[IndentingNewLine]{PolyLog[2,1-msq^2/mU3^2],0,2},\[IndentingNewLine]{Xt,0,8},

and the coefficient of these variables is a rational function of the Soft-SUSY parameters.

'DRratios', which contains the mass expansion expressed in terms of vu,yt,g3 and in powers of Log[MR^2/mQ3^2], just like 'DR' but the masses have been replaced by dimensionless parameters:

xEWS=mt/mQ3, stands as the ratio between the EW and the SUSY scales. At the order considered in this program, O((v^2/MS^2)^0), only appears as the argument of Log functions,
x12=mU3/mQ3,
x32=m3/mQ3,
xsq2=msq/mQ3,
xt=Xt/mQ3.

This expansion is particularly suitable for obtaing limiting cases by performing series expansions in {-1+x32^2), (-1+x12^2), and (-1+xsq2^2).

Soft-SUSY parameters:

m3, mQ3, mU3, msq, Xt and beta.

";

dMh2dMR::usage="

dMh2dMR[n] calculates the total derivative of the Higgs boson squared pole mass expansion with respect to MR^2.

It returns

(vu^2 yt^4 Sin[beta]^2)^-1 MR^2 (d/d(MR^2))Mx[h,NonDegenerate][n,1].

";

deltamh2::usage="

deltamh2[Alpha3Order,LogOrder] returns the loop correction of the Higgs boson DR mass at the desired order of alpha_t and alpha_3, and power of Log[MR^2/mt^2].
alphatOrder is set to 1 by defualt and can be ommited if no higher powers are desired to be computed.

This function extracts the coefficients from the already pre-loaded complete 3-loop order Higgs mass expansion stored in the symbol 'Mh2[EFT]'.

";

LogCoefficient::usage="

LogCoefficient is a function that can receive two types of input, namely, [Alpha3Order,LogOrder,m3LogOrder,mQ3LogOrder,mU3LogOrder,msqLogOrder] and [Alpha3Order,LogOrder,SUSYMass,SUSYMassLogOrder].
This functions is mainly intended to study the terms of O(Log[MR^2/mt^2]^0), which contain a big structure in SUSY logarithms.

Using the first type of input, it returns the term of the explicitly specified powers of the different types of logarithmic terms. This type of input is ableto return the purely non-logarithmic terms at each order by setting every power of the SUSY dependent logarithm to zero.
With the first input, however, it is not able to return e.g. the complete 3-loop order term of O(Log[MR^2/mt^2] Log[m3^2/MR^2]) at a time, since it includes various types of logarithms.
Using the second type of input, it returns the complete logarithmic term of the desired SUSY mass at the desired power. The possible SUSY masses are m3, mQ3, mU3, msq.

This function extracts the coefficients from the already pre-loaded complete 3-loop order Higgs mass expansion stored in the symbol 'Mh2[EFT]'.

";

K::usage="

K{Subindex}[n,p] returns the SM MS coefficient of Log[MR^2 MSyt^-2 MSv^-2] at all possible O(alpha_t^n alpha_3^m) at each loop order.

";

BetaFunction::usage="

BetaFunction[g,n] returns the beta function of the SM MS parameters.

The first argument gives the parameter in the following order: {MSg3,MSyt,MSlambda,MSv}, and the second argument the loop order, where order 0 is the parameter itself.

";

MSSMBetaFunction::usage="

MSSMBetaFunction[g,n] returns the beta function of the SM MS parameters.

The first argument gives the parameter in the following order: {g3(MR),yt(MR),m3(MR),mU3(MR),mQ3(MR),Xt(MR)}, and the second argument the loop order, where order 0 is the parameter itself.

";

ListOfParameters::usage="

Returns the list of relevant parameters in the mass expansions.

";



(*
Begin["Private`"];
*)


(* ::Subchapter::Closed:: *)
(*Timer start*)


IniPoleMass=DateList[];



(* ::Subchapter::Closed:: *)
(*Information & list of relevant functions and parameters*)


Info[]:=Print["

Use ListOfFunctions[] to see the list of available functions.

Use ListOfParameters[] to obtain a description on the parameters that appear in the results.

Use M[Subindex,LimitingCase][LoopOrder,AlphaTOrder] to obtain the Higgs boson squared mass expansion in the desired limit and order.

Use Mx[Subindex,LimitingCase,Scheme][LoopOrder,AlphaTOrder] to obtain the Higgs boson squared mass expansion in the desired limit and order.

Use deltamh2[LimitingCase,Alpha3Order,LogOrder,AlphaTOrder] to get the coefficient of the desired power of Log[MR^2/mt^2].

Use ? for each function for more information.

"];



ListOfFunctions[]:=Print["

ListOfParameters

M
Mx
Mh2

dMh2dMR

deltamh2
BetaFunction
MSSMBetaFunction

"];



ListOfParameters[]:=Print["

List of relevant parameters

k      : Loop factor 1/(4 Pi)^2

Energy scales

MR      : Renormalization scale
MS     : SUSY scale
MZ     : Z boson pole mass

SM MS parameters

MSg3      : Strong gauge coupling
MSyt      : Top Yukawa coupling
MSlambda  : Higgs field self-interaction coupling
MSv       : Higgs field vacuum expectation value
BetaFunction[g,n]     : SM parameter beta-function
SMConstat(n)as(m)     : SM loop constant of order at^n a3^m

MSmt = MSyt MSv         : Top quark mass, 

MSa3 = k^(1/2) MSg3^2   : Strong gauge coupling 'fine structure constant'
MSat = k^(1/2) MSyt^2   : Top Yukawa coupling 'fine structure constant'
MSSMBetaFunction[g,n]   : SM parameter beta-function

MSSM DR parameters

g3        : Strong gauge coupling
yt        : Top Yukawa coupling
vu        : Up Higgs field vacuum expectation value
TanBeta   : Ratio of the vacuum expectation values of the Higgs fields
beta      : arcTan of the ratio of the vacuum expectation values of the Higgs fields

mt = yt vu                      : Top quark mass, 

a3 = k^(1/2) g3^2               : Strong gauge coupling 'fine structure constant'
at = k^(1/2) yt^2 Sin[beta]^2   : Top Yukawa coupling 'fine structure constant'

m3     : Soft-SUSY breaking gluino mass parameter
mu     : Soft-SUSY breaking Higgsino mass parameter
mA     : Soft-SUSY breaking CP odd Higgs mass parameter
mQ3    : Soft-SUSY breaking doublet stop mass parameter
mU3    : Soft-SUSY breaking singlet stop mass parameter
msq    : Soft-SUSY breaking squark mass parameter
Xt = At - mu/Tan[beta]    : Soft-SUSY breaking stop mixing
Yt = At + mu/Tan[beta]    : Soft-SUSY breaking 'conjugate' stop mixing

MSSM MS bar threshold corrections

dlat(n)as(m)        : Higgs field self-interaction coupling threshold correction of order at^n a3^m
dytat(n)as(m)       : Top Yukawa coupling threshlold correction of order at^n a3^m
dg3at(n)as(n)       : Strong gauge coupling threshold correction of order at^n a3^m

MSSM DR bar threshold corrections

Dlat(n)as(m)        : Higgs field self-interaction coupling threshold correction of order at^n a3^m
Dytat(n)as(m)       : Top Yukawa coupling threshlold correction of order at^n a3^m
Dg3at(n)as(n)       : Strong gauge coupling threshold correction of order at^n a3^m

Misc

z2                  : Zeta[2]
z3                  : Zeta[3]
Kdlat2 = -0.1953256 : Constant that appears in dlat2

"];



(* ::Chapter::RGBColor[0., 0., 1.]::Closed:: *)
(*RG*)


(* ::Subchapter::Closed:: *)
(*List of beta-functions of the SM MS parameters*)


(*
3-loop order beta-functions taken from [Mihaila et al., 2012, arXiv:1201.5868], [Mihaila et al., 2012, arXiv:1208.3357], [Bednyakov et al., 2013, arXiv:1212.6829], and [Bednyakov et al., 2013, arXiv:1303.4364].
4-loop order lambda beta-function taken from [Martin, 2015, arXiv:1508.00912v1].
*)

BetaFunctions={

{
MSg3,
k(-7 MSg3^3),
k^2(-26 MSg3^5-2 MSg3^3 MSyt^2),
k^3((65 MSg3^7)/2-40 MSg3^5 MSyt^2+15 MSg3^3 MSyt^4),
k^4 betag34Loop (MSg3^9+MSg3^7 MSyt^2+MSg3^5 MSyt^4+MSg3^3 MSyt^6)
},

{
MSyt,
k(-8 MSg3^2 MSyt+(9 MSyt^3)/2),
k^2(-12 MSyt^5+MSyt^3 (36 MSg3^2-3 MSlambda)+MSyt (-108 MSg3^4+(3 MSlambda^2)/8)),
k^3((MSyt^5 (-3255552 MSg3^2+1026432 MSlambda))/20736+(MSyt^7 (878688+279936 Zeta[3]))/20736+(MSyt^3 (13226112 MSg3^4+82944 MSg3^2 MSlambda+4860 MSlambda^2-4727808 MSg3^4 Zeta[3]))/20736+(MSyt (-28795392 MSg3^6-11664 MSlambda^3+13271040 MSg3^6 Zeta[3]))/20736),
k^4 betayt4Loop (MSg3^8 MSyt +MSg3^6 MSyt^3 +MSg3^4 MSyt^5 +MSg3^2 MSyt^7)
},

{
MSlambda,
k(-24 MSyt^4+12 MSyt^2 MSlambda+6 MSlambda^2),
k^2(1/24 (-3072 MSg3^2 MSyt^4+2880 MSyt^6)+1/24 (1920 MSg3^2 MSyt^2-72 MSyt^4) MSlambda-36 MSyt^2 MSlambda^2-(39 MSlambda^3)/2),
k^3(873/8 MSyt^2 MSlambda^3+3/8 MSlambda^4 (299+168 Zeta[3])+1/16 MSlambda^2 (-9792 MSg3^2 MSyt^2+6876 MSyt^4+9216 MSg3^2 MSyt^2 Zeta[3]+6048 MSyt^4 Zeta[3])+8 (-38 MSg3^2 MSyt^6-(1599 MSyt^8)/8+240 MSg3^2 MSyt^6 Zeta[3]-36 MSyt^8 Zeta[3]+(MSyt^4 (-306432 MSg3^4+110592 MSg3^4 Zeta[3]))/3456)+8 MSlambda ((117 MSyt^6)/32-99/2 MSyt^6 Zeta[3]+1/4 (895 MSg3^2 MSyt^4-1296 MSg3^2 MSyt^4 Zeta[3]+(MSyt^2 (716544 MSg3^4-82944 MSg3^4 Zeta[3]))/1728))),
k^4 ((8/27 MSg3^6 MSyt^4 (13097+3888 Inactive[Zeta][2]^2+44568 Zeta[3]-47400 Zeta[5]))+betalabda4Loop(MSg3^4 MSyt^6 + MSg3^2 MSyt^8 + MSyt^10))
},

{
MSv,
k(-3 MSv MSyt^2),
k^2(-(1/32) MSv (640 MSg3^2 MSyt^2+12 (-18 MSyt^4+MSlambda^2))),
k^3 MSv betav3Loop (MSg3^4 MSyt^2+MSg3^2 MSyt^4),
0
}

};



(* ::Subchapter::Closed:: *)
(*List of beta-functions of the MSSM DR parameters*)


(* Provided by A. Voigt *)

MSSMBetaFunctions={

{
g3,
k(-3 g3^3),
k^2(14 g3^5-4 g3^3 yt^2),
k^3((347 g3^7)/3-(104 g3^5 yt^2)/3+30 g3^3 yt^4)
},

{
yt,
k(-((16 g3^2 yt)/3)+6 yt^3),
k^2(-((16 g3^4 yt)/9)+16 g3^2 yt^3-22 yt^5),
k^3((5440 g3^6 yt)/27-(296 g3^4 yt^3)/3+(272 g3^2 yt^5)/3+102 yt^7+640 g3^6 yt Zeta[3]-288 g3^4 yt^3 Zeta[3]+36 yt^7 Zeta[3])
},

{
m3,
k(-6 g3^2 m3),
k^2(56 g3^4 m3+8 At g3^2 yt^2-8 g3^2 m3 yt^2),
k^3(694 g3^6 m3+208/3 At g3^4 yt^2-416/3 g3^4 m3 yt^2-120 At g3^2 yt^4+60 g3^2 m3 yt^4)
},

{
mU3^2,
k(-(32/3) g3^2 m3^2+2 At^2 yt^2+2 mh^2 yt^2+2 mQ3^2 yt^2+2 mU3^2 yt^2),
k^2(-(128/3) g3^4 m3^2+(32 g3^4 mQ3^2)/3+48 g3^4 msq^2+(16 g3^4 mU3^2)/3-40 At^2 yt^4-20 mh^2 yt^4-20 mQ3^2 yt^4-20 mU3^2 yt^4)
},

{
mQ3^2,
k(-(32/3) g3^2 m3^2+4 At^2 yt^2+4 mh^2 yt^2+4 mQ3^2 yt^2+4 mU3^2 yt^2),
k^2(-(128/3) g3^4 m3^2+(32 g3^4 mQ3^2)/3+48 g3^4 msq^2+(16 g3^4 mU3^2)/3-64 At^2 yt^4-32 mh^2 yt^4-32 mQ3^2 yt^4-32 mU3^2 yt^4)
},

{
Xt,
32/3 g3^2 k m3,
64/9 g3^4 k^2 m3,
-(10880/9) g3^6 k^3 m3-3840 g3^6 k^3 m3 Zeta[3]
}

};



(* ::Subchapter::Closed:: *)
(*Function of the beta-functions*)


(*

The convention used here for the beta-functions is defined as follows:

beta_g_i({g_j}) = k beta_g_i^(1)({g_j}) + k^2 beta_g_i^(2)({g_j}) + k^3 beta_g_i^(3)({g_j}) + ... , where k=1/(4 Pi)^2

*)

(* 
The first argument gives the parameter in the following order: {MSg3(MR),MSyt(MR),MSlambda(MR),MSv(MR)}, and the second argument the loop order, where order 0 is the parameter itself 
*)

BetaFunction[g_,n_]:=BetaFunctions[[g]][[n+1]]/(k^n);

(* 
The first argument gives the parameter in the following order: {g3(MR),yt(MR),m3(MR),mU3(MR),mQ3(MR),Xt(MR)}, and the second argument the loop order, where order 0 is the parameter itself 
*)

MSSMBetaFunction[g_,n_]:=MSSMBetaFunctions[[g]][[n+1]]/(k^n);



(* ::Chapter::RGBColor[0., 0., 1.]::Closed:: *)
(*Pole Mass in the SM*)


(* ::Subchapter::Closed:: *)
(*C[n,p] ,p>0 : Coefficients of the powers of logs*)


(*
Recursion relation of the logs' coefficients,
extended result from [Martin, 2007, arXiv:hep-ph/0701051v1] Eqs. (3.3) - (3.4).
*)

K[Subindex_][n_Integer,p_Integer]/;p<=0:=Print["Error: 0<p\[LessEqual]n"];

K[Subindex_][n_Integer,p_Integer]/;p>n:=Print["Error: 0<p\[LessEqual]n"];

K[Subindex_][n_Integer,p_Integer]/;p==n:=Normal@Series[

-(1/(2 n)) 
Sum[
BetaFunction[G,1] D[K[Subindex][n-1,n-1],BetaFunction[G,0]],
{G,1,4}
],

{MSv,0,2},
{MSlambda,0,4},
{MSyt,0,12},
{MSg3,0,10}

];

K[Subindex_][n_Integer,p_Integer]/;p>0:=Normal@Series[

(Sum[
(BetaFunction[4,i]/MSv+BetaFunction[2,i]/MSyt) K[Subindex][n-i,p],
{i,1,n-p}
]
-(1/(2 p)) 
Sum[
Sum[
BetaFunction[G,j] D[K[Subindex][n-j,p-1],BetaFunction[G,0]],
{j,1,n-p+1}
],
{G,1,4}
]),

{MSv,0,2},
{MSlambda,0,4},
{MSyt,0,12},
{MSg3,0,10}

];



(* ::Subchapter::Closed:: *)
(*C[n,0] : Loop constants*)


(*
C[0,0],C[1,0], C[2,0], C[3,0] coefficients extracted from the 1-,2-, and 3-loop SM square pole mass of the Higss boson in the MS scheme
*)

K[h][0,0]:=MSlambda MSv^2;
K[h][1,0]:=(2MSlambda MSyt^2 + SMConstlambda2 MSlambda^2)MSv^2;
K[h][2,0]:=SMConstat2 MSyt^6 MSv^2;
K[h][3,0]:=SMConstatas2 MSg3^4 MSyt^4 MSv^2+SMConstat2as MSg3^2 MSyt^6 MSv^2;
K[h][4,0]:=SMConstatas3 MSg3^6 MSyt^4 MSv^2;

(* Taken from [Martin, 2007,arXiv:hep-ph/0701051] Eq. (3.16) *)

SM1LoopConst=SMConstlambda2->( -(1/4)(9Bh+2BW+BZ+6/5));

(* Taken from [Degrassi et al., 2013, arXiv:1205.6497v2[hep-ph]] Eq. (20) *)

SM2LoopConst=SMConstat2->24(-1+Inactive[Zeta][2]);

(* Taken from [Martin, Robertson, 2014, arXiv:1407.4336v2[hep-ph]] Eq. (3.3) *)

SM3LoopConst={
SMConstatas2->(-(3776/9)-2048/3 Inactive[PolyLog][4,1/2]+320 Zeta[3]+(2816 Inactive[Zeta][2]^2)/15+256/9 Log[2]^2 (-Log[2]^2+6 Inactive[Zeta][2]))
(*
,
SMConstat2as->2764.365
*)
};



(*
C[0,0],C[1,0], C[2,0], C[3,0] coefficients extracted from the 1-,2-, and 3-loop SM square pole mass of the Higss boson in the MS scheme
*)

K[t][0,0]:=MSyt MSv;
K[t][1,0]:=(16/3)MSg3^2 MSyt MSv+SMTopConstat MSyt^2 MSyt MSv;
K[t][2,0]:=SMTopConstas2 MSg3^4 MSyt MSv+SMTopConstatas MSyt^2 MSg3^2 MSyt MSv+SMTopConstat2 MSyt^4 MSyt MSv;
K[t][3,0]:=SMTopConstas3 MSg3^6 MSyt MSv+SMTopConstatas2 MSyt^2 MSg3^4 MSyt MSv+SMTopConstat2as MSyt^4 MSg3^2 MSyt MSv+SMTopConstat3 MSyt^6 MSyt MSv;
K[t][4,0]:=0;

(* Taken from [Fleischer et al.,1999,arXiv:hep-ph/9803493v5] *)

SMTop2LoopConst=SMTopConstas2->(4/3)^2 (-(71/8)+30 Inactive[Zeta][2]+12 Zeta[3]-48 Inactive[Zeta][2] Log[2])+(4/3) (-(83/2)-12 Inactive[Zeta][2]+3 (1111/24-8 Inactive[Zeta][2]-6 Zeta[3]+24 Inactive[Zeta][2] Log[2]));



(* ::Subchapter::Closed:: *)
(*Mass expansion*)


(*
Expansion of the scale (in)-dependent Higgs boson mass in powers of Log[MR^2/MSmt^2] using SM MS parameters
*)

Mx[Subindex_,GeneralSolution,SM][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

Sum[
k^n Sum[
K[Subindex][n,i] Log[MR^2 MSyt^-2 MSv^-2]^i,
{i,0,n}
],
{n,0,LoopOrder}
],

{MSv,0,2},
{k,0,LoopOrder},
{MSlambda,0,2},
{MSyt,0,2(AlphaTOrder+1)},
{MSg3,0,2(LoopOrder-AlphaTOrder)},
{Log[MR^2 MSyt^-2 MSv^-2],0,LoopOrder},
{SMTopConstas2,0,4}

];



(* ::Chapter::RGBColor[0., 0., 1.]::Closed:: *)
(*Matching SM MS param. & MSSM DR param.*)


(* ::Subchapter::Closed:: *)
(*Threshold corrections*)


(* ::Section::RGBColor[1, 0, 0]::Closed:: *)
(*1-loop order corrections*)


(* Non-degenerate *)

(* Taken from [Bagnaschi et al., 2014, arXiv:1407.4081v2], Eqs. (2.10), (2.21), and (2.24) *)

deltalambdaat=(
(12 Xt^4)/(mQ3^2-mU3^2)^2+(6+(12 Xt^2)/(mQ3^2-mU3^2)+(-((6 mQ3^2)/(mQ3^2-mU3^2)^3)-(6 mU3^2)/(mQ3^2-mU3^2)^3) Xt^4) Log[mQ3^2/MR^2]+(6-(12 Xt^2)/(mQ3^2-mU3^2)+((6 mQ3^2)/(mQ3^2-mU3^2)^3+(6 mU3^2)/(mQ3^2-mU3^2)^3) Xt^4) Log[mU3^2/MR^2]
);

deltaytas=(
(2 m3^4)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))-(2 mQ3^2 mU3^2)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))+(-((2 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(8 m3^3 Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))) Log[m3^2/MR^2]+((4 m3^2 mQ3^2)/(3 (m3^2-mQ3^2)^2)-(2 mQ3^4)/(3 (m3^2-mQ3^2)^2)-(8 m3 mQ3^2 Xt)/(3 (m3^2-mQ3^2) (mQ3^2-mU3^2))) Log[mQ3^2/MR^2]+((4 m3^2 mU3^2)/(3 (m3^2-mU3^2)^2)-(2 mU3^4)/(3 (m3^2-mU3^2)^2)-(8 m3 mU3^2 Xt)/(3 (m3^2-mU3^2) (-mQ3^2+mU3^2))) Log[mU3^2/MR^2]
);

deltaytat=(
CscBeta^2 ((3 CosBeta^2)/8+(5 mQ3^2 mu^2-9 mu^4-3 mQ3^2 mU3^2+7 mu^2 mU3^2)/(8 (mQ3^2-mu^2) (mu^2-mU3^2))+(3 (mQ3^2+mU3^2) SinBeta^2 Xt^2)/(4 (mQ3^2-mU3^2)^2)-3/4 CosBeta^2 Log[mA^2/MR^2]+(-(mQ3^4/(2 (mQ3^2-mu^2)^2))+(mQ3^2 mu^2)/(mQ3^2-mu^2)^2-(3 mQ3^2 mU3^2 SinBeta^2 Xt^2)/(2 (mQ3^2-mU3^2)^3)) Log[mQ3^2/MR^2]-(mu^4 (mQ3^4-2 mQ3^2 mu^2+3 mu^4-4 mu^2 mU3^2+2 mU3^4) Log[mu^2/MR^2])/(4 (-mQ3^2+mu^2)^2 (mu^2-mU3^2)^2)+((mu^2 mU3^2)/(2 (mu^2-mU3^2)^2)-mU3^4/(4 (mu^2-mU3^2)^2)+(3 mQ3^2 mU3^2 SinBeta^2 Xt^2)/(2 (mQ3^2-mU3^2)^3)) Log[mU3^2/MR^2])
);

deltag3as=(
1/2-Log[m3^2/MR^2]-1/12 (Log[mQ3^2/MR^2]+ Log[mU3^2/MR^2]+ 10 Log[msq^2/MR^2])
);

(* Taken from [Martin, 2007, arXiv:hep-ph/0701051v1] *)

deltavevat=(
(3 (mQ3^2+mU3^2) Xt^2)/(4 (mQ3^2-mU3^2)^2)-(3 mQ3^2 mU3^2 Xt^2 Log[mQ3^2/MR^2])/(2 (mQ3^2-mU3^2)^3)+(3 mQ3^2 mU3^2 Xt^2 Log[mU3^2/MR^2])/(2 (mQ3^2-mU3^2)^3)
);



(* Degenerate *)

deltalambdaatDeg=(
(12 Xt^2)/MS^2-Xt^4/MS^4+12 Log[MS^2/MR^2]
);

deltaytasDeg=(
-(4/3)+(4 Xt)/(3 MS)-4/3 Log[MS^2/MR^2]
);

deltaytatDeg=(
CscBeta^2 ((SinBeta^2 Xt^2)/(4 MS^2)+CosBeta^2 (3/8-3/4 Log[MS^2/MR^2])-3/4 Log[MS^2/MR^2])
);

deltag3asDeg=(
1/2-2 Log[MS^2/MR^2]
);

(* Taken from [Martin, 2007, arXiv:hep-ph/0701051v1] *)

deltavevatDeg=(
Xt^2/(4 MS^2)
);



(* ::Section::RGBColor[1, 0, 0]::Closed:: *)
(*2-loop order corrections*)


(* Non-degenerate *)

(* Taken from [Bagnaschi et al., 2014, arXiv:1407.4081v2], Eq. (2.36) and complete expression provided by A. Voigt *)

deltalambdaatas=(
(16 (-2 m3^6 mQ3^2+2 m3^4 mQ3^4-2 m3^6 mU3^2+9 m3^4 mQ3^2 mU3^2-6 m3^2 mQ3^4 mU3^2+2 m3^4 mU3^4-6 m3^2 mQ3^2 mU3^4+3 mQ3^4 mU3^4))/(mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2))-(64 m3^2 Xt^2)/(mQ3^2 mU3^2)-(512 m3 Xt^3)/(mQ3^2-mU3^2)^2+(32 (-m3^6 mQ3^2+m3^4 mQ3^4-m3^6 mU3^2-5 m3^4 mQ3^2 mU3^2+5 m3^2 mQ3^4 mU3^2+m3^4 mU3^4+5 m3^2 mQ3^2 mU3^4-5 mQ3^4 mU3^4) Xt^4)/(mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+(-((16 (3 m3^4-4 m3^2 mQ3^2+2 mQ3^4))/(m3^2-mQ3^2)^2)+(64 (2 m3^3-m3 mQ3^2) Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))-(32 (4 mQ3^6-2 mQ3^4 mU3^2+m3^4 (3 mQ3^2-mU3^2)+m3^2 (-8 mQ3^4+4 mQ3^2 mU3^2)) Xt^2)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)+(64 m3 (2 m3^4-mQ3^4+3 mQ3^2 mU3^2-m3^2 (3 mQ3^2+mU3^2)) Xt^3)/((m3^2-mQ3^2) (mQ3^2-mU3^2)^3)+(16 (5 mQ3^8+8 m3^4 mQ3^2 mU3^2+4 mQ3^6 mU3^2-mQ3^4 mU3^4+2 m3^6 (mQ3^2-mU3^2)-2 m3^2 (4 mQ3^6+5 mQ3^4 mU3^2-mQ3^2 mU3^4)) Xt^4)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)+(64 m3 mQ3^2 (mQ3^2+mU3^2) Xt^5)/((m3^2-mQ3^2) (mQ3^2-mU3^2)^4)) Log[mQ3^2/MR^2]^2+((16 (7 m3^6-2 mQ3^2 mU3^4-m3^4 (6 mQ3^2+7 mU3^2)+m3^2 (5 mQ3^2 mU3^2+3 mU3^4)))/((m3^2-mQ3^2) (m3^2-mU3^2)^2)+(128 m3 Xt)/(mQ3^2-mU3^2)-(32 (7 m3^4+3 mQ3^2 mU3^2-2 m3^2 (3 mQ3^2+2 mU3^2)) Xt^2)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))-(128 m3 (mQ3^2+3 mU3^2) Xt^3)/(mQ3^2-mU3^2)^3+(16 (4 m3^8-3 mQ3^2 mU3^4 (mQ3^2+5 mU3^2)+m3^6 (3 mQ3^2+7 mU3^2)-m3^4 (6 mQ3^4+15 mQ3^2 mU3^2+29 mU3^4)+m3^2 (7 mQ3^4 mU3^2+31 mQ3^2 mU3^4+16 mU3^6)) Xt^4)/((m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(128 m3 mU3^2 Xt^5)/((m3^2-mU3^2) (-mQ3^2+mU3^2)^3)) Log[mU3^2/MR^2]+(-((16 (3 m3^4-4 m3^2 mU3^2+2 mU3^4))/(m3^2-mU3^2)^2)+(64 (2 m3^3-m3 mU3^2) Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(32 (m3^4 (mQ3^2-3 mU3^2)+2 mU3^4 (mQ3^2-2 mU3^2)+m3^2 (-4 mQ3^2 mU3^2+8 mU3^4)) Xt^2)/((m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(64 m3 (2 m3^4+3 mQ3^2 mU3^2-mU3^4-m3^2 (mQ3^2+3 mU3^2)) Xt^3)/((m3^2-mU3^2) (-mQ3^2+mU3^2)^3)+(16 (8 m3^4 mQ3^2 mU3^2-mQ3^4 mU3^4+4 mQ3^2 mU3^6+5 mU3^8-2 m3^6 (mQ3^2-mU3^2)+2 m3^2 mU3^2 (mQ3^4-5 mQ3^2 mU3^2-4 mU3^4)) Xt^4)/((m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)+(64 m3 mU3^2 (mQ3^2+mU3^2) Xt^5)/((m3^2-mU3^2) (mQ3^2-mU3^2)^4)) Log[mU3^2/MR^2]^2+Log[m3^2/MR^2] (-((16 m3^4 (2 m3^6 (mQ3^2+mU3^2)+2 m3^2 (mQ3^2-mU3^2)^2 (mQ3^2+mU3^2)+m3^4 (-4 mQ3^4+2 mQ3^2 mU3^2-4 mU3^4)+mQ3^2 mU3^2 (mQ3^4+mU3^4)))/(mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2))+(64 m3^4 (-m3^2+mQ3^2+mU3^2) Xt^2)/(mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2))-(32 m3^2 (m3^8 (mQ3^2+mU3^2)+mQ3^4 mU3^4 (mQ3^2+mU3^2)+m3^4 (mQ3^2+mU3^2)^3-2 m3^6 (mQ3^4+mQ3^2 mU3^2+mU3^4)-m3^2 mQ3^2 mU3^2 (mQ3^4+4 mQ3^2 mU3^2+mU3^4)) Xt^4)/(mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(128 m3^3 Xt^5)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+((16 m3^4 (2 m3^2-mQ3^2-mU3^2) (mQ3^2-mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(64 m3^3 (-2 m3^2+mQ3^2+mU3^2) Xt)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))-(32 m3^4 (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2)) Xt^2)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(128 m3^3 (2 m3^4-mQ3^4+4 mQ3^2 mU3^2-mU3^4-2 m3^2 (mQ3^2+mU3^2)) Xt^3)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+(16 m3^4 (mQ3^2+mU3^2) (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2)) Xt^4)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(64 m3^3 (mQ3^2+mU3^2) Xt^5)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) Log[mQ3^2/MR^2]+(-((16 m3^4 (2 m3^2-mQ3^2-mU3^2) (mQ3^2-mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(64 m3^3 (2 m3^2-mQ3^2-mU3^2) Xt)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))+(32 m3^4 (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2)) Xt^2)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(128 m3^3 (2 m3^4-mQ3^4+4 mQ3^2 mU3^2-mU3^4-2 m3^2 (mQ3^2+mU3^2)) Xt^3)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(16 m3^4 (mQ3^2+mU3^2) (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2)) Xt^4)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(64 m3^3 (mQ3^2+mU3^2) Xt^5)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) Log[mU3^2/MR^2])+Log[mQ3^2/MR^2] ((16 (7 m3^6-2 mQ3^4 mU3^2-m3^4 (7 mQ3^2+6 mU3^2)+m3^2 (3 mQ3^4+5 mQ3^2 mU3^2)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2))-(128 m3 Xt)/(mQ3^2-mU3^2)+(32 (7 m3^4+3 mQ3^2 mU3^2-2 m3^2 (2 mQ3^2+3 mU3^2)) Xt^2)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))+(128 m3 (3 mQ3^2+mU3^2) Xt^3)/(mQ3^2-mU3^2)^3-(16 (4 m3^8-3 mQ3^4 mU3^2 (5 mQ3^2+mU3^2)+m3^6 (7 mQ3^2+3 mU3^2)-m3^4 (29 mQ3^4+15 mQ3^2 mU3^2+6 mU3^4)+m3^2 (16 mQ3^6+31 mQ3^4 mU3^2+7 mQ3^2 mU3^4)) Xt^4)/((m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(128 m3 mQ3^2 Xt^5)/((m3^2-mQ3^2) (mQ3^2-mU3^2)^3)+((16 (-2 mQ3^4 mU3^4+2 m3^6 (mQ3^2+mU3^2)+4 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)-m3^4 (mQ3^4+8 mQ3^2 mU3^2+mU3^4)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(64 m3^3 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))+(32 (2 m3^8 (mQ3^2+mU3^2)+2 mQ3^4 mU3^4 (mQ3^2+mU3^2)-4 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)^2+3 m3^4 (mQ3^2+mU3^2)^3-2 m3^6 (3 mQ3^4+2 mQ3^2 mU3^2+3 mU3^4)) Xt^2)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(128 (-2 m3 mQ3^2 mU3^2+m3^3 (mQ3^2+mU3^2)) Xt^3)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)-(16 (mQ3^2+mU3^2) (4 m3^8 (mQ3^2+mU3^2)+4 mQ3^4 mU3^4 (mQ3^2+mU3^2)-8 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)^2-2 m3^6 (5 mQ3^4+6 mQ3^2 mU3^2+5 mU3^4)+m3^4 (5 mQ3^6+19 mQ3^4 mU3^2+19 mQ3^2 mU3^4+5 mU3^6)) Xt^4)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)-(64 m3 (mQ3^2+mU3^2) (-2 mQ3^2 mU3^2+m3^2 (mQ3^2+mU3^2)) Xt^5)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^4)) Log[mU3^2/MR^2])+(-((64 m3^4)/(-m3^2+mQ3^2)^2)+(256 m3^3 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))+(128 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(mQ3^2-mU3^2)^3-(32 (-2 m3^2+mQ3^2+mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) PolyLog[2,1-m3^2/mQ3^2]+(-((64 m3^4)/(m3^2-mU3^2)^2)+(256 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(128 m3 (-2 m3^2+mQ3^2+mU3^2) Xt^3)/(mQ3^2-mU3^2)^3+(32 (-2 m3^2+mQ3^2+mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) PolyLog[2,1-m3^2/mU3^2]
);

(* Taken from [Pardo Vega and Villadoro, 2015, arXiv:1504.05200v1], Eq. (38) and complete expression provided by A. Voigt *)

(* Complete 2-loop order dlat2 correction *)

deltalambdaat2=(1/2)*Ndlat2*(
30-(6 (2 mQ3^4+3 mQ3^2 mU3^2-mU3^4) Xt^2)/(mQ3^2 mU3^2 (mQ3^2-mU3^2))+(12 (2 mQ3^4-27 mQ3^2 mU3^2+mU3^4) Xt^4)/(mQ3^2 mU3^2 (mQ3^2-mU3^2)^2)+(6 (-2 mQ3^4-11 mQ3^2 mU3^2+mU3^4) Xt^6)/(mQ3^2 mU3^2 (mQ3^2-mU3^2)^3)+(18+(6 (7 mQ3^2-9 mU3^2) Xt^2)/(mQ3^2-mU3^2)^2-(6 (7 mQ3^2+9 mU3^2) Xt^4)/(mQ3^2-mU3^2)^3-(6 (11 mQ3^2+3 mU3^2) Xt^6)/(mQ3^2-mU3^2)^4) Log[mQ3^2/MR^2]^2+(-24+(6 (16 mQ3^4-21 mQ3^2 mU3^2+mU3^4) Xt^2)/(mQ3^2 (mQ3^2-mU3^2)^2)-(12 (3 mQ3^4+22 mQ3^2 mU3^2-mU3^4) Xt^4)/(mQ3^2 (mQ3^2-mU3^2)^3)-(6 (10 mQ3^4+3 mQ3^2 mU3^2-mU3^4) Xt^6)/(mQ3^2 (mQ3^2-mU3^2)^4)) Log[mU3^2/MR^2]+(12-(54 Xt^2)/(mQ3^2-mU3^2)+((18 mQ3^2)/(mQ3^2-mU3^2)^3+(30 mU3^2)/(mQ3^2-mU3^2)^3) Xt^4+(-((30 mQ3^2)/(mQ3^2-mU3^2)^4)-(18 mU3^2)/(mQ3^2-mU3^2)^4) Xt^6) Log[mU3^2/MR^2]^2+Log[mQ3^2/MR^2] (-36+(6 (2 mQ3^4-17 mQ3^2 mU3^2+19 mU3^4) Xt^2)/(mU3^2 (-mQ3^2+mU3^2)^2)-(24 (-mQ3^4+10 mQ3^2 mU3^2+3 mU3^4) Xt^4)/(mU3^2 (-mQ3^2+mU3^2)^3)-(6 (-2 mQ3^4-13 mQ3^2 mU3^2+3 mU3^4) Xt^6)/(mU3^2 (mQ3^2-mU3^2)^4)+((12 mQ3^2 Xt^2)/(mQ3^2-mU3^2)^2+(24 (mQ3^2+mU3^2) Xt^4)/(mQ3^2-mU3^2)^3+(12 (8 mQ3^2+3 mU3^2) Xt^6)/(mQ3^2-mU3^2)^4) Log[mU3^2/MR^2])+(-((36 Xt^2)/(mQ3^2-mU3^2))-(36 Xt^4)/(mQ3^2-mU3^2)^2-(36 Xt^6)/(mQ3^2-mU3^2)^3) PolyLog[2,1-mQ3^2/mU3^2]+CscBeta^2 (-((3 (4 mQ3^6 mu^2-12 mQ3^4 mu^4+8 mQ3^2 mu^6-4 mQ3^6 mU3^2+17 mQ3^4 mu^2 mU3^2-21 mQ3^2 mu^4 mU3^2+4 mu^6 mU3^2-3 mQ3^4 mU3^4+13 mQ3^2 mu^2 mU3^4-6 mu^4 mU3^4-2 mQ3^2 mU3^6+2 mu^2 mU3^6))/(mQ3^2 (mQ3^2-mu^2) mU3^2 (mu^2-mU3^2)))+(12 (2 mQ3^4-4 mQ3^2 mu^2+3 mQ3^2 mU3^2+2 mu^2 mU3^2-mU3^4) Xt^2)/(mQ3^2 mU3^2 (mQ3^2-mU3^2))-(6 (2 mQ3^8 mu^2-6 mQ3^6 mu^4+4 mQ3^4 mu^6-2 mQ3^8 mU3^2+52 mQ3^6 mu^2 mU3^2-62 mQ3^4 mu^4 mU3^2+16 mQ3^2 mu^6 mU3^2-48 mQ3^6 mU3^4+31 mQ3^4 mu^2 mU3^4+11 mQ3^2 mu^4 mU3^4-2 mu^6 mU3^4+31 mQ3^4 mU3^6-30 mQ3^2 mu^2 mU3^6+3 mu^4 mU3^6+mQ3^2 mU3^8-mu^2 mU3^8) Xt^4)/(mQ3^2 (mQ3^2-mu^2) mU3^2 (mQ3^2-mU3^2)^3 (mu^2-mU3^2))+(-((12 (mQ3^4-2 mQ3^2 mu^2))/(mQ3^2-mu^2)^2)+(12 (7 mQ3^6-14 mQ3^4 mu^2+9 mQ3^2 mu^4+2 mQ3^4 mU3^2-4 mQ3^2 mu^2 mU3^2) Xt^2)/((mQ3^2-mu^2)^2 (mQ3^2-mU3^2)^2)-(12 (9 mQ3^8-18 mQ3^6 mu^2+10 mQ3^4 mu^4+9 mQ3^6 mU3^2-18 mQ3^4 mu^2 mU3^2+9 mQ3^2 mu^4 mU3^2+mQ3^4 mU3^4-2 mQ3^2 mu^2 mU3^4) Xt^4)/((mQ3^2-mu^2)^2 (mQ3^2-mU3^2)^4)) Log[mQ3^2/MR^2]^2+(12 (2 mu^8+2 mQ3^4 mu^2 mU3^2-4 mQ3^2 mu^4 mU3^2-2 mu^6 mU3^2-mQ3^4 mU3^4+2 mQ3^2 mu^2 mU3^4+mu^4 mU3^4) Log[mu^2/MR^2]^2)/((-mQ3^2+mu^2)^2 (mu^2-mU3^2)^2)+((3 (9 mQ3^4 mu^4-13 mQ3^2 mu^6-4 mQ3^4 mu^2 mU3^2+14 mQ3^2 mu^4 mU3^2-2 mu^6 mU3^2+mQ3^4 mU3^4-9 mQ3^2 mu^2 mU3^4+4 mu^4 mU3^4+2 mQ3^2 mU3^6-2 mu^2 mU3^6))/(mQ3^2 (mQ3^2-mu^2) (mu^2-mU3^2)^2)+(6 (13 mQ3^6 mu^2-13 mQ3^4 mu^4+4 mQ3^2 mu^6-19 mQ3^6 mU3^2+16 mQ3^4 mu^2 mU3^2-5 mQ3^2 mu^4 mU3^2+9 mQ3^4 mU3^4-7 mQ3^2 mu^2 mU3^4+2 mu^4 mU3^4+2 mQ3^2 mU3^6-2 mu^2 mU3^6) Xt^2)/(mQ3^2 (mQ3^2-mu^2) (mQ3^2-mU3^2)^2 (mu^2-mU3^2))+(3 (-43 mQ3^8 mu^4+63 mQ3^6 mu^6-24 mQ3^4 mu^8+96 mQ3^8 mu^2 mU3^2-226 mQ3^6 mu^4 mU3^2+150 mQ3^4 mu^6 mU3^2-12 mQ3^2 mu^8 mU3^2-49 mQ3^8 mU3^4+233 mQ3^6 mu^2 mU3^4-117 mQ3^4 mu^4 mU3^4-67 mQ3^2 mu^6 mU3^4-82 mQ3^6 mU3^6-90 mQ3^4 mu^2 mU3^6+166 mQ3^2 mu^4 mU3^6-2 mu^6 mU3^6+93 mQ3^4 mU3^8-93 mQ3^2 mu^2 mU3^8+4 mu^4 mU3^8+2 mQ3^2 mU3^10-2 mu^2 mU3^10) Xt^4)/(mQ3^2 (mQ3^2-mu^2) (mQ3^2-mU3^2)^4 (mu^2-mU3^2)^2)) Log[mU3^2/MR^2]+((6 (2 mu^2 mU3^2-mU3^4))/(mu^2-mU3^2)^2+(12 (-2 mQ3^2 mu^2 mU3^2+8 mu^4 mU3^2+mQ3^2 mU3^4-14 mu^2 mU3^4+7 mU3^6) Xt^2)/((mu^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)-(6 (-2 mQ3^4 mu^2 mU3^2+16 mQ3^2 mu^4 mU3^2+mQ3^4 mU3^4-32 mQ3^2 mu^2 mU3^4+16 mu^4 mU3^4+16 mQ3^2 mU3^6-30 mu^2 mU3^6+15 mU3^8) Xt^4)/((mu^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) Log[mU3^2/MR^2]^2+Log[mu^2/MR^2] (-((6 (4 mQ3^6 mu^6-8 mQ3^4 mu^8+4 mQ3^2 mu^10-3 mQ3^6 mu^4 mU3^2+8 mQ3^4 mu^6 mU3^2-mQ3^2 mu^8 mU3^2+2 mu^10 mU3^2+2 mQ3^6 mu^2 mU3^4-8 mQ3^4 mu^4 mU3^4-2 mQ3^2 mu^6 mU3^4-4 mu^8 mU3^4+2 mQ3^4 mu^2 mU3^6+2 mQ3^2 mu^4 mU3^6+2 mu^6 mU3^6))/(mQ3^2 (mQ3^2-mu^2)^2 mU3^2 (mu^2-mU3^2)^2))+(24 (2 mQ3^4 mu^4-2 mQ3^2 mu^6-mQ3^4 mu^2 mU3^2+mQ3^2 mu^4 mU3^2+mu^6 mU3^2-mu^4 mU3^4) Xt^2)/(mQ3^2 (mQ3^2-mu^2) mU3^2 (mQ3^2-mU3^2) (mu^2-mU3^2))-(6 (4 mQ3^8 mu^6-8 mQ3^6 mu^8+4 mQ3^4 mu^10-6 mQ3^8 mu^4 mU3^2+16 mQ3^6 mu^6 mU3^2-10 mQ3^4 mu^8 mU3^2+4 mQ3^2 mu^10 mU3^2+4 mQ3^8 mu^2 mU3^4-18 mQ3^6 mu^4 mU3^4+14 mQ3^4 mu^6 mU3^4-10 mQ3^2 mu^8 mU3^4-2 mu^10 mU3^4+4 mQ3^6 mu^2 mU3^6+4 mQ3^2 mu^6 mU3^6+4 mu^8 mU3^6-2 mQ3^4 mu^2 mU3^8-2 mu^6 mU3^8) Xt^4)/(mQ3^2 (mQ3^2-mu^2)^2 mU3^2 (mQ3^2-mU3^2)^3 (mu^2-mU3^2)^2)+((6 (mQ3^4 mu^4-2 mQ3^2 mu^6-5 mu^8-4 mQ3^4 mu^2 mU3^2+8 mQ3^2 mu^4 mU3^2+8 mu^6 mU3^2+2 mQ3^4 mU3^4-4 mQ3^2 mu^2 mU3^4-4 mu^4 mU3^4))/((mQ3^2-mu^2)^2 (mu^2-mU3^2)^2)+(12 (mQ3^6 mu^4-4 mQ3^4 mu^6+3 mQ3^2 mu^8-2 mu^10-4 mQ3^6 mu^2 mU3^2+13 mQ3^4 mu^4 mU3^2-10 mQ3^2 mu^6 mU3^2+7 mu^8 mU3^2+2 mQ3^6 mU3^4-6 mQ3^4 mu^2 mU3^4+4 mQ3^2 mu^4 mU3^4-6 mu^6 mU3^4+2 mu^4 mU3^6) Xt^2)/((mQ3^2-mu^2)^2 (mQ3^2-mU3^2)^2 (mu^2-mU3^2)^2)-(6 (3 mQ3^8 mu^4-18 mQ3^6 mu^6+25 mQ3^4 mu^8-12 mQ3^2 mu^10-8 mQ3^8 mu^2 mU3^2+44 mQ3^6 mu^4 mU3^2-60 mQ3^4 mu^6 mU3^2+28 mQ3^2 mu^8 mU3^2+4 mQ3^8 mU3^4-28 mQ3^6 mu^2 mU3^4+41 mQ3^4 mu^4 mU3^4-18 mQ3^2 mu^6 mU3^4+mu^8 mU3^4+4 mQ3^6 mU3^6-4 mQ3^4 mu^2 mU3^6-4 mQ3^2 mu^4 mU3^6-2 mQ3^4 mU3^8+4 mQ3^2 mu^2 mU3^8) Xt^4)/((mQ3^2-mu^2)^2 (mQ3^2-mU3^2)^4 (mu^2-mU3^2)^2)) Log[mQ3^2/MR^2]+(-((6 (mQ3^4 mu^4-2 mQ3^2 mu^6+3 mu^8+4 mQ3^4 mu^2 mU3^2-8 mQ3^2 mu^4 mU3^2-2 mQ3^4 mU3^4+4 mQ3^2 mu^2 mU3^4))/((mQ3^2-mu^2)^2 (mu^2-mU3^2)^2))-(12 (mQ3^6 mu^4-4 mQ3^4 mu^6+3 mQ3^2 mu^8-2 mu^10-4 mQ3^6 mu^2 mU3^2+13 mQ3^4 mu^4 mU3^2-10 mQ3^2 mu^6 mU3^2+7 mu^8 mU3^2+2 mQ3^6 mU3^4-6 mQ3^4 mu^2 mU3^4+4 mQ3^2 mu^4 mU3^4-6 mu^6 mU3^4+2 mu^4 mU3^6) Xt^2)/((mQ3^2-mu^2)^2 (mQ3^2-mU3^2)^2 (mu^2-mU3^2)^2)+(6 (3 mQ3^8 mu^4-18 mQ3^6 mu^6+25 mQ3^4 mu^8-12 mQ3^2 mu^10-8 mQ3^8 mu^2 mU3^2+44 mQ3^6 mu^4 mU3^2-60 mQ3^4 mu^6 mU3^2+28 mQ3^2 mu^8 mU3^2+4 mQ3^8 mU3^4-28 mQ3^6 mu^2 mU3^4+41 mQ3^4 mu^4 mU3^4-18 mQ3^2 mu^6 mU3^4+mu^8 mU3^4+4 mQ3^6 mU3^6-4 mQ3^4 mu^2 mU3^6-4 mQ3^2 mu^4 mU3^6-2 mQ3^4 mU3^8+4 mQ3^2 mu^2 mU3^8) Xt^4)/((mQ3^2-mu^2)^2 (mQ3^2-mU3^2)^4 (mu^2-mU3^2)^2)) Log[mU3^2/MR^2])+Log[mQ3^2/MR^2] ((3 (4 mQ3^6 mu^2-8 mQ3^4 mu^4+4 mQ3^2 mu^6-4 mQ3^6 mU3^2+15 mQ3^4 mu^2 mU3^2-14 mQ3^2 mu^4 mU3^2+15 mu^6 mU3^2-5 mQ3^4 mU3^4+6 mQ3^2 mu^2 mU3^4-13 mu^4 mU3^4))/((mQ3^2-mu^2)^2 mU3^2 (mu^2-mU3^2))-(6 (4 mQ3^6 mu^2-4 mQ3^4 mu^4-4 mQ3^6 mU3^2+21 mQ3^4 mu^2 mU3^2-13 mQ3^2 mu^4 mU3^2+4 mu^6 mU3^2-19 mQ3^4 mU3^4+2 mQ3^2 mu^2 mU3^4+mu^4 mU3^4+15 mQ3^2 mU3^6-7 mu^2 mU3^6) Xt^2)/((mQ3^2-mu^2) mU3^2 (mQ3^2-mU3^2)^2 (mu^2-mU3^2))+(3 (4 mQ3^10 mu^2-8 mQ3^8 mu^4+4 mQ3^6 mu^6-4 mQ3^10 mU3^2+125 mQ3^8 mu^2 mU3^2-270 mQ3^6 mu^4 mU3^2+193 mQ3^4 mu^6 mU3^2-36 mQ3^2 mu^8 mU3^2-119 mQ3^8 mU3^4+226 mQ3^6 mu^2 mU3^4-107 mQ3^4 mu^4 mU3^4-24 mQ3^2 mu^6 mU3^4+44 mQ3^6 mU3^6-129 mQ3^4 mu^2 mU3^6+138 mQ3^2 mu^4 mU3^6-29 mu^6 mU3^6+43 mQ3^4 mU3^8-82 mQ3^2 mu^2 mU3^8+31 mu^4 mU3^8) Xt^4)/((mQ3^2-mu^2)^2 mU3^2 (mQ3^2-mU3^2)^4 (mu^2-mU3^2))+((6 (2 mu^8+2 mQ3^4 mu^2 mU3^2-4 mQ3^2 mu^4 mU3^2-2 mu^6 mU3^2-mQ3^4 mU3^4+2 mQ3^2 mu^2 mU3^4+mu^4 mU3^4))/((-mQ3^2+mu^2)^2 (mu^2-mU3^2)^2)+2 (-((54 mQ3^2)/(mQ3^2-mU3^2)^2)-(48 mU3^2)/(mQ3^2-mU3^2)^2+(12 mQ3^4)/((-mQ3^2+mu^2)^2 (mQ3^2-mU3^2))-(24 mQ3^2 mu^2)/((-mQ3^2+mu^2)^2 (mQ3^2-mU3^2))-(12 mu^2 mU3^2)/((mu^2-mU3^2)^2 (-mQ3^2+mU3^2))+(6 mU3^4)/((mu^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^2+(6 (18 mQ3^8 mu^4-36 mQ3^6 mu^6+20 mQ3^4 mu^8-38 mQ3^8 mu^2 mU3^2+110 mQ3^6 mu^4 mU3^2-110 mQ3^4 mu^6 mU3^2+34 mQ3^2 mu^8 mU3^2+19 mQ3^8 mU3^4-106 mQ3^6 mu^2 mU3^4+175 mQ3^4 mu^4 mU3^4-104 mQ3^2 mu^6 mU3^4+16 mu^8 mU3^4+34 mQ3^6 mU3^6-102 mQ3^4 mu^2 mU3^6+102 mQ3^2 mu^4 mU3^6-30 mu^6 mU3^6+17 mQ3^4 mU3^8-34 mQ3^2 mu^2 mU3^8+15 mu^4 mU3^8) Xt^4)/((mQ3^2-mu^2)^2 (mQ3^2-mU3^2)^4 (-mu^2+mU3^2)^2)) Log[mU3^2/MR^2])+(-((12 (mQ3^4-2 mQ3^2 mu^2-3 mu^4))/(mQ3^2-mu^2)^2)-(24 (mQ3^2-mu^2) Xt^2)/(mQ3^2-mU3^2)^2+(12 (2 mQ3^4-6 mQ3^2 mu^2+3 mu^4+2 mQ3^2 mU3^2-mU3^4) Xt^4)/(mQ3^2-mU3^2)^4) PolyLog[2,1-mQ3^2/mu^2]+((12 (mu^4+2 mu^2 mU3^2-mU3^4))/(mu^2-mU3^2)^2+(24 (mQ3^2-mu^2) Xt^2)/(mQ3^2-mU3^2)^2-(12 (2 mQ3^4-6 mQ3^2 mu^2+3 mu^4+2 mQ3^2 mU3^2-mU3^4) Xt^4)/(mQ3^2-mU3^2)^4) PolyLog[2,1-mU3^2/mu^2]+CosBeta^2 (9-(6 mA^2)/mQ3^2-(12 mA^2)/mU3^2+18 Log[mA^2/MR^2]^2+(18+(36 Xt Yt)/(mQ3^2-mU3^2)+(12 (4 mA^2-5 mQ3^2-3 mU3^2) Xt^3 Yt)/(mQ3^2-mU3^2)^3+(6 (7 mA^2-11 mQ3^2-3 mU3^2) Xt^4 Yt^2)/(mQ3^2-mU3^2)^4+Xt^2 ((6 (mA^2+mQ3^2-3 mU3^2))/(mQ3^2-mU3^2)^2+(18 Yt^2)/(mQ3^2-mU3^2)^2)) Log[mQ3^2/MR^2]^2+(12-(24 Xt Yt)/(mQ3^2-mU3^2)-(24 (mA^2-mQ3^2-mU3^2) Xt^3 Yt)/(mQ3^2-mU3^2)^3+(12 (2 mA^2-mQ3^2-3 mU3^2) Xt^4 Yt^2)/(mQ3^2-mU3^2)^4+Xt^2 (-(12/(mQ3^2-mU3^2))+(12 Yt^2)/(mQ3^2-mU3^2)^2)) Log[mU3^2/MR^2]^2+Log[mA^2/MR^2] (42+(6 mA^2)/mQ3^2+(12 mA^2)/mU3^2+(-18+(12 Xt Yt)/(mQ3^2-mU3^2)-(12 (6 mA^2-mQ3^2+mU3^2) Xt^3 Yt)/(mQ3^2-mU3^2)^3+Xt^2 ((6 (mA^2-5 mQ3^2+5 mU3^2))/(mQ3^2-mU3^2)^2+(6 Yt^2)/(mQ3^2-mU3^2)^2)+Xt^4 ((18 (-2 mA^2 mQ3^2+mQ3^4-mU3^4))/(mQ3^2-mU3^2)^4+(18 (-mA^2+mQ3^2-mU3^2) Yt^2)/(mQ3^2-mU3^2)^4)) Log[mQ3^2/MR^2]+(-18-(12 Xt Yt)/(mQ3^2-mU3^2)+(12 (6 mA^2-mQ3^2+mU3^2) Xt^3 Yt)/(mQ3^2-mU3^2)^3+Xt^2 (-((6 (mA^2-5 mQ3^2+5 mU3^2))/(mQ3^2-mU3^2)^2)-(6 Yt^2)/(mQ3^2-mU3^2)^2)+Xt^4 (-((18 (-2 mA^2 mQ3^2+mQ3^4-mU3^4))/(mQ3^2-mU3^2)^4)-(18 (-mA^2+mQ3^2-mU3^2) Yt^2)/(mQ3^2-mU3^2)^4)) Log[mU3^2/MR^2]+Xt^4 ((30 mA^2)/(mQ3^2-mU3^2)^3-(36 mQ3^2)/(mQ3^2-mU3^2)^3+(12 mA^2 mQ3^2)/(mU3^2 (mQ3^2-mU3^2)^3)+(36 mU3^2)/(mQ3^2-mU3^2)^3-(6 mA^2 mU3^2)/(mQ3^2 (mQ3^2-mU3^2)^3)+Yt^2 (-((3 (-mA^2 mQ3^2+mQ3^4+mA^2 mU3^2-3 mQ3^2 mU3^2-3 mU3^4))/(mQ3^2 mU3^4 (mQ3^2-mU3^2)^2))+(3 (-mA^6+3 mA^4 mQ3^2-3 mA^2 mQ3^4+mQ3^6+3 mA^4 mU3^2-3 mQ3^4 mU3^2-5 mA^2 mU3^4-mQ3^2 mU3^4+3 mU3^6))/(mQ3^2 mU3^4 (mQ3^2-mU3^2) TDelta[mA^2,mQ3^2,mU3^2])-(3 (-mA^4+mQ3^4-2 mQ3^2 mU3^2+mU3^4))/(mU3^2 (mQ3^2-mU3^2)^2 TDelta[mA^2,mU3^2,mQ3^2])))+Xt^2 ((12 mA^2)/(mQ3^2 (mQ3^2-mU3^2))-(24 mA^2)/(mU3^2 (mQ3^2-mU3^2))+Yt^2 ((6 (-mA^2 mQ3^2+mQ3^4-3 mQ3^2 mU3^2+mU3^4))/(mQ3^2 mU3^4 (mQ3^2-mU3^2))-(6 (-mA^6 mQ3^2+3 mA^4 mQ3^4-3 mA^2 mQ3^6+mQ3^8+2 mA^4 mQ3^2 mU3^2+2 mA^2 mQ3^4 mU3^2-4 mQ3^6 mU3^2-mA^4 mU3^4+mA^2 mQ3^2 mU3^4+6 mQ3^4 mU3^4-4 mQ3^2 mU3^6+mU3^8))/(mQ3^2 mU3^4 (mQ3^2-mU3^2) TDelta[mA^2,mQ3^2,mU3^2])+(6 (-mA^4+mQ3^4-2 mQ3^2 mU3^2+mU3^4))/(mU3^2 (mQ3^2-mU3^2) TDelta[mA^2,mU3^2,mQ3^2])))+Yt^2 (-(3/mQ3^2)+(3 mA^2)/mU3^4-(3 mQ3^2)/mU3^4+9/mU3^2+(3 mA^2)/(mQ3^2 mU3^2)+(21 mA^2+(15 mA^4)/mQ3^2+18 mQ3^2-(3 mA^6)/mU3^4+(9 mA^4 mQ3^2)/mU3^4-(9 mA^2 mQ3^4)/mU3^4+(3 mQ3^6)/mU3^4+(12 mA^4)/mU3^2-(3 mA^6)/(mQ3^2 mU3^2)+(3 mA^2 mQ3^2)/mU3^2-(12 mQ3^4)/mU3^2-12 mU3^2-(15 mA^2 mU3^2)/mQ3^2+(3 mU3^4)/mQ3^2)/TDelta[mA^2,mQ3^2,mU3^2]+(6 mQ3^2+(3 mA^4)/mU3^2-(3 mQ3^4)/mU3^2-3 mU3^2)/TDelta[mA^2,mU3^2,mQ3^2]))+Log[mQ3^2/MR^2] (-27-(72 Xt Yt)/(mQ3^2-mU3^2)+(72 (3 mQ3^2+mU3^2) Xt^3 Yt)/(mQ3^2-mU3^2)^3+(-((12 Xt Yt)/(mQ3^2-mU3^2))-(12 (2 mA^2-3 mQ3^2-mU3^2) Xt^3 Yt)/(mQ3^2-mU3^2)^3-(6 (11 mA^2-13 mQ3^2-9 mU3^2) Xt^4 Yt^2)/(mQ3^2-mU3^2)^4+Xt^2 ((6 (-mA^2+mQ3^2+mU3^2))/(mQ3^2-mU3^2)^2-(30 Yt^2)/(mQ3^2-mU3^2)^2)) Log[mU3^2/MR^2]+Xt^4 ((36 mA^2 mQ3^2)/(mQ3^2-mU3^2)^4-(9 mQ3^4)/(mQ3^2-mU3^2)^4+(9 mU3^4)/(mQ3^2-mU3^2)^4+Yt^2 (-((3 (mA^2 mQ3^4-mQ3^6-2 mA^2 mQ3^2 mU3^2-2 mQ3^4 mU3^2+mA^2 mU3^4-17 mQ3^2 mU3^4-4 mU3^6))/(mU3^4 (-mQ3^2+mU3^2)^4))-(3 (-mA^6+3 mA^4 mQ3^2-3 mA^2 mQ3^4+mQ3^6+5 mA^4 mU3^2-4 mA^2 mQ3^2 mU3^2-mQ3^4 mU3^2-9 mA^2 mU3^4-9 mQ3^2 mU3^4+9 mU3^6))/(mU3^4 (-mQ3^2+mU3^2)^2 TDelta[mA^2,mQ3^2,mU3^2])-(3 (mA^4-mQ3^4-2 mA^2 mU3^2+mU3^4))/(mU3^2 (-mQ3^2+mU3^2)^2 TDelta[mA^2,mU3^2,mQ3^2])))+Xt^2 (-((12 mA^2)/(mQ3^2-mU3^2)^2)+(6 mQ3^2)/(mQ3^2-mU3^2)^2+(18 mU3^2)/(mQ3^2-mU3^2)^2+Yt^2 (-((6 (-mA^2 mQ3^2+mQ3^4+mA^2 mU3^2+2 mQ3^2 mU3^2-mU3^4))/(mU3^4 (mQ3^2-mU3^2)^2))-(6 (-mA^6+3 mA^4 mQ3^2-3 mA^2 mQ3^4+mQ3^6+4 mA^4 mU3^2-2 mA^2 mQ3^2 mU3^2-2 mQ3^4 mU3^2-7 mA^2 mU3^4+mQ3^2 mU3^4))/(mU3^4 (-mQ3^2+mU3^2) TDelta[mA^2,mQ3^2,mU3^2])-(6 (mA^4-mQ3^4-2 mA^2 mU3^2+mU3^4))/(mU3^2 (-mQ3^2+mU3^2) TDelta[mA^2,mU3^2,mQ3^2])))+Yt^2 (-((3 mA^2)/mU3^4)+(3 mQ3^2)/mU3^4+6/mU3^2+(-9 mA^2-9 mQ3^2+(3 mA^6)/mU3^4-(9 mA^4 mQ3^2)/mU3^4+(9 mA^2 mQ3^4)/mU3^4-(3 mQ3^6)/mU3^4-(9 mA^4)/mU3^2+(9 mQ3^4)/mU3^2+3 mU3^2)/TDelta[mA^2,mQ3^2,mU3^2]+(6 mA^2-(3 mA^4)/mU3^2+(3 mQ3^4)/mU3^2-3 mU3^2)/TDelta[mA^2,mU3^2,mQ3^2]))+Log[mU3^2/MR^2] (-15+(72 Xt Yt)/(mQ3^2-mU3^2)-(72 (mQ3^2+3 mU3^2) Xt^3 Yt)/(mQ3^2-mU3^2)^3+Xt^4 (-((36 mA^2 mQ3^2)/(mQ3^2-mU3^2)^4)+(9 mQ3^4)/(mQ3^2-mU3^2)^4-(9 mU3^4)/(mQ3^2-mU3^2)^4+Yt^2 (-((3 (-mA^2 mQ3^4+3 mQ3^6+2 mA^2 mQ3^2 mU3^2+7 mQ3^4 mU3^2-mA^2 mU3^4+13 mQ3^2 mU3^4+mU3^6))/(mQ3^2 mU3^2 (mQ3^2-mU3^2)^4))+(3 (-mA^6+5 mA^4 mQ3^2-7 mA^2 mQ3^4+3 mQ3^6+3 mA^4 mU3^2-4 mA^2 mQ3^2 mU3^2-11 mQ3^4 mU3^2-5 mA^2 mU3^4+5 mQ3^2 mU3^4+3 mU3^6))/(mQ3^2 mU3^2 (mQ3^2-mU3^2)^2 TDelta[mA^2,mQ3^2,mU3^2])-(6 (mA^2+mQ3^2-mU3^2))/((mQ3^2-mU3^2)^2 TDelta[mA^2,mU3^2,mQ3^2])))+Xt^2 ((12 mA^2)/(mQ3^2-mU3^2)^2+(6 mQ3^2)/(mQ3^2-mU3^2)^2-(30 mU3^2)/(mQ3^2-mU3^2)^2+Yt^2 (-((6 (-2 mQ3^4-mQ3^2 mU3^2+mU3^4))/(mQ3^2 mU3^2 (mQ3^2-mU3^2)^2))-(6 (2 mA^4 mQ3^2-4 mA^2 mQ3^4+2 mQ3^6+mA^4 mU3^2-8 mA^2 mQ3^2 mU3^2-5 mQ3^4 mU3^2+4 mQ3^2 mU3^4-mU3^6))/(mQ3^2 mU3^2 (mQ3^2-mU3^2) TDelta[mA^2,mQ3^2,mU3^2])+(12 (mA^2+mQ3^2-mU3^2))/((mQ3^2-mU3^2) TDelta[mA^2,mU3^2,mQ3^2])))+Yt^2 (9/mQ3^2-3/mU3^2-(3 mA^2)/(mQ3^2 mU3^2)+(-12 mA^2-(15 mA^4)/mQ3^2-9 mQ3^2-(3 mA^4)/mU3^2+(3 mA^6)/(mQ3^2 mU3^2)-(3 mA^2 mQ3^2)/mU3^2+(3 mQ3^4)/mU3^2+9 mU3^2+(15 mA^2 mU3^2)/mQ3^2-(3 mU3^4)/mQ3^2)/TDelta[mA^2,mQ3^2,mU3^2]+(-6 mA^2-6 mQ3^2+6 mU3^2)/TDelta[mA^2,mU3^2,mQ3^2]))+((6 mA^4)/mQ3^4-(42 mA^2)/mQ3^2) TPhi[mA^2,mQ3^2,mQ3^2]-(6 TDelta[mA^2,mQ3^2,mQ3^2] TPhi[mA^2,mQ3^2,mQ3^2])/mQ3^4+Yt^2 (-(6/mQ3^2)-12/mU3^2+(-((9 mA^4)/mU3^6)+(18 mA^2 mQ3^2)/mU3^6-(9 mQ3^4)/mU3^6+(18 mA^2)/mU3^4+(18 mQ3^2)/mU3^4-9/mU3^2) TPhi[mA^2,mQ3^2,mU3^2]+((-12 mA^2-12 mQ3^2+(3 mA^8)/mU3^6-(12 mA^6 mQ3^2)/mU3^6+(18 mA^4 mQ3^4)/mU3^6-(12 mA^2 mQ3^6)/mU3^6+(3 mQ3^8)/mU3^6-(12 mA^6)/mU3^4+(12 mA^4 mQ3^2)/mU3^4+(12 mA^2 mQ3^4)/mU3^4-(12 mQ3^6)/mU3^4+(6 mA^4)/mU3^2+(12 mA^2 mQ3^2)/mU3^2+(18 mQ3^4)/mU3^2+3 mU3^2) TPhi[mA^2,mQ3^2,mU3^2])/TDelta[mA^2,mQ3^2,mU3^2]+(6 TDelta[mA^2,mQ3^2,mU3^2] TPhi[mA^2,mQ3^2,mU3^2])/mU3^6+(3 TPhi[mA^2,mU3^2,mQ3^2])/mQ3^2+((-6 mA^2-(3 mA^4)/mQ3^2-3 mQ3^2+6 mU3^2+(6 mA^2 mU3^2)/mQ3^2-(3 mU3^4)/mQ3^2) TPhi[mA^2,mU3^2,mQ3^2])/TDelta[mA^2,mU3^2,mQ3^2])+((6 mA^4)/mU3^4-(36 mA^2)/mU3^2) TPhi[mA^2,mU3^2,mU3^2]-(6 TDelta[mA^2,mU3^2,mU3^2] TPhi[mA^2,mU3^2,mU3^2])/mU3^4+Xt^3 Yt (-(288/(mQ3^2-mU3^2)^2)+((12 mA^4)/(mQ3^4 (mQ3^2-mU3^2)^2)-(84 mA^2)/(mQ3^2 (mQ3^2-mU3^2)^2)) TPhi[mA^2,mQ3^2,mQ3^2]+(-(60/(mQ3^2 (mQ3^2-mU3^2)^3))+(12 mU3^2)/(mQ3^4 (mQ3^2-mU3^2)^3)) TDelta[mA^2,mQ3^2,mQ3^2] TPhi[mA^2,mQ3^2,mQ3^2]+(-(12/(mQ3^2-mU3^2)^2)-(12 mA^2)/(mQ3^2 (mQ3^2-mU3^2)^2)+(12 mU3^2)/(mQ3^2 (mQ3^2-mU3^2)^2)) TPhi[mA^2,mU3^2,mQ3^2]+(24 TDelta[mA^2,mU3^2,mQ3^2] TPhi[mA^2,mU3^2,mQ3^2])/(mQ3^2 (mQ3^2-mU3^2)^3)+((12 mA^4)/(mU3^4 (-mQ3^2+mU3^2)^2)-(72 mA^2)/(mU3^2 (-mQ3^2+mU3^2)^2)) TPhi[mA^2,mU3^2,mU3^2]-(12 (mQ3^2-3 mU3^2) TDelta[mA^2,mU3^2,mU3^2] TPhi[mA^2,mU3^2,mU3^2])/(mU3^4 (mQ3^2-mU3^2)^3))+Xt Yt (((12 mA^4)/(mQ3^4 (mQ3^2-mU3^2))-(84 mA^2)/(mQ3^2 (mQ3^2-mU3^2))) TPhi[mA^2,mQ3^2,mQ3^2]-(12 TDelta[mA^2,mQ3^2,mQ3^2] TPhi[mA^2,mQ3^2,mQ3^2])/(mQ3^4 (mQ3^2-mU3^2))+(12/(mQ3^2-mU3^2)+(12 mA^2)/(mQ3^2 (mQ3^2-mU3^2))-(12 mU3^2)/(mQ3^2 (mQ3^2-mU3^2))) TPhi[mA^2,mU3^2,mQ3^2]+((12 mA^4)/(mU3^4 (-mQ3^2+mU3^2))-(72 mA^2)/(mU3^2 (-mQ3^2+mU3^2))) TPhi[mA^2,mU3^2,mU3^2]+(12 TDelta[mA^2,mU3^2,mU3^2] TPhi[mA^2,mU3^2,mU3^2])/(mU3^4 (mQ3^2-mU3^2)))+Xt^4 ((6 (-2 mA^2 mQ3^4-5 mA^2 mQ3^2 mU3^2+3 mQ3^4 mU3^2+mA^2 mU3^4-3 mQ3^2 mU3^4))/(mQ3^2 mU3^2 (mQ3^2-mU3^2)^3)+Yt^2 (-((6 (2 mQ3^4+11 mQ3^2 mU3^2-mU3^4))/(mQ3^2 mU3^2 (mQ3^2-mU3^2)^3))+((6 mA^4)/(mQ3^4 (mQ3^2-mU3^2)^3)-(42 mA^2)/(mQ3^2 (mQ3^2-mU3^2)^3)) TPhi[mA^2,mQ3^2,mQ3^2]+(-(48/(mQ3^2 (mQ3^2-mU3^2)^4))+(6 mU3^2)/(mQ3^4 (mQ3^2-mU3^2)^4)) TDelta[mA^2,mQ3^2,mQ3^2] TPhi[mA^2,mQ3^2,mQ3^2]-1/(mU3^6 (mQ3^2-mU3^2)^3) 3 (3 mA^4 mQ3^2-6 mA^2 mQ3^4+3 mQ3^6-9 mA^4 mU3^2+12 mA^2 mQ3^2 mU3^2-15 mQ3^4 mU3^2+18 mA^2 mU3^4+29 mQ3^2 mU3^4-17 mU3^6) TPhi[mA^2,mQ3^2,mU3^2]+(3 (mA^8-4 mA^6 mQ3^2+6 mA^4 mQ3^4-4 mA^2 mQ3^6+mQ3^8-4 mA^6 mU3^2+4 mA^4 mQ3^2 mU3^2+4 mA^2 mQ3^4 mU3^2-4 mQ3^6 mU3^2+6 mA^4 mU3^4+4 mA^2 mQ3^2 mU3^4+2 mQ3^4 mU3^4-4 mA^2 mU3^6+4 mQ3^2 mU3^6-3 mU3^8) TPhi[mA^2,mQ3^2,mU3^2])/(mU3^6 (mQ3^2-mU3^2)^2 TDelta[mA^2,mQ3^2,mU3^2])+(6 (mQ3^4-5 mQ3^2 mU3^2+12 mU3^4) TDelta[mA^2,mQ3^2,mU3^2] TPhi[mA^2,mQ3^2,mU3^2])/(mU3^6 (mQ3^2-mU3^2)^4)+(-(9/(mQ3^2-mU3^2)^3)-(12 mA^2)/(mQ3^2 (mQ3^2-mU3^2)^3)+(9 mU3^2)/(mQ3^2 (mQ3^2-mU3^2)^3)) TPhi[mA^2,mU3^2,mQ3^2]-(3 (mA^4+2 mA^2 mQ3^2+mQ3^4-2 mA^2 mU3^2-2 mQ3^2 mU3^2+mU3^4) TPhi[mA^2,mU3^2,mQ3^2])/(mQ3^2 (mQ3^2-mU3^2)^2 TDelta[mA^2,mU3^2,mQ3^2])+(18 TDelta[mA^2,mU3^2,mQ3^2] TPhi[mA^2,mU3^2,mQ3^2])/(mQ3^2 (mQ3^2-mU3^2)^4)+(6 (mA^4-6 mA^2 mU3^2) TPhi[mA^2,mU3^2,mU3^2])/(mU3^4 (-mQ3^2+mU3^2)^3)+(6 (mQ3^2-5 mU3^2) TDelta[mA^2,mU3^2,mU3^2] TPhi[mA^2,mU3^2,mU3^2])/(mU3^4 (mQ3^2-mU3^2)^4)))+Xt^2 (-(24/(mQ3^2-mU3^2))-(12 mA^2)/(mQ3^2 (mQ3^2-mU3^2))+(24 mA^2)/(mU3^2 (mQ3^2-mU3^2))+((6 mA^4)/(mQ3^4 (mQ3^2-mU3^2))-(42 mA^2)/(mQ3^2 (mQ3^2-mU3^2))) TPhi[mA^2,mQ3^2,mQ3^2]+(-(12/(mQ3^2 (mQ3^2-mU3^2)^2))+(6 mU3^2)/(mQ3^4 (mQ3^2-mU3^2)^2)) TDelta[mA^2,mQ3^2,mQ3^2] TPhi[mA^2,mQ3^2,mQ3^2]+(6 TDelta[mA^2,mU3^2,mQ3^2] TPhi[mA^2,mU3^2,mQ3^2])/(mQ3^2 (mQ3^2-mU3^2)^2)+((6 mA^4)/(mU3^4 (-mQ3^2+mU3^2))-(36 mA^2)/(mU3^2 (-mQ3^2+mU3^2))) TPhi[mA^2,mU3^2,mU3^2]+(6 TDelta[mA^2,mU3^2,mU3^2] TPhi[mA^2,mU3^2,mU3^2])/(mU3^4 (mQ3^2-mU3^2))+Yt^2 (-(12/(mQ3^2 (mQ3^2-mU3^2)))+24/(mU3^2 (mQ3^2-mU3^2))+((6 mA^4)/(mQ3^4 (mQ3^2-mU3^2)^2)-(42 mA^2)/(mQ3^2 (mQ3^2-mU3^2)^2)) TPhi[mA^2,mQ3^2,mQ3^2]-(6 TDelta[mA^2,mQ3^2,mQ3^2] TPhi[mA^2,mQ3^2,mQ3^2])/(mQ3^4 (-mQ3^2+mU3^2)^2)+1/(mU3^6 (mQ3^2-mU3^2)^2) 6 (3 mA^4 mQ3^2-6 mA^2 mQ3^4+3 mQ3^6-5 mA^4 mU3^2+4 mA^2 mQ3^2 mU3^2-11 mQ3^4 mU3^2+14 mA^2 mU3^4+13 mQ3^2 mU3^4-5 mU3^6) TPhi[mA^2,mQ3^2,mU3^2]-(6 (mA^8-4 mA^6 mQ3^2+6 mA^4 mQ3^4-4 mA^2 mQ3^6+mQ3^8-4 mA^6 mU3^2+4 mA^4 mQ3^2 mU3^2+4 mA^2 mQ3^4 mU3^2-4 mQ3^6 mU3^2+6 mA^4 mU3^4+6 mQ3^4 mU3^4-4 mQ3^2 mU3^6+mU3^8) TPhi[mA^2,mQ3^2,mU3^2])/(mU3^6 (mQ3^2-mU3^2) TDelta[mA^2,mQ3^2,mU3^2])-(12 (mQ3^2-2 mU3^2) TDelta[mA^2,mQ3^2,mU3^2] TPhi[mA^2,mQ3^2,mU3^2])/(mU3^6 (mQ3^2-mU3^2)^2)+(6 mA^2 TPhi[mA^2,mU3^2,mQ3^2])/(mQ3^2 (mQ3^2-mU3^2)^2)+(6 (mA^4+2 mA^2 mQ3^2+mQ3^4-2 mA^2 mU3^2-2 mQ3^2 mU3^2+mU3^4) TPhi[mA^2,mU3^2,mQ3^2])/(mQ3^2 (mQ3^2-mU3^2) TDelta[mA^2,mU3^2,mQ3^2])+((6 mA^4)/(mU3^4 (-mQ3^2+mU3^2)^2)-(36 mA^2)/(mU3^2 (-mQ3^2+mU3^2)^2)) TPhi[mA^2,mU3^2,mU3^2]-(6 TDelta[mA^2,mU3^2,mU3^2] TPhi[mA^2,mU3^2,mU3^2])/(mU3^4 (mQ3^2-mU3^2)^2)))+36 Inactive[Zeta][2]))
);

(* Calculated based on [Fleischer et al., 1999, arXiv:hep-ph/9803493v5 ] & [Bednyakov et al., 2003, arXiv:hep-ph/0210258v2]*)

deltaytas2=(
(291 m3^8-282 m3^6 mQ3^2+93 m3^4 mQ3^4-18 m3^2 mQ3^6-360 m3^6 msq^2+300 m3^4 mQ3^2 msq^2-180 m3^2 mQ3^4 msq^2-282 m3^6 mU3^2-112 m3^4 mQ3^2 mU3^2+208 m3^2 mQ3^4 mU3^2-6 mQ3^6 mU3^2+300 m3^4 msq^2 mU3^2+240 m3^2 mQ3^2 msq^2 mU3^2-60 mQ3^4 msq^2 mU3^2+93 m3^4 mU3^4+208 m3^2 mQ3^2 mU3^4-169 mQ3^4 mU3^4-180 m3^2 msq^2 mU3^4-60 mQ3^2 msq^2 mU3^4-18 m3^2 mU3^6-6 mQ3^2 mU3^6)/(18 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(8 (14 m3^3-3 m3 mQ3^2-30 m3 msq^2-3 m3 mU3^2) Xt)/(9 (m3^2-mQ3^2) (m3^2-mU3^2))+((262 m3^16-878 m3^14 mQ3^2+885 m3^12 mQ3^4-320 m3^10 mQ3^6+49 m3^8 mQ3^8-878 m3^14 mU3^2+2972 m3^12 mQ3^2 mU3^2-2932 m3^10 mQ3^4 mU3^2+1004 m3^8 mQ3^6 mU3^2-158 m3^6 mQ3^8 mU3^2+885 m3^12 mU3^4-2932 m3^10 mQ3^2 mU3^4+2404 m3^8 mQ3^4 mU3^4-368 m3^6 mQ3^6 mU3^4-m3^4 mQ3^8 mU3^4-320 m3^10 mU3^6+1004 m3^8 mQ3^2 mU3^6-368 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6+49 m3^8 mU3^8-158 m3^6 mQ3^2 mU3^8-m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8)/(9 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4)-(8 (27 m3^11-82 m3^9 mQ3^2+57 m3^7 mQ3^4-82 m3^9 mU3^2+220 m3^7 mQ3^2 mU3^2-142 m3^5 mQ3^4 mU3^2+57 m3^7 mU3^4-142 m3^5 mQ3^2 mU3^4+87 m3^3 mQ3^4 mU3^4) Xt)/(9 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(64 m3^6 Xt^2)/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Log[m3^2/MR^2]^2+((128 m3^16-556 m3^14 mQ3^2+714 m3^12 mQ3^4-373 m3^10 mQ3^6+96 m3^8 mQ3^8-42 m3^6 mQ3^10+38 m3^4 mQ3^12-9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-340 m3^14 mU3^2+1712 m3^12 mQ3^2 mU3^2-2209 m3^10 mQ3^4 mU3^2+933 m3^8 mQ3^6 mU3^2+106 m3^6 mQ3^8 mU3^2-230 m3^4 mQ3^10 mU3^2+47 m3^2 mQ3^12 mU3^2-3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+262 m3^12 mU3^4-1903 m3^10 mQ3^2 mU3^4+2771 m3^8 mQ3^4 mU3^4-1506 m3^6 mQ3^6 mU3^4+316 m3^4 mQ3^8 mU3^4+37 m3^2 mQ3^10 mU3^4-mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4+5 m3^10 mU3^6+723 m3^8 mQ3^2 mU3^6-1186 m3^6 mQ3^4 mU3^6+610 m3^4 mQ3^6 mU3^6-119 m3^2 mQ3^8 mU3^6-17 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-43 m3^8 mU3^8-60 m3^6 mQ3^2 mU3^8+162 m3^4 mQ3^4 mU3^8-84 m3^2 mQ3^6 mU3^8+21 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8)/(18 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(4 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4))/(9 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))-(4 (5 m3^5 mQ3^2-5 m3^3 mQ3^4+3 m3 mQ3^6-5 m3^5 mU3^2-4 m3 mQ3^4 mU3^2+5 m3^3 mU3^4+4 m3 mQ3^2 mU3^4-3 m3 mU3^6))/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(4 (33 m3^9 mQ3^4-75 m3^7 mQ3^6+33 m3^5 mQ3^8+8 m3^3 mQ3^10-3 m3 mQ3^12-17 m3^9 mQ3^2 mU3^2-33 m3^7 mQ3^4 mU3^2+158 m3^5 mQ3^6 mU3^2-106 m3^3 mQ3^8 mU3^2+10 m3 mQ3^10 mU3^2+49 m3^7 mQ3^2 mU3^4-78 m3^5 mQ3^4 mU3^4-11 m3^3 mQ3^6 mU3^4+28 m3 mQ3^8 mU3^4-5 m3^7 mU3^6-22 m3^5 mQ3^2 mU3^6+43 m3^3 mQ3^4 mU3^6-12 m3 mQ3^6 mU3^6+5 m3^5 mU3^8+5 m3^3 mQ3^2 mU3^8-10 m3 mQ3^4 mU3^8-3 m3^3 mU3^10+3 m3 mQ3^2 mU3^10))/(9 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) Xt+(64 m3^2 mQ3^4 Xt^2)/(9 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)) Log[mQ3^2/MR^2]^2+(-((10 (12 m3^8-23 m3^6 mQ3^2+11 m3^4 mQ3^4+18 m3^6 msq^2-15 m3^4 mQ3^2 msq^2+9 m3^2 mQ3^4 msq^2-23 m3^6 mU3^2+44 m3^4 mQ3^2 mU3^2-21 m3^2 mQ3^4 mU3^2-15 m3^4 msq^2 mU3^2-12 m3^2 mQ3^2 msq^2 mU3^2+3 mQ3^4 msq^2 mU3^2+11 m3^4 mU3^4-21 m3^2 mQ3^2 mU3^4+10 mQ3^4 mU3^4+9 m3^2 msq^2 mU3^4+3 mQ3^2 msq^2 mU3^4))/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(80 m3 msq^2 Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))) Log[msq^2/MR^2]+(-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)+((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt) Log[msq^2/MR^2]^2+((52 m3^10 mQ3^2-105 m3^8 mQ3^4+53 m3^6 mQ3^6-308 m3^10 mU3^2+764 m3^8 mQ3^2 mU3^2-622 m3^6 mQ3^4 mU3^2+163 m3^4 mQ3^6 mU3^2-9 m3^2 mQ3^8 mU3^2-90 m3^6 mQ3^2 msq^2 mU3^2+180 m3^4 mQ3^4 msq^2 mU3^2-90 m3^2 mQ3^6 msq^2 mU3^2+365 m3^8 mU3^4-985 m3^6 mQ3^2 mU3^4+942 m3^4 mQ3^4 mU3^4-271 m3^2 mQ3^6 mU3^4-3 mQ3^8 mU3^4+90 m3^6 msq^2 mU3^4-210 m3^4 mQ3^2 msq^2 mU3^4+150 m3^2 mQ3^4 msq^2 mU3^4-30 mQ3^6 msq^2 mU3^4+18 m3^6 mU3^6+25 m3^4 mQ3^2 mU3^6-246 m3^2 mQ3^4 mU3^6+131 mQ3^6 mU3^6+30 m3^4 msq^2 mU3^6-60 m3^2 mQ3^2 msq^2 mU3^6+30 mQ3^4 msq^2 mU3^6-106 m3^4 mU3^8+279 m3^2 mQ3^2 mU3^8-125 mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10)/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(8 (16 m3^7-16 m3^5 mQ3^2+55 m3^5 mU3^2-53 m3^3 mQ3^2 mU3^2+3 m3 mQ3^4 mU3^2-30 m3^3 msq^2 mU3^2+30 m3 mQ3^2 msq^2 mU3^2-61 m3^3 mU3^4+53 m3 mQ3^2 mU3^4+3 m3 mU3^6) Xt)/(9 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(-((10 (6 m3^4 msq^2-9 m3^2 msq^4+2 m3^4 mU3^2+18 m3^2 msq^2 mU3^2-3 msq^4 mU3^2-3 m3^2 mU3^4+mU3^6))/(9 (m3^2-mU3^2)^3))+(40 (-6 m3 msq^4+m3^3 mU3^2+12 m3 msq^2 mU3^2-m3 mU3^4) Xt)/(9 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Log[msq^2/MR^2]) Log[mU3^2/MR^2]+((128 m3^10+44 m3^8 mQ3^2-2 m3^6 mQ3^4+9 m3^4 mQ3^6-60 m3^6 mQ3^2 msq^2+90 m3^4 mQ3^2 msq^4-556 m3^8 mU3^2+68 m3^6 mQ3^2 mU3^2-33 m3^4 mQ3^4 mU3^2-6 m3^2 mQ3^6 mU3^2+60 m3^6 msq^2 mU3^2-120 m3^4 mQ3^2 msq^2 mU3^2-90 m3^4 msq^4 mU3^2-60 m3^2 mQ3^2 msq^4 mU3^2+702 m3^6 mU3^4-141 m3^4 mQ3^2 mU3^4+36 m3^2 mQ3^4 mU3^4-3 mQ3^6 mU3^4+120 m3^4 msq^2 mU3^4+180 m3^2 mQ3^2 msq^2 mU3^4+60 m3^2 msq^4 mU3^4-30 mQ3^2 msq^4 mU3^4-347 m3^4 mU3^6+50 m3^2 mQ3^2 mU3^6-mQ3^4 mU3^6-180 m3^2 msq^2 mU3^6+30 msq^4 mU3^6+48 m3^2 mU3^8-17 mQ3^2 mU3^8+21 mU3^10)/(18 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2))-(4 (-18 m3^7 mQ3^2+m3^5 mQ3^4+3 m3^3 mQ3^6+30 m3^3 mQ3^2 msq^4+18 m3^7 mU3^2+69 m3^5 mQ3^2 mU3^2-11 m3^3 mQ3^4 mU3^2-3 m3 mQ3^6 mU3^2-60 m3^3 mQ3^2 msq^2 mU3^2-30 m3^3 msq^4 mU3^2-30 m3 mQ3^2 msq^4 mU3^2-86 m3^5 mU3^4-59 m3^3 mQ3^2 mU3^4+10 m3 mQ3^4 mU3^4+60 m3^3 msq^2 mU3^4+60 m3 mQ3^2 msq^2 mU3^4+30 m3 msq^4 mU3^4+99 m3^3 mU3^6+4 m3 mQ3^2 mU3^6-60 m3 msq^2 mU3^6-27 m3 mU3^8) Xt)/(9 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)+(64 m3^2 mU3^4 Xt^2)/(9 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)) Log[mU3^2/MR^2]^2+Log[m3^2/MR^2] ((-864 m3^12+1942 m3^10 mQ3^2-1460 m3^8 mQ3^4+351 m3^6 mQ3^6-9 m3^4 mQ3^8+180 m3^10 msq^2-240 m3^8 mQ3^2 msq^2+270 m3^6 mQ3^4 msq^2-90 m3^4 mQ3^6 msq^2+1942 m3^10 mU3^2-4088 m3^8 mQ3^2 mU3^2+2853 m3^6 mQ3^4 mU3^2-572 m3^4 mQ3^6 mU3^2-3 m3^2 mQ3^8 mU3^2-240 m3^8 msq^2 mU3^2-180 m3^6 mQ3^2 msq^2 mU3^2+90 m3^4 mQ3^4 msq^2 mU3^2-30 m3^2 mQ3^6 msq^2 mU3^2-1460 m3^8 mU3^4+2853 m3^6 mQ3^2 mU3^4-1894 m3^4 mQ3^4 mU3^4+345 m3^2 mQ3^6 mU3^4+270 m3^6 msq^2 mU3^4+90 m3^4 mQ3^2 msq^2 mU3^4+351 m3^6 mU3^6-572 m3^4 mQ3^2 mU3^6+345 m3^2 mQ3^4 mU3^6-48 mQ3^6 mU3^6-90 m3^4 msq^2 mU3^6-30 m3^2 mQ3^2 msq^2 mU3^6-9 m3^4 mU3^8-3 m3^2 mQ3^2 mU3^8)/(9 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(8 (85 m3^7-75 m3^5 mQ3^2+3 m3^3 mQ3^4-60 m3^5 msq^2+30 m3^3 mQ3^2 msq^2-75 m3^5 mU3^2+59 m3^3 mQ3^2 mU3^2+30 m3^3 msq^2 mU3^2+3 m3^3 mU3^4) Xt)/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(-((2 (64 m3^16-180 m3^14 mQ3^2+122 m3^12 mQ3^4-31 m3^10 mQ3^6+30 m3^8 mQ3^8-7 m3^6 mQ3^10-204 m3^14 mU3^2+604 m3^12 mQ3^2 mU3^2-369 m3^10 mQ3^4 mU3^2+13 m3^8 mQ3^6 mU3^2-49 m3^6 mQ3^8 mU3^2+13 m3^4 mQ3^10 mU3^2+234 m3^12 mU3^4-779 m3^10 mQ3^2 mU3^4+503 m3^8 mQ3^4 mU3^4-13 m3^6 mQ3^6 mU3^4+61 m3^4 mQ3^8 mU3^4-18 m3^2 mQ3^10 mU3^4-101 m3^10 mU3^6+401 m3^8 mQ3^2 mU3^6-239 m3^6 mQ3^4 mU3^6-53 m3^4 mQ3^6 mU3^6-6 m3^2 mQ3^8 mU3^6+6 mQ3^10 mU3^6+13 m3^8 mU3^8-76 m3^6 mQ3^2 mU3^8+43 m3^4 mQ3^4 mU3^8+24 m3^2 mQ3^6 mU3^8-6 mQ3^8 mU3^8))/(9 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(4 (68 m3^11-183 m3^9 mQ3^2+120 m3^7 mQ3^4+3 m3^5 mQ3^6-137 m3^9 mU3^2+353 m3^7 mQ3^2 mU3^2-215 m3^5 mQ3^4 mU3^2-17 m3^3 mQ3^6 mU3^2+75 m3^7 mU3^4-184 m3^5 mQ3^2 mU3^4+105 m3^3 mQ3^4 mU3^4+12 m3 mQ3^6 mU3^4) Xt)/(9 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(128 m3^4 mQ3^2 Xt^2)/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))) Log[mQ3^2/MR^2]+((-40 m3^12+200 m3^10 mQ3^2-60 m3^8 mQ3^4+20 m3^6 mQ3^6+200 m3^10 mU3^2-840 m3^8 mQ3^2 mU3^2+420 m3^6 mQ3^4 mU3^2-140 m3^4 mQ3^6 mU3^2-60 m3^8 mU3^4+420 m3^6 mQ3^2 mU3^4+20 m3^6 mU3^6-140 m3^4 mQ3^2 mU3^6)/(9 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(40 (m3^7+5 m3^5 mQ3^2+5 m3^5 mU3^2-11 m3^3 mQ3^2 mU3^2) Xt)/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Log[msq^2/MR^2]+((2 (64 m3^16-204 m3^14 mQ3^2+234 m3^12 mQ3^4-101 m3^10 mQ3^6+13 m3^8 mQ3^8-180 m3^14 mU3^2+604 m3^12 mQ3^2 mU3^2-779 m3^10 mQ3^4 mU3^2+401 m3^8 mQ3^6 mU3^2-76 m3^6 mQ3^8 mU3^2+122 m3^12 mU3^4-369 m3^10 mQ3^2 mU3^4+503 m3^8 mQ3^4 mU3^4-239 m3^6 mQ3^6 mU3^4+43 m3^4 mQ3^8 mU3^4-31 m3^10 mU3^6+13 m3^8 mQ3^2 mU3^6-13 m3^6 mQ3^4 mU3^6-53 m3^4 mQ3^6 mU3^6+24 m3^2 mQ3^8 mU3^6+30 m3^8 mU3^8-49 m3^6 mQ3^2 mU3^8+61 m3^4 mQ3^4 mU3^8-6 m3^2 mQ3^6 mU3^8-6 mQ3^8 mU3^8-7 m3^6 mU3^10+13 m3^4 mQ3^2 mU3^10-18 m3^2 mQ3^4 mU3^10+6 mQ3^6 mU3^10))/(9 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2))-(4 (68 m3^11-137 m3^9 mQ3^2+75 m3^7 mQ3^4-183 m3^9 mU3^2+353 m3^7 mQ3^2 mU3^2-184 m3^5 mQ3^4 mU3^2+120 m3^7 mU3^4-215 m3^5 mQ3^2 mU3^4+105 m3^3 mQ3^4 mU3^4+3 m3^5 mU3^6-17 m3^3 mQ3^2 mU3^6+12 m3 mQ3^4 mU3^6) Xt)/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(128 m3^4 mU3^2 Xt^2)/(9 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))) Log[mU3^2/MR^2])+Log[mQ3^2/MR^2] ((308 m3^10 mQ3^2-365 m3^8 mQ3^4-18 m3^6 mQ3^6+106 m3^4 mQ3^8+9 m3^2 mQ3^10-90 m3^6 mQ3^4 msq^2-30 m3^4 mQ3^6 msq^2-52 m3^10 mU3^2-764 m3^8 mQ3^2 mU3^2+985 m3^6 mQ3^4 mU3^2-25 m3^4 mQ3^6 mU3^2-279 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2+90 m3^6 mQ3^2 msq^2 mU3^2+210 m3^4 mQ3^4 msq^2 mU3^2+60 m3^2 mQ3^6 msq^2 mU3^2+105 m3^8 mU3^4+622 m3^6 mQ3^2 mU3^4-942 m3^4 mQ3^4 mU3^4+246 m3^2 mQ3^6 mU3^4+125 mQ3^8 mU3^4-180 m3^4 mQ3^2 msq^2 mU3^4-150 m3^2 mQ3^4 msq^2 mU3^4-30 mQ3^6 msq^2 mU3^4-53 m3^6 mU3^6-163 m3^4 mQ3^2 mU3^6+271 m3^2 mQ3^4 mU3^6-131 mQ3^6 mU3^6+90 m3^2 mQ3^2 msq^2 mU3^6+30 mQ3^4 msq^2 mU3^6+9 m3^2 mQ3^2 mU3^8+3 mQ3^4 mU3^8)/(9 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(8 (16 m3^7+55 m3^5 mQ3^2-61 m3^3 mQ3^4+3 m3 mQ3^6-30 m3^3 mQ3^2 msq^2-16 m3^5 mU3^2-53 m3^3 mQ3^2 mU3^2+53 m3 mQ3^4 mU3^2+30 m3 mQ3^2 msq^2 mU3^2+3 m3 mQ3^2 mU3^4) Xt)/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))+(-((10 (2 m3^4 mQ3^2-3 m3^2 mQ3^4+mQ3^6+6 m3^4 msq^2+18 m3^2 mQ3^2 msq^2-9 m3^2 msq^4-3 mQ3^2 msq^4))/(9 (m3^2-mQ3^2)^3))+(40 (m3^3 mQ3^2-m3 mQ3^4+12 m3 mQ3^2 msq^2-6 m3 msq^4) Xt)/(9 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) Log[msq^2/MR^2]+((-14 m3^10 mQ3^2+17 m3^8 mQ3^4-35 m3^6 mQ3^6+29 m3^4 mQ3^8-9 m3^2 mQ3^10-2 m3^10 mU3^2+28 m3^8 mQ3^2 mU3^2+53 m3^6 mQ3^4 mU3^2-75 m3^4 mQ3^6 mU3^2+35 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2+3 m3^8 mU3^4-57 m3^6 mQ3^2 mU3^4+23 m3^4 mQ3^4 mU3^4-m3^2 mQ3^6 mU3^4-4 mQ3^8 mU3^4-m3^6 mU3^6+19 m3^4 mQ3^2 mU3^6-9 m3^2 mQ3^4 mU3^6+3 mQ3^6 mU3^6)/(9 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(4 (m3^7 mQ3^4+9 m3^5 mQ3^6-10 m3^3 mQ3^8+6 m3 mQ3^10-34 m3^7 mQ3^2 mU3^2+29 m3^5 mQ3^4 mU3^2+11 m3^3 mQ3^6 mU3^2-20 m3 mQ3^8 mU3^2+m3^7 mU3^4+59 m3^5 mQ3^2 mU3^4-88 m3^3 mQ3^4 mU3^4+38 m3 mQ3^6 mU3^4-m3^5 mU3^6-9 m3^3 mQ3^2 mU3^6+8 m3 mQ3^4 mU3^6) Xt)/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(128 m3^2 mQ3^2 mU3^2 Xt^2)/(9 (m3^2-mQ3^2) (m3^2-mU3^2) (-mQ3^2+mU3^2)^2)) Log[mU3^2/MR^2])+(-((-128 m3^12+294 m3^10 mQ3^2-98 m3^8 mQ3^4-9 m3^6 mQ3^6-20 m3^4 mQ3^8+9 m3^2 mQ3^10+346 m3^10 mU3^2-902 m3^8 mQ3^2 mU3^2+251 m3^6 mQ3^4 mU3^2+151 m3^4 mQ3^6 mU3^2-41 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2-280 m3^8 mU3^4+1025 m3^6 mQ3^2 mU3^4-417 m3^4 mQ3^4 mU3^4-41 m3^2 mQ3^6 mU3^4+mQ3^8 mU3^4+13 m3^6 mU3^6-391 m3^4 mQ3^2 mU3^6+167 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6+37 m3^4 mU3^8+34 m3^2 mQ3^2 mU3^8-23 mQ3^4 mU3^8)/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(8 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4) Xt)/(9 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) PolyLog[2,1-m3^2/mQ3^2]+(1/2 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2)))+(80 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) PolyLog[2,1-m3^2/msq^2]+(-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(3 (-m3^2+mQ3^2)^3))+(80 m3 (mQ3^2-msq^2)^2 Xt)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-msq^2/mQ3^2]+((-128 m3^12+346 m3^10 mQ3^2-280 m3^8 mQ3^4+13 m3^6 mQ3^6+37 m3^4 mQ3^8+294 m3^10 mU3^2-902 m3^8 mQ3^2 mU3^2+1025 m3^6 mQ3^4 mU3^2-391 m3^4 mQ3^6 mU3^2+34 m3^2 mQ3^8 mU3^2-98 m3^8 mU3^4+251 m3^6 mQ3^2 mU3^4-417 m3^4 mQ3^4 mU3^4+167 m3^2 mQ3^6 mU3^4-23 mQ3^8 mU3^4-9 m3^6 mU3^6+151 m3^4 mQ3^2 mU3^6-41 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6-20 m3^4 mU3^8-41 m3^2 mQ3^2 mU3^8+mQ3^4 mU3^8+9 m3^2 mU3^10+3 mQ3^2 mU3^10)/(9 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(8 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt)/(9 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-m3^2/mU3^2]+((-24 m3^10 mQ3^2+28 m3^8 mQ3^4-68 m3^6 mQ3^6+58 m3^4 mQ3^8-18 m3^2 mQ3^10+24 m3^10 mU3^2+220 m3^6 mQ3^4 mU3^2-188 m3^4 mQ3^6 mU3^2+70 m3^2 mQ3^8 mU3^2-6 mQ3^10 mU3^2-28 m3^8 mU3^4-220 m3^6 mQ3^2 mU3^4+16 m3^2 mQ3^6 mU3^4-8 mQ3^8 mU3^4+68 m3^6 mU3^6+188 m3^4 mQ3^2 mU3^6-16 m3^2 mQ3^4 mU3^6-58 m3^4 mU3^8-70 m3^2 mQ3^2 mU3^8+8 mQ3^4 mU3^8+18 m3^2 mU3^10+6 mQ3^2 mU3^10)/(18 (-m3^2+mQ3^2)^3 (m3^2-mU3^2)^3)-(4 (10 m3^5 mQ3^2-10 m3^3 mQ3^4+6 m3 mQ3^6-10 m3^5 mU3^2-8 m3 mQ3^4 mU3^2+10 m3^3 mU3^4+8 m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt)/(9 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) PolyLog[2,1-mQ3^2/mU3^2]+(-((5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(3 (-m3^2+mU3^2)^3))+(80 m3 (-msq^2+mU3^2)^2 Xt)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) PolyLog[2,1-msq^2/mU3^2]
);



(* Degenerate *)

(* Taken from [Bagnaschi et al., 2014, arXiv:1407.4081v2], Eq. (2.36) and complete expression provided by A. Voigt *)

deltalambdaatasDeg=(
-((64 Xt)/MS)-(32 Xt^2)/MS^2+(224 Xt^3)/(3 MS^3)+(8 Xt^4)/(3 MS^4)-(16 Xt^5)/(3 MS^5)+((128 Xt)/MS-(128 Xt^2)/MS^2-(64 Xt^3)/(3 MS^3)+(16 Xt^4)/(3 MS^4)) Log[MS^2/MR^2]-96 Log[MS^2/MR^2]^2
);

(* Taken from [Pardo Vega and Villadoro, 2015, arXiv:1504.05200v1], Eq. (38) and complete expression provided by A. Voigt *)

deltalambdaat2Deg=
(1/2)*Ndlat2*(
-((3 Xt^6)/MS^6)+(-24+(3 Xt^6)/MS^6) Log[MS^2/MR^2]+18 Log[MS^2/MR^2]^2+CscBeta^2 ((9 (11 MS^4-5 MS^2 mu^2-4 mu^4))/(MS^2 (MS^2-mu^2))-(6 (11 MS^4+5 MS^2 mu^2-6 mu^4) Xt^2)/(MS^4 (MS^2-mu^2))+(3 (12 MS^4-8 MS^2 mu^2-2 mu^4) Xt^4)/(MS^6 (MS^2-mu^2))+(-((6 (2 MS^2 mu^2-11 mu^4))/(MS^2-mu^2)^2)+(6 (27 MS^4-48 MS^2 mu^2+31 mu^4) Xt^2)/(MS^2 (MS^2-mu^2)^2)-(3 (13 MS^4-24 MS^2 mu^2+13 mu^4) Xt^4)/(MS^4 (MS^2-mu^2)^2)) Log[MS^2/MR^2]-(12 (MS^4-2 MS^2 mu^2-2 mu^4) Log[MS^2/MR^2]^2)/(MS^2-mu^2)^2+(-((6 (4 MS^4 mu^2-MS^2 mu^4+6 mu^6))/(MS^2 (MS^2-mu^2)^2))-(12 (8 MS^2 mu^4-3 mu^6) Xt^2)/(MS^4 (MS^2-mu^2)^2)+(6 (2 MS^2 mu^4-mu^6) Xt^4)/(MS^6 (MS^2-mu^2)^2)+((24 MS^4)/(MS^2-mu^2)^2-(48 MS^2 mu^2)/(MS^2-mu^2)^2-(48 mu^4)/(MS^2-mu^2)^2) Log[MS^2/MR^2]) Log[mu^2/MR^2]-(12 (MS^4-2 MS^2 mu^2-2 mu^4) Log[mu^2/MR^2]^2)/(MS^2-mu^2)^2+(-((24 MS^4)/(MS^2-mu^2)^2)+(48 MS^2 mu^2)/(MS^2-mu^2)^2+(48 mu^4)/(MS^2-mu^2)^2) PolyLog[2,1-MS^2/mu^2]+CosBeta^2 (-39+360 Kdlat2+(-(72/MS^2)-(384 Kdlat2)/MS^2) Xt Yt+(24/MS^4+(96 Kdlat2)/MS^4) Xt^3 Yt+(-(18/MS^2)-(96 Kdlat2)/MS^2) Yt^2+Xt^4 (-(75/(2 MS^4))+(-(19/(2 MS^6))-(48 Kdlat2)/MS^6) Yt^2)+Xt^2 (90/MS^2-(144 Kdlat2)/MS^2+(28/MS^4+(144 Kdlat2)/MS^4) Yt^2)+(54+(72 Xt Yt)/MS^2-(12 Xt^3 Yt)/MS^4+(18 Yt^2)/MS^2+Xt^4 (36/MS^4+(3 Yt^2)/MS^6)+Xt^2 (-(144/MS^2)-(18 Yt^2)/MS^4)) Log[MS^2/MR^2]+6 Inactive[Zeta][2]))
);

(* Calculated based on [Fleischer et al., 1999, arXiv:hep-ph/9803493v5 ] & [Bednyakov et al., 2003, arXiv:hep-ph/0210258v2]*)

deltaytas2Deg=(
-(2075/54)+(356 Xt)/(27 MS)+(16 Xt^2)/(9 MS^2)+(-(670/27)-(208 Xt)/(27 MS)) Log[MS^2/MR^2]+2/9 Log[MS^2/MR^2]^2
);



(* ::Section::RGBColor[1, 0, 0]::Closed:: *)
(*3-loop order corrections*)


(* Calculated, based on the results from  [Harlander et al., 2008, arXiv:0803.0672v2] *)

deltalambdaatas2Deg=(
6980/27-SMConstatas2+(-(880/3)-(6784 Xt)/(3 MS)+(2432 Xt^2)/(3 MS^2)+(2368 Xt^3)/(9 MS^3)) Log[MS^2/MR^2]^2+736 Log[MS^2/MR^2]^3+Log[MS^2/MR^2] (7016/27+(12800 Xt)/(27 MS)-(3584 Xt^2)/(9 MS^2)-(26624 Xt^3)/(27 MS^3)-128 Zeta[3])-(3424 Zeta[3])/3+Xt^3 (57088/(27 MS^3)-(4016 Zeta[3])/(3 MS^3))+Xt^2 (1192/(9 MS^2)-(880 Zeta[3])/MS^2)+Xt (-(52672/(27 MS))+(3392 Zeta[3])/(3 MS))
);



(* ::Subchapter::Closed:: *)
(*Matching conditions*)


(* GeneralSolution implicit *)

LambdaMatchingMSImplicit={
MSlambda->MSyt^2 ((k dlat+k^2 dlatas MSg3^2+k^3 dlatas2 MSg3^4+k^4 dlatas3 MSg3^6) MSyt^2+(k^2 dlat2+k^3 dlat2as MSg3^2+k^4 dlat2as2 MSg3^4) MSyt^4+(k^3 dlat3+k^4 dlat3as MSg3^2) MSyt^6+k^4 dlat4 MSyt^8)
};

MatchingCondsMStoDRImplicit={
MSg3->g3 (1+g3^2 k Dg3as+g3^4 k^2 Dg3as2+g3^6 k^3 Dg3as3+yt^2 (g3^2 k^2 Dg3atas+g3^4 k^3 Dg3atas2) Sin[beta]^2+g3^2 k^3 yt^4 Dg3at2as Sin[beta]^4),
MSyt->yt Sin[beta] (1+g3^2 k Dytas+g3^4 k^2 Dytas2+g3^6 k^3 Dytas3+yt^2 (k Dytat+g3^2 k^2 Dytatas+g3^4 k^3 Dytatas2) Sin[beta]^2+yt^4 (k^2 Dytat2+g3^2 k^3 Dytat2as) Sin[beta]^4+k^3 yt^6 Dytat3 Sin[beta]^6),
MSv->Csc[beta] (1+yt^2 (k Dvat+g3^2 k^2 Dvatas+g3^4 k^3 Dvatas2) Sin[beta]^2+yt^4 (k^2 Dvat2+g3^2 k^3 Dvat2as) Sin[beta]^4+k^3 yt^6 Dvat3 Sin[beta]^6) vu
};

MatchingCondsDRtoMSImplicit={
g3->MSg3 (1+k dg3as MSg3^2+k^2 dg3as2 MSg3^4+k^3 dg3as3 MSg3^6+(k^2 dg3atas MSg3^2+k^3 dg3atas2 MSg3^4) MSyt^2+k^3 dg3at2as MSg3^2 MSyt^4),
yt->Csc[beta] MSyt (1+k dytas MSg3^2+k^2 dytas2 MSg3^4+k^3 dytas3 MSg3^6+(k dytat+k^2 dytatas MSg3^2+k^3 dytatas2 MSg3^4) MSyt^2+(k^2 dytat2+k^3 dytat2as MSg3^2) MSyt^4+k^3 dytat3 MSyt^6),
vu->MSv (1+(k dvat+k^2 dvatas MSg3^2) MSyt^2+k^2 dvat2 MSyt^4+k^3 dvat3 MSyt^6) Sin[beta]
};



(* Explicit *)

ExplicitizeCorrections={
dg3as->deltag3as,
dytas->deltaytas,
dytas2->deltaytas2,
dytat->deltaytat,
dlat->deltalambdaat,
dlatas->deltalambdaatas,
dlat2->deltalambdaat2,
dvat->deltavevat
(*
,
dlatas2->deltalambdaatas2,
*)
};

ExplicitizeCorrectionsDeg={
dg3as->deltag3asDeg,
dytas->deltaytasDeg,
dytas2->deltaytas2Deg,
dytat->deltaytatDeg,
dlat->deltalambdaatDeg,
dlatas->deltalambdaatasDeg,
(*dlatas2->deltalambdaatas2Deg,*)
dlat2->deltalambdaat2Deg,
dvat->deltavevatDeg
};



(* Conversions between the MS and the DR threshold corrections *)

LambdaCorrectionsFromMStoDR={

dlat->Dlat,
dlatas->-4 Dytas Dlat+Dlatas,
dlatas2->-(6 Dytas^2+4 Dytas2) Dlat-(2 Dg3as+4 Dytas) (-4 Dytas Dlat+Dlatas)+Dlatas2,
dlatas3->7 Dg3as^2 Dlatas-2 Dg3as2 Dlatas-4 Dg3as Dlatas2+Dlatas3-28 Dg3as^2 Dlat Dytas+8 Dg3as2 Dlat Dytas+16 Dg3as Dlatas Dytas-4 Dlatas2 Dytas-40 Dg3as Dlat Dytas^2+10 Dlatas Dytas^2-20 Dlat Dytas^3+16 Dg3as Dlat Dytas2-4 Dlatas Dytas2+20 Dlat Dytas Dytas2-4 Dlat Dytas3,

dlat2->Dlat2-4 Dlat Dytat,
dlat2as->Dlat2as-4 Dlatas Dytat+4 Dlat Dytas Dytat-6 Dytas (Dlat2-4 Dlat Dytat)-4 Dlat Dytatas,

dlat3->Dlat3-6 Dlat Dytat^2-6 Dytat (Dlat2-4 Dlat Dytat)-4 Dlat Dytat2,
dlat3as->Dlat3as-8 Dlat3 Dytas-6 Dlat2as Dytat+54 Dlat2 Dytas Dytat+18 Dlatas Dytat^2-180 Dlat Dytas Dytat^2-4 Dlatas Dytat2+36 Dlat Dytas Dytat2-4 Dlat Dytat2as-6 Dlat2 Dytatas+36 Dlat Dytat Dytatas

};

LambdaCorrectionsFromDRtoMS={

Dlat->dlat,
Dlatas->dlatas-4 dlat dytas,
Dlatas2->dlatas2+dlatas (-2 dg3as-4 dytas)+dlat (8 dg3as dytas+10 dytas^2-4 dytas2),
Dlatas3->dlatas3+dlatas2 (-4 dg3as-4 dytas)+dlatas (7 dg3as^2-2 dg3as2+16 dg3as dytas+10 dytas^2-4 dytas2)+dlat (-28 dg3as^2 dytas+8 dg3as2 dytas-40 dg3as dytas^2-20 dytas^3+16 dg3as dytas2+20 dytas dytas2-4 dytas3),

Dlat2->dlat2-4 dlat dytat,
Dlat2as->dlat2as-6 dlat2 dytas-4 dlatas dytat+dlat (28 dytas dytat-4 dytatas),
Dlat2as2->-2 dg3as dlat2as+dlat2as2+12 dg3as dlat2 dytas-6 dlat2as dytas+21 dlat2 dytas^2-6 dlat2 dytas2-4 dlatas2 dytat+dlatas (-2 dg3atas+8 dg3as dytat+28 dytas dytat-4 dytatas)+dlat (8 dg3atas dytas-56 dg3as dytas dytat-112 dytas^2 dytat+28 dytas2 dytat+8 dg3as dytatas+28 dytas dytatas-4 dytatas2),

Dlat3->dlat3-6 dlat2 dytat+dlat (18 dytat^2-4 dytat2),
Dlat3as->dlat3as-8 dlat3 dytas-6 dlat2as dytat+54 dlat2 dytas dytat+dlatas (18 dytat^2-4 dytat2)-6 dlat2 dytatas+dlat (-180 dytas dytat^2+36 dytas dytat2-4 dytat2as+36 dytat dytatas)

};

CorrectionsFromMStoDR={

dg3as->-Dg3as,
dg3as2->3 Dg3as^2-Dg3as2,
dg3as3->0,

dg3atas->-Dg3atas,
dg3atas2->6 Dg3as Dg3atas-Dg3atas2+2 Dg3atas Dytas,
dg3at2as->-Dg3at2as+2 Dg3atas Dytat,

dytas->-Dytas,
dytas2->2 Dg3as Dytas+Dytas^2-Dytas2,
dytas3->-7 Dg3as^2 Dytas+2 Dg3as2 Dytas-4 Dg3as Dytas^2-Dytas^3+4 Dg3as Dytas2+2 Dytas Dytas2-Dytas3,

dytat->-Dytat,
dytatas->4 Dytas Dytat-Dytatas,
dytatas2->2 Dg3atas Dytas-8 Dg3as Dytas Dytat-10 Dytas^2 Dytat+4 Dytas2 Dytat+2 Dg3as Dytatas+4 Dytas Dytatas-Dytatas2,

dytat2->3 Dytat^2-Dytat2,
dytat2as->-21 Dytas Dytat^2+6 Dytas Dytat2-Dytat2as+6 Dytat Dytatas,

dytat3->-12 Dytat^3+8 Dytat Dytat2-Dytat3,

dvat->-Dvat,
dvatas->-Dvatas+2 Dvat Dytas,

dvat2->Dvat^2-Dvat2+2 Dvat Dytat,
dvat2as->2 Dg3as Dvatas-4 Dg3as Dvat Dytas+2 Dvatas Dytas-3 Dvat Dytas^2+2 Dvat Dytas2

};

CorrectionsFromDRtoMS={

Dg3as->-dg3as,
Dg3as2->3 dg3as^2-dg3as2,
Dg3as3->-11 dg3as^3+6 dg3as dg3as2,

Dg3atas->-dg3atas,
Dg3atas2->6 dg3as dg3atas-dg3atas2+2 dg3atas dytas,
Dg3at2as->-dg3at2as+2 dg3atas dytat,

Dytas->-dytas,
Dytas2->2 dg3as dytas+dytas^2-dytas2,
Dytas3->-7 dg3as^2 dytas+2 dg3as2 dytas-4 dg3as dytas^2-dytas^3+4 dg3as dytas2+2 dytas dytas2-dytas3,

Dytat->-dytat,
Dytatas->4 dytas dytat-dytatas,
Dytatas2->2 dg3atas dytas-8 dg3as dytas dytat-10 dytas^2 dytat+4 dytas2 dytat+2 dg3as dytatas+4 dytas dytatas-dytatas2,

Dytat2->3 dytat^2-dytat2,
Dytat2as->-21 dytas dytat^2+6 dytas dytat2-dytat2as+6 dytat dytatas,

Dytat3->-12 dytat^3+8 dytat dytat2-dytat3,

Dvat->-dvat,
Dvatas->-dvatas+2 dvat dytas,

Dvat2->dvat^2-dvat2+2 dvat dytat,
Dvat2as->2 dg3as dvatas-4 dg3as dvat dytas+2 dvatas dytas-3 dvat dytas^2+2 dvat dytas2

};



(* Some replacement rules to change variables *)

FromkToPi={k->1/(4Pi)^2};
FromVuToMt={vu->mt/yt};
FromMtToVu={mt-> vu yt};
FromVToMt={MSv->MSmt/MSyt};
FromMtToV={MSmt->MSyt MSv};
FromCouplingToAlpha={g3->Sqrt[a3/Sqrt[k]],yt->Csc[beta] Sqrt[at/Sqrt[k]]};
FromAlphaToCoupling={a3->g3^2 Sqrt[k],at->Sqrt[k] yt^2 Sin[beta]^2};
FromCouplingToAlphaMS={MSg3->Sqrt[MSa3/Sqrt[k]],MSyt->Sqrt[MSat/Sqrt[k]]};
FromAlphaToCouplingMS={MSa3->Sqrt[k] MSg3^2,MSat->Sqrt[k] MSyt^2};
FromBetaVToBetaVTilde={betav3Loop->MSv betatildev3Loop};
FromBetaToTanBeta={beta->ArcTan[TanBeta]};
FromZetaToZ={Zeta[3]->z3,Inactive[Zeta][2]->z2,Zeta[5]->z5};
FromZToZeta={z2->Inactive[Zeta][2],z3->Zeta[3],z5->Inactive[Zeta][5]};
FromMassDiffToDm={(mQ3^2-mU3^2)->DmQ3U3,(-mQ3^2+mU3^2)->-DmQ3U3,(m3^2-mQ3^2)->Dm3Q3,(-m3^2+mQ3^2)->-Dm3Q3,(m3^2-mU3^2)->Dm3U3,(-m3^2+mU3^2)->-Dm3U3,(msq^2-mQ3^2)->DmsqQ3,(-msq^2+mQ3^2)->-DmsqQ3,(msq^2-mU3^2)->DmsqU3,(-msq^2+mU3^2)->-DmsqU3};
FromDmToMassDiff={Dm3Q3->(m3^2-mQ3^2),Dm3U3->(m3^2-mU3^2),DmQ3U3->(mQ3^2-mU3^2),DmsqQ3->(msq^2-mQ3^2),DmsqU3->(msq^2-mU3^2)};
FromKtoNumber={Kdlat2->N@Re@(-1/Sqrt[3] Inactive[Integrate][Log[2 Cos[equis]],{equis,0,\[Pi]/6}])};
FromLinearMassesToQuadratic={mQ3->(mmQ3^(1/2)),mU3->(mmU3^(1/2)),msq->(mmsq^(1/2)),mt->(mmt^(1/2))};
FromQuadraticMassesToLinear={mmQ3->mQ3^2,mmU3->mU3^2,mmsq->msq^2,mmt->mt^2};
FromFuncBetaToBeta={TanBeta->Tan[beta],SinBeta->Sin[beta],CosBeta->Cos[beta],CscBeta->Csc[beta]};



(* Replacement rule to change notation *)

NotationConversion={
k->\[Kappa],
MSyt->Subscript[\!\(\*OverscriptBox[\(y\), \(_\)]\), t],
MSg3->Subscript[\!\(\*OverscriptBox[\(g\), \(_\)]\), 3],
dlat->Subscript[\!\(\*OverscriptBox[\(\[Delta]\[Lambda]\), \(_\)]\), Subscript[\[Alpha], t]],
dlatas->Subscript[\!\(\*OverscriptBox[\(\[Delta]\[Lambda]\), \(_\)]\), Subscript[\[Alpha], 3] Subscript[\[Alpha], t]],
dlatas2->Subscript[\!\(\*OverscriptBox[\(\[Delta]\[Lambda]\), \(_\)]\), \!\(
\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\ 
\*SubscriptBox[\(\[Alpha]\), \(t\)]\)],
dlatas3->Subscript[\!\(\*OverscriptBox[\(\[Delta]\[Lambda]\), \(_\)]\), \!\(
\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(3\)]\ 
\*SubscriptBox[\(\[Alpha]\), \(t\)]\)],
dlat2->Subscript[\!\(\*OverscriptBox[\(\[Delta]\[Lambda]\), \(_\)]\), \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
dlat2as->Subscript[\!\(\*OverscriptBox[\(\[Delta]\[Lambda]\), \(_\)]\), Subscript[\[Alpha], 3] \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
dlat2as2->Subscript[\!\(\*OverscriptBox[\(\[Delta]\[Lambda]\), \(_\)]\), \!\(
\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\ 
\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
dlat3->Subscript[\!\(\*OverscriptBox[\(\[Delta]\[Lambda]\), \(_\)]\), \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(3\)]\)],
dlat3as->Subscript[\!\(\*OverscriptBox[\(\[Delta]\[Lambda]\), \(_\)]\), Subscript[\[Alpha], 3] \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(3\)]\)],
dlat4->Subscript[\!\(\*OverscriptBox[\(\[Delta]\[Lambda]\), \(_\)]\), \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(4\)]\)],
dg3as->Subscript[\!\(\*OverscriptBox[\(\[Delta]g\), \(_\)]\), 3,Subscript[\[Alpha], 3]],
dg3as2->Subscript[\!\(\*OverscriptBox[\(\[Delta]g\), \(_\)]\), 3,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\)],
dg3as3->Subscript[\!\(\*OverscriptBox[\(\[Delta]g\), \(_\)]\), 3,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(3\)]\)],
dg3atas->Subscript[\!\(\*OverscriptBox[\(\[Delta]g\), \(_\)]\), 3,Subscript[\[Alpha], 3] Subscript[\[Alpha], t]],
dg3atas2->Subscript[\!\(\*OverscriptBox[\(\[Delta]g\), \(_\)]\), 3,\!\(
\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\ 
\*SubscriptBox[\(\[Alpha]\), \(t\)]\)],
dg3at2as->Subscript[\!\(\*OverscriptBox[\(\[Delta]g\), \(_\)]\), 3,Subscript[\[Alpha], 3] \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
dytas->Subscript[\!\(\*OverscriptBox[\(\[Delta]y\), \(_\)]\), t,Subscript[\[Alpha], 3]],
dytas2->Subscript[\!\(\*OverscriptBox[\(\[Delta]y\), \(_\)]\), t,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\)],
dytas3->Subscript[\!\(\*OverscriptBox[\(\[Delta]y\), \(_\)]\), t,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(3\)]\)],
dytat->Subscript[\!\(\*OverscriptBox[\(\[Delta]y\), \(_\)]\), t,Subscript[\[Alpha], t]],
dytatas->Subscript[\!\(\*OverscriptBox[\(\[Delta]y\), \(_\)]\), t,Subscript[\[Alpha], 3] Subscript[\[Alpha], t]],
dytatas2->Subscript[\!\(\*OverscriptBox[\(\[Delta]y\), \(_\)]\), t,\!\(
\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\ 
\*SubscriptBox[\(\[Alpha]\), \(t\)]\)],
dytat2->Subscript[\!\(\*OverscriptBox[\(\[Delta]y\), \(_\)]\), t,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
dytat2as->Subscript[\!\(\*OverscriptBox[\(\[Delta]y\), \(_\)]\), t,Subscript[\[Alpha], 3] \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
dytat3->Subscript[\!\(\*OverscriptBox[\(\[Delta]y\), \(_\)]\), t,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(3\)]\)],
dvat->Subscript[\!\(\*OverscriptBox[\(\[Delta]v\), \(_\)]\), Subscript[\[Alpha], t]],
dvatas->Subscript[\!\(\*OverscriptBox[\(\[Delta]v\), \(_\)]\), Subscript[\[Alpha], 3] Subscript[\[Alpha], t]],
dvat2->Subscript[\!\(\*OverscriptBox[\(\[Delta]v\), \(_\)]\), \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
dvat3->Subscript[\!\(\*OverscriptBox[\(\[Delta]v\), \(_\)]\), \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(3\)]\)],
yt->Subscript[y, t],
g3->Subscript[g, 3],
vu->Subscript[v, u],
beta->\[Beta],
Dlat->Subscript[\[CapitalDelta]\[Lambda], Subscript[\[Alpha], t]],
Dlatas->Subscript[\[CapitalDelta]\[Lambda], Subscript[\[Alpha], 3] Subscript[\[Alpha], t]],
Dlatas2->Subscript[\[CapitalDelta]\[Lambda], \!\(
\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\ 
\*SubscriptBox[\(\[Alpha]\), \(t\)]\)],
Dlatas3->Subscript[\[CapitalDelta]\[Lambda], \!\(
\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(3\)]\ 
\*SubscriptBox[\(\[Alpha]\), \(t\)]\)],
Dlat2->Subscript[\[CapitalDelta]\[Lambda], \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
Dlat2as->Subscript[\[CapitalDelta]\[Lambda], Subscript[\[Alpha], 3] \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
Dlat3->Subscript[\[CapitalDelta]\[Lambda], \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(3\)]\)],
Dlat3as3->Subscript[\[CapitalDelta]\[Lambda], \!\(
\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(3\)]\ 
\*SubscriptBox[\(\[Alpha]\), \(t\)]\)],
Dlat3as->Subscript[\[CapitalDelta]\[Lambda], Subscript[\[Alpha], 3] \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(3\)]\)],
Dg3as->Subscript[\[CapitalDelta]g, 3,Subscript[\[Alpha], 3]],
Dg3as2->Subscript[\[CapitalDelta]g, 3,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\)],
Dg3as3->Subscript[\[CapitalDelta]g, 3,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(3\)]\)],
Dg3atas->Subscript[\[CapitalDelta]g, 3,Subscript[\[Alpha], 3] Subscript[\[Alpha], t]],
Dg3atas2->Subscript[\[CapitalDelta]g, 3,\!\(
\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\ 
\*SubscriptBox[\(\[Alpha]\), \(t\)]\)],
Dg3at2as->Subscript[\[CapitalDelta]g, 3,Subscript[\[Alpha], 3] \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
Dytas->Subscript[\[CapitalDelta]y, t,Subscript[\[Alpha], 3]],
Dytas2->Subscript[\[CapitalDelta]y, t,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\)],
Dytas3->Subscript[\[CapitalDelta]y, t,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(3\)]\)],
Dytat->Subscript[\[CapitalDelta]y, t,Subscript[\[Alpha], t]],
Dytatas->Subscript[\[CapitalDelta]y, t,Subscript[\[Alpha], 3] Subscript[\[Alpha], t]],
Dytatas2->Subscript[\[CapitalDelta]y, t,\!\(
\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\ 
\*SubscriptBox[\(\[Alpha]\), \(t\)]\)],
Dytat2->Subscript[\[CapitalDelta]y, t,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
Dytat2as->Subscript[\[CapitalDelta]y, t,Subscript[\[Alpha], 3] \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
Dytat3->Subscript[\[CapitalDelta]y, t,\!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(3\)]\)],
Dvat->Subscript[\[CapitalDelta]v, Subscript[\[Alpha], t]],
Dvatas->Subscript[\[CapitalDelta]v, Subscript[\[Alpha], 3] Subscript[\[Alpha], t]],
Dvat2->Subscript[\[CapitalDelta]v, \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
Dvat3->Subscript[\[CapitalDelta]v, \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(3\)]\)],
mu->\[Mu],
mA->Subscript[m, A],
m3->Subscript[m, 3],
mQ3->Subscript[m, Subscript[Q, 3]],
mU3->Subscript[m, Subscript[U, 3]],
msq->Subscript[m, sq],
mst1->Subscript[m, Subscript[t, 1]],
mst2->Subscript[m, Subscript[t, 2]]
Xt->Subscript[X, t],
Yt->Subscript[Y, t],
MS->Subscript[MR, S],
MR->MR,
mt->Subscript[m, t],
MZ->Subscript[MR, Z],
at->Subscript[\[Alpha], t],
a3->Subscript[\[Alpha], 3],
SMConstatas2->Subscript[\!\(\*OverscriptBox[\(c\), \(_\)]\), Subscript[\[Alpha], t] Subscript[\[Alpha], 3]],
SMConstatas3->Subscript[\!\(\*OverscriptBox[\(c\), \(_\)]\), Subscript[\[Alpha], t] \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(3\), \(2\)]\)],
SMConstat2->Subscript[\!\(\*OverscriptBox[\(c\), \(_\)]\), \!\(\*SubsuperscriptBox[\(\[Alpha]\), \(t\), \(2\)]\)],
DmQ3U3->Subscript[\[CapitalDelta]m,Subscript[Q, 3] Subscript[U, 3]],
Dm3Q3->Subscript[\[CapitalDelta]m,3Subscript[Q, 3]],
Dm3U3->Subscript[\[CapitalDelta]m,3Subscript[U, 3]],
DmsqQ3->Subscript[\[CapitalDelta]m,Subscript[sqQ, 3]],
DmsqU3->Subscript[\[CapitalDelta]m,Subscript[sqU, 3]],
z2->\[Zeta][2],
z3->\[Zeta][3]
};



ParameterValues[M3_,MQ3_,MU3_,Msq_,xt_,MRenorm_]:={
MZ->91,
yt->1,
g3->1.125,
vu->175(**Sin[\[Pi]/3]*),
beta->\[Pi]/3,
at->Sin[\[Pi]/3]^2/(4\[Pi]),
a3->1.125^2/(4\[Pi]),
mt->175,
k->1/(4\[Pi])^2,
MR->MRenorm,
m3->M3,
mu->5000,
mA->7500,
mQ3->MQ3,
mU3->MU3,
msq->Msq,
Xt->xt,
Yt->75000,
dlatas2->0,
SMConstatas2->(SMConstatas2/.SM3LoopConst),
z2->Zeta[2],
z3->Zeta[3],
Inactive[Zeta][2]->Zeta[2],
Inactive[PolyLog][4,1/2]->PolyLog[4,1/2]
};



(* ::Chapter::RGBColor[0., 0., 1.]::Closed:: *)
(*Pole Mass in the MSSM*)


(* ::Subchapter::Closed:: *)
(*Mass expansions*)


(* ::Section::RGBColor[1, 0, 0]::Closed:: *)
(*Functions that return complete expansions*)


(*
Expansion of the scale (in)-dependent Higgs boson mass in powers of Log[MR^2/MSmt^2] using SM MS parameters
*)

Mx[Subindex_,GeneralSolution,MS][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

Sum[
k^n Sum[
(K[Subindex][n,i] Log[MR^2 MSyt^-2 MSv^-2]^i/.LambdaMatchingMSImplicit),
{i,0,n}
],
{n,0,LoopOrder}
],

{MSv,0,2},
{MSyt,0,2(AlphaTOrder+1)},
{k,0,LoopOrder},
{MSg3,0,2(LoopOrder-AlphaTOrder)},
{Log[MR^2 MSyt^-2 MSv^-2],0,LoopOrder},
{SMTopConstas2,0,4}

];

(*
MS Degenerate limit
*)

Mx[Subindex_,Degenerate,MS][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

Mx[Subindex,GeneralSolution,MS][LoopOrder,AlphaTOrder]/.ExplicitizeCorrectionsDeg,

{MSv,0,2},
{MSyt,0,2(AlphaTOrder+1)},
{k,0,LoopOrder},
{MSg3,0,2(LoopOrder-AlphaTOrder)},
{Log[MR^2 MSyt^-2 MSv^-2],0,LoopOrder},
{Ndlat2,0,2},
{CscBeta,0,8},
{SinBeta,0,8},
{CosBeta,0,8},
{Log[MS^2 MR^-2],0,LoopOrder},
{Xt,0,10},
{Yt,0,10},
{Inactive[Zeta][2],0,4},
{SMTopConstas2,0,4}

];

(*
MS Non-Degenerate limit
*)

Mx[Subindex_,LimitingCase_,MS][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

Mx[Subindex,GeneralSolution,MS][LoopOrder,AlphaTOrder]/.ExplicitizeCorrections,

{MSv,0,2},
{MSyt,0,2(AlphaTOrder+1)},
{k,0,LoopOrder},
{MSg3,0,2(LoopOrder-AlphaTOrder)},
{Log[MR^2 MSyt^-2 MSv^-2],0,LoopOrder},
{Ndlat2,0,2},
{CscBeta,0,8},
{SinBeta,0,8},
{CosBeta,0,8},
{Log[m3^2 MR^-2],0,4},
{Log[mu^2 MR^-2],0,4},
{Log[mA^2 MR^-2],0,4},
{Log[mQ3^2 MR^-2],0,4},
{Log[mU3^2 MR^-2],0,4},
{Log[msq^2 MR^-2],0,4},
{PolyLog[2,1-m3^2/mQ3^2],0,2},
{PolyLog[2,1-m3^2/mU3^2],0,2},
{PolyLog[2,1-m3^2/msq^2],0,2},
{PolyLog[2,1-mQ3^2/mu^2],0,2},
{PolyLog[2,1-mU3^2/mu^2],0,2},
{PolyLog[2,1-msq^2/mQ3^2],0,2},
{PolyLog[2,1-msq^2/mU3^2],0,2},
{PolyLog[2,1-mQ3^2/mU3^2],0,2},
{Xt,0,10},
{Yt,0,10},
{Inactive[Zeta][2],0,4},
{SMTopConstas2,0,4},
{Ndlat2,0,1}

];

(*
Expansion of the scale (in)-dependent Higgs boson mass in powers of Log[MR^2/mt^2] using MSSM DR parameters
*)

Mx[Subindex_,GeneralSolution,DR][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

Normal@Series[

Sum[
k^n Sum[
(K[Subindex][n,i]Log[MR^2 MSyt^-2 MSv^-2]^i/.LambdaMatchingMSImplicit/.MatchingCondsMStoDRImplicit/.CorrectionsFromDRtoMS),
{i,0,n}
],
{n,0,LoopOrder}
],

{k,0,LoopOrder},
{vu,0,2},
{yt,0,2 (AlphaTOrder+1)},
{Sin[beta],0,2 AlphaTOrder},
{g3,0,2(LoopOrder-AlphaTOrder)},
{Log[MR^2 yt^-2 vu^-2],0,LoopOrder},
{SMTopConstas2,0,4}

],

{vu,0,2},
{yt,0,2 (AlphaTOrder+1)},
{Sin[beta],0,2 AlphaTOrder},
{k,0,LoopOrder},
{g3,0,2(LoopOrder-AlphaTOrder)},
{Log[MR^2 yt^-2 vu^-2],0,LoopOrder},
{dlat,0,4},
{dlatas,0,4},
{dlatas2,0,4},
{dlat2,0,4},
{dlat2as,0,4},
{dlat3,0,4},
{SMTopConstas2,0,4}

];

(*
Function that converts the DR expasion back to MS intended to verify the correctness of the DR expansion
*)

Mx[Subindex_,FromDRToMS,Scheme_:DR][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

Mx[Subindex,GeneralSolution,DR][LoopOrder,AlphaTOrder]/.MatchingCondsDRtoMSImplicit,

{MSv,0,2},
{MSyt,0,2(AlphaTOrder+1)},
{k,0,LoopOrder},
{MSg3,0,2(LoopOrder-AlphaTOrder)},
{Log[MR^2 MSyt^-2 MSv^-2],0,LoopOrder},
{dlat,0,4},
{dlatas,0,4},
{dlatas2,0,4}

];

(*
DR Degenerate limit
*)

Mx[Subindex_,Degenerate,Scheme_:DR][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

(Mx[Subindex,GeneralSolution,Scheme][LoopOrder,AlphaTOrder]/.ExplicitizeCorrectionsDeg),

{vu,0,2},
{yt,0,2 (AlphaTOrder+1)},
{Sin[beta],0,2 AlphaTOrder},
{k,0,LoopOrder},
{g3,0,2(LoopOrder-AlphaTOrder)},
{Log[MR^2 yt^-2 vu^-2],0,LoopOrder},
{CscBeta,0,8},
{SinBeta,0,8},
{CosBeta,0,8},
{Log[MS^2 MR^-2],0,LoopOrder},
{Xt,0,10},
{Yt,0,10},
{Inactive[Zeta][2],0,4},
{SMTopConstas2,0,4}

];

(*
DR Non-Degenerate limit
*)

Mx[Subindex_,LimitingCase_,Scheme_:DR][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

(Mx[Subindex,GeneralSolution,Scheme][LoopOrder,AlphaTOrder]/.ExplicitizeCorrections/.SM2LoopConst),

{vu,0,2},
{yt,0,2 (AlphaTOrder+1)},
{Sin[beta],0,2 AlphaTOrder},
{k,0,LoopOrder},
{g3,0,2(LoopOrder-AlphaTOrder)},
{Log[MR^2 yt^-2 vu^-2],0,LoopOrder},
{Ndlat2,0,2},
{CscBeta,0,8},
{SinBeta,0,8},
{CosBeta,0,8},
{Log[m3^2 MR^-2],0,4},
{Log[mu^2 MR^-2],0,4},
{Log[mA^2 MR^-2],0,4},
{Log[mQ3^2 MR^-2],0,4},
{Log[mU3^2 MR^-2],0,4},
{Log[msq^2 MR^-2],0,4},
{PolyLog[2,1-m3^2/mQ3^2],0,2},
{PolyLog[2,1-m3^2/mU3^2],0,2},
{PolyLog[2,1-m3^2/msq^2],0,2},
{PolyLog[2,1-mQ3^2/mu^2],0,2},
{PolyLog[2,1-mU3^2/mu^2],0,2},
{PolyLog[2,1-msq^2/mQ3^2],0,2},
{PolyLog[2,1-msq^2/mU3^2],0,2},
{PolyLog[2,1-mQ3^2/mU3^2],0,2},
{Xt,0,10},
{Yt,0,10},
{Inactive[Zeta][2],0,4},
{SMTopConstas2,0,4},
{Ndlat2,0,1}

]/.FromFuncBetaToBeta;



(*
Expansion of the scale (in)-dependent Higgs boson mass in powers of Log[MR^2/mt^2] using MSSM DR parameters
*)

M[Subindex_,GeneralSolution][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

Mx[Subindex,GeneralSolution,DR][LoopOrder,AlphaTOrder]/.LambdaCorrectionsFromMStoDR/.CorrectionsFromMStoDR/.FromVuToMt/.FromCouplingToAlpha/.FromZetaToZ,

{mt,0,2},
{at,0,AlphaTOrder},
{k,0,LoopOrder},
{a3,0,LoopOrder-AlphaTOrder},
{Log[MR^2 mt^-2],0,LoopOrder},
{Dlat,0,4},
{Dlatas,0,4},
{Dlatas2,0,4},
{SMTopConstas2,0,4}

];

M[Subindex_,Degenerate][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

Mx[Subindex,Degenerate,DR][LoopOrder,AlphaTOrder]/.FromVuToMt/.FromCouplingToAlpha/.FromZetaToZ/.SM2LoopConst,


{mt,0,2},
{at,0,AlphaTOrder},
{k,0,LoopOrder},
{a3,0,LoopOrder-AlphaTOrder},
{Log[MR^2 mt^-2],0,LoopOrder},
{CscBeta,0,8},
{SinBeta,0,8},
{CosBeta,0,8},
{Log[MS^2 MR^-2],0,LoopOrder},
{Xt,0,10},
{Yt,0,8},
{z2,0,4},
{z3,0,4},
{z5,0,4},
{SMTopConstas2,0,4}

]/.FromFuncBetaToBeta;

M[Subindex_,LimitingCase_][LoopOrder_Integer,AlphaTOrder_Integer:1]:=Normal@Series[

Mx[Subindex,GeneralSolution,DR][LoopOrder,AlphaTOrder]/.FromVuToMt/.FromCouplingToAlpha/.FromZetaToZ/.SM2LoopConst/.ExplicitizeCorrections,


{mt,0,2},
{at,0,AlphaTOrder},
{k,0,LoopOrder},
{a3,0,LoopOrder-AlphaTOrder},
{Log[MR^2 mt^-2],0,LoopOrder},
{Ndlat2,0,2},
{CscBeta,0,8},
{SinBeta,0,8},
{CosBeta,0,8},
{Log[m3^2 MR^-2],0,4},
{Log[mu^2 MR^-2],0,4},
{Log[mA^2 MR^-2],0,4},
{Log[mQ3^2 MR^-2],0,4},
{Log[mU3^2 MR^-2],0,4},
{Log[msq^2 MR^-2],0,4},
{PolyLog[2,1-m3^2/mQ3^2],0,2},
{PolyLog[2,1-m3^2/mU3^2],0,2},
{PolyLog[2,1-m3^2/msq^2],0,2},
{PolyLog[2,1-mQ3^2/mu^2],0,2},
{PolyLog[2,1-mU3^2/mu^2],0,2},
{PolyLog[2,1-msq^2/mQ3^2],0,2},
{PolyLog[2,1-msq^2/mU3^2],0,2},
{PolyLog[2,1-mQ3^2/mU3^2],0,2},
{Xt,0,10},
{Yt,0,10},
{z2,0,4},
{z3,0,4},
{z5,0,4},
{SMTopConstas2,0,4}

]/.FromFuncBetaToBeta;



(* ::Section::RGBColor[1, 0, 0]::Closed:: *)
(*Functions that return a single correction*)


(*
Functions that returns the coefficient of the desired order and log power
*)

deltamh2[Alpha3Order_Integer,LogOrder_Integer]/;LogOrder<=(Alpha3Order+1):=
Coefficient[
Coefficient[
Coefficient[
k^(-(Alpha3Order))mt^-2 Mh2[EFT],
at,1
],
a3,Alpha3Order
],
Log[MR^2/mt^2],LogOrder
];

LogCoefficient[Alpha3Order_Integer,LogOrder_Integer,m3LogOrder_,mQ3LogOrder_,mU3LogOrder_,msqLogOrder_]/;LogOrder<=(Alpha3Order+1):=
k^(-(Alpha3Order+1)/2)mt^-2 Log[MR^2/mt^2]^LogOrder Log[m3^2/MR^2]^m3LogOrder Log[mQ3^2/MR^2]^mQ3LogOrder Log[mU3^2/MR^2]^mU3LogOrder Log[msq^2/MR^2]^msqLogOrder Coefficient[
Coefficient[
Coefficient[
Coefficient[
Coefficient[
Coefficient[
Coefficient[
Mh2[EFT],
at,1
],
a3,Alpha3Order
],
Log[MR^2/mt^2],LogOrder
],
Log[m3^2/MR^2],m3LogOrder
],
Log[mQ3^2/MR^2],mQ3LogOrder
],
Log[mU3^2/MR^2],mU3LogOrder
],
Log[msq^2/MR^2],msqLogOrder
];

LogCoefficient[Alpha3Order_Integer,LogOrder_Integer,SUSYMass_,SUSYMassLogOrder_Integer]/;LogOrder<=(Alpha3Order+1):=Normal@Series[

k^(-(Alpha3Order+1)/2)mt^-2 Log[MR^2/mt^2]^LogOrder Log[SUSYMass^2/MR^2]^SUSYMassLogOrder Coefficient[
Coefficient[
Coefficient[
Coefficient[
Mh2[EFT],
at,1
],
a3,Alpha3Order
],
Log[MR^2/mt^2],LogOrder
],
Log[SUSYMass^2/MR^2],SUSYMassLogOrder
],

{Log[MR^2 mt^-2],0,4},
{Log[m3^2 MR^-2],0,4},
{Log[mu^2 MR^-2],0,4},
{Log[mA^2 MR^-2],0,4},
{Log[mQ3^2 MR^-2],0,4},
{Log[mU3^2 MR^-2],0,4},
{Log[msq^2 MR^-2],0,4},
{Xt,0,10},
{z2,0,4},
{z3,0,4},
{z5,0,4}

];



(* This function return the rational functions of each term in the mass expansion at a time *)

R[Alpha3Order_,LogOrder_,m3LogOrder_,mQ3LogOrder_,mU3LogOrder_,msqLogOrder_,PL3Q3_,PL3U3_,PLQ3U3_,PL3sq_,PLsqQ3_,PLsqU3_]:=
Normal@Series[
Simplify@Coefficient[Coefficient[Coefficient[
Coefficient[Coefficient[Coefficient[
LogCoefficient[Alpha3Order,LogOrder,m3LogOrder,mQ3LogOrder,mU3LogOrder,msqLogOrder],
PolyLog[2,1-m3^2/mQ3^2],PL3Q3],PolyLog[2,1-m3^2/mU3^2],PL3U3],PolyLog[2,1-mQ3^2/mU3^2],PLQ3U3],
PolyLog[2,1-m3^2/msq^2],PL3sq],PolyLog[2,1-msq^2/mQ3^2],PLsqQ3],PolyLog[2,1-msq^2/mU3^2],PLsqU3],
{Xt,0,8}
];



(* ::Section::RGBColor[1, 0, 0]::Closed:: *)
(*(d/dMR^2)Mh^2*)


bg3:=Normal@Series[
k*MSSMBetaFunction[1,1]+k^2*MSSMBetaFunction[1,2],
{k,0,2},{yt,0,1},{g3,0,6}
]/.{At->0};

byt:=Normal@Series[
k*MSSMBetaFunction[2,1]+k^2*MSSMBetaFunction[2,2],
{k,0,2},{yt,0,1},{g3,0,6}
]/.{At->0};

bm3:=Normal@Series[
k*MSSMBetaFunction[3,1]+k^2*MSSMBetaFunction[3,2],
{k,0,2},{yt,0,1},{g3,0,6}
]/.{At->0};

bmU32:=Normal@Series[
k*MSSMBetaFunction[4,1]+k^2*MSSMBetaFunction[4,2],
{k,0,2},{yt,0,1},{g3,0,6}
]/.{At->0,mQ3->mQ32^(1/2),mU3->mU32^(1/2)};

bmQ32:=Normal@Series[
k*MSSMBetaFunction[5,1]+k^2*MSSMBetaFunction[5,2],
{k,0,2},{yt,0,1},{g3,0,6}
]/.{At->0,mQ3->mQ32^(1/2),mU3->mU32^(1/2)};

bXt:=Normal@Series[
k*MSSMBetaFunction[6,1]+k^2*MSSMBetaFunction[6,2],
{k,0,2},{yt,0,1},{g3,0,6}
]/.{At->0};



SymbolChange={
MR->MR2^(1/2),
mQ3->mQ32^(1/2),
mU3->mU32^(1/2)
};

DConstants={
Dt[k,MR2]->0,
Dt[beta,MR2]->0,
Dt[vu,MR2]->0,
Dt[SMConstatas2,MR2]->0,
Dt[MZ,MR2]->0
};

FromDtToBeta={
Dt[g3,MR2]->(bg3/(2MR2)),
Dt[yt,MR2]->(byt/(2MR2)),
Dt[m3,MR2]->(bm3/(2MR2)),
Dt[mU32,MR2]->(bmU32/(2MR2)),
Dt[mQ32,MR2]->(bmQ32/(2MR2)),
Dt[Xt,MR2]->(bXt/(2MR2)),
Dt[msq,MR2]->k b1msq/(2MR2)
};

FromDtToDdx={
Dt[dlat,MR2]->((Ddlat+k g3^2 D2dlat+k^2 g3^4 D3dlat)/MR2),
Dt[dlatas,MR2]->((Ddlatas+k g3^2 D2dlatas)/MR2),
Dt[dytas,MR2]->((Ddytas+k g3^2 D2dytas+k^2 g3^4 D3dytas)/MR2),
Dt[dg3as,MR2]->((Ddg3as+k g3^2 D2dg3as)/MR2),
Dt[dytas2,MR2]->((Ddytas2+k g3^2 D2dytas2)/MR2),
Dt[dlatas2,MR2]->((Ddlatas2+k g3^2 D2dlatas2)/MR2)
};



OneOfManyParamChange={
mQ32->mQ3^2,
mU32->mU3^2,
MR2->MR^2
};

OneOfManyLogRules={
Log[m3^2/mU3^2]->Log[m3^2/MR^2]-Log[mU3^2/MR^2],
Log[m3^2/mQ3^2]->Log[m3^2/MR^2]-Log[mQ3^2/MR^2]
};



dMh2dMR2[n_]:=Normal@Series[

Simplify[

Normal@Series[
( vu^2 yt^4 Sin[beta]^2)^-1 MR2*Dt[
Mx[h,NonDegenerate][n]/.SymbolChange,
MR2
]/.DConstants/.FromDtToBeta/.FromDtToDdx/.OneOfManyParamChange/.OneOfManyLogRules,
{k,0,n},
{g3,0,2(n-1)},
{Log[MR2/mt^2],0,3},
{Log[m3^2/MR^2],0,3},
{Log[mQ3^2/MR^2],0,3},
{Log[mU3^2/MR^2],0,3},
{Log[msq^2/MR^2],0,3},
{PolyLog[2,1-m3^2/mQ3^2],0,2},
{PolyLog[2,1-m3^2/mU3^2],0,2},
{PolyLog[2,1-mU3^2/mQ3^2],0,2},
{PolyLog[2,1-mQ3^2/mU3^2],0,2},
{Xt,0,8};
{Ddlatas2,0,1}
]

],

{k,0,n},
{g3,0,2(n-1)},
{Log[MR2/mt^2],0,3},
{Log[m3^2/MR^2],0,3},
{Log[mQ3^2/MR^2],0,3},
{Log[mU3^2/MR^2],0,3},
{Log[msq^2/MR^2],0,3},
{PolyLog[2,1-m3^2/mQ3^2],0,2},
{PolyLog[2,1-m3^2/mU3^2],0,2},
{PolyLog[2,1-mQ3^2/mU3^2],0,2},
{Xt,0,8},
{Ddlatas2,0,1}

];



(* ::Subchapter::Closed:: *)
(*General result to be printed*)


(*
General result to be return when program is ran
*)

Module[{Res},

Res=Evaluate[M[h,GeneralSolution][3]];
Print["
Generic relation between the squared Higgs boson pole mass and the running squared mass at 3-loop order in the DR scheme:

\!\(\*SuperscriptBox[\(Mh\), \(2\)]\) = ",Res," ,

where k = ",HoldForm[1/(4Pi)^2],", a3 = ",k^(1/2)g3^2,", at = ",k^(1/2)yt^2 Sin[beta]^2,", 
and the corrections 'Dxat(n)as(m)', as all the parameters, are in the DR scheme.
"]

Print[
"Use M[Subindex,LimitingCase][LoopOrder,AlphaTOrder] to obtain the relation between the pole mass of the desire particle (Subindex) and its running DR mass expansion in the desired limiting case and order.
Available subindices are 't' and 'h', and available limiting cases are 'GeneralSolution', 'NonDegenerate', and 'Degenerate'. AlphaTOrder is set to 1 by default and can be omitted.

Use Info[] to obtain information about the functions and parameters."
];

];



(* ::Subchapter::Closed:: *)
(*Degenerate result to be printed*)


(*
Degenerate result to be return when program is ran
*)
(*
Module[{Res},

Res=Evaluate[M[h,Degenerate][3]];
Print[
"Relation between the squared Higgs boson pole mass and the running squared mass at 3-loop order in the DR scheme in the degenerate limit:

\!\(\*SuperscriptBox[\(Mh\), \(2\)]\) = ",Res," ,

where k = ",HoldForm[1/(4Pi)^2],", a3 = ",k^(1/2)g3^2,", at = ",k^(1/2)yt^2 Sin[beta]^2
]

Print[
"Use M[Subindex,LimitingCase][LoopOrder,AlphaTOrder] to obtain the relation between the pole mass of the desire particle (Subindex) and it running DR mass expansion in the desired limiting case and order.
Available subindices are 't' and 'h', and available limiting cases are 'GeneralSolution', 'NonDegenerate', and 'Degenerate'. AlphaTOrder is set to 1 by default.

Use Info[] to obtain information about the functions and parameters."
];

];
*)


(* ::Subchapter::Closed:: *)
(*Complete results*)


(*
Complete 3-loop order Higgs boson mass expansion of O(mt^2 at a3^2), but missing dlatas2
*)

(* Using gauge eigenstate masses *)

Mh2[EFT]=(
MZ^2 Cos[2 beta]^2+at mt^2 (Sqrt[k] ((12 Xt^4)/(mQ3^2-mU3^2)^2+(6+(12 Xt^2)/(mQ3^2-mU3^2)+(-((6 mQ3^2)/(mQ3^2-mU3^2)^3)-(6 mU3^2)/(mQ3^2-mU3^2)^3) Xt^4) Log[mQ3^2/MR^2]+12 Log[MR^2/mt^2]+(6-(12 Xt^2)/(mQ3^2-mU3^2)+((6 mQ3^2)/(mQ3^2-mU3^2)^3+(6 mU3^2)/(mQ3^2-mU3^2)^3) Xt^4) Log[mU3^2/MR^2])+a3 k (-64+(32 m3^2)/mQ3^2+(32 m3^2)/mU3^2-(64 m3^2 Xt^2)/(mQ3^2 mU3^2)-(512 m3 Xt^3)/(mQ3^2-mU3^2)^2+(192/(mQ3^2-mU3^2)^2+(32 m3^2)/(mQ3^2 (mQ3^2-mU3^2)^2)+(32 m3^2)/(mU3^2 (mQ3^2-mU3^2)^2)) Xt^4+(-((16 (3 m3^4-2 m3^2 mQ3^2+mQ3^4))/(m3^2-mQ3^2)^2)+(128 m3^3 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))-(32 (3 mQ3^2-mU3^2) Xt^2)/(mQ3^2-mU3^2)^2+(64 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(mQ3^2-mU3^2)^3+(32 (m3^2 mQ3^2+2 mQ3^4-m3^2 mU3^2+2 mQ3^2 mU3^2) Xt^4)/(mQ3^2-mU3^2)^4) Log[mQ3^2/MR^2]^2+96 Log[MR^2/mt^2]^2+((32 (3 m3^4-2 m3^2 mU3^2+mU3^4))/(m3^2-mU3^2)^2-(64 (2 m3^3-m3 mU3^2) Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(64 (3 m3^2-2 mU3^2) Xt^2)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(128 (m3 mQ3^2+3 m3 mU3^2) Xt^3)/(-mQ3^2+mU3^2)^3-(32 (2 m3^2+3 mQ3^2+7 mU3^2) Xt^4)/(-mQ3^2+mU3^2)^3) Log[mU3^2/MR^2]+(-((16 (3 m3^4-2 m3^2 mU3^2+mU3^4))/(m3^2-mU3^2)^2)+(128 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))-(32 (-mQ3^2+3 mU3^2) Xt^2)/(-mQ3^2+mU3^2)^2+(64 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(-mQ3^2+mU3^2)^3+(32 (-m3^2 mQ3^2+m3^2 mU3^2+2 mQ3^2 mU3^2+2 mU3^4) Xt^4)/(-mQ3^2+mU3^2)^4) Log[mU3^2/MR^2]^2+Log[MR^2/mt^2] (-((32 (2 m3^4-m3^2 mQ3^2-m3^2 mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2)))+((32 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(128 m3^3 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))) Log[m3^2/MR^2]+(-((64 m3^2 mQ3^2)/(m3^2-mQ3^2)^2)+(32 mQ3^4)/(m3^2-mQ3^2)^2+(128 m3 mQ3^2 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))) Log[mQ3^2/MR^2]+(-((64 m3^2 mU3^2)/(m3^2-mU3^2)^2)+(32 mU3^4)/(m3^2-mU3^2)^2+(128 m3 mU3^2 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))) Log[mU3^2/MR^2])+Log[m3^2/MR^2] (-((32 (m3^10 mQ3^2-2 m3^8 mQ3^4+m3^6 mQ3^6+m3^10 mU3^2+2 m3^8 mQ3^2 mU3^2-2 m3^6 mQ3^4 mU3^2+m3^4 mQ3^6 mU3^2-2 m3^8 mU3^4-2 m3^6 mQ3^2 mU3^4+m3^6 mU3^6+m3^4 mQ3^2 mU3^6))/(mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2))+(64 m3^3 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))+(64 m3^4 (-m3^2+mQ3^2+mU3^2) Xt^2)/(mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2))-(32 (m3^2 mQ3^2+m3^2 mU3^2) Xt^4)/(mQ3^2 mU3^2 (mQ3^2-mU3^2)^2)+((32 m3^4)/(m3^2-mQ3^2)^2-(128 m3^3 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))-(256 m3^3 Xt^3)/(mQ3^2-mU3^2)^3) Log[mQ3^2/MR^2]+((32 m3^4)/(m3^2-mU3^2)^2-(128 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))-(256 m3^3 Xt^3)/(-mQ3^2+mU3^2)^3) Log[mU3^2/MR^2])+Log[mQ3^2/MR^2] ((32 (3 m3^4-2 m3^2 mQ3^2+mQ3^4))/(m3^2-mQ3^2)^2-(64 (2 m3^3-m3 mQ3^2) Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))+(64 (3 m3^2-2 mQ3^2) Xt^2)/((m3^2-mQ3^2) (mQ3^2-mU3^2))+(128 (3 m3 mQ3^2+m3 mU3^2) Xt^3)/(mQ3^2-mU3^2)^3-(32 (2 m3^2+7 mQ3^2+3 mU3^2) Xt^4)/(mQ3^2-mU3^2)^3+((64 (mQ3^2+mU3^2) Xt^2)/(mQ3^2-mU3^2)^2-(64 (mQ3^4+2 mQ3^2 mU3^2+mU3^4) Xt^4)/(mQ3^2-mU3^2)^4) Log[mU3^2/MR^2])+(-((64 m3^4)/(-m3^2+mQ3^2)^2)+(256 m3^3 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))+(128 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(mQ3^2-mU3^2)^3-(32 (-2 m3^2+mQ3^2+mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) PolyLog[2,1-m3^2/mQ3^2]+(-((64 m3^4)/(m3^2-mU3^2)^2)+(256 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(128 m3 (-2 m3^2+mQ3^2+mU3^2) Xt^3)/(mQ3^2-mU3^2)^3+(32 (-2 m3^2+mQ3^2+mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) PolyLog[2,1-m3^2/mU3^2])+a3^2 k^(3/2) (dlatas2-(128 m3^4)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))+(128 mQ3^2 mU3^2)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))-(16 (m3^4-mQ3^2 mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2))-(48 (m3^4-mQ3^2 mU3^2)^2)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(16 (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2) (-2 m3^6 mQ3^2+2 m3^4 mQ3^4-2 m3^6 mU3^2+9 m3^4 mQ3^2 mU3^2-6 m3^2 mQ3^4 mU3^2+2 m3^4 mU3^4-6 m3^2 mQ3^2 mU3^4+3 mQ3^4 mU3^4))/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)+(4 (291 m3^8-282 m3^6 mQ3^2+93 m3^4 mQ3^4-18 m3^2 mQ3^6-360 m3^6 msq^2+300 m3^4 mQ3^2 msq^2-180 m3^2 mQ3^4 msq^2-282 m3^6 mU3^2-112 m3^4 mQ3^2 mU3^2+208 m3^2 mQ3^4 mU3^2-6 mQ3^6 mU3^2+300 m3^4 msq^2 mU3^2+240 m3^2 mQ3^2 msq^2 mU3^2-60 mQ3^4 msq^2 mU3^2+93 m3^4 mU3^4+208 m3^2 mQ3^2 mU3^4-169 mQ3^4 mU3^4-180 m3^2 msq^2 mU3^4-60 mQ3^2 msq^2 mU3^4-18 m3^2 mU3^6-6 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+SMConstatas2-(64 (14 m3^3-3 m3 mQ3^2-30 m3 msq^2-3 m3 mU3^2) Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))-(64 m3^2 (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2) mU3^2 (m3^2-mU3^2))-(512 m3 (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+(8 (-44 m3^10 mQ3^2+56 m3^8 mQ3^4-12 m3^6 mQ3^6-44 m3^10 mU3^2-455 m3^8 mQ3^2 mU3^2+558 m3^6 mQ3^4 mU3^2-173 m3^4 mQ3^6 mU3^2+18 m3^2 mQ3^8 mU3^2+360 m3^6 mQ3^2 msq^2 mU3^2-300 m3^4 mQ3^4 msq^2 mU3^2+180 m3^2 mQ3^6 msq^2 mU3^2+56 m3^8 mU3^4+558 m3^6 mQ3^2 mU3^4-168 m3^4 mQ3^4 mU3^4-236 m3^2 mQ3^6 mU3^4+6 mQ3^8 mU3^4-300 m3^4 mQ3^2 msq^2 mU3^4-240 m3^2 mQ3^4 msq^2 mU3^4+60 mQ3^6 msq^2 mU3^4-12 m3^6 mU3^6-173 m3^4 mQ3^2 mU3^6-236 m3^2 mQ3^4 mU3^6+277 mQ3^6 mU3^6+180 m3^2 mQ3^2 msq^2 mU3^6+60 mQ3^4 msq^2 mU3^6+18 m3^2 mQ3^2 mU3^8+6 mQ3^4 mU3^8) Xt^4)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(128 (14 m3^3-3 m3 mQ3^2-30 m3 msq^2-3 m3 mU3^2) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+(-((4 (128 m3^16-550 m3^14 mQ3^2+506 m3^12 mQ3^4-85 m3^10 mQ3^6-88 m3^8 mQ3^8+4 m3^6 mQ3^10+38 m3^4 mQ3^12-9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-346 m3^14 mU3^2+1902 m3^12 mQ3^2 mU3^2-1873 m3^10 mQ3^4 mU3^2+253 m3^8 mQ3^6 mU3^2+612 m3^6 mQ3^8 mU3^2-368 m3^4 mQ3^10 mU3^2+47 m3^2 mQ3^12 mU3^2-3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+280 m3^12 mU3^4-2509 m3^10 mQ3^2 mU3^4+3011 m3^8 mQ3^4 mU3^4-1194 m3^6 mQ3^6 mU3^4-98 m3^4 mQ3^8 mU3^4+175 m3^2 mQ3^10 mU3^4-mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4-13 m3^10 mU3^6+1341 m3^8 mQ3^2 mU3^6-1842 m3^6 mQ3^4 mU3^6+874 m3^4 mQ3^6 mU3^6-73 m3^2 mQ3^8 mU3^6-63 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-37 m3^8 mU3^8-268 m3^6 mQ3^2 mU3^8+450 m3^4 mQ3^4 mU3^8-268 m3^2 mQ3^6 mU3^8+67 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(32 (20 m3^11 mQ3^2-202 m3^9 mQ3^4+194 m3^7 mQ3^6-65 m3^5 mQ3^8+8 m3^3 mQ3^10-3 m3 mQ3^12+60 m3^7 mQ3^4 msq^2-60 m3^5 mQ3^6 msq^2-30 m3^7 mQ3^2 msq^4+30 m3^5 mQ3^4 msq^4-20 m3^11 mU3^2+145 m3^9 mQ3^2 mU3^2+240 m3^7 mQ3^4 mU3^2-321 m3^5 mQ3^6 mU3^2+90 m3^3 mQ3^8 mU3^2+10 m3 mQ3^10 mU3^2-60 m3^7 mQ3^2 msq^2 mU3^2-60 m3^5 mQ3^4 msq^2 mU3^2+120 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2+30 m3^5 mQ3^2 msq^4 mU3^2-60 m3^3 mQ3^4 msq^4 mU3^2+41 m3^9 mU3^4-346 m3^7 mQ3^2 mU3^4+91 m3^5 mQ3^4 mU3^4+140 m3^3 mQ3^6 mU3^4-70 m3 mQ3^8 mU3^4+120 m3^5 mQ3^2 msq^2 mU3^4-60 m3^3 mQ3^4 msq^2 mU3^4-60 m3 mQ3^6 msq^2 mU3^4-60 m3^5 msq^4 mU3^4+30 m3^3 mQ3^2 msq^4 mU3^4+30 m3 mQ3^4 msq^4 mU3^4-24 m3^7 mU3^6+199 m3^5 mQ3^2 mU3^6-174 m3^3 mQ3^4 mU3^6+47 m3 mQ3^6 mU3^6-60 m3^3 mQ3^2 msq^2 mU3^6+60 m3 mQ3^4 msq^2 mU3^6+30 m3^3 msq^4 mU3^6-30 m3 mQ3^2 msq^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(8 (128 m3^16-1062 m3^14 mQ3^2+1682 m3^12 mQ3^4-641 m3^10 mQ3^6-200 m3^8 mQ3^8+72 m3^6 mQ3^10+38 m3^4 mQ3^12-9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-342 m3^14 mU3^2+3302 m3^12 mQ3^2 mU3^2-5333 m3^10 mQ3^4 mU3^2+1921 m3^8 mQ3^6 mU3^2+948 m3^6 mQ3^8 mU3^2-572 m3^4 mQ3^10 mU3^2+47 m3^2 mQ3^12 mU3^2-3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+268 m3^12 mU3^4-3637 m3^10 mQ3^2 mU3^4+6335 m3^8 mQ3^4 mU3^4-2862 m3^6 mQ3^6 mU3^4-434 m3^4 mQ3^8 mU3^4+379 m3^2 mQ3^10 mU3^4-mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4-m3^10 mU3^6+1445 m3^8 mQ3^2 mU3^6-2814 m3^6 mQ3^4 mU3^6+1430 m3^4 mQ3^6 mU3^6+39 m3^2 mQ3^8 mU3^6-131 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-41 m3^8 mU3^8-132 m3^6 mQ3^2 mU3^8+382 m3^4 mQ3^4 mU3^8-268 m3^2 mQ3^6 mU3^8+67 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(32 (2 m3^13-35 m3^11 mQ3^2-131 m3^9 mQ3^4+363 m3^7 mQ3^6-177 m3^5 mQ3^8+16 m3^3 mQ3^10-6 m3 mQ3^12+120 m3^7 mQ3^4 msq^2-120 m3^5 mQ3^6 msq^2-60 m3^7 mQ3^2 msq^4+60 m3^5 mQ3^4 msq^4-41 m3^11 mU3^2+275 m3^9 mQ3^2 mU3^2-49 m3^7 mQ3^4 mU3^2-575 m3^5 mQ3^6 mU3^2+274 m3^3 mQ3^8 mU3^2+20 m3 mQ3^10 mU3^2-120 m3^7 mQ3^2 msq^2 mU3^2-120 m3^5 mQ3^4 msq^2 mU3^2+240 m3^3 mQ3^6 msq^2 mU3^2+60 m3^7 msq^4 mU3^2+60 m3^5 mQ3^2 msq^4 mU3^2-120 m3^3 mQ3^4 msq^4 mU3^2+78 m3^9 mU3^4-437 m3^7 mQ3^2 mU3^4+421 m3^5 mQ3^4 mU3^4+221 m3^3 mQ3^6 mU3^4-187 m3 mQ3^8 mU3^4+240 m3^5 mQ3^2 msq^2 mU3^4-120 m3^3 mQ3^4 msq^2 mU3^4-120 m3 mQ3^6 msq^2 mU3^4-120 m3^5 msq^4 mU3^4+60 m3^3 mQ3^2 msq^4 mU3^4+60 m3 mQ3^4 msq^4 mU3^4-45 m3^7 mU3^6+233 m3^5 mQ3^2 mU3^6-331 m3^3 mQ3^4 mU3^6+111 m3 mQ3^6 mU3^6-120 m3^3 mQ3^2 msq^2 mU3^6+120 m3 mQ3^4 msq^2 mU3^6+60 m3^3 msq^4 mU3^6-60 m3 mQ3^2 msq^4 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(4 (1156 m3^16 mQ3^2-4276 m3^14 mQ3^4+4738 m3^12 mQ3^6-1189 m3^10 mQ3^8-556 m3^8 mQ3^10+106 m3^6 mQ3^12+38 m3^4 mQ3^14-9 m3^2 mQ3^16+60 m3^12 mQ3^4 msq^2+120 m3^10 mQ3^6 msq^2-180 m3^8 mQ3^8 msq^2-90 m3^10 mQ3^4 msq^4+60 m3^8 mQ3^6 msq^4+30 m3^6 mQ3^8 msq^4+124 m3^16 mU3^2-4340 m3^14 mQ3^2 mU3^2+15130 m3^12 mQ3^4 mU3^2-16510 m3^10 mQ3^6 mU3^2+4017 m3^8 mQ3^8 mU3^2+2156 m3^6 mQ3^10 mU3^2-636 m3^4 mQ3^12 mU3^2+38 m3^2 mQ3^14 mU3^2-3 mQ3^16 mU3^2-180 m3^10 mQ3^4 msq^2 mU3^2-360 m3^8 mQ3^6 msq^2 mU3^2+540 m3^6 mQ3^8 msq^2 mU3^2+270 m3^8 mQ3^4 msq^4 mU3^2-180 m3^6 mQ3^6 msq^4 mU3^2-90 m3^4 mQ3^8 msq^4 mU3^2-328 m3^14 mU3^4+6138 m3^12 mQ3^2 mU3^4-19952 m3^10 mQ3^4 mU3^4+21392 m3^8 mQ3^6 mU3^4-5480 m3^6 mQ3^8 mU3^4-2278 m3^4 mQ3^10 mU3^4+528 m3^2 mQ3^12 mU3^4-4 mQ3^14 mU3^4-60 m3^12 msq^2 mU3^4-120 m3^10 mQ3^2 msq^2 mU3^4+360 m3^8 mQ3^4 msq^2 mU3^4+360 m3^6 mQ3^6 msq^2 mU3^4-540 m3^4 mQ3^8 msq^2 mU3^4+90 m3^10 msq^4 mU3^4-60 m3^8 mQ3^2 msq^4 mU3^4-300 m3^6 mQ3^4 msq^4 mU3^4+180 m3^4 mQ3^6 msq^4 mU3^4+90 m3^2 mQ3^8 msq^4 mU3^4+250 m3^12 mU3^6-4006 m3^10 mQ3^2 mU3^6+11990 m3^8 mQ3^4 mU3^6-12428 m3^6 mQ3^6 mU3^6+3398 m3^4 mQ3^8 mU3^6+978 m3^2 mQ3^10 mU3^6-166 mQ3^12 mU3^6+180 m3^10 msq^2 mU3^6+360 m3^8 mQ3^2 msq^2 mU3^6-600 m3^6 mQ3^4 msq^2 mU3^6-120 m3^4 mQ3^6 msq^2 mU3^6+180 m3^2 mQ3^8 msq^2 mU3^6-270 m3^8 msq^4 mU3^6+180 m3^6 mQ3^2 msq^4 mU3^6+180 m3^4 mQ3^4 msq^4 mU3^6-60 m3^2 mQ3^6 msq^4 mU3^6-30 mQ3^8 msq^4 mU3^6+9 m3^10 mU3^8+1040 m3^8 mQ3^2 mU3^8-2934 m3^6 mQ3^4 mU3^8+2772 m3^4 mQ3^6 mU3^8-779 m3^2 mQ3^8 mU3^8-132 mQ3^10 mU3^8-180 m3^8 msq^2 mU3^8-360 m3^6 mQ3^2 msq^2 mU3^8+540 m3^4 mQ3^4 msq^2 mU3^8+270 m3^6 msq^4 mU3^8-180 m3^4 mQ3^2 msq^4 mU3^8-90 m3^2 mQ3^4 msq^4 mU3^8-43 m3^8 mU3^10-60 m3^6 mQ3^2 mU3^10+210 m3^4 mQ3^4 mU3^10-132 m3^2 mQ3^6 mU3^10+33 mQ3^8 mU3^10+60 m3^6 msq^2 mU3^10+120 m3^4 mQ3^2 msq^2 mU3^10-180 m3^2 mQ3^4 msq^2 mU3^10-90 m3^4 msq^4 mU3^10+60 m3^2 mQ3^2 msq^4 mU3^10+30 mQ3^4 msq^4 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(32 (14 m3^11 mQ3^4+86 m3^9 mQ3^6-224 m3^7 mQ3^8+113 m3^5 mQ3^10-8 m3^3 mQ3^12+3 m3 mQ3^14-60 m3^7 mQ3^6 msq^2+60 m3^5 mQ3^8 msq^2+30 m3^7 mQ3^4 msq^4-30 m3^5 mQ3^6 msq^4-32 m3^11 mQ3^2 mU3^2+117 m3^9 mQ3^4 mU3^2-362 m3^7 mQ3^6 mU3^2+510 m3^5 mQ3^8 mU3^2-194 m3^3 mQ3^10 mU3^2-7 m3 mQ3^12 mU3^2+120 m3^5 mQ3^6 msq^2 mU3^2-120 m3^3 mQ3^8 msq^2 mU3^2-60 m3^5 mQ3^4 msq^4 mU3^2+60 m3^3 mQ3^6 msq^4 mU3^2+18 m3^11 mU3^4-6 m3^9 mQ3^2 mU3^4-182 m3^7 mQ3^4 mU3^4+450 m3^5 mQ3^6 mU3^4-388 m3^3 mQ3^8 mU3^4+108 m3 mQ3^10 mU3^4+60 m3^7 mQ3^2 msq^2 mU3^4-60 m3^5 mQ3^4 msq^2 mU3^4-60 m3^3 mQ3^6 msq^2 mU3^4+60 m3 mQ3^8 msq^2 mU3^4-30 m3^7 msq^4 mU3^4+30 m3^5 mQ3^2 msq^4 mU3^4+30 m3^3 mQ3^4 msq^4 mU3^4-30 m3 mQ3^6 msq^4 mU3^4-37 m3^9 mU3^6+106 m3^7 mQ3^2 mU3^6-30 m3^5 mQ3^4 mU3^6-158 m3^3 mQ3^6 mU3^6+87 m3 mQ3^8 mU3^6-120 m3^5 mQ3^2 msq^2 mU3^6+120 m3^3 mQ3^4 msq^2 mU3^6+60 m3^5 msq^4 mU3^6-60 m3^3 mQ3^2 msq^4 mU3^6+22 m3^7 mU3^8-83 m3^5 mQ3^2 mU3^8+108 m3^3 mQ3^4 mU3^8-31 m3 mQ3^6 mU3^8+60 m3^3 mQ3^2 msq^2 mU3^8-60 m3 mQ3^4 msq^2 mU3^8-30 m3^3 msq^4 mU3^8+30 m3 mQ3^2 msq^4 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^5)+(1280 m3^2 mQ3^4 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^5)) Log[mQ3^2/MR^2]^3+(24 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3))+24 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt-1/(mQ3^2-mU3^2)^2 48 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)) Xt^4-(48 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^5)/(mQ3^2-mU3^2)^2) Log[msq^2/MR^2]^2+736 Log[MR^2/mt^2]^3+((4 (128 m3^14-252 m3^12 mQ3^2+20 m3^10 mQ3^4+115 m3^8 mQ3^6-20 m3^6 mQ3^8+9 m3^4 mQ3^10-60 m3^10 mQ3^2 msq^2+120 m3^8 mQ3^4 msq^2-60 m3^6 mQ3^6 msq^2+90 m3^8 mQ3^2 msq^4-180 m3^6 mQ3^4 msq^4+90 m3^4 mQ3^6 msq^4-1028 m3^12 mU3^2+3132 m3^10 mQ3^2 mU3^2-2847 m3^8 mQ3^4 mU3^2+764 m3^6 mQ3^6 mU3^2-39 m3^4 mQ3^8 mU3^2-6 m3^2 mQ3^10 mU3^2+60 m3^10 msq^2 mU3^2-420 m3^8 mQ3^2 msq^2 mU3^2+660 m3^6 mQ3^4 msq^2 mU3^2-300 m3^4 mQ3^6 msq^2 mU3^2-90 m3^8 msq^4 mU3^2+120 m3^6 mQ3^2 msq^4 mU3^2+30 m3^4 mQ3^4 msq^4 mU3^2-60 m3^2 mQ3^6 msq^4 mU3^2+1328 m3^10 mU3^4-4635 m3^8 mQ3^2 mU3^4+4812 m3^6 mQ3^4 mU3^4-1436 m3^4 mQ3^6 mU3^4+54 m3^2 mQ3^8 mU3^4-3 mQ3^10 mU3^4+300 m3^8 msq^2 mU3^4-300 m3^6 mQ3^2 msq^2 mU3^4-300 m3^4 mQ3^4 msq^2 mU3^4+300 m3^2 mQ3^6 msq^2 mU3^4+60 m3^6 msq^4 mU3^4-150 m3^4 mQ3^2 msq^4 mU3^4+120 m3^2 mQ3^4 msq^4 mU3^4-30 mQ3^6 msq^4 mU3^4-313 m3^8 mU3^6+2044 m3^6 mQ3^2 mU3^6-2968 m3^4 mQ3^4 mU3^6+992 m3^2 mQ3^6 mU3^6+5 mQ3^8 mU3^6-300 m3^6 msq^2 mU3^6+660 m3^4 mQ3^2 msq^2 mU3^6-420 m3^2 mQ3^4 msq^2 mU3^6+60 mQ3^6 msq^2 mU3^6+30 m3^4 msq^4 mU3^6-60 m3^2 mQ3^2 msq^4 mU3^6+30 mQ3^4 msq^4 mU3^6-560 m3^6 mU3^8+815 m3^4 mQ3^2 mU3^8+284 m3^2 mQ3^4 mU3^8-299 mQ3^6 mU3^8-60 m3^4 msq^2 mU3^8+120 m3^2 mQ3^2 msq^2 mU3^8-60 mQ3^4 msq^2 mU3^8+291 m3^4 mU3^10-702 m3^2 mQ3^2 mU3^10+291 mQ3^4 mU3^10+18 m3^2 mU3^12+6 mQ3^2 mU3^12))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2))-(32 (-32 m3^9 mQ3^2+m3^7 mQ3^4+2 m3^5 mQ3^6-3 m3^3 mQ3^8+30 m3^5 mQ3^2 msq^4-30 m3^3 mQ3^4 msq^4+32 m3^9 mU3^2+380 m3^7 mQ3^2 mU3^2-311 m3^5 mQ3^4 mU3^2+14 m3^3 mQ3^6 mU3^2+3 m3 mQ3^8 mU3^2-120 m3^5 mQ3^2 msq^2 mU3^2+120 m3^3 mQ3^4 msq^2 mU3^2-30 m3^5 msq^4 mU3^2+30 m3 mQ3^4 msq^4 mU3^2-397 m3^7 mU3^4-122 m3^5 mQ3^2 mU3^4+367 m3^3 mQ3^4 mU3^4-16 m3 mQ3^6 mU3^4+120 m3^5 msq^2 mU3^4-120 m3 mQ3^4 msq^2 mU3^4+30 m3^3 msq^4 mU3^4-30 m3 mQ3^2 msq^4 mU3^4+479 m3^5 mU3^6-226 m3^3 mQ3^2 mU3^6-137 m3 mQ3^4 mU3^6-120 m3^3 msq^2 mU3^6+120 m3 mQ3^2 msq^2 mU3^6-200 m3^3 mU3^8+160 m3 mQ3^2 mU3^8+6 m3 mU3^10) Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)-(16 (-24 m3^10 mQ3^2+65 m3^8 mQ3^4-41 m3^6 mQ3^6+492 m3^10 mU3^2-1462 m3^8 mQ3^2 mU3^2+1346 m3^6 mQ3^4 mU3^2-373 m3^4 mQ3^6 mU3^2+9 m3^2 mQ3^8 mU3^2+90 m3^6 mQ3^2 msq^2 mU3^2-180 m3^4 mQ3^4 msq^2 mU3^2+90 m3^2 mQ3^6 msq^2 mU3^2-383 m3^8 mU3^4+1473 m3^6 mQ3^2 mU3^4-1612 m3^4 mQ3^4 mU3^4+471 m3^2 mQ3^6 mU3^4+3 mQ3^8 mU3^4-90 m3^6 msq^2 mU3^4+210 m3^4 mQ3^2 msq^2 mU3^4-150 m3^2 mQ3^4 msq^2 mU3^4+30 mQ3^6 msq^2 mU3^4-226 m3^6 mU3^6+137 m3^4 mQ3^2 mU3^6+338 m3^2 mQ3^4 mU3^6-177 mQ3^6 mU3^6-30 m3^4 msq^2 mU3^6+60 m3^2 mQ3^2 msq^2 mU3^6-30 mQ3^4 msq^2 mU3^6+192 m3^4 mU3^8-391 m3^2 mQ3^2 mU3^8+151 mQ3^4 mU3^8+9 m3^2 mU3^10+3 mQ3^2 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)+(128 (11 m3^9+7 m3^7 mQ3^2-14 m3^5 mQ3^4-37 m3^7 mU3^2+150 m3^5 mQ3^2 mU3^2-117 m3^3 mQ3^4 mU3^2+3 m3 mQ3^6 mU3^2-30 m3^3 mQ3^2 msq^2 mU3^2+30 m3 mQ3^4 msq^2 mU3^2-48 m3^5 mU3^4-47 m3^3 mQ3^2 mU3^4+82 m3 mQ3^4 mU3^4+30 m3^3 msq^2 mU3^4-30 m3 mQ3^2 msq^2 mU3^4+66 m3^3 mU3^6-53 m3 mQ3^2 mU3^6-3 m3 mU3^8) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(8 (-176 m3^14 mQ3^2+322 m3^12 mQ3^4-147 m3^10 mQ3^6-10 m3^8 mQ3^8+20 m3^6 mQ3^10-9 m3^4 mQ3^12+60 m3^10 mQ3^4 msq^2-120 m3^8 mQ3^6 msq^2+60 m3^6 mQ3^8 msq^2-90 m3^8 mQ3^4 msq^4+180 m3^6 mQ3^6 msq^4-90 m3^4 mQ3^8 msq^4+176 m3^14 mU3^2-120 m3^12 mQ3^2 mU3^2+201 m3^10 mQ3^4 mU3^2-474 m3^8 mQ3^6 mU3^2+178 m3^6 mQ3^8 mU3^2+21 m3^4 mQ3^10 mU3^2+6 m3^2 mQ3^12 mU3^2-120 m3^10 mQ3^2 msq^2 mU3^2+270 m3^8 mQ3^4 msq^2 mU3^2-180 m3^6 mQ3^6 msq^2 mU3^2+30 m3^4 mQ3^8 msq^2 mU3^2+180 m3^8 mQ3^2 msq^4 mU3^2-300 m3^6 mQ3^4 msq^4 mU3^2+60 m3^4 mQ3^6 msq^4 mU3^2+60 m3^2 mQ3^8 msq^4 mU3^2-2586 m3^12 mU3^4+6267 m3^10 mQ3^2 mU3^4-6104 m3^8 mQ3^4 mU3^4+3202 m3^6 mQ3^6 mU3^4-692 m3^4 mQ3^8 mU3^4-42 m3^2 mQ3^10 mU3^4+3 mQ3^12 mU3^4+60 m3^10 msq^2 mU3^4-360 m3^8 mQ3^2 msq^2 mU3^4+420 m3^6 mQ3^4 msq^2 mU3^4-120 m3^2 mQ3^8 msq^2 mU3^4-90 m3^8 msq^4 mU3^4+60 m3^6 mQ3^2 msq^4 mU3^4+180 m3^4 mQ3^4 msq^4 mU3^4-180 m3^2 mQ3^6 msq^4 mU3^4+30 mQ3^8 msq^4 mU3^4+5679 m3^10 mU3^6-13622 m3^8 mQ3^2 mU3^6+11390 m3^6 mQ3^4 mU3^6-4230 m3^4 mQ3^6 mU3^6+722 m3^2 mQ3^8 mU3^6+mQ3^10 mU3^6+210 m3^8 msq^2 mU3^6-60 m3^6 mQ3^2 msq^2 mU3^6-480 m3^4 mQ3^4 msq^2 mU3^6+300 m3^2 mQ3^6 msq^2 mU3^6+30 mQ3^8 msq^2 mU3^6+60 m3^6 msq^4 mU3^6-180 m3^4 mQ3^2 msq^4 mU3^6+180 m3^2 mQ3^4 msq^4 mU3^6-60 mQ3^6 msq^4 mU3^6-4030 m3^8 mU3^8+9482 m3^6 mQ3^2 mU3^8-7123 m3^4 mQ3^4 mU3^8+1861 m3^2 mQ3^6 mU3^8-190 mQ3^8 mU3^8-240 m3^6 msq^2 mU3^8+480 m3^4 mQ3^2 msq^2 mU3^8-240 m3^2 mQ3^4 msq^2 mU3^8+30 m3^4 msq^4 mU3^8-60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8+368 m3^6 mU3^10-1079 m3^4 mQ3^2 mU3^10+1015 m3^2 mQ3^4 mU3^10-244 mQ3^6 mU3^10-30 m3^4 msq^2 mU3^10+60 m3^2 mQ3^2 msq^2 mU3^10-30 mQ3^4 msq^2 mU3^10+392 m3^4 mU3^12-787 m3^2 mQ3^2 mU3^12+347 mQ3^4 mU3^12+9 m3^2 mU3^14+3 mQ3^2 mU3^14) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)^4)-(64 (34 m3^9 mQ3^2-35 m3^7 mQ3^4-2 m3^5 mQ3^6+3 m3^3 mQ3^8-30 m3^5 mQ3^2 msq^4+30 m3^3 mQ3^4 msq^4+30 m3^9 mU3^2-6 m3^7 mQ3^2 mU3^2-5 m3^5 mQ3^4 mU3^2-5 m3^3 mQ3^6 mU3^2-3 m3 mQ3^8 mU3^2+30 m3^5 mQ3^2 msq^2 mU3^2-30 m3^3 mQ3^4 msq^2 mU3^2+30 m3^5 msq^4 mU3^2-30 m3 mQ3^4 msq^4 mU3^2+179 m3^7 mU3^4-306 m3^5 mQ3^2 mU3^4+106 m3^3 mQ3^4 mU3^4+7 m3 mQ3^6 mU3^4-90 m3^5 msq^2 mU3^4+60 m3^3 mQ3^2 msq^2 mU3^4+30 m3 mQ3^4 msq^2 mU3^4-30 m3^3 msq^4 mU3^4+30 m3 mQ3^2 msq^4 mU3^4-437 m3^5 mU3^6+511 m3^3 mQ3^2 mU3^6-82 m3 mQ3^4 mU3^6+90 m3^3 msq^2 mU3^6-90 m3 mQ3^2 msq^2 mU3^6+215 m3^3 mU3^8-201 m3 mQ3^2 mU3^8-3 m3 mU3^10) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)-(2560 m3^2 mU3^4 Xt^6)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)+(-(1/(3 (m3^2-mU3^2)^3))80 (3 m3^6-6 m3^4 msq^2+9 m3^2 msq^4-7 m3^4 mU3^2-18 m3^2 msq^2 mU3^2+3 msq^4 mU3^2+6 m3^2 mU3^4-2 mU3^6)+(320 (2 m3^5+6 m3 msq^4-3 m3^3 mU3^2-12 m3 msq^2 mU3^2+m3 mU3^4) Xt)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+((160 (2 m3^4 msq^2-3 m3^2 msq^4+6 m3^2 msq^2 mU3^2-msq^4 mU3^2))/((m3^2-mU3^2)^3 (-mQ3^2+mU3^2))+(160 (m3^4 (mQ3^2-3 mU3^2)+2 mU3^4 (mQ3^2-2 mU3^2)+m3^2 (-4 mQ3^2 mU3^2+8 mU3^4)))/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) Xt^2+(-((1280 (-m3 msq^4+2 m3 msq^2 mU3^2))/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2))+(320 m3 (2 m3^4+3 mQ3^2 mU3^2-mU3^4-m3^2 (mQ3^2+3 mU3^2)))/(3 (m3^2-mU3^2) (-mQ3^2+mU3^2)^3)) Xt^3+(-((80 (mQ3^2+mU3^2) (2 m3^4 msq^2-3 m3^2 msq^4+6 m3^2 msq^2 mU3^2-msq^4 mU3^2))/((m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^3))+(80 (8 m3^4 mQ3^2 mU3^2-mQ3^4 mU3^4+4 mQ3^2 mU3^6+5 mU3^8-2 m3^6 (mQ3^2-mU3^2)+2 m3^2 mU3^2 (mQ3^4-5 mQ3^2 mU3^2-4 mU3^4)))/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) Xt^4+(320 (mQ3^2+mU3^2) (-6 m3 msq^4+m3^3 mU3^2+12 m3 msq^2 mU3^2-m3 mU3^4) Xt^5)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) Log[msq^2/MR^2]) Log[mU3^2/MR^2]^2+(-((4 (128 m3^10+38 m3^8 mQ3^2-2 m3^6 mQ3^4+9 m3^4 mQ3^6-60 m3^6 mQ3^2 msq^2+90 m3^4 mQ3^2 msq^4-550 m3^8 mU3^2+276 m3^6 mQ3^2 mU3^2-33 m3^4 mQ3^4 mU3^2-6 m3^2 mQ3^6 mU3^2+60 m3^6 msq^2 mU3^2-120 m3^4 mQ3^2 msq^2 mU3^2-90 m3^4 msq^4 mU3^2-60 m3^2 mQ3^2 msq^4 mU3^2+494 m3^6 mU3^4-429 m3^4 mQ3^2 mU3^4+36 m3^2 mQ3^4 mU3^4-3 mQ3^6 mU3^4+120 m3^4 msq^2 mU3^4+180 m3^2 mQ3^2 msq^2 mU3^4+60 m3^2 msq^4 mU3^4-30 mQ3^2 msq^4 mU3^4-59 m3^4 mU3^6+234 m3^2 mQ3^2 mU3^6-mQ3^4 mU3^6-180 m3^2 msq^2 mU3^6+30 msq^4 mU3^6-136 m3^2 mU3^8-63 mQ3^2 mU3^8+67 mU3^10))/(3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)))+(32 (-20 m3^7 mQ3^2+m3^5 mQ3^4+3 m3^3 mQ3^6+30 m3^3 mQ3^2 msq^4+20 m3^7 mU3^2+185 m3^5 mQ3^2 mU3^2-11 m3^3 mQ3^4 mU3^2-3 m3 mQ3^6 mU3^2-60 m3^3 mQ3^2 msq^2 mU3^2-30 m3^3 msq^4 mU3^2-30 m3 mQ3^2 msq^4 mU3^2-202 m3^5 mU3^4-149 m3^3 mQ3^2 mU3^4+10 m3 mQ3^4 mU3^4+60 m3^3 msq^2 mU3^4+60 m3 mQ3^2 msq^2 mU3^4+30 m3 msq^4 mU3^4+189 m3^3 mU3^6+32 m3 mQ3^2 mU3^6-60 m3 msq^2 mU3^6-55 m3 mU3^8) Xt)/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)-(8 (128 m3^10+42 m3^8 mQ3^2-2 m3^6 mQ3^4+9 m3^4 mQ3^6-60 m3^6 mQ3^2 msq^2+90 m3^4 mQ3^2 msq^4-1062 m3^8 mU3^2+140 m3^6 mQ3^2 mU3^2-33 m3^4 mQ3^4 mU3^2-6 m3^2 mQ3^6 mU3^2+60 m3^6 msq^2 mU3^2-120 m3^4 mQ3^2 msq^2 mU3^2-90 m3^4 msq^4 mU3^2-60 m3^2 mQ3^2 msq^4 mU3^2+1670 m3^6 mU3^4-361 m3^4 mQ3^2 mU3^4+36 m3^2 mQ3^4 mU3^4-3 mQ3^6 mU3^4+120 m3^4 msq^2 mU3^4+180 m3^2 mQ3^2 msq^2 mU3^4+60 m3^2 msq^4 mU3^4-30 mQ3^2 msq^4 mU3^4-615 m3^4 mU3^6+234 m3^2 mQ3^2 mU3^6-mQ3^4 mU3^6-180 m3^2 msq^2 mU3^6+30 msq^4 mU3^6-248 m3^2 mU3^8-63 mQ3^2 mU3^8+135 mU3^10) Xt^2)/(3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)^2)+(32 (2 m3^9-37 m3^7 mQ3^2+2 m3^5 mQ3^4+6 m3^3 mQ3^6+60 m3^3 mQ3^2 msq^4-35 m3^7 mU3^2+205 m3^5 mQ3^2 mU3^2-22 m3^3 mQ3^4 mU3^2-6 m3 mQ3^6 mU3^2-120 m3^3 mQ3^2 msq^2 mU3^2-60 m3^3 msq^4 mU3^2-60 m3 mQ3^2 msq^4 mU3^2-131 m3^5 mU3^4-281 m3^3 mQ3^2 mU3^4+20 m3 mQ3^4 mU3^4+120 m3^3 msq^2 mU3^4+120 m3 mQ3^2 msq^2 mU3^4+60 m3 msq^4 mU3^4+353 m3^3 mU3^6+81 m3 mQ3^2 mU3^6-120 m3 msq^2 mU3^6-157 m3 mU3^8) Xt^3)/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^3)+(4 (124 m3^10 mQ3^2+44 m3^8 mQ3^4-2 m3^6 mQ3^6+9 m3^4 mQ3^8-60 m3^6 mQ3^4 msq^2+90 m3^4 mQ3^4 msq^4+1156 m3^10 mU3^2-872 m3^8 mQ3^2 mU3^2+66 m3^6 mQ3^4 mU3^2-24 m3^4 mQ3^6 mU3^2-6 m3^2 mQ3^8 mU3^2-120 m3^4 mQ3^4 msq^2 mU3^2-60 m3^2 mQ3^4 msq^4 mU3^2-4276 m3^8 mU3^4+2314 m3^6 mQ3^2 mU3^4-222 m3^4 mQ3^4 mU3^4+30 m3^2 mQ3^6 mU3^4-3 mQ3^8 mU3^4+60 m3^6 msq^2 mU3^4+180 m3^2 mQ3^4 msq^2 mU3^4-90 m3^4 msq^4 mU3^4-30 mQ3^4 msq^4 mU3^4+4726 m3^6 mU3^6-2344 m3^4 mQ3^2 mU3^6+134 m3^2 mQ3^4 mU3^6-4 mQ3^6 mU3^6+120 m3^4 msq^2 mU3^6+60 m3^2 msq^4 mU3^6-1163 m3^4 mU3^8+638 m3^2 mQ3^2 mU3^8-30 mQ3^4 mU3^8-180 m3^2 msq^2 mU3^8+30 msq^4 mU3^8-604 m3^2 mU3^10+140 mQ3^2 mU3^10+169 mU3^12) Xt^4)/(3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)^4)+(32 (18 m3^7 mQ3^4-m3^5 mQ3^6-3 m3^3 mQ3^8-30 m3^3 mQ3^4 msq^4-32 m3^7 mQ3^2 mU3^2-70 m3^5 mQ3^4 mU3^2+8 m3^3 mQ3^6 mU3^2+3 m3 mQ3^8 mU3^2+60 m3^3 mQ3^4 msq^2 mU3^2+30 m3 mQ3^4 msq^4 mU3^2+14 m3^7 mU3^4+145 m3^5 mQ3^2 mU3^4+94 m3^3 mQ3^4 mU3^4-7 m3 mQ3^6 mU3^4-60 m3 mQ3^4 msq^2 mU3^4+30 m3^3 msq^4 mU3^4+86 m3^5 mU3^6-200 m3^3 mQ3^2 mU3^6-26 m3 mQ3^4 mU3^6-60 m3^3 msq^2 mU3^6-30 m3 msq^4 mU3^6-219 m3^3 mU3^8+87 m3 mQ3^2 mU3^8+60 m3 msq^2 mU3^8+103 m3 mU3^10) Xt^5)/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^5)+(1280 m3^2 mU3^4 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^5)) Log[mU3^2/MR^2]^3+Log[MR^2/mt^2]^2 (-((64 (3 m3^4+m3^2 mQ3^2+m3^2 mU3^2-5 mQ3^2 mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2)))+(192+(256 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(1024 m3^3 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))) Log[m3^2/MR^2]+((16 (m3^4-34 m3^2 mQ3^2+17 mQ3^4))/(m3^2-mQ3^2)^2+(1024 m3 mQ3^2 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))) Log[mQ3^2/MR^2]+160 Log[msq^2/MR^2]+((16 (m3^4-34 m3^2 mU3^2+17 mU3^4))/(m3^2-mU3^2)^2+(1024 m3 mU3^2 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))) Log[mU3^2/MR^2])+Log[m3^2/MR^2]^2 (-((8 (88 m3^18 mQ3^2-288 m3^16 mQ3^4+368 m3^14 mQ3^6-224 m3^12 mQ3^8+56 m3^10 mQ3^10+88 m3^18 mU3^2-302 m3^16 mQ3^2 mU3^2+838 m3^14 mQ3^4 mU3^2-797 m3^12 mQ3^6 mU3^2+280 m3^10 mQ3^8 mU3^2-39 m3^8 mQ3^10 mU3^2-288 m3^16 mU3^4+838 m3^14 mQ3^2 mU3^4-2636 m3^12 mQ3^4 mU3^4+2588 m3^10 mQ3^6 mU3^4-908 m3^8 mQ3^8 mU3^4+134 m3^6 mQ3^10 mU3^4+368 m3^14 mU3^6-797 m3^12 mQ3^2 mU3^6+2588 m3^10 mQ3^4 mU3^6-2048 m3^8 mQ3^6 mU3^6+272 m3^6 mQ3^8 mU3^6+25 m3^4 mQ3^10 mU3^6-224 m3^12 mU3^8+280 m3^10 mQ3^2 mU3^8-908 m3^8 mQ3^4 mU3^8+272 m3^6 mQ3^6 mU3^8+452 m3^4 mQ3^8 mU3^8-144 m3^2 mQ3^10 mU3^8+56 m3^10 mU3^10-39 m3^8 mQ3^2 mU3^10+134 m3^6 mQ3^4 mU3^10+25 m3^4 mQ3^6 mU3^10-144 m3^2 mQ3^8 mU3^10+36 mQ3^10 mU3^10))/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^4))+(64 (16 m3^13 mQ3^2-32 m3^11 mQ3^4+16 m3^9 mQ3^6+16 m3^13 mU3^2+31 m3^11 mQ3^2 mU3^2+18 m3^9 mQ3^4 mU3^2-25 m3^7 mQ3^6 mU3^2-32 m3^11 mU3^4+18 m3^9 mQ3^2 mU3^4-196 m3^7 mQ3^4 mU3^4+130 m3^5 mQ3^6 mU3^4+16 m3^9 mU3^6-25 m3^7 mQ3^2 mU3^6+130 m3^5 mQ3^4 mU3^6-81 m3^3 mQ3^6 mU3^6) Xt)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3)+(-((1792 m3^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(64 m3^4 (-m3^2+mQ3^2+mU3^2) (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)))/(mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2))) Xt^2-(2048 m3^7 (m3^2-mQ3^2-mU3^2) Xt^3)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)-(16 (44 m3^18 mQ3^2-144 m3^16 mQ3^4+184 m3^14 mQ3^6-112 m3^12 mQ3^8+28 m3^10 mQ3^10+44 m3^18 mU3^2-2 m3^16 mQ3^2 mU3^2-294 m3^14 mQ3^4 mU3^2+277 m3^12 mQ3^6 mU3^2+4 m3^10 mQ3^8 mU3^2-25 m3^8 mQ3^10 mU3^2-144 m3^16 mU3^4-294 m3^14 mQ3^2 mU3^4+1932 m3^12 mQ3^4 mU3^4-2012 m3^10 mQ3^6 mU3^4+572 m3^8 mQ3^8 mU3^4-70 m3^6 mQ3^10 mU3^4+184 m3^14 mU3^6+277 m3^12 mQ3^2 mU3^6-2012 m3^10 mQ3^4 mU3^6+1712 m3^8 mQ3^6 mU3^6-88 m3^6 mQ3^8 mU3^6-49 m3^4 mQ3^10 mU3^6-112 m3^12 mU3^8+4 m3^10 mQ3^2 mU3^8+572 m3^8 mQ3^4 mU3^8-88 m3^6 mQ3^6 mU3^8-548 m3^4 mQ3^8 mU3^8+156 m3^2 mQ3^10 mU3^8+28 m3^10 mU3^10-25 m3^8 mQ3^2 mU3^10-70 m3^6 mQ3^4 mU3^10-49 m3^4 mQ3^6 mU3^10+156 m3^2 mQ3^8 mU3^10-36 mQ3^10 mU3^10) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)+(128 (8 m3^13 mQ3^2-16 m3^11 mQ3^4+8 m3^9 mQ3^6+8 m3^13 mU3^2+7 m3^11 mQ3^2 mU3^2-54 m3^9 mQ3^4 mU3^2+47 m3^7 mQ3^6 mU3^2-16 m3^11 mU3^4-54 m3^9 mQ3^2 mU3^4+188 m3^7 mQ3^4 mU3^4-134 m3^5 mQ3^6 mU3^4+8 m3^9 mU3^6+47 m3^7 mQ3^2 mU3^6-134 m3^5 mQ3^4 mU3^6+87 m3^3 mQ3^6 mU3^6) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(2560 m3^6 Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(-((8 (198 m3^16-814 m3^14 mQ3^2+877 m3^12 mQ3^4-344 m3^10 mQ3^6+55 m3^8 mQ3^8-638 m3^14 mU3^2+2700 m3^12 mQ3^2 mU3^2-2796 m3^10 mQ3^4 mU3^2+1004 m3^8 mQ3^6 mU3^2-158 m3^6 mQ3^8 mU3^2+565 m3^12 mU3^4-2604 m3^10 mQ3^2 mU3^4+2240 m3^8 mQ3^4 mU3^4-368 m3^6 mQ3^6 mU3^4-m3^4 mQ3^8 mU3^4-120 m3^10 mU3^6+812 m3^8 mQ3^2 mU3^6-272 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6-m3^8 mU3^8-110 m3^6 mQ3^2 mU3^8-25 m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4))-(64 (44 m3^13-79 m3^11 mQ3^2+122 m3^9 mQ3^4-63 m3^7 mQ3^6-77 m3^11 mU3^2+72 m3^9 mQ3^2 mU3^2-209 m3^7 mQ3^4 mU3^2+142 m3^5 mQ3^6 mU3^2+22 m3^9 mU3^4+101 m3^7 mQ3^2 mU3^4+36 m3^5 mQ3^4 mU3^4-87 m3^3 mQ3^6 mU3^4+19 m3^7 mU3^6-118 m3^5 mQ3^2 mU3^6+75 m3^3 mQ3^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(16 (30 m3^16-334 m3^14 mQ3^2+517 m3^12 mQ3^4-264 m3^10 mQ3^6+55 m3^8 mQ3^8-238 m3^14 mU3^2+1484 m3^12 mQ3^2 mU3^2-1948 m3^10 mQ3^4 mU3^2+844 m3^8 mQ3^6 mU3^2-158 m3^6 mQ3^8 mU3^2+325 m3^12 mU3^4-1660 m3^10 mQ3^2 mU3^4+1648 m3^8 mQ3^4 mU3^4-288 m3^6 mQ3^6 mU3^4-m3^4 mQ3^8 mU3^4-168 m3^10 mU3^6+652 m3^8 mQ3^2 mU3^6-192 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6+55 m3^8 mU3^8-158 m3^6 mQ3^2 mU3^8-m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2))-(128 (44 m3^15-100 m3^13 mQ3^2+45 m3^11 mQ3^4+66 m3^9 mQ3^6-63 m3^7 mQ3^8-100 m3^13 mU3^2+282 m3^11 mQ3^2 mU3^2-250 m3^9 mQ3^4 mU3^2-42 m3^7 mQ3^6 mU3^2+142 m3^5 mQ3^8 mU3^2+45 m3^11 mU3^4-250 m3^9 mQ3^2 mU3^4+422 m3^7 mQ3^4 mU3^4-178 m3^5 mQ3^6 mU3^4-87 m3^3 mQ3^8 mU3^4+66 m3^9 mU3^6-42 m3^7 mQ3^2 mU3^6-178 m3^5 mQ3^4 mU3^6+186 m3^3 mQ3^6 mU3^6-63 m3^7 mU3^8+142 m3^5 mQ3^2 mU3^8-87 m3^3 mQ3^4 mU3^8) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(8 (1024 m3^18-2786 m3^16 mQ3^2+1826 m3^14 mQ3^4+549 m3^12 mQ3^6-664 m3^10 mQ3^8+55 m3^8 mQ3^10-2786 m3^16 mU3^2+8004 m3^14 mQ3^2 mU3^2-5903 m3^12 mQ3^4 mU3^2-868 m3^10 mQ3^6 mU3^2+1699 m3^8 mQ3^8 mU3^2-158 m3^6 mQ3^10 mU3^2+1826 m3^14 mU3^4-5903 m3^12 mQ3^2 mU3^4+5224 m3^10 mQ3^4 mU3^4-292 m3^8 mQ3^6 mU3^4-846 m3^6 mQ3^8 mU3^4-m3^4 mQ3^10 mU3^4+549 m3^12 mU3^6-868 m3^10 mQ3^2 mU3^6-292 m3^8 mQ3^4 mU3^6+928 m3^6 mQ3^6 mU3^6-453 m3^4 mQ3^8 mU3^6+144 m3^2 mQ3^10 mU3^6-664 m3^10 mU3^8+1699 m3^8 mQ3^2 mU3^8-846 m3^6 mQ3^4 mU3^8-453 m3^4 mQ3^6 mU3^8+288 m3^2 mQ3^8 mU3^8-36 mQ3^10 mU3^8+55 m3^8 mU3^10-158 m3^6 mQ3^2 mU3^10-m3^4 mQ3^4 mU3^10+144 m3^2 mQ3^6 mU3^10-36 mQ3^8 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^3)-(64 (39 m3^11 mQ3^2-94 m3^9 mQ3^4+63 m3^7 mQ3^6+39 m3^11 mU3^2-188 m3^9 mQ3^2 mU3^2+283 m3^7 mQ3^4 mU3^2-142 m3^5 mQ3^6 mU3^2-94 m3^9 mU3^4+283 m3^7 mQ3^2 mU3^4-284 m3^5 mQ3^4 mU3^4+87 m3^3 mQ3^6 mU3^4+63 m3^7 mU3^6-142 m3^5 mQ3^2 mU3^6+87 m3^3 mQ3^4 mU3^6) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(1280 m3^6 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[mQ3^2/MR^2]+(-((8 (198 m3^16-638 m3^14 mQ3^2+565 m3^12 mQ3^4-120 m3^10 mQ3^6-m3^8 mQ3^8-814 m3^14 mU3^2+2700 m3^12 mQ3^2 mU3^2-2604 m3^10 mQ3^4 mU3^2+812 m3^8 mQ3^6 mU3^2-110 m3^6 mQ3^8 mU3^2+877 m3^12 mU3^4-2796 m3^10 mQ3^2 mU3^4+2240 m3^8 mQ3^4 mU3^4-272 m3^6 mQ3^6 mU3^4-25 m3^4 mQ3^8 mU3^4-344 m3^10 mU3^6+1004 m3^8 mQ3^2 mU3^6-368 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6+55 m3^8 mU3^8-158 m3^6 mQ3^2 mU3^8-m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4))+(64 (44 m3^13-77 m3^11 mQ3^2+22 m3^9 mQ3^4+19 m3^7 mQ3^6-79 m3^11 mU3^2+72 m3^9 mQ3^2 mU3^2+101 m3^7 mQ3^4 mU3^2-118 m3^5 mQ3^6 mU3^2+122 m3^9 mU3^4-209 m3^7 mQ3^2 mU3^4+36 m3^5 mQ3^4 mU3^4+75 m3^3 mQ3^6 mU3^4-63 m3^7 mU3^6+142 m3^5 mQ3^2 mU3^6-87 m3^3 mQ3^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(16 (30 m3^16-238 m3^14 mQ3^2+325 m3^12 mQ3^4-168 m3^10 mQ3^6+55 m3^8 mQ3^8-334 m3^14 mU3^2+1484 m3^12 mQ3^2 mU3^2-1660 m3^10 mQ3^4 mU3^2+652 m3^8 mQ3^6 mU3^2-158 m3^6 mQ3^8 mU3^2+517 m3^12 mU3^4-1948 m3^10 mQ3^2 mU3^4+1648 m3^8 mQ3^4 mU3^4-192 m3^6 mQ3^6 mU3^4-m3^4 mQ3^8 mU3^4-264 m3^10 mU3^6+844 m3^8 mQ3^2 mU3^6-288 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6+55 m3^8 mU3^8-158 m3^6 mQ3^2 mU3^8-m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2))+(128 (44 m3^15-100 m3^13 mQ3^2+45 m3^11 mQ3^4+66 m3^9 mQ3^6-63 m3^7 mQ3^8-100 m3^13 mU3^2+282 m3^11 mQ3^2 mU3^2-250 m3^9 mQ3^4 mU3^2-42 m3^7 mQ3^6 mU3^2+142 m3^5 mQ3^8 mU3^2+45 m3^11 mU3^4-250 m3^9 mQ3^2 mU3^4+422 m3^7 mQ3^4 mU3^4-178 m3^5 mQ3^6 mU3^4-87 m3^3 mQ3^8 mU3^4+66 m3^9 mU3^6-42 m3^7 mQ3^2 mU3^6-178 m3^5 mQ3^4 mU3^6+186 m3^3 mQ3^6 mU3^6-63 m3^7 mU3^8+142 m3^5 mQ3^2 mU3^8-87 m3^3 mQ3^4 mU3^8) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(8 (1024 m3^18-2786 m3^16 mQ3^2+1826 m3^14 mQ3^4+549 m3^12 mQ3^6-664 m3^10 mQ3^8+55 m3^8 mQ3^10-2786 m3^16 mU3^2+8004 m3^14 mQ3^2 mU3^2-5903 m3^12 mQ3^4 mU3^2-868 m3^10 mQ3^6 mU3^2+1699 m3^8 mQ3^8 mU3^2-158 m3^6 mQ3^10 mU3^2+1826 m3^14 mU3^4-5903 m3^12 mQ3^2 mU3^4+5224 m3^10 mQ3^4 mU3^4-292 m3^8 mQ3^6 mU3^4-846 m3^6 mQ3^8 mU3^4-m3^4 mQ3^10 mU3^4+549 m3^12 mU3^6-868 m3^10 mQ3^2 mU3^6-292 m3^8 mQ3^4 mU3^6+928 m3^6 mQ3^6 mU3^6-453 m3^4 mQ3^8 mU3^6+144 m3^2 mQ3^10 mU3^6-664 m3^10 mU3^8+1699 m3^8 mQ3^2 mU3^8-846 m3^6 mQ3^4 mU3^8-453 m3^4 mQ3^6 mU3^8+288 m3^2 mQ3^8 mU3^8-36 mQ3^10 mU3^8+55 m3^8 mU3^10-158 m3^6 mQ3^2 mU3^10-m3^4 mQ3^4 mU3^10+144 m3^2 mQ3^6 mU3^10-36 mQ3^8 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^3)+(64 (39 m3^11 mQ3^2-94 m3^9 mQ3^4+63 m3^7 mQ3^6+39 m3^11 mU3^2-188 m3^9 mQ3^2 mU3^2+283 m3^7 mQ3^4 mU3^2-142 m3^5 mQ3^6 mU3^2-94 m3^9 mU3^4+283 m3^7 mQ3^2 mU3^4-284 m3^5 mQ3^4 mU3^4+87 m3^3 mQ3^6 mU3^4+63 m3^7 mU3^6-142 m3^5 mQ3^2 mU3^6+87 m3^3 mQ3^4 mU3^6) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(1280 m3^6 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[mU3^2/MR^2])+Log[mQ3^2/MR^2]^2 ((4 (128 m3^16-1028 m3^14 mQ3^2+1340 m3^12 mQ3^4-339 m3^10 mQ3^6-512 m3^8 mQ3^8+228 m3^6 mQ3^10+56 m3^4 mQ3^12-9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+300 m3^10 mQ3^4 msq^2-300 m3^8 mQ3^6 msq^2-60 m3^6 mQ3^8 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-380 m3^14 mU3^2+4136 m3^12 mQ3^2 mU3^2-5925 m3^10 mQ3^4 mU3^2+2199 m3^8 mQ3^6 mU3^2+1642 m3^6 mQ3^8 mU3^2-1160 m3^4 mQ3^10 mU3^2+35 m3^2 mQ3^12 mU3^2-3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-480 m3^10 mQ3^2 msq^2 mU3^2-600 m3^8 mQ3^4 msq^2 mU3^2+960 m3^6 mQ3^6 msq^2 mU3^2+180 m3^4 mQ3^8 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+284 m3^12 mU3^4-5977 m3^10 mQ3^2 mU3^4+9653 m3^8 mQ3^4 mU3^4-5326 m3^6 mQ3^6 mU3^4-410 m3^4 mQ3^8 mU3^4+967 m3^2 mQ3^10 mU3^4-7 mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+1080 m3^8 mQ3^2 msq^2 mU3^4-1080 m3^4 mQ3^6 msq^2 mU3^4-180 m3^2 mQ3^8 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4+81 m3^10 mU3^6+3481 m3^8 mQ3^2 mU3^6-6198 m3^6 mQ3^4 mU3^6+4070 m3^4 mQ3^6 mU3^6-603 m3^2 mQ3^8 mU3^6-287 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-960 m3^6 mQ3^2 msq^2 mU3^6+600 m3^4 mQ3^4 msq^2 mU3^6+480 m3^2 mQ3^6 msq^2 mU3^6+60 mQ3^8 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-101 m3^8 mU3^8-714 m3^6 mQ3^2 mU3^8+1394 m3^4 mQ3^4 mU3^8-1018 m3^2 mQ3^6 mU3^8+303 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+300 m3^4 mQ3^2 msq^2 mU3^8-300 m3^2 mQ3^4 msq^2 mU3^8-60 mQ3^6 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8+18 m3^4 mQ3^2 mU3^10-12 m3^2 mQ3^4 mU3^10-6 mQ3^6 mU3^10))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(32 (32 m3^11 mQ3^2-397 m3^9 mQ3^4+484 m3^7 mQ3^6-210 m3^5 mQ3^8+14 m3^3 mQ3^10-3 m3 mQ3^12+120 m3^7 mQ3^4 msq^2-120 m3^5 mQ3^6 msq^2-30 m3^7 mQ3^2 msq^4+30 m3^5 mQ3^4 msq^4-32 m3^11 mU3^2+348 m3^9 mQ3^2 mU3^2+260 m3^7 mQ3^4 mU3^2-680 m3^5 mQ3^6 mU3^2+340 m3^3 mQ3^8 mU3^2+4 m3 mQ3^10 mU3^2-120 m3^7 mQ3^2 msq^2 mU3^2-120 m3^5 mQ3^4 msq^2 mU3^2+240 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2+30 m3^5 mQ3^2 msq^4 mU3^2-60 m3^3 mQ3^4 msq^4 mU3^2+33 m3^9 mU3^4-676 m3^7 mQ3^2 mU3^4+474 m3^5 mQ3^4 mU3^4+104 m3^3 mQ3^6 mU3^4-175 m3 mQ3^8 mU3^4+240 m3^5 mQ3^2 msq^2 mU3^4-120 m3^3 mQ3^4 msq^2 mU3^4-120 m3 mQ3^6 msq^2 mU3^4-60 m3^5 msq^4 mU3^4+30 m3^3 mQ3^2 msq^4 mU3^4+30 m3 mQ3^4 msq^4 mU3^4-4 m3^7 mU3^6+320 m3^5 mQ3^2 mU3^6-388 m3^3 mQ3^4 mU3^6+152 m3 mQ3^6 mU3^6-120 m3^3 mQ3^2 msq^2 mU3^6+120 m3 mQ3^4 msq^2 mU3^6+30 m3^3 msq^4 mU3^6-30 m3 mQ3^2 msq^4 mU3^6-6 m3^3 mQ3^2 mU3^8+6 m3 mQ3^4 mU3^8) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(16 (492 m3^10 mQ3^2-383 m3^8 mQ3^4-226 m3^6 mQ3^6+192 m3^4 mQ3^8+9 m3^2 mQ3^10-90 m3^6 mQ3^4 msq^2-30 m3^4 mQ3^6 msq^2-24 m3^10 mU3^2-1462 m3^8 mQ3^2 mU3^2+1473 m3^6 mQ3^4 mU3^2+137 m3^4 mQ3^6 mU3^2-391 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2+90 m3^6 mQ3^2 msq^2 mU3^2+210 m3^4 mQ3^4 msq^2 mU3^2+60 m3^2 mQ3^6 msq^2 mU3^2+65 m3^8 mU3^4+1346 m3^6 mQ3^2 mU3^4-1612 m3^4 mQ3^4 mU3^4+338 m3^2 mQ3^6 mU3^4+151 mQ3^8 mU3^4-180 m3^4 mQ3^2 msq^2 mU3^4-150 m3^2 mQ3^4 msq^2 mU3^4-30 mQ3^6 msq^2 mU3^4-41 m3^6 mU3^6-373 m3^4 mQ3^2 mU3^6+471 m3^2 mQ3^4 mU3^6-177 mQ3^6 mU3^6+90 m3^2 mQ3^2 msq^2 mU3^6+30 mQ3^4 msq^2 mU3^6+9 m3^2 mQ3^2 mU3^8+3 mQ3^4 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(128 (11 m3^9-37 m3^7 mQ3^2-48 m3^5 mQ3^4+66 m3^3 mQ3^6-3 m3 mQ3^8+30 m3^3 mQ3^4 msq^2+7 m3^7 mU3^2+150 m3^5 mQ3^2 mU3^2-47 m3^3 mQ3^4 mU3^2-53 m3 mQ3^6 mU3^2-30 m3^3 mQ3^2 msq^2 mU3^2-30 m3 mQ3^4 msq^2 mU3^2-14 m3^5 mU3^4-117 m3^3 mQ3^2 mU3^4+82 m3 mQ3^4 mU3^4+30 m3 mQ3^2 msq^2 mU3^4+3 m3 mQ3^2 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(8 (176 m3^16 mQ3^2-2586 m3^14 mQ3^4+5691 m3^12 mQ3^6-4056 m3^10 mQ3^8+416 m3^8 mQ3^10+329 m3^6 mQ3^12+47 m3^4 mQ3^14-9 m3^2 mQ3^16+60 m3^12 mQ3^4 msq^2+210 m3^10 mQ3^6 msq^2-240 m3^8 mQ3^8 msq^2-30 m3^6 mQ3^10 msq^2-90 m3^10 mQ3^4 msq^4+60 m3^8 mQ3^6 msq^4+30 m3^6 mQ3^8 msq^4-176 m3^16 mU3^2-296 m3^14 mQ3^2 mU3^2+8817 m3^12 mQ3^4 mU3^2-19237 m3^10 mQ3^6 mU3^2+13306 m3^8 mQ3^8 mU3^2-1117 m3^6 mQ3^10 mU3^2-1384 m3^4 mQ3^12 mU3^2+50 m3^2 mQ3^14 mU3^2-3 mQ3^16 mU3^2-120 m3^12 mQ3^2 msq^2 mU3^2-420 m3^10 mQ3^4 msq^2 mU3^2-270 m3^8 mQ3^6 msq^2 mU3^2+720 m3^6 mQ3^8 msq^2 mU3^2+90 m3^4 mQ3^10 msq^2 mU3^2+180 m3^10 mQ3^2 msq^4 mU3^2+150 m3^8 mQ3^4 msq^4 mU3^2-240 m3^6 mQ3^6 msq^4 mU3^2-90 m3^4 mQ3^8 msq^4 mU3^2+498 m3^14 mU3^4+357 m3^12 mQ3^2 mU3^4-12407 m3^10 mQ3^4 mU3^4+25376 m3^8 mQ3^6 mU3^4-17186 m3^6 mQ3^8 mU3^4+2382 m3^4 mQ3^10 mU3^4+1061 m3^2 mQ3^12 mU3^4-mQ3^14 mU3^4+60 m3^12 msq^2 mU3^4+390 m3^10 mQ3^2 msq^2 mU3^4+780 m3^8 mQ3^4 msq^2 mU3^4-420 m3^6 mQ3^6 msq^2 mU3^4-720 m3^4 mQ3^8 msq^2 mU3^4-90 m3^2 mQ3^10 msq^2 mU3^4-90 m3^10 msq^4 mU3^4-480 m3^8 mQ3^2 msq^4 mU3^4+120 m3^6 mQ3^4 msq^4 mU3^4+360 m3^4 mQ3^6 msq^4 mU3^4+90 m3^2 mQ3^8 msq^4 mU3^4-481 m3^12 mU3^6-691 m3^10 mQ3^2 mU3^6+8970 m3^8 mQ3^4 mU3^6-15256 m3^6 mQ3^6 mU3^6+8973 m3^4 mQ3^8 mU3^6-1253 m3^2 mQ3^10 mU3^6-342 mQ3^12 mU3^6-180 m3^10 msq^2 mU3^6-450 m3^8 mQ3^2 msq^2 mU3^6-420 m3^6 mQ3^4 msq^2 mU3^6+780 m3^4 mQ3^6 msq^2 mU3^6+240 m3^2 mQ3^8 msq^2 mU3^6+30 mQ3^10 msq^2 mU3^6+270 m3^8 msq^4 mU3^6+360 m3^6 mQ3^2 msq^4 mU3^6-360 m3^4 mQ3^4 msq^4 mU3^6-240 m3^2 mQ3^6 msq^4 mU3^6-30 mQ3^8 msq^4 mU3^6+151 m3^10 mU3^8+816 m3^8 mQ3^2 mU3^8-3855 m3^6 mQ3^4 mU3^8+4746 m3^4 mQ3^6 mU3^8-2062 m3^2 mQ3^8 mU3^8+244 mQ3^10 mU3^8+180 m3^8 msq^2 mU3^8+210 m3^6 mQ3^2 msq^2 mU3^8-120 m3^4 mQ3^4 msq^2 mU3^8-270 m3^2 mQ3^6 msq^2 mU3^8-270 m3^6 msq^4 mU3^8+210 m3^2 mQ3^4 msq^4 mU3^8+60 mQ3^6 msq^4 mU3^8-4 m3^8 mU3^10-275 m3^6 mQ3^2 mU3^10+731 m3^4 mQ3^4 mU3^10-645 m3^2 mQ3^6 mU3^10+185 mQ3^8 mU3^10-60 m3^6 msq^2 mU3^10-30 m3^4 mQ3^2 msq^2 mU3^10+120 m3^2 mQ3^4 msq^2 mU3^10-30 mQ3^6 msq^2 mU3^10+90 m3^4 msq^4 mU3^10-60 m3^2 mQ3^2 msq^4 mU3^10-30 mQ3^4 msq^4 mU3^10+9 m3^4 mQ3^2 mU3^12-6 m3^2 mQ3^4 mU3^12-3 mQ3^6 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)-(64 (30 m3^11 mQ3^2+179 m3^9 mQ3^4-442 m3^7 mQ3^6+225 m3^5 mQ3^8-11 m3^3 mQ3^10+3 m3 mQ3^12-90 m3^7 mQ3^4 msq^2+90 m3^5 mQ3^6 msq^2+30 m3^7 mQ3^2 msq^4-30 m3^5 mQ3^4 msq^4+34 m3^11 mU3^2-36 m3^9 mQ3^2 mU3^2-470 m3^7 mQ3^4 mU3^2+923 m3^5 mQ3^6 mU3^2-396 m3^3 mQ3^8 mU3^2-7 m3 mQ3^10 mU3^2+30 m3^7 mQ3^2 msq^2 mU3^2+150 m3^5 mQ3^4 msq^2 mU3^2-180 m3^3 mQ3^6 msq^2 mU3^2-30 m3^7 msq^4 mU3^2-30 m3^5 mQ3^2 msq^4 mU3^2+60 m3^3 mQ3^4 msq^4 mU3^2-69 m3^9 mU3^4-14 m3^7 mQ3^2 mU3^4+427 m3^5 mQ3^4 mU3^4-608 m3^3 mQ3^6 mU3^4+216 m3 mQ3^8 mU3^4-60 m3^5 mQ3^2 msq^2 mU3^4-30 m3^3 mQ3^4 msq^2 mU3^4+90 m3 mQ3^6 msq^2 mU3^4+60 m3^5 msq^4 mU3^4-30 m3^3 mQ3^2 msq^4 mU3^4-30 m3 mQ3^4 msq^4 mU3^4+38 m3^7 mU3^6+5 m3^5 mQ3^2 mU3^6-94 m3^3 mQ3^4 mU3^6+67 m3 mQ3^6 mU3^6+30 m3^3 mQ3^2 msq^2 mU3^6-30 m3 mQ3^4 msq^2 mU3^6-30 m3^3 msq^4 mU3^6+30 m3 mQ3^2 msq^4 mU3^6-3 m3^3 mQ3^2 mU3^8+3 m3 mQ3^4 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)-(2560 m3^2 mQ3^4 Xt^6)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)+(-(1/(3 (m3^2-mQ3^2)^3))80 (3 m3^6-7 m3^4 mQ3^2+6 m3^2 mQ3^4-2 mQ3^6-6 m3^4 msq^2-18 m3^2 mQ3^2 msq^2+9 m3^2 msq^4+3 mQ3^2 msq^4)+(320 (2 m3^5-3 m3^3 mQ3^2+m3 mQ3^4-12 m3 mQ3^2 msq^2+6 m3 msq^4) Xt)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+((160 (2 m3^4 msq^2+6 m3^2 mQ3^2 msq^2-3 m3^2 msq^4-mQ3^2 msq^4))/((m3^2-mQ3^2)^3 (mQ3^2-mU3^2))-(160 (4 mQ3^6-2 mQ3^4 mU3^2+m3^4 (3 mQ3^2-mU3^2)+m3^2 (-8 mQ3^4+4 mQ3^2 mU3^2)))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)) Xt^2+(-((1280 (2 m3 mQ3^2 msq^2-m3 msq^4))/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2))+(320 m3 (2 m3^4-mQ3^4+3 mQ3^2 mU3^2-m3^2 (3 mQ3^2+mU3^2)))/(3 (m3^2-mQ3^2) (mQ3^2-mU3^2)^3)) Xt^3+(-((80 (2 m3^4 msq^2+6 m3^2 mQ3^2 msq^2-3 m3^2 msq^4-mQ3^2 msq^4) (mQ3^2+mU3^2))/((m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^3))+(80 (5 mQ3^8+8 m3^4 mQ3^2 mU3^2+4 mQ3^6 mU3^2-mQ3^4 mU3^4+2 m3^6 (mQ3^2-mU3^2)-2 m3^2 (4 mQ3^6+5 mQ3^4 mU3^2-mQ3^2 mU3^4)))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) Xt^4+(320 (m3^3 mQ3^2-m3 mQ3^4+12 m3 mQ3^2 msq^2-6 m3 msq^4) (mQ3^2+mU3^2) Xt^5)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) Log[msq^2/MR^2]+(-((4 (128 m3^16-550 m3^14 mQ3^2+670 m3^12 mQ3^4-247 m3^10 mQ3^6-64 m3^8 mQ3^8+100 m3^6 mQ3^10-38 m3^4 mQ3^12+9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-346 m3^14 mU3^2+1542 m3^12 mQ3^2 mU3^2-1619 m3^10 mQ3^4 mU3^2+199 m3^8 mQ3^6 mU3^2+324 m3^6 mQ3^8 mU3^2-88 m3^4 mQ3^10 mU3^2-47 m3^2 mQ3^12 mU3^2+3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+476 m3^12 mU3^4-2307 m3^10 mQ3^2 mU3^4+2585 m3^8 mQ3^4 mU3^4-386 m3^6 mQ3^6 mU3^4-638 m3^4 mQ3^8 mU3^4+317 m3^2 mQ3^10 mU3^4+mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4-307 m3^10 mU3^6+1699 m3^8 mQ3^2 mU3^6-2306 m3^6 mQ3^4 mU3^6+1002 m3^4 mQ3^6 mU3^6-19 m3^2 mQ3^8 mU3^6-101 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6+61 m3^8 mU3^8-420 m3^6 mQ3^2 mU3^8+658 m3^4 mQ3^4 mU3^8-388 m3^2 mQ3^6 mU3^8+97 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(32 (20 m3^11 mQ3^2-91 m3^9 mQ3^4+122 m3^7 mQ3^6-30 m3^5 mQ3^8-8 m3^3 mQ3^10+3 m3 mQ3^12+60 m3^7 mQ3^4 msq^2-60 m3^5 mQ3^6 msq^2-30 m3^7 mQ3^2 msq^4+30 m3^5 mQ3^4 msq^4-20 m3^11 mU3^2+51 m3^9 mQ3^2 mU3^2+65 m3^7 mQ3^4 mU3^2-219 m3^5 mQ3^6 mU3^2+85 m3^3 mQ3^8 mU3^2-10 m3 mQ3^10 mU3^2-60 m3^7 mQ3^2 msq^2 mU3^2-60 m3^5 mQ3^4 msq^2 mU3^2+120 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2+30 m3^5 mQ3^2 msq^4 mU3^2-60 m3^3 mQ3^4 msq^4 mU3^2+56 m3^9 mU3^4-244 m3^7 mQ3^2 mU3^4+278 m3^5 mQ3^4 mU3^4-42 m3^3 mQ3^6 mU3^4+120 m3^5 mQ3^2 msq^2 mU3^4-60 m3^3 mQ3^4 msq^2 mU3^4-60 m3 mQ3^6 msq^2 mU3^4-60 m3^5 msq^4 mU3^4+30 m3^3 mQ3^2 msq^4 mU3^4+30 m3 mQ3^4 msq^4 mU3^4-7 m3^7 mU3^6+67 m3^5 mQ3^2 mU3^6-99 m3^3 mQ3^4 mU3^6+23 m3 mQ3^6 mU3^6-60 m3^3 mQ3^2 msq^2 mU3^6+60 m3 mQ3^4 msq^2 mU3^6+30 m3^3 msq^4 mU3^6-30 m3 mQ3^2 msq^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(8 (128 m3^16-558 m3^14 mQ3^2+462 m3^12 mQ3^4+241 m3^10 mQ3^6-256 m3^8 mQ3^8-96 m3^6 mQ3^10+114 m3^4 mQ3^12-27 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-846 m3^14 mU3^2+3590 m3^12 mQ3^2 mU3^2-4419 m3^10 mQ3^4 mU3^2+1383 m3^8 mQ3^6 mU3^2+636 m3^6 mQ3^8 mU3^2-508 m3^4 mQ3^10 mU3^2+141 m3^2 mQ3^12 mU3^2-9 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+1200 m3^12 mU3^4-5047 m3^10 mQ3^2 mU3^4+6841 m3^8 mQ3^4 mU3^4-3782 m3^6 mQ3^6 mU3^4+1010 m3^4 mQ3^8 mU3^4-171 m3^2 mQ3^10 mU3^4-3 mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4-387 m3^10 mU3^6+1575 m3^8 mQ3^2 mU3^6-1854 m3^6 mQ3^4 mU3^6+742 m3^4 mQ3^6 mU3^6-151 m3^2 mQ3^8 mU3^6+43 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-83 m3^8 mU3^8+308 m3^6 mQ3^2 mU3^8-514 m3^4 mQ3^4 mU3^8+396 m3^2 mQ3^6 mU3^8-99 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(32 (2 m3^13-43 m3^11 mQ3^2+243 m3^9 mQ3^4-405 m3^7 mQ3^6+201 m3^5 mQ3^8-48 m3^3 mQ3^10+18 m3 mQ3^12-120 m3^7 mQ3^4 msq^2+120 m3^5 mQ3^6 msq^2+60 m3^7 mQ3^2 msq^4-60 m3^5 mQ3^4 msq^4-33 m3^11 mU3^2+403 m3^9 mQ3^2 mU3^2-1127 m3^7 mQ3^4 mU3^2+1201 m3^5 mQ3^6 mU3^2-288 m3^3 mQ3^8 mU3^2-60 m3 mQ3^10 mU3^2+120 m3^7 mQ3^2 msq^2 mU3^2+120 m3^5 mQ3^4 msq^2 mU3^2-240 m3^3 mQ3^6 msq^2 mU3^2-60 m3^7 msq^4 mU3^2-60 m3^5 mQ3^2 msq^4 mU3^2+120 m3^3 mQ3^4 msq^4 mU3^2-40 m3^9 mU3^4-233 m3^7 mQ3^2 mU3^4+959 m3^5 mQ3^4 mU3^4-1143 m3^3 mQ3^6 mU3^4+361 m3 mQ3^8 mU3^4-240 m3^5 mQ3^2 msq^2 mU3^4+120 m3^3 mQ3^4 msq^2 mU3^4+120 m3 mQ3^6 msq^2 mU3^4+120 m3^5 msq^4 mU3^4-60 m3^3 mQ3^2 msq^4 mU3^4-60 m3 mQ3^4 msq^4 mU3^4+61 m3^7 mU3^6-155 m3^5 mQ3^2 mU3^6+123 m3^3 mQ3^4 mU3^6+3 m3 mQ3^6 mU3^6+120 m3^3 mQ3^2 msq^2 mU3^6-120 m3 mQ3^4 msq^2 mU3^6-60 m3^3 msq^4 mU3^6+60 m3 mQ3^2 msq^4 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(4 (124 m3^16 mQ3^2-540 m3^14 mQ3^4-178 m3^12 mQ3^6+1549 m3^10 mQ3^8-1004 m3^8 mQ3^10-30 m3^6 mQ3^12+114 m3^4 mQ3^14-27 m3^2 mQ3^16+60 m3^12 mQ3^4 msq^2+120 m3^10 mQ3^6 msq^2-180 m3^8 mQ3^8 msq^2-90 m3^10 mQ3^4 msq^4+60 m3^8 mQ3^6 msq^4+30 m3^6 mQ3^8 msq^4+1156 m3^16 mU3^2-5372 m3^14 mQ3^2 mU3^2+9158 m3^12 mQ3^4 mU3^2-3530 m3^10 mQ3^6 mU3^2-4889 m3^8 mQ3^8 mU3^2+3876 m3^6 mQ3^10 mU3^2-528 m3^4 mQ3^12 mU3^2+114 m3^2 mQ3^14 mU3^2-9 mQ3^16 mU3^2-180 m3^10 mQ3^4 msq^2 mU3^2-360 m3^8 mQ3^6 msq^2 mU3^2+540 m3^6 mQ3^8 msq^2 mU3^2+270 m3^8 mQ3^4 msq^4 mU3^2-180 m3^6 mQ3^6 msq^4 mU3^2-90 m3^4 mQ3^8 msq^4 mU3^2-3032 m3^14 mU3^4+14790 m3^12 mQ3^2 mU3^4-25872 m3^10 mQ3^4 mU3^4+15488 m3^8 mQ3^6 mU3^4+2576 m3^6 mQ3^8 mU3^4-3994 m3^4 mQ3^10 mU3^4+72 m3^2 mQ3^12 mU3^4-12 mQ3^14 mU3^4-60 m3^12 msq^2 mU3^4-120 m3^10 mQ3^2 msq^2 mU3^4+360 m3^8 mQ3^4 msq^2 mU3^4+360 m3^6 mQ3^6 msq^2 mU3^4-540 m3^4 mQ3^8 msq^2 mU3^4+90 m3^10 msq^4 mU3^4-60 m3^8 mQ3^2 msq^4 mU3^4-300 m3^6 mQ3^4 msq^4 mU3^4+180 m3^4 mQ3^6 msq^4 mU3^4+90 m3^2 mQ3^8 msq^4 mU3^4+2486 m3^12 mU3^6-13250 m3^10 mQ3^2 mU3^6+25026 m3^8 mQ3^4 mU3^6-18644 m3^6 mQ3^6 mU3^6+2738 m3^4 mQ3^8 mU3^6+1654 m3^2 mQ3^10 mU3^6+6 mQ3^12 mU3^6+180 m3^10 msq^2 mU3^6+360 m3^8 mQ3^2 msq^2 mU3^6-600 m3^6 mQ3^4 msq^2 mU3^6-120 m3^4 mQ3^6 msq^2 mU3^6+180 m3^2 mQ3^8 msq^2 mU3^6-270 m3^8 msq^4 mU3^6+180 m3^6 mQ3^2 msq^4 mU3^6+180 m3^4 mQ3^4 msq^4 mU3^6-60 m3^2 mQ3^6 msq^4 mU3^6-30 mQ3^8 msq^4 mU3^6-545 m3^10 mU3^8+3272 m3^8 mQ3^2 mU3^8-6734 m3^6 mQ3^4 mU3^8+5828 m3^4 mQ3^6 mU3^8-1721 m3^2 mQ3^8 mU3^8-124 mQ3^10 mU3^8-180 m3^8 msq^2 mU3^8-360 m3^6 mQ3^2 msq^2 mU3^8+540 m3^4 mQ3^4 msq^2 mU3^8+270 m3^6 msq^4 mU3^8-180 m3^4 mQ3^2 msq^4 mU3^8-90 m3^2 mQ3^4 msq^4 mU3^8-53 m3^8 mU3^10+316 m3^6 mQ3^2 mU3^10-654 m3^4 mQ3^4 mU3^10+532 m3^2 mQ3^6 mU3^10-133 mQ3^8 mU3^10+60 m3^6 msq^2 mU3^10+120 m3^4 mQ3^2 msq^2 mU3^10-180 m3^2 mQ3^4 msq^2 mU3^10-90 m3^4 msq^4 mU3^10+60 m3^2 mQ3^2 msq^4 mU3^10+30 mQ3^4 msq^4 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(32 (18 m3^11 mQ3^4-149 m3^9 mQ3^6+264 m3^7 mQ3^8-132 m3^5 mQ3^10+24 m3^3 mQ3^12-9 m3 mQ3^14+60 m3^7 mQ3^6 msq^2-60 m3^5 mQ3^8 msq^2-30 m3^7 mQ3^4 msq^4+30 m3^5 mQ3^6 msq^4-32 m3^11 mQ3^2 mU3^2-182 m3^9 mQ3^4 mU3^2+753 m3^7 mQ3^6 mU3^2-823 m3^5 mQ3^8 mU3^2+231 m3^3 mQ3^10 mU3^2+21 m3 mQ3^12 mU3^2-120 m3^5 mQ3^6 msq^2 mU3^2+120 m3^3 mQ3^8 msq^2 mU3^2+60 m3^5 mQ3^4 msq^4 mU3^2-60 m3^3 mQ3^6 msq^4 mU3^2+14 m3^11 mU3^4-155 m3^9 mQ3^2 mU3^4+719 m3^7 mQ3^4 mU3^4-1267 m3^5 mQ3^6 mU3^4+879 m3^3 mQ3^8 mU3^4-190 m3 mQ3^10 mU3^4-60 m3^7 mQ3^2 msq^2 mU3^4+60 m3^5 mQ3^4 msq^2 mU3^4+60 m3^3 mQ3^6 msq^2 mU3^4-60 m3 mQ3^8 msq^2 mU3^4+30 m3^7 msq^4 mU3^4-30 m3^5 mQ3^2 msq^4 mU3^4-30 m3^3 mQ3^4 msq^4 mU3^4+30 m3 mQ3^6 msq^4 mU3^4+6 m3^9 mU3^6+207 m3^7 mQ3^2 mU3^6-681 m3^5 mQ3^4 mU3^6+761 m3^3 mQ3^6 mU3^6-261 m3 mQ3^8 mU3^6+120 m3^5 mQ3^2 msq^2 mU3^6-120 m3^3 mQ3^4 msq^2 mU3^6-60 m3^5 msq^4 mU3^6+60 m3^3 mQ3^2 msq^4 mU3^6-23 m3^7 mU3^8+23 m3^5 mQ3^2 mU3^8+25 m3^3 mQ3^4 mU3^8-41 m3 mQ3^6 mU3^8-60 m3^3 mQ3^2 msq^2 mU3^8+60 m3 mQ3^4 msq^2 mU3^8+30 m3^3 msq^4 mU3^8-30 m3 mQ3^2 msq^4 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^5)-(1280 m3^2 (mQ3^2+mU3^2) (m3^2 mQ3^4+2 m3^2 mQ3^2 mU3^2-3 mQ3^4 mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^5)) Log[mU3^2/MR^2])+((8 (128 m3^14-334 m3^12 mQ3^2+368 m3^10 mQ3^4-89 m3^8 mQ3^6+11 m3^6 mQ3^8-29 m3^4 mQ3^10+9 m3^2 mQ3^12-434 m3^12 mU3^2+1072 m3^10 mQ3^2 mU3^2-1145 m3^8 mQ3^4 mU3^2+100 m3^6 mQ3^6 mU3^2+192 m3^4 mQ3^8 mU3^2-44 m3^2 mQ3^10 mU3^2+3 mQ3^12 mU3^2+480 m3^10 mU3^4-1177 m3^8 mQ3^2 mU3^4+1498 m3^6 mQ3^4 mU3^4-376 m3^4 mQ3^6 mU3^4-42 m3^2 mQ3^8 mU3^4+mQ3^10 mU3^4-149 m3^8 mU3^6+324 m3^6 mQ3^2 mU3^6-598 m3^4 mQ3^4 mU3^6+148 m3^2 mQ3^6 mU3^6+19 mQ3^8 mU3^6-13 m3^6 mU3^8+43 m3^4 mQ3^2 mU3^8+57 m3^2 mQ3^4 mU3^8-23 mQ3^6 mU3^8))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+((256 m3^3 (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))+(64 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt+(128 m3 (2 m3^2-mQ3^2-mU3^2) (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(16 (172 m3^12-372 m3^10 mQ3^2+138 m3^8 mQ3^4+3 m3^6 mQ3^6+20 m3^4 mQ3^8-9 m3^2 mQ3^10-468 m3^10 mU3^2+1084 m3^8 mQ3^2 mU3^2-313 m3^6 mQ3^4 mU3^2-149 m3^4 mQ3^6 mU3^2+41 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2+398 m3^8 mU3^4-1151 m3^6 mQ3^2 mU3^4+411 m3^4 mQ3^4 mU3^4+55 m3^2 mQ3^6 mU3^4-mQ3^8 mU3^4-59 m3^6 mU3^6+409 m3^4 mQ3^2 mU3^6-129 m3^2 mQ3^4 mU3^6-29 mQ3^6 mU3^6-31 m3^4 mU3^8-30 m3^2 mQ3^2 mU3^8+13 mQ3^4 mU3^8) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(128 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4) Xt^5)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^3)) PolyLog[2,1-m3^2/mQ3^2]+((80 (m3^2-msq^2) (2 m3^10-8 m3^8 mQ3^2+3 m3^6 mQ3^4-m3^4 mQ3^6+6 m3^8 msq^2-8 m3^6 mQ3^2 msq^2+9 m3^4 mQ3^4 msq^2-3 m3^2 mQ3^6 msq^2-8 m3^8 mU3^2+30 m3^6 mQ3^2 mU3^2-15 m3^4 mQ3^4 mU3^2+5 m3^2 mQ3^6 mU3^2-8 m3^6 msq^2 mU3^2-6 m3^4 mQ3^2 msq^2 mU3^2+3 m3^2 mQ3^4 msq^2 mU3^2-mQ3^6 msq^2 mU3^2+3 m3^6 mU3^4-15 m3^4 mQ3^2 mU3^4+9 m3^4 msq^2 mU3^4+3 m3^2 mQ3^2 msq^2 mU3^4-m3^4 mU3^6+5 m3^2 mQ3^2 mU3^6-3 m3^2 msq^2 mU3^6-mQ3^2 msq^2 mU3^6))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(640 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(160 (m3^2-msq^2) (2 m3^10-8 m3^8 mQ3^2+3 m3^6 mQ3^4-m3^4 mQ3^6+6 m3^8 msq^2-8 m3^6 mQ3^2 msq^2+9 m3^4 mQ3^4 msq^2-3 m3^2 mQ3^6 msq^2-8 m3^8 mU3^2+30 m3^6 mQ3^2 mU3^2-15 m3^4 mQ3^4 mU3^2+5 m3^2 mQ3^6 mU3^2-8 m3^6 msq^2 mU3^2-6 m3^4 mQ3^2 msq^2 mU3^2+3 m3^2 mQ3^4 msq^2 mU3^2-mQ3^6 msq^2 mU3^2+3 m3^6 mU3^4-15 m3^4 mQ3^2 mU3^4+9 m3^4 msq^2 mU3^4+3 m3^2 mQ3^2 msq^2 mU3^4-m3^4 mU3^6+5 m3^2 mQ3^2 mU3^6-3 m3^2 msq^2 mU3^6-mQ3^2 msq^2 mU3^6) Xt^4)/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(1280 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) PolyLog[2,1-m3^2/msq^2]+(-((80 (mQ3^2-msq^2) (2 m3^4+3 m3^2 mQ3^2-mQ3^4-3 m3^2 msq^2-mQ3^2 msq^2))/(m3^2-mQ3^2)^3)+(640 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(160 (mQ3^2-msq^2) (2 m3^4+3 m3^2 mQ3^2-mQ3^4-3 m3^2 msq^2-mQ3^2 msq^2) Xt^4)/((m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^2)-(1280 m3 (mQ3^2-msq^2)^2 Xt^5)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-msq^2/mQ3^2]+(-((8 (128 m3^14-434 m3^12 mQ3^2+480 m3^10 mQ3^4-149 m3^8 mQ3^6-13 m3^6 mQ3^8-334 m3^12 mU3^2+1072 m3^10 mQ3^2 mU3^2-1177 m3^8 mQ3^4 mU3^2+324 m3^6 mQ3^6 mU3^2+43 m3^4 mQ3^8 mU3^2+368 m3^10 mU3^4-1145 m3^8 mQ3^2 mU3^4+1498 m3^6 mQ3^4 mU3^4-598 m3^4 mQ3^6 mU3^4+57 m3^2 mQ3^8 mU3^4-89 m3^8 mU3^6+100 m3^6 mQ3^2 mU3^6-376 m3^4 mQ3^4 mU3^6+148 m3^2 mQ3^6 mU3^6-23 mQ3^8 mU3^6+11 m3^6 mU3^8+192 m3^4 mQ3^2 mU3^8-42 m3^2 mQ3^4 mU3^8+19 mQ3^6 mU3^8-29 m3^4 mU3^10-44 m3^2 mQ3^2 mU3^10+mQ3^4 mU3^10+9 m3^2 mU3^12+3 mQ3^2 mU3^12))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+((256 m3^3 (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(64 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) Xt+(128 m3 (-2 m3^2+mQ3^2+mU3^2) (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+(16 (172 m3^12-468 m3^10 mQ3^2+398 m3^8 mQ3^4-59 m3^6 mQ3^6-31 m3^4 mQ3^8-372 m3^10 mU3^2+1084 m3^8 mQ3^2 mU3^2-1151 m3^6 mQ3^4 mU3^2+409 m3^4 mQ3^6 mU3^2-30 m3^2 mQ3^8 mU3^2+138 m3^8 mU3^4-313 m3^6 mQ3^2 mU3^4+411 m3^4 mQ3^4 mU3^4-129 m3^2 mQ3^6 mU3^4+13 mQ3^8 mU3^4+3 m3^6 mU3^6-149 m3^4 mQ3^2 mU3^6+55 m3^2 mQ3^4 mU3^6-29 mQ3^6 mU3^6+20 m3^4 mU3^8+41 m3^2 mQ3^2 mU3^8-mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(128 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^5)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mU3^2]+Log[msq^2/MR^2] ((80 (2 m3^10 mQ3^2-4 m3^8 mQ3^4+2 m3^6 mQ3^6+2 m3^10 mU3^2-24 m3^8 mQ3^2 mU3^2+41 m3^6 mQ3^4 mU3^2-19 m3^4 mQ3^6 mU3^2-18 m3^6 mQ3^2 msq^2 mU3^2+15 m3^4 mQ3^4 msq^2 mU3^2-9 m3^2 mQ3^6 msq^2 mU3^2-4 m3^8 mU3^4+41 m3^6 mQ3^2 mU3^4-68 m3^4 mQ3^4 mU3^4+31 m3^2 mQ3^6 mU3^4+15 m3^4 mQ3^2 msq^2 mU3^4+12 m3^2 mQ3^4 msq^2 mU3^4-3 mQ3^6 msq^2 mU3^4+2 m3^6 mU3^6-19 m3^4 mQ3^2 mU3^6+31 m3^2 mQ3^4 mU3^6-14 mQ3^6 mU3^6-9 m3^2 mQ3^2 msq^2 mU3^6-3 mQ3^4 msq^2 mU3^6))/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)+(640 m3 msq^2 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))-(320 m3^2 Xt^2)/(3 mQ3^2 mU3^2)-(2560 m3 Xt^3)/(3 (mQ3^2-mU3^2)^2)+(160 (m3^10 mQ3^2-2 m3^8 mQ3^4+m3^6 mQ3^6+m3^10 mU3^2+14 m3^8 mQ3^2 mU3^2-30 m3^6 mQ3^4 mU3^2+15 m3^4 mQ3^6 mU3^2+18 m3^6 mQ3^2 msq^2 mU3^2-15 m3^4 mQ3^4 msq^2 mU3^2+9 m3^2 mQ3^6 msq^2 mU3^2-2 m3^8 mU3^4-30 m3^6 mQ3^2 mU3^4+64 m3^4 mQ3^4 mU3^4-32 m3^2 mQ3^6 mU3^4-15 m3^4 mQ3^2 msq^2 mU3^4-12 m3^2 mQ3^4 msq^2 mU3^4+3 mQ3^6 msq^2 mU3^4+m3^6 mU3^6+15 m3^4 mQ3^2 mU3^6-32 m3^2 mQ3^4 mU3^6+16 mQ3^6 mU3^6+9 m3^2 mQ3^2 msq^2 mU3^6+3 mQ3^4 msq^2 mU3^6) Xt^4)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(1280 m3 msq^2 Xt^5)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+(-((320 m3^4)/(3 (-m3^2+mQ3^2)^2))+(1280 m3^3 Xt)/(3 (m3^2-mQ3^2) (mQ3^2-mU3^2))+(640 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(3 (mQ3^2-mU3^2)^3)-(160 (-2 m3^2+mQ3^2+mU3^2) Xt^4)/(3 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mQ3^2]+(-((320 m3^4)/(3 (m3^2-mU3^2)^2))+(1280 m3^3 Xt)/(3 (m3^2-mU3^2) (-mQ3^2+mU3^2))+(640 m3 (-2 m3^2+mQ3^2+mU3^2) Xt^3)/(3 (mQ3^2-mU3^2)^3)+(160 (-2 m3^2+mQ3^2+mU3^2) Xt^4)/(3 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mU3^2])+Log[m3^2/MR^2] ((8 (132 m3^14 mQ3^2-300 m3^12 mQ3^4+236 m3^10 mQ3^6-68 m3^8 mQ3^8+132 m3^14 mU3^2-1336 m3^12 mQ3^2 mU3^2+2738 m3^10 mQ3^4 mU3^2-2016 m3^8 mQ3^6 mU3^2+519 m3^6 mQ3^8 mU3^2-9 m3^4 mQ3^10 mU3^2+180 m3^10 mQ3^2 msq^2 mU3^2-240 m3^8 mQ3^4 msq^2 mU3^2+270 m3^6 mQ3^6 msq^2 mU3^2-90 m3^4 mQ3^8 msq^2 mU3^2-300 m3^12 mU3^4+2738 m3^10 mQ3^2 mU3^4-5432 m3^8 mQ3^4 mU3^4+3785 m3^6 mQ3^6 mU3^4-860 m3^4 mQ3^8 mU3^4-3 m3^2 mQ3^10 mU3^4-240 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4+90 m3^4 mQ3^6 msq^2 mU3^4-30 m3^2 mQ3^8 msq^2 mU3^4+236 m3^10 mU3^6-2016 m3^8 mQ3^2 mU3^6+3785 m3^6 mQ3^4 mU3^6-2470 m3^4 mQ3^6 mU3^6+513 m3^2 mQ3^8 mU3^6+270 m3^6 mQ3^2 msq^2 mU3^6+90 m3^4 mQ3^4 msq^2 mU3^6-68 m3^8 mU3^8+519 m3^6 mQ3^2 mU3^8-860 m3^4 mQ3^4 mU3^8+513 m3^2 mQ3^6 mU3^8-96 mQ3^8 mU3^8-90 m3^4 mQ3^2 msq^2 mU3^8-30 m3^2 mQ3^4 msq^2 mU3^8-9 m3^4 mQ3^2 mU3^10-3 m3^2 mQ3^4 mU3^10))/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3)-(64 (16 m3^9 mQ3^2-16 m3^7 mQ3^4+16 m3^9 mU3^2-128 m3^7 mQ3^2 mU3^2+112 m3^5 mQ3^4 mU3^2-3 m3^3 mQ3^6 mU3^2+60 m3^5 mQ3^2 msq^2 mU3^2-30 m3^3 mQ3^4 msq^2 mU3^2-16 m3^7 mU3^4+112 m3^5 mQ3^2 mU3^4-90 m3^3 mQ3^4 mU3^4-30 m3^3 mQ3^2 msq^2 mU3^4-3 m3^3 mQ3^2 mU3^6) Xt)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)-(64 (33 m3^10-42 m3^8 mQ3^2+17 m3^6 mQ3^4-42 m3^8 mU3^2+25 m3^6 mQ3^2 mU3^2-7 m3^4 mQ3^4 mU3^2+17 m3^6 mU3^4-7 m3^4 mQ3^2 mU3^4+6 m3^2 mQ3^4 mU3^4) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)+((2048 m3^5)/(3 mQ3^2 (m3^2-mQ3^2) mU3^2 (m3^2-mU3^2))-(512 m3 (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)))/(mQ3^2-mU3^2)^2) Xt^3+(16 (66 m3^14 mQ3^2-150 m3^12 mQ3^4+118 m3^10 mQ3^6-34 m3^8 mQ3^8+66 m3^14 mU3^2+1876 m3^12 mQ3^2 mU3^2-4190 m3^10 mQ3^4 mU3^2+2756 m3^8 mQ3^6 mU3^2-465 m3^6 mQ3^8 mU3^2+9 m3^4 mQ3^10 mU3^2-180 m3^10 mQ3^2 msq^2 mU3^2+240 m3^8 mQ3^4 msq^2 mU3^2-270 m3^6 mQ3^6 msq^2 mU3^2+90 m3^4 mQ3^8 msq^2 mU3^2-150 m3^12 mU3^4-4190 m3^10 mQ3^2 mU3^4+8964 m3^8 mQ3^4 mU3^4-5645 m3^6 mQ3^6 mU3^4+850 m3^4 mQ3^8 mU3^4+3 m3^2 mQ3^10 mU3^4+240 m3^8 mQ3^2 msq^2 mU3^4+180 m3^6 mQ3^4 msq^2 mU3^4-90 m3^4 mQ3^6 msq^2 mU3^4+30 m3^2 mQ3^8 msq^2 mU3^4+118 m3^10 mU3^6+2756 m3^8 mQ3^2 mU3^6-5645 m3^6 mQ3^4 mU3^6+3522 m3^4 mQ3^6 mU3^6-559 m3^2 mQ3^8 mU3^6-270 m3^6 mQ3^2 msq^2 mU3^6-90 m3^4 mQ3^4 msq^2 mU3^6-34 m3^8 mU3^8-465 m3^6 mQ3^2 mU3^8+850 m3^4 mQ3^4 mU3^8-559 m3^2 mQ3^6 mU3^8+120 mQ3^8 mU3^8+90 m3^4 mQ3^2 msq^2 mU3^8+30 m3^2 mQ3^4 msq^2 mU3^8+9 m3^4 mQ3^2 mU3^10+3 m3^2 mQ3^4 mU3^10) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(128 (8 m3^9 mQ3^2-8 m3^7 mQ3^4+8 m3^9 mU3^2+123 m3^7 mQ3^2 mU3^2-115 m3^5 mQ3^4 mU3^2+3 m3^3 mQ3^6 mU3^2-60 m3^5 mQ3^2 msq^2 mU3^2+30 m3^3 mQ3^4 msq^2 mU3^2-8 m3^7 mU3^4-115 m3^5 mQ3^2 mU3^4+101 m3^3 mQ3^4 mU3^4+30 m3^3 mQ3^2 msq^2 mU3^4+3 m3^3 mQ3^2 mU3^6) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+((16 (64 m3^16-245 m3^14 mQ3^2+240 m3^12 mQ3^4-170 m3^10 mQ3^6+110 m3^8 mQ3^8-27 m3^6 mQ3^10-139 m3^14 mU3^2+633 m3^12 mQ3^2 mU3^2-432 m3^10 mQ3^4 mU3^2+162 m3^8 mQ3^6 mU3^2-157 m3^6 mQ3^8 mU3^2+45 m3^4 mQ3^10 mU3^2+87 m3^12 mU3^4-700 m3^10 mQ3^2 mU3^4+400 m3^8 mQ3^4 mU3^4-20 m3^6 mQ3^6 mU3^4+101 m3^4 mQ3^8 mU3^4-36 m3^2 mQ3^10 mU3^4+22 m3^10 mU3^6+316 m3^8 mQ3^2 mU3^6-146 m3^6 mQ3^4 mU3^6-80 m3^4 mQ3^6 mU3^6-12 m3^2 mQ3^8 mU3^6+12 mQ3^10 mU3^6-28 m3^8 mU3^8-34 m3^6 mQ3^2 mU3^8-2 m3^4 mQ3^4 mU3^8+48 m3^2 mQ3^6 mU3^8-12 mQ3^8 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(32 (18 m3^11+195 m3^9 mQ3^2-130 m3^7 mQ3^4+13 m3^5 mQ3^6-19 m3^9 mU3^2-425 m3^7 mQ3^2 mU3^2+263 m3^5 mQ3^4 mU3^2-11 m3^3 mQ3^6 mU3^2+27 m3^7 mU3^4+168 m3^5 mQ3^2 mU3^4-87 m3^3 mQ3^4 mU3^4-12 m3 mQ3^6 mU3^4) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(32 (64 m3^16-502 m3^14 mQ3^2+954 m3^12 mQ3^4-759 m3^10 mQ3^6+302 m3^8 mQ3^8-55 m3^6 mQ3^10-54 m3^14 mU3^2+962 m3^12 mQ3^2 mU3^2-1897 m3^10 mQ3^4 mU3^2+1401 m3^8 mQ3^6 mU3^2-525 m3^6 mQ3^8 mU3^2+97 m3^4 mQ3^10 mU3^2-72 m3^12 mU3^4-545 m3^10 mQ3^2 mU3^4+1199 m3^8 mQ3^4 mU3^4-811 m3^6 mQ3^6 mU3^4+325 m3^4 mQ3^8 mU3^4-72 m3^2 mQ3^10 mU3^4+69 m3^10 mU3^6+79 m3^8 mQ3^2 mU3^6-189 m3^6 mQ3^4 mU3^6+61 m3^4 mQ3^6 mU3^6-60 m3^2 mQ3^8 mU3^6+24 mQ3^10 mU3^6-m3^8 mU3^8-24 m3^6 mQ3^2 mU3^8-7 m3^4 mQ3^4 mU3^8+48 m3^2 mQ3^6 mU3^8-12 mQ3^8 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(64 (42 m3^13-120 m3^11 mQ3^2+239 m3^9 mQ3^4-252 m3^7 mQ3^6+59 m3^5 mQ3^8-6 m3^11 mU3^2-44 m3^9 mQ3^2 mU3^2-67 m3^7 mQ3^4 mU3^2+268 m3^5 mQ3^6 mU3^2-55 m3^3 mQ3^8 mU3^2-67 m3^9 mU3^4+214 m3^7 mQ3^2 mU3^4-195 m3^5 mQ3^4 mU3^4-42 m3^3 mQ3^6 mU3^4-6 m3 mQ3^8 mU3^4+45 m3^7 mU3^6-90 m3^5 mQ3^2 mU3^6+59 m3^3 mQ3^4 mU3^6+18 m3 mQ3^6 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(16 (788 m3^16 mQ3^2-2556 m3^14 mQ3^4+2618 m3^12 mQ3^6-835 m3^10 mQ3^8+58 m3^8 mQ3^10-69 m3^6 mQ3^12-148 m3^16 mU3^2-1796 m3^14 mQ3^2 mU3^2+7366 m3^12 mQ3^4 mU3^2-8472 m3^10 mQ3^6 mU3^2+3031 m3^8 mQ3^8 mU3^2-116 m3^6 mQ3^10 mU3^2+123 m3^4 mQ3^12 mU3^2+336 m3^14 mU3^4+786 m3^12 mQ3^2 mU3^4-6232 m3^10 mQ3^4 mU3^4+8712 m3^8 mQ3^6 mU3^4-3742 m3^6 mQ3^8 mU3^4+238 m3^4 mQ3^10 mU3^4-90 m3^2 mQ3^12 mU3^4-194 m3^12 mU3^6+452 m3^10 mQ3^2 mU3^6+936 m3^8 mQ3^4 mU3^6-2648 m3^6 mQ3^6 mU3^6+1612 m3^4 mQ3^8 mU3^6-180 m3^2 mQ3^10 mU3^6+30 mQ3^12 mU3^6-m3^10 mU3^8-190 m3^8 mQ3^2 mU3^8+479 m3^6 mQ3^4 mU3^8-234 m3^4 mQ3^6 mU3^8-90 m3^2 mQ3^8 mU3^8+24 mQ3^10 mU3^8+13 m3^8 mU3^10-64 m3^6 mQ3^2 mU3^10+37 m3^4 mQ3^4 mU3^10+24 m3^2 mQ3^6 mU3^10-6 mQ3^8 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(32 (36 m3^11 mQ3^2-159 m3^9 mQ3^4+216 m3^7 mQ3^6-61 m3^5 mQ3^8+100 m3^11 mU3^2-392 m3^9 mQ3^2 mU3^2+585 m3^7 mQ3^4 mU3^2-384 m3^5 mQ3^6 mU3^2+59 m3^3 mQ3^8 mU3^2-169 m3^9 mU3^4+508 m3^7 mQ3^2 mU3^4-535 m3^5 mQ3^4 mU3^4+152 m3^3 mQ3^6 mU3^4+12 m3 mQ3^8 mU3^4+75 m3^7 mU3^6-148 m3^5 mQ3^2 mU3^6+93 m3^3 mQ3^4 mU3^6+12 m3 mQ3^6 mU3^6) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)-(2560 m3^4 mQ3^2 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^4)) Log[mQ3^2/MR^2]^2+(-((160 (m3^14 mQ3^2-3 m3^12 mQ3^4+3 m3^10 mQ3^6-m3^8 mQ3^8+m3^14 mU3^2+2 m3^12 mQ3^2 mU3^2-11 m3^10 mQ3^4 mU3^2+3 m3^8 mQ3^6 mU3^2-m3^6 mQ3^8 mU3^2-3 m3^12 mU3^4-11 m3^10 mQ3^2 mU3^4+48 m3^8 mQ3^4 mU3^4-24 m3^6 mQ3^6 mU3^4+8 m3^4 mQ3^8 mU3^4+3 m3^10 mU3^6+3 m3^8 mQ3^2 mU3^6-24 m3^6 mQ3^4 mU3^6-m3^8 mU3^8-m3^6 mQ3^2 mU3^8+8 m3^4 mQ3^4 mU3^8))/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3))-(640 (m3^5 mQ3^2+m3^5 mU3^2-2 m3^3 mQ3^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(320 m3^4 (-m3^2+mQ3^2+mU3^2) Xt^2)/(3 mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2))-(160 (m3^14 mQ3^2-3 m3^12 mQ3^4+3 m3^10 mQ3^6-m3^8 mQ3^8+m3^14 mU3^2-10 m3^12 mQ3^2 mU3^2+32 m3^10 mQ3^4 mU3^2-16 m3^8 mQ3^6 mU3^2+5 m3^6 mQ3^8 mU3^2-3 m3^12 mU3^4+32 m3^10 mQ3^2 mU3^4-102 m3^8 mQ3^4 mU3^4+54 m3^6 mQ3^6 mU3^4-17 m3^4 mQ3^8 mU3^4+3 m3^10 mU3^6-16 m3^8 mQ3^2 mU3^6+54 m3^6 mQ3^4 mU3^6-6 m3^4 mQ3^6 mU3^6+m3^2 mQ3^8 mU3^6-m3^8 mU3^8+5 m3^6 mQ3^2 mU3^8-17 m3^4 mQ3^4 mU3^8+m3^2 mQ3^6 mU3^8) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(640 (m3^7+5 m3^5 mQ3^2+5 m3^5 mU3^2-11 m3^3 mQ3^2 mU3^2) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) Log[msq^2/MR^2]+(-((8 (2 m3^16 mQ3^4-6 m3^14 mQ3^6+6 m3^12 mQ3^8-2 m3^10 mQ3^10-128 m3^16 mQ3^2 mU3^2-760 m3^14 mQ3^4 mU3^2+2188 m3^12 mQ3^6 mU3^2-1902 m3^10 mQ3^8 mU3^2+547 m3^8 mQ3^10 mU3^2-9 m3^6 mQ3^12 mU3^2+180 m3^12 mQ3^4 msq^2 mU3^2-240 m3^10 mQ3^6 msq^2 mU3^2+270 m3^8 mQ3^8 msq^2 mU3^2-90 m3^6 mQ3^10 msq^2 mU3^2-2 m3^16 mU3^4+1462 m3^14 mQ3^2 mU3^4-610 m3^12 mQ3^4 mU3^4-3080 m3^10 mQ3^6 mU3^4+3528 m3^8 mQ3^8 mU3^4-972 m3^6 mQ3^10 mU3^4+6 m3^4 mQ3^12 mU3^4-180 m3^12 mQ3^2 msq^2 mU3^4-180 m3^10 mQ3^4 msq^2 mU3^4-210 m3^8 mQ3^6 msq^2 mU3^4-90 m3^6 mQ3^8 msq^2 mU3^4+60 m3^4 mQ3^10 msq^2 mU3^4+72 m3^14 mU3^6-3332 m3^12 mQ3^2 mU3^6+3634 m3^10 mQ3^4 mU3^6+1602 m3^8 mQ3^6 mU3^6-3698 m3^6 mQ3^8 mU3^6+1019 m3^4 mQ3^10 mU3^6+3 m3^2 mQ3^12 mU3^6+420 m3^10 mQ3^2 msq^2 mU3^6+450 m3^8 mQ3^4 msq^2 mU3^6+450 m3^6 mQ3^6 msq^2 mU3^6-150 m3^4 mQ3^8 msq^2 mU3^6+30 m3^2 mQ3^10 msq^2 mU3^6-172 m3^12 mU3^8+3774 m3^10 mQ3^2 mU3^8-5506 m3^8 mQ3^4 mU3^8+1682 m3^6 mQ3^6 mU3^8+1522 m3^4 mQ3^8 mU3^8-540 m3^2 mQ3^10 mU3^8-510 m3^8 mQ3^2 msq^2 mU3^8-630 m3^6 mQ3^4 msq^2 mU3^8-30 m3^4 mQ3^6 msq^2 mU3^8-30 m3^2 mQ3^8 msq^2 mU3^8+136 m3^10 mU3^10-2057 m3^8 mQ3^2 mU3^10+3339 m3^6 mQ3^4 mU3^10-2026 m3^4 mQ3^6 mU3^10+84 m3^2 mQ3^8 mU3^10+84 mQ3^10 mU3^10+360 m3^6 mQ3^2 msq^2 mU3^10+210 m3^4 mQ3^4 msq^2 mU3^10+30 m3^2 mQ3^6 msq^2 mU3^10-34 m3^8 mU3^12+426 m3^6 mQ3^2 mU3^12-640 m3^4 mQ3^4 mU3^12+456 m3^2 mQ3^6 mU3^12-84 mQ3^8 mU3^12-90 m3^4 mQ3^2 msq^2 mU3^12-30 m3^2 mQ3^4 msq^2 mU3^12-9 m3^4 mQ3^2 mU3^14-3 m3^2 mQ3^4 mU3^14))/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)))+(64 (4 m3^11 mQ3^2-130 m3^9 mQ3^4+123 m3^7 mQ3^6-3 m3^5 mQ3^8+60 m3^7 mQ3^4 msq^2-30 m3^5 mQ3^6 msq^2+16 m3^11 mU3^2+216 m3^9 mQ3^2 mU3^2-106 m3^7 mQ3^4 mU3^2-68 m3^5 mQ3^6 mU3^2+3 m3^3 mQ3^8 mU3^2-60 m3^7 mQ3^2 msq^2 mU3^2-60 m3^5 mQ3^4 msq^2 mU3^2+30 m3^3 mQ3^6 msq^2 mU3^2-32 m3^9 mU3^4-223 m3^7 mQ3^2 mU3^4+111 m3^5 mQ3^4 mU3^4+37 m3^3 mQ3^6 mU3^4+90 m3^5 mQ3^2 msq^2 mU3^4+16 m3^7 mU3^6+86 m3^5 mQ3^2 mU3^6-35 m3^3 mQ3^4 mU3^6-12 m3 mQ3^6 mU3^6-30 m3^3 mQ3^2 msq^2 mU3^6-3 m3^3 mQ3^2 mU3^8) Xt)/(3 mQ3^2 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(16 (2 m3^14 mQ3^2-6 m3^12 mQ3^4+6 m3^10 mQ3^6-2 m3^8 mQ3^8-2 m3^14 mU3^2-1472 m3^12 mQ3^2 mU3^2+3280 m3^10 mQ3^4 mU3^2-2436 m3^8 mQ3^6 mU3^2+587 m3^6 mQ3^8 mU3^2-9 m3^4 mQ3^10 mU3^2+180 m3^10 mQ3^2 msq^2 mU3^2-240 m3^8 mQ3^4 msq^2 mU3^2+270 m3^6 mQ3^6 msq^2 mU3^2-90 m3^4 mQ3^8 msq^2 mU3^2+70 m3^12 mU3^4+2660 m3^10 mQ3^2 mU3^4-5696 m3^8 mQ3^4 mU3^4+3987 m3^6 mQ3^6 mU3^4-850 m3^4 mQ3^8 mU3^4-3 m3^2 mQ3^10 mU3^4-240 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4+90 m3^4 mQ3^6 msq^2 mU3^4-30 m3^2 mQ3^8 msq^2 mU3^4-102 m3^10 mU3^6-1692 m3^8 mQ3^2 mU3^6+3519 m3^6 mQ3^4 mU3^6-2430 m3^4 mQ3^6 mU3^6+513 m3^2 mQ3^8 mU3^6+270 m3^6 mQ3^2 msq^2 mU3^6+90 m3^4 mQ3^4 msq^2 mU3^6+34 m3^8 mU3^8+395 m3^6 mQ3^2 mU3^8-734 m3^4 mQ3^4 mU3^8+489 m3^2 mQ3^6 mU3^8-96 mQ3^8 mU3^8-90 m3^4 mQ3^2 msq^2 mU3^8-30 m3^2 mQ3^4 msq^2 mU3^8-9 m3^4 mQ3^2 mU3^10-3 m3^2 mQ3^4 mU3^10) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(128 (22 m3^11 mQ3^2+10 m3^9 mQ3^4-193 m3^7 mQ3^6+153 m3^5 mQ3^8-3 m3^3 mQ3^10+60 m3^5 mQ3^6 msq^2-30 m3^3 mQ3^8 msq^2+6 m3^9 mQ3^2 mU3^2+238 m3^7 mQ3^4 mU3^2-85 m3^5 mQ3^6 mU3^2-91 m3^3 mQ3^8 mU3^2-120 m3^5 mQ3^4 msq^2 mU3^2+30 m3^3 mQ3^6 msq^2 mU3^2+16 m3^9 mU3^4-217 m3^7 mQ3^2 mU3^4-17 m3^5 mQ3^4 mU3^4+106 m3^3 mQ3^6 mU3^4+6 m3 mQ3^8 mU3^4+60 m3^5 mQ3^2 msq^2 mU3^4+30 m3^3 mQ3^4 msq^2 mU3^4-16 m3^7 mU3^6+165 m3^5 mQ3^2 mU3^6-115 m3^3 mQ3^4 mU3^6+18 m3 mQ3^6 mU3^6-30 m3^3 mQ3^2 msq^2 mU3^6-3 m3^3 mQ3^2 mU3^8) Xt^3)/(3 mQ3^2 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(8 (2 m3^16 mQ3^4-6 m3^14 mQ3^6+6 m3^12 mQ3^8-2 m3^10 mQ3^10+80 m3^16 mQ3^2 mU3^2-2152 m3^14 mQ3^4 mU3^2+4406 m3^12 mQ3^6 mU3^2-2996 m3^10 mQ3^8 mU3^2+643 m3^8 mQ3^10 mU3^2-9 m3^6 mQ3^12 mU3^2+180 m3^12 mQ3^4 msq^2 mU3^2-240 m3^10 mQ3^6 msq^2 mU3^2+270 m3^8 mQ3^8 msq^2 mU3^2-90 m3^6 mQ3^10 msq^2 mU3^2-2 m3^16 mU3^4-3178 m3^14 mQ3^2 mU3^4+12962 m3^12 mQ3^4 mU3^4-17588 m3^10 mQ3^6 mU3^4+9676 m3^8 mQ3^8 mU3^4-1828 m3^6 mQ3^10 mU3^4+6 m3^4 mQ3^12 mU3^4+180 m3^12 mQ3^2 msq^2 mU3^4-660 m3^10 mQ3^4 msq^2 mU3^4+330 m3^8 mQ3^6 msq^2 mU3^4-270 m3^6 mQ3^8 msq^2 mU3^4+60 m3^4 mQ3^10 msq^2 mU3^4+72 m3^14 mU3^6+9110 m3^12 mQ3^2 mU3^6-25584 m3^10 mQ3^4 mU3^6+25638 m3^8 mQ3^6 mU3^6-10862 m3^6 mQ3^8 mU3^6+1723 m3^4 mQ3^10 mU3^6+3 m3^2 mQ3^12 mU3^6-420 m3^10 mQ3^2 msq^2 mU3^6+570 m3^8 mQ3^4 msq^2 mU3^6+90 m3^6 mQ3^6 msq^2 mU3^6-30 m3^4 mQ3^8 msq^2 mU3^6+30 m3^2 mQ3^10 msq^2 mU3^6-172 m3^12 mU3^8-9926 m3^10 mQ3^2 mU3^8+22990 m3^8 mQ3^4 mU3^8-18070 m3^6 mQ3^6 mU3^8+5474 m3^4 mQ3^8 mU3^8-616 m3^2 mQ3^10 mU3^8+510 m3^8 mQ3^2 msq^2 mU3^8-90 m3^6 mQ3^4 msq^2 mU3^8-210 m3^4 mQ3^6 msq^2 mU3^8+30 m3^2 mQ3^8 msq^2 mU3^8+136 m3^10 mU3^10+4623 m3^8 mQ3^2 mU3^10-9553 m3^6 mQ3^4 mU3^10+6600 m3^4 mQ3^6 mU3^10-1602 m3^2 mQ3^8 mU3^10+96 mQ3^10 mU3^10-360 m3^6 mQ3^2 msq^2 mU3^10-30 m3^4 mQ3^4 msq^2 mU3^10+30 m3^2 mQ3^6 msq^2 mU3^10-34 m3^8 mU3^12-702 m3^6 mQ3^2 mU3^12+1268 m3^4 mQ3^4 mU3^12-884 m3^2 mQ3^6 mU3^12+240 mQ3^8 mU3^12+90 m3^4 mQ3^2 msq^2 mU3^12+30 m3^2 mQ3^4 msq^2 mU3^12+9 m3^4 mQ3^2 mU3^14+3 m3^2 mQ3^4 mU3^14) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^3)+(64 (52 m3^11 mQ3^2-276 m3^9 mQ3^4+214 m3^7 mQ3^6-3 m3^5 mQ3^8+60 m3^7 mQ3^4 msq^2-30 m3^5 mQ3^6 msq^2+16 m3^11 mU3^2-362 m3^9 mQ3^2 mU3^2+770 m3^7 mQ3^4 mU3^2-407 m3^5 mQ3^6 mU3^2+3 m3^3 mQ3^8 mU3^2+60 m3^7 mQ3^2 msq^2 mU3^2-120 m3^5 mQ3^4 msq^2 mU3^2+30 m3^3 mQ3^6 msq^2 mU3^2-32 m3^9 mU3^4+558 m3^7 mQ3^2 mU3^4-737 m3^5 mQ3^4 mU3^4+213 m3^3 mQ3^6 mU3^4-90 m3^5 mQ3^2 msq^2 mU3^4+60 m3^3 mQ3^4 msq^2 mU3^4+16 m3^7 mU3^6-219 m3^5 mQ3^2 mU3^6+179 m3^3 mQ3^4 mU3^6+12 m3 mQ3^6 mU3^6+30 m3^3 mQ3^2 msq^2 mU3^6+3 m3^3 mQ3^2 mU3^8) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(5120 m3^4 mU3^2 Xt^6)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+((160 (3 m3^12-13 m3^10 mQ3^2+6 m3^8 mQ3^4-2 m3^6 mQ3^6-11 m3^10 mU3^2+45 m3^8 mQ3^2 mU3^2-24 m3^6 mQ3^4 mU3^2+8 m3^4 mQ3^6 mU3^2+3 m3^8 mU3^4-21 m3^6 mQ3^2 mU3^4-m3^6 mU3^6+7 m3^4 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+((320 m3^3 (2 m3^2-mQ3^2-mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))+(640 (m3^5 mQ3^2+m3^5 mU3^2-2 m3^3 mQ3^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Xt-(320 (2 m3^12-10 m3^10 mQ3^2+3 m3^8 mQ3^4-m3^6 mQ3^6-10 m3^10 mU3^2+42 m3^8 mQ3^2 mU3^2-21 m3^6 mQ3^4 mU3^2+7 m3^4 mQ3^6 mU3^2+3 m3^8 mU3^4-21 m3^6 mQ3^2 mU3^4-m3^6 mU3^6+7 m3^4 mQ3^2 mU3^6) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(-((1280 (m3^5 mQ3^2+m3^5 mU3^2-2 m3^3 mQ3^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)))+(640 m3^3 (2 m3^4-mQ3^4+4 mQ3^2 mU3^2-mU3^4-2 m3^2 (mQ3^2+mU3^2)))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) Xt^3+(160 (mQ3^2+mU3^2) (2 m3^12-10 m3^10 mQ3^2+3 m3^8 mQ3^4-m3^6 mQ3^6-10 m3^10 mU3^2+42 m3^8 mQ3^2 mU3^2-21 m3^6 mQ3^4 mU3^2+7 m3^4 mQ3^6 mU3^2+3 m3^8 mU3^4-21 m3^6 mQ3^2 mU3^4-m3^6 mU3^6+7 m3^4 mQ3^2 mU3^6) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(320 (mQ3^2+mU3^2) (m3^7+5 m3^5 mQ3^2+5 m3^5 mU3^2-11 m3^3 mQ3^2 mU3^2) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[msq^2/MR^2]) Log[mU3^2/MR^2]+(-((16 (64 m3^16-139 m3^14 mQ3^2+87 m3^12 mQ3^4+22 m3^10 mQ3^6-28 m3^8 mQ3^8-245 m3^14 mU3^2+633 m3^12 mQ3^2 mU3^2-700 m3^10 mQ3^4 mU3^2+316 m3^8 mQ3^6 mU3^2-34 m3^6 mQ3^8 mU3^2+240 m3^12 mU3^4-432 m3^10 mQ3^2 mU3^4+400 m3^8 mQ3^4 mU3^4-146 m3^6 mQ3^6 mU3^4-2 m3^4 mQ3^8 mU3^4-170 m3^10 mU3^6+162 m3^8 mQ3^2 mU3^6-20 m3^6 mQ3^4 mU3^6-80 m3^4 mQ3^6 mU3^6+48 m3^2 mQ3^8 mU3^6+110 m3^8 mU3^8-157 m3^6 mQ3^2 mU3^8+101 m3^4 mQ3^4 mU3^8-12 m3^2 mQ3^6 mU3^8-12 mQ3^8 mU3^8-27 m3^6 mU3^10+45 m3^4 mQ3^2 mU3^10-36 m3^2 mQ3^4 mU3^10+12 mQ3^6 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)))-(32 (18 m3^11-19 m3^9 mQ3^2+27 m3^7 mQ3^4+195 m3^9 mU3^2-425 m3^7 mQ3^2 mU3^2+168 m3^5 mQ3^4 mU3^2-130 m3^7 mU3^4+263 m3^5 mQ3^2 mU3^4-87 m3^3 mQ3^4 mU3^4+13 m3^5 mU3^6-11 m3^3 mQ3^2 mU3^6-12 m3 mQ3^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(32 (64 m3^16-54 m3^14 mQ3^2-72 m3^12 mQ3^4+69 m3^10 mQ3^6-m3^8 mQ3^8-502 m3^14 mU3^2+962 m3^12 mQ3^2 mU3^2-545 m3^10 mQ3^4 mU3^2+79 m3^8 mQ3^6 mU3^2-24 m3^6 mQ3^8 mU3^2+954 m3^12 mU3^4-1897 m3^10 mQ3^2 mU3^4+1199 m3^8 mQ3^4 mU3^4-189 m3^6 mQ3^6 mU3^4-7 m3^4 mQ3^8 mU3^4-759 m3^10 mU3^6+1401 m3^8 mQ3^2 mU3^6-811 m3^6 mQ3^4 mU3^6+61 m3^4 mQ3^6 mU3^6+48 m3^2 mQ3^8 mU3^6+302 m3^8 mU3^8-525 m3^6 mQ3^2 mU3^8+325 m3^4 mQ3^4 mU3^8-60 m3^2 mQ3^6 mU3^8-12 mQ3^8 mU3^8-55 m3^6 mU3^10+97 m3^4 mQ3^2 mU3^10-72 m3^2 mQ3^4 mU3^10+24 mQ3^6 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)-(64 (42 m3^13-6 m3^11 mQ3^2-67 m3^9 mQ3^4+45 m3^7 mQ3^6-120 m3^11 mU3^2-44 m3^9 mQ3^2 mU3^2+214 m3^7 mQ3^4 mU3^2-90 m3^5 mQ3^6 mU3^2+239 m3^9 mU3^4-67 m3^7 mQ3^2 mU3^4-195 m3^5 mQ3^4 mU3^4+59 m3^3 mQ3^6 mU3^4-252 m3^7 mU3^6+268 m3^5 mQ3^2 mU3^6-42 m3^3 mQ3^4 mU3^6+18 m3 mQ3^6 mU3^6+59 m3^5 mU3^8-55 m3^3 mQ3^2 mU3^8-6 m3 mQ3^4 mU3^8) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(16 (148 m3^16 mQ3^2-336 m3^14 mQ3^4+194 m3^12 mQ3^6+m3^10 mQ3^8-13 m3^8 mQ3^10-788 m3^16 mU3^2+1796 m3^14 mQ3^2 mU3^2-786 m3^12 mQ3^4 mU3^2-452 m3^10 mQ3^6 mU3^2+190 m3^8 mQ3^8 mU3^2+64 m3^6 mQ3^10 mU3^2+2556 m3^14 mU3^4-7366 m3^12 mQ3^2 mU3^4+6232 m3^10 mQ3^4 mU3^4-936 m3^8 mQ3^6 mU3^4-479 m3^6 mQ3^8 mU3^4-37 m3^4 mQ3^10 mU3^4-2618 m3^12 mU3^6+8472 m3^10 mQ3^2 mU3^6-8712 m3^8 mQ3^4 mU3^6+2648 m3^6 mQ3^6 mU3^6+234 m3^4 mQ3^8 mU3^6-24 m3^2 mQ3^10 mU3^6+835 m3^10 mU3^8-3031 m3^8 mQ3^2 mU3^8+3742 m3^6 mQ3^4 mU3^8-1612 m3^4 mQ3^6 mU3^8+90 m3^2 mQ3^8 mU3^8+6 mQ3^10 mU3^8-58 m3^8 mU3^10+116 m3^6 mQ3^2 mU3^10-238 m3^4 mQ3^4 mU3^10+180 m3^2 mQ3^6 mU3^10-24 mQ3^8 mU3^10+69 m3^6 mU3^12-123 m3^4 mQ3^2 mU3^12+90 m3^2 mQ3^4 mU3^12-30 mQ3^6 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^4)+(32 (100 m3^11 mQ3^2-169 m3^9 mQ3^4+75 m3^7 mQ3^6+36 m3^11 mU3^2-392 m3^9 mQ3^2 mU3^2+508 m3^7 mQ3^4 mU3^2-148 m3^5 mQ3^6 mU3^2-159 m3^9 mU3^4+585 m3^7 mQ3^2 mU3^4-535 m3^5 mQ3^4 mU3^4+93 m3^3 mQ3^6 mU3^4+216 m3^7 mU3^6-384 m3^5 mQ3^2 mU3^6+152 m3^3 mQ3^4 mU3^6+12 m3 mQ3^6 mU3^6-61 m3^5 mU3^8+59 m3^3 mQ3^2 mU3^8+12 m3 mQ3^4 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)-(2560 m3^4 mU3^2 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) Log[mU3^2/MR^2]^2+Log[mQ3^2/MR^2] (-((8 (2 m3^16 mQ3^4-72 m3^14 mQ3^6+172 m3^12 mQ3^8-136 m3^10 mQ3^10+34 m3^8 mQ3^12+128 m3^16 mQ3^2 mU3^2-1462 m3^14 mQ3^4 mU3^2+3332 m3^12 mQ3^6 mU3^2-3774 m3^10 mQ3^8 mU3^2+2057 m3^8 mQ3^10 mU3^2-426 m3^6 mQ3^12 mU3^2+9 m3^4 mQ3^14 mU3^2+180 m3^12 mQ3^4 msq^2 mU3^2-420 m3^10 mQ3^6 msq^2 mU3^2+510 m3^8 mQ3^8 msq^2 mU3^2-360 m3^6 mQ3^10 msq^2 mU3^2+90 m3^4 mQ3^12 msq^2 mU3^2-2 m3^16 mU3^4+760 m3^14 mQ3^2 mU3^4+610 m3^12 mQ3^4 mU3^4-3634 m3^10 mQ3^6 mU3^4+5506 m3^8 mQ3^8 mU3^4-3339 m3^6 mQ3^10 mU3^4+640 m3^4 mQ3^12 mU3^4+3 m3^2 mQ3^14 mU3^4-180 m3^12 mQ3^2 msq^2 mU3^4+180 m3^10 mQ3^4 msq^2 mU3^4-450 m3^8 mQ3^6 msq^2 mU3^4+630 m3^6 mQ3^8 msq^2 mU3^4-210 m3^4 mQ3^10 msq^2 mU3^4+30 m3^2 mQ3^12 msq^2 mU3^4+6 m3^14 mU3^6-2188 m3^12 mQ3^2 mU3^6+3080 m3^10 mQ3^4 mU3^6-1602 m3^8 mQ3^6 mU3^6-1682 m3^6 mQ3^8 mU3^6+2026 m3^4 mQ3^10 mU3^6-456 m3^2 mQ3^12 mU3^6+240 m3^10 mQ3^2 msq^2 mU3^6+210 m3^8 mQ3^4 msq^2 mU3^6-450 m3^6 mQ3^6 msq^2 mU3^6+30 m3^4 mQ3^8 msq^2 mU3^6-30 m3^2 mQ3^10 msq^2 mU3^6-6 m3^12 mU3^8+1902 m3^10 mQ3^2 mU3^8-3528 m3^8 mQ3^4 mU3^8+3698 m3^6 mQ3^6 mU3^8-1522 m3^4 mQ3^8 mU3^8-84 m3^2 mQ3^10 mU3^8+84 mQ3^12 mU3^8-270 m3^8 mQ3^2 msq^2 mU3^8+90 m3^6 mQ3^4 msq^2 mU3^8+150 m3^4 mQ3^6 msq^2 mU3^8+30 m3^2 mQ3^8 msq^2 mU3^8+2 m3^10 mU3^10-547 m3^8 mQ3^2 mU3^10+972 m3^6 mQ3^4 mU3^10-1019 m3^4 mQ3^6 mU3^10+540 m3^2 mQ3^8 mU3^10-84 mQ3^10 mU3^10+90 m3^6 mQ3^2 msq^2 mU3^10-60 m3^4 mQ3^4 msq^2 mU3^10-30 m3^2 mQ3^6 msq^2 mU3^10+9 m3^6 mQ3^2 mU3^12-6 m3^4 mQ3^4 mU3^12-3 m3^2 mQ3^6 mU3^12))/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))-(64 (16 m3^11 mQ3^2-32 m3^9 mQ3^4+16 m3^7 mQ3^6+4 m3^11 mU3^2+216 m3^9 mQ3^2 mU3^2-223 m3^7 mQ3^4 mU3^2+86 m3^5 mQ3^6 mU3^2-3 m3^3 mQ3^8 mU3^2-60 m3^7 mQ3^2 msq^2 mU3^2+90 m3^5 mQ3^4 msq^2 mU3^2-30 m3^3 mQ3^6 msq^2 mU3^2-130 m3^9 mU3^4-106 m3^7 mQ3^2 mU3^4+111 m3^5 mQ3^4 mU3^4-35 m3^3 mQ3^6 mU3^4+60 m3^7 msq^2 mU3^4-60 m3^5 mQ3^2 msq^2 mU3^4+123 m3^7 mU3^6-68 m3^5 mQ3^2 mU3^6+37 m3^3 mQ3^4 mU3^6-12 m3 mQ3^6 mU3^6-30 m3^5 msq^2 mU3^6+30 m3^3 mQ3^2 msq^2 mU3^6-3 m3^5 mU3^8+3 m3^3 mQ3^2 mU3^8) Xt)/(3 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(16 (2 m3^14 mQ3^2-70 m3^12 mQ3^4+102 m3^10 mQ3^6-34 m3^8 mQ3^8-2 m3^14 mU3^2+1472 m3^12 mQ3^2 mU3^2-2660 m3^10 mQ3^4 mU3^2+1692 m3^8 mQ3^6 mU3^2-395 m3^6 mQ3^8 mU3^2+9 m3^4 mQ3^10 mU3^2-180 m3^10 mQ3^2 msq^2 mU3^2+240 m3^8 mQ3^4 msq^2 mU3^2-270 m3^6 mQ3^6 msq^2 mU3^2+90 m3^4 mQ3^8 msq^2 mU3^2+6 m3^12 mU3^4-3280 m3^10 mQ3^2 mU3^4+5696 m3^8 mQ3^4 mU3^4-3519 m3^6 mQ3^6 mU3^4+734 m3^4 mQ3^8 mU3^4+3 m3^2 mQ3^10 mU3^4+240 m3^8 mQ3^2 msq^2 mU3^4+180 m3^6 mQ3^4 msq^2 mU3^4-90 m3^4 mQ3^6 msq^2 mU3^4+30 m3^2 mQ3^8 msq^2 mU3^4-6 m3^10 mU3^6+2436 m3^8 mQ3^2 mU3^6-3987 m3^6 mQ3^4 mU3^6+2430 m3^4 mQ3^6 mU3^6-489 m3^2 mQ3^8 mU3^6-270 m3^6 mQ3^2 msq^2 mU3^6-90 m3^4 mQ3^4 msq^2 mU3^6+2 m3^8 mU3^8-587 m3^6 mQ3^2 mU3^8+850 m3^4 mQ3^4 mU3^8-513 m3^2 mQ3^6 mU3^8+96 mQ3^8 mU3^8+90 m3^4 mQ3^2 msq^2 mU3^8+30 m3^2 mQ3^4 msq^2 mU3^8+9 m3^4 mQ3^2 mU3^10+3 m3^2 mQ3^4 mU3^10) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(128 (16 m3^9 mQ3^4-16 m3^7 mQ3^6+22 m3^11 mU3^2+6 m3^9 mQ3^2 mU3^2-217 m3^7 mQ3^4 mU3^2+165 m3^5 mQ3^6 mU3^2-3 m3^3 mQ3^8 mU3^2+60 m3^5 mQ3^4 msq^2 mU3^2-30 m3^3 mQ3^6 msq^2 mU3^2+10 m3^9 mU3^4+238 m3^7 mQ3^2 mU3^4-17 m3^5 mQ3^4 mU3^4-115 m3^3 mQ3^6 mU3^4-120 m3^5 mQ3^2 msq^2 mU3^4+30 m3^3 mQ3^4 msq^2 mU3^4-193 m3^7 mU3^6-85 m3^5 mQ3^2 mU3^6+106 m3^3 mQ3^4 mU3^6+18 m3 mQ3^6 mU3^6+60 m3^5 msq^2 mU3^6+30 m3^3 mQ3^2 msq^2 mU3^6+153 m3^5 mU3^8-91 m3^3 mQ3^2 mU3^8+6 m3 mQ3^4 mU3^8-30 m3^3 msq^2 mU3^8-3 m3^3 mU3^10) Xt^3)/(3 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(8 (2 m3^16 mQ3^4-72 m3^14 mQ3^6+172 m3^12 mQ3^8-136 m3^10 mQ3^10+34 m3^8 mQ3^12-80 m3^16 mQ3^2 mU3^2+3178 m3^14 mQ3^4 mU3^2-9110 m3^12 mQ3^6 mU3^2+9926 m3^10 mQ3^8 mU3^2-4623 m3^8 mQ3^10 mU3^2+702 m3^6 mQ3^12 mU3^2-9 m3^4 mQ3^14 mU3^2-180 m3^12 mQ3^4 msq^2 mU3^2+420 m3^10 mQ3^6 msq^2 mU3^2-510 m3^8 mQ3^8 msq^2 mU3^2+360 m3^6 mQ3^10 msq^2 mU3^2-90 m3^4 mQ3^12 msq^2 mU3^2-2 m3^16 mU3^4+2152 m3^14 mQ3^2 mU3^4-12962 m3^12 mQ3^4 mU3^4+25584 m3^10 mQ3^6 mU3^4-22990 m3^8 mQ3^8 mU3^4+9553 m3^6 mQ3^10 mU3^4-1268 m3^4 mQ3^12 mU3^4-3 m3^2 mQ3^14 mU3^4-180 m3^12 mQ3^2 msq^2 mU3^4+660 m3^10 mQ3^4 msq^2 mU3^4-570 m3^8 mQ3^6 msq^2 mU3^4+90 m3^6 mQ3^8 msq^2 mU3^4+30 m3^4 mQ3^10 msq^2 mU3^4-30 m3^2 mQ3^12 msq^2 mU3^4+6 m3^14 mU3^6-4406 m3^12 mQ3^2 mU3^6+17588 m3^10 mQ3^4 mU3^6-25638 m3^8 mQ3^6 mU3^6+18070 m3^6 mQ3^8 mU3^6-6600 m3^4 mQ3^10 mU3^6+884 m3^2 mQ3^12 mU3^6+240 m3^10 mQ3^2 msq^2 mU3^6-330 m3^8 mQ3^4 msq^2 mU3^6-90 m3^6 mQ3^6 msq^2 mU3^6+210 m3^4 mQ3^8 msq^2 mU3^6-30 m3^2 mQ3^10 msq^2 mU3^6-6 m3^12 mU3^8+2996 m3^10 mQ3^2 mU3^8-9676 m3^8 mQ3^4 mU3^8+10862 m3^6 mQ3^6 mU3^8-5474 m3^4 mQ3^8 mU3^8+1602 m3^2 mQ3^10 mU3^8-240 mQ3^12 mU3^8-270 m3^8 mQ3^2 msq^2 mU3^8+270 m3^6 mQ3^4 msq^2 mU3^8+30 m3^4 mQ3^6 msq^2 mU3^8-30 m3^2 mQ3^8 msq^2 mU3^8+2 m3^10 mU3^10-643 m3^8 mQ3^2 mU3^10+1828 m3^6 mQ3^4 mU3^10-1723 m3^4 mQ3^6 mU3^10+616 m3^2 mQ3^8 mU3^10-96 mQ3^10 mU3^10+90 m3^6 mQ3^2 msq^2 mU3^10-60 m3^4 mQ3^4 msq^2 mU3^10-30 m3^2 mQ3^6 msq^2 mU3^10+9 m3^6 mQ3^2 mU3^12-6 m3^4 mQ3^4 mU3^12-3 m3^2 mQ3^6 mU3^12) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(64 (16 m3^11 mQ3^2-32 m3^9 mQ3^4+16 m3^7 mQ3^6+52 m3^11 mU3^2-362 m3^9 mQ3^2 mU3^2+558 m3^7 mQ3^4 mU3^2-219 m3^5 mQ3^6 mU3^2+3 m3^3 mQ3^8 mU3^2+60 m3^7 mQ3^2 msq^2 mU3^2-90 m3^5 mQ3^4 msq^2 mU3^2+30 m3^3 mQ3^6 msq^2 mU3^2-276 m3^9 mU3^4+770 m3^7 mQ3^2 mU3^4-737 m3^5 mQ3^4 mU3^4+179 m3^3 mQ3^6 mU3^4+60 m3^7 msq^2 mU3^4-120 m3^5 mQ3^2 msq^2 mU3^4+60 m3^3 mQ3^4 msq^2 mU3^4+214 m3^7 mU3^6-407 m3^5 mQ3^2 mU3^6+213 m3^3 mQ3^4 mU3^6+12 m3 mQ3^6 mU3^6-30 m3^5 msq^2 mU3^6+30 m3^3 mQ3^2 msq^2 mU3^6-3 m3^5 mU3^8+3 m3^3 mQ3^2 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(5120 m3^4 mQ3^2 Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+((160 (3 m3^12-11 m3^10 mQ3^2+3 m3^8 mQ3^4-m3^6 mQ3^6-13 m3^10 mU3^2+45 m3^8 mQ3^2 mU3^2-21 m3^6 mQ3^4 mU3^2+7 m3^4 mQ3^6 mU3^2+6 m3^8 mU3^4-24 m3^6 mQ3^2 mU3^4-2 m3^6 mU3^6+8 m3^4 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+((320 m3^3 (-2 m3^2+mQ3^2+mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))+(640 (m3^5 mQ3^2+m3^5 mU3^2-2 m3^3 mQ3^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Xt+(320 (2 m3^12-10 m3^10 mQ3^2+3 m3^8 mQ3^4-m3^6 mQ3^6-10 m3^10 mU3^2+42 m3^8 mQ3^2 mU3^2-21 m3^6 mQ3^4 mU3^2+7 m3^4 mQ3^6 mU3^2+3 m3^8 mU3^4-21 m3^6 mQ3^2 mU3^4-m3^6 mU3^6+7 m3^4 mQ3^2 mU3^6) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+((1280 (m3^5 mQ3^2+m3^5 mU3^2-2 m3^3 mQ3^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(640 m3^3 (2 m3^4-mQ3^4+4 mQ3^2 mU3^2-mU3^4-2 m3^2 (mQ3^2+mU3^2)))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) Xt^3-(160 (mQ3^2+mU3^2) (2 m3^12-10 m3^10 mQ3^2+3 m3^8 mQ3^4-m3^6 mQ3^6-10 m3^10 mU3^2+42 m3^8 mQ3^2 mU3^2-21 m3^6 mQ3^4 mU3^2+7 m3^4 mQ3^6 mU3^2+3 m3^8 mU3^4-21 m3^6 mQ3^2 mU3^4-m3^6 mU3^6+7 m3^4 mQ3^2 mU3^6) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(320 (mQ3^2+mU3^2) (m3^7+5 m3^5 mQ3^2+5 m3^5 mU3^2-11 m3^3 mQ3^2 mU3^2) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[msq^2/MR^2]+((16 (90 m3^16-330 m3^14 mQ3^2+355 m3^12 mQ3^4-128 m3^10 mQ3^6+17 m3^8 mQ3^8-330 m3^14 mU3^2+1196 m3^12 mQ3^2 mU3^2-1260 m3^10 mQ3^4 mU3^2+444 m3^8 mQ3^6 mU3^2-66 m3^6 mQ3^8 mU3^2+355 m3^12 mU3^4-1260 m3^10 mQ3^2 mU3^4+1160 m3^8 mQ3^4 mU3^4-248 m3^6 mQ3^6 mU3^4+17 m3^4 mQ3^8 mU3^4-128 m3^10 mU3^6+444 m3^8 mQ3^2 mU3^6-248 m3^6 mQ3^4 mU3^6-132 m3^4 mQ3^6 mU3^6+48 m3^2 mQ3^8 mU3^6+17 m3^8 mU3^8-66 m3^6 mQ3^2 mU3^8+17 m3^4 mQ3^4 mU3^8+48 m3^2 mQ3^6 mU3^8-12 mQ3^8 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4)-(512 (5 m3^11-8 m3^9 mQ3^2+5 m3^7 mQ3^4-8 m3^9 mU3^2+16 m3^7 mQ3^2 mU3^2-12 m3^5 mQ3^4 mU3^2+5 m3^7 mU3^4-12 m3^5 mQ3^2 mU3^4+9 m3^3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(64 (64 m3^18-310 m3^16 mQ3^2+516 m3^14 mQ3^4-405 m3^12 mQ3^6+164 m3^10 mQ3^8-27 m3^8 mQ3^10-310 m3^16 mU3^2+1368 m3^14 mQ3^2 mU3^2-2083 m3^12 mQ3^4 mU3^2+1488 m3^10 mQ3^6 mU3^2-561 m3^8 mQ3^8 mU3^2+88 m3^6 mQ3^10 mU3^2+516 m3^14 mU3^4-2083 m3^12 mQ3^2 mU3^4+2808 m3^10 mQ3^4 mU3^4-1704 m3^8 mQ3^6 mU3^4+564 m3^6 mQ3^8 mU3^4-81 m3^4 mQ3^10 mU3^4-405 m3^12 mU3^6+1488 m3^10 mQ3^2 mU3^6-1704 m3^8 mQ3^4 mU3^6+776 m3^6 mQ3^6 mU3^6-199 m3^4 mQ3^8 mU3^6+24 m3^2 mQ3^10 mU3^6+164 m3^10 mU3^8-561 m3^8 mQ3^2 mU3^8+564 m3^6 mQ3^4 mU3^8-199 m3^4 mQ3^6 mU3^8+48 m3^2 mQ3^8 mU3^8-6 mQ3^10 mU3^8-27 m3^8 mU3^10+88 m3^6 mQ3^2 mU3^10-81 m3^4 mQ3^4 mU3^10+24 m3^2 mQ3^6 mU3^10-6 mQ3^8 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)+(256 (18 m3^13-75 m3^11 mQ3^2+91 m3^9 mQ3^4-26 m3^7 mQ3^6-75 m3^11 mU3^2+232 m3^9 mQ3^2 mU3^2-232 m3^7 mQ3^4 mU3^2+51 m3^5 mQ3^6 mU3^2+91 m3^9 mU3^4-232 m3^7 mQ3^2 mU3^4+192 m3^5 mQ3^4 mU3^4-27 m3^3 mQ3^6 mU3^4-26 m3^7 mU3^6+51 m3^5 mQ3^2 mU3^6-27 m3^3 mQ3^4 mU3^6-6 m3 mQ3^6 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(32 (320 m3^18 mQ3^2-1036 m3^16 mQ3^4+1044 m3^14 mQ3^6-321 m3^12 mQ3^8+36 m3^10 mQ3^10-41 m3^8 mQ3^12+320 m3^18 mU3^2-2584 m3^16 mQ3^2 mU3^2+6252 m3^14 mQ3^4 mU3^2-5712 m3^12 mQ3^6 mU3^2+1612 m3^10 mQ3^8 mU3^2-24 m3^8 mQ3^10 mU3^2+128 m3^6 mQ3^12 mU3^2-1036 m3^16 mU3^4+6252 m3^14 mQ3^2 mU3^4-13598 m3^12 mQ3^4 mU3^4+12176 m3^10 mQ3^6 mU3^4-3615 m3^8 mQ3^8 mU3^4-44 m3^6 mQ3^10 mU3^4-125 m3^4 mQ3^12 mU3^4+1044 m3^14 mU3^6-5712 m3^12 mQ3^2 mU3^6+12176 m3^10 mQ3^4 mU3^6-11360 m3^8 mQ3^6 mU3^6+3884 m3^6 mQ3^8 mU3^6-80 m3^4 mQ3^10 mU3^6+48 m3^2 mQ3^12 mU3^6-321 m3^12 mU3^8+1612 m3^10 mQ3^2 mU3^8-3615 m3^8 mQ3^4 mU3^8+3884 m3^6 mQ3^6 mU3^8-1702 m3^4 mQ3^8 mU3^8+144 m3^2 mQ3^10 mU3^8-12 mQ3^12 mU3^8+36 m3^10 mU3^10-24 m3^8 mQ3^2 mU3^10-44 m3^6 mQ3^4 mU3^10-80 m3^4 mQ3^6 mU3^10+144 m3^2 mQ3^8 mU3^10-24 mQ3^10 mU3^10-41 m3^8 mU3^12+128 m3^6 mQ3^2 mU3^12-125 m3^4 mQ3^4 mU3^12+48 m3^2 mQ3^6 mU3^12-12 mQ3^8 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^4)-(128 (34 m3^13 mQ3^2-107 m3^11 mQ3^4+115 m3^9 mQ3^6-34 m3^7 mQ3^8+34 m3^13 mU3^2-214 m3^11 mQ3^2 mU3^2+411 m3^9 mQ3^4 mU3^2-314 m3^7 mQ3^6 mU3^2+67 m3^5 mQ3^8 mU3^2-107 m3^11 mU3^4+411 m3^9 mQ3^2 mU3^4-560 m3^7 mQ3^4 mU3^4+291 m3^5 mQ3^6 mU3^4-35 m3^3 mQ3^8 mU3^4+115 m3^9 mU3^6-314 m3^7 mQ3^2 mU3^6+291 m3^5 mQ3^4 mU3^6-70 m3^3 mQ3^6 mU3^6-6 m3 mQ3^8 mU3^6-34 m3^7 mU3^8+67 m3^5 mQ3^2 mU3^8-35 m3^3 mQ3^4 mU3^8-6 m3 mQ3^6 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(2560 m3^4 (mQ3^2+mU3^2) (m3^2 mQ3^2+m3^2 mU3^2-2 mQ3^2 mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) Log[mU3^2/MR^2])+(-((64 m3^4 (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)))/(-m3^2+mQ3^2)^2)+(512 (11 m3^11-10 m3^9 mQ3^2+7 m3^7 mQ3^4-18 m3^9 mU3^2+8 m3^7 mQ3^2 mU3^2-6 m3^5 mQ3^4 mU3^2+11 m3^7 mU3^4-6 m3^5 mQ3^2 mU3^4+3 m3^3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(8192 m3^6 Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))+(128 m3 (2 m3^2-mQ3^2-mU3^2) (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Xt^3)/(mQ3^2-mU3^2)^3-(64 (2 m3^2-mQ3^2-mU3^2) (53 m3^8-50 m3^6 mQ3^2-7 m3^4 mQ3^4-50 m3^6 mU3^2+52 m3^4 mQ3^2 mU3^2+6 m3^2 mQ3^4 mU3^2-7 m3^4 mU3^4+6 m3^2 mQ3^2 mU3^4-3 mQ3^4 mU3^4) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(1024 m3^3 (-2 m3^2+mQ3^2+mU3^2) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mQ3^2]+(-((64 m3^4 (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)))/(m3^2-mU3^2)^2)-(512 (11 m3^11-18 m3^9 mQ3^2+11 m3^7 mQ3^4-10 m3^9 mU3^2+8 m3^7 mQ3^2 mU3^2-6 m3^5 mQ3^4 mU3^2+7 m3^7 mU3^4-6 m3^5 mQ3^2 mU3^4+3 m3^3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(8192 m3^6 Xt^2)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(128 m3 (-2 m3^2+mQ3^2+mU3^2) (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Xt^3)/(mQ3^2-mU3^2)^3+(64 (2 m3^2-mQ3^2-mU3^2) (53 m3^8-50 m3^6 mQ3^2-7 m3^4 mQ3^4-50 m3^6 mU3^2+52 m3^4 mQ3^2 mU3^2+6 m3^2 mQ3^4 mU3^2-7 m3^4 mU3^4+6 m3^2 mQ3^2 mU3^4-3 mQ3^4 mU3^4) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(1024 m3^3 (-2 m3^2+mQ3^2+mU3^2) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mU3^2])+((4 (-24 m3^10 mQ3^2+28 m3^8 mQ3^4-68 m3^6 mQ3^6+58 m3^4 mQ3^8-18 m3^2 mQ3^10+24 m3^10 mU3^2+220 m3^6 mQ3^4 mU3^2-188 m3^4 mQ3^6 mU3^2+70 m3^2 mQ3^8 mU3^2-6 mQ3^10 mU3^2-28 m3^8 mU3^4-220 m3^6 mQ3^2 mU3^4+16 m3^2 mQ3^6 mU3^4-8 mQ3^8 mU3^4+68 m3^6 mU3^6+188 m3^4 mQ3^2 mU3^6-16 m3^2 mQ3^4 mU3^6-58 m3^4 mU3^8-70 m3^2 mQ3^2 mU3^8+8 mQ3^4 mU3^8+18 m3^2 mU3^10+6 mQ3^2 mU3^10))/(3 (-m3^2+mQ3^2)^3 (m3^2-mU3^2)^3)-(32 (10 m3^5 mQ3^2-10 m3^3 mQ3^4+6 m3 mQ3^6-10 m3^5 mU3^2-8 m3 mQ3^4 mU3^2+10 m3^3 mU3^4+8 m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(16 (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(128 (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-mQ3^2/mU3^2]+(-((80 (-msq^2+mU3^2) (2 m3^4-3 m3^2 msq^2+3 m3^2 mU3^2-msq^2 mU3^2-mU3^4))/(m3^2-mU3^2)^3)+(640 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(160 (-msq^2+mU3^2) (2 m3^4-3 m3^2 msq^2+3 m3^2 mU3^2-msq^2 mU3^2-mU3^4) Xt^4)/((m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)-(1280 m3 (-msq^2+mU3^2)^2 Xt^5)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^3)) PolyLog[2,1-msq^2/mU3^2]+Log[MR^2/mt^2] (-((16 (188 m3^8-316 m3^6 mQ3^2+169 m3^4 mQ3^4-9 m3^2 mQ3^6-180 m3^6 msq^2+150 m3^4 mQ3^2 msq^2-90 m3^2 mQ3^4 msq^2-316 m3^6 mU3^2+454 m3^4 mQ3^2 mU3^2-211 m3^2 mQ3^4 mU3^2-3 mQ3^6 mU3^2+150 m3^4 msq^2 mU3^2+120 m3^2 mQ3^2 msq^2 mU3^2-30 mQ3^4 msq^2 mU3^2+169 m3^4 mU3^4-211 m3^2 mQ3^2 mU3^4+98 mQ3^4 mU3^4-90 m3^2 msq^2 mU3^4-30 mQ3^2 msq^2 mU3^4-9 m3^2 mU3^6-3 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+((1792 m3^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))-(128 m3 mQ3^2)/((m3^2-mQ3^2) (m3^2-mU3^2))-(1280 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2))-(128 m3 mU3^2)/((m3^2-mQ3^2) (m3^2-mU3^2))) Xt-128 z3+(-((16 (198 m3^16-726 m3^14 mQ3^2+721 m3^12 mQ3^4-232 m3^10 mQ3^6+27 m3^8 mQ3^8-726 m3^14 mU3^2+2700 m3^12 mQ3^2 mU3^2-2700 m3^10 mQ3^4 mU3^2+908 m3^8 mQ3^6 mU3^2-134 m3^6 mQ3^8 mU3^2+721 m3^12 mU3^4-2700 m3^10 mQ3^2 mU3^4+2240 m3^8 mQ3^4 mU3^4-320 m3^6 mQ3^6 mU3^4-13 m3^4 mQ3^8 mU3^4-232 m3^10 mU3^6+908 m3^8 mQ3^2 mU3^6-320 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6+27 m3^8 mU3^8-134 m3^6 mQ3^2 mU3^8-13 m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4))+(128 (m3^11-50 m3^9 mQ3^2+41 m3^7 mQ3^4-50 m3^9 mU3^2+196 m3^7 mQ3^2 mU3^2-130 m3^5 mQ3^4 mU3^2+41 m3^7 mU3^4-130 m3^5 mQ3^2 mU3^4+81 m3^3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(512 m3^6 Xt^2)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Log[m3^2/MR^2]^2+(-((8 (128 m3^16-556 m3^14 mQ3^2+718 m3^12 mQ3^4-463 m3^10 mQ3^6+184 m3^8 mQ3^8-64 m3^6 mQ3^10+38 m3^4 mQ3^12-9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-340 m3^14 mU3^2+1708 m3^12 mQ3^2 mU3^2-2131 m3^10 mQ3^4 mU3^2+1115 m3^8 mQ3^6 mU3^2-136 m3^6 mQ3^8 mU3^2-164 m3^4 mQ3^10 mU3^2+47 m3^2 mQ3^12 mU3^2-3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+262 m3^12 mU3^4-1891 m3^10 mQ3^2 mU3^4+2513 m3^8 mQ3^4 mU3^4-1512 m3^6 mQ3^6 mU3^4+514 m3^4 mQ3^8 mU3^4-29 m3^2 mQ3^10 mU3^4-mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4+5 m3^10 mU3^6+711 m3^8 mQ3^2 mU3^6-920 m3^6 mQ3^4 mU3^6+436 m3^4 mQ3^6 mU3^6-141 m3^2 mQ3^8 mU3^6+5 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-43 m3^8 mU3^8-56 m3^6 mQ3^2 mU3^8+72 m3^4 mQ3^4 mU3^8+4 m3^2 mQ3^6 mU3^8-mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(64 (18 m3^11 mQ3^2-85 m3^9 mQ3^4+62 m3^7 mQ3^6-16 m3^5 mQ3^8+8 m3^3 mQ3^10-3 m3 mQ3^12+60 m3^7 mQ3^4 msq^2-60 m3^5 mQ3^6 msq^2-30 m3^7 mQ3^2 msq^4+30 m3^5 mQ3^4 msq^4-18 m3^11 mU3^2+32 m3^9 mQ3^2 mU3^2+138 m3^7 mQ3^4 mU3^2-106 m3^5 mQ3^6 mU3^2-8 m3^3 mQ3^8 mU3^2+10 m3 mQ3^10 mU3^2-60 m3^7 mQ3^2 msq^2 mU3^2-60 m3^5 mQ3^4 msq^2 mU3^2+120 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2+30 m3^5 mQ3^2 msq^4 mU3^2-60 m3^3 mQ3^4 msq^4 mU3^2+37 m3^9 mU3^4-114 m3^7 mQ3^2 mU3^4-56 m3^5 mQ3^4 mU3^4+106 m3^3 mQ3^6 mU3^4-21 m3 mQ3^8 mU3^4+120 m3^5 mQ3^2 msq^2 mU3^4-60 m3^3 mQ3^4 msq^2 mU3^4-60 m3 mQ3^6 msq^2 mU3^4-60 m3^5 msq^4 mU3^4+30 m3^3 mQ3^2 msq^4 mU3^4+30 m3 mQ3^4 msq^4 mU3^4-22 m3^7 mU3^6+82 m3^5 mQ3^2 mU3^6-42 m3^3 mQ3^4 mU3^6-2 m3 mQ3^6 mU3^6-60 m3^3 mQ3^2 msq^2 mU3^6+60 m3 mQ3^4 msq^2 mU3^6+30 m3^3 msq^4 mU3^6-30 m3 mQ3^2 msq^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(512 m3^2 mQ3^4 Xt^2)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)) Log[mQ3^2/MR^2]^2+((160 (10 m3^8-20 m3^6 mQ3^2+10 m3^4 mQ3^4+18 m3^6 msq^2-15 m3^4 mQ3^2 msq^2+9 m3^2 mQ3^4 msq^2-20 m3^6 mU3^2+40 m3^4 mQ3^2 mU3^2-20 m3^2 mQ3^4 mU3^2-15 m3^4 msq^2 mU3^2-12 m3^2 mQ3^2 msq^2 mU3^2+3 mQ3^4 msq^2 mU3^2+10 m3^4 mU3^4-20 m3^2 mQ3^2 mU3^4+10 mQ3^4 mU3^4+9 m3^2 msq^2 mU3^4+3 mQ3^2 msq^2 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(1280 m3 msq^2 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))) Log[msq^2/MR^2]+(-48 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3))-48 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt) Log[msq^2/MR^2]^2+((16 (54 m3^10 mQ3^2-108 m3^8 mQ3^4+54 m3^6 mQ3^6-310 m3^10 mU3^2+582 m3^8 mQ3^2 mU3^2-296 m3^6 mQ3^4 mU3^2+21 m3^4 mQ3^6 mU3^2-9 m3^2 mQ3^8 mU3^2-90 m3^6 mQ3^2 msq^2 mU3^2+180 m3^4 mQ3^4 msq^2 mU3^2-90 m3^2 mQ3^6 msq^2 mU3^2+550 m3^8 mU3^4-1078 m3^6 mQ3^2 mU3^4+679 m3^4 mQ3^4 mU3^4-100 m3^2 mQ3^6 mU3^4-3 mQ3^8 mU3^4+90 m3^6 msq^2 mU3^4-210 m3^4 mQ3^2 msq^2 mU3^4+150 m3^2 mQ3^4 msq^2 mU3^4-30 mQ3^6 msq^2 mU3^4-216 m3^6 mU3^6+359 m3^4 mQ3^2 mU3^6-296 m3^2 mQ3^4 mU3^6+81 mQ3^6 mU3^6+30 m3^4 msq^2 mU3^6-60 m3^2 mQ3^2 msq^2 mU3^6+30 mQ3^4 msq^2 mU3^6-35 m3^4 mU3^8+158 m3^2 mQ3^2 mU3^8-75 mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2))-(128 (16 m3^7-16 m3^5 mQ3^2+10 m3^5 mU3^2-18 m3^3 mQ3^2 mU3^2+3 m3 mQ3^4 mU3^2-30 m3^3 msq^2 mU3^2+30 m3 mQ3^2 msq^2 mU3^2-26 m3^3 mU3^4+28 m3 mQ3^2 mU3^4+3 m3 mU3^6) Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+((160 (2 m3^4 msq^2-3 m3^2 msq^4+6 m3^2 msq^2 mU3^2-msq^4 mU3^2))/(m3^2-mU3^2)^3-(1280 (-m3 msq^4+2 m3 msq^2 mU3^2) Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Log[msq^2/MR^2]) Log[mU3^2/MR^2]+(-((8 (128 m3^10+44 m3^8 mQ3^2-2 m3^6 mQ3^4+9 m3^4 mQ3^6-60 m3^6 mQ3^2 msq^2+90 m3^4 mQ3^2 msq^4-556 m3^8 mU3^2+64 m3^6 mQ3^2 mU3^2-33 m3^4 mQ3^4 mU3^2-6 m3^2 mQ3^6 mU3^2+60 m3^6 msq^2 mU3^2-120 m3^4 mQ3^2 msq^2 mU3^2-90 m3^4 msq^4 mU3^2-60 m3^2 mQ3^2 msq^4 mU3^2+706 m3^6 mU3^4-51 m3^4 mQ3^2 mU3^4+36 m3^2 mQ3^4 mU3^4-3 mQ3^6 mU3^4+120 m3^4 msq^2 mU3^4+180 m3^2 mQ3^2 msq^2 mU3^4+60 m3^2 msq^4 mU3^4-30 mQ3^2 msq^4 mU3^4-437 m3^4 mU3^6-38 m3^2 mQ3^2 mU3^6-mQ3^4 mU3^6-180 m3^2 msq^2 mU3^6+30 msq^4 mU3^6+136 m3^2 mU3^8+5 mQ3^2 mU3^8-mU3^10))/(3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)))+(64 (-18 m3^7 mQ3^2+m3^5 mQ3^4+3 m3^3 mQ3^6+30 m3^3 mQ3^2 msq^4+18 m3^7 mU3^2+68 m3^5 mQ3^2 mU3^2-11 m3^3 mQ3^4 mU3^2-3 m3 mQ3^6 mU3^2-60 m3^3 mQ3^2 msq^2 mU3^2-30 m3^3 msq^4 mU3^2-30 m3 mQ3^2 msq^4 mU3^2-85 m3^5 mU3^4-17 m3^3 mQ3^2 mU3^4+10 m3 mQ3^4 mU3^4+60 m3^3 msq^2 mU3^4+60 m3 mQ3^2 msq^2 mU3^4+30 m3 msq^4 mU3^4+57 m3^3 mU3^6-17 m3 mQ3^2 mU3^6-60 m3 msq^2 mU3^6-6 m3 mU3^8) Xt)/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)+(512 m3^2 mU3^4 Xt^2)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)) Log[mU3^2/MR^2]^2+Log[mQ3^2/MR^2] ((16 (-310 m3^10 mQ3^2+550 m3^8 mQ3^4-216 m3^6 mQ3^6-35 m3^4 mQ3^8-9 m3^2 mQ3^10+90 m3^6 mQ3^4 msq^2+30 m3^4 mQ3^6 msq^2+54 m3^10 mU3^2+582 m3^8 mQ3^2 mU3^2-1078 m3^6 mQ3^4 mU3^2+359 m3^4 mQ3^6 mU3^2+158 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2-90 m3^6 mQ3^2 msq^2 mU3^2-210 m3^4 mQ3^4 msq^2 mU3^2-60 m3^2 mQ3^6 msq^2 mU3^2-108 m3^8 mU3^4-296 m3^6 mQ3^2 mU3^4+679 m3^4 mQ3^4 mU3^4-296 m3^2 mQ3^6 mU3^4-75 mQ3^8 mU3^4+180 m3^4 mQ3^2 msq^2 mU3^4+150 m3^2 mQ3^4 msq^2 mU3^4+30 mQ3^6 msq^2 mU3^4+54 m3^6 mU3^6+21 m3^4 mQ3^2 mU3^6-100 m3^2 mQ3^4 mU3^6+81 mQ3^6 mU3^6-90 m3^2 mQ3^2 msq^2 mU3^6-30 mQ3^4 msq^2 mU3^6-9 m3^2 mQ3^2 mU3^8-3 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(128 (16 m3^7+10 m3^5 mQ3^2-26 m3^3 mQ3^4+3 m3 mQ3^6-30 m3^3 mQ3^2 msq^2-16 m3^5 mU3^2-18 m3^3 mQ3^2 mU3^2+28 m3 mQ3^4 mU3^2+30 m3 mQ3^2 msq^2 mU3^2+3 m3 mQ3^2 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))+((160 (2 m3^4 msq^2+6 m3^2 mQ3^2 msq^2-3 m3^2 msq^4-mQ3^2 msq^4))/(m3^2-mQ3^2)^3-(1280 (2 m3 mQ3^2 msq^2-m3 msq^4) Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) Log[msq^2/MR^2]+((16 (12 m3^10 mQ3^2-14 m3^8 mQ3^4+34 m3^6 mQ3^6-29 m3^4 mQ3^8+9 m3^2 mQ3^10+64 m3^8 mQ3^2 mU3^2-188 m3^6 mQ3^4 mU3^2+120 m3^4 mQ3^6 mU3^2-35 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2-78 m3^6 mQ3^2 mU3^4+175 m3^4 mQ3^4 mU3^4-65 m3^2 mQ3^6 mU3^4+4 mQ3^8 mU3^4+26 m3^4 mQ3^2 mU3^6-57 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(128 (-5 m3^5 mQ3^6+5 m3^3 mQ3^8-3 m3 mQ3^10+16 m3^7 mQ3^2 mU3^2-14 m3^5 mQ3^4 mU3^2+5 m3^3 mQ3^6 mU3^2+10 m3 mQ3^8 mU3^2-29 m3^5 mQ3^2 mU3^4+23 m3^3 mQ3^4 mU3^4-19 m3 mQ3^6 mU3^4+15 m3^3 mQ3^2 mU3^6-4 m3 mQ3^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(1024 m3^2 mQ3^2 mU3^2 Xt^2)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)) Log[mU3^2/MR^2])+Log[m3^2/MR^2] ((16 (660 m3^12-1562 m3^10 mQ3^2+1182 m3^8 mQ3^4-269 m3^6 mQ3^6+9 m3^4 mQ3^8-180 m3^10 msq^2+240 m3^8 mQ3^2 msq^2-270 m3^6 mQ3^4 msq^2+90 m3^4 mQ3^6 msq^2-1562 m3^10 mU3^2+3564 m3^8 mQ3^2 mU3^2-2575 m3^6 mQ3^4 mU3^2+498 m3^4 mQ3^6 mU3^2+3 m3^2 mQ3^8 mU3^2+240 m3^8 msq^2 mU3^2+180 m3^6 mQ3^2 msq^2 mU3^2-90 m3^4 mQ3^4 msq^2 mU3^2+30 m3^2 mQ3^6 msq^2 mU3^2+1182 m3^8 mU3^4-2575 m3^6 mQ3^2 mU3^4+1822 m3^4 mQ3^4 mU3^4-333 m3^2 mQ3^6 mU3^4-270 m3^6 msq^2 mU3^4-90 m3^4 mQ3^2 msq^2 mU3^4-269 m3^6 mU3^6+498 m3^4 mQ3^2 mU3^6-333 m3^2 mQ3^4 mU3^6+48 mQ3^6 mU3^6+90 m3^4 msq^2 mU3^6+30 m3^2 mQ3^2 msq^2 mU3^6+9 m3^4 mU3^8+3 m3^2 mQ3^2 mU3^8))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(128 (40 m3^7-40 m3^5 mQ3^2+3 m3^3 mQ3^4-60 m3^5 msq^2+30 m3^3 mQ3^2 msq^2-40 m3^5 mU3^2+34 m3^3 mQ3^2 mU3^2+30 m3^3 msq^2 mU3^2+3 m3^3 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+((16 (128 m3^16-358 m3^14 mQ3^2+134 m3^12 mQ3^4+125 m3^10 mQ3^6-72 m3^8 mQ3^8+19 m3^6 mQ3^10-410 m3^14 mU3^2+1314 m3^12 mQ3^2 mU3^2-683 m3^10 mQ3^4 mU3^2-191 m3^8 mQ3^6 mU3^2+97 m3^6 mQ3^8 mU3^2-31 m3^4 mQ3^10 mU3^2+472 m3^12 mU3^4-1797 m3^10 mQ3^2 mU3^4+1157 m3^8 mQ3^4 mU3^4-11 m3^6 mQ3^6 mU3^4+35 m3^4 mQ3^8 mU3^4-205 m3^10 mU3^6+999 m3^8 mQ3^2 mU3^6-655 m3^6 mQ3^4 mU3^6-43 m3^4 mQ3^6 mU3^6+27 m3^8 mU3^8-218 m3^6 mQ3^2 mU3^8+167 m3^4 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(128 (34 m3^11-117 m3^9 mQ3^2+71 m3^7 mQ3^4-4 m3^5 mQ3^6-69 m3^9 mU3^2+229 m3^7 mQ3^2 mU3^2-121 m3^5 mQ3^4 mU3^2-7 m3^3 mQ3^6 mU3^2+38 m3^7 mU3^4-129 m3^5 mQ3^2 mU3^4+75 m3^3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(1024 m3^4 mQ3^2 Xt^2)/((m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))) Log[mQ3^2/MR^2]+((160 (2 m3^12-8 m3^10 mQ3^2+3 m3^8 mQ3^4-m3^6 mQ3^6-8 m3^10 mU3^2+30 m3^8 mQ3^2 mU3^2-15 m3^6 mQ3^4 mU3^2+5 m3^4 mQ3^6 mU3^2+3 m3^8 mU3^4-15 m3^6 mQ3^2 mU3^4-m3^6 mU3^6+5 m3^4 mQ3^2 mU3^6))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(1280 (m3^5 mQ3^2+m3^5 mU3^2-2 m3^3 mQ3^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Log[msq^2/MR^2]+(-((16 (128 m3^16-410 m3^14 mQ3^2+472 m3^12 mQ3^4-205 m3^10 mQ3^6+27 m3^8 mQ3^8-358 m3^14 mU3^2+1314 m3^12 mQ3^2 mU3^2-1797 m3^10 mQ3^4 mU3^2+999 m3^8 mQ3^6 mU3^2-218 m3^6 mQ3^8 mU3^2+134 m3^12 mU3^4-683 m3^10 mQ3^2 mU3^4+1157 m3^8 mQ3^4 mU3^4-655 m3^6 mQ3^6 mU3^4+167 m3^4 mQ3^8 mU3^4+125 m3^10 mU3^6-191 m3^8 mQ3^2 mU3^6-11 m3^6 mQ3^4 mU3^6-43 m3^4 mQ3^6 mU3^6-72 m3^8 mU3^8+97 m3^6 mQ3^2 mU3^8+35 m3^4 mQ3^4 mU3^8+19 m3^6 mU3^10-31 m3^4 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)))+(128 (34 m3^11-69 m3^9 mQ3^2+38 m3^7 mQ3^4-117 m3^9 mU3^2+229 m3^7 mQ3^2 mU3^2-129 m3^5 mQ3^4 mU3^2+71 m3^7 mU3^4-121 m3^5 mQ3^2 mU3^4+75 m3^3 mQ3^4 mU3^4-4 m3^5 mU3^6-7 m3^3 mQ3^2 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(1024 m3^4 mU3^2 Xt^2)/((m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))) Log[mU3^2/MR^2])+((16 (-128 m3^12+294 m3^10 mQ3^2-98 m3^8 mQ3^4-9 m3^6 mQ3^6-20 m3^4 mQ3^8+9 m3^2 mQ3^10+346 m3^10 mU3^2-902 m3^8 mQ3^2 mU3^2+251 m3^6 mQ3^4 mU3^2+151 m3^4 mQ3^6 mU3^2-41 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2-280 m3^8 mU3^4+1025 m3^6 mQ3^2 mU3^4-417 m3^4 mQ3^4 mU3^4-41 m3^2 mQ3^6 mU3^4+mQ3^8 mU3^4+13 m3^6 mU3^6-391 m3^4 mQ3^2 mU3^6+167 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6+37 m3^4 mU3^8+34 m3^2 mQ3^2 mU3^8-23 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(128 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4) Xt)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) PolyLog[2,1-m3^2/mQ3^2]+(-24 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2)))-(1280 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) PolyLog[2,1-m3^2/msq^2]+((80 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(-m3^2+mQ3^2)^3-(1280 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-msq^2/mQ3^2]+(-((16 (-128 m3^12+346 m3^10 mQ3^2-280 m3^8 mQ3^4+13 m3^6 mQ3^6+37 m3^4 mQ3^8+294 m3^10 mU3^2-902 m3^8 mQ3^2 mU3^2+1025 m3^6 mQ3^4 mU3^2-391 m3^4 mQ3^6 mU3^2+34 m3^2 mQ3^8 mU3^2-98 m3^8 mU3^4+251 m3^6 mQ3^2 mU3^4-417 m3^4 mQ3^4 mU3^4+167 m3^2 mQ3^6 mU3^4-23 mQ3^8 mU3^4-9 m3^6 mU3^6+151 m3^4 mQ3^2 mU3^6-41 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6-20 m3^4 mU3^8-41 m3^2 mQ3^2 mU3^8+mQ3^4 mU3^8+9 m3^2 mU3^10+3 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)))-(128 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-m3^2/mU3^2]+(-((8 (-24 m3^10 mQ3^2+28 m3^8 mQ3^4-68 m3^6 mQ3^6+58 m3^4 mQ3^8-18 m3^2 mQ3^10+24 m3^10 mU3^2+220 m3^6 mQ3^4 mU3^2-188 m3^4 mQ3^6 mU3^2+70 m3^2 mQ3^8 mU3^2-6 mQ3^10 mU3^2-28 m3^8 mU3^4-220 m3^6 mQ3^2 mU3^4+16 m3^2 mQ3^6 mU3^4-8 mQ3^8 mU3^4+68 m3^6 mU3^6+188 m3^4 mQ3^2 mU3^6-16 m3^2 mQ3^4 mU3^6-58 m3^4 mU3^8-70 m3^2 mQ3^2 mU3^8+8 mQ3^4 mU3^8+18 m3^2 mU3^10+6 mQ3^2 mU3^10))/(3 (-m3^2+mQ3^2)^3 (m3^2-mU3^2)^3))+(64 (10 m3^5 mQ3^2-10 m3^3 mQ3^4+6 m3 mQ3^6-10 m3^5 mU3^2-8 m3 mQ3^4 mU3^2+10 m3^3 mU3^4+8 m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) PolyLog[2,1-mQ3^2/mU3^2]+((80 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(-m3^2+mU3^2)^3-(1280 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) PolyLog[2,1-msq^2/mU3^2])+Log[mQ3^2/MR^2] ((4 (4 m3^12 mQ3^4-140 m3^10 mQ3^6+204 m3^8 mQ3^8-68 m3^6 mQ3^10+33 m3^10 mQ3^4 mU3^2+583 m3^8 mQ3^6 mU3^2-1259 m3^6 mQ3^8 mU3^2+587 m3^4 mQ3^10 mU3^2+360 m3^8 mQ3^4 msq^2 mU3^2-840 m3^6 mQ3^6 msq^2 mU3^2+420 m3^4 mQ3^8 msq^2 mU3^2-180 m3^2 mQ3^10 msq^2 mU3^2-4 m3^12 mU3^4+611 m3^10 mQ3^2 mU3^4-2127 m3^8 mQ3^4 mU3^4+1975 m3^6 mQ3^6 mU3^4+283 m3^4 mQ3^8 mU3^4-594 m3^2 mQ3^10 mU3^4-360 m3^8 mQ3^2 msq^2 mU3^4+540 m3^6 mQ3^4 msq^2 mU3^4+600 m3^2 mQ3^8 msq^2 mU3^4-60 mQ3^10 msq^2 mU3^4+8 m3^10 mU3^6-704 m3^8 mQ3^2 mU3^6+2217 m3^6 mQ3^4 mU3^6-2359 m3^4 mQ3^6 mU3^6+571 m3^2 mQ3^8 mU3^6+171 mQ3^10 mU3^6+300 m3^6 mQ3^2 msq^2 mU3^6-240 m3^4 mQ3^4 msq^2 mU3^6-720 m3^2 mQ3^6 msq^2 mU3^6-60 mQ3^8 msq^2 mU3^6-4 m3^8 mU3^8+207 m3^6 mQ3^2 mU3^8-541 m3^4 mQ3^4 mU3^8+505 m3^2 mQ3^6 mU3^8-183 mQ3^8 mU3^8-180 m3^4 mQ3^2 msq^2 mU3^8+300 m3^2 mQ3^4 msq^2 mU3^8+120 mQ3^6 msq^2 mU3^8-18 m3^4 mQ3^2 mU3^10+30 m3^2 mQ3^4 mU3^10+12 mQ3^6 mU3^10))/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(64 (16 m3^7 mQ3^2-16 m3^5 mQ3^4+22 m3^7 mU3^2-112 m3^5 mQ3^2 mU3^2+87 m3^3 mQ3^4 mU3^2+30 m3 mQ3^4 msq^2 mU3^2-20 m3^5 mU3^4+100 m3^3 mQ3^2 mU3^4-74 m3 mQ3^4 mU3^4+30 m3^3 msq^2 mU3^4-60 m3 mQ3^2 msq^2 mU3^4+3 m3^3 mU3^6-6 m3 mQ3^2 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2) (mQ3^2-mU3^2))-(8 (4 m3^10 mQ3^2-136 m3^8 mQ3^4+68 m3^6 mQ3^6-4 m3^10 mU3^2+695 m3^8 mQ3^2 mU3^2-326 m3^6 mQ3^4 mU3^2+5 m3^4 mQ3^6 mU3^2-18 m3^2 mQ3^8 mU3^2-360 m3^6 mQ3^2 msq^2 mU3^2+300 m3^4 mQ3^4 msq^2 mU3^2-180 m3^2 mQ3^6 msq^2 mU3^2+8 m3^8 mU3^4-886 m3^6 mQ3^2 mU3^4+40 m3^4 mQ3^4 mU3^4+308 m3^2 mQ3^6 mU3^4-6 mQ3^8 mU3^4+300 m3^4 mQ3^2 msq^2 mU3^4+240 m3^2 mQ3^4 msq^2 mU3^4-60 mQ3^6 msq^2 mU3^4-4 m3^6 mU3^6+301 m3^4 mQ3^2 mU3^6+212 m3^2 mQ3^4 mU3^6-237 mQ3^6 mU3^6-180 m3^2 mQ3^2 msq^2 mU3^6-60 mQ3^4 msq^2 mU3^6-18 m3^2 mQ3^2 mU3^8-6 mQ3^4 mU3^8) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+((128 (14 m3^3-3 m3 mQ3^2-30 m3 msq^2-3 m3 mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))+(128 m3 (3 mQ3^2+mU3^2) (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(256 (8 m3^5 mQ3^2-8 m3^3 mQ3^4-7 m3^5 mU3^2-26 m3^3 mQ3^2 mU3^2+17 m3 mQ3^4 mU3^2))/(3 (m3^2-mQ3^2)^2 mU3^2 (mQ3^2-mU3^2)^2)) Xt^3+(4 (4 m3^12 mQ3^4-140 m3^10 mQ3^6+204 m3^8 mQ3^8-68 m3^6 mQ3^10+176 m3^12 mQ3^2 mU3^2-4793 m3^10 mQ3^4 mU3^2+7283 m3^8 mQ3^6 mU3^2-1881 m3^6 mQ3^8 mU3^2-975 m3^4 mQ3^10 mU3^2-18 m3^2 mQ3^12 mU3^2-360 m3^8 mQ3^4 msq^2 mU3^2+1020 m3^6 mQ3^6 msq^2 mU3^2-360 m3^4 mQ3^8 msq^2 mU3^2+180 m3^2 mQ3^10 msq^2 mU3^2-4 m3^12 mU3^4+667 m3^10 mQ3^2 mU3^4+10101 m3^8 mQ3^4 mU3^4-17667 m3^6 mQ3^6 mU3^4+6073 m3^4 mQ3^8 mU3^4+1508 m3^2 mQ3^10 mU3^4-6 mQ3^12 mU3^4-360 m3^8 mQ3^2 msq^2 mU3^4+600 m3^6 mQ3^4 msq^2 mU3^4-1380 m3^4 mQ3^6 msq^2 mU3^4-360 m3^2 mQ3^8 msq^2 mU3^4+60 mQ3^10 msq^2 mU3^4+8 m3^10 mU3^6-1222 m3^8 mQ3^2 mU3^6-6913 m3^6 mQ3^4 mU3^6+13435 m3^4 mQ3^6 mU3^6-5659 m3^2 mQ3^8 mU3^6-417 mQ3^10 mU3^6+300 m3^6 mQ3^2 msq^2 mU3^6+480 m3^4 mQ3^4 msq^2 mU3^6+420 m3^2 mQ3^6 msq^2 mU3^6+240 mQ3^8 msq^2 mU3^6-4 m3^8 mU3^8+493 m3^6 mQ3^2 mU3^8+1361 m3^4 mQ3^4 mU3^8-2665 m3^2 mQ3^6 mU3^8+1167 mQ3^8 mU3^8-180 m3^4 mQ3^2 msq^2 mU3^8-240 m3^2 mQ3^4 msq^2 mU3^8-60 mQ3^6 msq^2 mU3^8-18 m3^4 mQ3^2 mU3^10-24 m3^2 mQ3^4 mU3^10-6 mQ3^6 mU3^10) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(64 (16 m3^7 mQ3^2-16 m3^5 mQ3^4+48 m3^7 mU3^2+172 m3^5 mQ3^2 mU3^2-185 m3^3 mQ3^4 mU3^2+3 m3 mQ3^6 mU3^2-30 m3^3 mQ3^2 msq^2 mU3^2-30 m3 mQ3^4 msq^2 mU3^2-62 m3^5 mU3^4-166 m3^3 mQ3^2 mU3^4+184 m3 mQ3^4 mU3^4+30 m3^3 msq^2 mU3^4+30 m3 mQ3^2 msq^2 mU3^4+3 m3^3 mU3^6+3 m3 mQ3^2 mU3^6) Xt^5)/(3 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+((80 (18 m3^10-47 m3^8 mQ3^2+43 m3^6 mQ3^4-14 m3^4 mQ3^6+12 m3^8 msq^2-51 m3^6 mQ3^2 msq^2+24 m3^4 mQ3^4 msq^2-9 m3^2 mQ3^6 msq^2+9 m3^6 msq^4+3 m3^4 mQ3^2 msq^4-35 m3^8 mU3^2+91 m3^6 mQ3^2 mU3^2-83 m3^4 mQ3^4 mU3^2+27 m3^2 mQ3^6 mU3^2-3 m3^6 msq^2 mU3^2+39 m3^4 mQ3^2 msq^2 mU3^2+15 m3^2 mQ3^4 msq^2 mU3^2-3 mQ3^6 msq^2 mU3^2-18 m3^4 msq^4 mU3^2-6 m3^2 mQ3^2 msq^4 mU3^2+17 m3^6 mU3^4-44 m3^4 mQ3^2 mU3^4+40 m3^2 mQ3^4 mU3^4-13 mQ3^6 mU3^4+3 m3^4 msq^2 mU3^4-24 m3^2 mQ3^2 msq^2 mU3^4-3 mQ3^4 msq^2 mU3^4+9 m3^2 msq^4 mU3^4+3 mQ3^2 msq^4 mU3^4))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2)+(-((640 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2)))-(640 m3)/(3 (mQ3^2-mU3^2))-(320 m3 mQ3^2)/(3 (m3^2-mQ3^2) (mQ3^2-mU3^2))+(320 (m3^3 mQ3^2-m3 mQ3^4+12 m3 mQ3^2 msq^2-6 m3 msq^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) Xt+(160 (18 m3^8-33 m3^6 mQ3^2+15 m3^4 mQ3^4+18 m3^6 msq^2-15 m3^4 mQ3^2 msq^2+9 m3^2 mQ3^4 msq^2-35 m3^6 mU3^2+64 m3^4 mQ3^2 mU3^2-29 m3^2 mQ3^4 mU3^2-15 m3^4 msq^2 mU3^2-12 m3^2 mQ3^2 msq^2 mU3^2+3 mQ3^4 msq^2 mU3^2+17 m3^4 mU3^4-31 m3^2 mQ3^2 mU3^4+14 mQ3^4 mU3^4+9 m3^2 msq^2 mU3^4+3 mQ3^2 msq^2 mU3^4) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(-((1280 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)))+(640 m3 (3 mQ3^2+mU3^2))/(3 (mQ3^2-mU3^2)^3)) Xt^3+((160 (2 m3^4 msq^2+6 m3^2 mQ3^2 msq^2-3 m3^2 msq^4-mQ3^2 msq^4))/((m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^2)-(80 (mQ3^2+mU3^2) (11 m3^8-22 m3^6 mQ3^2+11 m3^4 mQ3^4+18 m3^6 msq^2-15 m3^4 mQ3^2 msq^2+9 m3^2 mQ3^4 msq^2-22 m3^6 mU3^2+44 m3^4 mQ3^2 mU3^2-22 m3^2 mQ3^4 mU3^2-15 m3^4 msq^2 mU3^2-12 m3^2 mQ3^2 msq^2 mU3^2+3 mQ3^4 msq^2 mU3^2+11 m3^4 mU3^4-22 m3^2 mQ3^2 mU3^4+11 mQ3^4 mU3^4+9 m3^2 msq^2 mU3^4+3 mQ3^2 msq^2 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(80 (4 m3^8-3 mQ3^4 mU3^2 (5 mQ3^2+mU3^2)+m3^6 (7 mQ3^2+3 mU3^2)-m3^4 (29 mQ3^4+15 mQ3^2 mU3^2+6 mU3^4)+m3^2 (16 mQ3^6+31 mQ3^4 mU3^2+7 mQ3^2 mU3^4)))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) Xt^4+(-((640 m3 mQ3^2)/(3 (m3^2-mQ3^2) (mQ3^2-mU3^2)^3))-(1280 (2 m3 mQ3^2 msq^2-m3 msq^4))/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)+(640 m3 msq^2 (mQ3^2+mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) Xt^5) Log[msq^2/MR^2]+(-24 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3))-24 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt-1/(mQ3^2-mU3^2) 48 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)) Xt^2-(48 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^3)/(mQ3^2-mU3^2)+1/(mQ3^2-mU3^2)^3 24 (mQ3^2+mU3^2) (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)) Xt^4+(24 (mQ3^2+mU3^2) ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^5)/(mQ3^2-mU3^2)^3) Log[msq^2/MR^2]^2+(-((8 (348 m3^12-264 m3^10 mQ3^2-201 m3^8 mQ3^4+192 m3^6 mQ3^6-20 m3^4 mQ3^8+9 m3^2 mQ3^10-90 m3^8 mQ3^2 msq^2-30 m3^6 mQ3^4 msq^2-276 m3^10 mU3^2-704 m3^8 mQ3^2 mU3^2+1498 m3^6 mQ3^4 mU3^2-693 m3^4 mQ3^6 mU3^2-32 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2-90 m3^8 msq^2 mU3^2+540 m3^6 mQ3^2 msq^2 mU3^2-180 m3^4 mQ3^4 msq^2 mU3^2+90 m3^2 mQ3^6 msq^2 mU3^2-187 m3^8 mU3^4+1608 m3^6 mQ3^2 mU3^4-2017 m3^4 mQ3^4 mU3^4+820 m3^2 mQ3^6 mU3^4+4 mQ3^8 mU3^4-30 m3^6 msq^2 mU3^4-180 m3^4 mQ3^2 msq^2 mU3^4-180 m3^2 mQ3^4 msq^2 mU3^4+30 mQ3^6 msq^2 mU3^4+158 m3^6 mU3^6-787 m3^4 mQ3^2 mU3^6+828 m3^2 mQ3^4 mU3^6-299 mQ3^6 mU3^6+90 m3^2 mQ3^2 msq^2 mU3^6+30 mQ3^4 msq^2 mU3^6+9 m3^4 mU3^8+3 m3^2 mQ3^2 mU3^8))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))+(64 (109 m3^7 mQ3^4-104 m3^5 mQ3^6-2 m3^3 mQ3^8+3 m3 mQ3^10-30 m3^5 mQ3^4 msq^2-234 m3^7 mQ3^2 mU3^2+123 m3^5 mQ3^4 mU3^2+60 m3^3 mQ3^6 mU3^2-10 m3 mQ3^8 mU3^2+60 m3^5 mQ3^2 msq^2 mU3^2+30 m3 mQ3^6 msq^2 mU3^2+109 m3^7 mU3^4+138 m3^5 mQ3^2 mU3^4-159 m3^3 mQ3^4 mU3^4+19 m3 mQ3^6 mU3^4-30 m3^5 msq^2 mU3^4-60 m3 mQ3^4 msq^2 mU3^4-109 m3^5 mU3^6+50 m3^3 mQ3^2 mU3^6+4 m3 mQ3^4 mU3^6+30 m3 mQ3^2 msq^2 mU3^6+3 m3^3 mU3^8) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(16 (468 m3^12 mQ3^2-402 m3^10 mQ3^4-292 m3^8 mQ3^6+269 m3^6 mQ3^8+9 m3^4 mQ3^10-90 m3^8 mQ3^4 msq^2-30 m3^6 mQ3^6 msq^2+468 m3^12 mU3^2-3692 m3^10 mQ3^2 mU3^2+4624 m3^8 mQ3^4 mU3^2-1268 m3^6 mQ3^6 mU3^2-337 m3^4 mQ3^8 mU3^2-15 m3^2 mQ3^10 mU3^2+180 m3^8 mQ3^2 msq^2 mU3^2+30 m3^6 mQ3^4 msq^2 mU3^2+360 m3^4 mQ3^6 msq^2 mU3^2-90 m3^2 mQ3^8 msq^2 mU3^2-402 m3^10 mU3^4+4624 m3^8 mQ3^2 mU3^4-6418 m3^6 mQ3^4 mU3^4+2420 m3^4 mQ3^6 mU3^4+142 m3^2 mQ3^8 mU3^4-6 mQ3^10 mU3^4-90 m3^8 msq^2 mU3^4+30 m3^6 mQ3^2 msq^2 mU3^4-720 m3^4 mQ3^4 msq^2 mU3^4+90 m3^2 mQ3^6 msq^2 mU3^4-30 mQ3^8 msq^2 mU3^4-292 m3^8 mU3^6-1268 m3^6 mQ3^2 mU3^6+2420 m3^4 mQ3^4 mU3^6-1166 m3^2 mQ3^6 mU3^6+26 mQ3^8 mU3^6-30 m3^6 msq^2 mU3^6+360 m3^4 mQ3^2 msq^2 mU3^6+90 m3^2 mQ3^4 msq^2 mU3^6+60 mQ3^6 msq^2 mU3^6+269 m3^6 mU3^8-337 m3^4 mQ3^2 mU3^8+142 m3^2 mQ3^4 mU3^8+26 mQ3^6 mU3^8-90 m3^2 mQ3^2 msq^2 mU3^8-30 mQ3^4 msq^2 mU3^8+9 m3^4 mU3^10-15 m3^2 mQ3^2 mU3^10-6 mQ3^4 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(128 (31 m3^9+63 m3^7 mQ3^2-86 m3^5 mQ3^4+3 m3^3 mQ3^6-30 m3^5 mQ3^2 msq^2+63 m3^7 mU3^2-424 m3^5 mQ3^2 mU3^2+310 m3^3 mQ3^4 mU3^2-6 m3 mQ3^6 mU3^2-30 m3^5 msq^2 mU3^2+120 m3^3 mQ3^2 msq^2 mU3^2-30 m3 mQ3^4 msq^2 mU3^2-86 m3^5 mU3^4+310 m3^3 mQ3^2 mU3^4-175 m3 mQ3^4 mU3^4-30 m3 mQ3^2 msq^2 mU3^4+3 m3^3 mU3^6-6 m3 mQ3^2 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(8 (816 m3^12 mQ3^4-1060 m3^10 mQ3^6+22 m3^8 mQ3^8+201 m3^6 mQ3^10+67 m3^4 mQ3^12-18 m3^2 mQ3^14-90 m3^8 mQ3^6 msq^2-30 m3^6 mQ3^8 msq^2+3136 m3^12 mQ3^2 mU3^2-10916 m3^10 mQ3^4 mU3^2+11716 m3^8 mQ3^6 mU3^2-3835 m3^6 mQ3^8 mU3^2-234 m3^4 mQ3^10 mU3^2+91 m3^2 mQ3^12 mU3^2-6 mQ3^14 mU3^2+90 m3^8 mQ3^4 msq^2 mU3^2+360 m3^4 mQ3^8 msq^2 mU3^2-90 m3^2 mQ3^10 msq^2 mU3^2+816 m3^12 mU3^4-10988 m3^10 mQ3^2 mU3^4+24976 m3^8 mQ3^4 mU3^4-21226 m3^6 mQ3^6 mU3^4+6899 m3^4 mQ3^8 mU3^4-575 m3^2 mQ3^10 mU3^4-2 mQ3^12 mU3^4+90 m3^8 mQ3^2 msq^2 mU3^4+60 m3^6 mQ3^4 msq^2 mU3^4-360 m3^4 mQ3^6 msq^2 mU3^4-30 mQ3^10 msq^2 mU3^4-1036 m3^10 mU3^6+11772 m3^8 mQ3^2 mU3^6-20498 m3^6 mQ3^4 mU3^6+12164 m3^4 mQ3^6 mU3^6-2298 m3^2 mQ3^8 mU3^6+216 mQ3^10 mU3^6-90 m3^8 msq^2 mU3^6-360 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+30 mQ3^8 msq^2 mU3^6-6 m3^8 mU3^8-4191 m3^6 mQ3^2 mU3^8+6465 m3^4 mQ3^4 mU3^8-2320 m3^2 mQ3^6 mU3^8-248 mQ3^8 mU3^8-30 m3^6 msq^2 mU3^8+360 m3^4 mQ3^2 msq^2 mU3^8+30 mQ3^6 msq^2 mU3^8+269 m3^6 mU3^10+70 m3^4 mQ3^2 mU3^10-433 m3^2 mQ3^4 mU3^10+206 mQ3^6 mU3^10-90 m3^2 mQ3^2 msq^2 mU3^10-30 mQ3^4 msq^2 mU3^10+9 m3^4 mU3^12-15 m3^2 mQ3^2 mU3^12-6 mQ3^4 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^4)+(64 (64 m3^9 mQ3^2+44 m3^7 mQ3^4-102 m3^5 mQ3^6+13 m3^3 mQ3^8-6 m3 mQ3^10-30 m3^5 mQ3^4 msq^2+64 m3^9 mU3^2+188 m3^7 mQ3^2 mU3^2-638 m3^5 mQ3^4 mU3^2+352 m3^3 mQ3^6 mU3^2+14 m3 mQ3^8 mU3^2-60 m3^5 mQ3^2 msq^2 mU3^2+120 m3^3 mQ3^4 msq^2 mU3^2-30 m3 mQ3^6 msq^2 mU3^2+44 m3^7 mU3^4-668 m3^5 mQ3^2 mU3^4+920 m3^3 mQ3^4 mU3^4-298 m3 mQ3^6 mU3^4-30 m3^5 msq^2 mU3^4+120 m3^3 mQ3^2 msq^2 mU3^4-60 m3 mQ3^4 msq^2 mU3^4-92 m3^5 mU3^6+372 m3^3 mQ3^2 mU3^6-268 m3 mQ3^4 mU3^6-30 m3 mQ3^2 msq^2 mU3^6+3 m3^3 mU3^8-6 m3 mQ3^2 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)+(5120 m3^2 mQ3^2 mU3^2 Xt^6)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^4)+((80 (2 m3^4 msq^2+6 m3^2 mQ3^2 msq^2-3 m3^2 msq^4-mQ3^2 msq^4))/(m3^2-mQ3^2)^3+(80 (2 m3^4 msq^2-3 m3^2 msq^4+6 m3^2 msq^2 mU3^2-msq^4 mU3^2))/(m3^2-mU3^2)^3+(80 (-2 mQ3^4 mU3^4+2 m3^6 (mQ3^2+mU3^2)+4 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)-m3^4 (mQ3^4+8 mQ3^2 mU3^2+mU3^4)))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(320 (m3^7-m3^5 mQ3^2+12 m3^5 msq^2-12 m3^3 msq^4+6 m3 mQ3^2 msq^4-m3^5 mU3^2+m3^3 mQ3^2 mU3^2-12 m3 mQ3^2 msq^2 mU3^2+6 m3 msq^4 mU3^2) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(-((160 (2 m3^4 msq^2+6 m3^2 mQ3^2 msq^2-3 m3^2 msq^4-mQ3^2 msq^4))/((m3^2-mQ3^2)^3 (mQ3^2-mU3^2)))-(160 (2 m3^4 msq^2-3 m3^2 msq^4+6 m3^2 msq^2 mU3^2-msq^4 mU3^2))/((m3^2-mU3^2)^3 (-mQ3^2+mU3^2))+(160 (2 m3^8 (mQ3^2+mU3^2)+2 mQ3^4 mU3^4 (mQ3^2+mU3^2)-4 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)^2+3 m3^4 (mQ3^2+mU3^2)^3-2 m3^6 (3 mQ3^4+2 mQ3^2 mU3^2+3 mU3^4)))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) Xt^2+((1280 (2 m3 mQ3^2 msq^2-m3 msq^4))/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)+(1280 (-m3 msq^4+2 m3 msq^2 mU3^2))/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)+(640 (-2 m3 mQ3^2 mU3^2+m3^3 (mQ3^2+mU3^2)))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)) Xt^3+((80 (2 m3^4 msq^2+6 m3^2 mQ3^2 msq^2-3 m3^2 msq^4-mQ3^2 msq^4) (mQ3^2+mU3^2))/((m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^3)+(80 (mQ3^2+mU3^2) (2 m3^4 msq^2-3 m3^2 msq^4+6 m3^2 msq^2 mU3^2-msq^4 mU3^2))/((m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^3)-(80 (mQ3^2+mU3^2) (4 m3^8 (mQ3^2+mU3^2)+4 mQ3^4 mU3^4 (mQ3^2+mU3^2)-8 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)^2-2 m3^6 (5 mQ3^4+6 mQ3^2 mU3^2+5 mU3^4)+m3^4 (5 mQ3^6+19 mQ3^4 mU3^2+19 mQ3^2 mU3^4+5 mU3^6)))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) Xt^4+(-((640 (2 m3 mQ3^2 msq^2-m3 msq^4) (mQ3^2+mU3^2))/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4))-(640 (mQ3^2+mU3^2) (-m3 msq^4+2 m3 msq^2 mU3^2))/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)-(320 m3 (mQ3^2+mU3^2) (-2 mQ3^2 mU3^2+m3^2 (mQ3^2+mU3^2)))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^4)) Xt^5) Log[msq^2/MR^2]) Log[mU3^2/MR^2]+(-((8 (m3^4-34 m3^2 mQ3^2+17 mQ3^4) (3 m3^4-4 m3^2 mU3^2+2 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(8 (12 m3^10 mQ3^2-14 m3^8 mQ3^4+34 m3^6 mQ3^6-29 m3^4 mQ3^8+9 m3^2 mQ3^10+64 m3^8 mQ3^2 mU3^2-188 m3^6 mQ3^4 mU3^2+120 m3^4 mQ3^6 mU3^2-35 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2-78 m3^6 mQ3^2 mU3^4+175 m3^4 mQ3^4 mU3^4-65 m3^2 mQ3^6 mU3^4+4 mQ3^8 mU3^4+26 m3^4 mQ3^2 mU3^6-57 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(4 (128 m3^10+44 m3^8 mQ3^2-2 m3^6 mQ3^4+9 m3^4 mQ3^6-60 m3^6 mQ3^2 msq^2+90 m3^4 mQ3^2 msq^4-556 m3^8 mU3^2+64 m3^6 mQ3^2 mU3^2-33 m3^4 mQ3^4 mU3^2-6 m3^2 mQ3^6 mU3^2+60 m3^6 msq^2 mU3^2-120 m3^4 mQ3^2 msq^2 mU3^2-90 m3^4 msq^4 mU3^2-60 m3^2 mQ3^2 msq^4 mU3^2+706 m3^6 mU3^4-51 m3^4 mQ3^2 mU3^4+36 m3^2 mQ3^4 mU3^4-3 mQ3^6 mU3^4+120 m3^4 msq^2 mU3^4+180 m3^2 mQ3^2 msq^2 mU3^4+60 m3^2 msq^4 mU3^4-30 mQ3^2 msq^4 mU3^4-437 m3^4 mU3^6-38 m3^2 mQ3^2 mU3^6-mQ3^4 mU3^6-180 m3^2 msq^2 mU3^6+30 msq^4 mU3^6+136 m3^2 mU3^8+5 mQ3^2 mU3^8-mU3^10))/(3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2))+(8 (m3^4-34 m3^2 mU3^2+17 mU3^4) (-2 mQ3^4 mU3^4+2 m3^6 (mQ3^2+mU3^2)+4 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)-m3^4 (mQ3^4+8 mQ3^2 mU3^2+mU3^4)))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^4)-(32 (20 m3^11 mQ3^2-56 m3^9 mQ3^4+12 m3^7 mQ3^6-5 m3^5 mQ3^8+3 m3^3 mQ3^10-30 m3^7 mQ3^2 msq^4+60 m3^5 mQ3^4 msq^4-30 m3^3 mQ3^6 msq^4-20 m3^11 mU3^2-51 m3^9 mQ3^2 mU3^2+229 m3^7 mQ3^4 mU3^2-62 m3^5 mQ3^6 mU3^2-5 m3^3 mQ3^8 mU3^2-3 m3 mQ3^10 mU3^2+60 m3^7 mQ3^2 msq^2 mU3^2-120 m3^5 mQ3^4 msq^2 mU3^2+60 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2-30 m3^5 mQ3^2 msq^4 mU3^2-30 m3^3 mQ3^4 msq^4 mU3^2+30 m3 mQ3^6 msq^4 mU3^2+91 m3^9 mU3^4-50 m3^7 mQ3^2 mU3^4-263 m3^5 mQ3^4 mU3^4+104 m3^3 mQ3^6 mU3^4+10 m3 mQ3^8 mU3^4-60 m3^7 msq^2 mU3^4+60 m3^5 mQ3^2 msq^2 mU3^4+60 m3^3 mQ3^4 msq^2 mU3^4-60 m3 mQ3^6 msq^2 mU3^4-30 m3^5 msq^4 mU3^4+60 m3^3 mQ3^2 msq^4 mU3^4-30 m3 mQ3^4 msq^4 mU3^4-127 m3^7 mU3^6+194 m3^5 mQ3^2 mU3^6+27 m3^3 mQ3^4 mU3^6-38 m3 mQ3^6 mU3^6+60 m3^5 msq^2 mU3^6-120 m3^3 mQ3^2 msq^2 mU3^6+60 m3 mQ3^4 msq^2 mU3^6+40 m3^5 mU3^8-65 m3^3 mQ3^2 mU3^8+15 m3 mQ3^4 mU3^8) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(8 (128 m3^16-846 m3^14 mQ3^2+1164 m3^12 mQ3^4-345 m3^10 mQ3^6-185 m3^8 mQ3^8+87 m3^6 mQ3^10-27 m3^4 mQ3^12-60 m3^12 mQ3^2 msq^2+180 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2+60 m3^6 mQ3^8 msq^2+90 m3^10 mQ3^2 msq^4-270 m3^8 mQ3^4 msq^4+270 m3^6 mQ3^6 msq^4-90 m3^4 mQ3^8 msq^4-558 m3^14 mU3^2+3662 m3^12 mQ3^2 mU3^2-5053 m3^10 mQ3^4 mU3^2+1965 m3^8 mQ3^6 mU3^2+41 m3^6 mQ3^8 mU3^2+45 m3^4 mQ3^10 mU3^2+18 m3^2 mQ3^12 mU3^2+60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2+540 m3^8 mQ3^4 msq^2 mU3^2-420 m3^6 mQ3^6 msq^2 mU3^2+120 m3^4 mQ3^8 msq^2 mU3^2-90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-90 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+60 m3^2 mQ3^8 msq^4 mU3^2+426 m3^12 mU3^4-4533 m3^10 mQ3^2 mU3^4+6223 m3^8 mQ3^4 mU3^4-2004 m3^6 mQ3^6 mU3^4-226 m3^4 mQ3^8 mU3^4-135 m3^2 mQ3^10 mU3^4+9 mQ3^12 mU3^4+120 m3^10 msq^2 mU3^4-180 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4+420 m3^4 mQ3^6 msq^2 mU3^4-180 m3^2 mQ3^8 msq^2 mU3^4+60 m3^8 msq^4 mU3^4-210 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4-150 m3^2 mQ3^6 msq^4 mU3^4+30 mQ3^8 msq^4 mU3^4+319 m3^10 mU3^6+1857 m3^8 mQ3^2 mU3^6-2840 m3^6 mQ3^4 mU3^6+412 m3^4 mQ3^6 mU3^6+489 m3^2 mQ3^8 mU3^6+3 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6+540 m3^6 mQ3^2 msq^2 mU3^6-540 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+30 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6+90 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-400 m3^8 mU3^8-165 m3^6 mQ3^2 mU3^8+647 m3^4 mQ3^4 mU3^8-91 m3^2 mQ3^6 mU3^8-111 mQ3^8 mU3^8+93 m3^6 mU3^10-7 m3^4 mQ3^2 mU3^10-93 m3^2 mQ3^4 mU3^10+31 mQ3^6 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)-(32 (2 m3^13-33 m3^11 mQ3^2-40 m3^9 mQ3^4+31 m3^7 mQ3^6+30 m3^5 mQ3^8-18 m3^3 mQ3^10-60 m3^7 mQ3^2 msq^4+120 m3^5 mQ3^4 msq^4-60 m3^3 mQ3^6 msq^4-43 m3^11 mU3^2+403 m3^9 mQ3^2 mU3^2-143 m3^7 mQ3^4 mU3^2-185 m3^5 mQ3^6 mU3^2+30 m3^3 mQ3^8 mU3^2+18 m3 mQ3^10 mU3^2+120 m3^7 mQ3^2 msq^2 mU3^2-240 m3^5 mQ3^4 msq^2 mU3^2+120 m3^3 mQ3^6 msq^2 mU3^2+60 m3^7 msq^4 mU3^2-60 m3^5 mQ3^2 msq^4 mU3^2-60 m3^3 mQ3^4 msq^4 mU3^2+60 m3 mQ3^6 msq^4 mU3^2+243 m3^9 mU3^4-1217 m3^7 mQ3^2 mU3^4+869 m3^5 mQ3^4 mU3^4+93 m3^3 mQ3^6 mU3^4-60 m3 mQ3^8 mU3^4-120 m3^7 msq^2 mU3^4+120 m3^5 mQ3^2 msq^2 mU3^4+120 m3^3 mQ3^4 msq^2 mU3^4-120 m3 mQ3^6 msq^2 mU3^4-60 m3^5 msq^4 mU3^4+120 m3^3 mQ3^2 msq^4 mU3^4-60 m3 mQ3^4 msq^4 mU3^4-375 m3^7 mU3^6+1351 m3^5 mQ3^2 mU3^6-1053 m3^3 mQ3^4 mU3^6+93 m3 mQ3^6 mU3^6+120 m3^5 msq^2 mU3^6-240 m3^3 mQ3^2 msq^2 mU3^6+120 m3 mQ3^4 msq^2 mU3^6+141 m3^5 mU3^8-408 m3^3 mQ3^2 mU3^8+271 m3 mQ3^4 mU3^8) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(4 (1156 m3^16 mQ3^2-3032 m3^14 mQ3^4+2450 m3^12 mQ3^6-503 m3^10 mQ3^8-155 m3^8 mQ3^10+87 m3^6 mQ3^12-27 m3^4 mQ3^14-60 m3^12 mQ3^4 msq^2+180 m3^10 mQ3^6 msq^2-180 m3^8 mQ3^8 msq^2+60 m3^6 mQ3^10 msq^2+90 m3^10 mQ3^4 msq^4-270 m3^8 mQ3^6 msq^4+270 m3^6 mQ3^8 msq^4-90 m3^4 mQ3^10 msq^4+124 m3^16 mU3^2-5372 m3^14 mQ3^2 mU3^2+14826 m3^12 mQ3^4 mU3^2-13214 m3^10 mQ3^6 mU3^2+3560 m3^8 mQ3^8 mU3^2+136 m3^6 mQ3^10 mU3^2+18 m3^4 mQ3^12 mU3^2+18 m3^2 mQ3^14 mU3^2-120 m3^10 mQ3^4 msq^2 mU3^2+360 m3^8 mQ3^6 msq^2 mU3^2-360 m3^6 mQ3^8 msq^2 mU3^2+120 m3^4 mQ3^10 msq^2 mU3^2-60 m3^8 mQ3^4 msq^4 mU3^2+180 m3^6 mQ3^6 msq^4 mU3^2-180 m3^4 mQ3^8 msq^4 mU3^2+60 m3^2 mQ3^10 msq^4 mU3^2-540 m3^14 mU3^4+9194 m3^12 mQ3^2 mU3^4-25992 m3^10 mQ3^4 mU3^4+24798 m3^8 mQ3^6 mU3^4-7151 m3^6 mQ3^8 mU3^4-321 m3^4 mQ3^10 mU3^4-117 m3^2 mQ3^12 mU3^4+9 mQ3^14 mU3^4+60 m3^12 msq^2 mU3^4-180 m3^10 mQ3^2 msq^2 mU3^4+360 m3^8 mQ3^4 msq^2 mU3^4-600 m3^6 mQ3^6 msq^2 mU3^4+540 m3^4 mQ3^8 msq^2 mU3^4-180 m3^2 mQ3^10 msq^2 mU3^4-90 m3^10 msq^4 mU3^4+270 m3^8 mQ3^2 msq^4 mU3^4-300 m3^6 mQ3^4 msq^4 mU3^4+180 m3^4 mQ3^6 msq^4 mU3^4-90 m3^2 mQ3^8 msq^4 mU3^4+30 mQ3^10 msq^4 mU3^4-214 m3^12 mU3^6-3566 m3^10 mQ3^2 mU3^6+15344 m3^8 mQ3^4 mU3^6-17852 m3^6 mQ3^6 mU3^6+5786 m3^4 mQ3^8 mU3^6+490 m3^2 mQ3^10 mU3^6+12 mQ3^12 mU3^6+120 m3^10 msq^2 mU3^6-360 m3^8 mQ3^2 msq^2 mU3^6+360 m3^6 mQ3^4 msq^2 mU3^6-120 m3^4 mQ3^6 msq^2 mU3^6+60 m3^8 msq^4 mU3^6-180 m3^6 mQ3^2 msq^4 mU3^6+180 m3^4 mQ3^4 msq^4 mU3^6-60 m3^2 mQ3^6 msq^4 mU3^6+1627 m3^10 mU3^8-4559 m3^8 mQ3^2 mU3^8+2717 m3^6 mQ3^4 mU3^8+2045 m3^4 mQ3^6 mU3^8-1568 m3^2 mQ3^8 mU3^8-142 mQ3^10 mU3^8-180 m3^8 msq^2 mU3^8+540 m3^6 mQ3^2 msq^2 mU3^8-540 m3^4 mQ3^4 msq^2 mU3^8+180 m3^2 mQ3^6 msq^2 mU3^8+30 m3^6 msq^4 mU3^8-90 m3^4 mQ3^2 msq^4 mU3^8+90 m3^2 mQ3^4 msq^4 mU3^8-30 mQ3^6 msq^4 mU3^8-1148 m3^8 mU3^10+3264 m3^6 mQ3^2 mU3^10-3856 m3^4 mQ3^4 mU3^10+1792 m3^2 mQ3^6 mU3^10-148 mQ3^8 mU3^10+159 m3^6 mU3^12-141 m3^4 mQ3^2 mU3^12+9 m3^2 mQ3^4 mU3^12-3 mQ3^6 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)^4)-(32 (14 m3^11 mQ3^4+6 m3^9 mQ3^6-8 m3^7 mQ3^8-15 m3^5 mQ3^10+9 m3^3 mQ3^12+30 m3^7 mQ3^4 msq^4-60 m3^5 mQ3^6 msq^4+30 m3^3 mQ3^8 msq^4-32 m3^11 mQ3^2 mU3^2-155 m3^9 mQ3^4 mU3^2+177 m3^7 mQ3^6 mU3^2+23 m3^5 mQ3^8 mU3^2-6 m3^3 mQ3^10 mU3^2-9 m3 mQ3^12 mU3^2-60 m3^7 mQ3^4 msq^2 mU3^2+120 m3^5 mQ3^6 msq^2 mU3^2-60 m3^3 mQ3^8 msq^2 mU3^2-30 m3^5 mQ3^4 msq^4 mU3^2+60 m3^3 mQ3^6 msq^4 mU3^2-30 m3 mQ3^8 msq^4 mU3^2+18 m3^11 mU3^4-182 m3^9 mQ3^2 mU3^4+719 m3^7 mQ3^4 mU3^4-621 m3^5 mQ3^6 mU3^4+25 m3^3 mQ3^8 mU3^4+21 m3 mQ3^10 mU3^4+60 m3^5 mQ3^4 msq^2 mU3^4-120 m3^3 mQ3^6 msq^2 mU3^4+60 m3 mQ3^8 msq^2 mU3^4-30 m3^7 msq^4 mU3^4+60 m3^5 mQ3^2 msq^4 mU3^4-30 m3^3 mQ3^4 msq^4 mU3^4-149 m3^9 mU3^6+783 m3^7 mQ3^2 mU3^6-1297 m3^5 mQ3^4 mU3^6+731 m3^3 mQ3^6 mU3^6-56 m3 mQ3^8 mU3^6+60 m3^7 msq^2 mU3^6-120 m3^5 mQ3^2 msq^2 mU3^6+60 m3^3 mQ3^4 msq^2 mU3^6+30 m3^5 msq^4 mU3^6-60 m3^3 mQ3^2 msq^4 mU3^6+30 m3 mQ3^4 msq^4 mU3^6+249 m3^7 mU3^8-868 m3^5 mQ3^2 mU3^8+894 m3^3 mQ3^4 mU3^8-261 m3 mQ3^6 mU3^8-60 m3^5 msq^2 mU3^8+120 m3^3 mQ3^2 msq^2 mU3^8-60 m3 mQ3^4 msq^2 mU3^8-102 m3^5 mU3^10+267 m3^3 mQ3^2 mU3^10-175 m3 mQ3^4 mU3^10) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^5)+(1280 m3^2 (mQ3^2+mU3^2) (2 m3^2 mQ3^2 mU3^2+m3^2 mU3^4-3 mQ3^2 mU3^4) Xt^6)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^5)) Log[mU3^2/MR^2]^2+(-((32 m3^4 (m3^4-34 m3^2 mQ3^2+17 mQ3^4))/(3 (m3^2-mQ3^2)^4))+(8 (-128 m3^12+294 m3^10 mQ3^2-98 m3^8 mQ3^4-9 m3^6 mQ3^6-20 m3^4 mQ3^8+9 m3^2 mQ3^10+346 m3^10 mU3^2-902 m3^8 mQ3^2 mU3^2+251 m3^6 mQ3^4 mU3^2+151 m3^4 mQ3^6 mU3^2-41 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2-280 m3^8 mU3^4+1025 m3^6 mQ3^2 mU3^4-417 m3^4 mQ3^4 mU3^4-41 m3^2 mQ3^6 mU3^4+mQ3^8 mU3^4+13 m3^6 mU3^6-391 m3^4 mQ3^2 mU3^6+167 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6+37 m3^4 mU3^8+34 m3^2 mQ3^2 mU3^8-23 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+((128 (m3^7-50 m3^5 mQ3^2+17 m3^3 mQ3^4))/(3 (m3^2-mQ3^2)^3 (mQ3^2-mU3^2))-(64 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt-(16 (128 m3^12-806 m3^10 mQ3^2+98 m3^8 mQ3^4+9 m3^6 mQ3^6+20 m3^4 mQ3^8-9 m3^2 mQ3^10-346 m3^10 mU3^2+2438 m3^8 mQ3^2 mU3^2-251 m3^6 mQ3^4 mU3^2-151 m3^4 mQ3^6 mU3^2+41 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2+280 m3^8 mU3^4-2561 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4+41 m3^2 mQ3^6 mU3^4-mQ3^8 mU3^4-13 m3^6 mU3^6+903 m3^4 mQ3^2 mU3^6-167 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6-37 m3^4 mU3^8-34 m3^2 mQ3^2 mU3^8+23 mQ3^4 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+((64 m3 (m3^4-34 m3^2 mQ3^2+17 mQ3^4) (2 m3^2-mQ3^2-mU3^2))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)+(128 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)) Xt^3+(8 (1156 m3^12 mQ3^2-1968 m3^10 mQ3^4+746 m3^8 mQ3^6-25 m3^6 mQ3^8+20 m3^4 mQ3^10-9 m3^2 mQ3^12+124 m3^12 mU3^2-4100 m3^10 mQ3^2 mU3^2+6466 m3^8 mQ3^4 mU3^2-2186 m3^6 mQ3^6 mU3^2-29 m3^4 mQ3^8 mU3^2+32 m3^2 mQ3^10 mU3^2-3 mQ3^12 mU3^2-332 m3^10 mU3^4+5326 m3^8 mQ3^2 mU3^4-7596 m3^6 mQ3^4 mU3^4+2210 m3^4 mQ3^6 mU3^4-20 m3^2 mQ3^8 mU3^4-4 mQ3^10 mU3^4+262 m3^8 mU3^6-2990 m3^6 mQ3^2 mU3^6+3712 m3^4 mQ3^4 mU3^6-774 m3^2 mQ3^6 mU3^6+14 mQ3^8 mU3^6-3 m3^6 mU3^8+526 m3^4 mQ3^2 mU3^8-543 m3^2 mQ3^4 mU3^8+4 mQ3^6 mU3^8-39 m3^4 mU3^10+34 m3^2 mQ3^2 mU3^10-11 mQ3^4 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(-((1024 m3 mQ3^2 (-2 m3^2+mQ3^2+mU3^2))/(3 (m3^2-mQ3^2) (mQ3^2-mU3^2)^4))-(64 (mQ3^2+mU3^2) (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) Xt^5) PolyLog[2,1-m3^2/mQ3^2]+(-12 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2)))-(640 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-1/(mQ3^2-mU3^2) 24 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2))) Xt^2-(1280 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^3)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+1/(mQ3^2-mU3^2)^3 12 (2 m3^2-2 msq^2) (mQ3^2+mU3^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2))) Xt^4+(640 (m3^2-msq^2) (mQ3^2+mU3^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/msq^2]+((40 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(-m3^2+mQ3^2)^3-(640 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(80 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2) Xt^2)/((-m3^2+mQ3^2)^3 (mQ3^2-mU3^2))-(1280 m3 (mQ3^2-msq^2)^2 Xt^3)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)-(40 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2) (mQ3^2+mU3^2) Xt^4)/((-m3^2+mQ3^2)^3 (mQ3^2-mU3^2)^3)+(640 m3 (mQ3^2-msq^2)^2 (mQ3^2+mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) PolyLog[2,1-msq^2/mQ3^2]+((8 (128 m3^12-350 m3^10 mQ3^2+420 m3^8 mQ3^4-217 m3^6 mQ3^6+31 m3^4 mQ3^8-290 m3^10 mU3^2+762 m3^8 mQ3^2 mU3^2-821 m3^6 mQ3^4 mU3^2+323 m3^4 mQ3^6 mU3^2-34 m3^2 mQ3^8 mU3^2+98 m3^8 mU3^4-251 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4-167 m3^2 mQ3^6 mU3^4+23 mQ3^8 mU3^4+9 m3^6 mU3^6-151 m3^4 mQ3^2 mU3^6+41 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6+20 m3^4 mU3^8+41 m3^2 mQ3^2 mU3^8-mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(64 (20 m3^9-73 m3^7 mQ3^2+24 m3^5 mQ3^4-37 m3^7 mU3^2+135 m3^5 mQ3^2 mU3^2-78 m3^3 mQ3^4 mU3^2+19 m3^5 mU3^4-23 m3^3 mQ3^2 mU3^4+22 m3 mQ3^4 mU3^4-5 m3^3 mU3^6-7 m3 mQ3^2 mU3^6+3 m3 mU3^8) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(16 (128 m3^12-858 m3^10 mQ3^2+1304 m3^8 mQ3^4-525 m3^6 mQ3^6-37 m3^4 mQ3^8-294 m3^10 mU3^2+1414 m3^8 mQ3^2 mU3^2-2049 m3^6 mQ3^4 mU3^2+903 m3^4 mQ3^6 mU3^2-34 m3^2 mQ3^8 mU3^2+98 m3^8 mU3^4-251 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4-167 m3^2 mQ3^6 mU3^4+23 mQ3^8 mU3^4+9 m3^6 mU3^6-151 m3^4 mQ3^2 mU3^6+41 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6+20 m3^4 mU3^8+41 m3^2 mQ3^2 mU3^8-mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(64 (2 m3^7-33 m3^5 mQ3^2-6 m3^3 mQ3^4+27 m3 mQ3^6-37 m3^5 mU3^2+110 m3^3 mQ3^2 mU3^2-75 m3 mQ3^4 mU3^2-2 m3^3 mU3^4+20 m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)-(8 (1156 m3^12 mQ3^2-3048 m3^10 mQ3^4+2602 m3^8 mQ3^6-695 m3^6 mQ3^8-3 m3^4 mQ3^10+124 m3^12 mU3^2-3068 m3^10 mQ3^2 mU3^2+7406 m3^8 mQ3^4 mU3^2-6126 m3^6 mQ3^6 mU3^2+1718 m3^4 mQ3^8 mU3^2-102 m3^2 mQ3^10 mU3^2-284 m3^10 mU3^4+2702 m3^8 mQ3^2 mU3^4-5516 m3^6 mQ3^4 mU3^4+3984 m3^4 mQ3^6 mU3^4-883 m3^2 mQ3^8 mU3^4+57 mQ3^10 mU3^4+90 m3^8 mU3^6-474 m3^6 mQ3^2 mU3^6+882 m3^4 mQ3^4 mU3^6-502 m3^2 mQ3^6 mU3^6+4 mQ3^8 mU3^6+11 m3^6 mU3^8-201 m3^4 mQ3^2 mU3^8+184 m3^2 mQ3^4 mU3^8-54 mQ3^6 mU3^8+20 m3^4 mU3^10+32 m3^2 mQ3^2 mU3^10-4 mQ3^4 mU3^10-9 m3^2 mU3^12-3 mQ3^2 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)-(64 (14 m3^5 mQ3^2-11 m3^3 mQ3^4-6 m3 mQ3^6-18 m3^5 mU3^2+20 m3^3 mQ3^2 mU3^2+m3 mQ3^4 mU3^2-m3^3 mU3^4+4 m3 mQ3^2 mU3^4-3 m3 mU3^6) Xt^5)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) PolyLog[2,1-m3^2/mU3^2]+(-((4 (-24 m3^10 mQ3^2+28 m3^8 mQ3^4-68 m3^6 mQ3^6+58 m3^4 mQ3^8-18 m3^2 mQ3^10+24 m3^10 mU3^2+220 m3^6 mQ3^4 mU3^2-188 m3^4 mQ3^6 mU3^2+70 m3^2 mQ3^8 mU3^2-6 mQ3^10 mU3^2-28 m3^8 mU3^4-220 m3^6 mQ3^2 mU3^4+16 m3^2 mQ3^6 mU3^4-8 mQ3^8 mU3^4+68 m3^6 mU3^6+188 m3^4 mQ3^2 mU3^6-16 m3^2 mQ3^4 mU3^6-58 m3^4 mU3^8-70 m3^2 mQ3^2 mU3^8+8 mQ3^4 mU3^8+18 m3^2 mU3^10+6 mQ3^2 mU3^10))/(3 (-m3^2+mQ3^2)^3 (m3^2-mU3^2)^3))+(32 (10 m3^5 mQ3^2-10 m3^3 mQ3^4+6 m3 mQ3^6-10 m3^5 mU3^2-8 m3 mQ3^4 mU3^2+10 m3^3 mU3^4+8 m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(16 (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(128 (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(8 (mQ3^2+mU3^2) (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(64 (mQ3^2+mU3^2) (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) PolyLog[2,1-mQ3^2/mU3^2]+((40 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(-m3^2+mU3^2)^3-(640 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(160 (-msq^2+mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4) Xt^2)/((mQ3^2-mU3^2) (-m3^2+mU3^2)^3)+(1280 m3 (-msq^2+mU3^2)^2 Xt^3)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)-(80 (mQ3^2+mU3^2) (-msq^2+mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4) Xt^4)/((mQ3^2-mU3^2)^3 (-m3^2+mU3^2)^3)-(640 m3 (mQ3^2+mU3^2) (-msq^2+mU3^2)^2 Xt^5)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) PolyLog[2,1-msq^2/mU3^2])+Log[mU3^2/MR^2] ((4 (-4 m3^12 mQ3^4+8 m3^10 mQ3^6-4 m3^8 mQ3^8+611 m3^10 mQ3^4 mU3^2-704 m3^8 mQ3^6 mU3^2+207 m3^6 mQ3^8 mU3^2-18 m3^4 mQ3^10 mU3^2-360 m3^8 mQ3^4 msq^2 mU3^2+300 m3^6 mQ3^6 msq^2 mU3^2-180 m3^4 mQ3^8 msq^2 mU3^2+4 m3^12 mU3^4+33 m3^10 mQ3^2 mU3^4-2127 m3^8 mQ3^4 mU3^4+2217 m3^6 mQ3^6 mU3^4-541 m3^4 mQ3^8 mU3^4+30 m3^2 mQ3^10 mU3^4+360 m3^8 mQ3^2 msq^2 mU3^4+540 m3^6 mQ3^4 msq^2 mU3^4-240 m3^4 mQ3^6 msq^2 mU3^4+300 m3^2 mQ3^8 msq^2 mU3^4-140 m3^10 mU3^6+583 m3^8 mQ3^2 mU3^6+1975 m3^6 mQ3^4 mU3^6-2359 m3^4 mQ3^6 mU3^6+505 m3^2 mQ3^8 mU3^6+12 mQ3^10 mU3^6-840 m3^6 mQ3^2 msq^2 mU3^6-720 m3^2 mQ3^6 msq^2 mU3^6+120 mQ3^8 msq^2 mU3^6+204 m3^8 mU3^8-1259 m3^6 mQ3^2 mU3^8+283 m3^4 mQ3^4 mU3^8+571 m3^2 mQ3^6 mU3^8-183 mQ3^8 mU3^8+420 m3^4 mQ3^2 msq^2 mU3^8+600 m3^2 mQ3^4 msq^2 mU3^8-60 mQ3^6 msq^2 mU3^8-68 m3^6 mU3^10+587 m3^4 mQ3^2 mU3^10-594 m3^2 mQ3^4 mU3^10+171 mQ3^6 mU3^10-180 m3^2 mQ3^2 msq^2 mU3^10-60 mQ3^4 msq^2 mU3^10))/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2))+(64 (22 m3^7 mQ3^2-20 m3^5 mQ3^4+3 m3^3 mQ3^6+30 m3^3 mQ3^4 msq^2+16 m3^7 mU3^2-112 m3^5 mQ3^2 mU3^2+100 m3^3 mQ3^4 mU3^2-6 m3 mQ3^6 mU3^2-60 m3 mQ3^4 msq^2 mU3^2-16 m3^5 mU3^4+87 m3^3 mQ3^2 mU3^4-74 m3 mQ3^4 mU3^4+30 m3 mQ3^2 msq^2 mU3^4) Xt)/(3 mQ3^2 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))-(8 (4 m3^10 mQ3^2-8 m3^8 mQ3^4+4 m3^6 mQ3^6-4 m3^10 mU3^2-695 m3^8 mQ3^2 mU3^2+886 m3^6 mQ3^4 mU3^2-301 m3^4 mQ3^6 mU3^2+18 m3^2 mQ3^8 mU3^2+360 m3^6 mQ3^2 msq^2 mU3^2-300 m3^4 mQ3^4 msq^2 mU3^2+180 m3^2 mQ3^6 msq^2 mU3^2+136 m3^8 mU3^4+326 m3^6 mQ3^2 mU3^4-40 m3^4 mQ3^4 mU3^4-212 m3^2 mQ3^6 mU3^4+6 mQ3^8 mU3^4-300 m3^4 mQ3^2 msq^2 mU3^4-240 m3^2 mQ3^4 msq^2 mU3^4+60 mQ3^6 msq^2 mU3^4-68 m3^6 mU3^6-5 m3^4 mQ3^2 mU3^6-308 m3^2 mQ3^4 mU3^6+237 mQ3^6 mU3^6+180 m3^2 mQ3^2 msq^2 mU3^6+60 mQ3^4 msq^2 mU3^6+18 m3^2 mQ3^2 mU3^8+6 mQ3^4 mU3^8) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(-((128 (14 m3^3-3 m3 mQ3^2-30 m3 msq^2-3 m3 mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)))-(128 m3 (mQ3^2+3 mU3^2) (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+(256 (7 m3^5 mQ3^2-8 m3^5 mU3^2+26 m3^3 mQ3^2 mU3^2+8 m3^3 mU3^4-17 m3 mQ3^2 mU3^4))/(3 mQ3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) Xt^3+(4 (-4 m3^12 mQ3^4+8 m3^10 mQ3^6-4 m3^8 mQ3^8+176 m3^12 mQ3^2 mU3^2+667 m3^10 mQ3^4 mU3^2-1222 m3^8 mQ3^6 mU3^2+493 m3^6 mQ3^8 mU3^2-18 m3^4 mQ3^10 mU3^2-360 m3^8 mQ3^4 msq^2 mU3^2+300 m3^6 mQ3^6 msq^2 mU3^2-180 m3^4 mQ3^8 msq^2 mU3^2+4 m3^12 mU3^4-4793 m3^10 mQ3^2 mU3^4+10101 m3^8 mQ3^4 mU3^4-6913 m3^6 mQ3^6 mU3^4+1361 m3^4 mQ3^8 mU3^4-24 m3^2 mQ3^10 mU3^4-360 m3^8 mQ3^2 msq^2 mU3^4+600 m3^6 mQ3^4 msq^2 mU3^4+480 m3^4 mQ3^6 msq^2 mU3^4-240 m3^2 mQ3^8 msq^2 mU3^4-140 m3^10 mU3^6+7283 m3^8 mQ3^2 mU3^6-17667 m3^6 mQ3^4 mU3^6+13435 m3^4 mQ3^6 mU3^6-2665 m3^2 mQ3^8 mU3^6-6 mQ3^10 mU3^6+1020 m3^6 mQ3^2 msq^2 mU3^6-1380 m3^4 mQ3^4 msq^2 mU3^6+420 m3^2 mQ3^6 msq^2 mU3^6-60 mQ3^8 msq^2 mU3^6+204 m3^8 mU3^8-1881 m3^6 mQ3^2 mU3^8+6073 m3^4 mQ3^4 mU3^8-5659 m3^2 mQ3^6 mU3^8+1167 mQ3^8 mU3^8-360 m3^4 mQ3^2 msq^2 mU3^8-360 m3^2 mQ3^4 msq^2 mU3^8+240 mQ3^6 msq^2 mU3^8-68 m3^6 mU3^10-975 m3^4 mQ3^2 mU3^10+1508 m3^2 mQ3^4 mU3^10-417 mQ3^6 mU3^10+180 m3^2 mQ3^2 msq^2 mU3^10+60 mQ3^4 msq^2 mU3^10-18 m3^2 mQ3^2 mU3^12-6 mQ3^4 mU3^12) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^3)-(64 (48 m3^7 mQ3^2-62 m3^5 mQ3^4+3 m3^3 mQ3^6+30 m3^3 mQ3^4 msq^2+16 m3^7 mU3^2+172 m3^5 mQ3^2 mU3^2-166 m3^3 mQ3^4 mU3^2+3 m3 mQ3^6 mU3^2-30 m3^3 mQ3^2 msq^2 mU3^2+30 m3 mQ3^4 msq^2 mU3^2-16 m3^5 mU3^4-185 m3^3 mQ3^2 mU3^4+184 m3 mQ3^4 mU3^4-30 m3 mQ3^2 msq^2 mU3^4+3 m3 mQ3^2 mU3^6) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+((80 (18 m3^10-35 m3^8 mQ3^2+17 m3^6 mQ3^4+12 m3^8 msq^2-3 m3^6 mQ3^2 msq^2+3 m3^4 mQ3^4 msq^2+9 m3^6 msq^4-18 m3^4 mQ3^2 msq^4+9 m3^2 mQ3^4 msq^4-47 m3^8 mU3^2+91 m3^6 mQ3^2 mU3^2-44 m3^4 mQ3^4 mU3^2-51 m3^6 msq^2 mU3^2+39 m3^4 mQ3^2 msq^2 mU3^2-24 m3^2 mQ3^4 msq^2 mU3^2+3 m3^4 msq^4 mU3^2-6 m3^2 mQ3^2 msq^4 mU3^2+3 mQ3^4 msq^4 mU3^2+43 m3^6 mU3^4-83 m3^4 mQ3^2 mU3^4+40 m3^2 mQ3^4 mU3^4+24 m3^4 msq^2 mU3^4+15 m3^2 mQ3^2 msq^2 mU3^4-3 mQ3^4 msq^2 mU3^4-14 m3^4 mU3^6+27 m3^2 mQ3^2 mU3^6-13 mQ3^4 mU3^6-9 m3^2 msq^2 mU3^6-3 mQ3^2 msq^2 mU3^6))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3)+(-((640 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2)))+(640 m3)/(3 (mQ3^2-mU3^2))-(320 m3 mU3^2)/(3 (m3^2-mU3^2) (-mQ3^2+mU3^2))+(320 (-6 m3 msq^4+m3^3 mU3^2+12 m3 msq^2 mU3^2-m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt-(160 (18 m3^8-35 m3^6 mQ3^2+17 m3^4 mQ3^4+18 m3^6 msq^2-15 m3^4 mQ3^2 msq^2+9 m3^2 mQ3^4 msq^2-33 m3^6 mU3^2+64 m3^4 mQ3^2 mU3^2-31 m3^2 mQ3^4 mU3^2-15 m3^4 msq^2 mU3^2-12 m3^2 mQ3^2 msq^2 mU3^2+3 mQ3^4 msq^2 mU3^2+15 m3^4 mU3^4-29 m3^2 mQ3^2 mU3^4+14 mQ3^4 mU3^4+9 m3^2 msq^2 mU3^4+3 mQ3^2 msq^2 mU3^4) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+((1280 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))-(640 m3 (mQ3^2+3 mU3^2))/(3 (mQ3^2-mU3^2)^3)) Xt^3+((160 (2 m3^4 msq^2-3 m3^2 msq^4+6 m3^2 msq^2 mU3^2-msq^4 mU3^2))/((m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)+(80 (mQ3^2+mU3^2) (11 m3^8-22 m3^6 mQ3^2+11 m3^4 mQ3^4+18 m3^6 msq^2-15 m3^4 mQ3^2 msq^2+9 m3^2 mQ3^4 msq^2-22 m3^6 mU3^2+44 m3^4 mQ3^2 mU3^2-22 m3^2 mQ3^4 mU3^2-15 m3^4 msq^2 mU3^2-12 m3^2 mQ3^2 msq^2 mU3^2+3 mQ3^4 msq^2 mU3^2+11 m3^4 mU3^4-22 m3^2 mQ3^2 mU3^4+11 mQ3^4 mU3^4+9 m3^2 msq^2 mU3^4+3 mQ3^2 msq^2 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(80 (4 m3^8-3 mQ3^2 mU3^4 (mQ3^2+5 mU3^2)+m3^6 (3 mQ3^2+7 mU3^2)-m3^4 (6 mQ3^4+15 mQ3^2 mU3^2+29 mU3^4)+m3^2 (7 mQ3^4 mU3^2+31 mQ3^2 mU3^4+16 mU3^6)))/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Xt^4+(-((640 m3 mU3^2)/(3 (m3^2-mU3^2) (-mQ3^2+mU3^2)^3))-(640 m3 msq^2 (mQ3^2+mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(1280 (-m3 msq^4+2 m3 msq^2 mU3^2))/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^3)) Xt^5) Log[msq^2/MR^2]+(-24 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3))-24 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt+1/(mQ3^2-mU3^2) 48 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)) Xt^2+(48 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^3)/(mQ3^2-mU3^2)-1/(mQ3^2-mU3^2)^3 24 (mQ3^2+mU3^2) (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)) Xt^4-(24 (mQ3^2+mU3^2) ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^5)/(mQ3^2-mU3^2)^3) Log[msq^2/MR^2]^2+(-((8 (128 m3^12-290 m3^10 mQ3^2+98 m3^8 mQ3^4+9 m3^6 mQ3^6+20 m3^4 mQ3^8-9 m3^2 mQ3^10-350 m3^10 mU3^2+762 m3^8 mQ3^2 mU3^2-251 m3^6 mQ3^4 mU3^2-151 m3^4 mQ3^6 mU3^2+41 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2+420 m3^8 mU3^4-821 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4+41 m3^2 mQ3^6 mU3^4-mQ3^8 mU3^4-217 m3^6 mU3^6+323 m3^4 mQ3^2 mU3^6-167 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6+31 m3^4 mU3^8-34 m3^2 mQ3^2 mU3^8+23 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))-(64 (20 m3^9-37 m3^7 mQ3^2+19 m3^5 mQ3^4-5 m3^3 mQ3^6+3 m3 mQ3^8-73 m3^7 mU3^2+135 m3^5 mQ3^2 mU3^2-23 m3^3 mQ3^4 mU3^2-7 m3 mQ3^6 mU3^2+24 m3^5 mU3^4-78 m3^3 mQ3^2 mU3^4+22 m3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(16 (128 m3^12-294 m3^10 mQ3^2+98 m3^8 mQ3^4+9 m3^6 mQ3^6+20 m3^4 mQ3^8-9 m3^2 mQ3^10-858 m3^10 mU3^2+1414 m3^8 mQ3^2 mU3^2-251 m3^6 mQ3^4 mU3^2-151 m3^4 mQ3^6 mU3^2+41 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2+1304 m3^8 mU3^4-2049 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4+41 m3^2 mQ3^6 mU3^4-mQ3^8 mU3^4-525 m3^6 mU3^6+903 m3^4 mQ3^2 mU3^6-167 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6-37 m3^4 mU3^8-34 m3^2 mQ3^2 mU3^8+23 mQ3^4 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(64 (2 m3^7-37 m3^5 mQ3^2-2 m3^3 mQ3^4-6 m3 mQ3^6-33 m3^5 mU3^2+110 m3^3 mQ3^2 mU3^2+20 m3 mQ3^4 mU3^2-6 m3^3 mU3^4-75 m3 mQ3^2 mU3^4+27 m3 mU3^6) Xt^3)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^3)-(8 (124 m3^12 mQ3^2-284 m3^10 mQ3^4+90 m3^8 mQ3^6+11 m3^6 mQ3^8+20 m3^4 mQ3^10-9 m3^2 mQ3^12+1156 m3^12 mU3^2-3068 m3^10 mQ3^2 mU3^2+2702 m3^8 mQ3^4 mU3^2-474 m3^6 mQ3^6 mU3^2-201 m3^4 mQ3^8 mU3^2+32 m3^2 mQ3^10 mU3^2-3 mQ3^12 mU3^2-3048 m3^10 mU3^4+7406 m3^8 mQ3^2 mU3^4-5516 m3^6 mQ3^4 mU3^4+882 m3^4 mQ3^6 mU3^4+184 m3^2 mQ3^8 mU3^4-4 mQ3^10 mU3^4+2602 m3^8 mU3^6-6126 m3^6 mQ3^2 mU3^6+3984 m3^4 mQ3^4 mU3^6-502 m3^2 mQ3^6 mU3^6-54 mQ3^8 mU3^6-695 m3^6 mU3^8+1718 m3^4 mQ3^2 mU3^8-883 m3^2 mQ3^4 mU3^8+4 mQ3^6 mU3^8-3 m3^4 mU3^10-102 m3^2 mQ3^2 mU3^10+57 mQ3^4 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^4)-(64 (-18 m3^5 mQ3^2-m3^3 mQ3^4-3 m3 mQ3^6+14 m3^5 mU3^2+20 m3^3 mQ3^2 mU3^2+4 m3 mQ3^4 mU3^2-11 m3^3 mU3^4+m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt^5)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) PolyLog[2,1-m3^2/mQ3^2]+(-12 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2)))-(640 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+1/(mQ3^2-mU3^2) 24 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2))) Xt^2+(1280 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^3)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-1/(mQ3^2-mU3^2)^3 12 (2 m3^2-2 msq^2) (mQ3^2+mU3^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2))) Xt^4-(640 (m3^2-msq^2) (mQ3^2+mU3^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/msq^2]+((40 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(-m3^2+mQ3^2)^3-(640 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))-(80 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2) Xt^2)/((-m3^2+mQ3^2)^3 (mQ3^2-mU3^2))+(1280 m3 (mQ3^2-msq^2)^2 Xt^3)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)+(40 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2) (mQ3^2+mU3^2) Xt^4)/((-m3^2+mQ3^2)^3 (mQ3^2-mU3^2)^3)-(640 m3 (mQ3^2-msq^2)^2 (mQ3^2+mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) PolyLog[2,1-msq^2/mQ3^2]+(-((32 m3^4 (m3^4-34 m3^2 mU3^2+17 mU3^4))/(3 (m3^2-mU3^2)^4))-(8 (-128 m3^12+346 m3^10 mQ3^2-280 m3^8 mQ3^4+13 m3^6 mQ3^6+37 m3^4 mQ3^8+294 m3^10 mU3^2-902 m3^8 mQ3^2 mU3^2+1025 m3^6 mQ3^4 mU3^2-391 m3^4 mQ3^6 mU3^2+34 m3^2 mQ3^8 mU3^2-98 m3^8 mU3^4+251 m3^6 mQ3^2 mU3^4-417 m3^4 mQ3^4 mU3^4+167 m3^2 mQ3^6 mU3^4-23 mQ3^8 mU3^4-9 m3^6 mU3^6+151 m3^4 mQ3^2 mU3^6-41 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6-20 m3^4 mU3^8-41 m3^2 mQ3^2 mU3^8+mQ3^4 mU3^8+9 m3^2 mU3^10+3 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(-((64 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)))+(128 (m3^7-50 m3^5 mU3^2+17 m3^3 mU3^4))/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2))) Xt-(16 (128 m3^12-346 m3^10 mQ3^2+280 m3^8 mQ3^4-13 m3^6 mQ3^6-37 m3^4 mQ3^8-806 m3^10 mU3^2+2438 m3^8 mQ3^2 mU3^2-2561 m3^6 mQ3^4 mU3^2+903 m3^4 mQ3^6 mU3^2-34 m3^2 mQ3^8 mU3^2+98 m3^8 mU3^4-251 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4-167 m3^2 mQ3^6 mU3^4+23 mQ3^8 mU3^4+9 m3^6 mU3^6-151 m3^4 mQ3^2 mU3^6+41 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6+20 m3^4 mU3^8+41 m3^2 mQ3^2 mU3^8-mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)+((64 m3 (-2 m3^2+mQ3^2+mU3^2) (m3^4-34 m3^2 mU3^2+17 mU3^4))/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(128 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)) Xt^3+(8 (124 m3^12 mQ3^2-332 m3^10 mQ3^4+262 m3^8 mQ3^6-3 m3^6 mQ3^8-39 m3^4 mQ3^10+1156 m3^12 mU3^2-4100 m3^10 mQ3^2 mU3^2+5326 m3^8 mQ3^4 mU3^2-2990 m3^6 mQ3^6 mU3^2+526 m3^4 mQ3^8 mU3^2+34 m3^2 mQ3^10 mU3^2-1968 m3^10 mU3^4+6466 m3^8 mQ3^2 mU3^4-7596 m3^6 mQ3^4 mU3^4+3712 m3^4 mQ3^6 mU3^4-543 m3^2 mQ3^8 mU3^4-11 mQ3^10 mU3^4+746 m3^8 mU3^6-2186 m3^6 mQ3^2 mU3^6+2210 m3^4 mQ3^4 mU3^6-774 m3^2 mQ3^6 mU3^6+4 mQ3^8 mU3^6-25 m3^6 mU3^8-29 m3^4 mQ3^2 mU3^8-20 m3^2 mQ3^4 mU3^8+14 mQ3^6 mU3^8+20 m3^4 mU3^10+32 m3^2 mQ3^2 mU3^10-4 mQ3^4 mU3^10-9 m3^2 mU3^12-3 mQ3^2 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)+((1024 m3 mU3^2 (2 m3^2-mQ3^2-mU3^2))/(3 (m3^2-mU3^2) (-mQ3^2+mU3^2)^4)-(64 (mQ3^2+mU3^2) (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) Xt^5) PolyLog[2,1-m3^2/mU3^2]+(-((4 (-24 m3^10 mQ3^2+28 m3^8 mQ3^4-68 m3^6 mQ3^6+58 m3^4 mQ3^8-18 m3^2 mQ3^10+24 m3^10 mU3^2+220 m3^6 mQ3^4 mU3^2-188 m3^4 mQ3^6 mU3^2+70 m3^2 mQ3^8 mU3^2-6 mQ3^10 mU3^2-28 m3^8 mU3^4-220 m3^6 mQ3^2 mU3^4+16 m3^2 mQ3^6 mU3^4-8 mQ3^8 mU3^4+68 m3^6 mU3^6+188 m3^4 mQ3^2 mU3^6-16 m3^2 mQ3^4 mU3^6-58 m3^4 mU3^8-70 m3^2 mQ3^2 mU3^8+8 mQ3^4 mU3^8+18 m3^2 mU3^10+6 mQ3^2 mU3^10))/(3 (-m3^2+mQ3^2)^3 (m3^2-mU3^2)^3))+(32 (10 m3^5 mQ3^2-10 m3^3 mQ3^4+6 m3 mQ3^6-10 m3^5 mU3^2-8 m3 mQ3^4 mU3^2+10 m3^3 mU3^4+8 m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(16 (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(128 (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(8 (mQ3^2+mU3^2) (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(64 (mQ3^2+mU3^2) (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) PolyLog[2,1-mQ3^2/mU3^2]+((40 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(-m3^2+mU3^2)^3-(640 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))-(160 (-msq^2+mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4) Xt^2)/((mQ3^2-mU3^2) (-m3^2+mU3^2)^3)-(1280 m3 (-msq^2+mU3^2)^2 Xt^3)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)+(80 (mQ3^2+mU3^2) (-msq^2+mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4) Xt^4)/((mQ3^2-mU3^2)^3 (-m3^2+mU3^2)^3)+(640 m3 (mQ3^2+mU3^2) (-msq^2+mU3^2)^2 Xt^5)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) PolyLog[2,1-msq^2/mU3^2])))
);

(* Using gauge eigenstate masses *)

Mh2[DR]=(
MZ^2 Cos[2 beta]^2+vu^2 yt^4 Sin[beta]^2 (k ((12 Xt^4)/(mQ3^2-mU3^2)^2+(6-(12 Xt^2)/(mQ3^2-mU3^2)+((6 mQ3^2)/(mQ3^2-mU3^2)^3+(6 mU3^2)/(mQ3^2-mU3^2)^3) Xt^4) Log[mU3^2/mQ3^2]-12 Log[(vu^2 yt^2)/mQ3^2])+g3^2 k^2 (-64+(32 m3^2)/mQ3^2+(32 m3^2)/mU3^2-(64 m3^2 Xt^2)/(mQ3^2 mU3^2)-(512 m3 Xt^3)/(mQ3^2-mU3^2)^2+(192/(mQ3^2-mU3^2)^2+(32 m3^2)/(mQ3^2 (mQ3^2-mU3^2)^2)+(32 m3^2)/(mU3^2 (mQ3^2-mU3^2)^2)) Xt^4+((32 (3 m3^4-2 m3^2 mU3^2+mU3^4))/(m3^2-mU3^2)^2-(64 (2 m3^3-m3 mU3^2) Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(64 (3 m3^2-2 mU3^2) Xt^2)/((m3^2-mU3^2) (-mQ3^2+mU3^2))-(128 (m3 mQ3^2+3 m3 mU3^2) Xt^3)/(mQ3^2-mU3^2)^3+(32 (2 m3^2+3 mQ3^2+7 mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) Log[mU3^2/mQ3^2]+(-((16 (3 m3^4-2 m3^2 mU3^2+mU3^4))/(m3^2-mU3^2)^2)+(128 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(32 (mQ3^2-3 mU3^2) Xt^2)/(mQ3^2-mU3^2)^2+(64 (-2 m3^3+m3 mQ3^2+m3 mU3^2) Xt^3)/(mQ3^2-mU3^2)^3-(32 (m3^2 mQ3^2-m3^2 mU3^2-2 mQ3^2 mU3^2-2 mU3^4) Xt^4)/(mQ3^2-mU3^2)^4) Log[mU3^2/mQ3^2]^2+Log[m3^2/mQ3^2] (-((32 (m3^10 mQ3^2-2 m3^8 mQ3^4+m3^6 mQ3^6+m3^10 mU3^2+2 m3^8 mQ3^2 mU3^2-2 m3^6 mQ3^4 mU3^2+m3^4 mQ3^6 mU3^2-2 m3^8 mU3^4-2 m3^6 mQ3^2 mU3^4+m3^6 mU3^6+m3^4 mQ3^2 mU3^6))/(mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2))+(64 m3^3 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))+(64 m3^4 (m3^2-mQ3^2-mU3^2) Xt^2)/(mQ3^2 (m3^2-mQ3^2) mU3^2 (m3^2-mU3^2))-(32 m3^2 (mQ3^2+mU3^2) Xt^4)/(mQ3^2 mU3^2 (mQ3^2-mU3^2)^2)+((32 m3^4)/(m3^2-mU3^2)^2-(128 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(256 m3^3 Xt^3)/(mQ3^2-mU3^2)^3) Log[mU3^2/mQ3^2])+Log[MR^2/mQ3^2] (-64+(32 m3^2)/mQ3^2+(32 m3^2)/mU3^2-(64 m3^2 Xt^2)/(mQ3^2 mU3^2)-(256 m3 Xt^3)/(mQ3^2-mU3^2)^2+(128/(mQ3^2-mU3^2)^2+(32 m3^2)/(mQ3^2 (mQ3^2-mU3^2)^2)+(32 m3^2)/(mU3^2 (mQ3^2-mU3^2)^2)) Xt^4+(64+(128 m3 Xt)/(mQ3^2-mU3^2)-(128 Xt^2)/(mQ3^2-mU3^2)-(128 (m3 mQ3^2+m3 mU3^2) Xt^3)/(mQ3^2-mU3^2)^3+(64 (m3^2+mQ3^2+mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) Log[mU3^2/mQ3^2]-128 Log[(vu^2 yt^2)/mQ3^2])+((32 m3^2 (2 m3^2-mQ3^2-mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2))+(-((32 m3^4 (2 m3^4-2 m3^2 mQ3^2+mQ3^4-2 m3^2 mU3^2+mU3^4))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(128 m3^3 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))) Log[m3^2/mQ3^2]+((64 m3^2 mU3^2)/(m3^2-mU3^2)^2-(32 mU3^4)/(m3^2-mU3^2)^2+(128 m3 mU3^2 Xt)/((-m3^2+mU3^2) (-mQ3^2+mU3^2))) Log[mU3^2/mQ3^2]) Log[(vu^2 yt^2)/mQ3^2]+96 Log[(vu^2 yt^2)/mQ3^2]^2+(-((64 m3^4)/(m3^2-mQ3^2)^2)+(256 m3^3 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))-(128 (-2 m3^3+m3 mQ3^2+m3 mU3^2) Xt^3)/(mQ3^2-mU3^2)^3+(32 (2 m3^2-mQ3^2-mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) PolyLog[2,1-m3^2/mQ3^2]+(-((64 m3^4)/(m3^2-mU3^2)^2)+(256 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))-(128 (2 m3^3-m3 mQ3^2-m3 mU3^2) Xt^3)/(mQ3^2-mU3^2)^3-(32 (2 m3^2-mQ3^2-mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) PolyLog[2,1-m3^2/mU3^2])+g3^4 k^3 (dlatas2+(-352 m3^10 mQ3^2+448 m3^8 mQ3^4-96 m3^6 mQ3^6-352 m3^10 mU3^2+2620 m3^8 mQ3^2 mU3^2-2376 m3^6 mQ3^4 mU3^2+500 m3^4 mQ3^6 mU3^2-72 m3^2 mQ3^8 mU3^2-1440 m3^6 mQ3^2 msq^2 mU3^2+1200 m3^4 mQ3^4 msq^2 mU3^2-720 m3^2 mQ3^6 msq^2 mU3^2+448 m3^8 mU3^4-2376 m3^6 mQ3^2 mU3^4+224 m3^4 mQ3^4 mU3^4+992 m3^2 mQ3^6 mU3^4-24 mQ3^8 mU3^4+1200 m3^4 mQ3^2 msq^2 mU3^4+960 m3^2 mQ3^4 msq^2 mU3^4-240 mQ3^6 msq^2 mU3^4-96 m3^6 mU3^6+500 m3^4 mQ3^2 mU3^6+992 m3^2 mQ3^4 mU3^6-884 mQ3^6 mU3^6-720 m3^2 mQ3^2 msq^2 mU3^6-240 mQ3^4 msq^2 mU3^6-72 m3^2 mQ3^2 mU3^8-24 mQ3^4 mU3^8)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)+SMConstatas2-(64 (14 m3^3-3 m3 (mQ3^2+10 msq^2+mU3^2)) Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))-(64 m3^2 (11 m3^4-5 mQ3^2 mU3^2-3 m3^2 (mQ3^2+mU3^2)) Xt^2)/(3 mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2))+(512 m3 (11 m3^4-5 mQ3^2 mU3^2-3 m3^2 (mQ3^2+mU3^2)) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+(8 (-44 m3^10 (mQ3^2+mU3^2)+7 m3^8 (8 mQ3^4-65 mQ3^2 mU3^2+8 mU3^4)-m3^4 mQ3^2 mU3^2 (173 mQ3^4+300 msq^2 mU3^2+173 mU3^4+12 mQ3^2 (25 msq^2+14 mU3^2))+mQ3^4 mU3^4 (6 mQ3^4+6 mU3^2 (10 msq^2+mU3^2)+mQ3^2 (60 msq^2+277 mU3^2))-6 m3^6 (2 mQ3^6-93 mQ3^4 mU3^2+2 mU3^6-3 mQ3^2 (20 msq^2 mU3^2+31 mU3^4))+2 m3^2 mQ3^2 mU3^2 (9 mQ3^6+2 mQ3^4 (45 msq^2-59 mU3^2)+9 mU3^4 (10 msq^2+mU3^2)-2 mQ3^2 (60 msq^2 mU3^2+59 mU3^4))) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(128 (14 m3^3-3 m3 (mQ3^2+10 msq^2+mU3^2)) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+(-80+(160 Xt^4)/(mQ3^2-mU3^2)^2) Log[msq^2/mQ3^2]^2+(1/3 (-((4 (mQ3^2-mU3^2) (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))-(4 (128 m3^14-4 m3^12 (63 mQ3^2+257 mU3^2)+4 m3^10 (5 mQ3^4+15 msq^2 mU3^2+332 mU3^4+mQ3^2 (-15 msq^2+783 mU3^2))+m3^8 (115 mQ3^6-90 msq^4 mU3^2+300 msq^2 mU3^4-313 mU3^6+3 mQ3^4 (40 msq^2-949 mU3^2)+15 mQ3^2 (6 msq^4-28 msq^2 mU3^2-309 mU3^4))-mQ3^2 mU3^4 (mQ3^2-mU3^2) (3 mQ3^6-2 mQ3^4 mU3^2+6 mU3^6+3 mQ3^2 (10 msq^4-20 msq^2 mU3^2+99 mU3^4))-4 m3^6 (5 mQ3^8-15 msq^4 mU3^4+75 msq^2 mU3^6+140 mU3^8+mQ3^6 (15 msq^2-191 mU3^2)+3 mQ3^4 (15 msq^4-55 msq^2 mU3^2-401 mU3^4)+mQ3^2 (-30 msq^4 mU3^2+75 msq^2 mU3^4-511 mU3^6))+2 m3^2 mU3^2 (-3 mQ3^10+27 mQ3^8 mU3^2+9 mU3^10+mQ3^6 (-30 msq^4+150 msq^2 mU3^2+496 mU3^4)+2 mQ3^4 (30 msq^4 mU3^2-105 msq^2 mU3^4+71 mU3^6)+mQ3^2 (-30 msq^4 mU3^4+60 msq^2 mU3^6-351 mU3^8))+m3^4 (9 mQ3^10-39 mQ3^8 mU3^2+2 mQ3^6 (45 msq^4-150 msq^2 mU3^2-718 mU3^4)+3 mU3^6 (10 msq^4-20 msq^2 mU3^2+97 mU3^4)+mQ3^4 (30 msq^4 mU3^2-300 msq^2 mU3^4-2968 mU3^6)+mQ3^2 (-150 msq^4 mU3^4+660 msq^2 mU3^6+815 mU3^8))))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)))+1/3 ((32 (mQ3^2-mU3^2) (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(32 (32 m3^9 (mQ3^2-mU3^2)-m3^7 (mQ3^4+380 mQ3^2 mU3^2-397 mU3^4)+m3^3 (3 mQ3^8-14 mQ3^6 mU3^2-30 msq^4 mU3^4+226 mQ3^2 mU3^6+120 msq^2 mU3^6+200 mU3^8+mQ3^4 (30 msq^4-120 msq^2 mU3^2-367 mU3^4))+m3^5 (-2 mQ3^6+311 mQ3^4 mU3^2+30 msq^4 mU3^2-120 msq^2 mU3^4-479 mU3^6+mQ3^2 (-30 msq^4+120 msq^2 mU3^2+122 mU3^4))+m3 mU3^2 (-3 mQ3^8+16 mQ3^6 mU3^2-6 mU3^8+mQ3^4 (-30 msq^4+120 msq^2 mU3^2+137 mU3^4)+10 mQ3^2 (3 msq^4 mU3^2-12 msq^2 mU3^4-16 mU3^6))))/((m3^2-mQ3^2) (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)) Xt-(16 (m3^10 (-24 mQ3^2+492 mU3^2)+m3^8 (65 mQ3^4-1462 mQ3^2 mU3^2-383 mU3^4)+m3^4 mU3^2 (-373 mQ3^6-30 msq^2 mU3^4+192 mU3^6-4 mQ3^4 (45 msq^2+403 mU3^2)+mQ3^2 (210 msq^2 mU3^2+137 mU3^4))+mQ3^2 mU3^4 (3 mQ3^6+3 mU3^6+3 mQ3^4 (10 msq^2-59 mU3^2)+mQ3^2 (-30 msq^2 mU3^2+151 mU3^4))+m3^6 (-41 mQ3^6+1346 mQ3^4 mU3^2-90 msq^2 mU3^4-226 mU3^6+3 mQ3^2 (30 msq^2 mU3^2+491 mU3^4))+m3^2 mU3^2 (9 mQ3^8+9 mU3^8+mQ3^6 (90 msq^2+471 mU3^2)+mQ3^4 (-150 msq^2 mU3^2+338 mU3^4)+mQ3^2 (60 msq^2 mU3^4-391 mU3^6))) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(128 (11 m3^9+m3^7 (7 mQ3^2-37 mU3^2)-2 m3^5 (7 mQ3^4-75 mQ3^2 mU3^2+24 mU3^4)+m3^3 mU3^2 (-117 mQ3^4+30 msq^2 mU3^2+66 mU3^4-mQ3^2 (30 msq^2+47 mU3^2))+m3 mU3^2 (3 mQ3^6-3 mU3^6+mQ3^4 (30 msq^2+82 mU3^2)-mQ3^2 (30 msq^2 mU3^2+53 mU3^4))) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+1/3 ((8 (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(8 (176 m3^14 (mQ3^2-mU3^2)+m3^12 (-322 mQ3^4+120 mQ3^2 mU3^2+2586 mU3^4)+3 m3^10 (49 mQ3^6-20 msq^2 mU3^4-1893 mU3^6-mQ3^4 (20 msq^2+67 mU3^2)+mQ3^2 (40 msq^2 mU3^2-2089 mU3^4))-mQ3^2 mU3^4 (3 mQ3^10+mQ3^8 mU3^2+3 mU3^10+10 mQ3^6 (3 msq^4+3 msq^2 mU3^2-19 mU3^4)-4 mQ3^4 (15 msq^4 mU3^2+61 mU3^6)+mQ3^2 (30 msq^4 mU3^4-30 msq^2 mU3^6+347 mU3^8))+2 m3^8 (5 mQ3^8+3 mQ3^6 (20 msq^2+79 mU3^2)+mQ3^4 (45 msq^4-135 msq^2 mU3^2+3052 mU3^4)+mQ3^2 (-90 msq^4 mU3^2+180 msq^2 mU3^4+6811 mU3^6)+5 (9 msq^4 mU3^4-21 msq^2 mU3^6+403 mU3^8))-2 m3^6 (10 mQ3^10+mQ3^8 (30 msq^2+89 mU3^2)+2 mU3^6 (15 msq^4-60 msq^2 mU3^2+92 mU3^4)+mQ3^6 (90 msq^4-90 msq^2 mU3^2+1601 mU3^4)+mQ3^4 (-150 msq^4 mU3^2+210 msq^2 mU3^4+5695 mU3^6)+mQ3^2 (30 msq^4 mU3^4-30 msq^2 mU3^6+4741 mU3^8))+m3^4 (9 mQ3^12-21 mQ3^10 mU3^2-2 mU3^8 (15 msq^4-15 msq^2 mU3^2+196 mU3^4)+mQ3^8 (90 msq^4-30 msq^2 mU3^2+692 mU3^4)+mQ3^2 mU3^6 (180 msq^4-480 msq^2 mU3^2+1079 mU3^4)+mQ3^6 (-60 msq^4 mU3^2+4230 mU3^6)+mQ3^4 (-180 msq^4 mU3^4+480 msq^2 mU3^6+7123 mU3^8))-m3^2 mU3^2 (6 mQ3^12-42 mQ3^10 mU3^2+9 mU3^12+2 mQ3^8 (30 msq^4-60 msq^2 mU3^2+361 mU3^4)+mQ3^6 (-180 msq^4 mU3^2+300 msq^2 mU3^4+1861 mU3^6)+5 mQ3^4 (36 msq^4 mU3^4-48 msq^2 mU3^6+203 mU3^8)+mQ3^2 (-60 msq^4 mU3^6+60 msq^2 mU3^8-787 mU3^10))))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^4)) Xt^4+1/3 (-((64 (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)))-(64 (m3^9 (34 mQ3^2+30 mU3^2)+m3^7 (-35 mQ3^4-6 mQ3^2 mU3^2+179 mU3^4)-m3^5 (2 mQ3^6+5 mQ3^4 mU3^2-30 msq^4 mU3^2+90 msq^2 mU3^4+437 mU3^6+6 mQ3^2 (5 msq^4-5 msq^2 mU3^2+51 mU3^4))-m3 mU3^2 (3 mQ3^8-7 mQ3^6 mU3^2+3 mU3^8+mQ3^4 (30 msq^4-30 msq^2 mU3^2+82 mU3^4)+mQ3^2 (-30 msq^4 mU3^2+90 msq^2 mU3^4+201 mU3^6))+m3^3 (3 mQ3^8-5 mQ3^6 mU3^2-30 msq^4 mU3^4+90 msq^2 mU3^6+215 mU3^8+2 mQ3^4 (15 msq^4-15 msq^2 mU3^2+53 mU3^4)+mQ3^2 (60 msq^2 mU3^4+511 mU3^6))))/((m3^2-mQ3^2) (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)) Xt^5-(2560 m3^2 mU3^4 Xt^6)/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)+(-(1/(3 (m3^2-mU3^2)^3))80 (3 m3^6+3 msq^4 mU3^2-2 mU3^6-m3^4 (6 msq^2+7 mU3^2)+3 m3^2 (3 msq^4-6 msq^2 mU3^2+2 mU3^4))-(320 m3 (2 m3^4+6 msq^4-3 m3^2 mU3^2-12 msq^2 mU3^2+mU3^4) Xt)/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(160 (3 msq^2 (mQ3^2-mU3^2) (2 m3^4-msq^2 mU3^2-3 m3^2 (msq^2-2 mU3^2))-(m3^2-mU3^2) (m3^4 (mQ3^2-3 mU3^2)+2 mU3^4 (mQ3^2-2 mU3^2)+m3^2 (-4 mQ3^2 mU3^2+8 mU3^4))) Xt^2)/(3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(320 (12 m3 msq^2 (msq^2-2 mU3^2) (mQ3^2-mU3^2)-m3 (m3^2-mU3^2) (2 m3^4+3 mQ3^2 mU3^2-mU3^4-m3^2 (mQ3^2+3 mU3^2))) Xt^3)/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(80 (-3 msq^2 (mQ3^2-mU3^2) (mQ3^2+mU3^2) (2 m3^4-msq^2 mU3^2-3 m3^2 (msq^2-2 mU3^2))-(m3^2-mU3^2) (8 m3^4 mQ3^2 mU3^2-mQ3^4 mU3^4+4 mQ3^2 mU3^6+5 mU3^8-2 m3^6 (mQ3^2-mU3^2)+2 m3^2 mU3^2 (mQ3^4-5 mQ3^2 mU3^2-4 mU3^4))) Xt^4)/(3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)-(320 m3 (mQ3^2+mU3^2) (6 msq^4-m3^2 mU3^2-12 msq^2 mU3^2+mU3^4) Xt^5)/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) Log[msq^2/mQ3^2]) Log[mU3^2/mQ3^2]^2+(1/3 ((4 (mQ3^2-mU3^2) (12 m3^10+3 mQ3^8 mU3^2+7 mQ3^6 mU3^4+7 mQ3^4 mU3^6+3 mQ3^2 mU3^8-14 m3^8 (mQ3^2+mU3^2)+m3^6 (34 mQ3^4-76 mQ3^2 mU3^2+34 mU3^4)+m3^4 (-29 mQ3^6+65 mQ3^4 mU3^2+65 mQ3^2 mU3^4-29 mU3^6)+m3^2 (9 mQ3^8-26 mQ3^6 mU3^2-34 mQ3^4 mU3^4-26 mQ3^2 mU3^6+9 mU3^8)))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(4 (-128 m3^10+m3^8 (-38 mQ3^2+550 mU3^2)+mU3^4 (mQ3^2-mU3^2) (3 mQ3^4+30 msq^4+4 mQ3^2 mU3^2+67 mU3^4)+2 m3^6 (mQ3^4-30 msq^2 mU3^2-247 mU3^4+6 mQ3^2 (5 msq^2-23 mU3^2))+2 m3^2 mU3^2 (3 mQ3^6-18 mQ3^4 mU3^2-30 msq^4 mU3^2+90 msq^2 mU3^4+68 mU3^6+3 mQ3^2 (10 msq^4-30 msq^2 mU3^2-39 mU3^4))+m3^4 (-9 mQ3^6+33 mQ3^4 mU3^2+90 msq^4 mU3^2-120 msq^2 mU3^4+59 mU3^6+mQ3^2 (-90 msq^4+120 msq^2 mU3^2+429 mU3^4))))/((m3^2-mU3^2)^4 (mQ3^2-mU3^2)))+1/3 (-((32 m3 (mQ3^2-mU3^2) (5 m3^4+3 mQ3^4-mQ3^2 mU3^2+3 mU3^4-5 m3^2 (mQ3^2+mU3^2)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))-(32 (-mQ3^2+mU3^2)^3 (-20 m3^7 (mQ3^2-mU3^2)+m3^5 (mQ3^4+185 mQ3^2 mU3^2-202 mU3^4)+m3^3 (3 mQ3^6-11 mQ3^4 mU3^2-30 msq^4 mU3^2+60 msq^2 mU3^4+189 mU3^6+mQ3^2 (30 msq^4-60 msq^2 mU3^2-149 mU3^4))+m3 mU3^2 (-3 mQ3^6+10 mQ3^4 mU3^2+30 msq^4 mU3^2-60 msq^2 mU3^4-55 mU3^6+mQ3^2 (-30 msq^4+60 msq^2 mU3^2+32 mU3^4))))/((m3^2-mU3^2)^3 (mQ3^2-mU3^2)^5)) Xt+1/3 (-((8 (12 m3^10+3 mQ3^8 mU3^2+7 mQ3^6 mU3^4+7 mQ3^4 mU3^6+3 mQ3^2 mU3^8-14 m3^8 (mQ3^2+mU3^2)+m3^6 (34 mQ3^4-76 mQ3^2 mU3^2+34 mU3^4)+m3^4 (-29 mQ3^6+65 mQ3^4 mU3^2+65 mQ3^2 mU3^4-29 mU3^6)+m3^2 (9 mQ3^8-26 mQ3^6 mU3^2-34 mQ3^4 mU3^4-26 mQ3^2 mU3^6+9 mU3^8)))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))+(8 (-128 m3^10-6 m3^8 (7 mQ3^2-177 mU3^2)+2 m3^6 (mQ3^4+10 mQ3^2 (3 msq^2-7 mU3^2)-5 mU3^2 (6 msq^2+167 mU3^2))+2 m3^2 mU3^2 (3 mQ3^6-18 mQ3^4 mU3^2-30 msq^4 mU3^2+90 msq^2 mU3^4+124 mU3^6+3 mQ3^2 (10 msq^4-30 msq^2 mU3^2-39 mU3^4))+mU3^4 (3 mQ3^6+mQ3^4 mU3^2+mQ3^2 (30 msq^4+63 mU3^4)-15 (2 msq^4 mU3^2+9 mU3^6))+m3^4 (-9 mQ3^6+33 mQ3^4 mU3^2+mQ3^2 (-90 msq^4+120 msq^2 mU3^2+361 mU3^4)+15 (6 msq^4 mU3^2-8 msq^2 mU3^4+41 mU3^6))))/((m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)) Xt^2+1/3 ((64 m3 (5 m3^4+3 mQ3^4-mQ3^2 mU3^2+3 mU3^4-5 m3^2 (mQ3^2+mU3^2)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(32 (2 m3^9-m3^7 (37 mQ3^2+35 mU3^2)+m3^5 (2 mQ3^4+205 mQ3^2 mU3^2-131 mU3^4)+m3^3 (6 mQ3^6-22 mQ3^4 mU3^2-60 msq^4 mU3^2+120 msq^2 mU3^4+353 mU3^6+mQ3^2 (60 msq^4-120 msq^2 mU3^2-281 mU3^4))+m3 mU3^2 (-6 mQ3^6+20 mQ3^4 mU3^2+60 msq^4 mU3^2-120 msq^2 mU3^4-157 mU3^6+mQ3^2 (-60 msq^4+120 msq^2 mU3^2+81 mU3^4))))/((m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)) Xt^3+1/3 ((4 (mQ3^2+mU3^2) (12 m3^10+3 mQ3^8 mU3^2+7 mQ3^6 mU3^4+7 mQ3^4 mU3^6+3 mQ3^2 mU3^8-14 m3^8 (mQ3^2+mU3^2)+m3^6 (34 mQ3^4-76 mQ3^2 mU3^2+34 mU3^4)+m3^4 (-29 mQ3^6+65 mQ3^4 mU3^2+65 mQ3^2 mU3^4-29 mU3^6)+m3^2 (9 mQ3^8-26 mQ3^6 mU3^2-34 mQ3^4 mU3^4-26 mQ3^2 mU3^6+9 mU3^8)))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(4 (4 m3^10 (31 mQ3^2+289 mU3^2)+m3^8 (44 mQ3^4-872 mQ3^2 mU3^2-4276 mU3^4)+m3^6 (-2 mQ3^6+2314 mQ3^2 mU3^4+60 msq^2 mU3^4+4726 mU3^6+mQ3^4 (-60 msq^2+66 mU3^2))-2 m3^2 mU3^2 (3 mQ3^8-15 mQ3^6 mU3^2-30 msq^4 mU3^4-319 mQ3^2 mU3^6+90 msq^2 mU3^6+302 mU3^8+mQ3^4 (30 msq^4-90 msq^2 mU3^2-67 mU3^4))+m3^4 (9 mQ3^8-24 mQ3^6 mU3^2-90 msq^4 mU3^4-2344 mQ3^2 mU3^6+120 msq^2 mU3^6-1163 mU3^8+6 mQ3^4 (15 msq^4-20 msq^2 mU3^2-37 mU3^4))-mU3^4 (mQ3^2+mU3^2) (3 mQ3^6+mQ3^4 mU3^2-30 msq^4 mU3^2-169 mU3^6+mQ3^2 (30 msq^4+29 mU3^4))))/((m3^2-mU3^2)^4 (mQ3^2-mU3^2)^4)) Xt^4+1/3 (-((32 m3 (mQ3^2+mU3^2) (5 m3^4+3 mQ3^4-mQ3^2 mU3^2+3 mU3^4-5 m3^2 (mQ3^2+mU3^2)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2))-(32 (2 m3^7 (9 mQ3^4-16 mQ3^2 mU3^2+7 mU3^4)-m3^5 (mQ3^6+70 mQ3^4 mU3^2-145 mQ3^2 mU3^4-86 mU3^6)+m3 mU3^2 (mQ3^2+mU3^2) (3 mQ3^6-10 mQ3^4 mU3^2-30 msq^4 mU3^2+60 msq^2 mU3^4+103 mU3^6+2 mQ3^2 (15 msq^4-30 msq^2 mU3^2-8 mU3^4))+m3^3 (-3 mQ3^8+8 mQ3^6 mU3^2+30 msq^4 mU3^4-200 mQ3^2 mU3^6-60 msq^2 mU3^6-219 mU3^8+mQ3^4 (-30 msq^4+60 msq^2 mU3^2+94 mU3^4))))/((m3^2-mU3^2)^3 (mQ3^2-mU3^2)^5)) Xt^5+(1280 m3^2 mU3^4 (mQ3^2+mU3^2) Xt^6)/(3 (-m3^2+mU3^2)^2 (-mQ3^2+mU3^2)^5)) Log[mU3^2/mQ3^2]^3+Log[m3^2/mQ3^2]^2 (1/3 (-((8 (36 mQ3^10 mU3^10+88 m3^18 (mQ3^2+mU3^2)-144 m3^2 mQ3^8 mU3^8 (mQ3^2+mU3^2)+m3^4 mQ3^6 mU3^6 (25 mQ3^4+452 mQ3^2 mU3^2+25 mU3^4)-2 m3^16 (144 mQ3^4+151 mQ3^2 mU3^2+144 mU3^4)+2 m3^6 mQ3^4 mU3^4 (67 mQ3^6+136 mQ3^4 mU3^2+136 mQ3^2 mU3^4+67 mU3^6)+m3^14 (368 mQ3^6+838 mQ3^4 mU3^2+838 mQ3^2 mU3^4+368 mU3^6)-m3^8 mQ3^2 mU3^2 (39 mQ3^8+908 mQ3^6 mU3^2+2048 mQ3^4 mU3^4+908 mQ3^2 mU3^6+39 mU3^8)-m3^12 (224 mQ3^8+797 mQ3^6 mU3^2+2636 mQ3^4 mU3^4+797 mQ3^2 mU3^6+224 mU3^8)+4 m3^10 (14 mQ3^10+70 mQ3^8 mU3^2+647 mQ3^6 mU3^4+647 mQ3^4 mU3^6+70 mQ3^2 mU3^8+14 mU3^10)))/(mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^4))-(120 (m3^2-msq^2) (mQ3^4-2 mQ3^2 mU3^2+mU3^4) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2))+1/3 (-((960 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(64 m3^3 (-81 mQ3^6 mU3^6+16 m3^10 (mQ3^2+mU3^2)+130 m3^2 mQ3^4 mU3^4 (mQ3^2+mU3^2)+m3^8 (-32 mQ3^4+31 mQ3^2 mU3^2-32 mU3^4)+2 m3^6 (8 mQ3^6+9 mQ3^4 mU3^2+9 mQ3^2 mU3^4+8 mU3^6)-m3^4 (25 mQ3^6 mU3^2+196 mQ3^4 mU3^4+25 mQ3^2 mU3^6)))/(mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3)) Xt-(128 (14 m3^6 mQ3^2 (m3^2-mQ3^2) mU3^2 (m3^2-mU3^2)-m3^4 (m3^2-mQ3^2-mU3^2) (3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2+4 m3^4 (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2)))) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3)-(2048 m3^7 (m3^2-mQ3^2-mU3^2) Xt^3)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)+1/3 ((16 (36 mQ3^10 mU3^10-44 m3^18 (mQ3^2+mU3^2)-156 m3^2 mQ3^8 mU3^8 (mQ3^2+mU3^2)+m3^4 mQ3^6 mU3^6 (49 mQ3^4+548 mQ3^2 mU3^2+49 mU3^4)+2 m3^16 (72 mQ3^4+mQ3^2 mU3^2+72 mU3^4)+2 m3^6 mQ3^4 mU3^4 (35 mQ3^6+44 mQ3^4 mU3^2+44 mQ3^2 mU3^4+35 mU3^6)-2 m3^14 (92 mQ3^6-147 mQ3^4 mU3^2-147 mQ3^2 mU3^4+92 mU3^6)+m3^8 mQ3^2 mU3^2 (25 mQ3^8-572 mQ3^6 mU3^2-1712 mQ3^4 mU3^4-572 mQ3^2 mU3^6+25 mU3^8)+m3^12 (112 mQ3^8-277 mQ3^6 mU3^2-1932 mQ3^4 mU3^4-277 mQ3^2 mU3^6+112 mU3^8)-4 m3^10 (7 mQ3^10+mQ3^8 mU3^2-503 mQ3^6 mU3^4-503 mQ3^4 mU3^6+mQ3^2 mU3^8+7 mU3^10)))/(mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)+(240 (m3^2-msq^2) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)) Xt^4+1/3 ((1920 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(128 m3^3 (87 mQ3^6 mU3^6+8 m3^10 (mQ3^2+mU3^2)-134 m3^2 mQ3^4 mU3^4 (mQ3^2+mU3^2)+m3^8 (-16 mQ3^4+7 mQ3^2 mU3^2-16 mU3^4)+47 m3^4 mQ3^2 mU3^2 (mQ3^4+4 mQ3^2 mU3^2+mU3^4)+m3^6 (8 mQ3^6-54 mQ3^4 mU3^2-54 mQ3^2 mU3^4+8 mU3^6)))/(mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)) Xt^5-(2560 m3^6 Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(1/3 ((8 (-198 m3^16+36 mQ3^8 mU3^8-144 m3^2 mQ3^6 mU3^6 (mQ3^2+mU3^2)+22 m3^14 (29 mQ3^2+37 mU3^2)+m3^4 mQ3^4 mU3^4 (25 mQ3^4+452 mQ3^2 mU3^2+mU3^4)-m3^12 (565 mQ3^4+2700 mQ3^2 mU3^2+877 mU3^4)+4 m3^10 (30 mQ3^6+651 mQ3^4 mU3^2+699 mQ3^2 mU3^4+86 mU3^6)+m3^8 (mQ3^8-812 mQ3^6 mU3^2-2240 mQ3^4 mU3^4-1004 mQ3^2 mU3^6-55 mU3^8)+2 m3^6 (55 mQ3^8 mU3^2+136 mQ3^6 mU3^4+184 mQ3^4 mU3^6+79 mQ3^2 mU3^8)))/((m3^2-mQ3^2)^4 (m3^2-mU3^2)^4)+(120 (m3^2-msq^2) (mQ3^6-3 mQ3^4 mU3^2+3 mQ3^2 mU3^4-mU3^6) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3))+1/3 ((960 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(64 m3^3 (44 m3^10+75 mQ3^6 mU3^4-87 mQ3^4 mU3^6-m3^8 (77 mQ3^2+79 mU3^2)+2 m3^6 (11 mQ3^4+36 mQ3^2 mU3^2+61 mU3^4)+m3^4 (19 mQ3^6+101 mQ3^4 mU3^2-209 mQ3^2 mU3^4-63 mU3^6)+m3^2 (-118 mQ3^6 mU3^2+36 mQ3^4 mU3^4+142 mQ3^2 mU3^6)))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))) Xt+1/3 (-((16 (-30 m3^16+36 mQ3^8 mU3^8-144 m3^2 mQ3^6 mU3^6 (mQ3^2+mU3^2)+m3^14 (238 mQ3^2+334 mU3^2)+m3^4 mQ3^4 mU3^4 (mQ3^4+452 mQ3^2 mU3^2+mU3^4)-m3^12 (325 mQ3^4+1484 mQ3^2 mU3^2+517 mU3^4)+4 m3^10 (42 mQ3^6+415 mQ3^4 mU3^2+487 mQ3^2 mU3^4+66 mU3^6)-m3^8 (55 mQ3^8+652 mQ3^6 mU3^2+1648 mQ3^4 mU3^4+844 mQ3^2 mU3^6+55 mU3^8)+2 m3^6 (79 mQ3^8 mU3^2+96 mQ3^6 mU3^4+144 mQ3^4 mU3^6+79 mQ3^2 mU3^8)))/((m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)))-(240 (m3^2-msq^2) (mQ3^4-2 mQ3^2 mU3^2+mU3^4) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)) Xt^2+1/3 (-((1920 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)))+(128 m3^3 (44 m3^12-87 mQ3^8 mU3^4+186 mQ3^6 mU3^6-87 mQ3^4 mU3^8-100 m3^10 (mQ3^2+mU3^2)+3 m3^8 (15 mQ3^4+94 mQ3^2 mU3^2+15 mU3^4)+m3^6 (66 mQ3^6-250 mQ3^4 mU3^2-250 mQ3^2 mU3^4+66 mU3^6)-m3^4 (63 mQ3^8+42 mQ3^6 mU3^2-422 mQ3^4 mU3^4+42 mQ3^2 mU3^6+63 mU3^8)+2 m3^2 (71 mQ3^8 mU3^2-89 mQ3^6 mU3^4-89 mQ3^4 mU3^6+71 mQ3^2 mU3^8)))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)) Xt^3+1/3 (-((8 (1024 m3^18-2786 m3^16 mQ3^2+1826 m3^14 mQ3^4+549 m3^12 mQ3^6-664 m3^10 mQ3^8+55 m3^8 mQ3^10-2786 m3^16 mU3^2+8004 m3^14 mQ3^2 mU3^2-5903 m3^12 mQ3^4 mU3^2-868 m3^10 mQ3^6 mU3^2+1699 m3^8 mQ3^8 mU3^2-158 m3^6 mQ3^10 mU3^2+1826 m3^14 mU3^4-5903 m3^12 mQ3^2 mU3^4+5224 m3^10 mQ3^4 mU3^4-292 m3^8 mQ3^6 mU3^4-846 m3^6 mQ3^8 mU3^4-m3^4 mQ3^10 mU3^4+549 m3^12 mU3^6-868 m3^10 mQ3^2 mU3^6-292 m3^8 mQ3^4 mU3^6+928 m3^6 mQ3^6 mU3^6-453 m3^4 mQ3^8 mU3^6+144 m3^2 mQ3^10 mU3^6-664 m3^10 mU3^8+1699 m3^8 mQ3^2 mU3^8-846 m3^6 mQ3^4 mU3^8-453 m3^4 mQ3^6 mU3^8+288 m3^2 mQ3^8 mU3^8-36 mQ3^10 mU3^8+55 m3^8 mU3^10-158 m3^6 mQ3^2 mU3^10-m3^4 mQ3^4 mU3^10+144 m3^2 mQ3^6 mU3^10-36 mQ3^8 mU3^10))/((m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^3))+(120 (m3^2-msq^2) (mQ3^2+mU3^2) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)) Xt^4+1/3 ((960 (m3^2-msq^2) (mQ3^2+mU3^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(64 m3^3 (39 m3^8 (mQ3^2+mU3^2)+87 mQ3^4 mU3^4 (mQ3^2+mU3^2)-94 m3^6 (mQ3^2+mU3^2)^2-142 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)^2+m3^4 (63 mQ3^6+283 mQ3^4 mU3^2+283 mQ3^2 mU3^4+63 mU3^6)))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)) Xt^5-(1280 m3^6 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[mU3^2/mQ3^2])+Log[MR^2/mQ3^2]^2 (3760/3-(896 m3^2)/(3 mQ3^2)-(896 m3^2)/(3 mU3^2)+(1792 m3^2 Xt^2)/(3 mQ3^2 mU3^2)+(7168 m3 Xt^3)/(3 (mQ3^2-mU3^2)^2)+(-(4000/(3 (mQ3^2-mU3^2)^2))-(896 m3^2)/(3 mQ3^2 (mQ3^2-mU3^2)^2)-(896 m3^2)/(3 mU3^2 (mQ3^2-mU3^2)^2)) Xt^4+(-(2000/3)-(3584 m3 Xt)/(3 (mQ3^2-mU3^2))+(4000 Xt^2)/(3 (mQ3^2-mU3^2))+(3584 (m3 mQ3^2+m3 mU3^2) Xt^3)/(3 (mQ3^2-mU3^2)^3)-(16 (112 m3^2+125 mQ3^2+125 mU3^2) Xt^4)/(3 (mQ3^2-mU3^2)^3)) Log[mU3^2/mQ3^2]-2624/3 Log[(vu^2 yt^2)/mQ3^2])+(-((64 (3 m3^4-5 mQ3^2 mU3^2+m3^2 (mQ3^2+mU3^2)))/((m3^2-mQ3^2) (m3^2-mU3^2)))+(64 (3+(4 m3^4 (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))-(1024 m3^3 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))) Log[m3^2/mQ3^2]+160 Log[msq^2/mQ3^2]+((16 (m3^4-34 m3^2 mU3^2+17 mU3^4))/(m3^2-mU3^2)^2+(1024 m3 mU3^2 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))) Log[mU3^2/mQ3^2]) Log[(vu^2 yt^2)/mQ3^2]^2-736 Log[(vu^2 yt^2)/mQ3^2]^3+((8 (128 m3^14-2 m3^12 (167 mQ3^2+217 mU3^2)+16 m3^10 (23 mQ3^4+67 mQ3^2 mU3^2+30 mU3^4)+mQ3^6 mU3^2 (3 mQ3^6+mQ3^4 mU3^2+19 mQ3^2 mU3^4-23 mU3^6)-m3^8 (89 mQ3^6+1145 mQ3^4 mU3^2+1177 mQ3^2 mU3^4+149 mU3^6)+m3^6 (11 mQ3^8+100 mQ3^6 mU3^2+1498 mQ3^4 mU3^4+324 mQ3^2 mU3^6-13 mU3^8)+m3^2 mQ3^4 (9 mQ3^8-44 mQ3^6 mU3^2-42 mQ3^4 mU3^4+148 mQ3^2 mU3^6+57 mU3^8)+m3^4 (-29 mQ3^10+192 mQ3^8 mU3^2-376 mQ3^6 mU3^4-598 mQ3^4 mU3^6+43 mQ3^2 mU3^8)))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(64 (m3 (m3^2-mQ3^2)^2 (18 m3^4+3 mQ3^4-7 mQ3^2 mU3^2+22 mU3^4+m3^2 (mQ3^2-37 mU3^2))+4 m3^3 (m3^2-mU3^2) (11 m3^4-5 mQ3^2 mU3^2-3 m3^2 (mQ3^2+mU3^2))) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(128 m3 (2 m3^2-mQ3^2-mU3^2) (11 m3^4-5 mQ3^2 mU3^2-3 m3^2 (mQ3^2+mU3^2)) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(16 (172 m3^12-12 m3^10 (31 mQ3^2+39 mU3^2)+2 m3^8 (69 mQ3^4+542 mQ3^2 mU3^2+199 mU3^4)+m3^6 (3 mQ3^6-313 mQ3^4 mU3^2-1151 mQ3^2 mU3^4-59 mU3^6)-mQ3^4 mU3^2 (3 mQ3^6+mQ3^4 mU3^2+29 mQ3^2 mU3^4-13 mU3^6)+m3^4 (20 mQ3^8-149 mQ3^6 mU3^2+411 mQ3^4 mU3^4+409 mQ3^2 mU3^6-31 mU3^8)+m3^2 (-9 mQ3^10+41 mQ3^8 mU3^2+55 mQ3^6 mU3^4-129 mQ3^4 mU3^6-30 mQ3^2 mU3^8)) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(128 m3 (18 m3^4+3 mQ3^4-7 mQ3^2 mU3^2+22 mU3^4+m3^2 (mQ3^2-37 mU3^2)) Xt^5)/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mQ3^2]+(-((80 (m3^2-msq^2) (mQ3^4-2 mQ3^2 mU3^2+mU3^4) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2))-(640 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(160 (m3^2-msq^2) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))) Xt^4)/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(1280 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) PolyLog[2,1-msq^2/m3^2]+(-((80 (mQ3^2-msq^2) (2 m3^4+3 m3^2 mQ3^2-mQ3^4-3 m3^2 msq^2-mQ3^2 msq^2))/(m3^2-mQ3^2)^3)+(640 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(160 (mQ3^2-msq^2) (2 m3^4+3 m3^2 mQ3^2-mQ3^4-3 m3^2 msq^2-mQ3^2 msq^2) Xt^4)/((m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^2)-(1280 m3 (mQ3^2-msq^2)^2 Xt^5)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-msq^2/mQ3^2]+(-((8 (128 m3^14-2 m3^12 (217 mQ3^2+167 mU3^2)+16 m3^10 (30 mQ3^4+67 mQ3^2 mU3^2+23 mU3^4)+mQ3^2 mU3^6 (-23 mQ3^6+19 mQ3^4 mU3^2+mQ3^2 mU3^4+3 mU3^6)-m3^8 (149 mQ3^6+1177 mQ3^4 mU3^2+1145 mQ3^2 mU3^4+89 mU3^6)+m3^4 mU3^2 (43 mQ3^8-598 mQ3^6 mU3^2-376 mQ3^4 mU3^4+192 mQ3^2 mU3^6-29 mU3^8)+m3^2 mU3^4 (57 mQ3^8+148 mQ3^6 mU3^2-42 mQ3^4 mU3^4-44 mQ3^2 mU3^6+9 mU3^8)+m3^6 (-13 mQ3^8+324 mQ3^6 mU3^2+1498 mQ3^4 mU3^4+100 mQ3^2 mU3^6+11 mU3^8)))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(64 (m3 (m3^2-mU3^2)^2 (18 m3^4+22 mQ3^4-7 mQ3^2 mU3^2+3 mU3^4+m3^2 (-37 mQ3^2+mU3^2))+4 m3^3 (m3^2-mQ3^2) (11 m3^4-5 mQ3^2 mU3^2-3 m3^2 (mQ3^2+mU3^2))) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(128 m3 (2 m3^2-mQ3^2-mU3^2) (11 m3^4-5 mQ3^2 mU3^2-3 m3^2 (mQ3^2+mU3^2)) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+(16 (172 m3^12-12 m3^10 (39 mQ3^2+31 mU3^2)+2 m3^8 (199 mQ3^4+542 mQ3^2 mU3^2+69 mU3^4)+mQ3^2 mU3^4 (13 mQ3^6-29 mQ3^4 mU3^2-mQ3^2 mU3^4-3 mU3^6)-m3^6 (59 mQ3^6+1151 mQ3^4 mU3^2+313 mQ3^2 mU3^4-3 mU3^6)+m3^4 (-31 mQ3^8+409 mQ3^6 mU3^2+411 mQ3^4 mU3^4-149 mQ3^2 mU3^6+20 mU3^8)+m3^2 (-30 mQ3^8 mU3^2-129 mQ3^6 mU3^4+55 mQ3^4 mU3^6+41 mQ3^2 mU3^8-9 mU3^10)) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(128 m3 (18 m3^4+22 mQ3^4-7 mQ3^2 mU3^2+3 mU3^4+m3^2 (-37 mQ3^2+mU3^2)) Xt^5)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mU3^2]+Log[msq^2/mQ3^2] ((80 (2 m3^10 (mQ3^2+mU3^2)-4 m3^8 (mQ3^4+6 mQ3^2 mU3^2+mU3^4)+m3^4 mQ3^2 mU3^2 (-19 mQ3^4+15 msq^2 mU3^2-19 mU3^4+mQ3^2 (15 msq^2-68 mU3^2))-mQ3^4 mU3^4 (3 msq^2 mU3^2+mQ3^2 (3 msq^2+14 mU3^2))+m3^6 (2 mQ3^6+41 mQ3^4 mU3^2+2 mU3^6+mQ3^2 (-18 msq^2 mU3^2+41 mU3^4))+m3^2 (-9 mQ3^2 msq^2 mU3^6+mQ3^6 (-9 msq^2 mU3^2+31 mU3^4)+mQ3^4 (12 msq^2 mU3^4+31 mU3^6))))/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)+(640 m3 msq^2 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))-(320 m3^2 Xt^2)/(3 mQ3^2 mU3^2)-(2560 m3 Xt^3)/(3 (mQ3^2-mU3^2)^2)+(160 (3 mQ3^4 msq^2 mU3^6+m3^10 (mQ3^2+mU3^2)-2 m3^8 (mQ3^4-7 mQ3^2 mU3^2+mU3^4)+mQ3^6 (3 msq^2 mU3^4+16 mU3^6)+m3^4 mQ3^2 mU3^2 (15 mQ3^4+15 mU3^2 (-msq^2+mU3^2)+mQ3^2 (-15 msq^2+64 mU3^2))+m3^6 (mQ3^6-30 mQ3^4 mU3^2+mU3^6+6 mQ3^2 (3 msq^2 mU3^2-5 mU3^4))+m3^2 (9 mQ3^2 msq^2 mU3^6+mQ3^6 (9 msq^2 mU3^2-32 mU3^4)-4 mQ3^4 (3 msq^2 mU3^4+8 mU3^6))) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(1280 m3 msq^2 Xt^5)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+(-((320 m3^4)/(3 (m3^2-mQ3^2)^2))+(1280 m3^3 Xt)/(3 (m3^2-mQ3^2) (mQ3^2-mU3^2))+(640 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(3 (mQ3^2-mU3^2)^3)+(160 (2 m3^2-mQ3^2-mU3^2) Xt^4)/(3 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mQ3^2]+(-((320 m3^4)/(3 (m3^2-mU3^2)^2))-(1280 m3^3 Xt)/(3 (m3^2-mU3^2) (mQ3^2-mU3^2))-(640 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(3 (mQ3^2-mU3^2)^3)-(160 (2 m3^2-mQ3^2-mU3^2) Xt^4)/(3 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mU3^2])+Log[m3^2/mQ3^2] ((8 (-96 mQ3^8 mU3^8+132 m3^14 (mQ3^2+mU3^2)-4 m3^12 (75 mQ3^4+334 mQ3^2 mU3^2+75 mU3^4)-3 m3^2 mQ3^4 mU3^4 (mQ3^6-171 mQ3^2 mU3^4+10 msq^2 mU3^4+mU3^6+mQ3^4 (10 msq^2-171 mU3^2))+m3^6 mQ3^2 mU3^2 (519 mQ3^6+270 msq^2 mU3^4+519 mU3^6+5 mQ3^4 (54 msq^2+757 mU3^2)-5 mQ3^2 (36 msq^2 mU3^2-757 mU3^4))+2 m3^10 (118 mQ3^6+1369 mQ3^4 mU3^2+118 mU3^6+mQ3^2 (90 msq^2 mU3^2+1369 mU3^4))-4 m3^8 (17 mQ3^8+504 mQ3^6 mU3^2+17 mU3^8+2 mQ3^4 (30 msq^2 mU3^2+679 mU3^4)+12 mQ3^2 (5 msq^2 mU3^4+42 mU3^6))-m3^4 mQ3^2 mU3^2 (9 mQ3^8+9 mU3^6 (10 msq^2+mU3^2)+10 mQ3^6 (9 msq^2+86 mU3^2)+mQ3^4 (-90 msq^2 mU3^2+2470 mU3^4)+mQ3^2 (-90 msq^2 mU3^4+860 mU3^6))))/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3)-(64 (16 m3^9 (mQ3^2+mU3^2)+4 m3^5 mQ3^2 mU3^2 (28 mQ3^2+15 msq^2+28 mU3^2)-16 m3^7 (mQ3^4+8 mQ3^2 mU3^2+mU3^4)-3 m3^3 mQ3^2 mU3^2 (mQ3^4+10 msq^2 mU3^2+mU3^4+10 mQ3^2 (msq^2+3 mU3^2))) Xt)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)-(64 m3^2 (33 m3^8+6 mQ3^4 mU3^4-42 m3^6 (mQ3^2+mU3^2)-7 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)+m3^4 (17 mQ3^4+25 mQ3^2 mU3^2+17 mU3^4)) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)+(1024 (2 m3^5 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2-m3 mQ3^2 mU3^2 (3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2+4 m3^4 (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2)))) Xt^3)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(16 (120 mQ3^8 mU3^8+66 m3^14 (mQ3^2+mU3^2)-2 m3^12 (75 mQ3^4-938 mQ3^2 mU3^2+75 mU3^4)+m3^2 mQ3^4 mU3^4 (3 mQ3^6-559 mQ3^2 mU3^4+mQ3^4 (30 msq^2-559 mU3^2)+3 mU3^4 (10 msq^2+mU3^2))+2 m3^10 (59 mQ3^6-2095 mQ3^4 mU3^2+59 mU3^6-5 mQ3^2 (18 msq^2 mU3^2+419 mU3^4))-5 m3^6 mQ3^2 mU3^2 (93 mQ3^6+54 msq^2 mU3^4+93 mU3^6+mQ3^4 (54 msq^2+1129 mU3^2)+mQ3^2 (-36 msq^2 mU3^2+1129 mU3^4))+m3^8 (-34 mQ3^8+2756 mQ3^6 mU3^2-34 mU3^8+12 mQ3^4 (20 msq^2 mU3^2+747 mU3^4)+4 mQ3^2 (60 msq^2 mU3^4+689 mU3^6))+m3^4 mQ3^2 mU3^2 (9 mQ3^8+9 mU3^6 (10 msq^2+mU3^2)+10 mQ3^6 (9 msq^2+85 mU3^2)+mQ3^4 (-90 msq^2 mU3^2+3522 mU3^4)+mQ3^2 (-90 msq^2 mU3^4+850 mU3^6))) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(128 (8 m3^9 (mQ3^2+mU3^2)-5 m3^5 mQ3^2 mU3^2 (23 mQ3^2+12 msq^2+23 mU3^2)+m3^7 (-8 mQ3^4+123 mQ3^2 mU3^2-8 mU3^4)+m3^3 mQ3^2 mU3^2 (3 mQ3^4+3 mU3^2 (10 msq^2+mU3^2)+mQ3^2 (30 msq^2+101 mU3^2))) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(1/3 (-((160 m3^4 (m3^10 (mQ3^2+mU3^2)+m3^8 (-3 mQ3^4+2 mQ3^2 mU3^2-3 mU3^4)+8 mQ3^4 mU3^4 (mQ3^4+mU3^4)-m3^2 mQ3^2 mU3^2 (mQ3^6+24 mQ3^4 mU3^2+24 mQ3^2 mU3^4+mU3^6)+m3^6 (3 mQ3^6-11 mQ3^4 mU3^2-11 mQ3^2 mU3^4+3 mU3^6)-m3^4 (mQ3^8-3 mQ3^6 mU3^2-48 mQ3^4 mU3^4-3 mQ3^2 mU3^6+mU3^8)))/(mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3))+(240 (m3^2-msq^2) (mQ3^4-2 mQ3^2 mU3^2+mU3^4) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2))-(640 (2 m3^5 msq^2-2 m3^3 msq^4+m3 mQ3^2 msq^4-2 m3 mQ3^2 msq^2 mU3^2+m3 msq^4 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(320 m3^4 (m3^2-mQ3^2-mU3^2) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2) mU3^2 (m3^2-mU3^2))+1/3 (-((160 (m3^14 mQ3^2-3 m3^12 mQ3^4+3 m3^10 mQ3^6-m3^8 mQ3^8+m3^14 mU3^2-10 m3^12 mQ3^2 mU3^2+32 m3^10 mQ3^4 mU3^2-16 m3^8 mQ3^6 mU3^2+5 m3^6 mQ3^8 mU3^2-3 m3^12 mU3^4+32 m3^10 mQ3^2 mU3^4-102 m3^8 mQ3^4 mU3^4+54 m3^6 mQ3^6 mU3^4-17 m3^4 mQ3^8 mU3^4+3 m3^10 mU3^6-16 m3^8 mQ3^2 mU3^6+54 m3^6 mQ3^4 mU3^6-6 m3^4 mQ3^6 mU3^6+m3^2 mQ3^8 mU3^6-m3^8 mU3^8+5 m3^6 mQ3^2 mU3^8-17 m3^4 mQ3^4 mU3^8+m3^2 mQ3^6 mU3^8))/(mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2))-(480 (m3^2-msq^2) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)) Xt^4+(640 (m3^7-m3^5 mQ3^2+12 m3^5 msq^2-12 m3^3 msq^4+6 m3 mQ3^2 msq^4-m3^5 mU3^2+m3^3 mQ3^2 mU3^2-12 m3 mQ3^2 msq^2 mU3^2+6 m3 msq^4 mU3^2) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) Log[msq^2/mQ3^2]+((8 (-m3^2+mQ3^2) (84 mQ3^8 mU3^10 (mQ3^2-mU3^2)+2 m3^16 (mQ3^4-64 mQ3^2 mU3^2-mU3^4)+m3^14 (-6 mQ3^6-760 mQ3^4 mU3^2+1462 mQ3^2 mU3^4+72 mU3^6)+3 m3^2 mQ3^4 mU3^6 (mQ3^2-mU3^2) (mQ3^6-151 mQ3^2 mU3^4+10 msq^2 mU3^4+mU3^6+mQ3^4 (10 msq^2-179 mU3^2))+2 m3^12 (3 mQ3^8+1094 mQ3^6 mU3^2-86 mU3^8+5 mQ3^4 (18 msq^2 mU3^2-61 mU3^4)-2 mQ3^2 (45 msq^2 mU3^4+833 mU3^6))+m3^4 mQ3^2 mU3^4 (6 mQ3^10-9 mU3^8 (10 msq^2+mU3^2)+mQ3^8 (60 msq^2+1019 mU3^2)-2 mQ3^6 (75 msq^2 mU3^2-761 mU3^4)-2 mQ3^4 (15 msq^2 mU3^4+1013 mU3^6)+10 mQ3^2 (21 msq^2 mU3^6-64 mU3^8))+m3^6 mQ3^2 mU3^2 (-9 mQ3^10-18 mQ3^8 (5 msq^2+54 mU3^2)+6 mU3^8 (60 msq^2+71 mU3^2)-2 mQ3^6 (45 msq^2 mU3^2+1849 mU3^4)+2 mQ3^4 (225 msq^2 mU3^4+841 mU3^6)-63 mQ3^2 (10 msq^2 mU3^6-53 mU3^8))+m3^8 mU3^2 (547 mQ3^10-34 mU3^10+18 mQ3^8 (15 msq^2+196 mU3^2)-6 mQ3^6 (35 msq^2 mU3^2-267 mU3^4)+mQ3^4 (450 msq^2 mU3^4-5506 mU3^6)-17 mQ3^2 (30 msq^2 mU3^6+121 mU3^8))-2 m3^10 (mQ3^10+951 mQ3^8 mU3^2-68 mU3^10+20 mQ3^6 (6 msq^2 mU3^2+77 mU3^4)+mQ3^4 (90 msq^2 mU3^4-1817 mU3^6)-3 mQ3^2 (70 msq^2 mU3^6+629 mU3^8))))/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2))-(64 (-m3^2+mQ3^2) (-12 m3 mQ3^6 mU3^6+4 m3^11 (mQ3^2+4 mU3^2)-2 m3^9 (65 mQ3^4-108 mQ3^2 mU3^2+16 mU3^4)+m3^5 mQ3^2 (-3 mQ3^6+90 msq^2 mU3^4+86 mU3^6-2 mQ3^4 (15 msq^2+34 mU3^2)+mQ3^2 (-60 msq^2 mU3^2+111 mU3^4))+m3^7 (123 mQ3^6+16 mU3^6+2 mQ3^4 (30 msq^2-53 mU3^2)-mQ3^2 (60 msq^2 mU3^2+223 mU3^4))+m3^3 mQ3^2 mU3^2 (3 mQ3^6-35 mQ3^2 mU3^4+mQ3^4 (30 msq^2+37 mU3^2)-3 (10 msq^2 mU3^4+mU3^6))) Xt)/(3 mQ3^2 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(16 (-m3^2+mQ3^2) (96 mQ3^8 mU3^8-2 m3^14 (mQ3^2-mU3^2)+2 m3^12 (3 mQ3^4+736 mQ3^2 mU3^2-35 mU3^4)+3 m3^2 mQ3^4 mU3^4 (mQ3^6-163 mQ3^2 mU3^4+10 msq^2 mU3^4+mU3^6+mQ3^4 (10 msq^2-171 mU3^2))-m3^6 mQ3^2 mU3^2 (587 mQ3^6+5 mU3^4 (54 msq^2+79 mU3^2)+9 mQ3^4 (30 msq^2+443 mU3^2)-9 mQ3^2 (20 msq^2 mU3^2-391 mU3^4))-2 m3^10 (3 mQ3^6+1640 mQ3^4 mU3^2-51 mU3^6+10 mQ3^2 (9 msq^2 mU3^2+133 mU3^4))+2 m3^8 (mQ3^8+1218 mQ3^6 mU3^2-17 mU3^8+8 mQ3^4 (15 msq^2 mU3^2+356 mU3^4)+6 mQ3^2 (20 msq^2 mU3^4+141 mU3^6))+m3^4 mQ3^2 mU3^2 (9 mQ3^8-90 mQ3^4 mU3^2 (msq^2-27 mU3^2)+9 mU3^6 (10 msq^2+mU3^2)+10 mQ3^6 (9 msq^2+85 mU3^2)+mQ3^2 (-90 msq^2 mU3^4+734 mU3^6))) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(128 (-m3^2+mQ3^2) (22 m3^11 mQ3^2+6 m3 mQ3^6 mU3^4 (mQ3^2+3 mU3^2)+2 m3^9 (5 mQ3^4+3 mQ3^2 mU3^2+8 mU3^4)-m3^7 (193 mQ3^6-238 mQ3^4 mU3^2+217 mQ3^2 mU3^4+16 mU3^6)+m3^5 mQ3^2 (153 mQ3^6+60 msq^2 mU3^4+165 mU3^6+mQ3^4 (60 msq^2-85 mU3^2)-mQ3^2 (120 msq^2 mU3^2+17 mU3^4))-m3^3 mQ3^2 (3 mQ3^8+3 mU3^6 (10 msq^2+mU3^2)+mQ3^6 (30 msq^2+91 mU3^2)-2 mQ3^4 (15 msq^2 mU3^2+53 mU3^4)-5 mQ3^2 (6 msq^2 mU3^4-23 mU3^6))) Xt^3)/(3 mQ3^2 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(8 (-m3^2+mQ3^2) (48 mQ3^8 mU3^10 (2 mQ3^2+5 mU3^2)+2 m3^16 (mQ3^4+40 mQ3^2 mU3^2-mU3^4)-2 m3^14 (3 mQ3^6+1076 mQ3^4 mU3^2+1589 mQ3^2 mU3^4-36 mU3^6)+m3^2 mQ3^4 mU3^6 (3 mQ3^8+mQ3^6 (30 msq^2-616 mU3^2)+3 mU3^6 (10 msq^2+mU3^2)+6 mQ3^4 (5 msq^2 mU3^2-267 mU3^4)+mQ3^2 (30 msq^2 mU3^4-884 mU3^6))+2 m3^12 (3 mQ3^8+2203 mQ3^6 mU3^2-86 mU3^8+mQ3^4 (90 msq^2 mU3^2+6481 mU3^4)+5 mQ3^2 (18 msq^2 mU3^4+911 mU3^6))-2 m3^10 (mQ3^10+1498 mQ3^8 mU3^2-68 mU3^10+2 mQ3^6 (60 msq^2 mU3^2+4397 mU3^4)+6 mQ3^4 (55 msq^2 mU3^4+2132 mU3^6)+7 mQ3^2 (30 msq^2 mU3^6+709 mU3^8))+m3^4 mQ3^2 mU3^4 (6 mQ3^10+9 mU3^8 (10 msq^2+mU3^2)+mQ3^8 (60 msq^2+1723 mU3^2)+mQ3^6 (-30 msq^2 mU3^2+5474 mU3^4)-30 mQ3^4 (7 msq^2 mU3^4-220 mU3^6)+mQ3^2 (-30 msq^2 mU3^6+1268 mU3^8))+m3^8 mU3^2 (643 mQ3^10-34 mU3^10+mQ3^8 (270 msq^2+9676 mU3^2)+6 mQ3^6 (55 msq^2 mU3^2+4273 mU3^4)+190 mQ3^4 (3 msq^2 mU3^4+121 mU3^6)+mQ3^2 (510 msq^2 mU3^6+4623 mU3^8))-m3^6 mQ3^2 mU3^2 (9 mQ3^10+18 mU3^8 (20 msq^2+39 mU3^2)+2 mQ3^8 (45 msq^2+914 mU3^2)+2 mQ3^6 (135 msq^2 mU3^2+5431 mU3^4)+mQ3^4 (-90 msq^2 mU3^4+18070 mU3^6)+mQ3^2 (90 msq^2 mU3^6+9553 mU3^8))) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^3)-(64 (-m3^2+mQ3^2) (12 m3 mQ3^6 mU3^6+4 m3^11 (13 mQ3^2+4 mU3^2)-2 m3^9 (138 mQ3^4+181 mQ3^2 mU3^2+16 mU3^4)+2 m3^7 (107 mQ3^6+8 mU3^6+5 mQ3^4 (6 msq^2+77 mU3^2)+3 mQ3^2 (10 msq^2 mU3^2+93 mU3^4))+m3^3 mQ3^2 mU3^2 (3 mQ3^6+3 mU3^4 (10 msq^2+mU3^2)+3 mQ3^4 (10 msq^2+71 mU3^2)+mQ3^2 (60 msq^2 mU3^2+179 mU3^4))-m3^5 mQ3^2 (3 mQ3^6+90 msq^2 mU3^4+219 mU3^6+mQ3^4 (30 msq^2+407 mU3^2)+mQ3^2 (120 msq^2 mU3^2+737 mU3^4))) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(5120 m3^4 mU3^2 Xt^6)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(-((80 (2 m3^10 mQ3^2-3 m3^8 mQ3^4+m3^6 mQ3^6+12 m3^10 msq^2+18 m3^6 mQ3^4 msq^2-6 m3^4 mQ3^6 msq^2-18 m3^8 msq^4+24 m3^6 mQ3^2 msq^4-27 m3^4 mQ3^4 msq^4+9 m3^2 mQ3^6 msq^4-2 m3^10 mU3^2+3 m3^6 mQ3^4 mU3^2-m3^4 mQ3^6 mU3^2-108 m3^6 mQ3^2 msq^2 mU3^2+54 m3^4 mQ3^4 msq^2 mU3^2-18 m3^2 mQ3^6 msq^2 mU3^2+24 m3^6 msq^4 mU3^2+18 m3^4 mQ3^2 msq^4 mU3^2-9 m3^2 mQ3^4 msq^4 mU3^2+3 mQ3^6 msq^4 mU3^2+3 m3^8 mU3^4-3 m3^6 mQ3^2 mU3^4+18 m3^6 msq^2 mU3^4+54 m3^4 mQ3^2 msq^2 mU3^4-27 m3^4 msq^4 mU3^4-9 m3^2 mQ3^2 msq^4 mU3^4-m3^6 mU3^6+m3^4 mQ3^2 mU3^6-6 m3^4 msq^2 mU3^6-18 m3^2 mQ3^2 msq^2 mU3^6+9 m3^2 msq^4 mU3^6+3 mQ3^2 msq^4 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))+1/3 (-((1920 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(320 m3^3 (2 m3^6-3 m3^4 mQ3^2+7 m3^2 mQ3^4-3 m3^4 mU3^2+4 m3^2 mQ3^2 mU3^2-13 mQ3^4 mU3^2-5 m3^2 mU3^4+11 mQ3^2 mU3^4))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))) Xt+1/3 ((320 m3^4 (-m3^2+mQ3^2) (2 m3^8-10 m3^6 (mQ3^2+mU3^2)+7 mQ3^2 mU3^2 (mQ3^4+mU3^4)+3 m3^4 (mQ3^4+14 mQ3^2 mU3^2+mU3^4)-m3^2 (mQ3^6+21 mQ3^4 mU3^2+21 mQ3^2 mU3^4+mU3^6)))/((m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(480 (m3^2-msq^2) (mQ3^4-2 mQ3^2 mU3^2+mU3^4) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)) Xt^2+1/3 ((3840 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(640 (-m3^2+mQ3^2) (-m3^3 (m3^2-mQ3^2) (m3^2-mU3^2) (2 m3^4-mQ3^4+4 mQ3^2 mU3^2-mU3^4-2 m3^2 (mQ3^2+mU3^2))+6 (mQ3^2-mU3^2)^2 (-2 m3^3 mQ3^2 mU3^2+m3^5 (mQ3^2+mU3^2))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Xt^3+1/3 (-((160 m3^4 (-m3^2+mQ3^2) (mQ3^2+mU3^2) (2 m3^8-10 m3^6 (mQ3^2+mU3^2)+7 mQ3^2 mU3^2 (mQ3^4+mU3^4)+3 m3^4 (mQ3^4+14 mQ3^2 mU3^2+mU3^4)-m3^2 (mQ3^6+21 mQ3^4 mU3^2+21 mQ3^2 mU3^4+mU3^6)))/((m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3))-(240 (m3^2-msq^2) (mQ3^2+mU3^2) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)) Xt^4+1/3 ((320 (mQ3^2+mU3^2) (m3^7+5 m3^5 mQ3^2+5 m3^5 mU3^2-11 m3^3 mQ3^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(1920 (m3^2-msq^2) (mQ3^2+mU3^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Xt^5) Log[msq^2/mQ3^2]) Log[mU3^2/mQ3^2]+((16 (-64 m3^16+12 mQ3^6 mU3^8 (mQ3^2-mU3^2)+m3^14 (139 mQ3^2+245 mU3^2)+12 m3^2 mQ3^4 mU3^6 (-4 mQ3^4+mQ3^2 mU3^2+3 mU3^4)-3 m3^12 (29 mQ3^4+211 mQ3^2 mU3^2+80 mU3^4)+m3^4 mQ3^2 mU3^4 (2 mQ3^6+80 mQ3^4 mU3^2-101 mQ3^2 mU3^4-45 mU3^6)+m3^10 (-22 mQ3^6+700 mQ3^4 mU3^2+432 mQ3^2 mU3^4+170 mU3^6)+2 m3^8 (14 mQ3^8-158 mQ3^6 mU3^2-200 mQ3^4 mU3^4-81 mQ3^2 mU3^6-55 mU3^8)+m3^6 mU3^2 (34 mQ3^8+146 mQ3^6 mU3^2+20 mQ3^4 mU3^4+157 mQ3^2 mU3^6+27 mU3^8)))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2))-(32 (18 m3^11-12 m3 mQ3^4 mU3^6+m3^9 (-19 mQ3^2+195 mU3^2)+m3^7 (27 mQ3^4-425 mQ3^2 mU3^2-130 mU3^4)+m3^5 (168 mQ3^4 mU3^2+263 mQ3^2 mU3^4+13 mU3^6)-m3^3 (87 mQ3^4 mU3^4+11 mQ3^2 mU3^6)) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(32 (-64 m3^16+12 mQ3^6 mU3^8 (mQ3^2-2 mU3^2)+m3^14 (54 mQ3^2+502 mU3^2)+m3^12 (72 mQ3^4-962 mQ3^2 mU3^2-954 mU3^4)+12 m3^2 mQ3^4 mU3^6 (-4 mQ3^4+5 mQ3^2 mU3^2+6 mU3^4)+m3^4 mQ3^2 mU3^4 (7 mQ3^6-61 mQ3^4 mU3^2-325 mQ3^2 mU3^4-97 mU3^6)+m3^10 (-69 mQ3^6+545 mQ3^4 mU3^2+1897 mQ3^2 mU3^4+759 mU3^6)+m3^8 (mQ3^8-79 mQ3^6 mU3^2-1199 mQ3^4 mU3^4-1401 mQ3^2 mU3^6-302 mU3^8)+m3^6 mU3^2 (24 mQ3^8+189 mQ3^6 mU3^2+811 mQ3^4 mU3^4+525 mQ3^2 mU3^6+55 mU3^8)) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)-(64 m3 (42 m3^12+18 mQ3^6 mU3^6-6 mQ3^4 mU3^8-6 m3^10 (mQ3^2+20 mU3^2)+m3^8 (-67 mQ3^4-44 mQ3^2 mU3^2+239 mU3^4)+m3^6 (45 mQ3^6+214 mQ3^4 mU3^2-67 mQ3^2 mU3^4-252 mU3^6)+m3^4 (-90 mQ3^6 mU3^2-195 mQ3^4 mU3^4+268 mQ3^2 mU3^6+59 mU3^8)+m3^2 (59 mQ3^6 mU3^4-42 mQ3^4 mU3^6-55 mQ3^2 mU3^8)) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(16 (-4 m3^16 (37 mQ3^2-197 mU3^2)+4 m3^14 (84 mQ3^4-449 mQ3^2 mU3^2-639 mU3^4)-6 mQ3^6 mU3^8 (mQ3^4-4 mQ3^2 mU3^2-5 mU3^4)-2 m3^12 (97 mQ3^6-393 mQ3^4 mU3^2-3683 mQ3^2 mU3^4-1309 mU3^6)-6 m3^2 mQ3^4 mU3^6 (-4 mQ3^6+15 mQ3^4 mU3^2+30 mQ3^2 mU3^4+15 mU3^6)+m3^4 mQ3^2 mU3^4 (37 mQ3^8-234 mQ3^6 mU3^2+1612 mQ3^4 mU3^4+238 mQ3^2 mU3^6+123 mU3^8)-m3^10 (mQ3^8-452 mQ3^6 mU3^2+6232 mQ3^4 mU3^4+8472 mQ3^2 mU3^6+835 mU3^8)+m3^8 (13 mQ3^10-190 mQ3^8 mU3^2+936 mQ3^6 mU3^4+8712 mQ3^4 mU3^6+3031 mQ3^2 mU3^8+58 mU3^10)-m3^6 mU3^2 (64 mQ3^10-479 mQ3^8 mU3^2+2648 mQ3^6 mU3^4+3742 mQ3^4 mU3^6+116 mQ3^2 mU3^8+69 mU3^10)) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^4)+(32 (12 m3 mQ3^4 mU3^6 (mQ3^2+mU3^2)+4 m3^11 (25 mQ3^2+9 mU3^2)-m3^9 (169 mQ3^4+392 mQ3^2 mU3^2+159 mU3^4)+m3^7 (75 mQ3^6+508 mQ3^4 mU3^2+585 mQ3^2 mU3^4+216 mU3^6)-m3^5 (148 mQ3^6 mU3^2+535 mQ3^4 mU3^4+384 mQ3^2 mU3^6+61 mU3^8)+m3^3 (93 mQ3^6 mU3^4+152 mQ3^4 mU3^6+59 mQ3^2 mU3^8)) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)-(2560 m3^4 mU3^2 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) Log[mU3^2/mQ3^2]^2+(-((128 m3^4 (3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2+4 m3^4 (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2))))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^2))+(512 m3^3 (11 m3^8+3 mQ3^4 mU3^4-6 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)-2 m3^6 (5 mQ3^2+9 mU3^2)+m3^4 (7 mQ3^4+8 mQ3^2 mU3^2+11 mU3^4)) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(8192 m3^6 Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))+(256 m3 (2 m3^2-mQ3^2-mU3^2) (3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2+4 m3^4 (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2))) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(64 (2 m3^2-mQ3^2-mU3^2) (53 m3^8-3 mQ3^4 mU3^4-50 m3^6 (mQ3^2+mU3^2)+6 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)+m3^4 (-7 mQ3^4+52 mQ3^2 mU3^2-7 mU3^4)) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(1024 m3^3 (2 m3^2-mQ3^2-mU3^2) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mQ3^2]+(-((128 m3^4 (3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2+4 m3^4 (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2))))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^4))-(512 m3^3 (11 m3^8+3 mQ3^4 mU3^4-6 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)-2 m3^6 (9 mQ3^2+5 mU3^2)+m3^4 (11 mQ3^4+8 mQ3^2 mU3^2+7 mU3^4)) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(8192 m3^6 Xt^2)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(256 m3 (2 m3^2-mQ3^2-mU3^2) (3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2+4 m3^4 (2 m3^4+mQ3^4+mU3^4-2 m3^2 (mQ3^2+mU3^2))) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(64 (2 m3^2-mQ3^2-mU3^2) (53 m3^8-3 mQ3^4 mU3^4-50 m3^6 (mQ3^2+mU3^2)+6 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)+m3^4 (-7 mQ3^4+52 mQ3^2 mU3^2-7 mU3^4)) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(1024 m3^3 (2 m3^2-mQ3^2-mU3^2) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mU3^2])+(-((80 (-msq^2+mU3^2) (2 m3^4-3 m3^2 msq^2+3 m3^2 mU3^2-msq^2 mU3^2-mU3^4))/(m3^2-mU3^2)^3)+(640 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(160 (-msq^2+mU3^2) (2 m3^4-3 m3^2 msq^2+3 m3^2 mU3^2-msq^2 mU3^2-mU3^4) Xt^4)/((m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)-(1280 m3 (-msq^2+mU3^2)^2 Xt^5)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^3)) PolyLog[2,1-msq^2/mU3^2]+(-((8 (mQ3^2-mU3^2) (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))+(64 (mQ3^2-mU3^2) (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(16 (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(128 (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-mU3^2/mQ3^2]+Log[mU3^2/mQ3^2] ((4 (4 m3^12 (mQ3^4-mU3^4)-m3^10 (8 mQ3^6+611 mQ3^4 mU3^2+33 mQ3^2 mU3^4-140 mU3^6)-3 mQ3^4 mU3^6 (mQ3^2-mU3^2) (4 mQ3^4+20 msq^2 mU3^2+mQ3^2 (40 msq^2-57 mU3^2))+m3^4 mQ3^2 mU3^2 (18 mQ3^8-283 mQ3^2 mU3^6-420 msq^2 mU3^6-587 mU3^8+mQ3^6 (180 msq^2+541 mU3^2)+mQ3^4 (240 msq^2 mU3^2+2359 mU3^4))-m3^2 mQ3^2 mU3^4 (30 mQ3^8-180 msq^2 mU3^6+5 mQ3^6 (60 msq^2+101 mU3^2)+mQ3^4 (-720 msq^2 mU3^2+571 mU3^4)+6 mQ3^2 (100 msq^2 mU3^4-99 mU3^6))+m3^8 (4 mQ3^8+704 mQ3^6 mU3^2-204 mU3^8+3 mQ3^4 (120 msq^2 mU3^2+709 mU3^4)-mQ3^2 (360 msq^2 mU3^4+583 mU3^6))-m3^6 mU3^2 (207 mQ3^8-68 mU3^8+3 mQ3^6 (100 msq^2+739 mU3^2)+5 mQ3^4 (108 msq^2 mU3^2+395 mU3^4)-mQ3^2 (840 msq^2 mU3^4+1259 mU3^6))))/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(64 (2 m3^7 (11 mQ3^2+8 mU3^2)-4 m3^5 (5 mQ3^4+28 mQ3^2 mU3^2+4 mU3^4)+m3^3 (3 mQ3^6+87 mQ3^2 mU3^4+10 mQ3^4 (3 msq^2+10 mU3^2))-2 m3 mQ3^2 mU3^2 (3 mQ3^4-15 msq^2 mU3^2+mQ3^2 (30 msq^2+37 mU3^2))) Xt)/(3 mQ3^2 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(8 (4 m3^10 (mQ3^2-mU3^2)+m3^8 (-8 mQ3^4-695 mQ3^2 mU3^2+136 mU3^4)+3 mQ3^4 mU3^4 (2 mQ3^4+2 mU3^2 (10 msq^2+mU3^2)+mQ3^2 (20 msq^2+79 mU3^2))-m3^4 mQ3^2 mU3^2 (301 mQ3^4+20 mQ3^2 (15 msq^2+2 mU3^2)+5 (60 msq^2 mU3^2+mU3^4))+2 m3^2 mQ3^2 mU3^2 (9 mQ3^6+2 mQ3^4 (45 msq^2-53 mU3^2)+9 mU3^4 (10 msq^2+mU3^2)-2 mQ3^2 (60 msq^2 mU3^2+77 mU3^4))+m3^6 (4 mQ3^6+886 mQ3^4 mU3^2-68 mU3^6+mQ3^2 (360 msq^2 mU3^2+326 mU3^4))) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(128 (-m3 mQ3^2 (m3^2-mU3^2) (mQ3^2+3 mU3^2) (11 m3^4-5 mQ3^2 mU3^2-3 m3^2 (mQ3^2+mU3^2))-(m3^2-mU3^2) (mQ3^3-mQ3 mU3^2)^2 (-14 m3^3+3 m3 (mQ3^2+10 msq^2+mU3^2))-2 (m3^2-mQ3^2) (mQ3^2-mU3^2) (-17 m3 mQ3^2 mU3^4+m3^5 (7 mQ3^2-8 mU3^2)+m3^3 (26 mQ3^2 mU3^2+8 mU3^4))) Xt^3)/(3 mQ3^2 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(4 (4 m3^12 (mQ3^4-44 mQ3^2 mU3^2-mU3^4)+m3^10 (-8 mQ3^6-667 mQ3^4 mU3^2+4793 mQ3^2 mU3^4+140 mU3^6)+3 mQ3^4 mU3^6 (2 mQ3^6-20 msq^2 mU3^4+2 mU3^6+mQ3^4 (20 msq^2-389 mU3^2)+mQ3^2 (-80 msq^2 mU3^2+139 mU3^4))+m3^8 (4 mQ3^8+1222 mQ3^6 mU3^2-204 mU3^8+3 mQ3^4 (120 msq^2 mU3^2-3367 mU3^4)+mQ3^2 (360 msq^2 mU3^4-7283 mU3^6))+m3^4 mQ3^2 mU3^2 (18 mQ3^8+360 msq^2 mU3^6+975 mU3^8+mQ3^6 (180 msq^2-1361 mU3^2)-5 mQ3^4 (96 msq^2 mU3^2+2687 mU3^4)+mQ3^2 (1380 msq^2 mU3^4-6073 mU3^6))+m3^6 mU3^2 (-493 mQ3^8+68 mU3^8+mQ3^6 (-300 msq^2+6913 mU3^2)+mQ3^4 (-600 msq^2 mU3^2+17667 mU3^4)-3 mQ3^2 (340 msq^2 mU3^4-627 mU3^6))+m3^2 mQ3^2 mU3^4 (24 mQ3^8+18 mU3^6 (-10 msq^2+mU3^2)+5 mQ3^6 (48 msq^2+533 mU3^2)+mQ3^4 (-420 msq^2 mU3^2+5659 mU3^4)+4 mQ3^2 (90 msq^2 mU3^4-377 mU3^6))) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(64 (16 m3^7 (3 mQ3^2+mU3^2)-2 m3^5 (31 mQ3^4-86 mQ3^2 mU3^2+8 mU3^4)+m3 mQ3^2 mU3^2 (3 mQ3^4-30 msq^2 mU3^2+3 mU3^4+2 mQ3^2 (15 msq^2+92 mU3^2))+m3^3 (3 mQ3^6+2 mQ3^4 (15 msq^2-83 mU3^2)-5 mQ3^2 (6 msq^2 mU3^2+37 mU3^4))) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+((80 (18 m3^10-3 mQ3^2 msq^2 mU3^6+m3^8 (-35 mQ3^2+12 msq^2-47 mU3^2)+mQ3^4 (3 msq^4 mU3^2-3 msq^2 mU3^4-13 mU3^6)+m3^6 (17 mQ3^4+9 msq^4-51 msq^2 mU3^2+43 mU3^4+mQ3^2 (-3 msq^2+91 mU3^2))+m3^4 (3 msq^4 mU3^2+24 msq^2 mU3^4-14 mU3^6+mQ3^4 (3 msq^2-44 mU3^2)+mQ3^2 (-18 msq^4+39 msq^2 mU3^2-83 mU3^4))+m3^2 (-9 msq^2 mU3^6+mQ3^4 (9 msq^4-24 msq^2 mU3^2+40 mU3^4)+mQ3^2 (-6 msq^4 mU3^2+15 msq^2 mU3^4+27 mU3^6))))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3)+(640 m3 (m3^6-3 msq^2 mU3^4-m3^4 (mQ3^2+2 mU3^2)-mQ3^2 (3 msq^4-9 msq^2 mU3^2+mU3^4)+m3^2 (3 msq^4-3 msq^2 mU3^2+mU3^4+mQ3^2 (-3 msq^2+2 mU3^2))) Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(160 (18 m3^8+3 mQ3^2 msq^2 mU3^4+m3^6 (-35 mQ3^2+18 msq^2-33 mU3^2)+mQ3^4 (3 msq^2 mU3^2+14 mU3^4)+m3^4 (17 mQ3^4+15 mU3^2 (-msq^2+mU3^2)+mQ3^2 (-15 msq^2+64 mU3^2))+m3^2 (9 msq^2 mU3^4+mQ3^4 (9 msq^2-31 mU3^2)-mQ3^2 (12 msq^2 mU3^2+29 mU3^4))) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(640 m3 (6 msq^2 (mQ3^2-mU3^2)^2-(m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2+3 mU3^2)) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+(80 (6 (m3^2-mQ3^2)^2 msq^2 (mQ3^2-mU3^2) (2 m3^4-msq^2 mU3^2-3 m3^2 (msq^2-2 mU3^2))+(m3^2-mQ3^2) (m3^2-mU3^2) (4 m3^8-3 mQ3^2 mU3^4 (mQ3^2+5 mU3^2)+m3^6 (3 mQ3^2+7 mU3^2)-m3^4 (6 mQ3^4+15 mQ3^2 mU3^2+29 mU3^4)+m3^2 (7 mQ3^4 mU3^2+31 mQ3^2 mU3^4+16 mU3^6))+(m3^2-mU3^2) (mQ3^2+mU3^2) (11 m3^8+3 mQ3^2 msq^2 mU3^4-2 m3^6 (11 mQ3^2-9 msq^2+11 mU3^2)+mQ3^4 (3 msq^2 mU3^2+11 mU3^4)+m3^4 (11 mQ3^4-15 msq^2 mU3^2+11 mU3^4+mQ3^2 (-15 msq^2+44 mU3^2))+m3^2 (9 msq^2 mU3^4+mQ3^4 (9 msq^2-22 mU3^2)-2 mQ3^2 (6 msq^2 mU3^2+11 mU3^4)))) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(640 m3 (-6 (m3^2-mQ3^2) msq^2 (msq^2-2 mU3^2)+(m3^2-mQ3^2) mU3^2 (m3^2-mU3^2)-3 msq^2 (m3^2-mU3^2) (mQ3^2+mU3^2)) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[msq^2/mQ3^2]+(80-(160 Xt^2)/(mQ3^2-mU3^2)+((80 mQ3^2)/(mQ3^2-mU3^2)^3+(80 mU3^2)/(mQ3^2-mU3^2)^3) Xt^4) Log[msq^2/mQ3^2]^2+(-((8 (128 m3^12-10 m3^10 (29 mQ3^2+35 mU3^2)+m3^8 (98 mQ3^4+762 mQ3^2 mU3^2+420 mU3^4)+m3^6 (9 mQ3^6-251 mQ3^4 mU3^2-821 mQ3^2 mU3^4-217 mU3^6)-mQ3^4 mU3^2 (3 mQ3^6+mQ3^4 mU3^2+19 mQ3^2 mU3^4-23 mU3^6)+m3^4 (20 mQ3^8-151 mQ3^6 mU3^2+417 mQ3^4 mU3^4+323 mQ3^2 mU3^6+31 mU3^8)+m3^2 (-9 mQ3^10+41 mQ3^8 mU3^2+41 mQ3^6 mU3^4-167 mQ3^4 mU3^6-34 mQ3^2 mU3^8)))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(64 m3 (20 m3^8-37 m3^6 mQ3^2+19 m3^4 mQ3^4-5 m3^2 mQ3^6+3 mQ3^8-73 m3^6 mU3^2+135 m3^4 mQ3^2 mU3^2-23 m3^2 mQ3^4 mU3^2-7 mQ3^6 mU3^2+24 m3^4 mU3^4-78 m3^2 mQ3^2 mU3^4+22 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(16 (-128 m3^12+6 m3^10 (49 mQ3^2+143 mU3^2)-2 m3^8 (49 mQ3^4+707 mQ3^2 mU3^2+652 mU3^4)+mQ3^4 mU3^2 (3 mQ3^6+mQ3^4 mU3^2+19 mQ3^2 mU3^4-23 mU3^6)+m3^6 (-9 mQ3^6+251 mQ3^4 mU3^2+2049 mQ3^2 mU3^4+525 mU3^6)+m3^2 mQ3^2 (9 mQ3^8-41 mQ3^6 mU3^2-41 mQ3^4 mU3^4+167 mQ3^2 mU3^6+34 mU3^8)+m3^4 (-20 mQ3^8+151 mQ3^6 mU3^2-417 mQ3^4 mU3^4-903 mQ3^2 mU3^6+37 mU3^8)) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(64 m3 (2 m3^6-6 mQ3^6+20 mQ3^4 mU3^2-75 mQ3^2 mU3^4+27 mU3^6-m3^4 (37 mQ3^2+33 mU3^2)-2 m3^2 (mQ3^4-55 mQ3^2 mU3^2+3 mU3^4)) Xt^3)/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(8 (124 m3^12 mQ3^2-284 m3^10 mQ3^4+90 m3^8 mQ3^6+11 m3^6 mQ3^8+20 m3^4 mQ3^10-9 m3^2 mQ3^12+1156 m3^12 mU3^2-3068 m3^10 mQ3^2 mU3^2+2702 m3^8 mQ3^4 mU3^2-474 m3^6 mQ3^6 mU3^2-201 m3^4 mQ3^8 mU3^2+32 m3^2 mQ3^10 mU3^2-3 mQ3^12 mU3^2-3048 m3^10 mU3^4+7406 m3^8 mQ3^2 mU3^4-5516 m3^6 mQ3^4 mU3^4+882 m3^4 mQ3^6 mU3^4+184 m3^2 mQ3^8 mU3^4-4 mQ3^10 mU3^4+2602 m3^8 mU3^6-6126 m3^6 mQ3^2 mU3^6+3984 m3^4 mQ3^4 mU3^6-502 m3^2 mQ3^6 mU3^6-54 mQ3^8 mU3^6-695 m3^6 mU3^8+1718 m3^4 mQ3^2 mU3^8-883 m3^2 mQ3^4 mU3^8+4 mQ3^6 mU3^8-3 m3^4 mU3^10-102 m3^2 mQ3^2 mU3^10+57 mQ3^4 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(64 m3 (3 mQ3^6-4 mQ3^4 mU3^2-mQ3^2 mU3^4+6 mU3^6+2 m3^4 (9 mQ3^2-7 mU3^2)+m3^2 (mQ3^4-20 mQ3^2 mU3^2+11 mU3^4)) Xt^5)/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) PolyLog[2,1-m3^2/mQ3^2]+((80 (m3^2-msq^2) (mQ3^6-3 mQ3^4 mU3^2+3 mQ3^2 mU3^4-mU3^6) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(640 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(160 (m3^2-msq^2) (mQ3^4-2 mQ3^2 mU3^2+mU3^4) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))) Xt^2)/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(1280 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^3)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(80 (m3^2-msq^2) (mQ3^2+mU3^2) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))) Xt^4)/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(640 (m3^2-msq^2) (mQ3^2+mU3^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-msq^2/m3^2]+((80 (mQ3^2-msq^2) (2 m3^4+3 m3^2 mQ3^2-mQ3^4-3 m3^2 msq^2-mQ3^2 msq^2))/(m3^2-mQ3^2)^3-(640 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))-(160 (mQ3^2-msq^2) (2 m3^4+3 m3^2 mQ3^2-mQ3^4-3 m3^2 msq^2-mQ3^2 msq^2) Xt^2)/((m3^2-mQ3^2)^3 (mQ3^2-mU3^2))+(1280 m3 (mQ3^2-msq^2)^2 Xt^3)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)+(80 (mQ3^2-msq^2) (2 m3^4+3 m3^2 mQ3^2-mQ3^4-3 m3^2 msq^2-mQ3^2 msq^2) (mQ3^2+mU3^2) Xt^4)/((m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^3)-(640 m3 (mQ3^2-msq^2)^2 (mQ3^2+mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) PolyLog[2,1-msq^2/mQ3^2]+(-((8 (4 m3^4 (m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^4 (m3^4-34 m3^2 mU3^2+17 mU3^4)+(m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^3 (128 m3^12-2 m3^10 (173 mQ3^2+147 mU3^2)+m3^8 (280 mQ3^4+902 mQ3^2 mU3^2+98 mU3^4)-m3^6 (13 mQ3^6+1025 mQ3^4 mU3^2+251 mQ3^2 mU3^4-9 mU3^6)+mQ3^2 mU3^4 (23 mQ3^6-19 mQ3^4 mU3^2-mQ3^2 mU3^4-3 mU3^6)+m3^4 (-37 mQ3^8+391 mQ3^6 mU3^2+417 mQ3^4 mU3^4-151 mQ3^2 mU3^6+20 mU3^8)+m3^2 (-34 mQ3^8 mU3^2-167 mQ3^6 mU3^4+41 mQ3^4 mU3^6+41 mQ3^2 mU3^8-9 mU3^10))))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^4))-(64 (2 (m3^2-mQ3^2)^2 (m3^7-50 m3^5 mU3^2+17 m3^3 mU3^4)+m3 (m3^2-mU3^2)^3 (18 m3^4+22 mQ3^4-7 mQ3^2 mU3^2+3 mU3^4+m3^2 (-37 mQ3^2+mU3^2))) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(16 (128 m3^12-2 m3^10 (173 mQ3^2+403 mU3^2)+m3^8 (280 mQ3^4+2438 mQ3^2 mU3^2+98 mU3^4)-m3^6 (13 mQ3^6+2561 mQ3^4 mU3^2+251 mQ3^2 mU3^4-9 mU3^6)+mQ3^2 mU3^4 (23 mQ3^6-19 mQ3^4 mU3^2-mQ3^2 mU3^4-3 mU3^6)+m3^4 (-37 mQ3^8+903 mQ3^6 mU3^2+417 mQ3^4 mU3^4-151 mQ3^2 mU3^6+20 mU3^8)+m3^2 (-34 mQ3^8 mU3^2-167 mQ3^6 mU3^4+41 mQ3^4 mU3^6+41 mQ3^2 mU3^8-9 mU3^10)) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(64 (m3 (m3^2-mQ3^2)^2 (2 m3^2-mQ3^2-mU3^2) (m3^4-34 m3^2 mU3^2+17 mU3^4)+2 m3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2) (18 m3^4+22 mQ3^4-7 mQ3^2 mU3^2+3 mU3^4+m3^2 (-37 mQ3^2+mU3^2))) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(8 (4 m3^12 (31 mQ3^2+289 mU3^2)-4 m3^10 (83 mQ3^4+1025 mQ3^2 mU3^2+492 mU3^4)+m3^8 (262 mQ3^6+5326 mQ3^4 mU3^2+6466 mQ3^2 mU3^4+746 mU3^6)+mQ3^2 mU3^4 (-11 mQ3^8+4 mQ3^6 mU3^2+14 mQ3^4 mU3^4-4 mQ3^2 mU3^6-3 mU3^8)-m3^6 (3 mQ3^8+2990 mQ3^6 mU3^2+7596 mQ3^4 mU3^4+2186 mQ3^2 mU3^6+25 mU3^8)-m3^2 mU3^2 (-34 mQ3^10+543 mQ3^8 mU3^2+774 mQ3^6 mU3^4+20 mQ3^4 mU3^6-32 mQ3^2 mU3^8+9 mU3^10)+m3^4 (-39 mQ3^10+526 mQ3^8 mU3^2+3712 mQ3^6 mU3^4+2210 mQ3^4 mU3^6-29 mQ3^2 mU3^8+20 mU3^10)) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)+(64 (16 m3 (m3^2-mQ3^2)^2 mU3^2 (2 m3^2-mQ3^2-mU3^2)-m3 (m3^2-mU3^2) (mQ3^2+mU3^2) (18 m3^4+22 mQ3^4-7 mQ3^2 mU3^2+3 mU3^4+m3^2 (-37 mQ3^2+mU3^2))) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^4)) PolyLog[2,1-m3^2/mU3^2]+((80 (-msq^2+mU3^2) (2 m3^4-3 m3^2 msq^2+3 m3^2 mU3^2-msq^2 mU3^2-mU3^4))/(m3^2-mU3^2)^3-(640 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(160 (-msq^2+mU3^2) (2 m3^4-3 m3^2 msq^2+3 m3^2 mU3^2-msq^2 mU3^2-mU3^4) Xt^2)/((m3^2-mU3^2)^3 (-mQ3^2+mU3^2))-(1280 m3 (-msq^2+mU3^2)^2 Xt^3)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)-(80 (mQ3^2+mU3^2) (-msq^2+mU3^2) (2 m3^4-3 m3^2 msq^2+3 m3^2 mU3^2-msq^2 mU3^2-mU3^4) Xt^4)/((m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^3)+(640 m3 (mQ3^2+mU3^2) (-msq^2+mU3^2)^2 Xt^5)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) PolyLog[2,1-msq^2/mU3^2]+((8 (mQ3^2-mU3^2) (12 m3^10+3 mQ3^8 mU3^2+7 mQ3^6 mU3^4+7 mQ3^4 mU3^6+3 mQ3^2 mU3^8-14 m3^8 (mQ3^2+mU3^2)+m3^6 (34 mQ3^4-76 mQ3^2 mU3^2+34 mU3^4)+m3^4 (-29 mQ3^6+65 mQ3^4 mU3^2+65 mQ3^2 mU3^4-29 mU3^6)+m3^2 (9 mQ3^8-26 mQ3^6 mU3^2-34 mQ3^4 mU3^4-26 mQ3^2 mU3^6+9 mU3^8)))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(64 m3 (mQ3^2-mU3^2) (5 m3^4+3 mQ3^4-mQ3^2 mU3^2+3 mU3^4-5 m3^2 (mQ3^2+mU3^2)) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(16 (12 m3^10+3 mQ3^8 mU3^2+7 mQ3^6 mU3^4+7 mQ3^4 mU3^6+3 mQ3^2 mU3^8-14 m3^8 (mQ3^2+mU3^2)+m3^6 (34 mQ3^4-76 mQ3^2 mU3^2+34 mU3^4)+m3^4 (-29 mQ3^6+65 mQ3^4 mU3^2+65 mQ3^2 mU3^4-29 mU3^6)+m3^2 (9 mQ3^8-26 mQ3^6 mU3^2-34 mQ3^4 mU3^4-26 mQ3^2 mU3^6+9 mU3^8)) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(128 m3 (5 m3^4+3 mQ3^4-mQ3^2 mU3^2+3 mU3^4-5 m3^2 (mQ3^2+mU3^2)) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(8 (mQ3^2+mU3^2) (12 m3^10+3 mQ3^8 mU3^2+7 mQ3^6 mU3^4+7 mQ3^4 mU3^6+3 mQ3^2 mU3^8-14 m3^8 (mQ3^2+mU3^2)+m3^6 (34 mQ3^4-76 mQ3^2 mU3^2+34 mU3^4)+m3^4 (-29 mQ3^6+65 mQ3^4 mU3^2+65 mQ3^2 mU3^4-29 mU3^6)+m3^2 (9 mQ3^8-26 mQ3^6 mU3^2-34 mQ3^4 mU3^4-26 mQ3^2 mU3^6+9 mU3^8)) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(64 m3 (mQ3^2+mU3^2) (5 m3^4+3 mQ3^4-mQ3^2 mU3^2+3 mU3^4-5 m3^2 (mQ3^2+mU3^2)) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) PolyLog[2,1-mU3^2/mQ3^2])+Log[MR^2/mQ3^2] (-((8 (156 m3^10 mQ3^2-280 m3^8 mQ3^4+124 m3^6 mQ3^6+156 m3^10 mU3^2-1295 m3^8 mQ3^2 mU3^2+1836 m3^6 mQ3^4 mU3^2-689 m3^4 mQ3^6 mU3^2-280 m3^8 mU3^4+1836 m3^6 mQ3^2 mU3^4-2356 m3^4 mQ3^4 mU3^4+784 m3^2 mQ3^6 mU3^4+124 m3^6 mU3^6-689 m3^4 mQ3^2 mU3^6+784 m3^2 mQ3^4 mU3^6-211 mQ3^6 mU3^6))/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2))-(128 m3^3 Xt)/(3 (-m3^2+mQ3^2) (m3^2-mU3^2))+(64 (-39 m3^6+31 m3^4 mQ3^2+31 m3^4 mU3^2-23 m3^2 mQ3^2 mU3^2) Xt^2)/(3 mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2))+(256 (-67 m3^5+59 m3^3 mQ3^2+59 m3^3 mU3^2-51 m3 mQ3^2 mU3^2) Xt^3)/(3 (-m3^2+mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)-(32 (39 m3^10 mQ3^2-70 m3^8 mQ3^4+31 m3^6 mQ3^6+39 m3^10 mU3^2+396 m3^8 mQ3^2 mU3^2-822 m3^6 mQ3^4 mU3^2+383 m3^4 mQ3^6 mU3^2-70 m3^8 mU3^4-822 m3^6 mQ3^2 mU3^4+1656 m3^4 mQ3^4 mU3^4-756 m3^2 mQ3^6 mU3^4+31 m3^6 mU3^6+383 m3^4 mQ3^2 mU3^6-756 m3^2 mQ3^4 mU3^6+338 mQ3^6 mU3^6) Xt^4)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(256 m3^3 Xt^5)/(3 (-m3^2+mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+((160 m3^2)/(3 mQ3^2)+(160 m3^2)/(3 mU3^2)-(320 m3^2 Xt^2)/(3 mQ3^2 mU3^2)-(1280 m3 Xt^3)/(3 (mQ3^2-mU3^2)^2)+((160 m3^2)/(3 mQ3^2 (mQ3^2-mU3^2)^2)+(160 m3^2)/(3 mU3^2 (mQ3^2-mU3^2)^2)) Xt^4) Log[msq^2/mQ3^2]+((16 (m3^12 mQ3^4-2 m3^10 mQ3^6+m3^8 mQ3^8-539 m3^10 mQ3^4 mU3^2+983 m3^8 mQ3^6 mU3^2-440 m3^6 mQ3^8 mU3^2-m3^12 mU3^4+442 m3^10 mQ3^2 mU3^4+720 m3^8 mQ3^4 mU3^4-2386 m3^6 mQ3^6 mU3^4+1209 m3^4 mQ3^8 mU3^4+35 m3^10 mU3^6-1397 m3^8 mQ3^2 mU3^6+1006 m3^6 mQ3^4 mU3^6+1481 m3^4 mQ3^6 mU3^6-1101 m3^2 mQ3^8 mU3^6-51 m3^8 mU3^8+1419 m3^6 mQ3^2 mU3^8-1974 m3^4 mQ3^4 mU3^8+263 m3^2 mQ3^6 mU3^8+327 mQ3^8 mU3^8+17 m3^6 mU3^10-460 m3^4 mQ3^2 mU3^10+774 m3^2 mQ3^4 mU3^10-327 mQ3^6 mU3^10))/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(128 (-39 m3^7 mQ3^2+30 m3^5 mQ3^4-8 m3^7 mU3^2+119 m3^5 mQ3^2 mU3^2-93 m3^3 mQ3^4 mU3^2+8 m3^5 mU3^4-81 m3^3 mQ3^2 mU3^4+64 m3 mQ3^4 mU3^4) Xt)/(3 mQ3^2 (-m3^2+mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(32 (m3^10 mQ3^2-2 m3^8 mQ3^4+m3^6 mQ3^6-m3^10 mU3^2-568 m3^8 mQ3^2 mU3^2+1044 m3^6 mQ3^4 mU3^2-471 m3^4 mQ3^6 mU3^2+34 m3^8 mU3^4+870 m3^6 mQ3^2 mU3^4-1652 m3^4 mQ3^4 mU3^4+740 m3^2 mQ3^6 mU3^4-17 m3^6 mU3^6-347 m3^4 mQ3^2 mU3^6+650 m3^2 mQ3^4 mU3^6-282 mQ3^6 mU3^6) Xt^2)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(256 (-27 m3^7 mQ3^4+22 m3^5 mQ3^6-32 m3^7 mQ3^2 mU3^2+56 m3^5 mQ3^4 mU3^2-17 m3^3 mQ3^6 mU3^2-8 m3^7 mU3^4+99 m3^5 mQ3^2 mU3^4-93 m3^3 mQ3^4 mU3^4+3 m3 mQ3^6 mU3^4+8 m3^5 mU3^6-59 m3^3 mQ3^2 mU3^6+48 m3 mQ3^4 mU3^6) Xt^3)/(3 mQ3^2 (-m3^2+mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(16 (m3^12 mQ3^4-2 m3^10 mQ3^6+m3^8 mQ3^8-156 m3^12 mQ3^2 mU3^2-355 m3^10 mQ3^4 mU3^2+1051 m3^8 mQ3^6 mU3^2-536 m3^6 mQ3^8 mU3^2-m3^12 mU3^4+382 m3^10 mQ3^2 mU3^4+600 m3^8 mQ3^4 mU3^4-2052 m3^6 mQ3^6 mU3^4+1063 m3^4 mQ3^8 mU3^4+35 m3^10 mU3^6+635 m3^8 mQ3^2 mU3^6-2203 m3^6 mQ3^4 mU3^6+2253 m3^4 mQ3^6 mU3^6-720 m3^2 mQ3^8 mU3^6-51 m3^8 mU3^8-1634 m3^6 mQ3^2 mU3^8+3467 m3^4 mQ3^4 mU3^8-1974 m3^2 mQ3^6 mU3^8+200 mQ3^8 mU3^8+17 m3^6 mU3^10+765 m3^4 mQ3^2 mU3^10-1486 m3^2 mQ3^4 mU3^10+700 mQ3^6 mU3^10) Xt^4)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(128 (-24 m3^7 mQ3^2+23 m3^5 mQ3^4-8 m3^7 mU3^2-45 m3^5 mQ3^2 mU3^2+53 m3^3 mQ3^4 mU3^2+8 m3^5 mU3^4+71 m3^3 mQ3^2 mU3^4-78 m3 mQ3^4 mU3^4) Xt^5)/(3 mQ3^2 (-m3^2+mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+((640 m3 Xt)/(3 (mQ3^2-mU3^2))+(-((640 m3 mQ3^2)/(3 (mQ3^2-mU3^2)^3))-(640 m3 mU3^2)/(3 (mQ3^2-mU3^2)^3)) Xt^3+(320 m3^2 Xt^4)/(3 (mQ3^2-mU3^2)^3)) Log[msq^2/mQ3^2]) Log[mU3^2/mQ3^2]+((16 (-52 m3^6 mQ3^2-12 m3^6 mU3^2+242 m3^4 mQ3^2 mU3^2-114 m3^4 mU3^4-291 m3^2 mQ3^2 mU3^4+227 m3^2 mU3^6+97 mQ3^2 mU3^6-97 mU3^8))/(3 (mQ3^2-mU3^2) (-m3^2+mU3^2)^3)+(64 (41 m3^5-150 m3^3 mU3^2+95 m3 mU3^4) Xt)/(3 (mQ3^2-mU3^2) (-m3^2+mU3^2)^2)-(32 (4 m3^6 mQ3^2-140 m3^6 mU3^2+186 m3^4 mQ3^2 mU3^2+30 m3^4 mU3^4-291 m3^2 mQ3^2 mU3^4+267 m3^2 mU3^6+97 mQ3^2 mU3^6-153 mU3^8) Xt^2)/(3 (mQ3^2-mU3^2)^2 (-m3^2+mU3^2)^3)-(64 (-56 m3^7-3 m3^5 mQ3^2+173 m3^5 mU3^2-210 m3^3 mQ3^2 mU3^2-26 m3^3 mU3^4+201 m3 mQ3^2 mU3^4-111 m3 mU3^6) Xt^3)/(3 (mQ3^2-mU3^2)^3 (-m3^2+mU3^2)^2)-(16 (58 m3^8 mQ3^2-32 m3^6 mQ3^4-58 m3^8 mU3^2-30 m3^6 mQ3^2 mU3^2-102 m3^4 mQ3^4 mU3^2+478 m3^6 mU3^4-34 m3^4 mQ3^2 mU3^4+207 m3^2 mQ3^4 mU3^4-472 m3^4 mU3^6-106 m3^2 mQ3^2 mU3^6-69 mQ3^4 mU3^6-133 m3^2 mU3^8+112 mQ3^2 mU3^8+181 mU3^10) Xt^4)/(3 (mQ3^2-mU3^2)^4 (-m3^2+mU3^2)^3)+(128 (-8 m3^5 mQ3^2-24 m3^5 mU3^2-30 m3^3 mQ3^2 mU3^2-14 m3^3 mU3^4+39 m3 mQ3^2 mU3^4+39 m3 mU3^6) Xt^5)/(3 (mQ3^2-mU3^2)^4 (-m3^2+mU3^2)^2)) Log[mU3^2/mQ3^2]^2+Log[m3^2/mQ3^2] ((16 (-100 m3^14 mQ3^2+268 m3^12 mQ3^4-252 m3^10 mQ3^6+84 m3^8 mQ3^8-100 m3^14 mU3^2+530 m3^12 mQ3^2 mU3^2-896 m3^10 mQ3^4 mU3^2+647 m3^8 mQ3^6 mU3^2-185 m3^6 mQ3^8 mU3^2+268 m3^12 mU3^4-896 m3^10 mQ3^2 mU3^4+1054 m3^8 mQ3^4 mU3^4-531 m3^6 mQ3^6 mU3^4+117 m3^4 mQ3^8 mU3^4-252 m3^10 mU3^6+647 m3^8 mQ3^2 mU3^6-531 m3^6 mQ3^4 mU3^6+136 m3^4 mQ3^6 mU3^6-12 m3^2 mQ3^8 mU3^6+84 m3^8 mU3^8-185 m3^6 mQ3^2 mU3^8+117 m3^4 mQ3^4 mU3^8-12 m3^2 mQ3^6 mU3^8))/(3 mQ3^2 (-m3^2+mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3)+(128 (-8 m3^9 mQ3^2+8 m3^7 mQ3^4-8 m3^9 mU3^2+64 m3^7 mQ3^2 mU3^2-57 m3^5 mQ3^4 mU3^2+8 m3^7 mU3^4-57 m3^5 mQ3^2 mU3^4+50 m3^3 mQ3^4 mU3^4) Xt)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)-(128 (25 m3^10-42 m3^8 mQ3^2+21 m3^6 mQ3^4-42 m3^8 mU3^2+54 m3^6 mQ3^2 mU3^2-20 m3^4 mQ3^4 mU3^2+21 m3^6 mU3^4-20 m3^4 mQ3^2 mU3^4+3 m3^2 mQ3^4 mU3^4) Xt^2)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)-(512 (-4 m3^9 mQ3^4+4 m3^7 mQ3^6+19 m3^9 mQ3^2 mU3^2-18 m3^7 mQ3^4 mU3^2+3 m3^5 mQ3^6 mU3^2-4 m3^9 mU3^4-18 m3^7 mQ3^2 mU3^4+20 m3^5 mQ3^4 mU3^4-6 m3^3 mQ3^6 mU3^4+4 m3^7 mU3^6+3 m3^5 mQ3^2 mU3^6-6 m3^3 mQ3^4 mU3^6+3 m3 mQ3^6 mU3^6) Xt^3)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(16 (-100 m3^14 mQ3^2+268 m3^12 mQ3^4-252 m3^10 mQ3^6+84 m3^8 mQ3^8-100 m3^14 mU3^2-508 m3^12 mQ3^2 mU3^2+1136 m3^10 mQ3^4 mU3^2-502 m3^8 mQ3^6 mU3^2-18 m3^6 mQ3^8 mU3^2+268 m3^12 mU3^4+1136 m3^10 mQ3^2 mU3^4-2500 m3^8 mQ3^4 mU3^4+1078 m3^6 mQ3^6 mU3^4-6 m3^4 mQ3^8 mU3^4-252 m3^10 mU3^6-502 m3^8 mQ3^2 mU3^6+1078 m3^6 mQ3^4 mU3^6-232 m3^4 mQ3^6 mU3^6-68 m3^2 mQ3^8 mU3^6+84 m3^8 mU3^8-18 m3^6 mQ3^2 mU3^8-6 m3^4 mQ3^4 mU3^8-68 m3^2 mQ3^6 mU3^8) Xt^4)/(3 mQ3^2 (-m3^2+mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(256 (-4 m3^9 mQ3^2+4 m3^7 mQ3^4-4 m3^9 mU3^2-38 m3^7 mQ3^2 mU3^2+43 m3^5 mQ3^4 mU3^2+4 m3^7 mU3^4+43 m3^5 mQ3^2 mU3^4-48 m3^3 mQ3^4 mU3^4) Xt^5)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+((16 (210 m3^12-368 m3^10 mQ3^2+199 m3^8 mQ3^4-45 m3^6 mQ3^6-480 m3^10 mU3^2+718 m3^8 mQ3^2 mU3^2-275 m3^6 mQ3^4 mU3^2+49 m3^4 mQ3^6 mU3^2+367 m3^8 mU3^4-443 m3^6 mQ3^2 mU3^4+64 m3^4 mQ3^4 mU3^4-101 m3^6 mU3^6+105 m3^4 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(128 (6 m3^9+18 m3^7 mQ3^2-33 m3^5 mQ3^4-46 m3^7 mU3^2+32 m3^5 mQ3^2 mU3^2+32 m3^3 mQ3^4 mU3^2+33 m3^5 mU3^4-36 m3^3 mQ3^2 mU3^4-6 m3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(32 (394 m3^12-792 m3^10 mQ3^2+495 m3^8 mQ3^4-101 m3^6 mQ3^6-792 m3^10 mU3^2+1398 m3^8 mQ3^2 mU3^2-699 m3^6 mQ3^4 mU3^2+105 m3^4 mQ3^6 mU3^2+495 m3^8 mU3^4-699 m3^6 mQ3^2 mU3^4+192 m3^4 mQ3^4 mU3^4-101 m3^6 mU3^6+105 m3^4 mQ3^2 mU3^6) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(256 (28 m3^11-45 m3^9 mQ3^2-32 m3^7 mQ3^4+54 m3^5 mQ3^6-45 m3^9 mU3^2+176 m3^7 mQ3^2 mU3^2-84 m3^5 mQ3^4 mU3^2-54 m3^3 mQ3^6 mU3^2-32 m3^7 mU3^4-84 m3^5 mQ3^2 mU3^4+112 m3^3 mQ3^4 mU3^4+3 m3 mQ3^6 mU3^4+54 m3^5 mU3^6-54 m3^3 mQ3^2 mU3^6+3 m3 mQ3^4 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(16 (-88 m3^14-322 m3^12 mQ3^2+880 m3^10 mQ3^4-567 m3^8 mQ3^6+101 m3^6 mQ3^8-322 m3^12 mU3^2+1688 m3^10 mQ3^2 mU3^2-2221 m3^8 mQ3^4 mU3^2+952 m3^6 mQ3^6 mU3^2-105 m3^4 mQ3^8 mU3^2+880 m3^10 mU3^4-2221 m3^8 mQ3^2 mU3^4+1694 m3^6 mQ3^4 mU3^4-353 m3^4 mQ3^6 mU3^4-567 m3^8 mU3^6+952 m3^6 mQ3^2 mU3^6-353 m3^4 mQ3^4 mU3^6-24 m3^2 mQ3^6 mU3^6+101 m3^6 mU3^8-105 m3^4 mQ3^2 mU3^8) Xt^4)/(3 (-m3^2+mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(128 (16 m3^9+30 m3^7 mQ3^2-47 m3^5 mQ3^4+30 m3^7 mU3^2-78 m3^5 mQ3^2 mU3^2+48 m3^3 mQ3^4 mU3^2-47 m3^5 mU3^4+48 m3^3 mQ3^2 mU3^4) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[mU3^2/mQ3^2])+((32 (234 m3^8-401 m3^6 mQ3^2+163 m3^4 mQ3^4-401 m3^6 mU3^2+668 m3^4 mQ3^2 mU3^2-259 m3^2 mQ3^4 mU3^2+163 m3^4 mU3^4-259 m3^2 mQ3^2 mU3^4+92 mQ3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(256 m3^3 Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))+(-((32 m3^4 (210 m3^8-424 m3^6 mQ3^2+283 m3^4 mQ3^4-73 m3^2 mQ3^6-424 m3^6 mU3^2+718 m3^4 mQ3^2 mU3^2-359 m3^2 mQ3^4 mU3^2+77 mQ3^6 mU3^2+283 m3^4 mU3^4-359 m3^2 mQ3^2 mU3^4+64 mQ3^4 mU3^4-73 m3^2 mU3^6+77 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))+(256 m3^3 (32 m3^4-33 m3^2 mQ3^2-33 m3^2 mU3^2+34 mQ3^2 mU3^2) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Log[m3^2/mQ3^2]+(-((32 (32 m3^6 mQ3^2-96 m3^6 mU3^2+46 m3^4 mQ3^2 mU3^2+82 m3^4 mU3^4-123 m3^2 mQ3^2 mU3^4+59 m3^2 mU3^6+41 mQ3^2 mU3^6-41 mU3^8))/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)))-(256 (8 m3^5+16 m3^3 mU3^2-25 m3 mU3^4) Xt)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Log[mU3^2/mQ3^2]) Log[(vu^2 yt^2)/mQ3^2]+1312 Log[(vu^2 yt^2)/mQ3^2]^2+((1792 m3^4)/(3 (m3^2-mQ3^2)^2)-(7168 m3^3 Xt)/(3 (m3^2-mQ3^2) (mQ3^2-mU3^2))+(3584 (-2 m3^3+m3 mQ3^2+m3 mU3^2) Xt^3)/(3 (mQ3^2-mU3^2)^3)-(896 (2 m3^2-mQ3^2-mU3^2) Xt^4)/(3 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mQ3^2]+((1792 m3^4)/(3 (m3^2-mU3^2)^2)-(7168 m3^3 Xt)/(3 (m3^2-mU3^2) (-mQ3^2+mU3^2))-(3584 (-2 m3^3+m3 mQ3^2+m3 mU3^2) Xt^3)/(3 (mQ3^2-mU3^2)^3)-(896 (2 m3^2-mQ3^2-mU3^2) Xt^4)/(3 (-mQ3^2+mU3^2)^3)) PolyLog[2,1-m3^2/mU3^2]-128 Zeta[3])+Log[(vu^2 yt^2)/mQ3^2] ((8 (376 m3^8-8 m3^6 (79 mQ3^2+45 msq^2+79 mU3^2)-2 mQ3^2 mU3^2 (3 mQ3^4+mQ3^2 (30 msq^2-98 mU3^2)+3 mU3^2 (10 msq^2+mU3^2))+m3^4 (338 mQ3^4+300 msq^2 mU3^2+338 mU3^4+4 mQ3^2 (75 msq^2+227 mU3^2))-2 m3^2 (9 mQ3^6+9 mU3^4 (10 msq^2+mU3^2)+mQ3^4 (90 msq^2+211 mU3^2)+mQ3^2 (-120 msq^2 mU3^2+211 mU3^4))))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(-((1792 m3^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)))+(128 m3 mQ3^2)/((m3^2-mQ3^2) (m3^2-mU3^2))+(1280 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2))+(128 m3 mU3^2)/((m3^2-mQ3^2) (m3^2-mU3^2))) Xt+(8/3 ((2 (198 m3^16-726 m3^14 mQ3^2+721 m3^12 mQ3^4-232 m3^10 mQ3^6+27 m3^8 mQ3^8-726 m3^14 mU3^2+2700 m3^12 mQ3^2 mU3^2-2700 m3^10 mQ3^4 mU3^2+908 m3^8 mQ3^6 mU3^2-134 m3^6 mQ3^8 mU3^2+721 m3^12 mU3^4-2700 m3^10 mQ3^2 mU3^4+2240 m3^8 mQ3^4 mU3^4-320 m3^6 mQ3^6 mU3^4-13 m3^4 mQ3^8 mU3^4-232 m3^10 mU3^6+908 m3^8 mQ3^2 mU3^6-320 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6+27 m3^8 mU3^8-134 m3^6 mQ3^2 mU3^8-13 m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8))/((m3^2-mQ3^2)^4 (m3^2-mU3^2)^4)-(30 (m3^2-msq^2) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))+8/3 (-((240 m3 (m3^2-msq^2) (msq^2 mU3^2+mQ3^2 (msq^2-2 mU3^2)+m3^2 (mQ3^2-2 msq^2+mU3^2)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))-(16 m3^3 (m3^8+81 mQ3^4 mU3^4-50 m3^6 (mQ3^2+mU3^2)-130 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)+m3^4 (41 mQ3^4+196 mQ3^2 mU3^2+41 mU3^4)))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)) Xt-(512 m3^6 Xt^2)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Log[m3^2/mQ3^2]^2+(-((160 (10 m3^8-20 m3^6 mQ3^2+10 m3^4 mQ3^4+18 m3^6 msq^2-15 m3^4 mQ3^2 msq^2+9 m3^2 mQ3^4 msq^2-20 m3^6 mU3^2+40 m3^4 mQ3^2 mU3^2-20 m3^2 mQ3^4 mU3^2-15 m3^4 msq^2 mU3^2-12 m3^2 mQ3^2 msq^2 mU3^2+3 mQ3^4 msq^2 mU3^2+10 m3^4 mU3^4-20 m3^2 mQ3^2 mU3^4+10 mQ3^4 mU3^4+9 m3^2 msq^2 mU3^4+3 mQ3^2 msq^2 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(1280 m3 msq^2 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))) Log[msq^2/mQ3^2]-160 Log[msq^2/mQ3^2]^2+((16 (54 m3^10 mQ3^2-108 m3^8 mQ3^4+54 m3^6 mQ3^6-310 m3^10 mU3^2+582 m3^8 mQ3^2 mU3^2-296 m3^6 mQ3^4 mU3^2+21 m3^4 mQ3^6 mU3^2-9 m3^2 mQ3^8 mU3^2-90 m3^6 mQ3^2 msq^2 mU3^2+180 m3^4 mQ3^4 msq^2 mU3^2-90 m3^2 mQ3^6 msq^2 mU3^2+550 m3^8 mU3^4-1078 m3^6 mQ3^2 mU3^4+679 m3^4 mQ3^4 mU3^4-100 m3^2 mQ3^6 mU3^4-3 mQ3^8 mU3^4+90 m3^6 msq^2 mU3^4-210 m3^4 mQ3^2 msq^2 mU3^4+150 m3^2 mQ3^4 msq^2 mU3^4-30 mQ3^6 msq^2 mU3^4-216 m3^6 mU3^6+359 m3^4 mQ3^2 mU3^6-296 m3^2 mQ3^4 mU3^6+81 mQ3^6 mU3^6+30 m3^4 msq^2 mU3^6-60 m3^2 mQ3^2 msq^2 mU3^6+30 mQ3^4 msq^2 mU3^6-35 m3^4 mU3^8+158 m3^2 mQ3^2 mU3^8-75 mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(128 (16 m3^7-2 m3^5 (8 mQ3^2-5 mU3^2)-2 m3^3 mU3^2 (9 mQ3^2+15 msq^2+13 mU3^2)+m3 mU3^2 (3 mQ3^4+3 mU3^4+mQ3^2 (30 msq^2+28 mU3^2))) Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(-((160 msq^2 (2 m3^4-msq^2 mU3^2-3 m3^2 (msq^2-2 mU3^2)))/(m3^2-mU3^2)^3)+(1280 m3 msq^2 (msq^2-2 mU3^2) Xt)/((m3^2-mU3^2)^2 (mQ3^2-mU3^2))) Log[msq^2/mQ3^2]) Log[mU3^2/mQ3^2]+(8/3 (((mQ3^2-mU3^2) (-12 m3^10+14 m3^8 mQ3^2-34 m3^6 mQ3^4+29 m3^4 mQ3^6-9 m3^2 mQ3^8+14 m3^8 mU3^2+76 m3^6 mQ3^2 mU3^2-65 m3^4 mQ3^4 mU3^2+26 m3^2 mQ3^6 mU3^2-3 mQ3^8 mU3^2-34 m3^6 mU3^4-65 m3^4 mQ3^2 mU3^4+34 m3^2 mQ3^4 mU3^4-7 mQ3^6 mU3^4+29 m3^4 mU3^6+26 m3^2 mQ3^2 mU3^6-7 mQ3^4 mU3^6-9 m3^2 mU3^8-3 mQ3^2 mU3^8))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(-128 m3^10+m3^8 (-44 mQ3^2+556 mU3^2)+mU3^4 (-mQ3^2+mU3^2) (-3 mQ3^4-30 msq^4-4 mQ3^2 mU3^2+mU3^4)+2 m3^6 (mQ3^4-30 msq^2 mU3^2-353 mU3^4+mQ3^2 (30 msq^2-32 mU3^2))+2 m3^2 mU3^2 (3 mQ3^6-18 mQ3^4 mU3^2-30 msq^4 mU3^2+90 msq^2 mU3^4-68 mU3^6+mQ3^2 (30 msq^4-90 msq^2 mU3^2+19 mU3^4))+m3^4 (-9 mQ3^6+33 mQ3^4 mU3^2+90 msq^4 mU3^2-120 msq^2 mU3^4+437 mU3^6+mQ3^2 (-90 msq^4+120 msq^2 mU3^2+51 mU3^4)))/((m3^2-mU3^2)^4 (mQ3^2-mU3^2)))+8/3 ((8 (mQ3^2-mU3^2) (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(8 (-18 m3^7 (mQ3^2-mU3^2)+m3^5 (mQ3^4+68 mQ3^2 mU3^2-85 mU3^4)+m3^3 (3 mQ3^6-11 mQ3^4 mU3^2-30 msq^4 mU3^2+60 msq^2 mU3^4+57 mU3^6+mQ3^2 (30 msq^4-60 msq^2 mU3^2-17 mU3^4))-m3 mU3^2 (3 mQ3^6-10 mQ3^4 mU3^2+6 mU3^2 (-5 msq^4+10 msq^2 mU3^2+mU3^4)+mQ3^2 (30 msq^4-60 msq^2 mU3^2+17 mU3^4))))/((m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)) Xt-(512 m3^2 mU3^4 Xt^2)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)) Log[mU3^2/mQ3^2]^2+Log[m3^2/mQ3^2] (-((16 (660 m3^12+48 mQ3^6 mU3^6-2 m3^10 (781 mQ3^2+90 msq^2+781 mU3^2)+3 m3^2 mQ3^2 mU3^2 (mQ3^6-111 mQ3^2 mU3^4+10 msq^2 mU3^4+mU3^6+mQ3^4 (10 msq^2-111 mU3^2))+6 m3^8 (197 mQ3^4+40 msq^2 mU3^2+197 mU3^4+mQ3^2 (40 msq^2+594 mU3^2))-m3^6 (269 mQ3^6+270 msq^2 mU3^4+269 mU3^6+5 mQ3^4 (54 msq^2+515 mU3^2)-5 mQ3^2 (36 msq^2 mU3^2-515 mU3^4))+m3^4 (9 mQ3^8+9 mU3^6 (10 msq^2+mU3^2)+6 mQ3^6 (15 msq^2+83 mU3^2)+mQ3^4 (-90 msq^2 mU3^2+1822 mU3^4)+mQ3^2 (-90 msq^2 mU3^4+498 mU3^6))))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))+(128 m3^3 (40 m3^4+3 mQ3^4+3 mU3^2 (10 msq^2+mU3^2)-20 m3^2 (2 mQ3^2+3 msq^2+2 mU3^2)+mQ3^2 (30 msq^2+34 mU3^2)) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+((160 (4 m3^10 msq^2+6 m3^6 mQ3^4 msq^2-2 m3^4 mQ3^6 msq^2-6 m3^8 msq^4+8 m3^6 mQ3^2 msq^4-9 m3^4 mQ3^4 msq^4+3 m3^2 mQ3^6 msq^4-36 m3^6 mQ3^2 msq^2 mU3^2+18 m3^4 mQ3^4 msq^2 mU3^2-6 m3^2 mQ3^6 msq^2 mU3^2+8 m3^6 msq^4 mU3^2+6 m3^4 mQ3^2 msq^4 mU3^2-3 m3^2 mQ3^4 msq^4 mU3^2+mQ3^6 msq^4 mU3^2+6 m3^6 msq^2 mU3^4+18 m3^4 mQ3^2 msq^2 mU3^4-9 m3^4 msq^4 mU3^4-3 m3^2 mQ3^2 msq^4 mU3^4-2 m3^4 msq^2 mU3^6-6 m3^2 mQ3^2 msq^2 mU3^6+3 m3^2 msq^4 mU3^6+mQ3^2 msq^4 mU3^6))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(1280 (2 m3^5 msq^2-2 m3^3 msq^4+m3 mQ3^2 msq^4-2 m3 mQ3^2 msq^2 mU3^2+m3 msq^4 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Log[msq^2/mQ3^2]+((16 m3^4 (128 m3^12-410 m3^10 mQ3^2+472 m3^8 mQ3^4-205 m3^6 mQ3^6+27 m3^4 mQ3^8-358 m3^10 mU3^2+1314 m3^8 mQ3^2 mU3^2-1797 m3^6 mQ3^4 mU3^2+999 m3^4 mQ3^6 mU3^2-218 m3^2 mQ3^8 mU3^2+134 m3^8 mU3^4-683 m3^6 mQ3^2 mU3^4+1157 m3^4 mQ3^4 mU3^4-655 m3^2 mQ3^6 mU3^4+167 mQ3^8 mU3^4+125 m3^6 mU3^6-191 m3^4 mQ3^2 mU3^6-11 m3^2 mQ3^4 mU3^6-43 mQ3^6 mU3^6-72 m3^4 mU3^8+97 m3^2 mQ3^2 mU3^8+35 mQ3^4 mU3^8+19 m3^2 mU3^10-31 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2))-(128 m3^3 (34 m3^8+75 mQ3^4 mU3^4-7 mQ3^2 mU3^6-3 m3^6 (23 mQ3^2+39 mU3^2)+m3^4 (38 mQ3^4+229 mQ3^2 mU3^2+71 mU3^4)-m3^2 (129 mQ3^4 mU3^2+121 mQ3^2 mU3^4+4 mU3^6)) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(1024 m3^4 mU3^2 Xt^2)/((m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))) Log[mU3^2/mQ3^2])+((16 (128 m3^12-294 m3^10 mQ3^2+98 m3^8 mQ3^4+9 m3^6 mQ3^6+20 m3^4 mQ3^8-9 m3^2 mQ3^10-346 m3^10 mU3^2+902 m3^8 mQ3^2 mU3^2-251 m3^6 mQ3^4 mU3^2-151 m3^4 mQ3^6 mU3^2+41 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2+280 m3^8 mU3^4-1025 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4+41 m3^2 mQ3^6 mU3^4-mQ3^8 mU3^4-13 m3^6 mU3^6+391 m3^4 mQ3^2 mU3^6-167 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6-37 m3^4 mU3^8-34 m3^2 mQ3^2 mU3^8+23 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(128 m3 (18 m3^4+3 mQ3^4-7 mQ3^2 mU3^2+22 mU3^4+m3^2 (mQ3^2-37 mU3^2)) Xt)/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-m3^2/mQ3^2]+(-((160 (m3^2-msq^2) (2 m3^10+m3^8 (-8 mQ3^2+6 msq^2-8 mU3^2)-mQ3^2 msq^2 mU3^2 (mQ3^4+mU3^4)+m3^6 (3 mQ3^4-8 msq^2 mU3^2+3 mU3^4+mQ3^2 (-8 msq^2+30 mU3^2))-m3^4 (mQ3^6-9 msq^2 mU3^4+mU3^6+mQ3^4 (-9 msq^2+15 mU3^2)+3 mQ3^2 (2 msq^2 mU3^2+5 mU3^4))+m3^2 (3 mQ3^4 msq^2 mU3^2-3 msq^2 mU3^6+mQ3^6 (-3 msq^2+5 mU3^2)+mQ3^2 (3 msq^2 mU3^4+5 mU3^6))))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))-(1280 m3 (m3^2-msq^2) (msq^2 mU3^2+mQ3^2 (msq^2-2 mU3^2)+m3^2 (mQ3^2-2 msq^2+mU3^2)) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) PolyLog[2,1-msq^2/m3^2]+((160 (mQ3^2-msq^2) (-2 m3^4-3 m3^2 (mQ3^2-msq^2)+mQ3^2 (mQ3^2+msq^2)))/(m3^2-mQ3^2)^3+(1280 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-msq^2/mQ3^2]+(-((16 (128 m3^12-346 m3^10 mQ3^2+280 m3^8 mQ3^4-13 m3^6 mQ3^6-37 m3^4 mQ3^8-294 m3^10 mU3^2+902 m3^8 mQ3^2 mU3^2-1025 m3^6 mQ3^4 mU3^2+391 m3^4 mQ3^6 mU3^2-34 m3^2 mQ3^8 mU3^2+98 m3^8 mU3^4-251 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4-167 m3^2 mQ3^6 mU3^4+23 mQ3^8 mU3^4+9 m3^6 mU3^6-151 m3^4 mQ3^2 mU3^6+41 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6+20 m3^4 mU3^8+41 m3^2 mQ3^2 mU3^8-mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)))+(128 m3 (18 m3^4+22 mQ3^4-7 mQ3^2 mU3^2+3 mU3^4+m3^2 (-37 mQ3^2+mU3^2)) Xt)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-m3^2/mU3^2]+(-((160 (-msq^2+mU3^2) (2 m3^4-3 m3^2 msq^2+3 m3^2 mU3^2-msq^2 mU3^2-mU3^4))/(m3^2-mU3^2)^3)+(1280 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) PolyLog[2,1-msq^2/mU3^2]+(-((16 (mQ3^2-mU3^2) (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))+(128 (mQ3^2-mU3^2) (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) PolyLog[2,1-mU3^2/mQ3^2]+128 Zeta[3])))
);

(* Using gauge eigenstate masses *)

Mh2[DRratios]=(
MZ^2 Cos[2 beta]^2+vu^2 yt^4 Sin[beta]^2 (k ((12 xt^4)/(-1+x12^2)^2+(6+(12 xt^2)/(-1+x12^2)+(-(6/(-1+x12^2)^3)-(6 x12^2)/(-1+x12^2)^3) xt^4) Log[x12^2]-12 Log[xEWS^2])+g3^2 k^2 (-64+32 x32^2+(32 x32^2)/x12^2-(64 x32^2 xt^2)/x12^2-(512 x32 xt^3)/(-1+x12^2)^2+(192/(-1+x12^2)^2+(32 x32^2)/(-1+x12^2)^2+(32 x32^2)/(x12^2 (-1+x12^2)^2)) xt^4+(-((16 (x12^4-2 x12^2 x32^2+3 x32^4))/(x12^2-x32^2)^2)-(128 x32^3 xt)/((-1+x12^2) (x12^2-x32^2))-(32 (-1+3 x12^2) xt^2)/(-1+x12^2)^2-(64 (x32+x12^2 x32-2 x32^3) xt^3)/(-1+x12^2)^3+(32 (2 x12^2+2 x12^4-x32^2+x12^2 x32^2) xt^4)/(-1+x12^2)^4) Log[x12^2]^2+(-((32 (x12^2 x32^4+x12^6 x32^4+x32^6-2 x12^2 x32^6-2 x12^4 x32^6+x12^6 x32^6-2 x32^8+2 x12^2 x32^8-2 x12^4 x32^8+x32^10+x12^2 x32^10))/(x12^2 (x12^2-x32^2)^2 (-1+x32^2)^2))+(64 x32^3 xt)/((-1+x32^2) (-x12^2+x32^2))+(64 x32^4 (1+x12^2-x32^2) xt^2)/(x12^2 (x12^2-x32^2) (-1+x32^2))-(32 (1+x12^2) x32^2 xt^4)/(x12^2 (-1+x12^2)^2)) Log[x32^2]+Log[x12^2] ((32 (x12^4-2 x12^2 x32^2+3 x32^4))/(x12^2-x32^2)^2-(64 (x12^2 x32-2 x32^3) xt)/((-1+x12^2) (x12^2-x32^2))+(64 (2 x12^2-3 x32^2) xt^2)/((-1+x12^2) (x12^2-x32^2))+(128 (x32+3 x12^2 x32) xt^3)/(-1+x12^2)^3-(32 (3+7 x12^2+2 x32^2) xt^4)/(-1+x12^2)^3+((32 x32^4)/(-x12^2+x32^2)^2+(128 x32^3 xt)/((-1+x12^2) (x12^2-x32^2))-(256 x32^3 xt^3)/(-1+x12^2)^3) Log[x32^2])+Log[MR^2/mQ3^2] (-64+32 x32^2+(32 x32^2)/x12^2-(64 x32^2 xt^2)/x12^2-(256 x32 xt^3)/(-1+x12^2)^2+(128/(-1+x12^2)^2+(32 x32^2)/(-1+x12^2)^2+(32 x32^2)/(x12^2 (-1+x12^2)^2)) xt^4+(64-(128 x32 xt)/(-1+x12^2)+(128 xt^2)/(-1+x12^2)+(128 (x32+x12^2 x32) xt^3)/(-1+x12^2)^3-(64 (1+x12^2+x32^2) xt^4)/(-1+x12^2)^3) Log[x12^2]-128 Log[xEWS^2])+(-((32 x32^2)/((-1+x32^2) (-x12^2+x32^2)))-(32 x12^2 x32^2)/((-1+x32^2) (-x12^2+x32^2))+(64 x32^4)/((-1+x32^2) (-x12^2+x32^2))+(-((32 x12^4)/(x12^2-x32^2)^2)+(64 x12^2 x32^2)/(x12^2-x32^2)^2+(128 x12^2 x32 xt)/((-1+x12^2) (x12^2-x32^2))) Log[x12^2]+(-((32 (x32^4+x12^4 x32^4-2 x32^6-2 x12^2 x32^6+2 x32^8))/((-1+x32^2)^2 (-x12^2+x32^2)^2))+(128 x32^3 xt)/((-1+x32^2) (-x12^2+x32^2))) Log[x32^2]) Log[xEWS^2]+96 Log[xEWS^2]^2+(-((64 x32^4)/(-1+x32^2)^2)-(256 x32^3 xt)/((-1+x12^2) (-1+x32^2))+(128 (x32+x12^2 x32-2 x32^3) xt^3)/(-1+x12^2)^3+(32 (1+x12^2-2 x32^2) xt^4)/(-1+x12^2)^3) PolyLog[2,1-x32^2]+(-((64 x32^4)/(-x12^2+x32^2)^2)-(256 x32^3 xt)/((-1+x12^2) (x12^2-x32^2))-(128 (x32+x12^2 x32-2 x32^3) xt^3)/(-1+x12^2)^3-(32 (1+x12^2-2 x32^2) xt^4)/(-1+x12^2)^3) PolyLog[2,1-x32^2/x12^2])+g3^4 k^3 (dlatas2+SMConstatas2+(-24 x12^4-884 x12^6-24 x12^8-72 x12^2 x32^2+992 x12^4 x32^2+992 x12^6 x32^2-72 x12^8 x32^2+500 x12^2 x32^4+224 x12^4 x32^4+500 x12^6 x32^4-96 x32^6-2376 x12^2 x32^6-2376 x12^4 x32^6-96 x12^6 x32^6+448 x32^8+2620 x12^2 x32^8+448 x12^4 x32^8-352 x32^10-352 x12^2 x32^10-240 x12^4 xsq2^2-240 x12^6 xsq2^2-720 x12^2 x32^2 xsq2^2+960 x12^4 x32^2 xsq2^2-720 x12^6 x32^2 xsq2^2+1200 x12^2 x32^4 xsq2^2+1200 x12^4 x32^4 xsq2^2-1440 x12^2 x32^6 xsq2^2)/(3 x12^2 (x12^2-x32^2)^2 (-1+x32^2)^2)-(64 (-3 x32-3 x12^2 x32+14 x32^3-30 x32 xsq2^2) xt)/(3 (-1+x32^2) (-x12^2+x32^2))+(64 x32^2 (5 x12^2+3 x32^2+3 x12^2 x32^2-11 x32^4) xt^2)/(3 x12^2 (x12^2-x32^2) (-1+x32^2))+(512 x32 (5 x12^2+3 x32^2+3 x12^2 x32^2-11 x32^4) xt^3)/(3 (-1+x12^2)^2 (x12^2-x32^2) (-1+x32^2))+(8 (6 x12^4+277 x12^6+6 x12^8+18 x12^2 x32^2-236 x12^4 x32^2-236 x12^6 x32^2+18 x12^8 x32^2-173 x12^2 x32^4-168 x12^4 x32^4-173 x12^6 x32^4-12 x32^6+558 x12^2 x32^6+558 x12^4 x32^6-12 x12^6 x32^6+56 x32^8-455 x12^2 x32^8+56 x12^4 x32^8-44 x32^10-44 x12^2 x32^10+60 x12^4 xsq2^2+60 x12^6 xsq2^2+180 x12^2 x32^2 xsq2^2-240 x12^4 x32^2 xsq2^2+180 x12^6 x32^2 xsq2^2-300 x12^2 x32^4 xsq2^2-300 x12^4 x32^4 xsq2^2+360 x12^2 x32^6 xsq2^2) xt^4)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)+(128 (3 x32+3 x12^2 x32-14 x32^3+30 x32 xsq2^2) xt^5)/(3 (-1+x12^2)^2 (x12^2-x32^2) (-1+x32^2))+(1/3 ((4 (mQ3^2-mQ3^2 x12^2) (3 mQ3^10 x12^2+7 mQ3^10 x12^4+7 mQ3^10 x12^6+3 mQ3^10 x12^8+mQ3^2 (9 mQ3^8-26 mQ3^8 x12^2-34 mQ3^8 x12^4-26 mQ3^8 x12^6+9 mQ3^8 x12^8) x32^2+mQ3^4 (-29 mQ3^6+65 mQ3^6 x12^2+65 mQ3^6 x12^4-29 mQ3^6 x12^6) x32^4+mQ3^6 (34 mQ3^4-76 mQ3^4 x12^2+34 mQ3^4 x12^4) x32^6-14 mQ3^8 (mQ3^2+mQ3^2 x12^2) x32^8+12 mQ3^10 x32^10))/((-mQ3^2+mQ3^2 x32^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)-(4 (mQ3^8 (-38 mQ3^2+550 mQ3^2 x12^2) x32^8-128 mQ3^10 x32^10+mQ3^4 x12^4 (mQ3^2-mQ3^2 x12^2) (3 mQ3^4+4 mQ3^4 x12^2+67 mQ3^4 x12^4+30 mQ3^4 xsq2^4)+2 mQ3^6 x32^6 (mQ3^4-247 mQ3^4 x12^4-30 mQ3^4 x12^2 xsq2^2+6 mQ3^2 (-23 mQ3^2 x12^2+5 mQ3^2 xsq2^2))+mQ3^4 x32^4 (-9 mQ3^6+33 mQ3^6 x12^2+59 mQ3^6 x12^6-120 mQ3^6 x12^4 xsq2^2+90 mQ3^6 x12^2 xsq2^4+mQ3^2 (429 mQ3^4 x12^4+120 mQ3^4 x12^2 xsq2^2-90 mQ3^4 xsq2^4))+2 mQ3^4 x12^2 x32^2 (3 mQ3^6-18 mQ3^6 x12^2+68 mQ3^6 x12^6+90 mQ3^6 x12^4 xsq2^2-30 mQ3^6 x12^2 xsq2^4+3 mQ3^2 (-39 mQ3^4 x12^4-30 mQ3^4 x12^2 xsq2^2+10 mQ3^4 xsq2^4))))/((mQ3^2-mQ3^2 x12^2) (-mQ3^2 x12^2+mQ3^2 x32^2)^4))+1/3 mQ3 (-((32 mQ3 (mQ3^2-mQ3^2 x12^2) x32 (3 mQ3^4-mQ3^4 x12^2+3 mQ3^4 x12^4-5 mQ3^2 (mQ3^2+mQ3^2 x12^2) x32^2+5 mQ3^4 x32^4))/((-mQ3^2+mQ3^2 x32^2)^2 (-mQ3^2 x12^2+mQ3^2 x32^2)^2))-(32 (-mQ3^2+mQ3^2 x12^2)^3 (mQ3^5 (mQ3^4+185 mQ3^4 x12^2-202 mQ3^4 x12^4) x32^5-20 mQ3^7 (mQ3^2-mQ3^2 x12^2) x32^7+mQ3^3 x12^2 x32 (-3 mQ3^6+10 mQ3^6 x12^2-55 mQ3^6 x12^6-60 mQ3^6 x12^4 xsq2^2+30 mQ3^6 x12^2 xsq2^4+mQ3^2 (32 mQ3^4 x12^4+60 mQ3^4 x12^2 xsq2^2-30 mQ3^4 xsq2^4))+mQ3^3 x32^3 (3 mQ3^6-11 mQ3^6 x12^2+189 mQ3^6 x12^6+60 mQ3^6 x12^4 xsq2^2-30 mQ3^6 x12^2 xsq2^4+mQ3^2 (-149 mQ3^4 x12^4-60 mQ3^4 x12^2 xsq2^2+30 mQ3^4 xsq2^4))))/((mQ3^2-mQ3^2 x12^2)^5 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)) xt+1/3 mQ3^2 (-((8 (3 mQ3^10 x12^2+7 mQ3^10 x12^4+7 mQ3^10 x12^6+3 mQ3^10 x12^8+mQ3^2 (9 mQ3^8-26 mQ3^8 x12^2-34 mQ3^8 x12^4-26 mQ3^8 x12^6+9 mQ3^8 x12^8) x32^2+mQ3^4 (-29 mQ3^6+65 mQ3^6 x12^2+65 mQ3^6 x12^4-29 mQ3^6 x12^6) x32^4+mQ3^6 (34 mQ3^4-76 mQ3^4 x12^2+34 mQ3^4 x12^4) x32^6-14 mQ3^8 (mQ3^2+mQ3^2 x12^2) x32^8+12 mQ3^10 x32^10))/((-mQ3^2+mQ3^2 x32^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3))+(8 (-6 mQ3^8 (7 mQ3^2-177 mQ3^2 x12^2) x32^8-128 mQ3^10 x32^10+2 mQ3^6 x32^6 (mQ3^4+10 mQ3^2 (-7 mQ3^2 x12^2+3 mQ3^2 xsq2^2)-5 mQ3^2 x12^2 (167 mQ3^2 x12^2+6 mQ3^2 xsq2^2))+2 mQ3^4 x12^2 x32^2 (3 mQ3^6-18 mQ3^6 x12^2+124 mQ3^6 x12^6+90 mQ3^6 x12^4 xsq2^2-30 mQ3^6 x12^2 xsq2^4+3 mQ3^2 (-39 mQ3^4 x12^4-30 mQ3^4 x12^2 xsq2^2+10 mQ3^4 xsq2^4))+mQ3^4 x12^4 (3 mQ3^6+mQ3^6 x12^2+mQ3^2 (63 mQ3^4 x12^4+30 mQ3^4 xsq2^4)-15 (9 mQ3^6 x12^6+2 mQ3^6 x12^2 xsq2^4))+mQ3^4 x32^4 (-9 mQ3^6+33 mQ3^6 x12^2+mQ3^2 (361 mQ3^4 x12^4+120 mQ3^4 x12^2 xsq2^2-90 mQ3^4 xsq2^4)+15 (41 mQ3^6 x12^6-8 mQ3^6 x12^4 xsq2^2+6 mQ3^6 x12^2 xsq2^4))))/((mQ3^2-mQ3^2 x12^2)^2 (-mQ3^2 x12^2+mQ3^2 x32^2)^4)) xt^2+1/3 mQ3^3 ((64 mQ3 x32 (3 mQ3^4-mQ3^4 x12^2+3 mQ3^4 x12^4-5 mQ3^2 (mQ3^2+mQ3^2 x12^2) x32^2+5 mQ3^4 x32^4))/((-mQ3^2+mQ3^2 x32^2)^2 (-mQ3^2 x12^2+mQ3^2 x32^2)^2)-(32 (mQ3^5 (2 mQ3^4+205 mQ3^4 x12^2-131 mQ3^4 x12^4) x32^5-mQ3^7 (37 mQ3^2+35 mQ3^2 x12^2) x32^7+2 mQ3^9 x32^9+mQ3^3 x12^2 x32 (-6 mQ3^6+20 mQ3^6 x12^2-157 mQ3^6 x12^6-120 mQ3^6 x12^4 xsq2^2+60 mQ3^6 x12^2 xsq2^4+mQ3^2 (81 mQ3^4 x12^4+120 mQ3^4 x12^2 xsq2^2-60 mQ3^4 xsq2^4))+mQ3^3 x32^3 (6 mQ3^6-22 mQ3^6 x12^2+353 mQ3^6 x12^6+120 mQ3^6 x12^4 xsq2^2-60 mQ3^6 x12^2 xsq2^4+mQ3^2 (-281 mQ3^4 x12^4-120 mQ3^4 x12^2 xsq2^2+60 mQ3^4 xsq2^4))))/((mQ3^2-mQ3^2 x12^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)) xt^3+1/3 mQ3^4 ((4 (mQ3^2+mQ3^2 x12^2) (3 mQ3^10 x12^2+7 mQ3^10 x12^4+7 mQ3^10 x12^6+3 mQ3^10 x12^8+mQ3^2 (9 mQ3^8-26 mQ3^8 x12^2-34 mQ3^8 x12^4-26 mQ3^8 x12^6+9 mQ3^8 x12^8) x32^2+mQ3^4 (-29 mQ3^6+65 mQ3^6 x12^2+65 mQ3^6 x12^4-29 mQ3^6 x12^6) x32^4+mQ3^6 (34 mQ3^4-76 mQ3^4 x12^2+34 mQ3^4 x12^4) x32^6-14 mQ3^8 (mQ3^2+mQ3^2 x12^2) x32^8+12 mQ3^10 x32^10))/((mQ3^2-mQ3^2 x12^2)^2 (-mQ3^2+mQ3^2 x32^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)+(4 (mQ3^8 (44 mQ3^4-872 mQ3^4 x12^2-4276 mQ3^4 x12^4) x32^8+4 mQ3^10 (31 mQ3^2+289 mQ3^2 x12^2) x32^10+mQ3^6 x32^6 (-2 mQ3^6+2314 mQ3^6 x12^4+4726 mQ3^6 x12^6+60 mQ3^6 x12^4 xsq2^2+mQ3^4 (66 mQ3^2 x12^2-60 mQ3^2 xsq2^2))+mQ3^4 x32^4 (9 mQ3^8-24 mQ3^8 x12^2-2344 mQ3^8 x12^6-1163 mQ3^8 x12^8+120 mQ3^8 x12^6 xsq2^2-90 mQ3^8 x12^4 xsq2^4+6 mQ3^4 (-37 mQ3^4 x12^4-20 mQ3^4 x12^2 xsq2^2+15 mQ3^4 xsq2^4))-mQ3^4 x12^4 (mQ3^2+mQ3^2 x12^2) (3 mQ3^6+mQ3^6 x12^2-169 mQ3^6 x12^6-30 mQ3^6 x12^2 xsq2^4+mQ3^2 (29 mQ3^4 x12^4+30 mQ3^4 xsq2^4))-2 mQ3^4 x12^2 x32^2 (3 mQ3^8-15 mQ3^8 x12^2-319 mQ3^8 x12^6+302 mQ3^8 x12^8+90 mQ3^8 x12^6 xsq2^2-30 mQ3^8 x12^4 xsq2^4+mQ3^4 (-67 mQ3^4 x12^4-90 mQ3^4 x12^2 xsq2^2+30 mQ3^4 xsq2^4))))/((mQ3^2-mQ3^2 x12^2)^4 (-mQ3^2 x12^2+mQ3^2 x32^2)^4)) xt^4+1/3 mQ3^5 (-((32 mQ3 (mQ3^2+mQ3^2 x12^2) x32 (3 mQ3^4-mQ3^4 x12^2+3 mQ3^4 x12^4-5 mQ3^2 (mQ3^2+mQ3^2 x12^2) x32^2+5 mQ3^4 x32^4))/((mQ3^2-mQ3^2 x12^2)^2 (-mQ3^2+mQ3^2 x32^2)^2 (-mQ3^2 x12^2+mQ3^2 x32^2)^2))-(32 (-mQ3^5 (mQ3^6+70 mQ3^6 x12^2-145 mQ3^6 x12^4-86 mQ3^6 x12^6) x32^5+2 mQ3^7 (9 mQ3^4-16 mQ3^4 x12^2+7 mQ3^4 x12^4) x32^7+mQ3^3 x32^3 (-3 mQ3^8+8 mQ3^8 x12^2-200 mQ3^8 x12^6-219 mQ3^8 x12^8-60 mQ3^8 x12^6 xsq2^2+30 mQ3^8 x12^4 xsq2^4+mQ3^4 (94 mQ3^4 x12^4+60 mQ3^4 x12^2 xsq2^2-30 mQ3^4 xsq2^4))+mQ3^3 x12^2 (mQ3^2+mQ3^2 x12^2) x32 (3 mQ3^6-10 mQ3^6 x12^2+103 mQ3^6 x12^6+60 mQ3^6 x12^4 xsq2^2-30 mQ3^6 x12^2 xsq2^4+2 mQ3^2 (-8 mQ3^4 x12^4-30 mQ3^4 x12^2 xsq2^2+15 mQ3^4 xsq2^4))))/((mQ3^2-mQ3^2 x12^2)^5 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)) xt^5+(1280 x12^4 (1+x12^2) x32^2 xt^6)/(3 (-1+x12^2)^5 (x12^2-x32^2)^2)) Log[x12^2]^3+(1/3 (-((8 (36 mQ3^20 x12^10-144 mQ3^18 x12^8 (mQ3^2+mQ3^2 x12^2) x32^2+mQ3^16 x12^6 (25 mQ3^4+452 mQ3^4 x12^2+25 mQ3^4 x12^4) x32^4+2 mQ3^14 x12^4 (67 mQ3^6+136 mQ3^6 x12^2+136 mQ3^6 x12^4+67 mQ3^6 x12^6) x32^6-mQ3^12 x12^2 (39 mQ3^8+908 mQ3^8 x12^2+2048 mQ3^8 x12^4+908 mQ3^8 x12^6+39 mQ3^8 x12^8) x32^8+4 mQ3^10 (14 mQ3^10+70 mQ3^10 x12^2+647 mQ3^10 x12^4+647 mQ3^10 x12^6+70 mQ3^10 x12^8+14 mQ3^10 x12^10) x32^10-mQ3^12 (224 mQ3^8+797 mQ3^8 x12^2+2636 mQ3^8 x12^4+797 mQ3^8 x12^6+224 mQ3^8 x12^8) x32^12+mQ3^14 (368 mQ3^6+838 mQ3^6 x12^2+838 mQ3^6 x12^4+368 mQ3^6 x12^6) x32^14-2 mQ3^16 (144 mQ3^4+151 mQ3^4 x12^2+144 mQ3^4 x12^4) x32^16+88 mQ3^18 (mQ3^2+mQ3^2 x12^2) x32^18))/(mQ3^4 x12^2 (-mQ3^2+mQ3^2 x32^2)^4 (-mQ3^2 x12^2+mQ3^2 x32^2)^4))-(120 (mQ3^4-2 mQ3^4 x12^2+mQ3^4 x12^4) (mQ3^2 x32^2-mQ3^2 xsq2^2) (2 mQ3^10 x32^10-mQ3^6 x12^2 (mQ3^4+mQ3^4 x12^4) xsq2^2+mQ3^8 x32^8 (-8 mQ3^2-8 mQ3^2 x12^2+6 mQ3^2 xsq2^2)+mQ3^6 x32^6 (3 mQ3^4+3 mQ3^4 x12^4-8 mQ3^4 x12^2 xsq2^2+mQ3^2 (30 mQ3^2 x12^2-8 mQ3^2 xsq2^2))-mQ3^4 x32^4 (mQ3^6+mQ3^6 x12^6-9 mQ3^6 x12^4 xsq2^2+mQ3^4 (15 mQ3^2 x12^2-9 mQ3^2 xsq2^2)+3 mQ3^2 (5 mQ3^4 x12^4+2 mQ3^4 x12^2 xsq2^2))+mQ3^2 x32^2 (3 mQ3^8 x12^2 xsq2^2-3 mQ3^8 x12^6 xsq2^2+mQ3^6 (5 mQ3^2 x12^2-3 mQ3^2 xsq2^2)+mQ3^2 (5 mQ3^6 x12^6+3 mQ3^6 x12^4 xsq2^2))))/((mQ3^2-mQ3^2 x12^2)^2 (-mQ3^2+mQ3^2 x32^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3))-(64 (-51 x12^6 x32^3+85 x12^4 x32^5+85 x12^6 x32^5-10 x12^2 x32^7-136 x12^4 x32^7-10 x12^6 x32^7+16 x32^9+3 x12^2 x32^9+3 x12^4 x32^9+16 x12^6 x32^9-32 x32^11+31 x12^2 x32^11-32 x12^4 x32^11+16 x32^13+16 x12^2 x32^13-30 x12^6 x32 xsq2^2+30 x12^4 x32^3 xsq2^2+30 x12^6 x32^3 xsq2^2-30 x12^2 x32^7 xsq2^2-30 x12^4 x32^7 xsq2^2+30 x12^2 x32^9 xsq2^2+15 x12^4 x32 xsq2^4+15 x12^6 x32 xsq2^4-15 x12^2 x32^3 xsq2^4-60 x12^4 x32^3 xsq2^4-15 x12^6 x32^3 xsq2^4+45 x12^2 x32^5 xsq2^4+45 x12^4 x32^5 xsq2^4-30 x12^2 x32^7 xsq2^4) xt)/(3 x12^2 (x12^2-x32^2)^3 (-1+x32^2)^3)+(128 (3 x12^4 x32^4+3 x12^6 x32^4-6 x12^2 x32^6-x12^4 x32^6-6 x12^6 x32^6+7 x32^8+11 x12^2 x32^8+11 x12^4 x32^8+7 x12^6 x32^8-21 x32^10-26 x12^2 x32^10-21 x12^4 x32^10+25 x32^12+25 x12^2 x32^12-11 x32^14) xt^2)/(3 x12^2 (x12^2-x32^2)^3 (-1+x32^2)^3)+(2048 x32^7 (1+x12^2-x32^2) xt^3)/(3 x12^2 (x12^2-x32^2)^2 (-1+x32^2)^2)+1/3 mQ3^4 ((16 (36 mQ3^20 x12^10-156 mQ3^18 x12^8 (mQ3^2+mQ3^2 x12^2) x32^2+mQ3^16 x12^6 (49 mQ3^4+548 mQ3^4 x12^2+49 mQ3^4 x12^4) x32^4+2 mQ3^14 x12^4 (35 mQ3^6+44 mQ3^6 x12^2+44 mQ3^6 x12^4+35 mQ3^6 x12^6) x32^6+mQ3^12 x12^2 (25 mQ3^8-572 mQ3^8 x12^2-1712 mQ3^8 x12^4-572 mQ3^8 x12^6+25 mQ3^8 x12^8) x32^8-4 mQ3^10 (7 mQ3^10+mQ3^10 x12^2-503 mQ3^10 x12^4-503 mQ3^10 x12^6+mQ3^10 x12^8+7 mQ3^10 x12^10) x32^10+mQ3^12 (112 mQ3^8-277 mQ3^8 x12^2-1932 mQ3^8 x12^4-277 mQ3^8 x12^6+112 mQ3^8 x12^8) x32^12-2 mQ3^14 (92 mQ3^6-147 mQ3^6 x12^2-147 mQ3^6 x12^4+92 mQ3^6 x12^6) x32^14+2 mQ3^16 (72 mQ3^4+mQ3^4 x12^2+72 mQ3^4 x12^4) x32^16-44 mQ3^18 (mQ3^2+mQ3^2 x12^2) x32^18))/(mQ3^4 x12^2 (mQ3^2-mQ3^2 x12^2)^2 (-mQ3^2+mQ3^2 x32^2)^4 (-mQ3^2 x12^2+mQ3^2 x32^2)^4)+(240 (mQ3^2 x32^2-mQ3^2 xsq2^2) (2 mQ3^10 x32^10-mQ3^6 x12^2 (mQ3^4+mQ3^4 x12^4) xsq2^2+mQ3^8 x32^8 (-8 mQ3^2-8 mQ3^2 x12^2+6 mQ3^2 xsq2^2)+mQ3^6 x32^6 (3 mQ3^4+3 mQ3^4 x12^4-8 mQ3^4 x12^2 xsq2^2+mQ3^2 (30 mQ3^2 x12^2-8 mQ3^2 xsq2^2))-mQ3^4 x32^4 (mQ3^6+mQ3^6 x12^6-9 mQ3^6 x12^4 xsq2^2+mQ3^4 (15 mQ3^2 x12^2-9 mQ3^2 xsq2^2)+3 mQ3^2 (5 mQ3^4 x12^4+2 mQ3^4 x12^2 xsq2^2))+mQ3^2 x32^2 (3 mQ3^8 x12^2 xsq2^2-3 mQ3^8 x12^6 xsq2^2+mQ3^6 (5 mQ3^2 x12^2-3 mQ3^2 xsq2^2)+mQ3^2 (5 mQ3^6 x12^6+3 mQ3^6 x12^4 xsq2^2))))/((mQ3^2-mQ3^2 x12^2)^2 (-mQ3^2+mQ3^2 x32^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)) xt^4-(128 (57 x12^6 x32^3-89 x12^4 x32^5-89 x12^6 x32^5+32 x12^2 x32^7+128 x12^4 x32^7+32 x12^6 x32^7+8 x32^9-39 x12^2 x32^9-39 x12^4 x32^9+8 x12^6 x32^9-16 x32^11+7 x12^2 x32^11-16 x12^4 x32^11+8 x32^13+8 x12^2 x32^13+30 x12^6 x32 xsq2^2-30 x12^4 x32^3 xsq2^2-30 x12^6 x32^3 xsq2^2+30 x12^2 x32^7 xsq2^2+30 x12^4 x32^7 xsq2^2-30 x12^2 x32^9 xsq2^2-15 x12^4 x32 xsq2^4-15 x12^6 x32 xsq2^4+15 x12^2 x32^3 xsq2^4+60 x12^4 x32^3 xsq2^4+15 x12^6 x32^3 xsq2^4-45 x12^2 x32^5 xsq2^4-45 x12^4 x32^5 xsq2^4+30 x12^2 x32^7 xsq2^4) xt^5)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^3 (-1+x32^2)^3)-(2560 x32^6 xt^6)/(3 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)) Log[x32^2]^2+Log[MR^2/mQ3^2]^2 (3760/3-(896 x32^2)/3-(896 x32^2)/(3 x12^2)+(1792 x32^2 xt^2)/(3 x12^2)+(7168 x32 xt^3)/(3 (-1+x12^2)^2)+(-(4000/(3 (-1+x12^2)^2))-(896 x32^2)/(3 (-1+x12^2)^2)-(896 x32^2)/(3 x12^2 (-1+x12^2)^2)) xt^4+(-(2000/3)+(3584 x32 xt)/(3 (-1+x12^2))-(4000 xt^2)/(3 (-1+x12^2))-(3584 (x32+x12^2 x32) xt^3)/(3 (-1+x12^2)^3)-(16 (-125-125 x12^2-112 x32^2) xt^4)/(3 (-1+x12^2)^3)) Log[x12^2]-(2624 Log[xEWS^2])/3)-736 Log[xEWS^2]^3+(-80+(160 xt^4)/(-1+x12^2)^2) Log[xsq2^2]^2+Log[xEWS^2]^2 (-((64 (-5 x12^2+x32^2+x12^2 x32^2+3 x32^4))/((-1+x32^2) (-x12^2+x32^2)))+((16 (17 x12^4-34 x12^2 x32^2+x32^4))/(x12^2-x32^2)^2-(1024 x12^2 x32 xt)/((-1+x12^2) (x12^2-x32^2))) Log[x12^2]+((64 (3 x12^4-6 x12^2 x32^2-6 x12^4 x32^2+7 x32^4+12 x12^2 x32^4+7 x12^4 x32^4-14 x32^6-14 x12^2 x32^6+11 x32^8))/((-1+x32^2)^2 (-x12^2+x32^2)^2)-(1024 x32^3 xt)/((-1+x32^2) (-x12^2+x32^2))) Log[x32^2]+160 Log[xsq2^2])+Log[x12^2]^2 (-((4 (6 x12^6-303 x12^8+287 x12^10+7 x12^12+3 x12^14+12 x12^4 x32^2+1018 x12^6 x32^2+603 x12^8 x32^2-967 x12^10 x32^2-35 x12^12 x32^2+9 x12^14 x32^2-18 x12^2 x32^4-1394 x12^4 x32^4-4070 x12^6 x32^4+410 x12^8 x32^4+1160 x12^10 x32^4-56 x12^12 x32^4+714 x12^2 x32^6+6198 x12^4 x32^6+5326 x12^6 x32^6-1642 x12^8 x32^6-228 x12^10 x32^6+101 x32^8-3481 x12^2 x32^8-9653 x12^4 x32^8-2199 x12^6 x32^8+512 x12^8 x32^8-81 x32^10+5977 x12^2 x32^10+5925 x12^4 x32^10+339 x12^6 x32^10-284 x32^12-4136 x12^2 x32^12-1340 x12^4 x32^12+380 x32^14+1028 x12^2 x32^14-128 x32^16+60 x12^6 xsq2^2-60 x12^8 xsq2^2+300 x12^4 x32^2 xsq2^2-480 x12^6 x32^2 xsq2^2+180 x12^8 x32^2 xsq2^2-300 x12^2 x32^4 xsq2^2-600 x12^4 x32^4 xsq2^2+1080 x12^6 x32^4 xsq2^2-180 x12^8 x32^4 xsq2^2-60 x32^6 xsq2^2+960 x12^2 x32^6 xsq2^2-960 x12^6 x32^6 xsq2^2+60 x12^8 x32^6 xsq2^2+180 x32^8 xsq2^2-1080 x12^2 x32^8 xsq2^2+600 x12^4 x32^8 xsq2^2+300 x12^6 x32^8 xsq2^2-180 x32^10 xsq2^2+480 x12^2 x32^10 xsq2^2-300 x12^4 x32^10 xsq2^2+60 x32^12 xsq2^2-60 x12^2 x32^12 xsq2^2-30 x12^4 xsq2^4+30 x12^6 xsq2^4-60 x12^2 x32^2 xsq2^4+150 x12^4 x32^2 xsq2^4-90 x12^6 x32^2 xsq2^4+90 x32^4 xsq2^4+90 x12^2 x32^4 xsq2^4-270 x12^4 x32^4 xsq2^4+90 x12^6 x32^4 xsq2^4-270 x32^6 xsq2^4+90 x12^2 x32^6 xsq2^4+210 x12^4 x32^6 xsq2^4-30 x12^6 x32^6 xsq2^4+270 x32^8 xsq2^4-210 x12^2 x32^8 xsq2^4-60 x12^4 x32^8 xsq2^4-90 x32^10 xsq2^4+90 x12^2 x32^10 xsq2^4))/(3 (-1+x12^2) (x12^2-x32^2)^4 (-1+x32^2)^3))-(32 (-6 x12^4 x32-152 x12^6 x32+175 x12^8 x32-4 x12^10 x32+3 x12^12 x32+6 x12^2 x32^3+388 x12^4 x32^3-104 x12^6 x32^3-340 x12^8 x32^3-14 x12^10 x32^3-320 x12^2 x32^5-474 x12^4 x32^5+680 x12^6 x32^5+210 x12^8 x32^5+4 x32^7+676 x12^2 x32^7-260 x12^4 x32^7-484 x12^6 x32^7-33 x32^9-348 x12^2 x32^9+397 x12^4 x32^9+32 x32^11-32 x12^2 x32^11-120 x12^4 x32 xsq2^2+120 x12^6 x32 xsq2^2+120 x12^2 x32^3 xsq2^2+120 x12^4 x32^3 xsq2^2-240 x12^6 x32^3 xsq2^2-240 x12^2 x32^5 xsq2^2+120 x12^4 x32^5 xsq2^2+120 x12^6 x32^5 xsq2^2+120 x12^2 x32^7 xsq2^2-120 x12^4 x32^7 xsq2^2+30 x12^2 x32 xsq2^4-30 x12^4 x32 xsq2^4-30 x32^3 xsq2^4-30 x12^2 x32^3 xsq2^4+60 x12^4 x32^3 xsq2^4+60 x32^5 xsq2^4-30 x12^2 x32^5 xsq2^4-30 x12^4 x32^5 xsq2^4-30 x32^7 xsq2^4+30 x12^2 x32^7 xsq2^4) xt)/(3 (-1+x12^2)^2 (x12^2-x32^2)^3 (-1+x32^2)^2)+(16 (3 x12^4-177 x12^6+151 x12^8+3 x12^10+9 x12^2 x32^2+471 x12^4 x32^2+338 x12^6 x32^2-391 x12^8 x32^2+9 x12^10 x32^2-373 x12^2 x32^4-1612 x12^4 x32^4+137 x12^6 x32^4+192 x12^8 x32^4-41 x32^6+1346 x12^2 x32^6+1473 x12^4 x32^6-226 x12^6 x32^6+65 x32^8-1462 x12^2 x32^8-383 x12^4 x32^8-24 x32^10+492 x12^2 x32^10+30 x12^4 xsq2^2-30 x12^6 xsq2^2+90 x12^2 x32^2 xsq2^2-150 x12^4 x32^2 xsq2^2+60 x12^6 x32^2 xsq2^2-180 x12^2 x32^4 xsq2^2+210 x12^4 x32^4 xsq2^2-30 x12^6 x32^4 xsq2^2+90 x12^2 x32^6 xsq2^2-90 x12^4 x32^6 xsq2^2) xt^2)/(3 (-1+x12^2)^2 (x12^2-x32^2)^3 (-1+x32^2)^2)+(128 (-3 x12^2 x32-82 x12^4 x32+53 x12^6 x32+3 x12^8 x32+117 x12^2 x32^3+47 x12^4 x32^3-66 x12^6 x32^3+14 x32^5-150 x12^2 x32^5+48 x12^4 x32^5-7 x32^7+37 x12^2 x32^7-11 x32^9-30 x12^2 x32 xsq2^2+30 x12^4 x32 xsq2^2+30 x12^2 x32^3 xsq2^2-30 x12^4 x32^3 xsq2^2) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2))+(8 (3 x12^6-185 x12^8-244 x12^10+342 x12^12+x12^14+3 x12^16+6 x12^4 x32^2+645 x12^6 x32^2+2062 x12^8 x32^2+1253 x12^10 x32^2-1061 x12^12 x32^2-50 x12^14 x32^2+9 x12^16 x32^2-9 x12^2 x32^4-731 x12^4 x32^4-4746 x12^6 x32^4-8973 x12^8 x32^4-2382 x12^10 x32^4+1384 x12^12 x32^4-47 x12^14 x32^4+275 x12^2 x32^6+3855 x12^4 x32^6+15256 x12^6 x32^6+17186 x12^8 x32^6+1117 x12^10 x32^6-329 x12^12 x32^6+4 x32^8-816 x12^2 x32^8-8970 x12^4 x32^8-25376 x12^6 x32^8-13306 x12^8 x32^8-416 x12^10 x32^8-151 x32^10+691 x12^2 x32^10+12407 x12^4 x32^10+19237 x12^6 x32^10+4056 x12^8 x32^10+481 x32^12-357 x12^2 x32^12-8817 x12^4 x32^12-5691 x12^6 x32^12-498 x32^14+296 x12^2 x32^14+2586 x12^4 x32^14+176 x32^16-176 x12^2 x32^16+30 x12^6 xsq2^2-30 x12^10 xsq2^2-120 x12^4 x32^2 xsq2^2+270 x12^6 x32^2 xsq2^2-240 x12^8 x32^2 xsq2^2+90 x12^10 x32^2 xsq2^2+30 x12^2 x32^4 xsq2^2+120 x12^4 x32^4 xsq2^2-780 x12^6 x32^4 xsq2^2+720 x12^8 x32^4 xsq2^2-90 x12^10 x32^4 xsq2^2+60 x32^6 xsq2^2-210 x12^2 x32^6 xsq2^2+420 x12^4 x32^6 xsq2^2+420 x12^6 x32^6 xsq2^2-720 x12^8 x32^6 xsq2^2+30 x12^10 x32^6 xsq2^2-180 x32^8 xsq2^2+450 x12^2 x32^8 xsq2^2-780 x12^4 x32^8 xsq2^2+270 x12^6 x32^8 xsq2^2+240 x12^8 x32^8 xsq2^2+180 x32^10 xsq2^2-390 x12^2 x32^10 xsq2^2+420 x12^4 x32^10 xsq2^2-210 x12^6 x32^10 xsq2^2-60 x32^12 xsq2^2+120 x12^2 x32^12 xsq2^2-60 x12^4 x32^12 xsq2^2+30 x12^4 xsq2^4-60 x12^6 xsq2^4+30 x12^8 xsq2^4+60 x12^2 x32^2 xsq2^4-210 x12^4 x32^2 xsq2^4+240 x12^6 x32^2 xsq2^4-90 x12^8 x32^2 xsq2^4-90 x32^4 xsq2^4+360 x12^4 x32^4 xsq2^4-360 x12^6 x32^4 xsq2^4+90 x12^8 x32^4 xsq2^4+270 x32^6 xsq2^4-360 x12^2 x32^6 xsq2^4-120 x12^4 x32^6 xsq2^4+240 x12^6 x32^6 xsq2^4-30 x12^8 x32^6 xsq2^4-270 x32^8 xsq2^4+480 x12^2 x32^8 xsq2^4-150 x12^4 x32^8 xsq2^4-60 x12^6 x32^8 xsq2^4+90 x32^10 xsq2^4-180 x12^2 x32^10 xsq2^4+90 x12^4 x32^10 xsq2^4) xt^4)/(3 (-1+x12^2)^4 (x12^2-x32^2)^4 (-1+x32^2)^3)+(64 (3 x12^4 x32+67 x12^6 x32+216 x12^8 x32-7 x12^10 x32+3 x12^12 x32-3 x12^2 x32^3-94 x12^4 x32^3-608 x12^6 x32^3-396 x12^8 x32^3-11 x12^10 x32^3+5 x12^2 x32^5+427 x12^4 x32^5+923 x12^6 x32^5+225 x12^8 x32^5+38 x32^7-14 x12^2 x32^7-470 x12^4 x32^7-442 x12^6 x32^7-69 x32^9-36 x12^2 x32^9+179 x12^4 x32^9+34 x32^11+30 x12^2 x32^11-30 x12^4 x32 xsq2^2+90 x12^6 x32 xsq2^2+30 x12^2 x32^3 xsq2^2-30 x12^4 x32^3 xsq2^2-180 x12^6 x32^3 xsq2^2-60 x12^2 x32^5 xsq2^2+150 x12^4 x32^5 xsq2^2+90 x12^6 x32^5 xsq2^2+30 x12^2 x32^7 xsq2^2-90 x12^4 x32^7 xsq2^2+30 x12^2 x32 xsq2^4-30 x12^4 x32 xsq2^4-30 x32^3 xsq2^4-30 x12^2 x32^3 xsq2^4+60 x12^4 x32^3 xsq2^4+60 x32^5 xsq2^4-30 x12^2 x32^5 xsq2^4-30 x12^4 x32^5 xsq2^4-30 x32^7 xsq2^4+30 x12^2 x32^7 xsq2^4) xt^5)/(3 (-1+x12^2)^4 (x12^2-x32^2)^3 (-1+x32^2)^2)-(2560 x12^4 x32^2 xt^6)/(3 (-1+x12^2)^4 (x12^2-x32^2)^2)+(-((16 (12 x12^8-12 x12^10-48 x12^6 x32^2+12 x12^8 x32^2+36 x12^10 x32^2+2 x12^4 x32^4+80 x12^6 x32^4-101 x12^8 x32^4-45 x12^10 x32^4+34 x12^2 x32^6+146 x12^4 x32^6+20 x12^6 x32^6+157 x12^8 x32^6+27 x12^10 x32^6+28 x32^8-316 x12^2 x32^8-400 x12^4 x32^8-162 x12^6 x32^8-110 x12^8 x32^8-22 x32^10+700 x12^2 x32^10+432 x12^4 x32^10+170 x12^6 x32^10-87 x32^12-633 x12^2 x32^12-240 x12^4 x32^12+139 x32^14+245 x12^2 x32^14-64 x32^16))/(3 (-1+x12^2) (x12^2-x32^2)^4 (-1+x32^2)^3))-(32 (-12 x12^6 x32-87 x12^4 x32^3-11 x12^6 x32^3+168 x12^2 x32^5+263 x12^4 x32^5+13 x12^6 x32^5+27 x32^7-425 x12^2 x32^7-130 x12^4 x32^7-19 x32^9+195 x12^2 x32^9+18 x32^11) xt)/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^2)-(32 (12 x12^8-24 x12^10-48 x12^6 x32^2+60 x12^8 x32^2+72 x12^10 x32^2+7 x12^4 x32^4-61 x12^6 x32^4-325 x12^8 x32^4-97 x12^10 x32^4+24 x12^2 x32^6+189 x12^4 x32^6+811 x12^6 x32^6+525 x12^8 x32^6+55 x12^10 x32^6+x32^8-79 x12^2 x32^8-1199 x12^4 x32^8-1401 x12^6 x32^8-302 x12^8 x32^8-69 x32^10+545 x12^2 x32^10+1897 x12^4 x32^10+759 x12^6 x32^10+72 x32^12-962 x12^2 x32^12-954 x12^4 x32^12+54 x32^14+502 x12^2 x32^14-64 x32^16) xt^2)/(3 (-1+x12^2)^2 (x12^2-x32^2)^4 (-1+x32^2)^3)-(64 x32 (18 x12^6-6 x12^8+59 x12^4 x32^2-42 x12^6 x32^2-55 x12^8 x32^2-90 x12^2 x32^4-195 x12^4 x32^4+268 x12^6 x32^4+59 x12^8 x32^4+45 x32^6+214 x12^2 x32^6-67 x12^4 x32^6-252 x12^6 x32^6-67 x32^8-44 x12^2 x32^8+239 x12^4 x32^8-6 x32^10-120 x12^2 x32^10+42 x32^12) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^3 (-1+x32^2)^2)+(16 (6 x12^8-24 x12^10-30 x12^12-24 x12^6 x32^2+90 x12^8 x32^2+180 x12^10 x32^2+90 x12^12 x32^2-37 x12^4 x32^4+234 x12^6 x32^4-1612 x12^8 x32^4-238 x12^10 x32^4-123 x12^12 x32^4+64 x12^2 x32^6-479 x12^4 x32^6+2648 x12^6 x32^6+3742 x12^8 x32^6+116 x12^10 x32^6+69 x12^12 x32^6-13 x32^8+190 x12^2 x32^8-936 x12^4 x32^8-8712 x12^6 x32^8-3031 x12^8 x32^8-58 x12^10 x32^8+x32^10-452 x12^2 x32^10+6232 x12^4 x32^10+8472 x12^6 x32^10+835 x12^8 x32^10+194 x32^12-786 x12^2 x32^12-7366 x12^4 x32^12-2618 x12^6 x32^12-336 x32^14+1796 x12^2 x32^14+2556 x12^4 x32^14+148 x32^16-788 x12^2 x32^16) xt^4)/(3 (-1+x12^2)^4 (x12^2-x32^2)^4 (-1+x32^2)^3)+(32 (-12 x12^6 x32-12 x12^8 x32-93 x12^4 x32^3-152 x12^6 x32^3-59 x12^8 x32^3+148 x12^2 x32^5+535 x12^4 x32^5+384 x12^6 x32^5+61 x12^8 x32^5-75 x32^7-508 x12^2 x32^7-585 x12^4 x32^7-216 x12^6 x32^7+169 x32^9+392 x12^2 x32^9+159 x12^4 x32^9-100 x32^11-36 x12^2 x32^11) xt^5)/(3 (-1+x12^2)^4 (x12^2-x32^2)^3 (-1+x32^2)^2)-(2560 x12^2 (1+x12^2) x32^4 xt^6)/(3 (-1+x12^2)^4 (x12^2-x32^2)^2 (-1+x32^2))) Log[x32^2]+(-(1/(3 (x12^2-x32^2)^3))80 (2 x12^6-6 x12^4 x32^2+7 x12^2 x32^4-3 x32^6+18 x12^2 x32^2 xsq2^2+6 x32^4 xsq2^2-3 x12^2 xsq2^4-9 x32^2 xsq2^4)+(320 x32 (x12^4-3 x12^2 x32^2+2 x32^4-12 x12^2 xsq2^2+6 xsq2^4) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2)-(160 (-2 x12^6+4 x12^8+6 x12^4 x32^2-12 x12^6 x32^2-5 x12^2 x32^4+11 x12^4 x32^4+x32^6-3 x12^2 x32^6-18 x12^2 x32^2 xsq2^2+18 x12^4 x32^2 xsq2^2-6 x32^4 xsq2^2+6 x12^2 x32^4 xsq2^2+3 x12^2 xsq2^4-3 x12^4 xsq2^4+9 x32^2 xsq2^4-9 x12^2 x32^2 xsq2^4) xt^2)/(3 (-1+x12^2)^2 (x12^2-x32^2)^3)+(320 (-3 x12^4 x32+x12^6 x32+4 x12^2 x32^3+2 x12^4 x32^3-x32^5-5 x12^2 x32^5+2 x32^7+24 x12^2 x32 xsq2^2-24 x12^4 x32 xsq2^2-12 x32 xsq2^4+12 x12^2 x32 xsq2^4) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2)+(80 (-x12^6+4 x12^8+5 x12^10+3 x12^4 x32^2-14 x12^6 x32^2-13 x12^8 x32^2-2 x12^2 x32^4+18 x12^4 x32^4+8 x12^6 x32^4-10 x12^2 x32^6+2 x12^4 x32^6+2 x32^8-2 x12^2 x32^8-18 x12^2 x32^2 xsq2^2+18 x12^6 x32^2 xsq2^2-6 x32^4 xsq2^2+6 x12^4 x32^4 xsq2^2+3 x12^2 xsq2^4-3 x12^6 xsq2^4+9 x32^2 xsq2^4-9 x12^4 x32^2 xsq2^4) xt^4)/(3 (-1+x12^2)^4 (x12^2-x32^2)^3)-(320 (1+x12^2) x32 (x12^4-x12^2 x32^2-12 x12^2 xsq2^2+6 xsq2^4) xt^5)/(3 (-1+x12^2)^4 (x12^2-x32^2)^2)) Log[xsq2^2])+(-((8 (-1+x12^2) (3 x12^2+7 x12^4+7 x12^6+3 x12^8+9 x32^2-26 x12^2 x32^2-34 x12^4 x32^2-26 x12^6 x32^2+9 x12^8 x32^2-29 x32^4+65 x12^2 x32^4+65 x12^4 x32^4-29 x12^6 x32^4+34 x32^6-76 x12^2 x32^6+34 x12^4 x32^6-14 x32^8-14 x12^2 x32^8+12 x32^10))/(3 (x12^2-x32^2)^3 (-1+x32^2)^3))-(64 (-1+x12^2) (3 x32-x12^2 x32+3 x12^4 x32-5 x32^3-5 x12^2 x32^3+5 x32^5) xt)/(3 (x12^2-x32^2)^2 (-1+x32^2)^2)+(16 (3 x12^2+7 x12^4+7 x12^6+3 x12^8+9 x32^2-26 x12^2 x32^2-34 x12^4 x32^2-26 x12^6 x32^2+9 x12^8 x32^2-29 x32^4+65 x12^2 x32^4+65 x12^4 x32^4-29 x12^6 x32^4+34 x32^6-76 x12^2 x32^6+34 x12^4 x32^6-14 x32^8-14 x12^2 x32^8+12 x32^10) xt^4)/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^3)+(128 (3 x32-x12^2 x32+3 x12^4 x32-5 x32^3-5 x12^2 x32^3+5 x32^5) xt^5)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^2)) PolyLog[2,1-x12^2]+(-((8 (-3 x12^2-x12^4-19 x12^6+23 x12^8-9 x32^2+44 x12^2 x32^2+42 x12^4 x32^2-148 x12^6 x32^2-57 x12^8 x32^2+29 x32^4-192 x12^2 x32^4+376 x12^4 x32^4+598 x12^6 x32^4-43 x12^8 x32^4-11 x32^6-100 x12^2 x32^6-1498 x12^4 x32^6-324 x12^6 x32^6+13 x12^8 x32^6+89 x32^8+1145 x12^2 x32^8+1177 x12^4 x32^8+149 x12^6 x32^8-368 x32^10-1072 x12^2 x32^10-480 x12^4 x32^10+334 x32^12+434 x12^2 x32^12-128 x32^14))/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^3))+(64 (3 x32-7 x12^2 x32+22 x12^4 x32-5 x32^3-23 x12^2 x32^3-24 x12^4 x32^3+19 x32^5+59 x12^2 x32^5+34 x12^4 x32^5-47 x32^7-93 x12^2 x32^7+62 x32^9) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^2)-(128 x32 (1+x12^2-2 x32^2) (5 x12^2+3 x32^2+3 x12^2 x32^2-11 x32^4) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2) (-1+x32^2))+(16 (3 x12^2+x12^4+29 x12^6-13 x12^8+9 x32^2-41 x12^2 x32^2-55 x12^4 x32^2+129 x12^6 x32^2+30 x12^8 x32^2-20 x32^4+149 x12^2 x32^4-411 x12^4 x32^4-409 x12^6 x32^4+31 x12^8 x32^4-3 x32^6+313 x12^2 x32^6+1151 x12^4 x32^6+59 x12^6 x32^6-138 x32^8-1084 x12^2 x32^8-398 x12^4 x32^8+372 x32^10+468 x12^2 x32^10-172 x32^12) xt^4)/(3 (-1+x12^2)^3 (x12^2-x32^2)^3 (-1+x32^2)^2)-(128 x32 (3-7 x12^2+22 x12^4+x32^2-37 x12^2 x32^2+18 x32^4) xt^5)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2)) PolyLog[2,1-x32^2]+(-((8 (-23 x12^6+19 x12^8+x12^10+3 x12^12+57 x12^4 x32^2+148 x12^6 x32^2-42 x12^8 x32^2-44 x12^10 x32^2+9 x12^12 x32^2+43 x12^2 x32^4-598 x12^4 x32^4-376 x12^6 x32^4+192 x12^8 x32^4-29 x12^10 x32^4-13 x32^6+324 x12^2 x32^6+1498 x12^4 x32^6+100 x12^6 x32^6+11 x12^8 x32^6-149 x32^8-1177 x12^2 x32^8-1145 x12^4 x32^8-89 x12^6 x32^8+480 x32^10+1072 x12^2 x32^10+368 x12^4 x32^10-434 x32^12-334 x12^2 x32^12+128 x32^14))/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^3))-(64 (22 x12^4 x32-7 x12^6 x32+3 x12^8 x32-24 x12^2 x32^3-23 x12^4 x32^3-5 x12^6 x32^3+34 x32^5+59 x12^2 x32^5+19 x12^4 x32^5-93 x32^7-47 x12^2 x32^7+62 x32^9) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^2)+(128 x32 (1+x12^2-2 x32^2) (5 x12^2+3 x32^2+3 x12^2 x32^2-11 x32^4) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2) (-1+x32^2))+(16 (-13 x12^4+29 x12^6+x12^8+3 x12^10+30 x12^2 x32^2+129 x12^4 x32^2-55 x12^6 x32^2-41 x12^8 x32^2+9 x12^10 x32^2+31 x32^4-409 x12^2 x32^4-411 x12^4 x32^4+149 x12^6 x32^4-20 x12^8 x32^4+59 x32^6+1151 x12^2 x32^6+313 x12^4 x32^6-3 x12^6 x32^6-398 x32^8-1084 x12^2 x32^8-138 x12^4 x32^8+468 x32^10+372 x12^2 x32^10-172 x32^12) xt^4)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^3)+(128 x32 (22-7 x12^2+3 x12^4-37 x32^2+x12^2 x32^2+18 x32^4) xt^5)/(3 (-1+x12^2)^3 (-1+x32^2)^2)) PolyLog[2,1-x32^2/x12^2]+Log[xsq2^2] ((80 (-14 x12^6+31 x12^4 x32^2+31 x12^6 x32^2-19 x12^2 x32^4-68 x12^4 x32^4-19 x12^6 x32^4+2 x32^6+41 x12^2 x32^6+41 x12^4 x32^6+2 x12^6 x32^6-4 x32^8-24 x12^2 x32^8-4 x12^4 x32^8+2 x32^10+2 x12^2 x32^10-3 x12^4 xsq2^2-3 x12^6 xsq2^2-9 x12^2 x32^2 xsq2^2+12 x12^4 x32^2 xsq2^2-9 x12^6 x32^2 xsq2^2+15 x12^2 x32^4 xsq2^2+15 x12^4 x32^4 xsq2^2-18 x12^2 x32^6 xsq2^2))/(3 x12^2 (x12^2-x32^2)^2 (-1+x32^2)^2)+(640 x32 xsq2^2 xt)/((-1+x32^2) (-x12^2+x32^2))-(320 x32^2 xt^2)/(3 x12^2)-(2560 x32 xt^3)/(3 (-1+x12^2)^2)+(160 (16 x12^6-32 x12^4 x32^2-32 x12^6 x32^2+15 x12^2 x32^4+64 x12^4 x32^4+15 x12^6 x32^4+x32^6-30 x12^2 x32^6-30 x12^4 x32^6+x12^6 x32^6-2 x32^8+14 x12^2 x32^8-2 x12^4 x32^8+x32^10+x12^2 x32^10+3 x12^4 xsq2^2+3 x12^6 xsq2^2+9 x12^2 x32^2 xsq2^2-12 x12^4 x32^2 xsq2^2+9 x12^6 x32^2 xsq2^2-15 x12^2 x32^4 xsq2^2-15 x12^4 x32^4 xsq2^2+18 x12^2 x32^6 xsq2^2) xt^4)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)+(1280 x32 xsq2^2 xt^5)/((-1+x12^2)^2 (x12^2-x32^2) (-1+x32^2))+(-((320 x32^4)/(3 (-1+x32^2)^2))-(1280 x32^3 xt)/(3 (-1+x12^2) (-1+x32^2))+(640 x32 (1+x12^2-2 x32^2) xt^3)/(3 (-1+x12^2)^3)+(160 (1+x12^2-2 x32^2) xt^4)/(3 (-1+x12^2)^3)) PolyLog[2,1-x32^2]+(-((320 x32^4)/(3 (-x12^2+x32^2)^2))-(1280 x32^3 xt)/(3 (-1+x12^2) (x12^2-x32^2))-(640 x32 (1+x12^2-2 x32^2) xt^3)/(3 (-1+x12^2)^3)-(160 (1+x12^2-2 x32^2) xt^4)/(3 (-1+x12^2)^3)) PolyLog[2,1-x32^2/x12^2])+Log[x32^2] ((8 (96 x12^8+3 x12^4 x32^2-513 x12^6 x32^2-513 x12^8 x32^2+3 x12^10 x32^2+9 x12^2 x32^4+860 x12^4 x32^4+2470 x12^6 x32^4+860 x12^8 x32^4+9 x12^10 x32^4-519 x12^2 x32^6-3785 x12^4 x32^6-3785 x12^6 x32^6-519 x12^8 x32^6+68 x32^8+2016 x12^2 x32^8+5432 x12^4 x32^8+2016 x12^6 x32^8+68 x12^8 x32^8-236 x32^10-2738 x12^2 x32^10-2738 x12^4 x32^10-236 x12^6 x32^10+300 x32^12+1336 x12^2 x32^12+300 x12^4 x32^12-132 x32^14-132 x12^2 x32^14+30 x12^4 x32^2 xsq2^2+30 x12^8 x32^2 xsq2^2+90 x12^2 x32^4 xsq2^2-90 x12^4 x32^4 xsq2^2-90 x12^6 x32^4 xsq2^2+90 x12^8 x32^4 xsq2^2-270 x12^2 x32^6 xsq2^2+180 x12^4 x32^6 xsq2^2-270 x12^6 x32^6 xsq2^2+240 x12^2 x32^8 xsq2^2+240 x12^4 x32^8 xsq2^2-180 x12^2 x32^10 xsq2^2))/(3 x12^2 (x12^2-x32^2)^3 (-1+x32^2)^3)+(64 (3 x12^2 x32^3+90 x12^4 x32^3+3 x12^6 x32^3-112 x12^2 x32^5-112 x12^4 x32^5+16 x32^7+128 x12^2 x32^7+16 x12^4 x32^7-16 x32^9-16 x12^2 x32^9+30 x12^2 x32^3 xsq2^2+30 x12^4 x32^3 xsq2^2-60 x12^2 x32^5 xsq2^2) xt)/(3 x12^2 (x12^2-x32^2)^2 (-1+x32^2)^2)-(64 x32^2 (6 x12^4-7 x12^2 x32^2-7 x12^4 x32^2+17 x32^4+25 x12^2 x32^4+17 x12^4 x32^4-42 x32^6-42 x12^2 x32^6+33 x32^8) xt^2)/(3 x12^2 (x12^2-x32^2)^2 (-1+x32^2)^2)-(1024 (3 x12^6 x32-6 x12^4 x32^3-6 x12^6 x32^3+5 x12^2 x32^5+16 x12^4 x32^5+5 x12^6 x32^5+2 x32^7-16 x12^2 x32^7-16 x12^4 x32^7+2 x12^6 x32^7-2 x32^9+15 x12^2 x32^9-2 x12^4 x32^9) xt^3)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)-(16 (120 x12^8+3 x12^4 x32^2-559 x12^6 x32^2-559 x12^8 x32^2+3 x12^10 x32^2+9 x12^2 x32^4+850 x12^4 x32^4+3522 x12^6 x32^4+850 x12^8 x32^4+9 x12^10 x32^4-465 x12^2 x32^6-5645 x12^4 x32^6-5645 x12^6 x32^6-465 x12^8 x32^6-34 x32^8+2756 x12^2 x32^8+8964 x12^4 x32^8+2756 x12^6 x32^8-34 x12^8 x32^8+118 x32^10-4190 x12^2 x32^10-4190 x12^4 x32^10+118 x12^6 x32^10-150 x32^12+1876 x12^2 x32^12-150 x12^4 x32^12+66 x32^14+66 x12^2 x32^14+30 x12^4 x32^2 xsq2^2+30 x12^8 x32^2 xsq2^2+90 x12^2 x32^4 xsq2^2-90 x12^4 x32^4 xsq2^2-90 x12^6 x32^4 xsq2^2+90 x12^8 x32^4 xsq2^2-270 x12^2 x32^6 xsq2^2+180 x12^4 x32^6 xsq2^2-270 x12^6 x32^6 xsq2^2+240 x12^2 x32^8 xsq2^2+240 x12^4 x32^8 xsq2^2-180 x12^2 x32^10 xsq2^2) xt^4)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^3 (-1+x32^2)^3)-(128 (3 x12^2 x32^3+101 x12^4 x32^3+3 x12^6 x32^3-115 x12^2 x32^5-115 x12^4 x32^5-8 x32^7+123 x12^2 x32^7-8 x12^4 x32^7+8 x32^9+8 x12^2 x32^9+30 x12^2 x32^3 xsq2^2+30 x12^4 x32^3 xsq2^2-60 x12^2 x32^5 xsq2^2) xt^5)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)+(-((80 (-x12^4 x32^4-x12^8 x32^4-x12^2 x32^6+3 x12^4 x32^6+3 x12^6 x32^6-x12^8 x32^6+2 x32^8+3 x12^2 x32^8-6 x12^4 x32^8+3 x12^6 x32^8+2 x12^8 x32^8-6 x32^10-2 x12^2 x32^10-2 x12^4 x32^10-6 x12^6 x32^10+6 x32^12+2 x12^2 x32^12+6 x12^4 x32^12-2 x32^14-2 x12^2 x32^14-18 x12^4 x32^2 xsq2^2-18 x12^8 x32^2 xsq2^2-6 x12^2 x32^4 xsq2^2+54 x12^4 x32^4 xsq2^2+54 x12^6 x32^4 xsq2^2-6 x12^8 x32^4 xsq2^2+18 x12^2 x32^6 xsq2^2-108 x12^4 x32^6 xsq2^2+18 x12^6 x32^6 xsq2^2+12 x12^2 x32^10 xsq2^2+3 x12^4 xsq2^4+3 x12^8 xsq2^4+9 x12^2 x32^2 xsq2^4-9 x12^4 x32^2 xsq2^4-9 x12^6 x32^2 xsq2^4+9 x12^8 x32^2 xsq2^4-27 x12^2 x32^4 xsq2^4+18 x12^4 x32^4 xsq2^4-27 x12^6 x32^4 xsq2^4+24 x12^2 x32^6 xsq2^4+24 x12^4 x32^6 xsq2^4-18 x12^2 x32^8 xsq2^4))/(3 x12^2 (x12^2-x32^2)^3 (-1+x32^2)^3))-(640 (-2 x12^2 x32 xsq2^2+2 x32^5 xsq2^2+x32 xsq2^4+x12^2 x32 xsq2^4-2 x32^3 xsq2^4) xt)/((-1+x32^2)^2 (-x12^2+x32^2)^2)+(320 x32^4 (1+x12^2-x32^2) xt^2)/(3 x12^2 (x12^2-x32^2) (-1+x32^2))-(160 (-x12^6 x32^2-x12^8 x32^2+2 x12^4 x32^4+6 x12^6 x32^4+2 x12^8 x32^4-2 x12^2 x32^6-9 x12^4 x32^6-9 x12^6 x32^6-2 x12^8 x32^6+x32^8+7 x12^2 x32^8+12 x12^4 x32^8+7 x12^6 x32^8+x12^8 x32^8-3 x32^10-8 x12^2 x32^10-8 x12^4 x32^10-3 x12^6 x32^10+3 x32^12+4 x12^2 x32^12+3 x12^4 x32^12-x32^14-x12^2 x32^14+18 x12^4 x32^2 xsq2^2+18 x12^8 x32^2 xsq2^2+6 x12^2 x32^4 xsq2^2-54 x12^4 x32^4 xsq2^2-54 x12^6 x32^4 xsq2^2+6 x12^8 x32^4 xsq2^2-18 x12^2 x32^6 xsq2^2+108 x12^4 x32^6 xsq2^2-18 x12^6 x32^6 xsq2^2-12 x12^2 x32^10 xsq2^2-3 x12^4 xsq2^4-3 x12^8 xsq2^4-9 x12^2 x32^2 xsq2^4+9 x12^4 x32^2 xsq2^4+9 x12^6 x32^2 xsq2^4-9 x12^8 x32^2 xsq2^4+27 x12^2 x32^4 xsq2^4-18 x12^4 x32^4 xsq2^4+27 x12^6 x32^4 xsq2^4-24 x12^2 x32^6 xsq2^4-24 x12^4 x32^6 xsq2^4+18 x12^2 x32^8 xsq2^4) xt^4)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^3 (-1+x32^2)^3)-(640 (-x12^2 x32^3+x32^5+x12^2 x32^5-x32^7+12 x12^2 x32 xsq2^2-12 x32^5 xsq2^2-6 x32 xsq2^4-6 x12^2 x32 xsq2^4+12 x32^3 xsq2^4) xt^5)/(3 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)) Log[xsq2^2]+(-((128 x32^4 (3 x12^4-6 x12^2 x32^2-6 x12^4 x32^2+7 x32^4+12 x12^2 x32^4+7 x12^4 x32^4-14 x32^6-14 x12^2 x32^6+11 x32^8))/(3 (-1+x32^2)^4 (-x12^2+x32^2)^2))-(512 x32^3 (3 x12^4-6 x12^2 x32^2-6 x12^4 x32^2+7 x32^4+8 x12^2 x32^4+11 x12^4 x32^4-10 x32^6-18 x12^2 x32^6+11 x32^8) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^3)-(8192 x32^6 xt^2)/(3 (-1+x12^2) (x12^2-x32^2) (-1+x32^2)^2)+(256 x32 (1+x12^2-2 x32^2) (3 x12^4-6 x12^2 x32^2-6 x12^4 x32^2+7 x32^4+12 x12^2 x32^4+7 x12^4 x32^4-14 x32^6-14 x12^2 x32^6+11 x32^8) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)+(64 (1+x12^2-2 x32^2) (3 x12^4-6 x12^2 x32^2-6 x12^4 x32^2+7 x32^4-52 x12^2 x32^4+7 x12^4 x32^4+50 x32^6+50 x12^2 x32^6-53 x32^8) xt^4)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)+(1024 x32^3 (1+x12^2-2 x32^2) xt^5)/(3 (-1+x12^2)^3 (x12^2-x32^2) (-1+x32^2))) PolyLog[2,1-x32^2]+(-((128 x32^4 (3 x12^4-6 x12^2 x32^2-6 x12^4 x32^2+7 x32^4+12 x12^2 x32^4+7 x12^4 x32^4-14 x32^6-14 x12^2 x32^6+11 x32^8))/(3 (-1+x32^2)^2 (-x12^2+x32^2)^4))-(512 x32^3 (3 x12^4-6 x12^2 x32^2-6 x12^4 x32^2+11 x32^4+8 x12^2 x32^4+7 x12^4 x32^4-18 x32^6-10 x12^2 x32^6+11 x32^8) xt)/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^2)-(8192 x32^6 xt^2)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2))-(256 x32 (1+x12^2-2 x32^2) (3 x12^4-6 x12^2 x32^2-6 x12^4 x32^2+7 x32^4+12 x12^2 x32^4+7 x12^4 x32^4-14 x32^6-14 x12^2 x32^6+11 x32^8) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)-(64 (1+x12^2-2 x32^2) (3 x12^4-6 x12^2 x32^2-6 x12^4 x32^2+7 x32^4-52 x12^2 x32^4+7 x12^4 x32^4+50 x32^6+50 x12^2 x32^6-53 x32^8) xt^4)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)-(1024 x32^3 (1+x12^2-2 x32^2) xt^5)/(3 (-1+x12^2)^3 (x12^2-x32^2) (-1+x32^2))) PolyLog[2,1-x32^2/x12^2])+((80 (-1+xsq2^2) (-1+3 x32^2+2 x32^4-xsq2^2-3 x32^2 xsq2^2))/(-1+x32^2)^3-(640 x32 (-1+xsq2^2)^2 xt)/((-1+x12^2) (-1+x32^2)^2)-(160 (-1+xsq2^2) (-1+3 x32^2+2 x32^4-xsq2^2-3 x32^2 xsq2^2) xt^4)/((-1+x12^2)^2 (-1+x32^2)^3)+(1280 x32 (-1+xsq2^2)^2 xt^5)/((-1+x12^2)^3 (-1+x32^2)^2)) PolyLog[2,1-xsq2^2]+(-((80 (x12^2-xsq2^2) (x12^4-3 x12^2 x32^2-2 x32^4+x12^2 xsq2^2+3 x32^2 xsq2^2))/(x12^2-x32^2)^3)+(640 x32 (x12^2-xsq2^2)^2 xt)/((-1+x12^2) (x12^2-x32^2)^2)+(160 (x12^2-xsq2^2) (x12^4-3 x12^2 x32^2-2 x32^4+x12^2 xsq2^2+3 x32^2 xsq2^2) xt^4)/((-1+x12^2)^2 (x12^2-x32^2)^3)-(1280 x32 (x12^2-xsq2^2)^2 xt^5)/((-1+x12^2)^3 (x12^2-x32^2)^2)) PolyLog[2,1-xsq2^2/x12^2]+(-((80 (x32^2-xsq2^2) (-5 x12^2 x32^2-5 x12^6 x32^2+x32^4+15 x12^2 x32^4+15 x12^4 x32^4+x12^6 x32^4-3 x32^6-30 x12^2 x32^6-3 x12^4 x32^6+8 x32^8+8 x12^2 x32^8-2 x32^10+x12^2 xsq2^2+x12^6 xsq2^2+3 x32^2 xsq2^2-3 x12^2 x32^2 xsq2^2-3 x12^4 x32^2 xsq2^2+3 x12^6 x32^2 xsq2^2-9 x32^4 xsq2^2+6 x12^2 x32^4 xsq2^2-9 x12^4 x32^4 xsq2^2+8 x32^6 xsq2^2+8 x12^2 x32^6 xsq2^2-6 x32^8 xsq2^2))/((x12^2-x32^2)^3 (-1+x32^2)^3))-(640 (x32^2-xsq2^2) (-2 x12^2 x32+x32^3+x12^2 x32^3+x32 xsq2^2+x12^2 x32 xsq2^2-2 x32^3 xsq2^2) xt)/((-1+x32^2)^2 (-x12^2+x32^2)^2)+(160 (x32^2-xsq2^2) (-5 x12^2 x32^2-5 x12^6 x32^2+x32^4+15 x12^2 x32^4+15 x12^4 x32^4+x12^6 x32^4-3 x32^6-30 x12^2 x32^6-3 x12^4 x32^6+8 x32^8+8 x12^2 x32^8-2 x32^10+x12^2 xsq2^2+x12^6 xsq2^2+3 x32^2 xsq2^2-3 x12^2 x32^2 xsq2^2-3 x12^4 x32^2 xsq2^2+3 x12^6 x32^2 xsq2^2-9 x32^4 xsq2^2+6 x12^2 x32^4 xsq2^2-9 x12^4 x32^4 xsq2^2+8 x32^6 xsq2^2+8 x12^2 x32^6 xsq2^2-6 x32^8 xsq2^2) xt^4)/((-1+x12^2)^2 (x12^2-x32^2)^3 (-1+x32^2)^3)+(1280 (x32^2-xsq2^2) (-2 x12^2 x32+x32^3+x12^2 x32^3+x32 xsq2^2+x12^2 x32 xsq2^2-2 x32^3 xsq2^2) xt^5)/((-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)) PolyLog[2,1-xsq2^2/x32^2]+Log[x12^2] ((4 (-12 x12^6+183 x12^8-171 x12^10-30 x12^4 x32^2-505 x12^6 x32^2-571 x12^8 x32^2+594 x12^10 x32^2+18 x12^2 x32^4+541 x12^4 x32^4+2359 x12^6 x32^4-283 x12^8 x32^4-587 x12^10 x32^4-207 x12^2 x32^6-2217 x12^4 x32^6-1975 x12^6 x32^6+1259 x12^8 x32^6+68 x12^10 x32^6+4 x32^8+704 x12^2 x32^8+2127 x12^4 x32^8-583 x12^6 x32^8-204 x12^8 x32^8-8 x32^10-611 x12^2 x32^10-33 x12^4 x32^10+140 x12^6 x32^10+4 x32^12-4 x12^4 x32^12-120 x12^6 xsq2^2+60 x12^8 xsq2^2+60 x12^10 xsq2^2-300 x12^4 x32^2 xsq2^2+720 x12^6 x32^2 xsq2^2-600 x12^8 x32^2 xsq2^2+180 x12^10 x32^2 xsq2^2+180 x12^2 x32^4 xsq2^2+240 x12^4 x32^4 xsq2^2-420 x12^8 x32^4 xsq2^2-300 x12^2 x32^6 xsq2^2-540 x12^4 x32^6 xsq2^2+840 x12^6 x32^6 xsq2^2+360 x12^2 x32^8 xsq2^2-360 x12^4 x32^8 xsq2^2))/(3 x12^2 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^2)-(64 (6 x12^2 x32+74 x12^4 x32-3 x32^3-100 x12^2 x32^3-87 x12^4 x32^3+20 x32^5+112 x12^2 x32^5+16 x12^4 x32^5-22 x32^7-16 x12^2 x32^7+60 x12^2 x32 xsq2^2-30 x12^4 x32 xsq2^2-30 x32^3 xsq2^2) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2))+(8 (6 x12^4+237 x12^6+6 x12^8+18 x12^2 x32^2-212 x12^4 x32^2-308 x12^6 x32^2+18 x12^8 x32^2-301 x12^2 x32^4-40 x12^4 x32^4-5 x12^6 x32^4+4 x32^6+886 x12^2 x32^6+326 x12^4 x32^6-68 x12^6 x32^6-8 x32^8-695 x12^2 x32^8+136 x12^4 x32^8+4 x32^10-4 x12^2 x32^10+60 x12^4 xsq2^2+60 x12^6 xsq2^2+180 x12^2 x32^2 xsq2^2-240 x12^4 x32^2 xsq2^2+180 x12^6 x32^2 xsq2^2-300 x12^2 x32^4 xsq2^2-300 x12^4 x32^4 xsq2^2+360 x12^2 x32^6 xsq2^2) xt^2)/(3 x12^2 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^2)+(128 (3 x12^2 x32-42 x12^4 x32+16 x12^6 x32+3 x12^8 x32-3 x32^3+43 x12^2 x32^3+32 x12^4 x32^3-76 x12^6 x32^3+31 x32^5-87 x12^2 x32^5+108 x12^4 x32^5+16 x12^6 x32^5-25 x32^7-3 x12^2 x32^7-16 x12^4 x32^7+30 x12^2 x32 xsq2^2-60 x12^4 x32 xsq2^2+30 x12^6 x32 xsq2^2-30 x32^3 xsq2^2+60 x12^2 x32^3 xsq2^2-30 x12^4 x32^3 xsq2^2) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2))+(4 (6 x12^6-1167 x12^8+417 x12^10+6 x12^12+24 x12^4 x32^2+2665 x12^6 x32^2+5659 x12^8 x32^2-1508 x12^10 x32^2+18 x12^12 x32^2+18 x12^2 x32^4-1361 x12^4 x32^4-13435 x12^6 x32^4-6073 x12^8 x32^4+975 x12^10 x32^4-493 x12^2 x32^6+6913 x12^4 x32^6+17667 x12^6 x32^6+1881 x12^8 x32^6+68 x12^10 x32^6+4 x32^8+1222 x12^2 x32^8-10101 x12^4 x32^8-7283 x12^6 x32^8-204 x12^8 x32^8-8 x32^10-667 x12^2 x32^10+4793 x12^4 x32^10+140 x12^6 x32^10+4 x32^12-176 x12^2 x32^12-4 x12^4 x32^12+60 x12^6 xsq2^2-240 x12^8 xsq2^2-60 x12^10 xsq2^2+240 x12^4 x32^2 xsq2^2-420 x12^6 x32^2 xsq2^2+360 x12^8 x32^2 xsq2^2-180 x12^10 x32^2 xsq2^2+180 x12^2 x32^4 xsq2^2-480 x12^4 x32^4 xsq2^2+1380 x12^6 x32^4 xsq2^2+360 x12^8 x32^4 xsq2^2-300 x12^2 x32^6 xsq2^2-600 x12^4 x32^6 xsq2^2-1020 x12^6 x32^6 xsq2^2+360 x12^2 x32^8 xsq2^2+360 x12^4 x32^8 xsq2^2) xt^4)/(3 x12^2 (-1+x12^2)^3 (x12^2-x32^2)^3 (-1+x32^2)^2)+(64 (3 x12^2 x32+184 x12^4 x32+3 x12^6 x32+3 x32^3-166 x12^2 x32^3-185 x12^4 x32^3-62 x32^5+172 x12^2 x32^5-16 x12^4 x32^5+48 x32^7+16 x12^2 x32^7+30 x12^2 x32 xsq2^2-30 x12^4 x32 xsq2^2+30 x32^3 xsq2^2-30 x12^2 x32^3 xsq2^2) xt^5)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2))+(-((8 (-36 x12^8+144 x12^6 x32^2+144 x12^8 x32^2-100 x12^4 x32^4-452 x12^6 x32^4-76 x12^8 x32^4-20 x12^2 x32^6+28 x12^4 x32^6-68 x12^6 x32^6-68 x12^8 x32^6-16 x32^8+452 x12^2 x32^8+1340 x12^4 x32^8+644 x12^6 x32^8+40 x12^8 x32^8-60 x32^10-1764 x12^2 x32^10-1956 x12^4 x32^10-284 x12^6 x32^10+400 x32^12+1980 x12^2 x32^12+712 x12^4 x32^12-488 x32^14-664 x12^2 x32^14+168 x32^16+90 x12^4 x32^2 xsq2^2+90 x12^8 x32^2 xsq2^2-60 x12^2 x32^4 xsq2^2-360 x12^4 x32^4 xsq2^2-360 x12^6 x32^4 xsq2^2-60 x12^8 x32^4 xsq2^2-30 x32^6 xsq2^2+240 x12^2 x32^6 xsq2^2+1080 x12^4 x32^6 xsq2^2+240 x12^6 x32^6 xsq2^2-30 x12^8 x32^6 xsq2^2+120 x32^8 xsq2^2-720 x12^2 x32^8 xsq2^2-720 x12^4 x32^8 xsq2^2+120 x12^6 x32^8 xsq2^2-90 x32^10 xsq2^2+480 x12^2 x32^10 xsq2^2-90 x12^4 x32^10 xsq2^2+60 x32^12 xsq2^2+60 x12^2 x32^12 xsq2^2-60 x32^14 xsq2^2-15 x12^4 xsq2^4-15 x12^8 xsq2^4-30 x12^2 x32^2 xsq2^4+60 x12^4 x32^2 xsq2^4+60 x12^6 x32^2 xsq2^4-30 x12^8 x32^2 xsq2^4+45 x32^4 xsq2^4+120 x12^2 x32^4 xsq2^4-180 x12^4 x32^4 xsq2^4+120 x12^6 x32^4 xsq2^4+45 x12^8 x32^4 xsq2^4-180 x32^6 xsq2^4-120 x12^2 x32^6 xsq2^4-120 x12^4 x32^6 xsq2^4-180 x12^6 x32^6 xsq2^4+255 x32^8 xsq2^4+240 x12^2 x32^8 xsq2^4+255 x12^4 x32^8 xsq2^4-210 x32^10 xsq2^4-210 x12^2 x32^10 xsq2^4+90 x32^12 xsq2^4))/(3 (-1+x32^2)^4 (-x12^2+x32^2)^4))-(64 (-45 x12^4 x32^3+57 x12^6 x32^3+73 x12^2 x32^5-36 x12^4 x32^5-97 x12^6 x32^5-4 x32^7-56 x12^2 x32^7+164 x12^4 x32^7+48 x12^6 x32^7-37 x32^9-72 x12^2 x32^9-107 x12^4 x32^9+77 x32^11+79 x12^2 x32^11-44 x32^13-30 x12^4 x32 xsq2^2+30 x12^6 x32 xsq2^2+30 x12^2 x32^3 xsq2^2-30 x12^6 x32^3 xsq2^2-30 x32^7 xsq2^2+30 x12^4 x32^7 xsq2^2+30 x32^9 xsq2^2-30 x12^2 x32^9 xsq2^2+15 x12^2 x32 xsq2^4-15 x12^6 x32 xsq2^4-15 x32^3 xsq2^4-45 x12^2 x32^3 xsq2^4+45 x12^4 x32^3 xsq2^4+15 x12^6 x32^3 xsq2^4+45 x32^5 xsq2^4-45 x12^4 x32^5 xsq2^4-30 x32^7 xsq2^4+30 x12^2 x32^7 xsq2^4) xt)/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^3)-(16 (-36 x12^8+144 x12^6 x32^2+144 x12^8 x32^2-76 x12^4 x32^4-452 x12^6 x32^4-76 x12^8 x32^4-68 x12^2 x32^6+108 x12^4 x32^6+12 x12^6 x32^6-68 x12^8 x32^6+40 x32^8+292 x12^2 x32^8+748 x12^4 x32^8+484 x12^6 x32^8+40 x12^8 x32^8-108 x32^10-820 x12^2 x32^10-1108 x12^4 x32^10-204 x12^6 x32^10+160 x32^12+764 x12^2 x32^12+352 x12^4 x32^12-88 x32^14-184 x12^2 x32^14+90 x12^4 x32^2 xsq2^2+90 x12^8 x32^2 xsq2^2-60 x12^2 x32^4 xsq2^2-360 x12^4 x32^4 xsq2^2-360 x12^6 x32^4 xsq2^2-60 x12^8 x32^4 xsq2^2-30 x32^6 xsq2^2+240 x12^2 x32^6 xsq2^2+1080 x12^4 x32^6 xsq2^2+240 x12^6 x32^6 xsq2^2-30 x12^8 x32^6 xsq2^2+120 x32^8 xsq2^2-720 x12^2 x32^8 xsq2^2-720 x12^4 x32^8 xsq2^2+120 x12^6 x32^8 xsq2^2-90 x32^10 xsq2^2+480 x12^2 x32^10 xsq2^2-90 x12^4 x32^10 xsq2^2+60 x32^12 xsq2^2+60 x12^2 x32^12 xsq2^2-60 x32^14 xsq2^2-15 x12^4 xsq2^4-15 x12^8 xsq2^4-30 x12^2 x32^2 xsq2^4+60 x12^4 x32^2 xsq2^4+60 x12^6 x32^2 xsq2^4-30 x12^8 x32^2 xsq2^4+45 x32^4 xsq2^4+120 x12^2 x32^4 xsq2^4-180 x12^4 x32^4 xsq2^4+120 x12^6 x32^4 xsq2^4+45 x12^8 x32^4 xsq2^4-180 x32^6 xsq2^4-120 x12^2 x32^6 xsq2^4-120 x12^4 x32^6 xsq2^4-180 x12^6 x32^6 xsq2^4+255 x32^8 xsq2^4+240 x12^2 x32^8 xsq2^4+255 x12^4 x32^8 xsq2^4-210 x32^10 xsq2^4-210 x12^2 x32^10 xsq2^4+90 x32^12 xsq2^4) xt^2)/(3 (-1+x12^2) (x12^2-x32^2)^4 (-1+x32^2)^4)+1/3 mQ3^3 ((128 mQ3^3 x32^3 (-87 mQ3^12 x12^4+186 mQ3^12 x12^6-87 mQ3^12 x12^8+2 mQ3^2 (71 mQ3^10 x12^2-89 mQ3^10 x12^4-89 mQ3^10 x12^6+71 mQ3^10 x12^8) x32^2-mQ3^4 (63 mQ3^8+42 mQ3^8 x12^2-422 mQ3^8 x12^4+42 mQ3^8 x12^6+63 mQ3^8 x12^8) x32^4+mQ3^6 (66 mQ3^6-250 mQ3^6 x12^2-250 mQ3^6 x12^4+66 mQ3^6 x12^6) x32^6+3 mQ3^8 (15 mQ3^4+94 mQ3^4 x12^2+15 mQ3^4 x12^4) x32^8-100 mQ3^10 (mQ3^2+mQ3^2 x12^2) x32^10+44 mQ3^12 x32^12))/((mQ3^2-mQ3^2 x12^2)^3 (-mQ3^2+mQ3^2 x32^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)-(1920 (mQ3^2 x32^2-mQ3^2 xsq2^2) (-2 mQ3^5 x12^2 x32+mQ3^5 x32^3+mQ3^5 x12^2 x32^3+mQ3^5 x32 xsq2^2+mQ3^5 x12^2 x32 xsq2^2-2 mQ3^5 x32^3 xsq2^2))/((mQ3^2-mQ3^2 x12^2) (-mQ3^2+mQ3^2 x32^2)^2 (-mQ3^2 x12^2+mQ3^2 x32^2)^2)) xt^3+1/3 mQ3^4 (-((8 (-36 mQ3^18 x12^8-36 mQ3^18 x12^10+144 mQ3^18 x12^6 x32^2+288 mQ3^18 x12^8 x32^2+144 mQ3^18 x12^10 x32^2-mQ3^18 x12^4 x32^4-453 mQ3^18 x12^6 x32^4-453 mQ3^18 x12^8 x32^4-mQ3^18 x12^10 x32^4-158 mQ3^18 x12^2 x32^6-846 mQ3^18 x12^4 x32^6+928 mQ3^18 x12^6 x32^6-846 mQ3^18 x12^8 x32^6-158 mQ3^18 x12^10 x32^6+55 mQ3^18 x32^8+1699 mQ3^18 x12^2 x32^8-292 mQ3^18 x12^4 x32^8-292 mQ3^18 x12^6 x32^8+1699 mQ3^18 x12^8 x32^8+55 mQ3^18 x12^10 x32^8-664 mQ3^18 x32^10-868 mQ3^18 x12^2 x32^10+5224 mQ3^18 x12^4 x32^10-868 mQ3^18 x12^6 x32^10-664 mQ3^18 x12^8 x32^10+549 mQ3^18 x32^12-5903 mQ3^18 x12^2 x32^12-5903 mQ3^18 x12^4 x32^12+549 mQ3^18 x12^6 x32^12+1826 mQ3^18 x32^14+8004 mQ3^18 x12^2 x32^14+1826 mQ3^18 x12^4 x32^14-2786 mQ3^18 x32^16-2786 mQ3^18 x12^2 x32^16+1024 mQ3^18 x32^18))/((mQ3^2-mQ3^2 x12^2)^3 (-mQ3^2+mQ3^2 x32^2)^4 (-mQ3^2 x12^2+mQ3^2 x32^2)^4))+(120 (mQ3^2+mQ3^2 x12^2) (mQ3^2 x32^2-mQ3^2 xsq2^2) (2 mQ3^10 x32^10-mQ3^6 x12^2 (mQ3^4+mQ3^4 x12^4) xsq2^2+mQ3^8 x32^8 (-8 mQ3^2-8 mQ3^2 x12^2+6 mQ3^2 xsq2^2)+mQ3^6 x32^6 (3 mQ3^4+3 mQ3^4 x12^4-8 mQ3^4 x12^2 xsq2^2+mQ3^2 (30 mQ3^2 x12^2-8 mQ3^2 xsq2^2))-mQ3^4 x32^4 (mQ3^6+mQ3^6 x12^6-9 mQ3^6 x12^4 xsq2^2+mQ3^4 (15 mQ3^2 x12^2-9 mQ3^2 xsq2^2)+3 mQ3^2 (5 mQ3^4 x12^4+2 mQ3^4 x12^2 xsq2^2))+mQ3^2 x32^2 (3 mQ3^8 x12^2 xsq2^2-3 mQ3^8 x12^6 xsq2^2+mQ3^6 (5 mQ3^2 x12^2-3 mQ3^2 xsq2^2)+mQ3^2 (5 mQ3^6 x12^6+3 mQ3^6 x12^4 xsq2^2))))/((mQ3^2-mQ3^2 x12^2)^3 (-mQ3^2+mQ3^2 x32^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)) xt^4+1/3 mQ3^5 ((64 mQ3^3 x32^3 (87 mQ3^8 x12^4 (mQ3^2+mQ3^2 x12^2)-142 mQ3^6 x12^2 (mQ3^2+mQ3^2 x12^2)^2 x32^2+mQ3^4 (63 mQ3^6+283 mQ3^6 x12^2+283 mQ3^6 x12^4+63 mQ3^6 x12^6) x32^4-94 mQ3^6 (mQ3^2+mQ3^2 x12^2)^2 x32^6+39 mQ3^8 (mQ3^2+mQ3^2 x12^2) x32^8))/((mQ3^2-mQ3^2 x12^2)^3 (-mQ3^2+mQ3^2 x32^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)+(960 (mQ3^2+mQ3^2 x12^2) (mQ3^2 x32^2-mQ3^2 xsq2^2) (-2 mQ3^5 x12^2 x32+mQ3^5 x32^3+mQ3^5 x12^2 x32^3+mQ3^5 x32 xsq2^2+mQ3^5 x12^2 x32 xsq2^2-2 mQ3^5 x32^3 xsq2^2))/((mQ3^2-mQ3^2 x12^2)^3 (-mQ3^2+mQ3^2 x32^2)^2 (-mQ3^2 x12^2+mQ3^2 x32^2)^2)) xt^5+(1280 (1+x12^2) x32^6 xt^6)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)) Log[x32^2]^2+((80 (-13 x12^6+40 x12^4 x32^2+27 x12^6 x32^2-44 x12^2 x32^4-83 x12^4 x32^4-14 x12^6 x32^4+17 x32^6+91 x12^2 x32^6+43 x12^4 x32^6-35 x32^8-47 x12^2 x32^8+18 x32^10-3 x12^4 xsq2^2-3 x12^6 xsq2^2-24 x12^2 x32^2 xsq2^2+15 x12^4 x32^2 xsq2^2-9 x12^6 x32^2 xsq2^2+3 x32^4 xsq2^2+39 x12^2 x32^4 xsq2^2+24 x12^4 x32^4 xsq2^2-3 x32^6 xsq2^2-51 x12^2 x32^6 xsq2^2+12 x32^8 xsq2^2+3 x12^2 xsq2^4+9 x32^2 xsq2^4-6 x12^2 x32^2 xsq2^4-18 x32^4 xsq2^4+3 x12^2 x32^4 xsq2^4+9 x32^6 xsq2^4))/(3 (-1+x32^2)^2 (-x12^2+x32^2)^3)-(640 x32 (-x12^4+2 x12^2 x32^2+x12^4 x32^2-x32^4-2 x12^2 x32^4+x32^6+9 x12^2 xsq2^2-3 x12^4 xsq2^2-3 x32^2 xsq2^2-3 x12^2 x32^2 xsq2^2-3 xsq2^4+3 x32^2 xsq2^4) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2))+(160 (14 x12^4-31 x12^2 x32^2-29 x12^4 x32^2+17 x32^4+64 x12^2 x32^4+15 x12^4 x32^4-35 x32^6-33 x12^2 x32^6+18 x32^8+3 x12^2 xsq2^2+3 x12^4 xsq2^2+9 x32^2 xsq2^2-12 x12^2 x32^2 xsq2^2+9 x12^4 x32^2 xsq2^2-15 x32^4 xsq2^2-15 x12^2 x32^4 xsq2^2+18 x32^6 xsq2^2) xt^2)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^2)+(640 x32 (-x12^2-3 x12^4+x32^2+4 x12^2 x32^2+3 x12^4 x32^2-x32^4-3 x12^2 x32^4+6 xsq2^2-12 x12^2 xsq2^2+6 x12^4 xsq2^2) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2) (-1+x32^2))-(80 (14 x12^6+26 x12^8-43 x12^4 x32^2-104 x12^6 x32^2-53 x12^8 x32^2+46 x12^2 x32^4+155 x12^4 x32^4+168 x12^6 x32^4+27 x12^8 x32^4-17 x32^6-108 x12^2 x32^6-181 x12^4 x32^6-78 x12^6 x32^6+31 x32^8+76 x12^2 x32^8+69 x12^4 x32^8-10 x32^10-14 x12^2 x32^10-4 x32^12+3 x12^4 xsq2^2+6 x12^6 xsq2^2+3 x12^8 xsq2^2-30 x12^2 x32^2 xsq2^2+27 x12^4 x32^2 xsq2^2-6 x12^6 x32^2 xsq2^2+9 x12^8 x32^2 xsq2^2-21 x32^4 xsq2^2+72 x12^2 x32^4 xsq2^2-99 x12^4 x32^4 xsq2^2-24 x12^6 x32^4 xsq2^2+39 x32^6 xsq2^2-12 x12^2 x32^6 xsq2^2+69 x12^4 x32^6 xsq2^2-30 x32^8 xsq2^2-6 x12^2 x32^8 xsq2^2+6 x12^2 xsq2^4-6 x12^4 xsq2^4+18 x32^2 xsq2^4-30 x12^2 x32^2 xsq2^4+12 x12^4 x32^2 xsq2^4-36 x32^4 xsq2^4+42 x12^2 x32^4 xsq2^4-6 x12^4 x32^4 xsq2^4+18 x32^6 xsq2^4-18 x12^2 x32^6 xsq2^4) xt^4)/(3 (-1+x12^2)^3 (x12^2-x32^2)^3 (-1+x32^2)^2)+(640 x32 (-x12^4+x12^2 x32^2+x12^4 x32^2-x12^2 x32^4+9 x12^2 xsq2^2-3 x12^4 xsq2^2+3 x32^2 xsq2^2-9 x12^2 x32^2 xsq2^2-6 xsq2^4+6 x32^2 xsq2^4) xt^5)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2))) Log[xsq2^2]+(80+(160 xt^2)/(-1+x12^2)+(-(80/(-1+x12^2)^3)-(80 x12^2)/(-1+x12^2)^3) xt^4) Log[xsq2^2]^2+Log[x32^2] (-((8 (-84 x12^10+84 x12^12-3 x12^6 x32^2+540 x12^8 x32^2-84 x12^10 x32^2-456 x12^12 x32^2+3 x12^14 x32^2-6 x12^4 x32^4-1019 x12^6 x32^4-1522 x12^8 x32^4+2026 x12^10 x32^4+640 x12^12 x32^4+9 x12^14 x32^4+9 x12^2 x32^6+972 x12^4 x32^6+3698 x12^6 x32^6-1682 x12^8 x32^6-3339 x12^10 x32^6-426 x12^12 x32^6-547 x12^2 x32^8-3528 x12^4 x32^8-1602 x12^6 x32^8+5506 x12^8 x32^8+2057 x12^10 x32^8+34 x12^12 x32^8+2 x32^10+1902 x12^2 x32^10+3080 x12^4 x32^10-3634 x12^6 x32^10-3774 x12^8 x32^10-136 x12^10 x32^10-6 x32^12-2188 x12^2 x32^12+610 x12^4 x32^12+3332 x12^6 x32^12+172 x12^8 x32^12+6 x32^14+760 x12^2 x32^14-1462 x12^4 x32^14-72 x12^6 x32^14-2 x32^16+128 x12^2 x32^16+2 x12^4 x32^16-30 x12^6 x32^2 xsq2^2+30 x12^8 x32^2 xsq2^2-30 x12^10 x32^2 xsq2^2+30 x12^12 x32^2 xsq2^2-60 x12^4 x32^4 xsq2^2+150 x12^6 x32^4 xsq2^2+30 x12^8 x32^4 xsq2^2-210 x12^10 x32^4 xsq2^2+90 x12^12 x32^4 xsq2^2+90 x12^2 x32^6 xsq2^2+90 x12^4 x32^6 xsq2^2-450 x12^6 x32^6 xsq2^2+630 x12^8 x32^6 xsq2^2-360 x12^10 x32^6 xsq2^2-270 x12^2 x32^8 xsq2^2+210 x12^4 x32^8 xsq2^2-450 x12^6 x32^8 xsq2^2+510 x12^8 x32^8 xsq2^2+240 x12^2 x32^10 xsq2^2+180 x12^4 x32^10 xsq2^2-420 x12^6 x32^10 xsq2^2-180 x12^2 x32^12 xsq2^2+180 x12^4 x32^12 xsq2^2))/(3 x12^2 (-1+x12^2) (x12^2-x32^2)^4 (-1+x32^2)^3))-(64 (12 x12^6 x32-3 x12^2 x32^3-37 x12^4 x32^3+35 x12^6 x32^3+3 x12^8 x32^3+3 x32^5+68 x12^2 x32^5-111 x12^4 x32^5-86 x12^6 x32^5-123 x32^7+106 x12^2 x32^7+223 x12^4 x32^7-16 x12^6 x32^7+130 x32^9-216 x12^2 x32^9+32 x12^4 x32^9-4 x32^11-16 x12^2 x32^11-30 x12^2 x32^3 xsq2^2+30 x12^6 x32^3 xsq2^2+30 x32^5 xsq2^2+60 x12^2 x32^5 xsq2^2-90 x12^4 x32^5 xsq2^2-60 x32^7 xsq2^2+60 x12^2 x32^7 xsq2^2) xt)/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^2)-(16 (96 x12^8+3 x12^4 x32^2-513 x12^6 x32^2-489 x12^8 x32^2+3 x12^10 x32^2+9 x12^2 x32^4+850 x12^4 x32^4+2430 x12^6 x32^4+734 x12^8 x32^4+9 x12^10 x32^4-587 x12^2 x32^6-3987 x12^4 x32^6-3519 x12^6 x32^6-395 x12^8 x32^6+2 x32^8+2436 x12^2 x32^8+5696 x12^4 x32^8+1692 x12^6 x32^8-34 x12^8 x32^8-6 x32^10-3280 x12^2 x32^10-2660 x12^4 x32^10+102 x12^6 x32^10+6 x32^12+1472 x12^2 x32^12-70 x12^4 x32^12-2 x32^14+2 x12^2 x32^14+30 x12^4 x32^2 xsq2^2+30 x12^8 x32^2 xsq2^2+90 x12^2 x32^4 xsq2^2-90 x12^4 x32^4 xsq2^2-90 x12^6 x32^4 xsq2^2+90 x12^8 x32^4 xsq2^2-270 x12^2 x32^6 xsq2^2+180 x12^4 x32^6 xsq2^2-270 x12^6 x32^6 xsq2^2+240 x12^2 x32^8 xsq2^2+240 x12^4 x32^8 xsq2^2-180 x12^2 x32^10 xsq2^2) xt^2)/(3 x12^2 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^3)-(128 (-6 x12^4 x32-18 x12^6 x32+3 x32^3+91 x12^2 x32^3-106 x12^4 x32^3+115 x12^6 x32^3+3 x12^8 x32^3-153 x32^5+85 x12^2 x32^5+17 x12^4 x32^5-165 x12^6 x32^5+193 x32^7-238 x12^2 x32^7+217 x12^4 x32^7+16 x12^6 x32^7-10 x32^9-6 x12^2 x32^9-16 x12^4 x32^9-22 x32^11+30 x32^3 xsq2^2-30 x12^2 x32^3 xsq2^2-30 x12^4 x32^3 xsq2^2+30 x12^6 x32^3 xsq2^2-60 x32^5 xsq2^2+120 x12^2 x32^5 xsq2^2-60 x12^4 x32^5 xsq2^2) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)+(8 (96 x12^10+240 x12^12+3 x12^6 x32^2-616 x12^8 x32^2-1602 x12^10 x32^2-884 x12^12 x32^2+3 x12^14 x32^2+6 x12^4 x32^4+1723 x12^6 x32^4+5474 x12^8 x32^4+6600 x12^10 x32^4+1268 x12^12 x32^4+9 x12^14 x32^4-9 x12^2 x32^6-1828 x12^4 x32^6-10862 x12^6 x32^6-18070 x12^8 x32^6-9553 x12^10 x32^6-702 x12^12 x32^6+643 x12^2 x32^8+9676 x12^4 x32^8+25638 x12^6 x32^8+22990 x12^8 x32^8+4623 x12^10 x32^8-34 x12^12 x32^8-2 x32^10-2996 x12^2 x32^10-17588 x12^4 x32^10-25584 x12^6 x32^10-9926 x12^8 x32^10+136 x12^10 x32^10+6 x32^12+4406 x12^2 x32^12+12962 x12^4 x32^12+9110 x12^6 x32^12-172 x12^8 x32^12-6 x32^14-2152 x12^2 x32^14-3178 x12^4 x32^14+72 x12^6 x32^14+2 x32^16+80 x12^2 x32^16-2 x12^4 x32^16+30 x12^6 x32^2 xsq2^2+30 x12^8 x32^2 xsq2^2+30 x12^10 x32^2 xsq2^2+30 x12^12 x32^2 xsq2^2+60 x12^4 x32^4 xsq2^2-30 x12^6 x32^4 xsq2^2-210 x12^8 x32^4 xsq2^2-30 x12^10 x32^4 xsq2^2+90 x12^12 x32^4 xsq2^2-90 x12^2 x32^6 xsq2^2-270 x12^4 x32^6 xsq2^2+90 x12^6 x32^6 xsq2^2-90 x12^8 x32^6 xsq2^2-360 x12^10 x32^6 xsq2^2+270 x12^2 x32^8 xsq2^2+330 x12^4 x32^8 xsq2^2+570 x12^6 x32^8 xsq2^2+510 x12^8 x32^8 xsq2^2-240 x12^2 x32^10 xsq2^2-660 x12^4 x32^10 xsq2^2-420 x12^6 x32^10 xsq2^2+180 x12^2 x32^12 xsq2^2+180 x12^4 x32^12 xsq2^2) xt^4)/(3 x12^2 (-1+x12^2)^3 (x12^2-x32^2)^4 (-1+x32^2)^3)+(64 (12 x12^6 x32+3 x12^2 x32^3+213 x12^4 x32^3+179 x12^6 x32^3+3 x12^8 x32^3-3 x32^5-407 x12^2 x32^5-737 x12^4 x32^5-219 x12^6 x32^5+214 x32^7+770 x12^2 x32^7+558 x12^4 x32^7+16 x12^6 x32^7-276 x32^9-362 x12^2 x32^9-32 x12^4 x32^9+52 x32^11+16 x12^2 x32^11+30 x12^2 x32^3 xsq2^2+60 x12^4 x32^3 xsq2^2+30 x12^6 x32^3 xsq2^2-30 x32^5 xsq2^2-120 x12^2 x32^5 xsq2^2-90 x12^4 x32^5 xsq2^2+60 x32^7 xsq2^2+60 x12^2 x32^7 xsq2^2) xt^5)/(3 (-1+x12^2)^3 (x12^2-x32^2)^3 (-1+x32^2)^2)+(5120 x12^2 x32^4 xt^6)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2))+((80 (x12^2 x32^4-x12^6 x32^4-x32^6-3 x12^2 x32^6+3 x12^4 x32^6+x12^6 x32^6+3 x32^8-3 x12^4 x32^8-2 x32^10+2 x12^2 x32^10+18 x12^2 x32^2 xsq2^2+18 x12^6 x32^2 xsq2^2+6 x32^4 xsq2^2-54 x12^2 x32^4 xsq2^2-54 x12^4 x32^4 xsq2^2+6 x12^6 x32^4 xsq2^2-18 x32^6 xsq2^2+108 x12^2 x32^6 xsq2^2-18 x12^4 x32^6 xsq2^2-12 x32^10 xsq2^2-3 x12^2 xsq2^4-3 x12^6 xsq2^4-9 x32^2 xsq2^4+9 x12^2 x32^2 xsq2^4+9 x12^4 x32^2 xsq2^4-9 x12^6 x32^2 xsq2^4+27 x32^4 xsq2^4-18 x12^2 x32^4 xsq2^4+27 x12^4 x32^4 xsq2^4-24 x32^6 xsq2^4-24 x12^2 x32^6 xsq2^4+18 x32^8 xsq2^4))/(3 (-1+x32^2)^3 (-x12^2+x32^2)^3)-(320 (-x12^2 x32^3-x12^4 x32^3+x32^5+4 x12^2 x32^5+x12^4 x32^5-3 x32^7-3 x12^2 x32^7+2 x32^9-12 x12^2 x32 xsq2^2+12 x12^4 x32 xsq2^2+12 x32^5 xsq2^2-12 x12^2 x32^5 xsq2^2+6 x32 xsq2^4-6 x12^4 x32 xsq2^4-12 x32^3 xsq2^4+12 x12^2 x32^3 xsq2^4) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^2)-(160 (-x12^2 x32^4-x12^6 x32^4+x32^6+3 x12^2 x32^6+3 x12^4 x32^6+x12^6 x32^6-3 x32^8-6 x12^2 x32^8-3 x12^4 x32^8+4 x32^10+4 x12^2 x32^10-2 x32^12+18 x12^2 x32^2 xsq2^2+18 x12^6 x32^2 xsq2^2+6 x32^4 xsq2^2-54 x12^2 x32^4 xsq2^2-54 x12^4 x32^4 xsq2^2+6 x12^6 x32^4 xsq2^2-18 x32^6 xsq2^2+108 x12^2 x32^6 xsq2^2-18 x12^4 x32^6 xsq2^2-12 x32^10 xsq2^2-3 x12^2 xsq2^4-3 x12^6 xsq2^4-9 x32^2 xsq2^4+9 x12^2 x32^2 xsq2^4+9 x12^4 x32^2 xsq2^4-9 x12^6 x32^2 xsq2^4+27 x32^4 xsq2^4-18 x12^2 x32^4 xsq2^4+27 x12^4 x32^4 xsq2^4-24 x32^6 xsq2^4-24 x12^2 x32^6 xsq2^4+18 x32^8 xsq2^4) xt^2)/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^3)-(640 (-x12^2 x32^3+4 x12^4 x32^3-x12^6 x32^3+x32^5-5 x12^2 x32^5-5 x12^4 x32^5+x12^6 x32^5+x32^7+10 x12^2 x32^7+x12^4 x32^7-4 x32^9-4 x12^2 x32^9+2 x32^11+12 x12^2 x32 xsq2^2-24 x12^4 x32 xsq2^2+12 x12^6 x32 xsq2^2-12 x32^5 xsq2^2+24 x12^2 x32^5 xsq2^2-12 x12^4 x32^5 xsq2^2-6 x32 xsq2^4+6 x12^2 x32 xsq2^4+6 x12^4 x32 xsq2^4-6 x12^6 x32 xsq2^4+12 x32^3 xsq2^4-24 x12^2 x32^3 xsq2^4+12 x12^4 x32^3 xsq2^4) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)+(80 (1+x12^2) (-x12^2 x32^4-x12^6 x32^4+x32^6+3 x12^2 x32^6+3 x12^4 x32^6+x12^6 x32^6-3 x32^8-6 x12^2 x32^8-3 x12^4 x32^8+4 x32^10+4 x12^2 x32^10-2 x32^12+18 x12^2 x32^2 xsq2^2+18 x12^6 x32^2 xsq2^2+6 x32^4 xsq2^2-54 x12^2 x32^4 xsq2^2-54 x12^4 x32^4 xsq2^2+6 x12^6 x32^4 xsq2^2-18 x32^6 xsq2^2+108 x12^2 x32^6 xsq2^2-18 x12^4 x32^6 xsq2^2-12 x32^10 xsq2^2-3 x12^2 xsq2^4-3 x12^6 xsq2^4-9 x32^2 xsq2^4+9 x12^2 x32^2 xsq2^4+9 x12^4 x32^2 xsq2^4-9 x12^6 x32^2 xsq2^4+27 x32^4 xsq2^4-18 x12^2 x32^4 xsq2^4+27 x12^4 x32^4 xsq2^4-24 x32^6 xsq2^4-24 x12^2 x32^6 xsq2^4+18 x32^8 xsq2^4) xt^4)/(3 (-1+x12^2)^3 (x12^2-x32^2)^3 (-1+x32^2)^3)+(320 (-x12^2 x32^3-x12^4 x32^3+x32^5+2 x12^2 x32^5+x12^4 x32^5-x32^7-x12^2 x32^7+12 x12^2 x32 xsq2^2+12 x12^4 x32 xsq2^2-12 x32^5 xsq2^2-12 x12^2 x32^5 xsq2^2-6 x32 xsq2^4-12 x12^2 x32 xsq2^4-6 x12^4 x32 xsq2^4+12 x32^3 xsq2^4+12 x12^2 x32^3 xsq2^4) xt^5)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)) Log[xsq2^2])+((8 (-1+x12^2) (3 x12^2+7 x12^4+7 x12^6+3 x12^8+9 x32^2-26 x12^2 x32^2-34 x12^4 x32^2-26 x12^6 x32^2+9 x12^8 x32^2-29 x32^4+65 x12^2 x32^4+65 x12^4 x32^4-29 x12^6 x32^4+34 x32^6-76 x12^2 x32^6+34 x12^4 x32^6-14 x32^8-14 x12^2 x32^8+12 x32^10))/(3 (x12^2-x32^2)^3 (-1+x32^2)^3)+(64 (-1+x12^2) x32 (3-x12^2+3 x12^4-5 x32^2-5 x12^2 x32^2+5 x32^4) xt)/(3 (x12^2-x32^2)^2 (-1+x32^2)^2)-(16 (3 x12^2+7 x12^4+7 x12^6+3 x12^8+9 x32^2-26 x12^2 x32^2-34 x12^4 x32^2-26 x12^6 x32^2+9 x12^8 x32^2-29 x32^4+65 x12^2 x32^4+65 x12^4 x32^4-29 x12^6 x32^4+34 x32^6-76 x12^2 x32^6+34 x12^4 x32^6-14 x32^8-14 x12^2 x32^8+12 x32^10) xt^2)/(3 (-1+x32^2)^3 (-x12^2+x32^2)^3)+(128 x32 (3-x12^2+3 x12^4-5 x32^2-5 x12^2 x32^2+5 x32^4) xt^3)/(3 (-1+x32^2)^2 (-x12^2+x32^2)^2)-(8 (1+x12^2) (3 x12^2+7 x12^4+7 x12^6+3 x12^8+9 x32^2-26 x12^2 x32^2-34 x12^4 x32^2-26 x12^6 x32^2+9 x12^8 x32^2-29 x32^4+65 x12^2 x32^4+65 x12^4 x32^4-29 x12^6 x32^4+34 x32^6-76 x12^2 x32^6+34 x12^4 x32^6-14 x32^8-14 x12^2 x32^8+12 x32^10) xt^4)/(3 (-1+x12^2)^2 (x12^2-x32^2)^3 (-1+x32^2)^3)-(64 (1+x12^2) x32 (3-x12^2+3 x12^4-5 x32^2-5 x12^2 x32^2+5 x32^4) xt^5)/(3 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)) PolyLog[2,1-x12^2]+(-((8 (-3 x12^2-x12^4-19 x12^6+23 x12^8-9 x32^2+41 x12^2 x32^2+41 x12^4 x32^2-167 x12^6 x32^2-34 x12^8 x32^2+20 x32^4-151 x12^2 x32^4+417 x12^4 x32^4+323 x12^6 x32^4+31 x12^8 x32^4+9 x32^6-251 x12^2 x32^6-821 x12^4 x32^6-217 x12^6 x32^6+98 x32^8+762 x12^2 x32^8+420 x12^4 x32^8-290 x32^10-350 x12^2 x32^10+128 x32^12))/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^2))-(64 x32 (3-7 x12^2+22 x12^4-5 x32^2-23 x12^2 x32^2-78 x12^4 x32^2+19 x32^4+135 x12^2 x32^4+24 x12^4 x32^4-37 x32^6-73 x12^2 x32^6+20 x32^8) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^2)+(16 (3 x12^2+x12^4+19 x12^6-23 x12^8+9 x32^2-41 x12^2 x32^2-41 x12^4 x32^2+167 x12^6 x32^2+34 x12^8 x32^2-20 x32^4+151 x12^2 x32^4-417 x12^4 x32^4-903 x12^6 x32^4+37 x12^8 x32^4-9 x32^6+251 x12^2 x32^6+2049 x12^4 x32^6+525 x12^6 x32^6-98 x32^8-1414 x12^2 x32^8-1304 x12^4 x32^8+294 x32^10+858 x12^2 x32^10-128 x32^12) xt^2)/(3 (-1+x12^2)^2 (x12^2-x32^2)^3 (-1+x32^2)^2)-(64 x32 (-6+20 x12^2-75 x12^4+27 x12^6-2 x32^2+110 x12^2 x32^2-6 x12^4 x32^2-37 x32^4-33 x12^2 x32^4+2 x32^6) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2)-(8 (3 x12^2+4 x12^4+54 x12^6-4 x12^8-57 x12^10+9 x32^2-32 x12^2 x32^2-184 x12^4 x32^2+502 x12^6 x32^2+883 x12^8 x32^2+102 x12^10 x32^2-20 x32^4+201 x12^2 x32^4-882 x12^4 x32^4-3984 x12^6 x32^4-1718 x12^8 x32^4+3 x12^10 x32^4-11 x32^6+474 x12^2 x32^6+5516 x12^4 x32^6+6126 x12^6 x32^6+695 x12^8 x32^6-90 x32^8-2702 x12^2 x32^8-7406 x12^4 x32^8-2602 x12^6 x32^8+284 x32^10+3068 x12^2 x32^10+3048 x12^4 x32^10-124 x32^12-1156 x12^2 x32^12) xt^4)/(3 (-1+x12^2)^4 (x12^2-x32^2)^3 (-1+x32^2)^2)+(64 x32 (3-4 x12^2-x12^4+6 x12^6+x32^2-20 x12^2 x32^2+11 x12^4 x32^2+18 x32^4-14 x12^2 x32^4) xt^5)/(3 (-1+x12^2)^4 (x12^2-x32^2)^2)) PolyLog[2,1-x32^2]+((8 (-23 x12^8+19 x12^10+x12^12+3 x12^14+80 x12^6 x32^2+129 x12^8 x32^2-43 x12^10 x32^2-47 x12^12 x32^2+9 x12^14 x32^2-122 x12^4 x32^4-638 x12^6 x32^4-334 x12^8 x32^4+236 x12^10 x32^4-38 x12^12 x32^4+96 x12^2 x32^6+1030 x12^4 x32^6+1614 x12^6 x32^6-92 x12^8 x32^6+40 x12^10 x32^6+33 x32^8-821 x12^2 x32^8-2543 x12^4 x32^8-1049 x12^6 x32^8-100 x12^8 x32^8+25 x32^10+1981 x12^2 x32^10+2061 x12^4 x32^10+413 x12^6 x32^10-292 x32^12-1718 x12^2 x32^12-678 x12^4 x32^12+350 x32^14+546 x12^2 x32^14-128 x32^16))/(3 (-1+x12^2) (x12^2-x32^2)^4 (-1+x32^2)^3)+(64 (22 x12^6 x32-7 x12^8 x32+3 x12^10 x32-100 x12^4 x32^3-16 x12^6 x32^3-8 x12^8 x32^3+166 x12^2 x32^5+158 x12^4 x32^5+24 x12^6 x32^5-24 x32^7-304 x12^2 x32^7-88 x12^4 x32^7+41 x32^9+153 x12^2 x32^9-20 x32^11) xt)/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^2)+(16 (-23 x12^4+19 x12^6+x12^8+3 x12^10+34 x12^2 x32^2+167 x12^4 x32^2-41 x12^6 x32^2-41 x12^8 x32^2+9 x12^10 x32^2+37 x32^4-903 x12^2 x32^4-417 x12^4 x32^4+151 x12^6 x32^4-20 x12^8 x32^4+13 x32^6+2561 x12^2 x32^6+251 x12^4 x32^6-9 x12^6 x32^6-280 x32^8-2438 x12^2 x32^8-98 x12^4 x32^8+346 x32^10+806 x12^2 x32^10-128 x32^12) xt^2)/(3 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^3)+(64 (-61 x12^4 x32+41 x12^6 x32-20 x12^8 x32+6 x12^10 x32+122 x12^2 x32^3+60 x12^4 x32^3-2 x12^6 x32^3-10 x12^8 x32^3-45 x32^5-227 x12^2 x32^5-57 x12^4 x32^5+21 x12^6 x32^5+78 x32^7+168 x12^2 x32^7-2 x12^4 x32^7-41 x32^9-33 x12^2 x32^9+2 x32^11) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)-(8 (11 x12^4-4 x12^6-14 x12^8+4 x12^10+3 x12^12-34 x12^2 x32^2+543 x12^4 x32^2+774 x12^6 x32^2+20 x12^8 x32^2-32 x12^10 x32^2+9 x12^12 x32^2+39 x32^4-526 x12^2 x32^4-3712 x12^4 x32^4-2210 x12^6 x32^4+29 x12^8 x32^4-20 x12^10 x32^4+3 x32^6+2990 x12^2 x32^6+7596 x12^4 x32^6+2186 x12^6 x32^6+25 x12^8 x32^6-262 x32^8-5326 x12^2 x32^8-6466 x12^4 x32^8-746 x12^6 x32^8+332 x32^10+4100 x12^2 x32^10+1968 x12^4 x32^10-124 x32^12-1156 x12^2 x32^12) xt^4)/(3 (-1+x12^2)^4 (x12^2-x32^2)^2 (-1+x32^2)^3)-(64 (6 x12^2 x32-x12^4 x32-4 x12^6 x32+3 x12^8 x32-22 x32^3+12 x12^2 x32^3-2 x12^6 x32^3+37 x32^5-26 x12^2 x32^5+x12^4 x32^5-18 x32^7+14 x12^2 x32^7) xt^5)/(3 (-1+x12^2)^4 (x12^2-x32^2) (-1+x32^2)^2)) PolyLog[2,1-x32^2/x12^2]+(-((80 (-1+xsq2^2) (-1+3 x32^2+2 x32^4-xsq2^2-3 x32^2 xsq2^2))/(-1+x32^2)^3)+(640 x32 (-1+xsq2^2)^2 xt)/((-1+x12^2) (-1+x32^2)^2)-(160 (-1+xsq2^2) (-1+3 x32^2+2 x32^4-xsq2^2-3 x32^2 xsq2^2) xt^2)/((-1+x12^2) (-1+x32^2)^3)+(1280 x32 (-1+xsq2^2)^2 xt^3)/((-1+x12^2)^2 (-1+x32^2)^2)+(80 (1+x12^2) (-1+xsq2^2) (-1+3 x32^2+2 x32^4-xsq2^2-3 x32^2 xsq2^2) xt^4)/((-1+x12^2)^3 (-1+x32^2)^3)-(640 (1+x12^2) x32 (-1+xsq2^2)^2 xt^5)/((-1+x12^2)^4 (-1+x32^2)^2)) PolyLog[2,1-xsq2^2]+((80 (x12^2-xsq2^2) (x12^4-3 x12^2 x32^2-2 x32^4+x12^2 xsq2^2+3 x32^2 xsq2^2))/(x12^2-x32^2)^3-(640 x32 (x12^2-xsq2^2)^2 xt)/((-1+x12^2) (x12^2-x32^2)^2)+(160 (x12^2-xsq2^2) (x12^4-3 x12^2 x32^2-2 x32^4+x12^2 xsq2^2+3 x32^2 xsq2^2) xt^2)/((-1+x12^2) (x12^2-x32^2)^3)-(1280 x32 (x12^2-xsq2^2)^2 xt^3)/((-1+x12^2)^2 (x12^2-x32^2)^2)-(80 (1+x12^2) (x12^2-xsq2^2) (x12^4-3 x12^2 x32^2-2 x32^4+x12^2 xsq2^2+3 x32^2 xsq2^2) xt^4)/((-1+x12^2)^3 (x12^2-x32^2)^3)+(640 (1+x12^2) x32 (x12^2-xsq2^2)^2 xt^5)/((-1+x12^2)^4 (x12^2-x32^2)^2)) PolyLog[2,1-xsq2^2/x12^2]+((80 (x32^2-xsq2^2) (-5 x12^2 x32^2-5 x12^6 x32^2+x32^4+15 x12^2 x32^4+15 x12^4 x32^4+x12^6 x32^4-3 x32^6-30 x12^2 x32^6-3 x12^4 x32^6+8 x32^8+8 x12^2 x32^8-2 x32^10+x12^2 xsq2^2+x12^6 xsq2^2+3 x32^2 xsq2^2-3 x12^2 x32^2 xsq2^2-3 x12^4 x32^2 xsq2^2+3 x12^6 x32^2 xsq2^2-9 x32^4 xsq2^2+6 x12^2 x32^4 xsq2^2-9 x12^4 x32^4 xsq2^2+8 x32^6 xsq2^2+8 x12^2 x32^6 xsq2^2-6 x32^8 xsq2^2))/((x12^2-x32^2)^3 (-1+x32^2)^3)+(640 (x32^2-xsq2^2) (-2 x12^2 x32+x32^3+x12^2 x32^3+x32 xsq2^2+x12^2 x32 xsq2^2-2 x32^3 xsq2^2) xt)/((-1+x32^2)^2 (-x12^2+x32^2)^2)+(160 (x32^2-xsq2^2) (-5 x12^2 x32^2-5 x12^6 x32^2+x32^4+15 x12^2 x32^4+15 x12^4 x32^4+x12^6 x32^4-3 x32^6-30 x12^2 x32^6-3 x12^4 x32^6+8 x32^8+8 x12^2 x32^8-2 x32^10+x12^2 xsq2^2+x12^6 xsq2^2+3 x32^2 xsq2^2-3 x12^2 x32^2 xsq2^2-3 x12^4 x32^2 xsq2^2+3 x12^6 x32^2 xsq2^2-9 x32^4 xsq2^2+6 x12^2 x32^4 xsq2^2-9 x12^4 x32^4 xsq2^2+8 x32^6 xsq2^2+8 x12^2 x32^6 xsq2^2-6 x32^8 xsq2^2) xt^2)/((-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^3)+(1280 (x32^2-xsq2^2) (-2 x12^2 x32+x32^3+x12^2 x32^3+x32 xsq2^2+x12^2 x32 xsq2^2-2 x32^3 xsq2^2) xt^3)/((-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^2)-(80 (1+x12^2) (x32^2-xsq2^2) (-5 x12^2 x32^2-5 x12^6 x32^2+x32^4+15 x12^2 x32^4+15 x12^4 x32^4+x12^6 x32^4-3 x32^6-30 x12^2 x32^6-3 x12^4 x32^6+8 x32^8+8 x12^2 x32^8-2 x32^10+x12^2 xsq2^2+x12^6 xsq2^2+3 x32^2 xsq2^2-3 x12^2 x32^2 xsq2^2-3 x12^4 x32^2 xsq2^2+3 x12^6 x32^2 xsq2^2-9 x32^4 xsq2^2+6 x12^2 x32^4 xsq2^2-9 x12^4 x32^4 xsq2^2+8 x32^6 xsq2^2+8 x12^2 x32^6 xsq2^2-6 x32^8 xsq2^2) xt^4)/((-1+x12^2)^3 (x12^2-x32^2)^3 (-1+x32^2)^3)-(640 (1+x12^2) (x32^2-xsq2^2) (-2 x12^2 x32+x32^3+x12^2 x32^3+x32 xsq2^2+x12^2 x32 xsq2^2-2 x32^3 xsq2^2) xt^5)/((-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)) PolyLog[2,1-xsq2^2/x32^2])+Log[MR^2/mQ3^2] (-((8 (-211 mQ3^12 x12^6+784 mQ3^12 x12^4 x32^2+784 mQ3^12 x12^6 x32^2-689 mQ3^12 x12^2 x32^4-2356 mQ3^12 x12^4 x32^4-689 mQ3^12 x12^6 x32^4+124 mQ3^12 x32^6+1836 mQ3^12 x12^2 x32^6+1836 mQ3^12 x12^4 x32^6+124 mQ3^12 x12^6 x32^6-280 mQ3^12 x32^8-1295 mQ3^12 x12^2 x32^8-280 mQ3^12 x12^4 x32^8+156 mQ3^12 x32^10+156 mQ3^12 x12^2 x32^10))/(3 mQ3^4 x12^2 (mQ3^2-mQ3^2 x32^2)^2 (-mQ3^2 x12^2+mQ3^2 x32^2)^2))+(128 x32^3 xt)/(3 (-1+x32^2) (-x12^2+x32^2))+(64 (-23 x12^2 x32^2+31 x32^4+31 x12^2 x32^4-39 x32^6) xt^2)/(3 x12^2 (x12^2-x32^2) (-1+x32^2))+(256 (-51 x12^2 x32+59 x32^3+59 x12^2 x32^3-67 x32^5) xt^3)/(3 (-1+x12^2)^2 (x12^2-x32^2) (-1+x32^2))-(32 (338 x12^6-756 x12^4 x32^2-756 x12^6 x32^2+383 x12^2 x32^4+1656 x12^4 x32^4+383 x12^6 x32^4+31 x32^6-822 x12^2 x32^6-822 x12^4 x32^6+31 x12^6 x32^6-70 x32^8+396 x12^2 x32^8-70 x12^4 x32^8+39 x32^10+39 x12^2 x32^10) xt^4)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)+(256 x32^3 xt^5)/(3 (-1+x12^2)^2 (x12^2-x32^2) (-1+x32^2))+((16 (-97 x12^6+97 x12^8+291 x12^4 x32^2-227 x12^6 x32^2-242 x12^2 x32^4+114 x12^4 x32^4+52 x32^6+12 x12^2 x32^6))/(3 (-1+x12^2) (x12^2-x32^2)^3)-(64 (95 x12^4 x32-150 x12^2 x32^3+41 x32^5) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2)+(32 (-97 x12^6+153 x12^8+291 x12^4 x32^2-267 x12^6 x32^2-186 x12^2 x32^4-30 x12^4 x32^4-4 x32^6+140 x12^2 x32^6) xt^2)/(3 (-1+x12^2)^2 (x12^2-x32^2)^3)-(64 (-201 x12^4 x32+111 x12^6 x32+210 x12^2 x32^3+26 x12^4 x32^3+3 x32^5-173 x12^2 x32^5+56 x32^7) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2)-(16 (-69 x12^6+112 x12^8+181 x12^10+207 x12^4 x32^2-106 x12^6 x32^2-133 x12^8 x32^2-102 x12^2 x32^4-34 x12^4 x32^4-472 x12^6 x32^4-32 x32^6-30 x12^2 x32^6+478 x12^4 x32^6+58 x32^8-58 x12^2 x32^8) xt^4)/(3 (-1+x12^2)^4 (x12^2-x32^2)^3)+(128 (39 x12^4 x32+39 x12^6 x32-30 x12^2 x32^3-14 x12^4 x32^3-8 x32^5-24 x12^2 x32^5) xt^5)/(3 (-1+x12^2)^4 (x12^2-x32^2)^2)) Log[x12^2]^2+((16 (-12 x12^6 x32^2-12 x12^8 x32^2+117 x12^4 x32^4+136 x12^6 x32^4+117 x12^8 x32^4-185 x12^2 x32^6-531 x12^4 x32^6-531 x12^6 x32^6-185 x12^8 x32^6+84 x32^8+647 x12^2 x32^8+1054 x12^4 x32^8+647 x12^6 x32^8+84 x12^8 x32^8-252 x32^10-896 x12^2 x32^10-896 x12^4 x32^10-252 x12^6 x32^10+268 x32^12+530 x12^2 x32^12+268 x12^4 x32^12-100 x32^14-100 x12^2 x32^14))/(3 x12^2 (x12^2-x32^2)^3 (-1+x32^2)^3)+(128 (50 x12^4 x32^3-57 x12^2 x32^5-57 x12^4 x32^5+8 x32^7+64 x12^2 x32^7+8 x12^4 x32^7-8 x32^9-8 x12^2 x32^9) xt)/(3 x12^2 (x12^2-x32^2)^2 (-1+x32^2)^2)-(128 (3 x12^4 x32^2-20 x12^2 x32^4-20 x12^4 x32^4+21 x32^6+54 x12^2 x32^6+21 x12^4 x32^6-42 x32^8-42 x12^2 x32^8+25 x32^10) xt^2)/(3 x12^2 (x12^2-x32^2)^2 (-1+x32^2)^2)-(512 (3 x12^6 x32-6 x12^4 x32^3-6 x12^6 x32^3+3 x12^2 x32^5+20 x12^4 x32^5+3 x12^6 x32^5+4 x32^7-18 x12^2 x32^7-18 x12^4 x32^7+4 x12^6 x32^7-4 x32^9+19 x12^2 x32^9-4 x12^4 x32^9) xt^3)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)+(32 (-34 x12^6 x32^2-34 x12^8 x32^2-3 x12^4 x32^4-116 x12^6 x32^4-3 x12^8 x32^4-9 x12^2 x32^6+539 x12^4 x32^6+539 x12^6 x32^6-9 x12^8 x32^6+42 x32^8-251 x12^2 x32^8-1250 x12^4 x32^8-251 x12^6 x32^8+42 x12^8 x32^8-126 x32^10+568 x12^2 x32^10+568 x12^4 x32^10-126 x12^6 x32^10+134 x32^12-254 x12^2 x32^12+134 x12^4 x32^12-50 x32^14-50 x12^2 x32^14) xt^4)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^3 (-1+x32^2)^3)+(256 (-48 x12^4 x32^3+43 x12^2 x32^5+43 x12^4 x32^5+4 x32^7-38 x12^2 x32^7+4 x12^4 x32^7-4 x32^9-4 x12^2 x32^9) xt^5)/(3 x12^2 (-1+x12^2)^2 (x12^2-x32^2)^2 (-1+x32^2)^2)) Log[x32^2]+((32 (92 x12^4-259 x12^2 x32^2-259 x12^4 x32^2+163 x32^4+668 x12^2 x32^4+163 x12^4 x32^4-401 x32^6-401 x12^2 x32^6+234 x32^8))/(3 (-1+x32^2)^2 (-x12^2+x32^2)^2)+(256 x32^3 xt)/(3 (-1+x32^2) (-x12^2+x32^2))+(-((32 (-41 x12^6+41 x12^8+123 x12^4 x32^2-59 x12^6 x32^2-46 x12^2 x32^4-82 x12^4 x32^4-32 x32^6+96 x12^2 x32^6))/(3 (-1+x12^2) (x12^2-x32^2)^3))+(256 (25 x12^4 x32-16 x12^2 x32^3-8 x32^5) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2)) Log[x12^2]+(-((32 (77 x12^2 x32^4+64 x12^4 x32^4+77 x12^6 x32^4-73 x32^6-359 x12^2 x32^6-359 x12^4 x32^6-73 x12^6 x32^6+283 x32^8+718 x12^2 x32^8+283 x12^4 x32^8-424 x32^10-424 x12^2 x32^10+210 x32^12))/(3 (-1+x32^2)^3 (-x12^2+x32^2)^3))+(256 (34 x12^2 x32^3-33 x32^5-33 x12^2 x32^5+32 x32^7) xt)/(3 (-1+x32^2)^2 (-x12^2+x32^2)^2)) Log[x32^2]) Log[xEWS^2]+1312 Log[xEWS^2]^2+((160 x32^2)/3+(160 x32^2)/(3 x12^2)-(320 x32^2 xt^2)/(3 x12^2)-(1280 x32 xt^3)/(3 (-1+x12^2)^2)+((160 x32^2)/(3 (-1+x12^2)^2)+(160 x32^2)/(3 x12^2 (-1+x12^2)^2)) xt^4) Log[xsq2^2]+Log[x12^2] ((16 (327 x12^8-327 x12^10-1101 x12^6 x32^2+263 x12^8 x32^2+774 x12^10 x32^2+1209 x12^4 x32^4+1481 x12^6 x32^4-1974 x12^8 x32^4-460 x12^10 x32^4-440 x12^2 x32^6-2386 x12^4 x32^6+1006 x12^6 x32^6+1419 x12^8 x32^6+17 x12^10 x32^6+x32^8+983 x12^2 x32^8+720 x12^4 x32^8-1397 x12^6 x32^8-51 x12^8 x32^8-2 x32^10-539 x12^2 x32^10+442 x12^4 x32^10+35 x12^6 x32^10+x32^12-x12^4 x32^12))/(3 x12^2 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^2)-(128 (64 x12^4 x32-93 x12^2 x32^3-81 x12^4 x32^3+30 x32^5+119 x12^2 x32^5+8 x12^4 x32^5-39 x32^7-8 x12^2 x32^7) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2))-(32 (282 x12^6-740 x12^4 x32^2-650 x12^6 x32^2+471 x12^2 x32^4+1652 x12^4 x32^4+347 x12^6 x32^4-x32^6-1044 x12^2 x32^6-870 x12^4 x32^6+17 x12^6 x32^6+2 x32^8+568 x12^2 x32^8-34 x12^4 x32^8-x32^10+x12^2 x32^10) xt^2)/(3 x12^2 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^2)+(256 (3 x12^4 x32+48 x12^6 x32-17 x12^2 x32^3-93 x12^4 x32^3-59 x12^6 x32^3+22 x32^5+56 x12^2 x32^5+99 x12^4 x32^5+8 x12^6 x32^5-27 x32^7-32 x12^2 x32^7-8 x12^4 x32^7) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2))+(16 (200 x12^8+700 x12^10-720 x12^6 x32^2-1974 x12^8 x32^2-1486 x12^10 x32^2+1063 x12^4 x32^4+2253 x12^6 x32^4+3467 x12^8 x32^4+765 x12^10 x32^4-536 x12^2 x32^6-2052 x12^4 x32^6-2203 x12^6 x32^6-1634 x12^8 x32^6+17 x12^10 x32^6+x32^8+1051 x12^2 x32^8+600 x12^4 x32^8+635 x12^6 x32^8-51 x12^8 x32^8-2 x32^10-355 x12^2 x32^10+382 x12^4 x32^10+35 x12^6 x32^10+x32^12-156 x12^2 x32^12-x12^4 x32^12) xt^4)/(3 x12^2 (-1+x12^2)^3 (x12^2-x32^2)^3 (-1+x32^2)^2)-(128 (-78 x12^4 x32+53 x12^2 x32^3+71 x12^4 x32^3+23 x32^5-45 x12^2 x32^5+8 x12^4 x32^5-24 x32^7-8 x12^2 x32^7) xt^5)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2))+((16 (49 x12^2 x32^4+64 x12^4 x32^4+105 x12^6 x32^4-45 x32^6-275 x12^2 x32^6-443 x12^4 x32^6-101 x12^6 x32^6+199 x32^8+718 x12^2 x32^8+367 x12^4 x32^8-368 x32^10-480 x12^2 x32^10+210 x32^12))/(3 (-1+x32^2)^3 (-x12^2+x32^2)^3)+(128 (-6 x12^4 x32+32 x12^2 x32^3-36 x12^4 x32^3-33 x32^5+32 x12^2 x32^5+33 x12^4 x32^5+18 x32^7-46 x12^2 x32^7+6 x32^9) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^2)+(32 (-105 x12^2 x32^4-192 x12^4 x32^4-105 x12^6 x32^4+101 x32^6+699 x12^2 x32^6+699 x12^4 x32^6+101 x12^6 x32^6-495 x32^8-1398 x12^2 x32^8-495 x12^4 x32^8+792 x32^10+792 x12^2 x32^10-394 x32^12) xt^2)/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^3)+(256 (3 x12^4 x32+3 x12^6 x32-54 x12^2 x32^3+112 x12^4 x32^3-54 x12^6 x32^3+54 x32^5-84 x12^2 x32^5-84 x12^4 x32^5+54 x12^6 x32^5-32 x32^7+176 x12^2 x32^7-32 x12^4 x32^7-45 x32^9-45 x12^2 x32^9+28 x32^11) xt^3)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)-(16 (-24 x12^6 x32^2-105 x12^2 x32^4-353 x12^4 x32^4-353 x12^6 x32^4-105 x12^8 x32^4+101 x32^6+952 x12^2 x32^6+1694 x12^4 x32^6+952 x12^6 x32^6+101 x12^8 x32^6-567 x32^8-2221 x12^2 x32^8-2221 x12^4 x32^8-567 x12^6 x32^8+880 x32^10+1688 x12^2 x32^10+880 x12^4 x32^10-322 x32^12-322 x12^2 x32^12-88 x32^14) xt^4)/(3 (-1+x12^2)^3 (x12^2-x32^2)^3 (-1+x32^2)^3)-(128 (-48 x12^2 x32^3-48 x12^4 x32^3+47 x32^5+78 x12^2 x32^5+47 x12^4 x32^5-30 x32^7-30 x12^2 x32^7-16 x32^9) xt^5)/(3 (-1+x12^2)^3 (x12^2-x32^2)^2 (-1+x32^2)^2)) Log[x32^2]+(-((640 x32 xt)/(3 (-1+x12^2)))+((640 x32)/(3 (-1+x12^2)^3)+(640 x12^2 x32)/(3 (-1+x12^2)^3)) xt^3-(320 x32^2 xt^4)/(3 (-1+x12^2)^3)) Log[xsq2^2])+((1792 x32^4)/(3 (-1+x32^2)^2)+(7168 x32^3 xt)/(3 (-1+x12^2) (-1+x32^2))-(3584 (x32+x12^2 x32-2 x32^3) xt^3)/(3 (-1+x12^2)^3)-(896 (1+x12^2-2 x32^2) xt^4)/(3 (-1+x12^2)^3)) PolyLog[2,1-x32^2]+((1792 x32^4)/(3 (-x12^2+x32^2)^2)+(7168 x32^3 xt)/(3 (-1+x12^2) (x12^2-x32^2))+(3584 (x32+x12^2 x32-2 x32^3) xt^3)/(3 (-1+x12^2)^3)+(896 (1+x12^2-2 x32^2) xt^4)/(3 (-1+x12^2)^3)) PolyLog[2,1-x32^2/x12^2]-128 Zeta[3])+Log[xEWS^2] (-((128 (-3 x32-3 x12^2 x32+14 x32^3-30 x32 xsq2^2) xt)/(3 (-1+x32^2) (-x12^2+x32^2)))+(8/3 (((mQ3^2-mQ3^2 x12^2) (-3 mQ3^10 x12^2-7 mQ3^10 x12^4-7 mQ3^10 x12^6-3 mQ3^10 x12^8-9 mQ3^10 x32^2+26 mQ3^10 x12^2 x32^2+34 mQ3^10 x12^4 x32^2+26 mQ3^10 x12^6 x32^2-9 mQ3^10 x12^8 x32^2+29 mQ3^10 x32^4-65 mQ3^10 x12^2 x32^4-65 mQ3^10 x12^4 x32^4+29 mQ3^10 x12^6 x32^4-34 mQ3^10 x32^6+76 mQ3^10 x12^2 x32^6-34 mQ3^10 x12^4 x32^6+14 mQ3^10 x32^8+14 mQ3^10 x12^2 x32^8-12 mQ3^10 x32^10))/((-mQ3^2+mQ3^2 x32^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)+(mQ3^8 (-44 mQ3^2+556 mQ3^2 x12^2) x32^8-128 mQ3^10 x32^10+mQ3^4 x12^4 (-mQ3^2+mQ3^2 x12^2) (-3 mQ3^4-4 mQ3^4 x12^2+mQ3^4 x12^4-30 mQ3^4 xsq2^4)+2 mQ3^6 x32^6 (mQ3^4-353 mQ3^4 x12^4-30 mQ3^4 x12^2 xsq2^2+mQ3^2 (-32 mQ3^2 x12^2+30 mQ3^2 xsq2^2))+mQ3^4 x32^4 (-9 mQ3^6+33 mQ3^6 x12^2+437 mQ3^6 x12^6-120 mQ3^6 x12^4 xsq2^2+90 mQ3^6 x12^2 xsq2^4+mQ3^2 (51 mQ3^4 x12^4+120 mQ3^4 x12^2 xsq2^2-90 mQ3^4 xsq2^4))+2 mQ3^4 x12^2 x32^2 (3 mQ3^6-18 mQ3^6 x12^2-68 mQ3^6 x12^6+90 mQ3^6 x12^4 xsq2^2-30 mQ3^6 x12^2 xsq2^4+mQ3^2 (19 mQ3^4 x12^4-90 mQ3^4 x12^2 xsq2^2+30 mQ3^4 xsq2^4)))/((mQ3^2-mQ3^2 x12^2) (-mQ3^2 x12^2+mQ3^2 x32^2)^4))+8/3 mQ3 ((8 (mQ3^2-mQ3^2 x12^2) (3 mQ3^5 x32-mQ3^5 x12^2 x32+3 mQ3^5 x12^4 x32-5 mQ3^5 x32^3-5 mQ3^5 x12^2 x32^3+5 mQ3^5 x32^5))/((-mQ3^2+mQ3^2 x32^2)^2 (-mQ3^2 x12^2+mQ3^2 x32^2)^2)-(8 (mQ3^5 (mQ3^4+68 mQ3^4 x12^2-85 mQ3^4 x12^4) x32^5-18 mQ3^7 (mQ3^2-mQ3^2 x12^2) x32^7+mQ3^3 x32^3 (3 mQ3^6-11 mQ3^6 x12^2+57 mQ3^6 x12^6+60 mQ3^6 x12^4 xsq2^2-30 mQ3^6 x12^2 xsq2^4+mQ3^2 (-17 mQ3^4 x12^4-60 mQ3^4 x12^2 xsq2^2+30 mQ3^4 xsq2^4))-mQ3^3 x12^2 x32 (3 mQ3^6-10 mQ3^6 x12^2+6 mQ3^2 x12^2 (mQ3^4 x12^4+10 mQ3^4 x12^2 xsq2^2-5 mQ3^4 xsq2^4)+mQ3^2 (17 mQ3^4 x12^4-60 mQ3^4 x12^2 xsq2^2+30 mQ3^4 xsq2^4))))/((mQ3^2-mQ3^2 x12^2)^2 (-mQ3^2 x12^2+mQ3^2 x32^2)^3)) xt-(512 x12^4 x32^2 xt^2)/((-1+x12^2)^2 (x12^2-x32^2)^2)) Log[x12^2]^2+(8/3 ((2 (-36 mQ3^16 x12^8+144 mQ3^16 x12^6 x32^2+144 mQ3^16 x12^8 x32^2-13 mQ3^16 x12^4 x32^4-452 mQ3^16 x12^6 x32^4-13 mQ3^16 x12^8 x32^4-134 mQ3^16 x12^2 x32^6-320 mQ3^16 x12^4 x32^6-320 mQ3^16 x12^6 x32^6-134 mQ3^16 x12^8 x32^6+27 mQ3^16 x32^8+908 mQ3^16 x12^2 x32^8+2240 mQ3^16 x12^4 x32^8+908 mQ3^16 x12^6 x32^8+27 mQ3^16 x12^8 x32^8-232 mQ3^16 x32^10-2700 mQ3^16 x12^2 x32^10-2700 mQ3^16 x12^4 x32^10-232 mQ3^16 x12^6 x32^10+721 mQ3^16 x32^12+2700 mQ3^16 x12^2 x32^12+721 mQ3^16 x12^4 x32^12-726 mQ3^16 x32^14-726 mQ3^16 x12^2 x32^14+198 mQ3^16 x32^16))/((-mQ3^2+mQ3^2 x32^2)^4 (-mQ3^2 x12^2+mQ3^2 x32^2)^4)-(30 (mQ3^2 x32^2-mQ3^2 xsq2^2) (2 mQ3^10 x32^10-mQ3^6 x12^2 (mQ3^4+mQ3^4 x12^4) xsq2^2+mQ3^8 x32^8 (-8 mQ3^2-8 mQ3^2 x12^2+6 mQ3^2 xsq2^2)+mQ3^6 x32^6 (3 mQ3^4+3 mQ3^4 x12^4-8 mQ3^4 x12^2 xsq2^2+mQ3^2 (30 mQ3^2 x12^2-8 mQ3^2 xsq2^2))-mQ3^4 x32^4 (mQ3^6+mQ3^6 x12^6-9 mQ3^6 x12^4 xsq2^2+mQ3^4 (15 mQ3^2 x12^2-9 mQ3^2 xsq2^2)+3 mQ3^2 (5 mQ3^4 x12^4+2 mQ3^4 x12^2 xsq2^2))+mQ3^2 x32^2 (3 mQ3^8 x12^2 xsq2^2-3 mQ3^8 x12^6 xsq2^2+mQ3^6 (5 mQ3^2 x12^2-3 mQ3^2 xsq2^2)+mQ3^2 (5 mQ3^6 x12^6+3 mQ3^6 x12^4 xsq2^2))))/((-mQ3^2+mQ3^2 x32^2)^3 (-mQ3^2 x12^2+mQ3^2 x32^2)^3))-(128 (51 x12^4 x32^3-85 x12^2 x32^5-85 x12^4 x32^5+26 x32^7+136 x12^2 x32^7+26 x12^4 x32^7-35 x32^9-35 x12^2 x32^9+x32^11+30 x12^4 x32 xsq2^2-30 x12^2 x32^3 xsq2^2-30 x12^4 x32^3 xsq2^2+30 x32^7 xsq2^2+30 x12^2 x32^7 xsq2^2-30 x32^9 xsq2^2-15 x12^2 x32 xsq2^4-15 x12^4 x32 xsq2^4+15 x32^3 xsq2^4+60 x12^2 x32^3 xsq2^4+15 x12^4 x32^3 xsq2^4-45 x32^5 xsq2^4-45 x12^2 x32^5 xsq2^4+30 x32^7 xsq2^4) xt)/(3 (-1+x32^2)^3 (-x12^2+x32^2)^3)-(512 x32^6 xt^2)/((-1+x32^2)^2 (-x12^2+x32^2)^2)) Log[x32^2]^2+(-((160 (10 x12^4-20 x12^2 x32^2-20 x12^4 x32^2+10 x32^4+40 x12^2 x32^4+10 x12^4 x32^4-20 x32^6-20 x12^2 x32^6+10 x32^8+3 x12^2 xsq2^2+3 x12^4 xsq2^2+9 x32^2 xsq2^2-12 x12^2 x32^2 xsq2^2+9 x12^4 x32^2 xsq2^2-15 x32^4 xsq2^2-15 x12^2 x32^4 xsq2^2+18 x32^6 xsq2^2))/(3 (-1+x32^2)^2 (-x12^2+x32^2)^2))+(1280 x32 xsq2^2 xt)/((-1+x32^2) (-x12^2+x32^2))) Log[xsq2^2]-160 Log[xsq2^2]^2+Log[x12^2] (-((16 (3 x12^4-81 x12^6+75 x12^8+3 x12^10+9 x12^2 x32^2+100 x12^4 x32^2+296 x12^6 x32^2-158 x12^8 x32^2+9 x12^10 x32^2-21 x12^2 x32^4-679 x12^4 x32^4-359 x12^6 x32^4+35 x12^8 x32^4-54 x32^6+296 x12^2 x32^6+1078 x12^4 x32^6+216 x12^6 x32^6+108 x32^8-582 x12^2 x32^8-550 x12^4 x32^8-54 x32^10+310 x12^2 x32^10+30 x12^4 xsq2^2-30 x12^6 xsq2^2+90 x12^2 x32^2 xsq2^2-150 x12^4 x32^2 xsq2^2+60 x12^6 x32^2 xsq2^2-180 x12^2 x32^4 xsq2^2+210 x12^4 x32^4 xsq2^2-30 x12^6 x32^4 xsq2^2+90 x12^2 x32^6 xsq2^2-90 x12^4 x32^6 xsq2^2))/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^2))-(128 (3 x12^2 x32+28 x12^4 x32+3 x12^6 x32-18 x12^2 x32^3-26 x12^4 x32^3-16 x32^5+10 x12^2 x32^5+16 x32^7+30 x12^2 x32 xsq2^2-30 x12^2 x32^3 xsq2^2) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2))+(-((16 (167 x12^4 x32^4-43 x12^6 x32^4+35 x12^8 x32^4-31 x12^10 x32^4-218 x12^2 x32^6-655 x12^4 x32^6-11 x12^6 x32^6+97 x12^8 x32^6+19 x12^10 x32^6+27 x32^8+999 x12^2 x32^8+1157 x12^4 x32^8-191 x12^6 x32^8-72 x12^8 x32^8-205 x32^10-1797 x12^2 x32^10-683 x12^4 x32^10+125 x12^6 x32^10+472 x32^12+1314 x12^2 x32^12+134 x12^4 x32^12-410 x32^14-358 x12^2 x32^14+128 x32^16))/(3 (-1+x12^2) (x12^2-x32^2)^4 (-1+x32^2)^3))+(128 (-75 x12^4 x32^3+7 x12^6 x32^3+129 x12^2 x32^5+121 x12^4 x32^5+4 x12^6 x32^5-38 x32^7-229 x12^2 x32^7-71 x12^4 x32^7+69 x32^9+117 x12^2 x32^9-34 x32^11) xt)/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^2)+(1024 x12^2 x32^4 xt^2)/((-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2))) Log[x32^2]+((160 xsq2^2 (6 x12^2 x32^2+2 x32^4-x12^2 xsq2^2-3 x32^2 xsq2^2))/(x12^2-x32^2)^3+(1280 xsq2^2 (2 x12^2 x32-x32 xsq2^2) xt)/((-1+x12^2) (x12^2-x32^2)^2)) Log[xsq2^2])+Log[x32^2] (-((16 (48 x12^6+3 x12^2 x32^2-333 x12^4 x32^2-333 x12^6 x32^2+3 x12^8 x32^2+9 x32^4+498 x12^2 x32^4+1822 x12^4 x32^4+498 x12^6 x32^4+9 x12^8 x32^4-269 x32^6-2575 x12^2 x32^6-2575 x12^4 x32^6-269 x12^6 x32^6+1182 x32^8+3564 x12^2 x32^8+1182 x12^4 x32^8-1562 x32^10-1562 x12^2 x32^10+660 x32^12+30 x12^2 x32^2 xsq2^2+30 x12^6 x32^2 xsq2^2+90 x32^4 xsq2^2-90 x12^2 x32^4 xsq2^2-90 x12^4 x32^4 xsq2^2+90 x12^6 x32^4 xsq2^2-270 x32^6 xsq2^2+180 x12^2 x32^6 xsq2^2-270 x12^4 x32^6 xsq2^2+240 x32^8 xsq2^2+240 x12^2 x32^8 xsq2^2-180 x32^10 xsq2^2))/(3 (-1+x32^2)^3 (-x12^2+x32^2)^3))+(128 (3 x32^3+34 x12^2 x32^3+3 x12^4 x32^3-40 x32^5-40 x12^2 x32^5+40 x32^7+30 x32^3 xsq2^2+30 x12^2 x32^3 xsq2^2-60 x32^5 xsq2^2) xt)/(3 (-1+x32^2)^2 (-x12^2+x32^2)^2)+((160 (-6 x12^2 x32^2 xsq2^2-6 x12^6 x32^2 xsq2^2-2 x32^4 xsq2^2+18 x12^2 x32^4 xsq2^2+18 x12^4 x32^4 xsq2^2-2 x12^6 x32^4 xsq2^2+6 x32^6 xsq2^2-36 x12^2 x32^6 xsq2^2+6 x12^4 x32^6 xsq2^2+4 x32^10 xsq2^2+x12^2 xsq2^4+x12^6 xsq2^4+3 x32^2 xsq2^4-3 x12^2 x32^2 xsq2^4-3 x12^4 x32^2 xsq2^4+3 x12^6 x32^2 xsq2^4-9 x32^4 xsq2^4+6 x12^2 x32^4 xsq2^4-9 x12^4 x32^4 xsq2^4+8 x32^6 xsq2^4+8 x12^2 x32^6 xsq2^4-6 x32^8 xsq2^4))/((-1+x32^2)^3 (-x12^2+x32^2)^3)-(1280 (-2 x12^2 x32 xsq2^2+2 x32^5 xsq2^2+x32 xsq2^4+x12^2 x32 xsq2^4-2 x32^3 xsq2^4) xt)/((-1+x32^2)^2 (-x12^2+x32^2)^2)) Log[xsq2^2])+(-((16 (-1+x12^2) (3 x12^2+7 x12^4+7 x12^6+3 x12^8+9 x32^2-26 x12^2 x32^2-34 x12^4 x32^2-26 x12^6 x32^2+9 x12^8 x32^2-29 x32^4+65 x12^2 x32^4+65 x12^4 x32^4-29 x12^6 x32^4+34 x32^6-76 x12^2 x32^6+34 x12^4 x32^6-14 x32^8-14 x12^2 x32^8+12 x32^10))/(3 (x12^2-x32^2)^3 (-1+x32^2)^3))-(128 (-1+x12^2) (3 x32-x12^2 x32+3 x12^4 x32-5 x32^3-5 x12^2 x32^3+5 x32^5) xt)/(3 (x12^2-x32^2)^2 (-1+x32^2)^2)) PolyLog[2,1-x12^2]+(-((16 (3 x12^2+x12^4+19 x12^6-23 x12^8+9 x32^2-41 x12^2 x32^2-41 x12^4 x32^2+167 x12^6 x32^2+34 x12^8 x32^2-20 x32^4+151 x12^2 x32^4-417 x12^4 x32^4-391 x12^6 x32^4+37 x12^8 x32^4-9 x32^6+251 x12^2 x32^6+1025 x12^4 x32^6+13 x12^6 x32^6-98 x32^8-902 x12^2 x32^8-280 x12^4 x32^8+294 x32^10+346 x12^2 x32^10-128 x32^12))/(3 (-1+x12^2) (x12^2-x32^2)^3 (-1+x32^2)^2))+(128 (3 x32-7 x12^2 x32+22 x12^4 x32+x32^3-37 x12^2 x32^3+18 x32^5) xt)/(3 (-1+x12^2) (x12^2-x32^2)^2)) PolyLog[2,1-x32^2]+(-((16 (-23 x12^4+19 x12^6+x12^8+3 x12^10+34 x12^2 x32^2+167 x12^4 x32^2-41 x12^6 x32^2-41 x12^8 x32^2+9 x12^10 x32^2+37 x32^4-391 x12^2 x32^4-417 x12^4 x32^4+151 x12^6 x32^4-20 x12^8 x32^4+13 x32^6+1025 x12^2 x32^6+251 x12^4 x32^6-9 x12^6 x32^6-280 x32^8-902 x12^2 x32^8-98 x12^4 x32^8+346 x32^10+294 x12^2 x32^10-128 x32^12))/(3 (-1+x12^2) (x12^2-x32^2)^2 (-1+x32^2)^3))-(128 (22 x32-7 x12^2 x32+3 x12^4 x32-37 x32^3+x12^2 x32^3+18 x32^5) xt)/(3 (-1+x12^2) (-1+x32^2)^2)) PolyLog[2,1-x32^2/x12^2]+((160 (-1+xsq2^2) (-1+3 x32^2+2 x32^4-xsq2^2-3 x32^2 xsq2^2))/(-1+x32^2)^3-(1280 x32 (-1+xsq2^2)^2 xt)/((-1+x12^2) (-1+x32^2)^2)) PolyLog[2,1-xsq2^2]+(-((160 (x12^2-xsq2^2) (x12^4-3 x12^2 x32^2-2 x32^4+x12^2 xsq2^2+3 x32^2 xsq2^2))/(x12^2-x32^2)^3)+(1280 x32 (x12^2-xsq2^2)^2 xt)/((-1+x12^2) (x12^2-x32^2)^2)) PolyLog[2,1-xsq2^2/x12^2]+(-((160 (x32^2-xsq2^2) (5 x12^2 x32^2+5 x12^6 x32^2-x32^4-15 x12^2 x32^4-15 x12^4 x32^4-x12^6 x32^4+3 x32^6+30 x12^2 x32^6+3 x12^4 x32^6-8 x32^8-8 x12^2 x32^8+2 x32^10-x12^2 xsq2^2-x12^6 xsq2^2-3 x32^2 xsq2^2+3 x12^2 x32^2 xsq2^2+3 x12^4 x32^2 xsq2^2-3 x12^6 x32^2 xsq2^2+9 x32^4 xsq2^2-6 x12^2 x32^4 xsq2^2+9 x12^4 x32^4 xsq2^2-8 x32^6 xsq2^2-8 x12^2 x32^6 xsq2^2+6 x32^8 xsq2^2))/((-1+x32^2)^3 (-x12^2+x32^2)^3))-(1280 (x32^2-xsq2^2) (-2 x12^2 x32+x32^3+x12^2 x32^3+x32 xsq2^2+x12^2 x32 xsq2^2-2 x32^3 xsq2^2) xt)/((-1+x32^2)^2 (-x12^2+x32^2)^2)) PolyLog[2,1-xsq2^2/x32^2]+(16 (-3 x12^2+98 x12^4-3 x12^6-9 x32^2-211 x12^2 x32^2-211 x12^4 x32^2-9 x12^6 x32^2+169 x32^4+454 x12^2 x32^4+169 x12^4 x32^4-316 x32^6-316 x12^2 x32^6+188 x32^8-30 x12^2 xsq2^2-30 x12^4 xsq2^2-90 x32^2 xsq2^2+120 x12^2 x32^2 xsq2^2-90 x12^4 x32^2 xsq2^2+150 x32^4 xsq2^2+150 x12^2 x32^4 xsq2^2-180 x32^6 xsq2^2+24 x12^4 Zeta[3]-48 x12^2 x32^2 Zeta[3]-48 x12^4 x32^2 Zeta[3]+24 x32^4 Zeta[3]+96 x12^2 x32^4 Zeta[3]+24 x12^4 x32^4 Zeta[3]-48 x32^6 Zeta[3]-48 x12^2 x32^6 Zeta[3]+24 x32^8 Zeta[3]))/(3 (-1+x32^2)^2 (-x12^2+x32^2)^2))))
);

(* Using mass eigenstate masses *)

Mh2[MassEigen]=(
MZ^2 Cos[2 beta]^2+at mt^2 (Sqrt[k] ((12 Xt^4)/(-mst1^2+mst2^2)^2+(6-(12 Xt^2)/(-mst1^2+mst2^2)+((6 mst1^2)/(-mst1^2+mst2^2)^3+(6 mst2^2)/(-mst1^2+mst2^2)^3) Xt^4) Log[mst1^2/MR^2]+(6+(12 Xt^2)/(-mst1^2+mst2^2)+(-((6 mst1^2)/(-mst1^2+mst2^2)^3)-(6 mst2^2)/(-mst1^2+mst2^2)^3) Xt^4) Log[mst2^2/MR^2]+12 Log[MR^2/mt^2])+a3 k (-64+(32 mg^2)/mst1^2+(32 mg^2)/mst2^2-(64 mg^2 Xt^2)/(mst1^2 mst2^2)-(512 mg Xt^3)/(-mst1^2+mst2^2)^2+(192/(-mst1^2+mst2^2)^2+(32 mg^2)/(mst1^2 (-mst1^2+mst2^2)^2)+(32 mg^2)/(mst2^2 (-mst1^2+mst2^2)^2)) Xt^4+((32 (3 mg^4-2 mg^2 mst1^2+mst1^4))/(mg^2-mst1^2)^2-(64 (2 mg^3-mg mst1^2) Xt)/((mg^2-mst1^2) (mst1^2-mst2^2))+(64 (3 mg^2-2 mst1^2) Xt^2)/((mg^2-mst1^2) (mst1^2-mst2^2))+(128 (3 mg mst1^2+mg mst2^2) Xt^3)/(mst1^2-mst2^2)^3-(32 (2 mg^2+7 mst1^2+3 mst2^2) Xt^4)/(mst1^2-mst2^2)^3) Log[mst1^2/MR^2]+(-((16 (3 mg^4-2 mg^2 mst1^2+mst1^4))/(mg^2-mst1^2)^2)+(128 mg^3 Xt)/((mg^2-mst1^2) (mst1^2-mst2^2))-(32 (3 mst1^2-mst2^2) Xt^2)/(mst1^2-mst2^2)^2+(64 mg (2 mg^2-mst1^2-mst2^2) Xt^3)/(mst1^2-mst2^2)^3+(32 (mg^2 mst1^2+2 mst1^4-mg^2 mst2^2+2 mst1^2 mst2^2) Xt^4)/(mst1^2-mst2^2)^4) Log[mst1^2/MR^2]^2+((32 (3 mg^4-2 mg^2 mst2^2+mst2^4))/(mg^2-mst2^2)^2-(64 (2 mg^3-mg mst2^2) Xt)/((mg^2-mst2^2) (-mst1^2+mst2^2))+(64 (3 mg^2-2 mst2^2) Xt^2)/((mg^2-mst2^2) (-mst1^2+mst2^2))+(128 (mg mst1^2+3 mg mst2^2) Xt^3)/(-mst1^2+mst2^2)^3-(32 (2 mg^2+3 mst1^2+7 mst2^2) Xt^4)/(-mst1^2+mst2^2)^3+((64 (mst1^2+mst2^2) Xt^2)/(-mst1^2+mst2^2)^2-(64 (mst1^4+2 mst1^2 mst2^2+mst2^4) Xt^4)/(-mst1^2+mst2^2)^4) Log[mst1^2/MR^2]) Log[mst2^2/MR^2]+(-((16 (3 mg^4-2 mg^2 mst2^2+mst2^4))/(mg^2-mst2^2)^2)+(128 mg^3 Xt)/((mg^2-mst2^2) (-mst1^2+mst2^2))-(32 (-mst1^2+3 mst2^2) Xt^2)/(-mst1^2+mst2^2)^2+(64 mg (2 mg^2-mst1^2-mst2^2) Xt^3)/(-mst1^2+mst2^2)^3+(32 (-mg^2 mst1^2+mg^2 mst2^2+2 mst1^2 mst2^2+2 mst2^4) Xt^4)/(-mst1^2+mst2^2)^4) Log[mst2^2/MR^2]^2+Log[mg^2/MR^2] (-((32 (mg^10 mst1^2-2 mg^8 mst1^4+mg^6 mst1^6+mg^10 mst2^2+2 mg^8 mst1^2 mst2^2-2 mg^6 mst1^4 mst2^2+mg^4 mst1^6 mst2^2-2 mg^8 mst2^4-2 mg^6 mst1^2 mst2^4+mg^6 mst2^6+mg^4 mst1^2 mst2^6))/(mst1^2 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^2))+(64 mg^3 Xt)/((mg^2-mst1^2) (mg^2-mst2^2))+(64 mg^4 (-mg^2+mst1^2+mst2^2) Xt^2)/(mst1^2 (mg^2-mst1^2) mst2^2 (-mg^2+mst2^2))-(32 (mg^2 mst1^2+mg^2 mst2^2) Xt^4)/(mst1^2 mst2^2 (-mst1^2+mst2^2)^2)+((32 mg^4)/(mg^2-mst1^2)^2-(128 mg^3 Xt)/((mg^2-mst1^2) (mst1^2-mst2^2))-(256 mg^3 Xt^3)/(mst1^2-mst2^2)^3) Log[mst1^2/MR^2]+((32 mg^4)/(mg^2-mst2^2)^2-(128 mg^3 Xt)/((mg^2-mst2^2) (-mst1^2+mst2^2))-(256 mg^3 Xt^3)/(-mst1^2+mst2^2)^3) Log[mst2^2/MR^2])+(-((32 (2 mg^4-mg^2 mst1^2-mg^2 mst2^2))/((mg^2-mst1^2) (mg^2-mst2^2)))+((32 (2 mg^8-2 mg^6 mst1^2+mg^4 mst1^4-2 mg^6 mst2^2+mg^4 mst2^4))/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)-(128 mg^3 Xt)/((mg^2-mst1^2) (mg^2-mst2^2))) Log[mg^2/MR^2]+(-((64 mg^2 mst1^2)/(mg^2-mst1^2)^2)+(32 mst1^4)/(mg^2-mst1^2)^2+(128 mg mst1^2 Xt)/((mg^2-mst1^2) (mst1^2-mst2^2))) Log[mst1^2/MR^2]+(-((64 mg^2 mst2^2)/(mg^2-mst2^2)^2)+(32 mst2^4)/(mg^2-mst2^2)^2+(128 mg mst2^2 Xt)/((mg^2-mst2^2) (-mst1^2+mst2^2))) Log[mst2^2/MR^2]) Log[MR^2/mt^2]+96 Log[MR^2/mt^2]^2+(-((64 mg^4)/(mg^2-mst1^2)^2)+(256 mg^3 Xt)/((mg^2-mst1^2) (mst1^2-mst2^2))+(128 mg (-2 mg^2+mst1^2+mst2^2) Xt^3)/(-mst1^2+mst2^2)^3+(32 (-2 mg^2+mst1^2+mst2^2) Xt^4)/(-mst1^2+mst2^2)^3) PolyLog[2,1-mg^2/mst1^2]+(-((64 mg^4)/(-mg^2+mst2^2)^2)+(256 mg^3 Xt)/((mg^2-mst2^2) (-mst1^2+mst2^2))+(128 mg (2 mg^2-mst1^2-mst2^2) Xt^3)/(-mst1^2+mst2^2)^3-(32 (-2 mg^2+mst1^2+mst2^2) Xt^4)/(-mst1^2+mst2^2)^3) PolyLog[2,1-mg^2/mst2^2])+a3^2 k^(3/2) (dlatas2-(128 mg^4)/(3 (mg^2-mst1^2) (mg^2-mst2^2))+(128 mst1^2 mst2^2)/(3 (mg^2-mst1^2) (mg^2-mst2^2))-(16 (mg^4-mst1^2 mst2^2))/((mg^2-mst1^2) (mg^2-mst2^2))-(48 (mg^4-mst1^2 mst2^2)^2)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)-(16 (-11 mg^4+3 mg^2 mst1^2+3 mg^2 mst2^2+5 mst1^2 mst2^2) (-2 mg^6 mst1^2+2 mg^4 mst1^4-2 mg^6 mst2^2+9 mg^4 mst1^2 mst2^2-6 mg^2 mst1^4 mst2^2+2 mg^4 mst2^4-6 mg^2 mst1^2 mst2^4+3 mst1^4 mst2^4))/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (-mg^2+mst2^2)^2)+(4 (291 mg^8-360 mg^6 msq^2-282 mg^6 mst1^2+300 mg^4 msq^2 mst1^2+93 mg^4 mst1^4-180 mg^2 msq^2 mst1^4-18 mg^2 mst1^6-282 mg^6 mst2^2+300 mg^4 msq^2 mst2^2-112 mg^4 mst1^2 mst2^2+240 mg^2 msq^2 mst1^2 mst2^2+208 mg^2 mst1^4 mst2^2-60 msq^2 mst1^4 mst2^2-6 mst1^6 mst2^2+93 mg^4 mst2^4-180 mg^2 msq^2 mst2^4+208 mg^2 mst1^2 mst2^4-60 msq^2 mst1^2 mst2^4-169 mst1^4 mst2^4-18 mg^2 mst2^6-6 mst1^2 mst2^6))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)+SMConstatas2-(64 (14 mg^3-30 mg msq^2-3 mg mst1^2-3 mg mst2^2) Xt)/(3 (mg^2-mst1^2) (mg^2-mst2^2))-(64 mg^2 (-11 mg^4+3 mg^2 mst1^2+3 mg^2 mst2^2+5 mst1^2 mst2^2) Xt^2)/(3 mst1^2 (mg^2-mst1^2) mst2^2 (mg^2-mst2^2))-(512 mg (-11 mg^4+3 mg^2 mst1^2+3 mg^2 mst2^2+5 mst1^2 mst2^2) Xt^3)/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^2)+(8 (-44 mg^10 mst1^2+56 mg^8 mst1^4-12 mg^6 mst1^6-44 mg^10 mst2^2-455 mg^8 mst1^2 mst2^2+360 mg^6 msq^2 mst1^2 mst2^2+558 mg^6 mst1^4 mst2^2-300 mg^4 msq^2 mst1^4 mst2^2-173 mg^4 mst1^6 mst2^2+180 mg^2 msq^2 mst1^6 mst2^2+18 mg^2 mst1^8 mst2^2+56 mg^8 mst2^4+558 mg^6 mst1^2 mst2^4-300 mg^4 msq^2 mst1^2 mst2^4-168 mg^4 mst1^4 mst2^4-240 mg^2 msq^2 mst1^4 mst2^4-236 mg^2 mst1^6 mst2^4+60 msq^2 mst1^6 mst2^4+6 mst1^8 mst2^4-12 mg^6 mst2^6-173 mg^4 mst1^2 mst2^6+180 mg^2 msq^2 mst1^2 mst2^6-236 mg^2 mst1^4 mst2^6+60 msq^2 mst1^4 mst2^6+277 mst1^6 mst2^6+18 mg^2 mst1^2 mst2^8+6 mst1^4 mst2^8) Xt^4)/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (-mg^2+mst2^2)^2 (-mst1^2+mst2^2)^2)+(128 (14 mg^3-30 mg msq^2-3 mg mst1^2-3 mg mst2^2) Xt^5)/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^2)+(24 (-((5 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(6 (-mg^2+mst1^2)^3))-(5 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(6 (-mg^2+mst2^2)^3))+24 ((40 mg (-msq^2+mst1^2)^2)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+(40 mg (-msq^2+mst2^2)^2)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt-1/(-mst1^2+mst2^2)^2 48 (-((5 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(6 (-mg^2+mst1^2)^3))-(5 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(6 (-mg^2+mst2^2)^3)) Xt^4-(48 ((40 mg (-msq^2+mst1^2)^2)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+(40 mg (-msq^2+mst2^2)^2)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt^5)/(-mst1^2+mst2^2)^2) Log[msq^2/MR^2]^2+((4 (128 mg^14-1028 mg^12 mst1^2+60 mg^10 msq^2 mst1^2-90 mg^8 msq^4 mst1^2+1328 mg^10 mst1^4+300 mg^8 msq^2 mst1^4+60 mg^6 msq^4 mst1^4-313 mg^8 mst1^6-300 mg^6 msq^2 mst1^6+30 mg^4 msq^4 mst1^6-560 mg^6 mst1^8-60 mg^4 msq^2 mst1^8+291 mg^4 mst1^10+18 mg^2 mst1^12-252 mg^12 mst2^2-60 mg^10 msq^2 mst2^2+90 mg^8 msq^4 mst2^2+3132 mg^10 mst1^2 mst2^2-420 mg^8 msq^2 mst1^2 mst2^2+120 mg^6 msq^4 mst1^2 mst2^2-4635 mg^8 mst1^4 mst2^2-300 mg^6 msq^2 mst1^4 mst2^2-150 mg^4 msq^4 mst1^4 mst2^2+2044 mg^6 mst1^6 mst2^2+660 mg^4 msq^2 mst1^6 mst2^2-60 mg^2 msq^4 mst1^6 mst2^2+815 mg^4 mst1^8 mst2^2+120 mg^2 msq^2 mst1^8 mst2^2-702 mg^2 mst1^10 mst2^2+6 mst1^12 mst2^2+20 mg^10 mst2^4+120 mg^8 msq^2 mst2^4-180 mg^6 msq^4 mst2^4-2847 mg^8 mst1^2 mst2^4+660 mg^6 msq^2 mst1^2 mst2^4+30 mg^4 msq^4 mst1^2 mst2^4+4812 mg^6 mst1^4 mst2^4-300 mg^4 msq^2 mst1^4 mst2^4+120 mg^2 msq^4 mst1^4 mst2^4-2968 mg^4 mst1^6 mst2^4-420 mg^2 msq^2 mst1^6 mst2^4+30 msq^4 mst1^6 mst2^4+284 mg^2 mst1^8 mst2^4-60 msq^2 mst1^8 mst2^4+291 mst1^10 mst2^4+115 mg^8 mst2^6-60 mg^6 msq^2 mst2^6+90 mg^4 msq^4 mst2^6+764 mg^6 mst1^2 mst2^6-300 mg^4 msq^2 mst1^2 mst2^6-60 mg^2 msq^4 mst1^2 mst2^6-1436 mg^4 mst1^4 mst2^6+300 mg^2 msq^2 mst1^4 mst2^6-30 msq^4 mst1^4 mst2^6+992 mg^2 mst1^6 mst2^6+60 msq^2 mst1^6 mst2^6-299 mst1^8 mst2^6-20 mg^6 mst2^8-39 mg^4 mst1^2 mst2^8+54 mg^2 mst1^4 mst2^8+5 mst1^6 mst2^8+9 mg^4 mst2^10-6 mg^2 mst1^2 mst2^10-3 mst1^4 mst2^10))/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^2 (mst1^2-mst2^2))-(32 (32 mg^9 mst1^2-30 mg^5 msq^4 mst1^2-397 mg^7 mst1^4+120 mg^5 msq^2 mst1^4+30 mg^3 msq^4 mst1^4+479 mg^5 mst1^6-120 mg^3 msq^2 mst1^6-200 mg^3 mst1^8+6 mg mst1^10-32 mg^9 mst2^2+30 mg^5 msq^4 mst2^2+380 mg^7 mst1^2 mst2^2-120 mg^5 msq^2 mst1^2 mst2^2-122 mg^5 mst1^4 mst2^2-30 mg msq^4 mst1^4 mst2^2-226 mg^3 mst1^6 mst2^2+120 mg msq^2 mst1^6 mst2^2+160 mg mst1^8 mst2^2+mg^7 mst2^4-30 mg^3 msq^4 mst2^4-311 mg^5 mst1^2 mst2^4+120 mg^3 msq^2 mst1^2 mst2^4+30 mg msq^4 mst1^2 mst2^4+367 mg^3 mst1^4 mst2^4-120 mg msq^2 mst1^4 mst2^4-137 mg mst1^6 mst2^4+2 mg^5 mst2^6+14 mg^3 mst1^2 mst2^6-16 mg mst1^4 mst2^6-3 mg^3 mst2^8+3 mg mst1^2 mst2^8) Xt)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2) (mst1^2-mst2^2)^2)-(16 (492 mg^10 mst1^2-383 mg^8 mst1^4-90 mg^6 msq^2 mst1^4-226 mg^6 mst1^6-30 mg^4 msq^2 mst1^6+192 mg^4 mst1^8+9 mg^2 mst1^10-24 mg^10 mst2^2-1462 mg^8 mst1^2 mst2^2+90 mg^6 msq^2 mst1^2 mst2^2+1473 mg^6 mst1^4 mst2^2+210 mg^4 msq^2 mst1^4 mst2^2+137 mg^4 mst1^6 mst2^2+60 mg^2 msq^2 mst1^6 mst2^2-391 mg^2 mst1^8 mst2^2+3 mst1^10 mst2^2+65 mg^8 mst2^4+1346 mg^6 mst1^2 mst2^4-180 mg^4 msq^2 mst1^2 mst2^4-1612 mg^4 mst1^4 mst2^4-150 mg^2 msq^2 mst1^4 mst2^4+338 mg^2 mst1^6 mst2^4-30 msq^2 mst1^6 mst2^4+151 mst1^8 mst2^4-41 mg^6 mst2^6-373 mg^4 mst1^2 mst2^6+90 mg^2 msq^2 mst1^2 mst2^6+471 mg^2 mst1^4 mst2^6+30 msq^2 mst1^4 mst2^6-177 mst1^6 mst2^6+9 mg^2 mst1^2 mst2^8+3 mst1^4 mst2^8) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (mst1^2-mst2^2)^2)+(128 (11 mg^9-37 mg^7 mst1^2-48 mg^5 mst1^4+30 mg^3 msq^2 mst1^4+66 mg^3 mst1^6-3 mg mst1^8+7 mg^7 mst2^2+150 mg^5 mst1^2 mst2^2-30 mg^3 msq^2 mst1^2 mst2^2-47 mg^3 mst1^4 mst2^2-30 mg msq^2 mst1^4 mst2^2-53 mg mst1^6 mst2^2-14 mg^5 mst2^4-117 mg^3 mst1^2 mst2^4+30 mg msq^2 mst1^2 mst2^4+82 mg mst1^4 mst2^4+3 mg mst1^2 mst2^6) Xt^3)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2) (-mst1^2+mst2^2)^3)-(8 (176 mg^14 mst1^2-2586 mg^12 mst1^4+60 mg^10 msq^2 mst1^4-90 mg^8 msq^4 mst1^4+5679 mg^10 mst1^6+210 mg^8 msq^2 mst1^6+60 mg^6 msq^4 mst1^6-4030 mg^8 mst1^8-240 mg^6 msq^2 mst1^8+30 mg^4 msq^4 mst1^8+368 mg^6 mst1^10-30 mg^4 msq^2 mst1^10+392 mg^4 mst1^12+9 mg^2 mst1^14-176 mg^14 mst2^2-120 mg^12 mst1^2 mst2^2-120 mg^10 msq^2 mst1^2 mst2^2+180 mg^8 msq^4 mst1^2 mst2^2+6267 mg^10 mst1^4 mst2^2-360 mg^8 msq^2 mst1^4 mst2^2+60 mg^6 msq^4 mst1^4 mst2^2-13622 mg^8 mst1^6 mst2^2-60 mg^6 msq^2 mst1^6 mst2^2-180 mg^4 msq^4 mst1^6 mst2^2+9482 mg^6 mst1^8 mst2^2+480 mg^4 msq^2 mst1^8 mst2^2-60 mg^2 msq^4 mst1^8 mst2^2-1079 mg^4 mst1^10 mst2^2+60 mg^2 msq^2 mst1^10 mst2^2-787 mg^2 mst1^12 mst2^2+3 mst1^14 mst2^2+322 mg^12 mst2^4+60 mg^10 msq^2 mst2^4-90 mg^8 msq^4 mst2^4+201 mg^10 mst1^2 mst2^4+270 mg^8 msq^2 mst1^2 mst2^4-300 mg^6 msq^4 mst1^2 mst2^4-6104 mg^8 mst1^4 mst2^4+420 mg^6 msq^2 mst1^4 mst2^4+180 mg^4 msq^4 mst1^4 mst2^4+11390 mg^6 mst1^6 mst2^4-480 mg^4 msq^2 mst1^6 mst2^4+180 mg^2 msq^4 mst1^6 mst2^4-7123 mg^4 mst1^8 mst2^4-240 mg^2 msq^2 mst1^8 mst2^4+30 msq^4 mst1^8 mst2^4+1015 mg^2 mst1^10 mst2^4-30 msq^2 mst1^10 mst2^4+347 mst1^12 mst2^4-147 mg^10 mst2^6-120 mg^8 msq^2 mst2^6+180 mg^6 msq^4 mst2^6-474 mg^8 mst1^2 mst2^6-180 mg^6 msq^2 mst1^2 mst2^6+60 mg^4 msq^4 mst1^2 mst2^6+3202 mg^6 mst1^4 mst2^6-180 mg^2 msq^4 mst1^4 mst2^6-4230 mg^4 mst1^6 mst2^6+300 mg^2 msq^2 mst1^6 mst2^6-60 msq^4 mst1^6 mst2^6+1861 mg^2 mst1^8 mst2^6-244 mst1^10 mst2^6-10 mg^8 mst2^8+60 mg^6 msq^2 mst2^8-90 mg^4 msq^4 mst2^8+178 mg^6 mst1^2 mst2^8+30 mg^4 msq^2 mst1^2 mst2^8+60 mg^2 msq^4 mst1^2 mst2^8-692 mg^4 mst1^4 mst2^8-120 mg^2 msq^2 mst1^4 mst2^8+30 msq^4 mst1^4 mst2^8+722 mg^2 mst1^6 mst2^8+30 msq^2 mst1^6 mst2^8-190 mst1^8 mst2^8+20 mg^6 mst2^10+21 mg^4 mst1^2 mst2^10-42 mg^2 mst1^4 mst2^10+mst1^6 mst2^10-9 mg^4 mst2^12+6 mg^2 mst1^2 mst2^12+3 mst1^4 mst2^12) Xt^4)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^2 (mst1^2-mst2^2)^4)-(64 (30 mg^9 mst1^2+30 mg^5 msq^4 mst1^2+179 mg^7 mst1^4-90 mg^5 msq^2 mst1^4-30 mg^3 msq^4 mst1^4-437 mg^5 mst1^6+90 mg^3 msq^2 mst1^6+215 mg^3 mst1^8-3 mg mst1^10+34 mg^9 mst2^2-30 mg^5 msq^4 mst2^2-6 mg^7 mst1^2 mst2^2+30 mg^5 msq^2 mst1^2 mst2^2-306 mg^5 mst1^4 mst2^2+60 mg^3 msq^2 mst1^4 mst2^2+30 mg msq^4 mst1^4 mst2^2+511 mg^3 mst1^6 mst2^2-90 mg msq^2 mst1^6 mst2^2-201 mg mst1^8 mst2^2-35 mg^7 mst2^4+30 mg^3 msq^4 mst2^4-5 mg^5 mst1^2 mst2^4-30 mg^3 msq^2 mst1^2 mst2^4-30 mg msq^4 mst1^2 mst2^4+106 mg^3 mst1^4 mst2^4+30 mg msq^2 mst1^4 mst2^4-82 mg mst1^6 mst2^4-2 mg^5 mst2^6-5 mg^3 mst1^2 mst2^6+7 mg mst1^4 mst2^6+3 mg^3 mst2^8-3 mg mst1^2 mst2^8) Xt^5)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2) (-mst1^2+mst2^2)^4)-(2560 mg^2 mst1^4 Xt^6)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2)^4)+(-(1/(3 (mg^2-mst1^2)^3))80 (3 mg^6-6 mg^4 msq^2+9 mg^2 msq^4-7 mg^4 mst1^2-18 mg^2 msq^2 mst1^2+3 msq^4 mst1^2+6 mg^2 mst1^4-2 mst1^6)+(320 (2 mg^5+6 mg msq^4-3 mg^3 mst1^2-12 mg msq^2 mst1^2+mg mst1^4) Xt)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+((160 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst1^2-msq^4 mst1^2))/((mg^2-mst1^2)^3 (mst1^2-mst2^2))+(160 (mg^4 (-3 mst1^2+mst2^2)+2 mst1^4 (-2 mst1^2+mst2^2)+mg^2 (8 mst1^4-4 mst1^2 mst2^2)))/(3 (mg^2-mst1^2)^2 (-mst1^2+mst2^2)^2)) Xt^2+(-((1280 (-mg msq^4+2 mg msq^2 mst1^2))/((mg^2-mst1^2)^2 (mst1^2-mst2^2)^2))+(320 mg (2 mg^4-mst1^4+3 mst1^2 mst2^2-mg^2 (3 mst1^2+mst2^2)))/(3 (mg^2-mst1^2) (mst1^2-mst2^2)^3)) Xt^3+(-((80 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst1^2-msq^4 mst1^2) (mst1^2+mst2^2))/((mg^2-mst1^2)^3 (mst1^2-mst2^2)^3))+(80 (5 mst1^8+8 mg^4 mst1^2 mst2^2+4 mst1^6 mst2^2-mst1^4 mst2^4-2 mg^6 (-mst1^2+mst2^2)+2 mg^2 mst1^2 (-4 mst1^4-5 mst1^2 mst2^2+mst2^4)))/(3 (mg^2-mst1^2)^2 (-mst1^2+mst2^2)^4)) Xt^4+(320 (-6 mg msq^4+mg^3 mst1^2+12 mg msq^2 mst1^2-mg mst1^4) (mst1^2+mst2^2) Xt^5)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2)^4)) Log[msq^2/MR^2]) Log[mst1^2/MR^2]^2+(-((4 (128 mg^10-550 mg^8 mst1^2+60 mg^6 msq^2 mst1^2-90 mg^4 msq^4 mst1^2+494 mg^6 mst1^4+120 mg^4 msq^2 mst1^4+60 mg^2 msq^4 mst1^4-59 mg^4 mst1^6-180 mg^2 msq^2 mst1^6+30 msq^4 mst1^6-136 mg^2 mst1^8+67 mst1^10+38 mg^8 mst2^2-60 mg^6 msq^2 mst2^2+90 mg^4 msq^4 mst2^2+276 mg^6 mst1^2 mst2^2-120 mg^4 msq^2 mst1^2 mst2^2-60 mg^2 msq^4 mst1^2 mst2^2-429 mg^4 mst1^4 mst2^2+180 mg^2 msq^2 mst1^4 mst2^2-30 msq^4 mst1^4 mst2^2+234 mg^2 mst1^6 mst2^2-63 mst1^8 mst2^2-2 mg^6 mst2^4-33 mg^4 mst1^2 mst2^4+36 mg^2 mst1^4 mst2^4-mst1^6 mst2^4+9 mg^4 mst2^6-6 mg^2 mst1^2 mst2^6-3 mst1^4 mst2^6))/(3 (mg^2-mst1^2)^4 (mst1^2-mst2^2)))+(32 (20 mg^7 mst1^2-30 mg^3 msq^4 mst1^2-202 mg^5 mst1^4+60 mg^3 msq^2 mst1^4+30 mg msq^4 mst1^4+189 mg^3 mst1^6-60 mg msq^2 mst1^6-55 mg mst1^8-20 mg^7 mst2^2+30 mg^3 msq^4 mst2^2+185 mg^5 mst1^2 mst2^2-60 mg^3 msq^2 mst1^2 mst2^2-30 mg msq^4 mst1^2 mst2^2-149 mg^3 mst1^4 mst2^2+60 mg msq^2 mst1^4 mst2^2+32 mg mst1^6 mst2^2+mg^5 mst2^4-11 mg^3 mst1^2 mst2^4+10 mg mst1^4 mst2^4+3 mg^3 mst2^6-3 mg mst1^2 mst2^6) Xt)/(3 (mg^2-mst1^2)^3 (mst1^2-mst2^2)^2)-(8 (128 mg^10-1062 mg^8 mst1^2+60 mg^6 msq^2 mst1^2-90 mg^4 msq^4 mst1^2+1670 mg^6 mst1^4+120 mg^4 msq^2 mst1^4+60 mg^2 msq^4 mst1^4-615 mg^4 mst1^6-180 mg^2 msq^2 mst1^6+30 msq^4 mst1^6-248 mg^2 mst1^8+135 mst1^10+42 mg^8 mst2^2-60 mg^6 msq^2 mst2^2+90 mg^4 msq^4 mst2^2+140 mg^6 mst1^2 mst2^2-120 mg^4 msq^2 mst1^2 mst2^2-60 mg^2 msq^4 mst1^2 mst2^2-361 mg^4 mst1^4 mst2^2+180 mg^2 msq^2 mst1^4 mst2^2-30 msq^4 mst1^4 mst2^2+234 mg^2 mst1^6 mst2^2-63 mst1^8 mst2^2-2 mg^6 mst2^4-33 mg^4 mst1^2 mst2^4+36 mg^2 mst1^4 mst2^4-mst1^6 mst2^4+9 mg^4 mst2^6-6 mg^2 mst1^2 mst2^6-3 mst1^4 mst2^6) Xt^2)/(3 (mg^2-mst1^2)^4 (mst1^2-mst2^2)^2)+(32 (2 mg^9-35 mg^7 mst1^2-60 mg^3 msq^4 mst1^2-131 mg^5 mst1^4+120 mg^3 msq^2 mst1^4+60 mg msq^4 mst1^4+353 mg^3 mst1^6-120 mg msq^2 mst1^6-157 mg mst1^8-37 mg^7 mst2^2+60 mg^3 msq^4 mst2^2+205 mg^5 mst1^2 mst2^2-120 mg^3 msq^2 mst1^2 mst2^2-60 mg msq^4 mst1^2 mst2^2-281 mg^3 mst1^4 mst2^2+120 mg msq^2 mst1^4 mst2^2+81 mg mst1^6 mst2^2+2 mg^5 mst2^4-22 mg^3 mst1^2 mst2^4+20 mg mst1^4 mst2^4+6 mg^3 mst2^6-6 mg mst1^2 mst2^6) Xt^3)/(3 (mg^2-mst1^2)^3 (mst1^2-mst2^2)^3)+(4 (1156 mg^10 mst1^2-4276 mg^8 mst1^4+60 mg^6 msq^2 mst1^4-90 mg^4 msq^4 mst1^4+4726 mg^6 mst1^6+120 mg^4 msq^2 mst1^6+60 mg^2 msq^4 mst1^6-1163 mg^4 mst1^8-180 mg^2 msq^2 mst1^8+30 msq^4 mst1^8-604 mg^2 mst1^10+169 mst1^12+124 mg^10 mst2^2-872 mg^8 mst1^2 mst2^2+2314 mg^6 mst1^4 mst2^2-2344 mg^4 mst1^6 mst2^2+638 mg^2 mst1^8 mst2^2+140 mst1^10 mst2^2+44 mg^8 mst2^4-60 mg^6 msq^2 mst2^4+90 mg^4 msq^4 mst2^4+66 mg^6 mst1^2 mst2^4-120 mg^4 msq^2 mst1^2 mst2^4-60 mg^2 msq^4 mst1^2 mst2^4-222 mg^4 mst1^4 mst2^4+180 mg^2 msq^2 mst1^4 mst2^4-30 msq^4 mst1^4 mst2^4+134 mg^2 mst1^6 mst2^4-30 mst1^8 mst2^4-2 mg^6 mst2^6-24 mg^4 mst1^2 mst2^6+30 mg^2 mst1^4 mst2^6-4 mst1^6 mst2^6+9 mg^4 mst2^8-6 mg^2 mst1^2 mst2^8-3 mst1^4 mst2^8) Xt^4)/(3 (mg^2-mst1^2)^4 (mst1^2-mst2^2)^4)+(32 (14 mg^7 mst1^4+30 mg^3 msq^4 mst1^4+86 mg^5 mst1^6-60 mg^3 msq^2 mst1^6-30 mg msq^4 mst1^6-219 mg^3 mst1^8+60 mg msq^2 mst1^8+103 mg mst1^10-32 mg^7 mst1^2 mst2^2+145 mg^5 mst1^4 mst2^2-200 mg^3 mst1^6 mst2^2+87 mg mst1^8 mst2^2+18 mg^7 mst2^4-30 mg^3 msq^4 mst2^4-70 mg^5 mst1^2 mst2^4+60 mg^3 msq^2 mst1^2 mst2^4+30 mg msq^4 mst1^2 mst2^4+94 mg^3 mst1^4 mst2^4-60 mg msq^2 mst1^4 mst2^4-26 mg mst1^6 mst2^4-mg^5 mst2^6+8 mg^3 mst1^2 mst2^6-7 mg mst1^4 mst2^6-3 mg^3 mst2^8+3 mg mst1^2 mst2^8) Xt^5)/(3 (mg^2-mst1^2)^3 (mst1^2-mst2^2)^5)+(1280 mg^2 mst1^4 (mst1^2+mst2^2) Xt^6)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2)^5)) Log[mst1^2/MR^2]^3+((4 (128 mg^16-380 mg^14 mst1^2-60 mg^12 msq^2 mst1^2+90 mg^10 msq^4 mst1^2+284 mg^12 mst1^4+180 mg^10 msq^2 mst1^4-270 mg^8 msq^4 mst1^4+81 mg^10 mst1^6-180 mg^8 msq^2 mst1^6+270 mg^6 msq^4 mst1^6-101 mg^8 mst1^8+60 mg^6 msq^2 mst1^8-90 mg^4 msq^4 mst1^8-1028 mg^14 mst2^2+60 mg^12 msq^2 mst2^2-90 mg^10 msq^4 mst2^2+4136 mg^12 mst1^2 mst2^2-480 mg^10 msq^2 mst1^2 mst2^2+210 mg^8 msq^4 mst1^2 mst2^2-5977 mg^10 mst1^4 mst2^2+1080 mg^8 msq^2 mst1^4 mst2^2-90 mg^6 msq^4 mst1^4 mst2^2+3481 mg^8 mst1^6 mst2^2-960 mg^6 msq^2 mst1^6 mst2^2-90 mg^4 msq^4 mst1^6 mst2^2-714 mg^6 mst1^8 mst2^2+300 mg^4 msq^2 mst1^8 mst2^2+60 mg^2 msq^4 mst1^8 mst2^2+18 mg^4 mst1^10 mst2^2+1340 mg^12 mst2^4+300 mg^10 msq^2 mst2^4+60 mg^8 msq^4 mst2^4-5925 mg^10 mst1^2 mst2^4-600 mg^8 msq^2 mst1^2 mst2^4-210 mg^6 msq^4 mst1^2 mst2^4+9653 mg^8 mst1^4 mst2^4+270 mg^4 msq^4 mst1^4 mst2^4-6198 mg^6 mst1^6 mst2^4+600 mg^4 msq^2 mst1^6 mst2^4-150 mg^2 msq^4 mst1^6 mst2^4+1394 mg^4 mst1^8 mst2^4-300 mg^2 msq^2 mst1^8 mst2^4+30 msq^4 mst1^8 mst2^4-12 mg^2 mst1^10 mst2^4-339 mg^10 mst2^6-300 mg^8 msq^2 mst2^6+30 mg^6 msq^4 mst2^6+2199 mg^8 mst1^2 mst2^6+960 mg^6 msq^2 mst1^2 mst2^6-90 mg^4 msq^4 mst1^2 mst2^6-5326 mg^6 mst1^4 mst2^6-1080 mg^4 msq^2 mst1^4 mst2^6+90 mg^2 msq^4 mst1^4 mst2^6+4070 mg^4 mst1^6 mst2^6+480 mg^2 msq^2 mst1^6 mst2^6-30 msq^4 mst1^6 mst2^6-1018 mg^2 mst1^8 mst2^6-60 msq^2 mst1^8 mst2^6-6 mst1^10 mst2^6-512 mg^8 mst2^8-60 mg^6 msq^2 mst2^8+1642 mg^6 mst1^2 mst2^8+180 mg^4 msq^2 mst1^2 mst2^8-410 mg^4 mst1^4 mst2^8-180 mg^2 msq^2 mst1^4 mst2^8-603 mg^2 mst1^6 mst2^8+60 msq^2 mst1^6 mst2^8+303 mst1^8 mst2^8+228 mg^6 mst2^10-1160 mg^4 mst1^2 mst2^10+967 mg^2 mst1^4 mst2^10-287 mst1^6 mst2^10+56 mg^4 mst2^12+35 mg^2 mst1^2 mst2^12-7 mst1^4 mst2^12-9 mg^2 mst2^14-3 mst1^2 mst2^14))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2))-(32 (-32 mg^11 mst1^2+30 mg^7 msq^4 mst1^2+33 mg^9 mst1^4-60 mg^5 msq^4 mst1^4-4 mg^7 mst1^6+30 mg^3 msq^4 mst1^6+32 mg^11 mst2^2-30 mg^7 msq^4 mst2^2+348 mg^9 mst1^2 mst2^2-120 mg^7 msq^2 mst1^2 mst2^2+30 mg^5 msq^4 mst1^2 mst2^2-676 mg^7 mst1^4 mst2^2+240 mg^5 msq^2 mst1^4 mst2^2+30 mg^3 msq^4 mst1^4 mst2^2+320 mg^5 mst1^6 mst2^2-120 mg^3 msq^2 mst1^6 mst2^2-30 mg msq^4 mst1^6 mst2^2-6 mg^3 mst1^8 mst2^2-397 mg^9 mst2^4+120 mg^7 msq^2 mst2^4+30 mg^5 msq^4 mst2^4+260 mg^7 mst1^2 mst2^4-120 mg^5 msq^2 mst1^2 mst2^4-60 mg^3 msq^4 mst1^2 mst2^4+474 mg^5 mst1^4 mst2^4-120 mg^3 msq^2 mst1^4 mst2^4+30 mg msq^4 mst1^4 mst2^4-388 mg^3 mst1^6 mst2^4+120 mg msq^2 mst1^6 mst2^4+6 mg mst1^8 mst2^4+484 mg^7 mst2^6-120 mg^5 msq^2 mst2^6-680 mg^5 mst1^2 mst2^6+240 mg^3 msq^2 mst1^2 mst2^6+104 mg^3 mst1^4 mst2^6-120 mg msq^2 mst1^4 mst2^6+152 mg mst1^6 mst2^6-210 mg^5 mst2^8+340 mg^3 mst1^2 mst2^8-175 mg mst1^4 mst2^8+14 mg^3 mst2^10+4 mg mst1^2 mst2^10-3 mg mst2^12) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(16 (-24 mg^10 mst1^2+65 mg^8 mst1^4-41 mg^6 mst1^6+492 mg^10 mst2^2-1462 mg^8 mst1^2 mst2^2+90 mg^6 msq^2 mst1^2 mst2^2+1346 mg^6 mst1^4 mst2^2-180 mg^4 msq^2 mst1^4 mst2^2-373 mg^4 mst1^6 mst2^2+90 mg^2 msq^2 mst1^6 mst2^2+9 mg^2 mst1^8 mst2^2-383 mg^8 mst2^4-90 mg^6 msq^2 mst2^4+1473 mg^6 mst1^2 mst2^4+210 mg^4 msq^2 mst1^2 mst2^4-1612 mg^4 mst1^4 mst2^4-150 mg^2 msq^2 mst1^4 mst2^4+471 mg^2 mst1^6 mst2^4+30 msq^2 mst1^6 mst2^4+3 mst1^8 mst2^4-226 mg^6 mst2^6-30 mg^4 msq^2 mst2^6+137 mg^4 mst1^2 mst2^6+60 mg^2 msq^2 mst1^2 mst2^6+338 mg^2 mst1^4 mst2^6-30 msq^2 mst1^4 mst2^6-177 mst1^6 mst2^6+192 mg^4 mst2^8-391 mg^2 mst1^2 mst2^8+151 mst1^4 mst2^8+9 mg^2 mst2^10+3 mst1^2 mst2^10) Xt^2)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(128 (11 mg^9+7 mg^7 mst1^2-14 mg^5 mst1^4-37 mg^7 mst2^2+150 mg^5 mst1^2 mst2^2-30 mg^3 msq^2 mst1^2 mst2^2-117 mg^3 mst1^4 mst2^2+30 mg msq^2 mst1^4 mst2^2+3 mg mst1^6 mst2^2-48 mg^5 mst2^4+30 mg^3 msq^2 mst2^4-47 mg^3 mst1^2 mst2^4-30 mg msq^2 mst1^2 mst2^4+82 mg mst1^4 mst2^4+66 mg^3 mst2^6-53 mg mst1^2 mst2^6-3 mg mst2^8) Xt^3)/(3 (mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)-(8 (-176 mg^16 mst1^2+498 mg^14 mst1^4+60 mg^12 msq^2 mst1^4-90 mg^10 msq^4 mst1^4-481 mg^12 mst1^6-180 mg^10 msq^2 mst1^6+270 mg^8 msq^4 mst1^6+151 mg^10 mst1^8+180 mg^8 msq^2 mst1^8-270 mg^6 msq^4 mst1^8-4 mg^8 mst1^10-60 mg^6 msq^2 mst1^10+90 mg^4 msq^4 mst1^10+176 mg^16 mst2^2-296 mg^14 mst1^2 mst2^2-120 mg^12 msq^2 mst1^2 mst2^2+180 mg^10 msq^4 mst1^2 mst2^2+357 mg^12 mst1^4 mst2^2+390 mg^10 msq^2 mst1^4 mst2^2-480 mg^8 msq^4 mst1^4 mst2^2-691 mg^10 mst1^6 mst2^2-450 mg^8 msq^2 mst1^6 mst2^2+360 mg^6 msq^4 mst1^6 mst2^2+816 mg^8 mst1^8 mst2^2+210 mg^6 msq^2 mst1^8 mst2^2-275 mg^6 mst1^10 mst2^2-30 mg^4 msq^2 mst1^10 mst2^2-60 mg^2 msq^4 mst1^10 mst2^2+9 mg^4 mst1^12 mst2^2-2586 mg^14 mst2^4+60 mg^12 msq^2 mst2^4-90 mg^10 msq^4 mst2^4+8817 mg^12 mst1^2 mst2^4-420 mg^10 msq^2 mst1^2 mst2^4+150 mg^8 msq^4 mst1^2 mst2^4-12407 mg^10 mst1^4 mst2^4+780 mg^8 msq^2 mst1^4 mst2^4+120 mg^6 msq^4 mst1^4 mst2^4+8970 mg^8 mst1^6 mst2^4-420 mg^6 msq^2 mst1^6 mst2^4-360 mg^4 msq^4 mst1^6 mst2^4-3855 mg^6 mst1^8 mst2^4-120 mg^4 msq^2 mst1^8 mst2^4+210 mg^2 msq^4 mst1^8 mst2^4+731 mg^4 mst1^10 mst2^4+120 mg^2 msq^2 mst1^10 mst2^4-30 msq^4 mst1^10 mst2^4-6 mg^2 mst1^12 mst2^4+5691 mg^12 mst2^6+210 mg^10 msq^2 mst2^6+60 mg^8 msq^4 mst2^6-19237 mg^10 mst1^2 mst2^6-270 mg^8 msq^2 mst1^2 mst2^6-240 mg^6 msq^4 mst1^2 mst2^6+25376 mg^8 mst1^4 mst2^6-420 mg^6 msq^2 mst1^4 mst2^6+360 mg^4 msq^4 mst1^4 mst2^6-15256 mg^6 mst1^6 mst2^6+780 mg^4 msq^2 mst1^6 mst2^6-240 mg^2 msq^4 mst1^6 mst2^6+4746 mg^4 mst1^8 mst2^6-270 mg^2 msq^2 mst1^8 mst2^6+60 msq^4 mst1^8 mst2^6-645 mg^2 mst1^10 mst2^6-30 msq^2 mst1^10 mst2^6-3 mst1^12 mst2^6-4056 mg^10 mst2^8-240 mg^8 msq^2 mst2^8+30 mg^6 msq^4 mst2^8+13306 mg^8 mst1^2 mst2^8+720 mg^6 msq^2 mst1^2 mst2^8-90 mg^4 msq^4 mst1^2 mst2^8-17186 mg^6 mst1^4 mst2^8-720 mg^4 msq^2 mst1^4 mst2^8+90 mg^2 msq^4 mst1^4 mst2^8+8973 mg^4 mst1^6 mst2^8+240 mg^2 msq^2 mst1^6 mst2^8-30 msq^4 mst1^6 mst2^8-2062 mg^2 mst1^8 mst2^8+185 mst1^10 mst2^8+416 mg^8 mst2^10-30 mg^6 msq^2 mst2^10-1117 mg^6 mst1^2 mst2^10+90 mg^4 msq^2 mst1^2 mst2^10+2382 mg^4 mst1^4 mst2^10-90 mg^2 msq^2 mst1^4 mst2^10-1253 mg^2 mst1^6 mst2^10+30 msq^2 mst1^6 mst2^10+244 mst1^8 mst2^10+329 mg^6 mst2^12-1384 mg^4 mst1^2 mst2^12+1061 mg^2 mst1^4 mst2^12-342 mst1^6 mst2^12+47 mg^4 mst2^14+50 mg^2 mst1^2 mst2^14-mst1^4 mst2^14-9 mg^2 mst2^16-3 mst1^2 mst2^16) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^4)-(64 (34 mg^11 mst1^2-30 mg^7 msq^4 mst1^2-69 mg^9 mst1^4+60 mg^5 msq^4 mst1^4+38 mg^7 mst1^6-30 mg^3 msq^4 mst1^6+30 mg^11 mst2^2+30 mg^7 msq^4 mst2^2-36 mg^9 mst1^2 mst2^2+30 mg^7 msq^2 mst1^2 mst2^2-30 mg^5 msq^4 mst1^2 mst2^2-14 mg^7 mst1^4 mst2^2-60 mg^5 msq^2 mst1^4 mst2^2-30 mg^3 msq^4 mst1^4 mst2^2+5 mg^5 mst1^6 mst2^2+30 mg^3 msq^2 mst1^6 mst2^2+30 mg msq^4 mst1^6 mst2^2-3 mg^3 mst1^8 mst2^2+179 mg^9 mst2^4-90 mg^7 msq^2 mst2^4-30 mg^5 msq^4 mst2^4-470 mg^7 mst1^2 mst2^4+150 mg^5 msq^2 mst1^2 mst2^4+60 mg^3 msq^4 mst1^2 mst2^4+427 mg^5 mst1^4 mst2^4-30 mg^3 msq^2 mst1^4 mst2^4-30 mg msq^4 mst1^4 mst2^4-94 mg^3 mst1^6 mst2^4-30 mg msq^2 mst1^6 mst2^4+3 mg mst1^8 mst2^4-442 mg^7 mst2^6+90 mg^5 msq^2 mst2^6+923 mg^5 mst1^2 mst2^6-180 mg^3 msq^2 mst1^2 mst2^6-608 mg^3 mst1^4 mst2^6+90 mg msq^2 mst1^4 mst2^6+67 mg mst1^6 mst2^6+225 mg^5 mst2^8-396 mg^3 mst1^2 mst2^8+216 mg mst1^4 mst2^8-11 mg^3 mst2^10-7 mg mst1^2 mst2^10+3 mg mst2^12) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^4)-(2560 mg^2 mst2^4 Xt^6)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)+(-(1/(3 (mg^2-mst2^2)^3))80 (3 mg^6-6 mg^4 msq^2+9 mg^2 msq^4-7 mg^4 mst2^2-18 mg^2 msq^2 mst2^2+3 msq^4 mst2^2+6 mg^2 mst2^4-2 mst2^6)+(320 (2 mg^5+6 mg msq^4-3 mg^3 mst2^2-12 mg msq^2 mst2^2+mg mst2^4) Xt)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+((160 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst2^2-msq^4 mst2^2))/((mg^2-mst2^2)^3 (-mst1^2+mst2^2))-(160 (-2 mst1^2 mst2^4+4 mst2^6+mg^4 (-mst1^2+3 mst2^2)+mg^2 (4 mst1^2 mst2^2-8 mst2^4)))/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)) Xt^2+(-((1280 (-mg msq^4+2 mg msq^2 mst2^2))/((mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2))+(320 mg (2 mg^4+3 mst1^2 mst2^2-mst2^4-mg^2 (mst1^2+3 mst2^2)))/(3 (mg^2-mst2^2) (-mst1^2+mst2^2)^3)) Xt^3+(-((80 (mst1^2+mst2^2) (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst2^2-msq^4 mst2^2))/((mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3))+(80 (8 mg^4 mst1^2 mst2^2-mst1^4 mst2^4+4 mst1^2 mst2^6+5 mst2^8+2 mg^6 (-mst1^2+mst2^2)-2 mg^2 (-mst1^4 mst2^2+5 mst1^2 mst2^4+4 mst2^6)))/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)) Xt^4+(320 (mst1^2+mst2^2) (-6 mg msq^4+mg^3 mst2^2+12 mg msq^2 mst2^2-mg mst2^4) Xt^5)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)) Log[msq^2/MR^2]+(-((4 (128 mg^16-346 mg^14 mst1^2-60 mg^12 msq^2 mst1^2+90 mg^10 msq^4 mst1^2+476 mg^12 mst1^4+180 mg^10 msq^2 mst1^4-270 mg^8 msq^4 mst1^4-307 mg^10 mst1^6-180 mg^8 msq^2 mst1^6+270 mg^6 msq^4 mst1^6+61 mg^8 mst1^8+60 mg^6 msq^2 mst1^8-90 mg^4 msq^4 mst1^8-550 mg^14 mst2^2+60 mg^12 msq^2 mst2^2-90 mg^10 msq^4 mst2^2+1542 mg^12 mst1^2 mst2^2-300 mg^10 msq^2 mst1^2 mst2^2+210 mg^8 msq^4 mst1^2 mst2^2-2307 mg^10 mst1^4 mst2^2+540 mg^8 msq^2 mst1^4 mst2^2-90 mg^6 msq^4 mst1^4 mst2^2+1699 mg^8 mst1^6 mst2^2-420 mg^6 msq^2 mst1^6 mst2^2-90 mg^4 msq^4 mst1^6 mst2^2-420 mg^6 mst1^8 mst2^2+120 mg^4 msq^2 mst1^8 mst2^2+60 mg^2 msq^4 mst1^8 mst2^2+670 mg^12 mst2^4+120 mg^10 msq^2 mst2^4+60 mg^8 msq^4 mst2^4-1619 mg^10 mst1^2 mst2^4-180 mg^8 msq^2 mst1^2 mst2^4-210 mg^6 msq^4 mst1^2 mst2^4+2585 mg^8 mst1^4 mst2^4-180 mg^6 msq^2 mst1^4 mst2^4+270 mg^4 msq^4 mst1^4 mst2^4-2306 mg^6 mst1^6 mst2^4+420 mg^4 msq^2 mst1^6 mst2^4-150 mg^2 msq^4 mst1^6 mst2^4+658 mg^4 mst1^8 mst2^4-180 mg^2 msq^2 mst1^8 mst2^4+30 msq^4 mst1^8 mst2^4-247 mg^10 mst2^6-180 mg^8 msq^2 mst2^6+30 mg^6 msq^4 mst2^6+199 mg^8 mst1^2 mst2^6+540 mg^6 msq^2 mst1^2 mst2^6-90 mg^4 msq^4 mst1^2 mst2^6-386 mg^6 mst1^4 mst2^6-540 mg^4 msq^2 mst1^4 mst2^6+90 mg^2 msq^4 mst1^4 mst2^6+1002 mg^4 mst1^6 mst2^6+180 mg^2 msq^2 mst1^6 mst2^6-30 msq^4 mst1^6 mst2^6-388 mg^2 mst1^8 mst2^6-64 mg^8 mst2^8+324 mg^6 mst1^2 mst2^8-638 mg^4 mst1^4 mst2^8-19 mg^2 mst1^6 mst2^8+97 mst1^8 mst2^8+100 mg^6 mst2^10-88 mg^4 mst1^2 mst2^10+317 mg^2 mst1^4 mst2^10-101 mst1^6 mst2^10-38 mg^4 mst2^12-47 mg^2 mst1^2 mst2^12+mst1^4 mst2^12+9 mg^2 mst2^14+3 mst1^2 mst2^14))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)))+(32 (-20 mg^11 mst1^2+30 mg^7 msq^4 mst1^2+56 mg^9 mst1^4-60 mg^5 msq^4 mst1^4-7 mg^7 mst1^6+30 mg^3 msq^4 mst1^6+20 mg^11 mst2^2-30 mg^7 msq^4 mst2^2+51 mg^9 mst1^2 mst2^2-60 mg^7 msq^2 mst1^2 mst2^2+30 mg^5 msq^4 mst1^2 mst2^2-244 mg^7 mst1^4 mst2^2+120 mg^5 msq^2 mst1^4 mst2^2+30 mg^3 msq^4 mst1^4 mst2^2+67 mg^5 mst1^6 mst2^2-60 mg^3 msq^2 mst1^6 mst2^2-30 mg msq^4 mst1^6 mst2^2-91 mg^9 mst2^4+60 mg^7 msq^2 mst2^4+30 mg^5 msq^4 mst2^4+65 mg^7 mst1^2 mst2^4-60 mg^5 msq^2 mst1^2 mst2^4-60 mg^3 msq^4 mst1^2 mst2^4+278 mg^5 mst1^4 mst2^4-60 mg^3 msq^2 mst1^4 mst2^4+30 mg msq^4 mst1^4 mst2^4-99 mg^3 mst1^6 mst2^4+60 mg msq^2 mst1^6 mst2^4+122 mg^7 mst2^6-60 mg^5 msq^2 mst2^6-219 mg^5 mst1^2 mst2^6+120 mg^3 msq^2 mst1^2 mst2^6-42 mg^3 mst1^4 mst2^6-60 mg msq^2 mst1^4 mst2^6+23 mg mst1^6 mst2^6-30 mg^5 mst2^8+85 mg^3 mst1^2 mst2^8-8 mg^3 mst2^10-10 mg mst1^2 mst2^10+3 mg mst2^12) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)+(8 (128 mg^16-846 mg^14 mst1^2-60 mg^12 msq^2 mst1^2+90 mg^10 msq^4 mst1^2+1200 mg^12 mst1^4+180 mg^10 msq^2 mst1^4-270 mg^8 msq^4 mst1^4-387 mg^10 mst1^6-180 mg^8 msq^2 mst1^6+270 mg^6 msq^4 mst1^6-83 mg^8 mst1^8+60 mg^6 msq^2 mst1^8-90 mg^4 msq^4 mst1^8-558 mg^14 mst2^2+60 mg^12 msq^2 mst2^2-90 mg^10 msq^4 mst2^2+3590 mg^12 mst1^2 mst2^2-300 mg^10 msq^2 mst1^2 mst2^2+210 mg^8 msq^4 mst1^2 mst2^2-5047 mg^10 mst1^4 mst2^2+540 mg^8 msq^2 mst1^4 mst2^2-90 mg^6 msq^4 mst1^4 mst2^2+1575 mg^8 mst1^6 mst2^2-420 mg^6 msq^2 mst1^6 mst2^2-90 mg^4 msq^4 mst1^6 mst2^2+308 mg^6 mst1^8 mst2^2+120 mg^4 msq^2 mst1^8 mst2^2+60 mg^2 msq^4 mst1^8 mst2^2+462 mg^12 mst2^4+120 mg^10 msq^2 mst2^4+60 mg^8 msq^4 mst2^4-4419 mg^10 mst1^2 mst2^4-180 mg^8 msq^2 mst1^2 mst2^4-210 mg^6 msq^4 mst1^2 mst2^4+6841 mg^8 mst1^4 mst2^4-180 mg^6 msq^2 mst1^4 mst2^4+270 mg^4 msq^4 mst1^4 mst2^4-1854 mg^6 mst1^6 mst2^4+420 mg^4 msq^2 mst1^6 mst2^4-150 mg^2 msq^4 mst1^6 mst2^4-514 mg^4 mst1^8 mst2^4-180 mg^2 msq^2 mst1^8 mst2^4+30 msq^4 mst1^8 mst2^4+241 mg^10 mst2^6-180 mg^8 msq^2 mst2^6+30 mg^6 msq^4 mst2^6+1383 mg^8 mst1^2 mst2^6+540 mg^6 msq^2 mst1^2 mst2^6-90 mg^4 msq^4 mst1^2 mst2^6-3782 mg^6 mst1^4 mst2^6-540 mg^4 msq^2 mst1^4 mst2^6+90 mg^2 msq^4 mst1^4 mst2^6+742 mg^4 mst1^6 mst2^6+180 mg^2 msq^2 mst1^6 mst2^6-30 msq^4 mst1^6 mst2^6+396 mg^2 mst1^8 mst2^6-256 mg^8 mst2^8+636 mg^6 mst1^2 mst2^8+1010 mg^4 mst1^4 mst2^8-151 mg^2 mst1^6 mst2^8-99 mst1^8 mst2^8-96 mg^6 mst2^10-508 mg^4 mst1^2 mst2^10-171 mg^2 mst1^4 mst2^10+43 mst1^6 mst2^10+114 mg^4 mst2^12+141 mg^2 mst1^2 mst2^12-3 mst1^4 mst2^12-27 mg^2 mst2^14-9 mst1^2 mst2^14) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^2)+(32 (2 mg^13-33 mg^11 mst1^2-60 mg^7 msq^4 mst1^2-40 mg^9 mst1^4+120 mg^5 msq^4 mst1^4+61 mg^7 mst1^6-60 mg^3 msq^4 mst1^6-43 mg^11 mst2^2+60 mg^7 msq^4 mst2^2+403 mg^9 mst1^2 mst2^2+120 mg^7 msq^2 mst1^2 mst2^2-60 mg^5 msq^4 mst1^2 mst2^2-233 mg^7 mst1^4 mst2^2-240 mg^5 msq^2 mst1^4 mst2^2-60 mg^3 msq^4 mst1^4 mst2^2-155 mg^5 mst1^6 mst2^2+120 mg^3 msq^2 mst1^6 mst2^2+60 mg msq^4 mst1^6 mst2^2+243 mg^9 mst2^4-120 mg^7 msq^2 mst2^4-60 mg^5 msq^4 mst2^4-1127 mg^7 mst1^2 mst2^4+120 mg^5 msq^2 mst1^2 mst2^4+120 mg^3 msq^4 mst1^2 mst2^4+959 mg^5 mst1^4 mst2^4+120 mg^3 msq^2 mst1^4 mst2^4-60 mg msq^4 mst1^4 mst2^4+123 mg^3 mst1^6 mst2^4-120 mg msq^2 mst1^6 mst2^4-405 mg^7 mst2^6+120 mg^5 msq^2 mst2^6+1201 mg^5 mst1^2 mst2^6-240 mg^3 msq^2 mst1^2 mst2^6-1143 mg^3 mst1^4 mst2^6+120 mg msq^2 mst1^4 mst2^6+3 mg mst1^6 mst2^6+201 mg^5 mst2^8-288 mg^3 mst1^2 mst2^8+361 mg mst1^4 mst2^8-48 mg^3 mst2^10-60 mg mst1^2 mst2^10+18 mg mst2^12) Xt^3)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)-(4 (1156 mg^16 mst1^2-3032 mg^14 mst1^4-60 mg^12 msq^2 mst1^4+90 mg^10 msq^4 mst1^4+2486 mg^12 mst1^6+180 mg^10 msq^2 mst1^6-270 mg^8 msq^4 mst1^6-545 mg^10 mst1^8-180 mg^8 msq^2 mst1^8+270 mg^6 msq^4 mst1^8-53 mg^8 mst1^10+60 mg^6 msq^2 mst1^10-90 mg^4 msq^4 mst1^10+124 mg^16 mst2^2-5372 mg^14 mst1^2 mst2^2+14790 mg^12 mst1^4 mst2^2-120 mg^10 msq^2 mst1^4 mst2^2-60 mg^8 msq^4 mst1^4 mst2^2-13250 mg^10 mst1^6 mst2^2+360 mg^8 msq^2 mst1^6 mst2^2+180 mg^6 msq^4 mst1^6 mst2^2+3272 mg^8 mst1^8 mst2^2-360 mg^6 msq^2 mst1^8 mst2^2-180 mg^4 msq^4 mst1^8 mst2^2+316 mg^6 mst1^10 mst2^2+120 mg^4 msq^2 mst1^10 mst2^2+60 mg^2 msq^4 mst1^10 mst2^2-540 mg^14 mst2^4+60 mg^12 msq^2 mst2^4-90 mg^10 msq^4 mst2^4+9158 mg^12 mst1^2 mst2^4-180 mg^10 msq^2 mst1^2 mst2^4+270 mg^8 msq^4 mst1^2 mst2^4-25872 mg^10 mst1^4 mst2^4+360 mg^8 msq^2 mst1^4 mst2^4-300 mg^6 msq^4 mst1^4 mst2^4+25026 mg^8 mst1^6 mst2^4-600 mg^6 msq^2 mst1^6 mst2^4+180 mg^4 msq^4 mst1^6 mst2^4-6734 mg^6 mst1^8 mst2^4+540 mg^4 msq^2 mst1^8 mst2^4-90 mg^2 msq^4 mst1^8 mst2^4-654 mg^4 mst1^10 mst2^4-180 mg^2 msq^2 mst1^10 mst2^4+30 msq^4 mst1^10 mst2^4-178 mg^12 mst2^6+120 mg^10 msq^2 mst2^6+60 mg^8 msq^4 mst2^6-3530 mg^10 mst1^2 mst2^6-360 mg^8 msq^2 mst1^2 mst2^6-180 mg^6 msq^4 mst1^2 mst2^6+15488 mg^8 mst1^4 mst2^6+360 mg^6 msq^2 mst1^4 mst2^6+180 mg^4 msq^4 mst1^4 mst2^6-18644 mg^6 mst1^6 mst2^6-120 mg^4 msq^2 mst1^6 mst2^6-60 mg^2 msq^4 mst1^6 mst2^6+5828 mg^4 mst1^8 mst2^6+532 mg^2 mst1^10 mst2^6+1549 mg^10 mst2^8-180 mg^8 msq^2 mst2^8+30 mg^6 msq^4 mst2^8-4889 mg^8 mst1^2 mst2^8+540 mg^6 msq^2 mst1^2 mst2^8-90 mg^4 msq^4 mst1^2 mst2^8+2576 mg^6 mst1^4 mst2^8-540 mg^4 msq^2 mst1^4 mst2^8+90 mg^2 msq^4 mst1^4 mst2^8+2738 mg^4 mst1^6 mst2^8+180 mg^2 msq^2 mst1^6 mst2^8-30 msq^4 mst1^6 mst2^8-1721 mg^2 mst1^8 mst2^8-133 mst1^10 mst2^8-1004 mg^8 mst2^10+3876 mg^6 mst1^2 mst2^10-3994 mg^4 mst1^4 mst2^10+1654 mg^2 mst1^6 mst2^10-124 mst1^8 mst2^10-30 mg^6 mst2^12-528 mg^4 mst1^2 mst2^12+72 mg^2 mst1^4 mst2^12+6 mst1^6 mst2^12+114 mg^4 mst2^14+114 mg^2 mst1^2 mst2^14-12 mst1^4 mst2^14-27 mg^2 mst2^16-9 mst1^2 mst2^16) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^4)+(32 (14 mg^11 mst1^4+30 mg^7 msq^4 mst1^4+6 mg^9 mst1^6-60 mg^5 msq^4 mst1^6-23 mg^7 mst1^8+30 mg^3 msq^4 mst1^8-32 mg^11 mst1^2 mst2^2-155 mg^9 mst1^4 mst2^2-60 mg^7 msq^2 mst1^4 mst2^2-30 mg^5 msq^4 mst1^4 mst2^2+207 mg^7 mst1^6 mst2^2+120 mg^5 msq^2 mst1^6 mst2^2+60 mg^3 msq^4 mst1^6 mst2^2+23 mg^5 mst1^8 mst2^2-60 mg^3 msq^2 mst1^8 mst2^2-30 mg msq^4 mst1^8 mst2^2+18 mg^11 mst2^4-30 mg^7 msq^4 mst2^4-182 mg^9 mst1^2 mst2^4+60 mg^5 msq^4 mst1^2 mst2^4+719 mg^7 mst1^4 mst2^4+60 mg^5 msq^2 mst1^4 mst2^4-30 mg^3 msq^4 mst1^4 mst2^4-681 mg^5 mst1^6 mst2^4-120 mg^3 msq^2 mst1^6 mst2^4+25 mg^3 mst1^8 mst2^4+60 mg msq^2 mst1^8 mst2^4-149 mg^9 mst2^6+60 mg^7 msq^2 mst2^6+30 mg^5 msq^4 mst2^6+753 mg^7 mst1^2 mst2^6-120 mg^5 msq^2 mst1^2 mst2^6-60 mg^3 msq^4 mst1^2 mst2^6-1267 mg^5 mst1^4 mst2^6+60 mg^3 msq^2 mst1^4 mst2^6+30 mg msq^4 mst1^4 mst2^6+761 mg^3 mst1^6 mst2^6-41 mg mst1^8 mst2^6+264 mg^7 mst2^8-60 mg^5 msq^2 mst2^8-823 mg^5 mst1^2 mst2^8+120 mg^3 msq^2 mst1^2 mst2^8+879 mg^3 mst1^4 mst2^8-60 mg msq^2 mst1^4 mst2^8-261 mg mst1^6 mst2^8-132 mg^5 mst2^10+231 mg^3 mst1^2 mst2^10-190 mg mst1^4 mst2^10+24 mg^3 mst2^12+21 mg mst1^2 mst2^12-9 mg mst2^14) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^5)-(1280 mg^2 (mst1^2+mst2^2) (2 mg^2 mst1^2 mst2^2+mg^2 mst2^4-3 mst1^2 mst2^4) Xt^6)/(3 (mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^5)) Log[mst1^2/MR^2]) Log[mst2^2/MR^2]^2+(-((4 (128 mg^16-346 mg^14 mst1^2-60 mg^12 msq^2 mst1^2+90 mg^10 msq^4 mst1^2+280 mg^12 mst1^4+180 mg^10 msq^2 mst1^4-270 mg^8 msq^4 mst1^4-13 mg^10 mst1^6-180 mg^8 msq^2 mst1^6+270 mg^6 msq^4 mst1^6-37 mg^8 mst1^8+60 mg^6 msq^2 mst1^8-90 mg^4 msq^4 mst1^8-550 mg^14 mst2^2+60 mg^12 msq^2 mst2^2-90 mg^10 msq^4 mst2^2+1902 mg^12 mst1^2 mst2^2-300 mg^10 msq^2 mst1^2 mst2^2+210 mg^8 msq^4 mst1^2 mst2^2-2509 mg^10 mst1^4 mst2^2+540 mg^8 msq^2 mst1^4 mst2^2-90 mg^6 msq^4 mst1^4 mst2^2+1341 mg^8 mst1^6 mst2^2-420 mg^6 msq^2 mst1^6 mst2^2-90 mg^4 msq^4 mst1^6 mst2^2-268 mg^6 mst1^8 mst2^2+120 mg^4 msq^2 mst1^8 mst2^2+60 mg^2 msq^4 mst1^8 mst2^2+506 mg^12 mst2^4+120 mg^10 msq^2 mst2^4+60 mg^8 msq^4 mst2^4-1873 mg^10 mst1^2 mst2^4-180 mg^8 msq^2 mst1^2 mst2^4-210 mg^6 msq^4 mst1^2 mst2^4+3011 mg^8 mst1^4 mst2^4-180 mg^6 msq^2 mst1^4 mst2^4+270 mg^4 msq^4 mst1^4 mst2^4-1842 mg^6 mst1^6 mst2^4+420 mg^4 msq^2 mst1^6 mst2^4-150 mg^2 msq^4 mst1^6 mst2^4+450 mg^4 mst1^8 mst2^4-180 mg^2 msq^2 mst1^8 mst2^4+30 msq^4 mst1^8 mst2^4-85 mg^10 mst2^6-180 mg^8 msq^2 mst2^6+30 mg^6 msq^4 mst2^6+253 mg^8 mst1^2 mst2^6+540 mg^6 msq^2 mst1^2 mst2^6-90 mg^4 msq^4 mst1^2 mst2^6-1194 mg^6 mst1^4 mst2^6-540 mg^4 msq^2 mst1^4 mst2^6+90 mg^2 msq^4 mst1^4 mst2^6+874 mg^4 mst1^6 mst2^6+180 mg^2 msq^2 mst1^6 mst2^6-30 msq^4 mst1^6 mst2^6-268 mg^2 mst1^8 mst2^6-88 mg^8 mst2^8+612 mg^6 mst1^2 mst2^8-98 mg^4 mst1^4 mst2^8-73 mg^2 mst1^6 mst2^8+67 mst1^8 mst2^8+4 mg^6 mst2^10-368 mg^4 mst1^2 mst2^10+175 mg^2 mst1^4 mst2^10-63 mst1^6 mst2^10+38 mg^4 mst2^12+47 mg^2 mst1^2 mst2^12-mst1^4 mst2^12-9 mg^2 mst2^14-3 mst1^2 mst2^14))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)))+(32 (-20 mg^11 mst1^2+30 mg^7 msq^4 mst1^2+41 mg^9 mst1^4-60 mg^5 msq^4 mst1^4-24 mg^7 mst1^6+30 mg^3 msq^4 mst1^6+20 mg^11 mst2^2-30 mg^7 msq^4 mst2^2+145 mg^9 mst1^2 mst2^2-60 mg^7 msq^2 mst1^2 mst2^2+30 mg^5 msq^4 mst1^2 mst2^2-346 mg^7 mst1^4 mst2^2+120 mg^5 msq^2 mst1^4 mst2^2+30 mg^3 msq^4 mst1^4 mst2^2+199 mg^5 mst1^6 mst2^2-60 mg^3 msq^2 mst1^6 mst2^2-30 mg msq^4 mst1^6 mst2^2-202 mg^9 mst2^4+60 mg^7 msq^2 mst2^4+30 mg^5 msq^4 mst2^4+240 mg^7 mst1^2 mst2^4-60 mg^5 msq^2 mst1^2 mst2^4-60 mg^3 msq^4 mst1^2 mst2^4+91 mg^5 mst1^4 mst2^4-60 mg^3 msq^2 mst1^4 mst2^4+30 mg msq^4 mst1^4 mst2^4-174 mg^3 mst1^6 mst2^4+60 mg msq^2 mst1^6 mst2^4+194 mg^7 mst2^6-60 mg^5 msq^2 mst2^6-321 mg^5 mst1^2 mst2^6+120 mg^3 msq^2 mst1^2 mst2^6+140 mg^3 mst1^4 mst2^6-60 mg msq^2 mst1^4 mst2^6+47 mg mst1^6 mst2^6-65 mg^5 mst2^8+90 mg^3 mst1^2 mst2^8-70 mg mst1^4 mst2^8+8 mg^3 mst2^10+10 mg mst1^2 mst2^10-3 mg mst2^12) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(8 (128 mg^16-342 mg^14 mst1^2-60 mg^12 msq^2 mst1^2+90 mg^10 msq^4 mst1^2+268 mg^12 mst1^4+180 mg^10 msq^2 mst1^4-270 mg^8 msq^4 mst1^4-mg^10 mst1^6-180 mg^8 msq^2 mst1^6+270 mg^6 msq^4 mst1^6-41 mg^8 mst1^8+60 mg^6 msq^2 mst1^8-90 mg^4 msq^4 mst1^8-1062 mg^14 mst2^2+60 mg^12 msq^2 mst2^2-90 mg^10 msq^4 mst2^2+3302 mg^12 mst1^2 mst2^2-300 mg^10 msq^2 mst1^2 mst2^2+210 mg^8 msq^4 mst1^2 mst2^2-3637 mg^10 mst1^4 mst2^2+540 mg^8 msq^2 mst1^4 mst2^2-90 mg^6 msq^4 mst1^4 mst2^2+1445 mg^8 mst1^6 mst2^2-420 mg^6 msq^2 mst1^6 mst2^2-90 mg^4 msq^4 mst1^6 mst2^2-132 mg^6 mst1^8 mst2^2+120 mg^4 msq^2 mst1^8 mst2^2+60 mg^2 msq^4 mst1^8 mst2^2+1682 mg^12 mst2^4+120 mg^10 msq^2 mst2^4+60 mg^8 msq^4 mst2^4-5333 mg^10 mst1^2 mst2^4-180 mg^8 msq^2 mst1^2 mst2^4-210 mg^6 msq^4 mst1^2 mst2^4+6335 mg^8 mst1^4 mst2^4-180 mg^6 msq^2 mst1^4 mst2^4+270 mg^4 msq^4 mst1^4 mst2^4-2814 mg^6 mst1^6 mst2^4+420 mg^4 msq^2 mst1^6 mst2^4-150 mg^2 msq^4 mst1^6 mst2^4+382 mg^4 mst1^8 mst2^4-180 mg^2 msq^2 mst1^8 mst2^4+30 msq^4 mst1^8 mst2^4-641 mg^10 mst2^6-180 mg^8 msq^2 mst2^6+30 mg^6 msq^4 mst2^6+1921 mg^8 mst1^2 mst2^6+540 mg^6 msq^2 mst1^2 mst2^6-90 mg^4 msq^4 mst1^2 mst2^6-2862 mg^6 mst1^4 mst2^6-540 mg^4 msq^2 mst1^4 mst2^6+90 mg^2 msq^4 mst1^4 mst2^6+1430 mg^4 mst1^6 mst2^6+180 mg^2 msq^2 mst1^6 mst2^6-30 msq^4 mst1^6 mst2^6-268 mg^2 mst1^8 mst2^6-200 mg^8 mst2^8+948 mg^6 mst1^2 mst2^8-434 mg^4 mst1^4 mst2^8+39 mg^2 mst1^6 mst2^8+67 mst1^8 mst2^8+72 mg^6 mst2^10-572 mg^4 mst1^2 mst2^10+379 mg^2 mst1^4 mst2^10-131 mst1^6 mst2^10+38 mg^4 mst2^12+47 mg^2 mst1^2 mst2^12-mst1^4 mst2^12-9 mg^2 mst2^14-3 mst1^2 mst2^14) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^2)+(32 (2 mg^13-41 mg^11 mst1^2+60 mg^7 msq^4 mst1^2+78 mg^9 mst1^4-120 mg^5 msq^4 mst1^4-45 mg^7 mst1^6+60 mg^3 msq^4 mst1^6-35 mg^11 mst2^2-60 mg^7 msq^4 mst2^2+275 mg^9 mst1^2 mst2^2-120 mg^7 msq^2 mst1^2 mst2^2+60 mg^5 msq^4 mst1^2 mst2^2-437 mg^7 mst1^4 mst2^2+240 mg^5 msq^2 mst1^4 mst2^2+60 mg^3 msq^4 mst1^4 mst2^2+233 mg^5 mst1^6 mst2^2-120 mg^3 msq^2 mst1^6 mst2^2-60 mg msq^4 mst1^6 mst2^2-131 mg^9 mst2^4+120 mg^7 msq^2 mst2^4+60 mg^5 msq^4 mst2^4-49 mg^7 mst1^2 mst2^4-120 mg^5 msq^2 mst1^2 mst2^4-120 mg^3 msq^4 mst1^2 mst2^4+421 mg^5 mst1^4 mst2^4-120 mg^3 msq^2 mst1^4 mst2^4+60 mg msq^4 mst1^4 mst2^4-331 mg^3 mst1^6 mst2^4+120 mg msq^2 mst1^6 mst2^4+363 mg^7 mst2^6-120 mg^5 msq^2 mst2^6-575 mg^5 mst1^2 mst2^6+240 mg^3 msq^2 mst1^2 mst2^6+221 mg^3 mst1^4 mst2^6-120 mg msq^2 mst1^4 mst2^6+111 mg mst1^6 mst2^6-177 mg^5 mst2^8+274 mg^3 mst1^2 mst2^8-187 mg mst1^4 mst2^8+16 mg^3 mst2^10+20 mg mst1^2 mst2^10-6 mg mst2^12) Xt^3)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)+(4 (124 mg^16 mst1^2-328 mg^14 mst1^4-60 mg^12 msq^2 mst1^4+90 mg^10 msq^4 mst1^4+250 mg^12 mst1^6+180 mg^10 msq^2 mst1^6-270 mg^8 msq^4 mst1^6+9 mg^10 mst1^8-180 mg^8 msq^2 mst1^8+270 mg^6 msq^4 mst1^8-43 mg^8 mst1^10+60 mg^6 msq^2 mst1^10-90 mg^4 msq^4 mst1^10+1156 mg^16 mst2^2-4340 mg^14 mst1^2 mst2^2+6138 mg^12 mst1^4 mst2^2-120 mg^10 msq^2 mst1^4 mst2^2-60 mg^8 msq^4 mst1^4 mst2^2-4006 mg^10 mst1^6 mst2^2+360 mg^8 msq^2 mst1^6 mst2^2+180 mg^6 msq^4 mst1^6 mst2^2+1040 mg^8 mst1^8 mst2^2-360 mg^6 msq^2 mst1^8 mst2^2-180 mg^4 msq^4 mst1^8 mst2^2-60 mg^6 mst1^10 mst2^2+120 mg^4 msq^2 mst1^10 mst2^2+60 mg^2 msq^4 mst1^10 mst2^2-4276 mg^14 mst2^4+60 mg^12 msq^2 mst2^4-90 mg^10 msq^4 mst2^4+15130 mg^12 mst1^2 mst2^4-180 mg^10 msq^2 mst1^2 mst2^4+270 mg^8 msq^4 mst1^2 mst2^4-19952 mg^10 mst1^4 mst2^4+360 mg^8 msq^2 mst1^4 mst2^4-300 mg^6 msq^4 mst1^4 mst2^4+11990 mg^8 mst1^6 mst2^4-600 mg^6 msq^2 mst1^6 mst2^4+180 mg^4 msq^4 mst1^6 mst2^4-2934 mg^6 mst1^8 mst2^4+540 mg^4 msq^2 mst1^8 mst2^4-90 mg^2 msq^4 mst1^8 mst2^4+210 mg^4 mst1^10 mst2^4-180 mg^2 msq^2 mst1^10 mst2^4+30 msq^4 mst1^10 mst2^4+4738 mg^12 mst2^6+120 mg^10 msq^2 mst2^6+60 mg^8 msq^4 mst2^6-16510 mg^10 mst1^2 mst2^6-360 mg^8 msq^2 mst1^2 mst2^6-180 mg^6 msq^4 mst1^2 mst2^6+21392 mg^8 mst1^4 mst2^6+360 mg^6 msq^2 mst1^4 mst2^6+180 mg^4 msq^4 mst1^4 mst2^6-12428 mg^6 mst1^6 mst2^6-120 mg^4 msq^2 mst1^6 mst2^6-60 mg^2 msq^4 mst1^6 mst2^6+2772 mg^4 mst1^8 mst2^6-132 mg^2 mst1^10 mst2^6-1189 mg^10 mst2^8-180 mg^8 msq^2 mst2^8+30 mg^6 msq^4 mst2^8+4017 mg^8 mst1^2 mst2^8+540 mg^6 msq^2 mst1^2 mst2^8-90 mg^4 msq^4 mst1^2 mst2^8-5480 mg^6 mst1^4 mst2^8-540 mg^4 msq^2 mst1^4 mst2^8+90 mg^2 msq^4 mst1^4 mst2^8+3398 mg^4 mst1^6 mst2^8+180 mg^2 msq^2 mst1^6 mst2^8-30 msq^4 mst1^6 mst2^8-779 mg^2 mst1^8 mst2^8+33 mst1^10 mst2^8-556 mg^8 mst2^10+2156 mg^6 mst1^2 mst2^10-2278 mg^4 mst1^4 mst2^10+978 mg^2 mst1^6 mst2^10-132 mst1^8 mst2^10+106 mg^6 mst2^12-636 mg^4 mst1^2 mst2^12+528 mg^2 mst1^4 mst2^12-166 mst1^6 mst2^12+38 mg^4 mst2^14+38 mg^2 mst1^2 mst2^14-4 mst1^4 mst2^14-9 mg^2 mst2^16-3 mst1^2 mst2^16) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^4)+(32 (18 mg^11 mst1^4-30 mg^7 msq^4 mst1^4-37 mg^9 mst1^6+60 mg^5 msq^4 mst1^6+22 mg^7 mst1^8-30 mg^3 msq^4 mst1^8-32 mg^11 mst1^2 mst2^2-6 mg^9 mst1^4 mst2^2+60 mg^7 msq^2 mst1^4 mst2^2+30 mg^5 msq^4 mst1^4 mst2^2+106 mg^7 mst1^6 mst2^2-120 mg^5 msq^2 mst1^6 mst2^2-60 mg^3 msq^4 mst1^6 mst2^2-83 mg^5 mst1^8 mst2^2+60 mg^3 msq^2 mst1^8 mst2^2+30 mg msq^4 mst1^8 mst2^2+14 mg^11 mst2^4+30 mg^7 msq^4 mst2^4+117 mg^9 mst1^2 mst2^4-60 mg^5 msq^4 mst1^2 mst2^4-182 mg^7 mst1^4 mst2^4-60 mg^5 msq^2 mst1^4 mst2^4+30 mg^3 msq^4 mst1^4 mst2^4-30 mg^5 mst1^6 mst2^4+120 mg^3 msq^2 mst1^6 mst2^4+108 mg^3 mst1^8 mst2^4-60 mg msq^2 mst1^8 mst2^4+86 mg^9 mst2^6-60 mg^7 msq^2 mst2^6-30 mg^5 msq^4 mst2^6-362 mg^7 mst1^2 mst2^6+120 mg^5 msq^2 mst1^2 mst2^6+60 mg^3 msq^4 mst1^2 mst2^6+450 mg^5 mst1^4 mst2^6-60 mg^3 msq^2 mst1^4 mst2^6-30 mg msq^4 mst1^4 mst2^6-158 mg^3 mst1^6 mst2^6-31 mg mst1^8 mst2^6-224 mg^7 mst2^8+60 mg^5 msq^2 mst2^8+510 mg^5 mst1^2 mst2^8-120 mg^3 msq^2 mst1^2 mst2^8-388 mg^3 mst1^4 mst2^8+60 mg msq^2 mst1^4 mst2^8+87 mg mst1^6 mst2^8+113 mg^5 mst2^10-194 mg^3 mst1^2 mst2^10+108 mg mst1^4 mst2^10-8 mg^3 mst2^12-7 mg mst1^2 mst2^12+3 mg mst2^14) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^5)+(1280 mg^2 mst2^4 (mst1^2+mst2^2) Xt^6)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^5)) Log[mst2^2/MR^2]^3+Log[mg^2/MR^2]^2 (-((8 (88 mg^18 mst1^2-288 mg^16 mst1^4+368 mg^14 mst1^6-224 mg^12 mst1^8+56 mg^10 mst1^10+88 mg^18 mst2^2-302 mg^16 mst1^2 mst2^2+838 mg^14 mst1^4 mst2^2-797 mg^12 mst1^6 mst2^2+280 mg^10 mst1^8 mst2^2-39 mg^8 mst1^10 mst2^2-288 mg^16 mst2^4+838 mg^14 mst1^2 mst2^4-2636 mg^12 mst1^4 mst2^4+2588 mg^10 mst1^6 mst2^4-908 mg^8 mst1^8 mst2^4+134 mg^6 mst1^10 mst2^4+368 mg^14 mst2^6-797 mg^12 mst1^2 mst2^6+2588 mg^10 mst1^4 mst2^6-2048 mg^8 mst1^6 mst2^6+272 mg^6 mst1^8 mst2^6+25 mg^4 mst1^10 mst2^6-224 mg^12 mst2^8+280 mg^10 mst1^2 mst2^8-908 mg^8 mst1^4 mst2^8+272 mg^6 mst1^6 mst2^8+452 mg^4 mst1^8 mst2^8-144 mg^2 mst1^10 mst2^8+56 mg^10 mst2^10-39 mg^8 mst1^2 mst2^10+134 mg^6 mst1^4 mst2^10+25 mg^4 mst1^6 mst2^10-144 mg^2 mst1^8 mst2^10+36 mst1^10 mst2^10))/(3 mst1^2 (mg^2-mst1^2)^4 mst2^2 (mg^2-mst2^2)^4))+(64 (16 mg^13 mst1^2-32 mg^11 mst1^4+16 mg^9 mst1^6+16 mg^13 mst2^2+31 mg^11 mst1^2 mst2^2+18 mg^9 mst1^4 mst2^2-25 mg^7 mst1^6 mst2^2-32 mg^11 mst2^4+18 mg^9 mst1^2 mst2^4-196 mg^7 mst1^4 mst2^4+130 mg^5 mst1^6 mst2^4+16 mg^9 mst2^6-25 mg^7 mst1^2 mst2^6+130 mg^5 mst1^4 mst2^6-81 mg^3 mst1^6 mst2^6) Xt)/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^3)+(-((1792 mg^6)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2))+(64 mg^4 (-mg^2+mst1^2+mst2^2) (2+(8 (2 mg^8-2 mg^6 mst1^2+mg^4 mst1^4-2 mg^6 mst2^2+mg^4 mst2^4))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)))/(mst1^2 (mg^2-mst1^2) mst2^2 (-mg^2+mst2^2))) Xt^2-(2048 mg^7 (mg^2-mst1^2-mst2^2) Xt^3)/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^2)-(16 (44 mg^18 mst1^2-144 mg^16 mst1^4+184 mg^14 mst1^6-112 mg^12 mst1^8+28 mg^10 mst1^10+44 mg^18 mst2^2-2 mg^16 mst1^2 mst2^2-294 mg^14 mst1^4 mst2^2+277 mg^12 mst1^6 mst2^2+4 mg^10 mst1^8 mst2^2-25 mg^8 mst1^10 mst2^2-144 mg^16 mst2^4-294 mg^14 mst1^2 mst2^4+1932 mg^12 mst1^4 mst2^4-2012 mg^10 mst1^6 mst2^4+572 mg^8 mst1^8 mst2^4-70 mg^6 mst1^10 mst2^4+184 mg^14 mst2^6+277 mg^12 mst1^2 mst2^6-2012 mg^10 mst1^4 mst2^6+1712 mg^8 mst1^6 mst2^6-88 mg^6 mst1^8 mst2^6-49 mg^4 mst1^10 mst2^6-112 mg^12 mst2^8+4 mg^10 mst1^2 mst2^8+572 mg^8 mst1^4 mst2^8-88 mg^6 mst1^6 mst2^8-548 mg^4 mst1^8 mst2^8+156 mg^2 mst1^10 mst2^8+28 mg^10 mst2^10-25 mg^8 mst1^2 mst2^10-70 mg^6 mst1^4 mst2^10-49 mg^4 mst1^6 mst2^10+156 mg^2 mst1^8 mst2^10-36 mst1^10 mst2^10) Xt^4)/(3 mst1^2 (mg^2-mst1^2)^4 mst2^2 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^2)+(128 (8 mg^13 mst1^2-16 mg^11 mst1^4+8 mg^9 mst1^6+8 mg^13 mst2^2+7 mg^11 mst1^2 mst2^2-54 mg^9 mst1^4 mst2^2+47 mg^7 mst1^6 mst2^2-16 mg^11 mst2^4-54 mg^9 mst1^2 mst2^4+188 mg^7 mst1^4 mst2^4-134 mg^5 mst1^6 mst2^4+8 mg^9 mst2^6+47 mg^7 mst1^2 mst2^6-134 mg^5 mst1^4 mst2^6+87 mg^3 mst1^6 mst2^6) Xt^5)/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(2560 mg^6 Xt^6)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)+(-((8 (198 mg^16-814 mg^14 mst1^2+877 mg^12 mst1^4-344 mg^10 mst1^6+55 mg^8 mst1^8-638 mg^14 mst2^2+2700 mg^12 mst1^2 mst2^2-2796 mg^10 mst1^4 mst2^2+1004 mg^8 mst1^6 mst2^2-158 mg^6 mst1^8 mst2^2+565 mg^12 mst2^4-2604 mg^10 mst1^2 mst2^4+2240 mg^8 mst1^4 mst2^4-368 mg^6 mst1^6 mst2^4-mg^4 mst1^8 mst2^4-120 mg^10 mst2^6+812 mg^8 mst1^2 mst2^6-272 mg^6 mst1^4 mst2^6-452 mg^4 mst1^6 mst2^6+144 mg^2 mst1^8 mst2^6-mg^8 mst2^8-110 mg^6 mst1^2 mst2^8-25 mg^4 mst1^4 mst2^8+144 mg^2 mst1^6 mst2^8-36 mst1^8 mst2^8))/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^4))+(64 (44 mg^13-79 mg^11 mst1^2+122 mg^9 mst1^4-63 mg^7 mst1^6-77 mg^11 mst2^2+72 mg^9 mst1^2 mst2^2-209 mg^7 mst1^4 mst2^2+142 mg^5 mst1^6 mst2^2+22 mg^9 mst2^4+101 mg^7 mst1^2 mst2^4+36 mg^5 mst1^4 mst2^4-87 mg^3 mst1^6 mst2^4+19 mg^7 mst2^6-118 mg^5 mst1^2 mst2^6+75 mg^3 mst1^4 mst2^6) Xt)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+(16 (30 mg^16-334 mg^14 mst1^2+517 mg^12 mst1^4-264 mg^10 mst1^6+55 mg^8 mst1^8-238 mg^14 mst2^2+1484 mg^12 mst1^2 mst2^2-1948 mg^10 mst1^4 mst2^2+844 mg^8 mst1^6 mst2^2-158 mg^6 mst1^8 mst2^2+325 mg^12 mst2^4-1660 mg^10 mst1^2 mst2^4+1648 mg^8 mst1^4 mst2^4-288 mg^6 mst1^6 mst2^4-mg^4 mst1^8 mst2^4-168 mg^10 mst2^6+652 mg^8 mst1^2 mst2^6-192 mg^6 mst1^4 mst2^6-452 mg^4 mst1^6 mst2^6+144 mg^2 mst1^8 mst2^6+55 mg^8 mst2^8-158 mg^6 mst1^2 mst2^8-mg^4 mst1^4 mst2^8+144 mg^2 mst1^6 mst2^8-36 mst1^8 mst2^8) Xt^2)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^4 (-mst1^2+mst2^2))+(128 (44 mg^15-100 mg^13 mst1^2+45 mg^11 mst1^4+66 mg^9 mst1^6-63 mg^7 mst1^8-100 mg^13 mst2^2+282 mg^11 mst1^2 mst2^2-250 mg^9 mst1^4 mst2^2-42 mg^7 mst1^6 mst2^2+142 mg^5 mst1^8 mst2^2+45 mg^11 mst2^4-250 mg^9 mst1^2 mst2^4+422 mg^7 mst1^4 mst2^4-178 mg^5 mst1^6 mst2^4-87 mg^3 mst1^8 mst2^4+66 mg^9 mst2^6-42 mg^7 mst1^2 mst2^6-178 mg^5 mst1^4 mst2^6+186 mg^3 mst1^6 mst2^6-63 mg^7 mst2^8+142 mg^5 mst1^2 mst2^8-87 mg^3 mst1^4 mst2^8) Xt^3)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)-(8 (1024 mg^18-2786 mg^16 mst1^2+1826 mg^14 mst1^4+549 mg^12 mst1^6-664 mg^10 mst1^8+55 mg^8 mst1^10-2786 mg^16 mst2^2+8004 mg^14 mst1^2 mst2^2-5903 mg^12 mst1^4 mst2^2-868 mg^10 mst1^6 mst2^2+1699 mg^8 mst1^8 mst2^2-158 mg^6 mst1^10 mst2^2+1826 mg^14 mst2^4-5903 mg^12 mst1^2 mst2^4+5224 mg^10 mst1^4 mst2^4-292 mg^8 mst1^6 mst2^4-846 mg^6 mst1^8 mst2^4-mg^4 mst1^10 mst2^4+549 mg^12 mst2^6-868 mg^10 mst1^2 mst2^6-292 mg^8 mst1^4 mst2^6+928 mg^6 mst1^6 mst2^6-453 mg^4 mst1^8 mst2^6+144 mg^2 mst1^10 mst2^6-664 mg^10 mst2^8+1699 mg^8 mst1^2 mst2^8-846 mg^6 mst1^4 mst2^8-453 mg^4 mst1^6 mst2^8+288 mg^2 mst1^8 mst2^8-36 mst1^10 mst2^8+55 mg^8 mst2^10-158 mg^6 mst1^2 mst2^10-mg^4 mst1^4 mst2^10+144 mg^2 mst1^6 mst2^10-36 mst1^8 mst2^10) Xt^4)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^3)+(64 (39 mg^11 mst1^2-94 mg^9 mst1^4+63 mg^7 mst1^6+39 mg^11 mst2^2-188 mg^9 mst1^2 mst2^2+283 mg^7 mst1^4 mst2^2-142 mg^5 mst1^6 mst2^2-94 mg^9 mst2^4+283 mg^7 mst1^2 mst2^4-284 mg^5 mst1^4 mst2^4+87 mg^3 mst1^6 mst2^4+63 mg^7 mst2^6-142 mg^5 mst1^2 mst2^6+87 mg^3 mst1^4 mst2^6) Xt^5)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)-(1280 mg^6 (mst1^2+mst2^2) Xt^6)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)) Log[mst1^2/MR^2]+(-((8 (198 mg^16-638 mg^14 mst1^2+565 mg^12 mst1^4-120 mg^10 mst1^6-mg^8 mst1^8-814 mg^14 mst2^2+2700 mg^12 mst1^2 mst2^2-2604 mg^10 mst1^4 mst2^2+812 mg^8 mst1^6 mst2^2-110 mg^6 mst1^8 mst2^2+877 mg^12 mst2^4-2796 mg^10 mst1^2 mst2^4+2240 mg^8 mst1^4 mst2^4-272 mg^6 mst1^6 mst2^4-25 mg^4 mst1^8 mst2^4-344 mg^10 mst2^6+1004 mg^8 mst1^2 mst2^6-368 mg^6 mst1^4 mst2^6-452 mg^4 mst1^6 mst2^6+144 mg^2 mst1^8 mst2^6+55 mg^8 mst2^8-158 mg^6 mst1^2 mst2^8-mg^4 mst1^4 mst2^8+144 mg^2 mst1^6 mst2^8-36 mst1^8 mst2^8))/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^4))-(64 (44 mg^13-77 mg^11 mst1^2+22 mg^9 mst1^4+19 mg^7 mst1^6-79 mg^11 mst2^2+72 mg^9 mst1^2 mst2^2+101 mg^7 mst1^4 mst2^2-118 mg^5 mst1^6 mst2^2+122 mg^9 mst2^4-209 mg^7 mst1^2 mst2^4+36 mg^5 mst1^4 mst2^4+75 mg^3 mst1^6 mst2^4-63 mg^7 mst2^6+142 mg^5 mst1^2 mst2^6-87 mg^3 mst1^4 mst2^6) Xt)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))-(16 (30 mg^16-238 mg^14 mst1^2+325 mg^12 mst1^4-168 mg^10 mst1^6+55 mg^8 mst1^8-334 mg^14 mst2^2+1484 mg^12 mst1^2 mst2^2-1660 mg^10 mst1^4 mst2^2+652 mg^8 mst1^6 mst2^2-158 mg^6 mst1^8 mst2^2+517 mg^12 mst2^4-1948 mg^10 mst1^2 mst2^4+1648 mg^8 mst1^4 mst2^4-192 mg^6 mst1^6 mst2^4-mg^4 mst1^8 mst2^4-264 mg^10 mst2^6+844 mg^8 mst1^2 mst2^6-288 mg^6 mst1^4 mst2^6-452 mg^4 mst1^6 mst2^6+144 mg^2 mst1^8 mst2^6+55 mg^8 mst2^8-158 mg^6 mst1^2 mst2^8-mg^4 mst1^4 mst2^8+144 mg^2 mst1^6 mst2^8-36 mst1^8 mst2^8) Xt^2)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^4 (-mst1^2+mst2^2))-(128 (44 mg^15-100 mg^13 mst1^2+45 mg^11 mst1^4+66 mg^9 mst1^6-63 mg^7 mst1^8-100 mg^13 mst2^2+282 mg^11 mst1^2 mst2^2-250 mg^9 mst1^4 mst2^2-42 mg^7 mst1^6 mst2^2+142 mg^5 mst1^8 mst2^2+45 mg^11 mst2^4-250 mg^9 mst1^2 mst2^4+422 mg^7 mst1^4 mst2^4-178 mg^5 mst1^6 mst2^4-87 mg^3 mst1^8 mst2^4+66 mg^9 mst2^6-42 mg^7 mst1^2 mst2^6-178 mg^5 mst1^4 mst2^6+186 mg^3 mst1^6 mst2^6-63 mg^7 mst2^8+142 mg^5 mst1^2 mst2^8-87 mg^3 mst1^4 mst2^8) Xt^3)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)+(8 (1024 mg^18-2786 mg^16 mst1^2+1826 mg^14 mst1^4+549 mg^12 mst1^6-664 mg^10 mst1^8+55 mg^8 mst1^10-2786 mg^16 mst2^2+8004 mg^14 mst1^2 mst2^2-5903 mg^12 mst1^4 mst2^2-868 mg^10 mst1^6 mst2^2+1699 mg^8 mst1^8 mst2^2-158 mg^6 mst1^10 mst2^2+1826 mg^14 mst2^4-5903 mg^12 mst1^2 mst2^4+5224 mg^10 mst1^4 mst2^4-292 mg^8 mst1^6 mst2^4-846 mg^6 mst1^8 mst2^4-mg^4 mst1^10 mst2^4+549 mg^12 mst2^6-868 mg^10 mst1^2 mst2^6-292 mg^8 mst1^4 mst2^6+928 mg^6 mst1^6 mst2^6-453 mg^4 mst1^8 mst2^6+144 mg^2 mst1^10 mst2^6-664 mg^10 mst2^8+1699 mg^8 mst1^2 mst2^8-846 mg^6 mst1^4 mst2^8-453 mg^4 mst1^6 mst2^8+288 mg^2 mst1^8 mst2^8-36 mst1^10 mst2^8+55 mg^8 mst2^10-158 mg^6 mst1^2 mst2^10-mg^4 mst1^4 mst2^10+144 mg^2 mst1^6 mst2^10-36 mst1^8 mst2^10) Xt^4)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^3)-(64 (39 mg^11 mst1^2-94 mg^9 mst1^4+63 mg^7 mst1^6+39 mg^11 mst2^2-188 mg^9 mst1^2 mst2^2+283 mg^7 mst1^4 mst2^2-142 mg^5 mst1^6 mst2^2-94 mg^9 mst2^4+283 mg^7 mst1^2 mst2^4-284 mg^5 mst1^4 mst2^4+87 mg^3 mst1^6 mst2^4+63 mg^7 mst2^6-142 mg^5 mst1^2 mst2^6+87 mg^3 mst1^4 mst2^6) Xt^5)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)+(1280 mg^6 (mst1^2+mst2^2) Xt^6)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)) Log[mst2^2/MR^2])+(-((64 (3 mg^4+mg^2 mst1^2+mg^2 mst2^2-5 mst1^2 mst2^2))/((mg^2-mst1^2) (mg^2-mst2^2)))+(192+(256 (2 mg^8-2 mg^6 mst1^2+mg^4 mst1^4-2 mg^6 mst2^2+mg^4 mst2^4))/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)-(1024 mg^3 Xt)/((mg^2-mst1^2) (mg^2-mst2^2))) Log[mg^2/MR^2]+160 Log[msq^2/MR^2]+((16 (mg^4-34 mg^2 mst1^2+17 mst1^4))/(mg^2-mst1^2)^2+(1024 mg mst1^2 Xt)/((mg^2-mst1^2) (mst1^2-mst2^2))) Log[mst1^2/MR^2]+((16 (mg^4-34 mg^2 mst2^2+17 mst2^4))/(mg^2-mst2^2)^2+(1024 mg mst2^2 Xt)/((mg^2-mst2^2) (-mst1^2+mst2^2))) Log[mst2^2/MR^2]) Log[MR^2/mt^2]^2+736 Log[MR^2/mt^2]^3+((80 (mg^2-msq^2) (2 mg^10+6 mg^8 msq^2-8 mg^8 mst1^2-8 mg^6 msq^2 mst1^2+3 mg^6 mst1^4+9 mg^4 msq^2 mst1^4-mg^4 mst1^6-3 mg^2 msq^2 mst1^6-8 mg^8 mst2^2-8 mg^6 msq^2 mst2^2+30 mg^6 mst1^2 mst2^2-6 mg^4 msq^2 mst1^2 mst2^2-15 mg^4 mst1^4 mst2^2+3 mg^2 msq^2 mst1^4 mst2^2+5 mg^2 mst1^6 mst2^2-msq^2 mst1^6 mst2^2+3 mg^6 mst2^4+9 mg^4 msq^2 mst2^4-15 mg^4 mst1^2 mst2^4+3 mg^2 msq^2 mst1^2 mst2^4-mg^4 mst2^6-3 mg^2 msq^2 mst2^6+5 mg^2 mst1^2 mst2^6-msq^2 mst1^2 mst2^6))/((mg^2-mst1^2)^3 (mg^2-mst2^2)^3)+(640 (mg^2-msq^2) (-2 mg^3 msq^2+mg^3 mst1^2+mg msq^2 mst1^2+mg^3 mst2^2+mg msq^2 mst2^2-2 mg mst1^2 mst2^2) Xt)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)-(160 (mg^2-msq^2) (2 mg^10+6 mg^8 msq^2-8 mg^8 mst1^2-8 mg^6 msq^2 mst1^2+3 mg^6 mst1^4+9 mg^4 msq^2 mst1^4-mg^4 mst1^6-3 mg^2 msq^2 mst1^6-8 mg^8 mst2^2-8 mg^6 msq^2 mst2^2+30 mg^6 mst1^2 mst2^2-6 mg^4 msq^2 mst1^2 mst2^2-15 mg^4 mst1^4 mst2^2+3 mg^2 msq^2 mst1^4 mst2^2+5 mg^2 mst1^6 mst2^2-msq^2 mst1^6 mst2^2+3 mg^6 mst2^4+9 mg^4 msq^2 mst2^4-15 mg^4 mst1^2 mst2^4+3 mg^2 msq^2 mst1^2 mst2^4-mg^4 mst2^6-3 mg^2 msq^2 mst2^6+5 mg^2 mst1^2 mst2^6-msq^2 mst1^2 mst2^6) Xt^4)/((mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(1280 (mg^2-msq^2) (-2 mg^3 msq^2+mg^3 mst1^2+mg msq^2 mst1^2+mg^3 mst2^2+mg msq^2 mst2^2-2 mg mst1^2 mst2^2) Xt^5)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)) PolyLog[2,1-mg^2/msq^2]+(-((8 (128 mg^14-334 mg^12 mst1^2+368 mg^10 mst1^4-89 mg^8 mst1^6+11 mg^6 mst1^8-29 mg^4 mst1^10+9 mg^2 mst1^12-434 mg^12 mst2^2+1072 mg^10 mst1^2 mst2^2-1145 mg^8 mst1^4 mst2^2+100 mg^6 mst1^6 mst2^2+192 mg^4 mst1^8 mst2^2-44 mg^2 mst1^10 mst2^2+3 mst1^12 mst2^2+480 mg^10 mst2^4-1177 mg^8 mst1^2 mst2^4+1498 mg^6 mst1^4 mst2^4-376 mg^4 mst1^6 mst2^4-42 mg^2 mst1^8 mst2^4+mst1^10 mst2^4-149 mg^8 mst2^6+324 mg^6 mst1^2 mst2^6-598 mg^4 mst1^4 mst2^6+148 mg^2 mst1^6 mst2^6+19 mst1^8 mst2^6-13 mg^6 mst2^8+43 mg^4 mst1^2 mst2^8+57 mg^2 mst1^4 mst2^8-23 mst1^6 mst2^8))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)))+((256 mg^3 (-11 mg^4+3 mg^2 mst1^2+3 mg^2 mst2^2+5 mst1^2 mst2^2))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2) (mst1^2-mst2^2))+(64 (18 mg^5+mg^3 mst1^2+3 mg mst1^4-37 mg^3 mst2^2-7 mg mst1^2 mst2^2+22 mg mst2^4))/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt+(128 mg (-2 mg^2+mst1^2+mst2^2) (-11 mg^4+3 mg^2 mst1^2+3 mg^2 mst2^2+5 mst1^2 mst2^2) Xt^3)/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^3)+(16 (172 mg^12-372 mg^10 mst1^2+138 mg^8 mst1^4+3 mg^6 mst1^6+20 mg^4 mst1^8-9 mg^2 mst1^10-468 mg^10 mst2^2+1084 mg^8 mst1^2 mst2^2-313 mg^6 mst1^4 mst2^2-149 mg^4 mst1^6 mst2^2+41 mg^2 mst1^8 mst2^2-3 mst1^10 mst2^2+398 mg^8 mst2^4-1151 mg^6 mst1^2 mst2^4+411 mg^4 mst1^4 mst2^4+55 mg^2 mst1^6 mst2^4-mst1^8 mst2^4-59 mg^6 mst2^6+409 mg^4 mst1^2 mst2^6-129 mg^2 mst1^4 mst2^6-29 mst1^6 mst2^6-31 mg^4 mst2^8-30 mg^2 mst1^2 mst2^8+13 mst1^4 mst2^8) Xt^4)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)-(128 (18 mg^5+mg^3 mst1^2+3 mg mst1^4-37 mg^3 mst2^2-7 mg mst1^2 mst2^2+22 mg mst2^4) Xt^5)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)) PolyLog[2,1-mg^2/mst1^2]+(-((80 (-msq^2+mst1^2) (2 mg^4-3 mg^2 msq^2+3 mg^2 mst1^2-msq^2 mst1^2-mst1^4))/(mg^2-mst1^2)^3)+(640 mg (-msq^2+mst1^2)^2 Xt)/((mg^2-mst1^2)^2 (mst1^2-mst2^2))+(160 (-msq^2+mst1^2) (2 mg^4-3 mg^2 msq^2+3 mg^2 mst1^2-msq^2 mst1^2-mst1^4) Xt^4)/((mg^2-mst1^2)^3 (mst1^2-mst2^2)^2)-(1280 mg (-msq^2+mst1^2)^2 Xt^5)/((mg^2-mst1^2)^2 (mst1^2-mst2^2)^3)) PolyLog[2,1-msq^2/mst1^2]+((8 (128 mg^14-434 mg^12 mst1^2+480 mg^10 mst1^4-149 mg^8 mst1^6-13 mg^6 mst1^8-334 mg^12 mst2^2+1072 mg^10 mst1^2 mst2^2-1177 mg^8 mst1^4 mst2^2+324 mg^6 mst1^6 mst2^2+43 mg^4 mst1^8 mst2^2+368 mg^10 mst2^4-1145 mg^8 mst1^2 mst2^4+1498 mg^6 mst1^4 mst2^4-598 mg^4 mst1^6 mst2^4+57 mg^2 mst1^8 mst2^4-89 mg^8 mst2^6+100 mg^6 mst1^2 mst2^6-376 mg^4 mst1^4 mst2^6+148 mg^2 mst1^6 mst2^6-23 mst1^8 mst2^6+11 mg^6 mst2^8+192 mg^4 mst1^2 mst2^8-42 mg^2 mst1^4 mst2^8+19 mst1^6 mst2^8-29 mg^4 mst2^10-44 mg^2 mst1^2 mst2^10+mst1^4 mst2^10+9 mg^2 mst2^12+3 mst1^2 mst2^12))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+((256 mg^3 (-11 mg^4+3 mg^2 mst1^2+3 mg^2 mst2^2+5 mst1^2 mst2^2))/(3 (mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(64 (18 mg^5-37 mg^3 mst1^2+22 mg mst1^4+mg^3 mst2^2-7 mg mst1^2 mst2^2+3 mg mst2^4))/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))) Xt+(128 mg (2 mg^2-mst1^2-mst2^2) (-11 mg^4+3 mg^2 mst1^2+3 mg^2 mst2^2+5 mst1^2 mst2^2) Xt^3)/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^3)-(16 (172 mg^12-468 mg^10 mst1^2+398 mg^8 mst1^4-59 mg^6 mst1^6-31 mg^4 mst1^8-372 mg^10 mst2^2+1084 mg^8 mst1^2 mst2^2-1151 mg^6 mst1^4 mst2^2+409 mg^4 mst1^6 mst2^2-30 mg^2 mst1^8 mst2^2+138 mg^8 mst2^4-313 mg^6 mst1^2 mst2^4+411 mg^4 mst1^4 mst2^4-129 mg^2 mst1^6 mst2^4+13 mst1^8 mst2^4+3 mg^6 mst2^6-149 mg^4 mst1^2 mst2^6+55 mg^2 mst1^4 mst2^6-29 mst1^6 mst2^6+20 mg^4 mst2^8+41 mg^2 mst1^2 mst2^8-mst1^4 mst2^8-9 mg^2 mst2^10-3 mst1^2 mst2^10) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)-(128 (18 mg^5-37 mg^3 mst1^2+22 mg mst1^4+mg^3 mst2^2-7 mg mst1^2 mst2^2+3 mg mst2^4) Xt^5)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2)^3)) PolyLog[2,1-mg^2/mst2^2]+Log[msq^2/MR^2] ((80 (2 mg^10 mst1^2-4 mg^8 mst1^4+2 mg^6 mst1^6+2 mg^10 mst2^2-24 mg^8 mst1^2 mst2^2-18 mg^6 msq^2 mst1^2 mst2^2+41 mg^6 mst1^4 mst2^2+15 mg^4 msq^2 mst1^4 mst2^2-19 mg^4 mst1^6 mst2^2-9 mg^2 msq^2 mst1^6 mst2^2-4 mg^8 mst2^4+41 mg^6 mst1^2 mst2^4+15 mg^4 msq^2 mst1^2 mst2^4-68 mg^4 mst1^4 mst2^4+12 mg^2 msq^2 mst1^4 mst2^4+31 mg^2 mst1^6 mst2^4-3 msq^2 mst1^6 mst2^4+2 mg^6 mst2^6-19 mg^4 mst1^2 mst2^6-9 mg^2 msq^2 mst1^2 mst2^6+31 mg^2 mst1^4 mst2^6-3 msq^2 mst1^4 mst2^6-14 mst1^6 mst2^6))/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^2)+(640 mg msq^2 Xt)/((mg^2-mst1^2) (mg^2-mst2^2))-(320 mg^2 Xt^2)/(3 mst1^2 mst2^2)-(2560 mg Xt^3)/(3 (-mst1^2+mst2^2)^2)+(160 (mg^10 mst1^2-2 mg^8 mst1^4+mg^6 mst1^6+mg^10 mst2^2+14 mg^8 mst1^2 mst2^2+18 mg^6 msq^2 mst1^2 mst2^2-30 mg^6 mst1^4 mst2^2-15 mg^4 msq^2 mst1^4 mst2^2+15 mg^4 mst1^6 mst2^2+9 mg^2 msq^2 mst1^6 mst2^2-2 mg^8 mst2^4-30 mg^6 mst1^2 mst2^4-15 mg^4 msq^2 mst1^2 mst2^4+64 mg^4 mst1^4 mst2^4-12 mg^2 msq^2 mst1^4 mst2^4-32 mg^2 mst1^6 mst2^4+3 msq^2 mst1^6 mst2^4+mg^6 mst2^6+15 mg^4 mst1^2 mst2^6+9 mg^2 msq^2 mst1^2 mst2^6-32 mg^2 mst1^4 mst2^6+3 msq^2 mst1^4 mst2^6+16 mst1^6 mst2^6) Xt^4)/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (-mg^2+mst2^2)^2 (-mst1^2+mst2^2)^2)-(1280 mg msq^2 Xt^5)/((mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^2)+(-((320 mg^4)/(3 (mg^2-mst1^2)^2))+(1280 mg^3 Xt)/(3 (mg^2-mst1^2) (mst1^2-mst2^2))+(640 mg (-2 mg^2+mst1^2+mst2^2) Xt^3)/(3 (-mst1^2+mst2^2)^3)+(160 (-2 mg^2+mst1^2+mst2^2) Xt^4)/(3 (-mst1^2+mst2^2)^3)) PolyLog[2,1-mg^2/mst1^2]+(-((320 mg^4)/(3 (-mg^2+mst2^2)^2))+(1280 mg^3 Xt)/(3 (mg^2-mst2^2) (-mst1^2+mst2^2))+(640 mg (2 mg^2-mst1^2-mst2^2) Xt^3)/(3 (-mst1^2+mst2^2)^3)-(160 (-2 mg^2+mst1^2+mst2^2) Xt^4)/(3 (-mst1^2+mst2^2)^3)) PolyLog[2,1-mg^2/mst2^2])+Log[mg^2/MR^2] ((8 (132 mg^14 mst1^2-300 mg^12 mst1^4+236 mg^10 mst1^6-68 mg^8 mst1^8+132 mg^14 mst2^2-1336 mg^12 mst1^2 mst2^2+180 mg^10 msq^2 mst1^2 mst2^2+2738 mg^10 mst1^4 mst2^2-240 mg^8 msq^2 mst1^4 mst2^2-2016 mg^8 mst1^6 mst2^2+270 mg^6 msq^2 mst1^6 mst2^2+519 mg^6 mst1^8 mst2^2-90 mg^4 msq^2 mst1^8 mst2^2-9 mg^4 mst1^10 mst2^2-300 mg^12 mst2^4+2738 mg^10 mst1^2 mst2^4-240 mg^8 msq^2 mst1^2 mst2^4-5432 mg^8 mst1^4 mst2^4-180 mg^6 msq^2 mst1^4 mst2^4+3785 mg^6 mst1^6 mst2^4+90 mg^4 msq^2 mst1^6 mst2^4-860 mg^4 mst1^8 mst2^4-30 mg^2 msq^2 mst1^8 mst2^4-3 mg^2 mst1^10 mst2^4+236 mg^10 mst2^6-2016 mg^8 mst1^2 mst2^6+270 mg^6 msq^2 mst1^2 mst2^6+3785 mg^6 mst1^4 mst2^6+90 mg^4 msq^2 mst1^4 mst2^6-2470 mg^4 mst1^6 mst2^6+513 mg^2 mst1^8 mst2^6-68 mg^8 mst2^8+519 mg^6 mst1^2 mst2^8-90 mg^4 msq^2 mst1^2 mst2^8-860 mg^4 mst1^4 mst2^8-30 mg^2 msq^2 mst1^4 mst2^8+513 mg^2 mst1^6 mst2^8-96 mst1^8 mst2^8-9 mg^4 mst1^2 mst2^10-3 mg^2 mst1^4 mst2^10))/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^3)-(64 (16 mg^9 mst1^2-16 mg^7 mst1^4+16 mg^9 mst2^2-128 mg^7 mst1^2 mst2^2+60 mg^5 msq^2 mst1^2 mst2^2+112 mg^5 mst1^4 mst2^2-30 mg^3 msq^2 mst1^4 mst2^2-3 mg^3 mst1^6 mst2^2-16 mg^7 mst2^4+112 mg^5 mst1^2 mst2^4-30 mg^3 msq^2 mst1^2 mst2^4-90 mg^3 mst1^4 mst2^4-3 mg^3 mst1^2 mst2^6) Xt)/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^2)-(64 (33 mg^10-42 mg^8 mst1^2+17 mg^6 mst1^4-42 mg^8 mst2^2+25 mg^6 mst1^2 mst2^2-7 mg^4 mst1^4 mst2^2+17 mg^6 mst2^4-7 mg^4 mst1^2 mst2^4+6 mg^2 mst1^4 mst2^4) Xt^2)/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^2)+((2048 mg^5)/(3 mst1^2 (mg^2-mst1^2) mst2^2 (mg^2-mst2^2))-(512 mg (2+(8 (2 mg^8-2 mg^6 mst1^2+mg^4 mst1^4-2 mg^6 mst2^2+mg^4 mst2^4))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)))/(-mst1^2+mst2^2)^2) Xt^3+(16 (66 mg^14 mst1^2-150 mg^12 mst1^4+118 mg^10 mst1^6-34 mg^8 mst1^8+66 mg^14 mst2^2+1876 mg^12 mst1^2 mst2^2-180 mg^10 msq^2 mst1^2 mst2^2-4190 mg^10 mst1^4 mst2^2+240 mg^8 msq^2 mst1^4 mst2^2+2756 mg^8 mst1^6 mst2^2-270 mg^6 msq^2 mst1^6 mst2^2-465 mg^6 mst1^8 mst2^2+90 mg^4 msq^2 mst1^8 mst2^2+9 mg^4 mst1^10 mst2^2-150 mg^12 mst2^4-4190 mg^10 mst1^2 mst2^4+240 mg^8 msq^2 mst1^2 mst2^4+8964 mg^8 mst1^4 mst2^4+180 mg^6 msq^2 mst1^4 mst2^4-5645 mg^6 mst1^6 mst2^4-90 mg^4 msq^2 mst1^6 mst2^4+850 mg^4 mst1^8 mst2^4+30 mg^2 msq^2 mst1^8 mst2^4+3 mg^2 mst1^10 mst2^4+118 mg^10 mst2^6+2756 mg^8 mst1^2 mst2^6-270 mg^6 msq^2 mst1^2 mst2^6-5645 mg^6 mst1^4 mst2^6-90 mg^4 msq^2 mst1^4 mst2^6+3522 mg^4 mst1^6 mst2^6-559 mg^2 mst1^8 mst2^6-34 mg^8 mst2^8-465 mg^6 mst1^2 mst2^8+90 mg^4 msq^2 mst1^2 mst2^8+850 mg^4 mst1^4 mst2^8+30 mg^2 msq^2 mst1^4 mst2^8-559 mg^2 mst1^6 mst2^8+120 mst1^8 mst2^8+9 mg^4 mst1^2 mst2^10+3 mg^2 mst1^4 mst2^10) Xt^4)/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(128 (8 mg^9 mst1^2-8 mg^7 mst1^4+8 mg^9 mst2^2+123 mg^7 mst1^2 mst2^2-60 mg^5 msq^2 mst1^2 mst2^2-115 mg^5 mst1^4 mst2^2+30 mg^3 msq^2 mst1^4 mst2^2+3 mg^3 mst1^6 mst2^2-8 mg^7 mst2^4-115 mg^5 mst1^2 mst2^4+30 mg^3 msq^2 mst1^2 mst2^4+101 mg^3 mst1^4 mst2^4+3 mg^3 mst1^2 mst2^6) Xt^5)/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)+(-((160 (mg^14 mst1^2-3 mg^12 mst1^4+3 mg^10 mst1^6-mg^8 mst1^8+mg^14 mst2^2+2 mg^12 mst1^2 mst2^2-11 mg^10 mst1^4 mst2^2+3 mg^8 mst1^6 mst2^2-mg^6 mst1^8 mst2^2-3 mg^12 mst2^4-11 mg^10 mst1^2 mst2^4+48 mg^8 mst1^4 mst2^4-24 mg^6 mst1^6 mst2^4+8 mg^4 mst1^8 mst2^4+3 mg^10 mst2^6+3 mg^8 mst1^2 mst2^6-24 mg^6 mst1^4 mst2^6-mg^8 mst2^8-mg^6 mst1^2 mst2^8+8 mg^4 mst1^4 mst2^8))/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^3))-(640 (mg^5 mst1^2+mg^5 mst2^2-2 mg^3 mst1^2 mst2^2) Xt)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)+(320 mg^4 (-mg^2+mst1^2+mst2^2) Xt^2)/(3 mst1^2 (mg^2-mst1^2) mst2^2 (-mg^2+mst2^2))-(160 (mg^14 mst1^2-3 mg^12 mst1^4+3 mg^10 mst1^6-mg^8 mst1^8+mg^14 mst2^2-10 mg^12 mst1^2 mst2^2+32 mg^10 mst1^4 mst2^2-16 mg^8 mst1^6 mst2^2+5 mg^6 mst1^8 mst2^2-3 mg^12 mst2^4+32 mg^10 mst1^2 mst2^4-102 mg^8 mst1^4 mst2^4+54 mg^6 mst1^6 mst2^4-17 mg^4 mst1^8 mst2^4+3 mg^10 mst2^6-16 mg^8 mst1^2 mst2^6+54 mg^6 mst1^4 mst2^6-6 mg^4 mst1^6 mst2^6+mg^2 mst1^8 mst2^6-mg^8 mst2^8+5 mg^6 mst1^2 mst2^8-17 mg^4 mst1^4 mst2^8+mg^2 mst1^6 mst2^8) Xt^4)/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)+(640 (mg^7+5 mg^5 mst1^2+5 mg^5 mst2^2-11 mg^3 mst1^2 mst2^2) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)) Log[msq^2/MR^2]+(-((8 (-2 mg^16 mst1^4+72 mg^14 mst1^6-172 mg^12 mst1^8+136 mg^10 mst1^10-34 mg^8 mst1^12-128 mg^16 mst1^2 mst2^2+1462 mg^14 mst1^4 mst2^2-180 mg^12 msq^2 mst1^4 mst2^2-3332 mg^12 mst1^6 mst2^2+420 mg^10 msq^2 mst1^6 mst2^2+3774 mg^10 mst1^8 mst2^2-510 mg^8 msq^2 mst1^8 mst2^2-2057 mg^8 mst1^10 mst2^2+360 mg^6 msq^2 mst1^10 mst2^2+426 mg^6 mst1^12 mst2^2-90 mg^4 msq^2 mst1^12 mst2^2-9 mg^4 mst1^14 mst2^2+2 mg^16 mst2^4-760 mg^14 mst1^2 mst2^4+180 mg^12 msq^2 mst1^2 mst2^4-610 mg^12 mst1^4 mst2^4-180 mg^10 msq^2 mst1^4 mst2^4+3634 mg^10 mst1^6 mst2^4+450 mg^8 msq^2 mst1^6 mst2^4-5506 mg^8 mst1^8 mst2^4-630 mg^6 msq^2 mst1^8 mst2^4+3339 mg^6 mst1^10 mst2^4+210 mg^4 msq^2 mst1^10 mst2^4-640 mg^4 mst1^12 mst2^4-30 mg^2 msq^2 mst1^12 mst2^4-3 mg^2 mst1^14 mst2^4-6 mg^14 mst2^6+2188 mg^12 mst1^2 mst2^6-240 mg^10 msq^2 mst1^2 mst2^6-3080 mg^10 mst1^4 mst2^6-210 mg^8 msq^2 mst1^4 mst2^6+1602 mg^8 mst1^6 mst2^6+450 mg^6 msq^2 mst1^6 mst2^6+1682 mg^6 mst1^8 mst2^6-30 mg^4 msq^2 mst1^8 mst2^6-2026 mg^4 mst1^10 mst2^6+30 mg^2 msq^2 mst1^10 mst2^6+456 mg^2 mst1^12 mst2^6+6 mg^12 mst2^8-1902 mg^10 mst1^2 mst2^8+270 mg^8 msq^2 mst1^2 mst2^8+3528 mg^8 mst1^4 mst2^8-90 mg^6 msq^2 mst1^4 mst2^8-3698 mg^6 mst1^6 mst2^8-150 mg^4 msq^2 mst1^6 mst2^8+1522 mg^4 mst1^8 mst2^8-30 mg^2 msq^2 mst1^8 mst2^8+84 mg^2 mst1^10 mst2^8-84 mst1^12 mst2^8-2 mg^10 mst2^10+547 mg^8 mst1^2 mst2^10-90 mg^6 msq^2 mst1^2 mst2^10-972 mg^6 mst1^4 mst2^10+60 mg^4 msq^2 mst1^4 mst2^10+1019 mg^4 mst1^6 mst2^10+30 mg^2 msq^2 mst1^6 mst2^10-540 mg^2 mst1^8 mst2^10+84 mst1^10 mst2^10-9 mg^6 mst1^2 mst2^12+6 mg^4 mst1^4 mst2^12+3 mg^2 mst1^6 mst2^12))/(3 mst1^2 (mg^2-mst1^2)^4 mst2^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)))+(64 (16 mg^11 mst1^2-32 mg^9 mst1^4+16 mg^7 mst1^6+4 mg^11 mst2^2+216 mg^9 mst1^2 mst2^2-60 mg^7 msq^2 mst1^2 mst2^2-223 mg^7 mst1^4 mst2^2+90 mg^5 msq^2 mst1^4 mst2^2+86 mg^5 mst1^6 mst2^2-30 mg^3 msq^2 mst1^6 mst2^2-3 mg^3 mst1^8 mst2^2-130 mg^9 mst2^4+60 mg^7 msq^2 mst2^4-106 mg^7 mst1^2 mst2^4-60 mg^5 msq^2 mst1^2 mst2^4+111 mg^5 mst1^4 mst2^4-35 mg^3 mst1^6 mst2^4+123 mg^7 mst2^6-30 mg^5 msq^2 mst2^6-68 mg^5 mst1^2 mst2^6+30 mg^3 msq^2 mst1^2 mst2^6+37 mg^3 mst1^4 mst2^6-12 mg mst1^6 mst2^6-3 mg^5 mst2^8+3 mg^3 mst1^2 mst2^8) Xt)/(3 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(16 (-2 mg^14 mst1^2+70 mg^12 mst1^4-102 mg^10 mst1^6+34 mg^8 mst1^8+2 mg^14 mst2^2-1472 mg^12 mst1^2 mst2^2+180 mg^10 msq^2 mst1^2 mst2^2+2660 mg^10 mst1^4 mst2^2-240 mg^8 msq^2 mst1^4 mst2^2-1692 mg^8 mst1^6 mst2^2+270 mg^6 msq^2 mst1^6 mst2^2+395 mg^6 mst1^8 mst2^2-90 mg^4 msq^2 mst1^8 mst2^2-9 mg^4 mst1^10 mst2^2-6 mg^12 mst2^4+3280 mg^10 mst1^2 mst2^4-240 mg^8 msq^2 mst1^2 mst2^4-5696 mg^8 mst1^4 mst2^4-180 mg^6 msq^2 mst1^4 mst2^4+3519 mg^6 mst1^6 mst2^4+90 mg^4 msq^2 mst1^6 mst2^4-734 mg^4 mst1^8 mst2^4-30 mg^2 msq^2 mst1^8 mst2^4-3 mg^2 mst1^10 mst2^4+6 mg^10 mst2^6-2436 mg^8 mst1^2 mst2^6+270 mg^6 msq^2 mst1^2 mst2^6+3987 mg^6 mst1^4 mst2^6+90 mg^4 msq^2 mst1^4 mst2^6-2430 mg^4 mst1^6 mst2^6+489 mg^2 mst1^8 mst2^6-2 mg^8 mst2^8+587 mg^6 mst1^2 mst2^8-90 mg^4 msq^2 mst1^2 mst2^8-850 mg^4 mst1^4 mst2^8-30 mg^2 msq^2 mst1^4 mst2^8+513 mg^2 mst1^6 mst2^8-96 mst1^8 mst2^8-9 mg^4 mst1^2 mst2^10-3 mg^2 mst1^4 mst2^10) Xt^2)/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))-(128 (16 mg^9 mst1^4-16 mg^7 mst1^6+22 mg^11 mst2^2+6 mg^9 mst1^2 mst2^2-217 mg^7 mst1^4 mst2^2+60 mg^5 msq^2 mst1^4 mst2^2+165 mg^5 mst1^6 mst2^2-30 mg^3 msq^2 mst1^6 mst2^2-3 mg^3 mst1^8 mst2^2+10 mg^9 mst2^4+238 mg^7 mst1^2 mst2^4-120 mg^5 msq^2 mst1^2 mst2^4-17 mg^5 mst1^4 mst2^4+30 mg^3 msq^2 mst1^4 mst2^4-115 mg^3 mst1^6 mst2^4-193 mg^7 mst2^6+60 mg^5 msq^2 mst2^6-85 mg^5 mst1^2 mst2^6+30 mg^3 msq^2 mst1^2 mst2^6+106 mg^3 mst1^4 mst2^6+18 mg mst1^6 mst2^6+153 mg^5 mst2^8-30 mg^3 msq^2 mst2^8-91 mg^3 mst1^2 mst2^8+6 mg mst1^4 mst2^8-3 mg^3 mst2^10) Xt^3)/(3 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)-(8 (-2 mg^16 mst1^4+72 mg^14 mst1^6-172 mg^12 mst1^8+136 mg^10 mst1^10-34 mg^8 mst1^12+80 mg^16 mst1^2 mst2^2-3178 mg^14 mst1^4 mst2^2+180 mg^12 msq^2 mst1^4 mst2^2+9110 mg^12 mst1^6 mst2^2-420 mg^10 msq^2 mst1^6 mst2^2-9926 mg^10 mst1^8 mst2^2+510 mg^8 msq^2 mst1^8 mst2^2+4623 mg^8 mst1^10 mst2^2-360 mg^6 msq^2 mst1^10 mst2^2-702 mg^6 mst1^12 mst2^2+90 mg^4 msq^2 mst1^12 mst2^2+9 mg^4 mst1^14 mst2^2+2 mg^16 mst2^4-2152 mg^14 mst1^2 mst2^4+180 mg^12 msq^2 mst1^2 mst2^4+12962 mg^12 mst1^4 mst2^4-660 mg^10 msq^2 mst1^4 mst2^4-25584 mg^10 mst1^6 mst2^4+570 mg^8 msq^2 mst1^6 mst2^4+22990 mg^8 mst1^8 mst2^4-90 mg^6 msq^2 mst1^8 mst2^4-9553 mg^6 mst1^10 mst2^4-30 mg^4 msq^2 mst1^10 mst2^4+1268 mg^4 mst1^12 mst2^4+30 mg^2 msq^2 mst1^12 mst2^4+3 mg^2 mst1^14 mst2^4-6 mg^14 mst2^6+4406 mg^12 mst1^2 mst2^6-240 mg^10 msq^2 mst1^2 mst2^6-17588 mg^10 mst1^4 mst2^6+330 mg^8 msq^2 mst1^4 mst2^6+25638 mg^8 mst1^6 mst2^6+90 mg^6 msq^2 mst1^6 mst2^6-18070 mg^6 mst1^8 mst2^6-210 mg^4 msq^2 mst1^8 mst2^6+6600 mg^4 mst1^10 mst2^6+30 mg^2 msq^2 mst1^10 mst2^6-884 mg^2 mst1^12 mst2^6+6 mg^12 mst2^8-2996 mg^10 mst1^2 mst2^8+270 mg^8 msq^2 mst1^2 mst2^8+9676 mg^8 mst1^4 mst2^8-270 mg^6 msq^2 mst1^4 mst2^8-10862 mg^6 mst1^6 mst2^8-30 mg^4 msq^2 mst1^6 mst2^8+5474 mg^4 mst1^8 mst2^8+30 mg^2 msq^2 mst1^8 mst2^8-1602 mg^2 mst1^10 mst2^8+240 mst1^12 mst2^8-2 mg^10 mst2^10+643 mg^8 mst1^2 mst2^10-90 mg^6 msq^2 mst1^2 mst2^10-1828 mg^6 mst1^4 mst2^10+60 mg^4 msq^2 mst1^4 mst2^10+1723 mg^4 mst1^6 mst2^10+30 mg^2 msq^2 mst1^6 mst2^10-616 mg^2 mst1^8 mst2^10+96 mst1^10 mst2^10-9 mg^6 mst1^2 mst2^12+6 mg^4 mst1^4 mst2^12+3 mg^2 mst1^6 mst2^12) Xt^4)/(3 mst1^2 (mg^2-mst1^2)^4 mst2^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)+(64 (16 mg^11 mst1^2-32 mg^9 mst1^4+16 mg^7 mst1^6+52 mg^11 mst2^2-362 mg^9 mst1^2 mst2^2+60 mg^7 msq^2 mst1^2 mst2^2+558 mg^7 mst1^4 mst2^2-90 mg^5 msq^2 mst1^4 mst2^2-219 mg^5 mst1^6 mst2^2+30 mg^3 msq^2 mst1^6 mst2^2+3 mg^3 mst1^8 mst2^2-276 mg^9 mst2^4+60 mg^7 msq^2 mst2^4+770 mg^7 mst1^2 mst2^4-120 mg^5 msq^2 mst1^2 mst2^4-737 mg^5 mst1^4 mst2^4+60 mg^3 msq^2 mst1^4 mst2^4+179 mg^3 mst1^6 mst2^4+214 mg^7 mst2^6-30 mg^5 msq^2 mst2^6-407 mg^5 mst1^2 mst2^6+30 mg^3 msq^2 mst1^2 mst2^6+213 mg^3 mst1^4 mst2^6+12 mg mst1^6 mst2^6-3 mg^5 mst2^8+3 mg^3 mst1^2 mst2^8) Xt^5)/(3 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)-(5120 mg^4 mst1^2 Xt^6)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2) (-mst1^2+mst2^2)^3)+((160 (3 mg^12-11 mg^10 mst1^2+3 mg^8 mst1^4-mg^6 mst1^6-13 mg^10 mst2^2+45 mg^8 mst1^2 mst2^2-21 mg^6 mst1^4 mst2^2+7 mg^4 mst1^6 mst2^2+6 mg^8 mst2^4-24 mg^6 mst1^2 mst2^4-2 mg^6 mst2^6+8 mg^4 mst1^2 mst2^6))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3)+((320 mg^3 (2 mg^2-mst1^2-mst2^2))/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2))+(640 (mg^5 mst1^2+mg^5 mst2^2-2 mg^3 mst1^2 mst2^2))/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)) Xt-(320 (2 mg^12-10 mg^10 mst1^2+3 mg^8 mst1^4-mg^6 mst1^6-10 mg^10 mst2^2+42 mg^8 mst1^2 mst2^2-21 mg^6 mst1^4 mst2^2+7 mg^4 mst1^6 mst2^2+3 mg^8 mst2^4-21 mg^6 mst1^2 mst2^4-mg^6 mst2^6+7 mg^4 mst1^2 mst2^6) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+(-((1280 (mg^5 mst1^2+mg^5 mst2^2-2 mg^3 mst1^2 mst2^2))/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)))+(640 mg^3 (2 mg^4-mst1^4+4 mst1^2 mst2^2-mst2^4-2 mg^2 (mst1^2+mst2^2)))/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^3)) Xt^3+(160 (mst1^2+mst2^2) (2 mg^12-10 mg^10 mst1^2+3 mg^8 mst1^4-mg^6 mst1^6-10 mg^10 mst2^2+42 mg^8 mst1^2 mst2^2-21 mg^6 mst1^4 mst2^2+7 mg^4 mst1^6 mst2^2+3 mg^8 mst2^4-21 mg^6 mst1^2 mst2^4-mg^6 mst2^6+7 mg^4 mst1^2 mst2^6) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)+(320 (mst1^2+mst2^2) (mg^7+5 mg^5 mst1^2+5 mg^5 mst2^2-11 mg^3 mst1^2 mst2^2) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)) Log[msq^2/MR^2]) Log[mst1^2/MR^2]+(-((16 (64 mg^16-245 mg^14 mst1^2+240 mg^12 mst1^4-170 mg^10 mst1^6+110 mg^8 mst1^8-27 mg^6 mst1^10-139 mg^14 mst2^2+633 mg^12 mst1^2 mst2^2-432 mg^10 mst1^4 mst2^2+162 mg^8 mst1^6 mst2^2-157 mg^6 mst1^8 mst2^2+45 mg^4 mst1^10 mst2^2+87 mg^12 mst2^4-700 mg^10 mst1^2 mst2^4+400 mg^8 mst1^4 mst2^4-20 mg^6 mst1^6 mst2^4+101 mg^4 mst1^8 mst2^4-36 mg^2 mst1^10 mst2^4+22 mg^10 mst2^6+316 mg^8 mst1^2 mst2^6-146 mg^6 mst1^4 mst2^6-80 mg^4 mst1^6 mst2^6-12 mg^2 mst1^8 mst2^6+12 mst1^10 mst2^6-28 mg^8 mst2^8-34 mg^6 mst1^2 mst2^8-2 mg^4 mst1^4 mst2^8+48 mg^2 mst1^6 mst2^8-12 mst1^8 mst2^8))/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)))-(32 (18 mg^11+195 mg^9 mst1^2-130 mg^7 mst1^4+13 mg^5 mst1^6-19 mg^9 mst2^2-425 mg^7 mst1^2 mst2^2+263 mg^5 mst1^4 mst2^2-11 mg^3 mst1^6 mst2^2+27 mg^7 mst2^4+168 mg^5 mst1^2 mst2^4-87 mg^3 mst1^4 mst2^4-12 mg mst1^6 mst2^4) Xt)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(32 (64 mg^16-502 mg^14 mst1^2+954 mg^12 mst1^4-759 mg^10 mst1^6+302 mg^8 mst1^8-55 mg^6 mst1^10-54 mg^14 mst2^2+962 mg^12 mst1^2 mst2^2-1897 mg^10 mst1^4 mst2^2+1401 mg^8 mst1^6 mst2^2-525 mg^6 mst1^8 mst2^2+97 mg^4 mst1^10 mst2^2-72 mg^12 mst2^4-545 mg^10 mst1^2 mst2^4+1199 mg^8 mst1^4 mst2^4-811 mg^6 mst1^6 mst2^4+325 mg^4 mst1^8 mst2^4-72 mg^2 mst1^10 mst2^4+69 mg^10 mst2^6+79 mg^8 mst1^2 mst2^6-189 mg^6 mst1^4 mst2^6+61 mg^4 mst1^6 mst2^6-60 mg^2 mst1^8 mst2^6+24 mst1^10 mst2^6-mg^8 mst2^8-24 mg^6 mst1^2 mst2^8-7 mg^4 mst1^4 mst2^8+48 mg^2 mst1^6 mst2^8-12 mst1^8 mst2^8) Xt^2)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(64 (42 mg^13-120 mg^11 mst1^2+239 mg^9 mst1^4-252 mg^7 mst1^6+59 mg^5 mst1^8-6 mg^11 mst2^2-44 mg^9 mst1^2 mst2^2-67 mg^7 mst1^4 mst2^2+268 mg^5 mst1^6 mst2^2-55 mg^3 mst1^8 mst2^2-67 mg^9 mst2^4+214 mg^7 mst1^2 mst2^4-195 mg^5 mst1^4 mst2^4-42 mg^3 mst1^6 mst2^4-6 mg mst1^8 mst2^4+45 mg^7 mst2^6-90 mg^5 mst1^2 mst2^6+59 mg^3 mst1^4 mst2^6+18 mg mst1^6 mst2^6) Xt^3)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)+(16 (-788 mg^16 mst1^2+2556 mg^14 mst1^4-2618 mg^12 mst1^6+835 mg^10 mst1^8-58 mg^8 mst1^10+69 mg^6 mst1^12+148 mg^16 mst2^2+1796 mg^14 mst1^2 mst2^2-7366 mg^12 mst1^4 mst2^2+8472 mg^10 mst1^6 mst2^2-3031 mg^8 mst1^8 mst2^2+116 mg^6 mst1^10 mst2^2-123 mg^4 mst1^12 mst2^2-336 mg^14 mst2^4-786 mg^12 mst1^2 mst2^4+6232 mg^10 mst1^4 mst2^4-8712 mg^8 mst1^6 mst2^4+3742 mg^6 mst1^8 mst2^4-238 mg^4 mst1^10 mst2^4+90 mg^2 mst1^12 mst2^4+194 mg^12 mst2^6-452 mg^10 mst1^2 mst2^6-936 mg^8 mst1^4 mst2^6+2648 mg^6 mst1^6 mst2^6-1612 mg^4 mst1^8 mst2^6+180 mg^2 mst1^10 mst2^6-30 mst1^12 mst2^6+mg^10 mst2^8+190 mg^8 mst1^2 mst2^8-479 mg^6 mst1^4 mst2^8+234 mg^4 mst1^6 mst2^8+90 mg^2 mst1^8 mst2^8-24 mst1^10 mst2^8-13 mg^8 mst2^10+64 mg^6 mst1^2 mst2^10-37 mg^4 mst1^4 mst2^10-24 mg^2 mst1^6 mst2^10+6 mst1^8 mst2^10) Xt^4)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^4)+(32 (36 mg^11 mst1^2-159 mg^9 mst1^4+216 mg^7 mst1^6-61 mg^5 mst1^8+100 mg^11 mst2^2-392 mg^9 mst1^2 mst2^2+585 mg^7 mst1^4 mst2^2-384 mg^5 mst1^6 mst2^2+59 mg^3 mst1^8 mst2^2-169 mg^9 mst2^4+508 mg^7 mst1^2 mst2^4-535 mg^5 mst1^4 mst2^4+152 mg^3 mst1^6 mst2^4+12 mg mst1^8 mst2^4+75 mg^7 mst2^6-148 mg^5 mst1^2 mst2^6+93 mg^3 mst1^4 mst2^6+12 mg mst1^6 mst2^6) Xt^5)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)-(2560 mg^4 mst1^2 (mst1^2+mst2^2) Xt^6)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2) (-mst1^2+mst2^2)^4)) Log[mst1^2/MR^2]^2+(-((8 (-2 mg^16 mst1^4+6 mg^14 mst1^6-6 mg^12 mst1^8+2 mg^10 mst1^10+128 mg^16 mst1^2 mst2^2+760 mg^14 mst1^4 mst2^2-180 mg^12 msq^2 mst1^4 mst2^2-2188 mg^12 mst1^6 mst2^2+240 mg^10 msq^2 mst1^6 mst2^2+1902 mg^10 mst1^8 mst2^2-270 mg^8 msq^2 mst1^8 mst2^2-547 mg^8 mst1^10 mst2^2+90 mg^6 msq^2 mst1^10 mst2^2+9 mg^6 mst1^12 mst2^2+2 mg^16 mst2^4-1462 mg^14 mst1^2 mst2^4+180 mg^12 msq^2 mst1^2 mst2^4+610 mg^12 mst1^4 mst2^4+180 mg^10 msq^2 mst1^4 mst2^4+3080 mg^10 mst1^6 mst2^4+210 mg^8 msq^2 mst1^6 mst2^4-3528 mg^8 mst1^8 mst2^4+90 mg^6 msq^2 mst1^8 mst2^4+972 mg^6 mst1^10 mst2^4-60 mg^4 msq^2 mst1^10 mst2^4-6 mg^4 mst1^12 mst2^4-72 mg^14 mst2^6+3332 mg^12 mst1^2 mst2^6-420 mg^10 msq^2 mst1^2 mst2^6-3634 mg^10 mst1^4 mst2^6-450 mg^8 msq^2 mst1^4 mst2^6-1602 mg^8 mst1^6 mst2^6-450 mg^6 msq^2 mst1^6 mst2^6+3698 mg^6 mst1^8 mst2^6+150 mg^4 msq^2 mst1^8 mst2^6-1019 mg^4 mst1^10 mst2^6-30 mg^2 msq^2 mst1^10 mst2^6-3 mg^2 mst1^12 mst2^6+172 mg^12 mst2^8-3774 mg^10 mst1^2 mst2^8+510 mg^8 msq^2 mst1^2 mst2^8+5506 mg^8 mst1^4 mst2^8+630 mg^6 msq^2 mst1^4 mst2^8-1682 mg^6 mst1^6 mst2^8+30 mg^4 msq^2 mst1^6 mst2^8-1522 mg^4 mst1^8 mst2^8+30 mg^2 msq^2 mst1^8 mst2^8+540 mg^2 mst1^10 mst2^8-136 mg^10 mst2^10+2057 mg^8 mst1^2 mst2^10-360 mg^6 msq^2 mst1^2 mst2^10-3339 mg^6 mst1^4 mst2^10-210 mg^4 msq^2 mst1^4 mst2^10+2026 mg^4 mst1^6 mst2^10-30 mg^2 msq^2 mst1^6 mst2^10-84 mg^2 mst1^8 mst2^10-84 mst1^10 mst2^10+34 mg^8 mst2^12-426 mg^6 mst1^2 mst2^12+90 mg^4 msq^2 mst1^2 mst2^12+640 mg^4 mst1^4 mst2^12+30 mg^2 msq^2 mst1^4 mst2^12-456 mg^2 mst1^6 mst2^12+84 mst1^8 mst2^12+9 mg^4 mst1^2 mst2^14+3 mg^2 mst1^4 mst2^14))/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)))-(64 (4 mg^11 mst1^2-130 mg^9 mst1^4+60 mg^7 msq^2 mst1^4+123 mg^7 mst1^6-30 mg^5 msq^2 mst1^6-3 mg^5 mst1^8+16 mg^11 mst2^2+216 mg^9 mst1^2 mst2^2-60 mg^7 msq^2 mst1^2 mst2^2-106 mg^7 mst1^4 mst2^2-60 mg^5 msq^2 mst1^4 mst2^2-68 mg^5 mst1^6 mst2^2+30 mg^3 msq^2 mst1^6 mst2^2+3 mg^3 mst1^8 mst2^2-32 mg^9 mst2^4-223 mg^7 mst1^2 mst2^4+90 mg^5 msq^2 mst1^2 mst2^4+111 mg^5 mst1^4 mst2^4+37 mg^3 mst1^6 mst2^4+16 mg^7 mst2^6+86 mg^5 mst1^2 mst2^6-30 mg^3 msq^2 mst1^2 mst2^6-35 mg^3 mst1^4 mst2^6-12 mg mst1^6 mst2^6-3 mg^3 mst1^2 mst2^8) Xt)/(3 mst1^2 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+(16 (-2 mg^14 mst1^2+6 mg^12 mst1^4-6 mg^10 mst1^6+2 mg^8 mst1^8+2 mg^14 mst2^2+1472 mg^12 mst1^2 mst2^2-180 mg^10 msq^2 mst1^2 mst2^2-3280 mg^10 mst1^4 mst2^2+240 mg^8 msq^2 mst1^4 mst2^2+2436 mg^8 mst1^6 mst2^2-270 mg^6 msq^2 mst1^6 mst2^2-587 mg^6 mst1^8 mst2^2+90 mg^4 msq^2 mst1^8 mst2^2+9 mg^4 mst1^10 mst2^2-70 mg^12 mst2^4-2660 mg^10 mst1^2 mst2^4+240 mg^8 msq^2 mst1^2 mst2^4+5696 mg^8 mst1^4 mst2^4+180 mg^6 msq^2 mst1^4 mst2^4-3987 mg^6 mst1^6 mst2^4-90 mg^4 msq^2 mst1^6 mst2^4+850 mg^4 mst1^8 mst2^4+30 mg^2 msq^2 mst1^8 mst2^4+3 mg^2 mst1^10 mst2^4+102 mg^10 mst2^6+1692 mg^8 mst1^2 mst2^6-270 mg^6 msq^2 mst1^2 mst2^6-3519 mg^6 mst1^4 mst2^6-90 mg^4 msq^2 mst1^4 mst2^6+2430 mg^4 mst1^6 mst2^6-513 mg^2 mst1^8 mst2^6-34 mg^8 mst2^8-395 mg^6 mst1^2 mst2^8+90 mg^4 msq^2 mst1^2 mst2^8+734 mg^4 mst1^4 mst2^8+30 mg^2 msq^2 mst1^4 mst2^8-489 mg^2 mst1^6 mst2^8+96 mst1^8 mst2^8+9 mg^4 mst1^2 mst2^10+3 mg^2 mst1^4 mst2^10) Xt^2)/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+(128 (22 mg^11 mst1^2+10 mg^9 mst1^4-193 mg^7 mst1^6+60 mg^5 msq^2 mst1^6+153 mg^5 mst1^8-30 mg^3 msq^2 mst1^8-3 mg^3 mst1^10+6 mg^9 mst1^2 mst2^2+238 mg^7 mst1^4 mst2^2-120 mg^5 msq^2 mst1^4 mst2^2-85 mg^5 mst1^6 mst2^2+30 mg^3 msq^2 mst1^6 mst2^2-91 mg^3 mst1^8 mst2^2+16 mg^9 mst2^4-217 mg^7 mst1^2 mst2^4+60 mg^5 msq^2 mst1^2 mst2^4-17 mg^5 mst1^4 mst2^4+30 mg^3 msq^2 mst1^4 mst2^4+106 mg^3 mst1^6 mst2^4+6 mg mst1^8 mst2^4-16 mg^7 mst2^6+165 mg^5 mst1^2 mst2^6-30 mg^3 msq^2 mst1^2 mst2^6-115 mg^3 mst1^4 mst2^6+18 mg mst1^6 mst2^6-3 mg^3 mst1^2 mst2^8) Xt^3)/(3 mst1^2 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)-(8 (-2 mg^16 mst1^4+6 mg^14 mst1^6-6 mg^12 mst1^8+2 mg^10 mst1^10-80 mg^16 mst1^2 mst2^2+2152 mg^14 mst1^4 mst2^2-180 mg^12 msq^2 mst1^4 mst2^2-4406 mg^12 mst1^6 mst2^2+240 mg^10 msq^2 mst1^6 mst2^2+2996 mg^10 mst1^8 mst2^2-270 mg^8 msq^2 mst1^8 mst2^2-643 mg^8 mst1^10 mst2^2+90 mg^6 msq^2 mst1^10 mst2^2+9 mg^6 mst1^12 mst2^2+2 mg^16 mst2^4+3178 mg^14 mst1^2 mst2^4-180 mg^12 msq^2 mst1^2 mst2^4-12962 mg^12 mst1^4 mst2^4+660 mg^10 msq^2 mst1^4 mst2^4+17588 mg^10 mst1^6 mst2^4-330 mg^8 msq^2 mst1^6 mst2^4-9676 mg^8 mst1^8 mst2^4+270 mg^6 msq^2 mst1^8 mst2^4+1828 mg^6 mst1^10 mst2^4-60 mg^4 msq^2 mst1^10 mst2^4-6 mg^4 mst1^12 mst2^4-72 mg^14 mst2^6-9110 mg^12 mst1^2 mst2^6+420 mg^10 msq^2 mst1^2 mst2^6+25584 mg^10 mst1^4 mst2^6-570 mg^8 msq^2 mst1^4 mst2^6-25638 mg^8 mst1^6 mst2^6-90 mg^6 msq^2 mst1^6 mst2^6+10862 mg^6 mst1^8 mst2^6+30 mg^4 msq^2 mst1^8 mst2^6-1723 mg^4 mst1^10 mst2^6-30 mg^2 msq^2 mst1^10 mst2^6-3 mg^2 mst1^12 mst2^6+172 mg^12 mst2^8+9926 mg^10 mst1^2 mst2^8-510 mg^8 msq^2 mst1^2 mst2^8-22990 mg^8 mst1^4 mst2^8+90 mg^6 msq^2 mst1^4 mst2^8+18070 mg^6 mst1^6 mst2^8+210 mg^4 msq^2 mst1^6 mst2^8-5474 mg^4 mst1^8 mst2^8-30 mg^2 msq^2 mst1^8 mst2^8+616 mg^2 mst1^10 mst2^8-136 mg^10 mst2^10-4623 mg^8 mst1^2 mst2^10+360 mg^6 msq^2 mst1^2 mst2^10+9553 mg^6 mst1^4 mst2^10+30 mg^4 msq^2 mst1^4 mst2^10-6600 mg^4 mst1^6 mst2^10-30 mg^2 msq^2 mst1^6 mst2^10+1602 mg^2 mst1^8 mst2^10-96 mst1^10 mst2^10+34 mg^8 mst2^12+702 mg^6 mst1^2 mst2^12-90 mg^4 msq^2 mst1^2 mst2^12-1268 mg^4 mst1^4 mst2^12-30 mg^2 msq^2 mst1^4 mst2^12+884 mg^2 mst1^6 mst2^12-240 mst1^8 mst2^12-9 mg^4 mst1^2 mst2^14-3 mg^2 mst1^4 mst2^14) Xt^4)/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^3)-(64 (52 mg^11 mst1^2-276 mg^9 mst1^4+60 mg^7 msq^2 mst1^4+214 mg^7 mst1^6-30 mg^5 msq^2 mst1^6-3 mg^5 mst1^8+16 mg^11 mst2^2-362 mg^9 mst1^2 mst2^2+60 mg^7 msq^2 mst1^2 mst2^2+770 mg^7 mst1^4 mst2^2-120 mg^5 msq^2 mst1^4 mst2^2-407 mg^5 mst1^6 mst2^2+30 mg^3 msq^2 mst1^6 mst2^2+3 mg^3 mst1^8 mst2^2-32 mg^9 mst2^4+558 mg^7 mst1^2 mst2^4-90 mg^5 msq^2 mst1^2 mst2^4-737 mg^5 mst1^4 mst2^4+60 mg^3 msq^2 mst1^4 mst2^4+213 mg^3 mst1^6 mst2^4+16 mg^7 mst2^6-219 mg^5 mst1^2 mst2^6+30 mg^3 msq^2 mst1^2 mst2^6+179 mg^3 mst1^4 mst2^6+12 mg mst1^6 mst2^6+3 mg^3 mst1^2 mst2^8) Xt^5)/(3 mst1^2 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)+(5120 mg^4 mst2^2 Xt^6)/(3 (mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)+((160 (3 mg^12-13 mg^10 mst1^2+6 mg^8 mst1^4-2 mg^6 mst1^6-11 mg^10 mst2^2+45 mg^8 mst1^2 mst2^2-24 mg^6 mst1^4 mst2^2+8 mg^4 mst1^6 mst2^2+3 mg^8 mst2^4-21 mg^6 mst1^2 mst2^4-mg^6 mst2^6+7 mg^4 mst1^2 mst2^6))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3)+((320 mg^3 (-2 mg^2+mst1^2+mst2^2))/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2))+(640 (mg^5 mst1^2+mg^5 mst2^2-2 mg^3 mst1^2 mst2^2))/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)) Xt+(320 (2 mg^12-10 mg^10 mst1^2+3 mg^8 mst1^4-mg^6 mst1^6-10 mg^10 mst2^2+42 mg^8 mst1^2 mst2^2-21 mg^6 mst1^4 mst2^2+7 mg^4 mst1^6 mst2^2+3 mg^8 mst2^4-21 mg^6 mst1^2 mst2^4-mg^6 mst2^6+7 mg^4 mst1^2 mst2^6) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+((1280 (mg^5 mst1^2+mg^5 mst2^2-2 mg^3 mst1^2 mst2^2))/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))-(640 mg^3 (2 mg^4-mst1^4+4 mst1^2 mst2^2-mst2^4-2 mg^2 (mst1^2+mst2^2)))/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^3)) Xt^3-(160 (mst1^2+mst2^2) (2 mg^12-10 mg^10 mst1^2+3 mg^8 mst1^4-mg^6 mst1^6-10 mg^10 mst2^2+42 mg^8 mst1^2 mst2^2-21 mg^6 mst1^4 mst2^2+7 mg^4 mst1^6 mst2^2+3 mg^8 mst2^4-21 mg^6 mst1^2 mst2^4-mg^6 mst2^6+7 mg^4 mst1^2 mst2^6) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)-(320 (mst1^2+mst2^2) (mg^7+5 mg^5 mst1^2+5 mg^5 mst2^2-11 mg^3 mst1^2 mst2^2) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)) Log[msq^2/MR^2]+((16 (90 mg^16-330 mg^14 mst1^2+355 mg^12 mst1^4-128 mg^10 mst1^6+17 mg^8 mst1^8-330 mg^14 mst2^2+1196 mg^12 mst1^2 mst2^2-1260 mg^10 mst1^4 mst2^2+444 mg^8 mst1^6 mst2^2-66 mg^6 mst1^8 mst2^2+355 mg^12 mst2^4-1260 mg^10 mst1^2 mst2^4+1160 mg^8 mst1^4 mst2^4-248 mg^6 mst1^6 mst2^4+17 mg^4 mst1^8 mst2^4-128 mg^10 mst2^6+444 mg^8 mst1^2 mst2^6-248 mg^6 mst1^4 mst2^6-132 mg^4 mst1^6 mst2^6+48 mg^2 mst1^8 mst2^6+17 mg^8 mst2^8-66 mg^6 mst1^2 mst2^8+17 mg^4 mst1^4 mst2^8+48 mg^2 mst1^6 mst2^8-12 mst1^8 mst2^8))/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^4)-(512 (5 mg^11-8 mg^9 mst1^2+5 mg^7 mst1^4-8 mg^9 mst2^2+16 mg^7 mst1^2 mst2^2-12 mg^5 mst1^4 mst2^2+5 mg^7 mst2^4-12 mg^5 mst1^2 mst2^4+9 mg^3 mst1^4 mst2^4) Xt)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3)-(64 (64 mg^18-310 mg^16 mst1^2+516 mg^14 mst1^4-405 mg^12 mst1^6+164 mg^10 mst1^8-27 mg^8 mst1^10-310 mg^16 mst2^2+1368 mg^14 mst1^2 mst2^2-2083 mg^12 mst1^4 mst2^2+1488 mg^10 mst1^6 mst2^2-561 mg^8 mst1^8 mst2^2+88 mg^6 mst1^10 mst2^2+516 mg^14 mst2^4-2083 mg^12 mst1^2 mst2^4+2808 mg^10 mst1^4 mst2^4-1704 mg^8 mst1^6 mst2^4+564 mg^6 mst1^8 mst2^4-81 mg^4 mst1^10 mst2^4-405 mg^12 mst2^6+1488 mg^10 mst1^2 mst2^6-1704 mg^8 mst1^4 mst2^6+776 mg^6 mst1^6 mst2^6-199 mg^4 mst1^8 mst2^6+24 mg^2 mst1^10 mst2^6+164 mg^10 mst2^8-561 mg^8 mst1^2 mst2^8+564 mg^6 mst1^4 mst2^8-199 mg^4 mst1^6 mst2^8+48 mg^2 mst1^8 mst2^8-6 mst1^10 mst2^8-27 mg^8 mst2^10+88 mg^6 mst1^2 mst2^10-81 mg^4 mst1^4 mst2^10+24 mg^2 mst1^6 mst2^10-6 mst1^8 mst2^10) Xt^2)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^2)+(256 (18 mg^13-75 mg^11 mst1^2+91 mg^9 mst1^4-26 mg^7 mst1^6-75 mg^11 mst2^2+232 mg^9 mst1^2 mst2^2-232 mg^7 mst1^4 mst2^2+51 mg^5 mst1^6 mst2^2+91 mg^9 mst2^4-232 mg^7 mst1^2 mst2^4+192 mg^5 mst1^4 mst2^4-27 mg^3 mst1^6 mst2^4-26 mg^7 mst2^6+51 mg^5 mst1^2 mst2^6-27 mg^3 mst1^4 mst2^6-6 mg mst1^6 mst2^6) Xt^3)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)+(32 (320 mg^18 mst1^2-1036 mg^16 mst1^4+1044 mg^14 mst1^6-321 mg^12 mst1^8+36 mg^10 mst1^10-41 mg^8 mst1^12+320 mg^18 mst2^2-2584 mg^16 mst1^2 mst2^2+6252 mg^14 mst1^4 mst2^2-5712 mg^12 mst1^6 mst2^2+1612 mg^10 mst1^8 mst2^2-24 mg^8 mst1^10 mst2^2+128 mg^6 mst1^12 mst2^2-1036 mg^16 mst2^4+6252 mg^14 mst1^2 mst2^4-13598 mg^12 mst1^4 mst2^4+12176 mg^10 mst1^6 mst2^4-3615 mg^8 mst1^8 mst2^4-44 mg^6 mst1^10 mst2^4-125 mg^4 mst1^12 mst2^4+1044 mg^14 mst2^6-5712 mg^12 mst1^2 mst2^6+12176 mg^10 mst1^4 mst2^6-11360 mg^8 mst1^6 mst2^6+3884 mg^6 mst1^8 mst2^6-80 mg^4 mst1^10 mst2^6+48 mg^2 mst1^12 mst2^6-321 mg^12 mst2^8+1612 mg^10 mst1^2 mst2^8-3615 mg^8 mst1^4 mst2^8+3884 mg^6 mst1^6 mst2^8-1702 mg^4 mst1^8 mst2^8+144 mg^2 mst1^10 mst2^8-12 mst1^12 mst2^8+36 mg^10 mst2^10-24 mg^8 mst1^2 mst2^10-44 mg^6 mst1^4 mst2^10-80 mg^4 mst1^6 mst2^10+144 mg^2 mst1^8 mst2^10-24 mst1^10 mst2^10-41 mg^8 mst2^12+128 mg^6 mst1^2 mst2^12-125 mg^4 mst1^4 mst2^12+48 mg^2 mst1^6 mst2^12-12 mst1^8 mst2^12) Xt^4)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^4)-(128 (34 mg^13 mst1^2-107 mg^11 mst1^4+115 mg^9 mst1^6-34 mg^7 mst1^8+34 mg^13 mst2^2-214 mg^11 mst1^2 mst2^2+411 mg^9 mst1^4 mst2^2-314 mg^7 mst1^6 mst2^2+67 mg^5 mst1^8 mst2^2-107 mg^11 mst2^4+411 mg^9 mst1^2 mst2^4-560 mg^7 mst1^4 mst2^4+291 mg^5 mst1^6 mst2^4-35 mg^3 mst1^8 mst2^4+115 mg^9 mst2^6-314 mg^7 mst1^2 mst2^6+291 mg^5 mst1^4 mst2^6-70 mg^3 mst1^6 mst2^6-6 mg mst1^8 mst2^6-34 mg^7 mst2^8+67 mg^5 mst1^2 mst2^8-35 mg^3 mst1^4 mst2^8-6 mg mst1^6 mst2^8) Xt^5)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^4)+(2560 mg^4 (mst1^2+mst2^2) (mg^2 mst1^2+mg^2 mst2^2-2 mst1^2 mst2^2) Xt^6)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)) Log[mst1^2/MR^2]) Log[mst2^2/MR^2]+((16 (64 mg^16-139 mg^14 mst1^2+87 mg^12 mst1^4+22 mg^10 mst1^6-28 mg^8 mst1^8-245 mg^14 mst2^2+633 mg^12 mst1^2 mst2^2-700 mg^10 mst1^4 mst2^2+316 mg^8 mst1^6 mst2^2-34 mg^6 mst1^8 mst2^2+240 mg^12 mst2^4-432 mg^10 mst1^2 mst2^4+400 mg^8 mst1^4 mst2^4-146 mg^6 mst1^6 mst2^4-2 mg^4 mst1^8 mst2^4-170 mg^10 mst2^6+162 mg^8 mst1^2 mst2^6-20 mg^6 mst1^4 mst2^6-80 mg^4 mst1^6 mst2^6+48 mg^2 mst1^8 mst2^6+110 mg^8 mst2^8-157 mg^6 mst1^2 mst2^8+101 mg^4 mst1^4 mst2^8-12 mg^2 mst1^6 mst2^8-12 mst1^8 mst2^8-27 mg^6 mst2^10+45 mg^4 mst1^2 mst2^10-36 mg^2 mst1^4 mst2^10+12 mst1^6 mst2^10))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2))+(32 (18 mg^11-19 mg^9 mst1^2+27 mg^7 mst1^4+195 mg^9 mst2^2-425 mg^7 mst1^2 mst2^2+168 mg^5 mst1^4 mst2^2-130 mg^7 mst2^4+263 mg^5 mst1^2 mst2^4-87 mg^3 mst1^4 mst2^4+13 mg^5 mst2^6-11 mg^3 mst1^2 mst2^6-12 mg mst1^4 mst2^6) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+(32 (64 mg^16-54 mg^14 mst1^2-72 mg^12 mst1^4+69 mg^10 mst1^6-mg^8 mst1^8-502 mg^14 mst2^2+962 mg^12 mst1^2 mst2^2-545 mg^10 mst1^4 mst2^2+79 mg^8 mst1^6 mst2^2-24 mg^6 mst1^8 mst2^2+954 mg^12 mst2^4-1897 mg^10 mst1^2 mst2^4+1199 mg^8 mst1^4 mst2^4-189 mg^6 mst1^6 mst2^4-7 mg^4 mst1^8 mst2^4-759 mg^10 mst2^6+1401 mg^8 mst1^2 mst2^6-811 mg^6 mst1^4 mst2^6+61 mg^4 mst1^6 mst2^6+48 mg^2 mst1^8 mst2^6+302 mg^8 mst2^8-525 mg^6 mst1^2 mst2^8+325 mg^4 mst1^4 mst2^8-60 mg^2 mst1^6 mst2^8-12 mst1^8 mst2^8-55 mg^6 mst2^10+97 mg^4 mst1^2 mst2^10-72 mg^2 mst1^4 mst2^10+24 mst1^6 mst2^10) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^2)+(64 (42 mg^13-6 mg^11 mst1^2-67 mg^9 mst1^4+45 mg^7 mst1^6-120 mg^11 mst2^2-44 mg^9 mst1^2 mst2^2+214 mg^7 mst1^4 mst2^2-90 mg^5 mst1^6 mst2^2+239 mg^9 mst2^4-67 mg^7 mst1^2 mst2^4-195 mg^5 mst1^4 mst2^4+59 mg^3 mst1^6 mst2^4-252 mg^7 mst2^6+268 mg^5 mst1^2 mst2^6-42 mg^3 mst1^4 mst2^6+18 mg mst1^6 mst2^6+59 mg^5 mst2^8-55 mg^3 mst1^2 mst2^8-6 mg mst1^4 mst2^8) Xt^3)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)-(16 (-148 mg^16 mst1^2+336 mg^14 mst1^4-194 mg^12 mst1^6-mg^10 mst1^8+13 mg^8 mst1^10+788 mg^16 mst2^2-1796 mg^14 mst1^2 mst2^2+786 mg^12 mst1^4 mst2^2+452 mg^10 mst1^6 mst2^2-190 mg^8 mst1^8 mst2^2-64 mg^6 mst1^10 mst2^2-2556 mg^14 mst2^4+7366 mg^12 mst1^2 mst2^4-6232 mg^10 mst1^4 mst2^4+936 mg^8 mst1^6 mst2^4+479 mg^6 mst1^8 mst2^4+37 mg^4 mst1^10 mst2^4+2618 mg^12 mst2^6-8472 mg^10 mst1^2 mst2^6+8712 mg^8 mst1^4 mst2^6-2648 mg^6 mst1^6 mst2^6-234 mg^4 mst1^8 mst2^6+24 mg^2 mst1^10 mst2^6-835 mg^10 mst2^8+3031 mg^8 mst1^2 mst2^8-3742 mg^6 mst1^4 mst2^8+1612 mg^4 mst1^6 mst2^8-90 mg^2 mst1^8 mst2^8-6 mst1^10 mst2^8+58 mg^8 mst2^10-116 mg^6 mst1^2 mst2^10+238 mg^4 mst1^4 mst2^10-180 mg^2 mst1^6 mst2^10+24 mst1^8 mst2^10-69 mg^6 mst2^12+123 mg^4 mst1^2 mst2^12-90 mg^2 mst1^4 mst2^12+30 mst1^6 mst2^12) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)^4)+(32 (100 mg^11 mst1^2-169 mg^9 mst1^4+75 mg^7 mst1^6+36 mg^11 mst2^2-392 mg^9 mst1^2 mst2^2+508 mg^7 mst1^4 mst2^2-148 mg^5 mst1^6 mst2^2-159 mg^9 mst2^4+585 mg^7 mst1^2 mst2^4-535 mg^5 mst1^4 mst2^4+93 mg^3 mst1^6 mst2^4+216 mg^7 mst2^6-384 mg^5 mst1^2 mst2^6+152 mg^3 mst1^4 mst2^6+12 mg mst1^6 mst2^6-61 mg^5 mst2^8+59 mg^3 mst1^2 mst2^8+12 mg mst1^4 mst2^8) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^4)-(2560 mg^4 mst2^2 (mst1^2+mst2^2) Xt^6)/(3 (mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)) Log[mst2^2/MR^2]^2+(-((64 mg^4 (2+(8 (2 mg^8-2 mg^6 mst1^2+mg^4 mst1^4-2 mg^6 mst2^2+mg^4 mst2^4))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)))/(mg^2-mst1^2)^2)-(512 (11 mg^11-10 mg^9 mst1^2+7 mg^7 mst1^4-18 mg^9 mst2^2+8 mg^7 mst1^2 mst2^2-6 mg^5 mst1^4 mst2^2+11 mg^7 mst2^4-6 mg^5 mst1^2 mst2^4+3 mg^3 mst1^4 mst2^4) Xt)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))-(8192 mg^6 Xt^2)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2) (mst1^2-mst2^2))+(128 mg (-2 mg^2+mst1^2+mst2^2) (2+(8 (2 mg^8-2 mg^6 mst1^2+mg^4 mst1^4-2 mg^6 mst2^2+mg^4 mst2^4))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)) Xt^3)/(-mst1^2+mst2^2)^3+(64 (2 mg^2-mst1^2-mst2^2) (53 mg^8-50 mg^6 mst1^2-7 mg^4 mst1^4-50 mg^6 mst2^2+52 mg^4 mst1^2 mst2^2+6 mg^2 mst1^4 mst2^2-7 mg^4 mst2^4+6 mg^2 mst1^2 mst2^4-3 mst1^4 mst2^4) Xt^4)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)-(1024 mg^3 (-2 mg^2+mst1^2+mst2^2) Xt^5)/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^3)) PolyLog[2,1-mg^2/mst1^2]+(-((64 mg^4 (2+(8 (2 mg^8-2 mg^6 mst1^2+mg^4 mst1^4-2 mg^6 mst2^2+mg^4 mst2^4))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)))/(-mg^2+mst2^2)^2)+(512 (11 mg^11-18 mg^9 mst1^2+11 mg^7 mst1^4-10 mg^9 mst2^2+8 mg^7 mst1^2 mst2^2-6 mg^5 mst1^4 mst2^2+7 mg^7 mst2^4-6 mg^5 mst1^2 mst2^4+3 mg^3 mst1^4 mst2^4) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))-(8192 mg^6 Xt^2)/(3 (mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(128 mg (2 mg^2-mst1^2-mst2^2) (2+(8 (2 mg^8-2 mg^6 mst1^2+mg^4 mst1^4-2 mg^6 mst2^2+mg^4 mst2^4))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)) Xt^3)/(-mst1^2+mst2^2)^3-(64 (2 mg^2-mst1^2-mst2^2) (53 mg^8-50 mg^6 mst1^2-7 mg^4 mst1^4-50 mg^6 mst2^2+52 mg^4 mst1^2 mst2^2+6 mg^2 mst1^4 mst2^2-7 mg^4 mst2^4+6 mg^2 mst1^2 mst2^4-3 mst1^4 mst2^4) Xt^4)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)+(1024 mg^3 (-2 mg^2+mst1^2+mst2^2) Xt^5)/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^3)) PolyLog[2,1-mg^2/mst2^2])+(-((80 (-msq^2+mst2^2) (2 mg^4-3 mg^2 msq^2+3 mg^2 mst2^2-msq^2 mst2^2-mst2^4))/(mg^2-mst2^2)^3)+(640 mg (-msq^2+mst2^2)^2 Xt)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(160 (-msq^2+mst2^2) (2 mg^4-3 mg^2 msq^2+3 mg^2 mst2^2-msq^2 mst2^2-mst2^4) Xt^4)/((mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(1280 mg (-msq^2+mst2^2)^2 Xt^5)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)) PolyLog[2,1-msq^2/mst2^2]+((4 (24 mg^10 mst1^2-28 mg^8 mst1^4+68 mg^6 mst1^6-58 mg^4 mst1^8+18 mg^2 mst1^10-24 mg^10 mst2^2-220 mg^6 mst1^4 mst2^2+188 mg^4 mst1^6 mst2^2-70 mg^2 mst1^8 mst2^2+6 mst1^10 mst2^2+28 mg^8 mst2^4+220 mg^6 mst1^2 mst2^4-16 mg^2 mst1^6 mst2^4+8 mst1^8 mst2^4-68 mg^6 mst2^6-188 mg^4 mst1^2 mst2^6+16 mg^2 mst1^4 mst2^6+58 mg^4 mst2^8+70 mg^2 mst1^2 mst2^8-8 mst1^4 mst2^8-18 mg^2 mst2^10-6 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (-mg^2+mst2^2)^3)-(32 (-10 mg^5 mst1^2+10 mg^3 mst1^4-6 mg mst1^6+10 mg^5 mst2^2+8 mg mst1^4 mst2^2-10 mg^3 mst2^4-8 mg mst1^2 mst2^4+6 mg mst2^6) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)-(16 (12 mg^10-14 mg^8 mst1^2+34 mg^6 mst1^4-29 mg^4 mst1^6+9 mg^2 mst1^8-14 mg^8 mst2^2-76 mg^6 mst1^2 mst2^2+65 mg^4 mst1^4 mst2^2-26 mg^2 mst1^6 mst2^2+3 mst1^8 mst2^2+34 mg^6 mst2^4+65 mg^4 mst1^2 mst2^4-34 mg^2 mst1^4 mst2^4+7 mst1^6 mst2^4-29 mg^4 mst2^6-26 mg^2 mst1^2 mst2^6+7 mst1^4 mst2^6+9 mg^2 mst2^8+3 mst1^2 mst2^8) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+(128 (5 mg^5-5 mg^3 mst1^2+3 mg mst1^4-5 mg^3 mst2^2-mg mst1^2 mst2^2+3 mg mst2^4) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) PolyLog[2,1-mst2^2/mst1^2]+Log[MR^2/mt^2] (-((16 (188 mg^8-180 mg^6 msq^2-316 mg^6 mst1^2+150 mg^4 msq^2 mst1^2+169 mg^4 mst1^4-90 mg^2 msq^2 mst1^4-9 mg^2 mst1^6-316 mg^6 mst2^2+150 mg^4 msq^2 mst2^2+454 mg^4 mst1^2 mst2^2+120 mg^2 msq^2 mst1^2 mst2^2-211 mg^2 mst1^4 mst2^2-30 msq^2 mst1^4 mst2^2-3 mst1^6 mst2^2+169 mg^4 mst2^4-90 mg^2 msq^2 mst2^4-211 mg^2 mst1^2 mst2^4-30 msq^2 mst1^2 mst2^4+98 mst1^4 mst2^4-9 mg^2 mst2^6-3 mst1^2 mst2^6))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2))+((1792 mg^3)/(3 (mg^2-mst1^2) (mg^2-mst2^2))-(1280 mg msq^2)/((mg^2-mst1^2) (mg^2-mst2^2))-(128 mg mst1^2)/((mg^2-mst1^2) (mg^2-mst2^2))-(128 mg mst2^2)/((mg^2-mst1^2) (mg^2-mst2^2))) Xt-128 z3+(-((16 (198 mg^16-726 mg^14 mst1^2+721 mg^12 mst1^4-232 mg^10 mst1^6+27 mg^8 mst1^8-726 mg^14 mst2^2+2700 mg^12 mst1^2 mst2^2-2700 mg^10 mst1^4 mst2^2+908 mg^8 mst1^6 mst2^2-134 mg^6 mst1^8 mst2^2+721 mg^12 mst2^4-2700 mg^10 mst1^2 mst2^4+2240 mg^8 mst1^4 mst2^4-320 mg^6 mst1^6 mst2^4-13 mg^4 mst1^8 mst2^4-232 mg^10 mst2^6+908 mg^8 mst1^2 mst2^6-320 mg^6 mst1^4 mst2^6-452 mg^4 mst1^6 mst2^6+144 mg^2 mst1^8 mst2^6+27 mg^8 mst2^8-134 mg^6 mst1^2 mst2^8-13 mg^4 mst1^4 mst2^8+144 mg^2 mst1^6 mst2^8-36 mst1^8 mst2^8))/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^4))+(128 (mg^11-50 mg^9 mst1^2+41 mg^7 mst1^4-50 mg^9 mst2^2+196 mg^7 mst1^2 mst2^2-130 mg^5 mst1^4 mst2^2+41 mg^7 mst2^4-130 mg^5 mst1^2 mst2^4+81 mg^3 mst1^4 mst2^4) Xt)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3)+(512 mg^6 Xt^2)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)) Log[mg^2/MR^2]^2+((160 (10 mg^8+18 mg^6 msq^2-20 mg^6 mst1^2-15 mg^4 msq^2 mst1^2+10 mg^4 mst1^4+9 mg^2 msq^2 mst1^4-20 mg^6 mst2^2-15 mg^4 msq^2 mst2^2+40 mg^4 mst1^2 mst2^2-12 mg^2 msq^2 mst1^2 mst2^2-20 mg^2 mst1^4 mst2^2+3 msq^2 mst1^4 mst2^2+10 mg^4 mst2^4+9 mg^2 msq^2 mst2^4-20 mg^2 mst1^2 mst2^4+3 msq^2 mst1^2 mst2^4+10 mst1^4 mst2^4))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)-(1280 mg msq^2 Xt)/((mg^2-mst1^2) (mg^2-mst2^2))) Log[msq^2/MR^2]+(-48 (-((5 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(6 (-mg^2+mst1^2)^3))-(5 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(6 (-mg^2+mst2^2)^3))-48 ((40 mg (-msq^2+mst1^2)^2)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+(40 mg (-msq^2+mst2^2)^2)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt) Log[msq^2/MR^2]^2+((16 (-310 mg^10 mst1^2+550 mg^8 mst1^4+90 mg^6 msq^2 mst1^4-216 mg^6 mst1^6+30 mg^4 msq^2 mst1^6-35 mg^4 mst1^8-9 mg^2 mst1^10+54 mg^10 mst2^2+582 mg^8 mst1^2 mst2^2-90 mg^6 msq^2 mst1^2 mst2^2-1078 mg^6 mst1^4 mst2^2-210 mg^4 msq^2 mst1^4 mst2^2+359 mg^4 mst1^6 mst2^2-60 mg^2 msq^2 mst1^6 mst2^2+158 mg^2 mst1^8 mst2^2-3 mst1^10 mst2^2-108 mg^8 mst2^4-296 mg^6 mst1^2 mst2^4+180 mg^4 msq^2 mst1^2 mst2^4+679 mg^4 mst1^4 mst2^4+150 mg^2 msq^2 mst1^4 mst2^4-296 mg^2 mst1^6 mst2^4+30 msq^2 mst1^6 mst2^4-75 mst1^8 mst2^4+54 mg^6 mst2^6+21 mg^4 mst1^2 mst2^6-90 mg^2 msq^2 mst1^2 mst2^6-100 mg^2 mst1^4 mst2^6-30 msq^2 mst1^4 mst2^6+81 mst1^6 mst2^6-9 mg^2 mst1^2 mst2^8-3 mst1^4 mst2^8))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (mst1^2-mst2^2))-(128 (16 mg^7+10 mg^5 mst1^2-30 mg^3 msq^2 mst1^2-26 mg^3 mst1^4+3 mg mst1^6-16 mg^5 mst2^2-18 mg^3 mst1^2 mst2^2+30 mg msq^2 mst1^2 mst2^2+28 mg mst1^4 mst2^2+3 mg mst1^2 mst2^4) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2) (-mst1^2+mst2^2))+((160 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst1^2-msq^4 mst1^2))/(mg^2-mst1^2)^3-(1280 (-mg msq^4+2 mg msq^2 mst1^2) Xt)/((mg^2-mst1^2)^2 (mst1^2-mst2^2))) Log[msq^2/MR^2]) Log[mst1^2/MR^2]+(-((8 (128 mg^10-556 mg^8 mst1^2+60 mg^6 msq^2 mst1^2-90 mg^4 msq^4 mst1^2+706 mg^6 mst1^4+120 mg^4 msq^2 mst1^4+60 mg^2 msq^4 mst1^4-437 mg^4 mst1^6-180 mg^2 msq^2 mst1^6+30 msq^4 mst1^6+136 mg^2 mst1^8-mst1^10+44 mg^8 mst2^2-60 mg^6 msq^2 mst2^2+90 mg^4 msq^4 mst2^2+64 mg^6 mst1^2 mst2^2-120 mg^4 msq^2 mst1^2 mst2^2-60 mg^2 msq^4 mst1^2 mst2^2-51 mg^4 mst1^4 mst2^2+180 mg^2 msq^2 mst1^4 mst2^2-30 msq^4 mst1^4 mst2^2-38 mg^2 mst1^6 mst2^2+5 mst1^8 mst2^2-2 mg^6 mst2^4-33 mg^4 mst1^2 mst2^4+36 mg^2 mst1^4 mst2^4-mst1^6 mst2^4+9 mg^4 mst2^6-6 mg^2 mst1^2 mst2^6-3 mst1^4 mst2^6))/(3 (mg^2-mst1^2)^4 (mst1^2-mst2^2)))+(64 (18 mg^7 mst1^2-30 mg^3 msq^4 mst1^2-85 mg^5 mst1^4+60 mg^3 msq^2 mst1^4+30 mg msq^4 mst1^4+57 mg^3 mst1^6-60 mg msq^2 mst1^6-6 mg mst1^8-18 mg^7 mst2^2+30 mg^3 msq^4 mst2^2+68 mg^5 mst1^2 mst2^2-60 mg^3 msq^2 mst1^2 mst2^2-30 mg msq^4 mst1^2 mst2^2-17 mg^3 mst1^4 mst2^2+60 mg msq^2 mst1^4 mst2^2-17 mg mst1^6 mst2^2+mg^5 mst2^4-11 mg^3 mst1^2 mst2^4+10 mg mst1^4 mst2^4+3 mg^3 mst2^6-3 mg mst1^2 mst2^6) Xt)/(3 (mg^2-mst1^2)^3 (mst1^2-mst2^2)^2)+(512 mg^2 mst1^4 Xt^2)/((mg^2-mst1^2)^2 (mst1^2-mst2^2)^2)) Log[mst1^2/MR^2]^2+((16 (54 mg^10 mst1^2-108 mg^8 mst1^4+54 mg^6 mst1^6-310 mg^10 mst2^2+582 mg^8 mst1^2 mst2^2-90 mg^6 msq^2 mst1^2 mst2^2-296 mg^6 mst1^4 mst2^2+180 mg^4 msq^2 mst1^4 mst2^2+21 mg^4 mst1^6 mst2^2-90 mg^2 msq^2 mst1^6 mst2^2-9 mg^2 mst1^8 mst2^2+550 mg^8 mst2^4+90 mg^6 msq^2 mst2^4-1078 mg^6 mst1^2 mst2^4-210 mg^4 msq^2 mst1^2 mst2^4+679 mg^4 mst1^4 mst2^4+150 mg^2 msq^2 mst1^4 mst2^4-100 mg^2 mst1^6 mst2^4-30 msq^2 mst1^6 mst2^4-3 mst1^8 mst2^4-216 mg^6 mst2^6+30 mg^4 msq^2 mst2^6+359 mg^4 mst1^2 mst2^6-60 mg^2 msq^2 mst1^2 mst2^6-296 mg^2 mst1^4 mst2^6+30 msq^2 mst1^4 mst2^6+81 mst1^6 mst2^6-35 mg^4 mst2^8+158 mg^2 mst1^2 mst2^8-75 mst1^4 mst2^8-9 mg^2 mst2^10-3 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+(128 (16 mg^7-16 mg^5 mst1^2+10 mg^5 mst2^2-30 mg^3 msq^2 mst2^2-18 mg^3 mst1^2 mst2^2+30 mg msq^2 mst1^2 mst2^2+3 mg mst1^4 mst2^2-26 mg^3 mst2^4+28 mg mst1^2 mst2^4+3 mg mst2^6) Xt)/(3 (mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+((160 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst2^2-msq^4 mst2^2))/(mg^2-mst2^2)^3-(1280 (-mg msq^4+2 mg msq^2 mst2^2) Xt)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Log[msq^2/MR^2]+((16 (12 mg^10 mst2^2+64 mg^8 mst1^2 mst2^2-78 mg^6 mst1^4 mst2^2+26 mg^4 mst1^6 mst2^2-14 mg^8 mst2^4-188 mg^6 mst1^2 mst2^4+175 mg^4 mst1^4 mst2^4-57 mg^2 mst1^6 mst2^4+34 mg^6 mst2^6+120 mg^4 mst1^2 mst2^6-65 mg^2 mst1^4 mst2^6+19 mst1^6 mst2^6-29 mg^4 mst2^8-35 mg^2 mst1^2 mst2^8+4 mst1^4 mst2^8+9 mg^2 mst2^10+3 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3)+(128 (16 mg^7 mst1^2 mst2^2-29 mg^5 mst1^4 mst2^2+15 mg^3 mst1^6 mst2^2-14 mg^5 mst1^2 mst2^4+23 mg^3 mst1^4 mst2^4-4 mg mst1^6 mst2^4-5 mg^5 mst2^6+5 mg^3 mst1^2 mst2^6-19 mg mst1^4 mst2^6+5 mg^3 mst2^8+10 mg mst1^2 mst2^8-3 mg mst2^10) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)-(1024 mg^2 mst1^2 mst2^2 Xt^2)/((mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^2)) Log[mst1^2/MR^2]) Log[mst2^2/MR^2]+(-((8 (128 mg^16-340 mg^14 mst1^2-60 mg^12 msq^2 mst1^2+90 mg^10 msq^4 mst1^2+262 mg^12 mst1^4+180 mg^10 msq^2 mst1^4-270 mg^8 msq^4 mst1^4+5 mg^10 mst1^6-180 mg^8 msq^2 mst1^6+270 mg^6 msq^4 mst1^6-43 mg^8 mst1^8+60 mg^6 msq^2 mst1^8-90 mg^4 msq^4 mst1^8-556 mg^14 mst2^2+60 mg^12 msq^2 mst2^2-90 mg^10 msq^4 mst2^2+1708 mg^12 mst1^2 mst2^2-300 mg^10 msq^2 mst1^2 mst2^2+210 mg^8 msq^4 mst1^2 mst2^2-1891 mg^10 mst1^4 mst2^2+540 mg^8 msq^2 mst1^4 mst2^2-90 mg^6 msq^4 mst1^4 mst2^2+711 mg^8 mst1^6 mst2^2-420 mg^6 msq^2 mst1^6 mst2^2-90 mg^4 msq^4 mst1^6 mst2^2-56 mg^6 mst1^8 mst2^2+120 mg^4 msq^2 mst1^8 mst2^2+60 mg^2 msq^4 mst1^8 mst2^2+718 mg^12 mst2^4+120 mg^10 msq^2 mst2^4+60 mg^8 msq^4 mst2^4-2131 mg^10 mst1^2 mst2^4-180 mg^8 msq^2 mst1^2 mst2^4-210 mg^6 msq^4 mst1^2 mst2^4+2513 mg^8 mst1^4 mst2^4-180 mg^6 msq^2 mst1^4 mst2^4+270 mg^4 msq^4 mst1^4 mst2^4-920 mg^6 mst1^6 mst2^4+420 mg^4 msq^2 mst1^6 mst2^4-150 mg^2 msq^4 mst1^6 mst2^4+72 mg^4 mst1^8 mst2^4-180 mg^2 msq^2 mst1^8 mst2^4+30 msq^4 mst1^8 mst2^4-463 mg^10 mst2^6-180 mg^8 msq^2 mst2^6+30 mg^6 msq^4 mst2^6+1115 mg^8 mst1^2 mst2^6+540 mg^6 msq^2 mst1^2 mst2^6-90 mg^4 msq^4 mst1^2 mst2^6-1512 mg^6 mst1^4 mst2^6-540 mg^4 msq^2 mst1^4 mst2^6+90 mg^2 msq^4 mst1^4 mst2^6+436 mg^4 mst1^6 mst2^6+180 mg^2 msq^2 mst1^6 mst2^6-30 msq^4 mst1^6 mst2^6+4 mg^2 mst1^8 mst2^6+184 mg^8 mst2^8-136 mg^6 mst1^2 mst2^8+514 mg^4 mst1^4 mst2^8-141 mg^2 mst1^6 mst2^8-mst1^8 mst2^8-64 mg^6 mst2^10-164 mg^4 mst1^2 mst2^10-29 mg^2 mst1^4 mst2^10+5 mst1^6 mst2^10+38 mg^4 mst2^12+47 mg^2 mst1^2 mst2^12-mst1^4 mst2^12-9 mg^2 mst2^14-3 mst1^2 mst2^14))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2)))+(64 (-18 mg^11 mst1^2+30 mg^7 msq^4 mst1^2+37 mg^9 mst1^4-60 mg^5 msq^4 mst1^4-22 mg^7 mst1^6+30 mg^3 msq^4 mst1^6+18 mg^11 mst2^2-30 mg^7 msq^4 mst2^2+32 mg^9 mst1^2 mst2^2-60 mg^7 msq^2 mst1^2 mst2^2+30 mg^5 msq^4 mst1^2 mst2^2-114 mg^7 mst1^4 mst2^2+120 mg^5 msq^2 mst1^4 mst2^2+30 mg^3 msq^4 mst1^4 mst2^2+82 mg^5 mst1^6 mst2^2-60 mg^3 msq^2 mst1^6 mst2^2-30 mg msq^4 mst1^6 mst2^2-85 mg^9 mst2^4+60 mg^7 msq^2 mst2^4+30 mg^5 msq^4 mst2^4+138 mg^7 mst1^2 mst2^4-60 mg^5 msq^2 mst1^2 mst2^4-60 mg^3 msq^4 mst1^2 mst2^4-56 mg^5 mst1^4 mst2^4-60 mg^3 msq^2 mst1^4 mst2^4+30 mg msq^4 mst1^4 mst2^4-42 mg^3 mst1^6 mst2^4+60 mg msq^2 mst1^6 mst2^4+62 mg^7 mst2^6-60 mg^5 msq^2 mst2^6-106 mg^5 mst1^2 mst2^6+120 mg^3 msq^2 mst1^2 mst2^6+106 mg^3 mst1^4 mst2^6-60 mg msq^2 mst1^4 mst2^6-2 mg mst1^6 mst2^6-16 mg^5 mst2^8-8 mg^3 mst1^2 mst2^8-21 mg mst1^4 mst2^8+8 mg^3 mst2^10+10 mg mst1^2 mst2^10-3 mg mst2^12) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)+(512 mg^2 mst2^4 Xt^2)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)) Log[mst2^2/MR^2]^2+Log[mg^2/MR^2] ((16 (660 mg^12-180 mg^10 msq^2-1562 mg^10 mst1^2+240 mg^8 msq^2 mst1^2+1182 mg^8 mst1^4-270 mg^6 msq^2 mst1^4-269 mg^6 mst1^6+90 mg^4 msq^2 mst1^6+9 mg^4 mst1^8-1562 mg^10 mst2^2+240 mg^8 msq^2 mst2^2+3564 mg^8 mst1^2 mst2^2+180 mg^6 msq^2 mst1^2 mst2^2-2575 mg^6 mst1^4 mst2^2-90 mg^4 msq^2 mst1^4 mst2^2+498 mg^4 mst1^6 mst2^2+30 mg^2 msq^2 mst1^6 mst2^2+3 mg^2 mst1^8 mst2^2+1182 mg^8 mst2^4-270 mg^6 msq^2 mst2^4-2575 mg^6 mst1^2 mst2^4-90 mg^4 msq^2 mst1^2 mst2^4+1822 mg^4 mst1^4 mst2^4-333 mg^2 mst1^6 mst2^4-269 mg^6 mst2^6+90 mg^4 msq^2 mst2^6+498 mg^4 mst1^2 mst2^6+30 mg^2 msq^2 mst1^2 mst2^6-333 mg^2 mst1^4 mst2^6+48 mst1^6 mst2^6+9 mg^4 mst2^8+3 mg^2 mst1^2 mst2^8))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3)-(128 (40 mg^7-60 mg^5 msq^2-40 mg^5 mst1^2+30 mg^3 msq^2 mst1^2+3 mg^3 mst1^4-40 mg^5 mst2^2+30 mg^3 msq^2 mst2^2+34 mg^3 mst1^2 mst2^2+3 mg^3 mst2^4) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)+((160 (2 mg^12-8 mg^10 mst1^2+3 mg^8 mst1^4-mg^6 mst1^6-8 mg^10 mst2^2+30 mg^8 mst1^2 mst2^2-15 mg^6 mst1^4 mst2^2+5 mg^4 mst1^6 mst2^2+3 mg^8 mst2^4-15 mg^6 mst1^2 mst2^4-mg^6 mst2^6+5 mg^4 mst1^2 mst2^6))/((mg^2-mst1^2)^3 (mg^2-mst2^2)^3)+(1280 (mg^5 mst1^2+mg^5 mst2^2-2 mg^3 mst1^2 mst2^2) Xt)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)) Log[msq^2/MR^2]+(-((16 (128 mg^16-358 mg^14 mst1^2+134 mg^12 mst1^4+125 mg^10 mst1^6-72 mg^8 mst1^8+19 mg^6 mst1^10-410 mg^14 mst2^2+1314 mg^12 mst1^2 mst2^2-683 mg^10 mst1^4 mst2^2-191 mg^8 mst1^6 mst2^2+97 mg^6 mst1^8 mst2^2-31 mg^4 mst1^10 mst2^2+472 mg^12 mst2^4-1797 mg^10 mst1^2 mst2^4+1157 mg^8 mst1^4 mst2^4-11 mg^6 mst1^6 mst2^4+35 mg^4 mst1^8 mst2^4-205 mg^10 mst2^6+999 mg^8 mst1^2 mst2^6-655 mg^6 mst1^4 mst2^6-43 mg^4 mst1^6 mst2^6+27 mg^8 mst2^8-218 mg^6 mst1^2 mst2^8+167 mg^4 mst1^4 mst2^8))/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)))+(128 (34 mg^11-117 mg^9 mst1^2+71 mg^7 mst1^4-4 mg^5 mst1^6-69 mg^9 mst2^2+229 mg^7 mst1^2 mst2^2-121 mg^5 mst1^4 mst2^2-7 mg^3 mst1^6 mst2^2+38 mg^7 mst2^4-129 mg^5 mst1^2 mst2^4+75 mg^3 mst1^4 mst2^4) Xt)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(1024 mg^4 mst1^2 Xt^2)/((mg^2-mst1^2)^2 (mg^2-mst2^2) (-mst1^2+mst2^2))) Log[mst1^2/MR^2]+((16 (128 mg^16-410 mg^14 mst1^2+472 mg^12 mst1^4-205 mg^10 mst1^6+27 mg^8 mst1^8-358 mg^14 mst2^2+1314 mg^12 mst1^2 mst2^2-1797 mg^10 mst1^4 mst2^2+999 mg^8 mst1^6 mst2^2-218 mg^6 mst1^8 mst2^2+134 mg^12 mst2^4-683 mg^10 mst1^2 mst2^4+1157 mg^8 mst1^4 mst2^4-655 mg^6 mst1^6 mst2^4+167 mg^4 mst1^8 mst2^4+125 mg^10 mst2^6-191 mg^8 mst1^2 mst2^6-11 mg^6 mst1^4 mst2^6-43 mg^4 mst1^6 mst2^6-72 mg^8 mst2^8+97 mg^6 mst1^2 mst2^8+35 mg^4 mst1^4 mst2^8+19 mg^6 mst2^10-31 mg^4 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^4 (-mst1^2+mst2^2))-(128 (34 mg^11-69 mg^9 mst1^2+38 mg^7 mst1^4-117 mg^9 mst2^2+229 mg^7 mst1^2 mst2^2-129 mg^5 mst1^4 mst2^2+71 mg^7 mst2^4-121 mg^5 mst1^2 mst2^4+75 mg^3 mst1^4 mst2^4-4 mg^5 mst2^6-7 mg^3 mst1^2 mst2^6) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))-(1024 mg^4 mst2^2 Xt^2)/((mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Log[mst2^2/MR^2])+(-24 (2 mg^2-2 msq^2) (-((40 msq^2 mst1^2)/(3 (-mg^2+mst1^2)^3))+(40 mst1^4)/(3 (-mg^2+mst1^2)^3)+(10 msq^2)/(-mg^2+mst1^2)^2-(10 mst1^2)/(-mg^2+mst1^2)^2-10/(3 (-mg^2+mst1^2))-(40 msq^2 mst2^2)/(3 (-mg^2+mst2^2)^3)+(40 mst2^4)/(3 (-mg^2+mst2^2)^3)+(10 msq^2)/(-mg^2+mst2^2)^2-(10 mst2^2)/(-mg^2+mst2^2)^2-10/(3 (-mg^2+mst2^2)))-(1280 (mg^2-msq^2) (-2 mg^3 msq^2+mg^3 mst1^2+mg msq^2 mst1^2+mg^3 mst2^2+mg msq^2 mst2^2-2 mg mst1^2 mst2^2) Xt)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)) PolyLog[2,1-mg^2/msq^2]+(-((16 (-128 mg^12+294 mg^10 mst1^2-98 mg^8 mst1^4-9 mg^6 mst1^6-20 mg^4 mst1^8+9 mg^2 mst1^10+346 mg^10 mst2^2-902 mg^8 mst1^2 mst2^2+251 mg^6 mst1^4 mst2^2+151 mg^4 mst1^6 mst2^2-41 mg^2 mst1^8 mst2^2+3 mst1^10 mst2^2-280 mg^8 mst2^4+1025 mg^6 mst1^2 mst2^4-417 mg^4 mst1^4 mst2^4-41 mg^2 mst1^6 mst2^4+mst1^8 mst2^4+13 mg^6 mst2^6-391 mg^4 mst1^2 mst2^6+167 mg^2 mst1^4 mst2^6+19 mst1^6 mst2^6+37 mg^4 mst2^8+34 mg^2 mst1^2 mst2^8-23 mst1^4 mst2^8))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)))-(128 (18 mg^5+mg^3 mst1^2+3 mg mst1^4-37 mg^3 mst2^2-7 mg mst1^2 mst2^2+22 mg mst2^4) Xt)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) PolyLog[2,1-mg^2/mst1^2]+((80 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(-mg^2+mst1^2)^3-(1280 mg (-msq^2+mst1^2)^2 Xt)/((mg^2-mst1^2)^2 (mst1^2-mst2^2))) PolyLog[2,1-msq^2/mst1^2]+((16 (-128 mg^12+346 mg^10 mst1^2-280 mg^8 mst1^4+13 mg^6 mst1^6+37 mg^4 mst1^8+294 mg^10 mst2^2-902 mg^8 mst1^2 mst2^2+1025 mg^6 mst1^4 mst2^2-391 mg^4 mst1^6 mst2^2+34 mg^2 mst1^8 mst2^2-98 mg^8 mst2^4+251 mg^6 mst1^2 mst2^4-417 mg^4 mst1^4 mst2^4+167 mg^2 mst1^6 mst2^4-23 mst1^8 mst2^4-9 mg^6 mst2^6+151 mg^4 mst1^2 mst2^6-41 mg^2 mst1^4 mst2^6+19 mst1^6 mst2^6-20 mg^4 mst2^8-41 mg^2 mst1^2 mst2^8+mst1^4 mst2^8+9 mg^2 mst2^10+3 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))-(128 (18 mg^5-37 mg^3 mst1^2+22 mg mst1^4+mg^3 mst2^2-7 mg mst1^2 mst2^2+3 mg mst2^4) Xt)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))) PolyLog[2,1-mg^2/mst2^2]+((80 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(-mg^2+mst2^2)^3-(1280 mg (-msq^2+mst2^2)^2 Xt)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2))) PolyLog[2,1-msq^2/mst2^2]+(-((8 (24 mg^10 mst1^2-28 mg^8 mst1^4+68 mg^6 mst1^6-58 mg^4 mst1^8+18 mg^2 mst1^10-24 mg^10 mst2^2-220 mg^6 mst1^4 mst2^2+188 mg^4 mst1^6 mst2^2-70 mg^2 mst1^8 mst2^2+6 mst1^10 mst2^2+28 mg^8 mst2^4+220 mg^6 mst1^2 mst2^4-16 mg^2 mst1^6 mst2^4+8 mst1^8 mst2^4-68 mg^6 mst2^6-188 mg^4 mst1^2 mst2^6+16 mg^2 mst1^4 mst2^6+58 mg^4 mst2^8+70 mg^2 mst1^2 mst2^8-8 mst1^4 mst2^8-18 mg^2 mst2^10-6 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (-mg^2+mst2^2)^3))+(64 (-10 mg^5 mst1^2+10 mg^3 mst1^4-6 mg mst1^6+10 mg^5 mst2^2+8 mg mst1^4 mst2^2-10 mg^3 mst2^4-8 mg mst1^2 mst2^4+6 mg mst2^6) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)) PolyLog[2,1-mst2^2/mst1^2])+Log[mst2^2/MR^2] ((4 (-4 mg^12 mst1^4+8 mg^10 mst1^6-4 mg^8 mst1^8+611 mg^10 mst1^4 mst2^2-360 mg^8 msq^2 mst1^4 mst2^2-704 mg^8 mst1^6 mst2^2+300 mg^6 msq^2 mst1^6 mst2^2+207 mg^6 mst1^8 mst2^2-180 mg^4 msq^2 mst1^8 mst2^2-18 mg^4 mst1^10 mst2^2+4 mg^12 mst2^4+33 mg^10 mst1^2 mst2^4+360 mg^8 msq^2 mst1^2 mst2^4-2127 mg^8 mst1^4 mst2^4+540 mg^6 msq^2 mst1^4 mst2^4+2217 mg^6 mst1^6 mst2^4-240 mg^4 msq^2 mst1^6 mst2^4-541 mg^4 mst1^8 mst2^4+300 mg^2 msq^2 mst1^8 mst2^4+30 mg^2 mst1^10 mst2^4-140 mg^10 mst2^6+583 mg^8 mst1^2 mst2^6-840 mg^6 msq^2 mst1^2 mst2^6+1975 mg^6 mst1^4 mst2^6-2359 mg^4 mst1^6 mst2^6-720 mg^2 msq^2 mst1^6 mst2^6+505 mg^2 mst1^8 mst2^6+120 msq^2 mst1^8 mst2^6+12 mst1^10 mst2^6+204 mg^8 mst2^8-1259 mg^6 mst1^2 mst2^8+420 mg^4 msq^2 mst1^2 mst2^8+283 mg^4 mst1^4 mst2^8+600 mg^2 msq^2 mst1^4 mst2^8+571 mg^2 mst1^6 mst2^8-60 msq^2 mst1^6 mst2^8-183 mst1^8 mst2^8-68 mg^6 mst2^10+587 mg^4 mst1^2 mst2^10-180 mg^2 msq^2 mst1^2 mst2^10-594 mg^2 mst1^4 mst2^10-60 msq^2 mst1^4 mst2^10+171 mst1^6 mst2^10))/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+(64 (22 mg^7 mst1^2-20 mg^5 mst1^4+30 mg^3 msq^2 mst1^4+3 mg^3 mst1^6+16 mg^7 mst2^2-112 mg^5 mst1^2 mst2^2+100 mg^3 mst1^4 mst2^2-60 mg msq^2 mst1^4 mst2^2-6 mg mst1^6 mst2^2-16 mg^5 mst2^4+87 mg^3 mst1^2 mst2^4+30 mg msq^2 mst1^2 mst2^4-74 mg mst1^4 mst2^4) Xt)/(3 mst1^2 (mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2))-(8 (-4 mg^10 mst1^2+8 mg^8 mst1^4-4 mg^6 mst1^6+4 mg^10 mst2^2+695 mg^8 mst1^2 mst2^2-360 mg^6 msq^2 mst1^2 mst2^2-886 mg^6 mst1^4 mst2^2+300 mg^4 msq^2 mst1^4 mst2^2+301 mg^4 mst1^6 mst2^2-180 mg^2 msq^2 mst1^6 mst2^2-18 mg^2 mst1^8 mst2^2-136 mg^8 mst2^4-326 mg^6 mst1^2 mst2^4+300 mg^4 msq^2 mst1^2 mst2^4+40 mg^4 mst1^4 mst2^4+240 mg^2 msq^2 mst1^4 mst2^4+212 mg^2 mst1^6 mst2^4-60 msq^2 mst1^6 mst2^4-6 mst1^8 mst2^4+68 mg^6 mst2^6+5 mg^4 mst1^2 mst2^6-180 mg^2 msq^2 mst1^2 mst2^6+308 mg^2 mst1^4 mst2^6-60 msq^2 mst1^4 mst2^6-237 mst1^6 mst2^6-18 mg^2 mst1^2 mst2^8-6 mst1^4 mst2^8) Xt^2)/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+((128 (14 mg^3-30 mg msq^2-3 mg mst1^2-3 mg mst2^2))/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2))+(128 mg (mst1^2+3 mst2^2) (-11 mg^4+3 mg^2 mst1^2+3 mg^2 mst2^2+5 mst1^2 mst2^2))/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^3)-(256 (-7 mg^5 mst1^2+8 mg^5 mst2^2-26 mg^3 mst1^2 mst2^2-8 mg^3 mst2^4+17 mg mst1^2 mst2^4))/(3 mst1^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)) Xt^3+(4 (-4 mg^12 mst1^4+8 mg^10 mst1^6-4 mg^8 mst1^8+176 mg^12 mst1^2 mst2^2+667 mg^10 mst1^4 mst2^2-360 mg^8 msq^2 mst1^4 mst2^2-1222 mg^8 mst1^6 mst2^2+300 mg^6 msq^2 mst1^6 mst2^2+493 mg^6 mst1^8 mst2^2-180 mg^4 msq^2 mst1^8 mst2^2-18 mg^4 mst1^10 mst2^2+4 mg^12 mst2^4-4793 mg^10 mst1^2 mst2^4-360 mg^8 msq^2 mst1^2 mst2^4+10101 mg^8 mst1^4 mst2^4+600 mg^6 msq^2 mst1^4 mst2^4-6913 mg^6 mst1^6 mst2^4+480 mg^4 msq^2 mst1^6 mst2^4+1361 mg^4 mst1^8 mst2^4-240 mg^2 msq^2 mst1^8 mst2^4-24 mg^2 mst1^10 mst2^4-140 mg^10 mst2^6+7283 mg^8 mst1^2 mst2^6+1020 mg^6 msq^2 mst1^2 mst2^6-17667 mg^6 mst1^4 mst2^6-1380 mg^4 msq^2 mst1^4 mst2^6+13435 mg^4 mst1^6 mst2^6+420 mg^2 msq^2 mst1^6 mst2^6-2665 mg^2 mst1^8 mst2^6-60 msq^2 mst1^8 mst2^6-6 mst1^10 mst2^6+204 mg^8 mst2^8-1881 mg^6 mst1^2 mst2^8-360 mg^4 msq^2 mst1^2 mst2^8+6073 mg^4 mst1^4 mst2^8-360 mg^2 msq^2 mst1^4 mst2^8-5659 mg^2 mst1^6 mst2^8+240 msq^2 mst1^6 mst2^8+1167 mst1^8 mst2^8-68 mg^6 mst2^10-975 mg^4 mst1^2 mst2^10+180 mg^2 msq^2 mst1^2 mst2^10+1508 mg^2 mst1^4 mst2^10+60 msq^2 mst1^4 mst2^10-417 mst1^6 mst2^10-18 mg^2 mst1^2 mst2^12-6 mst1^4 mst2^12) Xt^4)/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)+(64 (48 mg^7 mst1^2-62 mg^5 mst1^4+30 mg^3 msq^2 mst1^4+3 mg^3 mst1^6+16 mg^7 mst2^2+172 mg^5 mst1^2 mst2^2-30 mg^3 msq^2 mst1^2 mst2^2-166 mg^3 mst1^4 mst2^2+30 mg msq^2 mst1^4 mst2^2+3 mg mst1^6 mst2^2-16 mg^5 mst2^4-185 mg^3 mst1^2 mst2^4-30 mg msq^2 mst1^2 mst2^4+184 mg mst1^4 mst2^4+3 mg mst1^2 mst2^6) Xt^5)/(3 mst1^2 (mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)+((80 (18 mg^10+12 mg^8 msq^2+9 mg^6 msq^4-35 mg^8 mst1^2-3 mg^6 msq^2 mst1^2-18 mg^4 msq^4 mst1^2+17 mg^6 mst1^4+3 mg^4 msq^2 mst1^4+9 mg^2 msq^4 mst1^4-47 mg^8 mst2^2-51 mg^6 msq^2 mst2^2+3 mg^4 msq^4 mst2^2+91 mg^6 mst1^2 mst2^2+39 mg^4 msq^2 mst1^2 mst2^2-6 mg^2 msq^4 mst1^2 mst2^2-44 mg^4 mst1^4 mst2^2-24 mg^2 msq^2 mst1^4 mst2^2+3 msq^4 mst1^4 mst2^2+43 mg^6 mst2^4+24 mg^4 msq^2 mst2^4-83 mg^4 mst1^2 mst2^4+15 mg^2 msq^2 mst1^2 mst2^4+40 mg^2 mst1^4 mst2^4-3 msq^2 mst1^4 mst2^4-14 mg^4 mst2^6-9 mg^2 msq^2 mst2^6+27 mg^2 mst1^2 mst2^6-3 msq^2 mst1^2 mst2^6-13 mst1^4 mst2^6))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3)+(-((640 mg msq^2)/((mg^2-mst1^2) (mg^2-mst2^2)))-(640 mg)/(3 (-mst1^2+mst2^2))-(320 mg mst2^2)/(3 (mg^2-mst2^2) (-mst1^2+mst2^2))+(320 (-6 mg msq^4+mg^3 mst2^2+12 mg msq^2 mst2^2-mg mst2^4))/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt+(160 (18 mg^8+18 mg^6 msq^2-35 mg^6 mst1^2-15 mg^4 msq^2 mst1^2+17 mg^4 mst1^4+9 mg^2 msq^2 mst1^4-33 mg^6 mst2^2-15 mg^4 msq^2 mst2^2+64 mg^4 mst1^2 mst2^2-12 mg^2 msq^2 mst1^2 mst2^2-31 mg^2 mst1^4 mst2^2+3 msq^2 mst1^4 mst2^2+15 mg^4 mst2^4+9 mg^2 msq^2 mst2^4-29 mg^2 mst1^2 mst2^4+3 msq^2 mst1^2 mst2^4+14 mst1^4 mst2^4) Xt^2)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(-((1280 mg msq^2)/((mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)))+(640 mg (mst1^2+3 mst2^2))/(3 (-mst1^2+mst2^2)^3)) Xt^3+((160 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst2^2-msq^4 mst2^2))/((mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(80 (mst1^2+mst2^2) (11 mg^8+18 mg^6 msq^2-22 mg^6 mst1^2-15 mg^4 msq^2 mst1^2+11 mg^4 mst1^4+9 mg^2 msq^2 mst1^4-22 mg^6 mst2^2-15 mg^4 msq^2 mst2^2+44 mg^4 mst1^2 mst2^2-12 mg^2 msq^2 mst1^2 mst2^2-22 mg^2 mst1^4 mst2^2+3 msq^2 mst1^4 mst2^2+11 mg^4 mst2^4+9 mg^2 msq^2 mst2^4-22 mg^2 mst1^2 mst2^4+3 msq^2 mst1^2 mst2^4+11 mst1^4 mst2^4))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)-(80 (4 mg^8-3 mst1^2 mst2^4 (mst1^2+5 mst2^2)+mg^6 (3 mst1^2+7 mst2^2)-mg^4 (6 mst1^4+15 mst1^2 mst2^2+29 mst2^4)+mg^2 (7 mst1^4 mst2^2+31 mst1^2 mst2^4+16 mst2^6)))/(3 (mg^2-mst1^2) (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)) Xt^4+(-((640 mg mst2^2)/(3 (mg^2-mst2^2) (-mst1^2+mst2^2)^3))+(640 mg msq^2 (mst1^2+mst2^2))/((mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^3)-(1280 (-mg msq^4+2 mg msq^2 mst2^2))/((mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)) Xt^5) Log[msq^2/MR^2]+(-24 (-((5 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(6 (-mg^2+mst1^2)^3))-(5 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(6 (-mg^2+mst2^2)^3))-24 ((40 mg (-msq^2+mst1^2)^2)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+(40 mg (-msq^2+mst2^2)^2)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt-1/(-mst1^2+mst2^2) 48 (-((5 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(6 (-mg^2+mst1^2)^3))-(5 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(6 (-mg^2+mst2^2)^3)) Xt^2-(48 ((40 mg (-msq^2+mst1^2)^2)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+(40 mg (-msq^2+mst2^2)^2)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt^3)/(-mst1^2+mst2^2)+1/(-mst1^2+mst2^2)^3 24 (mst1^2+mst2^2) (-((5 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(6 (-mg^2+mst1^2)^3))-(5 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(6 (-mg^2+mst2^2)^3)) Xt^4+(24 (mst1^2+mst2^2) ((40 mg (-msq^2+mst1^2)^2)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+(40 mg (-msq^2+mst2^2)^2)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt^5)/(-mst1^2+mst2^2)^3) Log[msq^2/MR^2]^2+(-((8 (348 mg^12-276 mg^10 mst1^2-90 mg^8 msq^2 mst1^2-187 mg^8 mst1^4-30 mg^6 msq^2 mst1^4+158 mg^6 mst1^6+9 mg^4 mst1^8-264 mg^10 mst2^2-90 mg^8 msq^2 mst2^2-704 mg^8 mst1^2 mst2^2+540 mg^6 msq^2 mst1^2 mst2^2+1608 mg^6 mst1^4 mst2^2-180 mg^4 msq^2 mst1^4 mst2^2-787 mg^4 mst1^6 mst2^2+90 mg^2 msq^2 mst1^6 mst2^2+3 mg^2 mst1^8 mst2^2-201 mg^8 mst2^4-30 mg^6 msq^2 mst2^4+1498 mg^6 mst1^2 mst2^4-180 mg^4 msq^2 mst1^2 mst2^4-2017 mg^4 mst1^4 mst2^4-180 mg^2 msq^2 mst1^4 mst2^4+828 mg^2 mst1^6 mst2^4+30 msq^2 mst1^6 mst2^4+192 mg^6 mst2^6-693 mg^4 mst1^2 mst2^6+90 mg^2 msq^2 mst1^2 mst2^6+820 mg^2 mst1^4 mst2^6+30 msq^2 mst1^4 mst2^6-299 mst1^6 mst2^6-20 mg^4 mst2^8-32 mg^2 mst1^2 mst2^8+4 mst1^4 mst2^8+9 mg^2 mst2^10+3 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3))+(64 (109 mg^7 mst1^4-30 mg^5 msq^2 mst1^4-109 mg^5 mst1^6+3 mg^3 mst1^8-234 mg^7 mst1^2 mst2^2+60 mg^5 msq^2 mst1^2 mst2^2+138 mg^5 mst1^4 mst2^2+50 mg^3 mst1^6 mst2^2+30 mg msq^2 mst1^6 mst2^2+109 mg^7 mst2^4-30 mg^5 msq^2 mst2^4+123 mg^5 mst1^2 mst2^4-159 mg^3 mst1^4 mst2^4-60 mg msq^2 mst1^4 mst2^4+4 mg mst1^6 mst2^4-104 mg^5 mst2^6+60 mg^3 mst1^2 mst2^6+30 mg msq^2 mst1^2 mst2^6+19 mg mst1^4 mst2^6-2 mg^3 mst2^8-10 mg mst1^2 mst2^8+3 mg mst2^10) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)+(16 (468 mg^12 mst1^2-402 mg^10 mst1^4-90 mg^8 msq^2 mst1^4-292 mg^8 mst1^6-30 mg^6 msq^2 mst1^6+269 mg^6 mst1^8+9 mg^4 mst1^10+468 mg^12 mst2^2-3692 mg^10 mst1^2 mst2^2+180 mg^8 msq^2 mst1^2 mst2^2+4624 mg^8 mst1^4 mst2^2+30 mg^6 msq^2 mst1^4 mst2^2-1268 mg^6 mst1^6 mst2^2+360 mg^4 msq^2 mst1^6 mst2^2-337 mg^4 mst1^8 mst2^2-90 mg^2 msq^2 mst1^8 mst2^2-15 mg^2 mst1^10 mst2^2-402 mg^10 mst2^4-90 mg^8 msq^2 mst2^4+4624 mg^8 mst1^2 mst2^4+30 mg^6 msq^2 mst1^2 mst2^4-6418 mg^6 mst1^4 mst2^4-720 mg^4 msq^2 mst1^4 mst2^4+2420 mg^4 mst1^6 mst2^4+90 mg^2 msq^2 mst1^6 mst2^4+142 mg^2 mst1^8 mst2^4-30 msq^2 mst1^8 mst2^4-6 mst1^10 mst2^4-292 mg^8 mst2^6-30 mg^6 msq^2 mst2^6-1268 mg^6 mst1^2 mst2^6+360 mg^4 msq^2 mst1^2 mst2^6+2420 mg^4 mst1^4 mst2^6+90 mg^2 msq^2 mst1^4 mst2^6-1166 mg^2 mst1^6 mst2^6+60 msq^2 mst1^6 mst2^6+26 mst1^8 mst2^6+269 mg^6 mst2^8-337 mg^4 mst1^2 mst2^8-90 mg^2 msq^2 mst1^2 mst2^8+142 mg^2 mst1^4 mst2^8-30 msq^2 mst1^4 mst2^8+26 mst1^6 mst2^8+9 mg^4 mst2^10-15 mg^2 mst1^2 mst2^10-6 mst1^4 mst2^10) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(128 (31 mg^9+63 mg^7 mst1^2-30 mg^5 msq^2 mst1^2-86 mg^5 mst1^4+3 mg^3 mst1^6+63 mg^7 mst2^2-30 mg^5 msq^2 mst2^2-424 mg^5 mst1^2 mst2^2+120 mg^3 msq^2 mst1^2 mst2^2+310 mg^3 mst1^4 mst2^2-30 mg msq^2 mst1^4 mst2^2-6 mg mst1^6 mst2^2-86 mg^5 mst2^4+310 mg^3 mst1^2 mst2^4-30 mg msq^2 mst1^2 mst2^4-175 mg mst1^4 mst2^4+3 mg^3 mst2^6-6 mg mst1^2 mst2^6) Xt^3)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)-(8 (816 mg^12 mst1^4-1036 mg^10 mst1^6-90 mg^8 msq^2 mst1^6-6 mg^8 mst1^8-30 mg^6 msq^2 mst1^8+269 mg^6 mst1^10+9 mg^4 mst1^12+3136 mg^12 mst1^2 mst2^2-10988 mg^10 mst1^4 mst2^2+90 mg^8 msq^2 mst1^4 mst2^2+11772 mg^8 mst1^6 mst2^2-4191 mg^6 mst1^8 mst2^2+360 mg^4 msq^2 mst1^8 mst2^2+70 mg^4 mst1^10 mst2^2-90 mg^2 msq^2 mst1^10 mst2^2-15 mg^2 mst1^12 mst2^2+816 mg^12 mst2^4-10916 mg^10 mst1^2 mst2^4+90 mg^8 msq^2 mst1^2 mst2^4+24976 mg^8 mst1^4 mst2^4+60 mg^6 msq^2 mst1^4 mst2^4-20498 mg^6 mst1^6 mst2^4-360 mg^4 msq^2 mst1^6 mst2^4+6465 mg^4 mst1^8 mst2^4-433 mg^2 mst1^10 mst2^4-30 msq^2 mst1^10 mst2^4-6 mst1^12 mst2^4-1060 mg^10 mst2^6-90 mg^8 msq^2 mst2^6+11716 mg^8 mst1^2 mst2^6-21226 mg^6 mst1^4 mst2^6-360 mg^4 msq^2 mst1^4 mst2^6+12164 mg^4 mst1^6 mst2^6+180 mg^2 msq^2 mst1^6 mst2^6-2320 mg^2 mst1^8 mst2^6+30 msq^2 mst1^8 mst2^6+206 mst1^10 mst2^6+22 mg^8 mst2^8-30 mg^6 msq^2 mst2^8-3835 mg^6 mst1^2 mst2^8+360 mg^4 msq^2 mst1^2 mst2^8+6899 mg^4 mst1^4 mst2^8-2298 mg^2 mst1^6 mst2^8+30 msq^2 mst1^6 mst2^8-248 mst1^8 mst2^8+201 mg^6 mst2^10-234 mg^4 mst1^2 mst2^10-90 mg^2 msq^2 mst1^2 mst2^10-575 mg^2 mst1^4 mst2^10-30 msq^2 mst1^4 mst2^10+216 mst1^6 mst2^10+67 mg^4 mst2^12+91 mg^2 mst1^2 mst2^12-2 mst1^4 mst2^12-18 mg^2 mst2^14-6 mst1^2 mst2^14) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (mst1^2-mst2^2)^4)+(64 (64 mg^9 mst1^2+44 mg^7 mst1^4-30 mg^5 msq^2 mst1^4-92 mg^5 mst1^6+3 mg^3 mst1^8+64 mg^9 mst2^2+188 mg^7 mst1^2 mst2^2-60 mg^5 msq^2 mst1^2 mst2^2-668 mg^5 mst1^4 mst2^2+120 mg^3 msq^2 mst1^4 mst2^2+372 mg^3 mst1^6 mst2^2-30 mg msq^2 mst1^6 mst2^2-6 mg mst1^8 mst2^2+44 mg^7 mst2^4-30 mg^5 msq^2 mst2^4-638 mg^5 mst1^2 mst2^4+120 mg^3 msq^2 mst1^2 mst2^4+920 mg^3 mst1^4 mst2^4-60 mg msq^2 mst1^4 mst2^4-268 mg mst1^6 mst2^4-102 mg^5 mst2^6+352 mg^3 mst1^2 mst2^6-30 mg msq^2 mst1^2 mst2^6-298 mg mst1^4 mst2^6+13 mg^3 mst2^8+14 mg mst1^2 mst2^8-6 mg mst2^10) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)+(5120 mg^2 mst1^2 mst2^2 Xt^6)/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^4)+((80 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst1^2-msq^4 mst1^2))/(mg^2-mst1^2)^3+(80 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst2^2-msq^4 mst2^2))/(mg^2-mst2^2)^3+(80 (-2 mst1^4 mst2^4+2 mg^6 (mst1^2+mst2^2)+4 mg^2 mst1^2 mst2^2 (mst1^2+mst2^2)-mg^4 (mst1^4+8 mst1^2 mst2^2+mst2^4)))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)-(320 (mg^7+12 mg^5 msq^2-12 mg^3 msq^4-mg^5 mst1^2+6 mg msq^4 mst1^2-mg^5 mst2^2+6 mg msq^4 mst2^2+mg^3 mst1^2 mst2^2-12 mg msq^2 mst1^2 mst2^2) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)+(-((160 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst1^2-msq^4 mst1^2))/((mg^2-mst1^2)^3 (mst1^2-mst2^2)))-(160 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst2^2-msq^4 mst2^2))/((mg^2-mst2^2)^3 (-mst1^2+mst2^2))+(160 (2 mg^8 (mst1^2+mst2^2)+2 mst1^4 mst2^4 (mst1^2+mst2^2)-4 mg^2 mst1^2 mst2^2 (mst1^2+mst2^2)^2+3 mg^4 (mst1^2+mst2^2)^3-2 mg^6 (3 mst1^4+2 mst1^2 mst2^2+3 mst2^4)))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)) Xt^2+((1280 (-mg msq^4+2 mg msq^2 mst1^2))/((mg^2-mst1^2)^2 (mst1^2-mst2^2)^2)+(1280 (-mg msq^4+2 mg msq^2 mst2^2))/((mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)+(640 (-2 mg mst1^2 mst2^2+mg^3 (mst1^2+mst2^2)))/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^2)) Xt^3+((80 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst1^2-msq^4 mst1^2) (mst1^2+mst2^2))/((mg^2-mst1^2)^3 (mst1^2-mst2^2)^3)+(80 (mst1^2+mst2^2) (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst2^2-msq^4 mst2^2))/((mg^2-mst2^2)^3 (-mst1^2+mst2^2)^3)-(80 (mst1^2+mst2^2) (4 mg^8 (mst1^2+mst2^2)+4 mst1^4 mst2^4 (mst1^2+mst2^2)-8 mg^2 mst1^2 mst2^2 (mst1^2+mst2^2)^2-2 mg^6 (5 mst1^4+6 mst1^2 mst2^2+5 mst2^4)+mg^4 (5 mst1^6+19 mst1^4 mst2^2+19 mst1^2 mst2^4+5 mst2^6)))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)) Xt^4+(-((640 (-mg msq^4+2 mg msq^2 mst1^2) (mst1^2+mst2^2))/((mg^2-mst1^2)^2 (mst1^2-mst2^2)^4))-(640 (mst1^2+mst2^2) (-mg msq^4+2 mg msq^2 mst2^2))/((mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)-(320 mg (mst1^2+mst2^2) (-2 mst1^2 mst2^2+mg^2 (mst1^2+mst2^2)))/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^4)) Xt^5) Log[msq^2/MR^2]) Log[mst1^2/MR^2]+(-((8 (3 mg^4-4 mg^2 mst1^2+2 mst1^4) (mg^4-34 mg^2 mst2^2+17 mst2^4))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2))-(4 (128 mg^10-556 mg^8 mst1^2+60 mg^6 msq^2 mst1^2-90 mg^4 msq^4 mst1^2+706 mg^6 mst1^4+120 mg^4 msq^2 mst1^4+60 mg^2 msq^4 mst1^4-437 mg^4 mst1^6-180 mg^2 msq^2 mst1^6+30 msq^4 mst1^6+136 mg^2 mst1^8-mst1^10+44 mg^8 mst2^2-60 mg^6 msq^2 mst2^2+90 mg^4 msq^4 mst2^2+64 mg^6 mst1^2 mst2^2-120 mg^4 msq^2 mst1^2 mst2^2-60 mg^2 msq^4 mst1^2 mst2^2-51 mg^4 mst1^4 mst2^2+180 mg^2 msq^2 mst1^4 mst2^2-30 msq^4 mst1^4 mst2^2-38 mg^2 mst1^6 mst2^2+5 mst1^8 mst2^2-2 mg^6 mst2^4-33 mg^4 mst1^2 mst2^4+36 mg^2 mst1^4 mst2^4-mst1^6 mst2^4+9 mg^4 mst2^6-6 mg^2 mst1^2 mst2^6-3 mst1^4 mst2^6))/(3 (mg^2-mst1^2)^4 (mst1^2-mst2^2))+(8 (12 mg^10 mst2^2+64 mg^8 mst1^2 mst2^2-78 mg^6 mst1^4 mst2^2+26 mg^4 mst1^6 mst2^2-14 mg^8 mst2^4-188 mg^6 mst1^2 mst2^4+175 mg^4 mst1^4 mst2^4-57 mg^2 mst1^6 mst2^4+34 mg^6 mst2^6+120 mg^4 mst1^2 mst2^6-65 mg^2 mst1^4 mst2^6+19 mst1^6 mst2^6-29 mg^4 mst2^8-35 mg^2 mst1^2 mst2^8+4 mst1^4 mst2^8+9 mg^2 mst2^10+3 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3)+(8 (mg^4-34 mg^2 mst1^2+17 mst1^4) (-2 mst1^4 mst2^4+2 mg^6 (mst1^2+mst2^2)+4 mg^2 mst1^2 mst2^2 (mst1^2+mst2^2)-mg^4 (mst1^4+8 mst1^2 mst2^2+mst2^4)))/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^2)-(32 (-20 mg^11 mst1^2+30 mg^7 msq^4 mst1^2+91 mg^9 mst1^4-60 mg^7 msq^2 mst1^4-30 mg^5 msq^4 mst1^4-127 mg^7 mst1^6+60 mg^5 msq^2 mst1^6+40 mg^5 mst1^8+20 mg^11 mst2^2-30 mg^7 msq^4 mst2^2-51 mg^9 mst1^2 mst2^2+60 mg^7 msq^2 mst1^2 mst2^2-30 mg^5 msq^4 mst1^2 mst2^2-50 mg^7 mst1^4 mst2^2+60 mg^5 msq^2 mst1^4 mst2^2+60 mg^3 msq^4 mst1^4 mst2^2+194 mg^5 mst1^6 mst2^2-120 mg^3 msq^2 mst1^6 mst2^2-65 mg^3 mst1^8 mst2^2-56 mg^9 mst2^4+60 mg^5 msq^4 mst2^4+229 mg^7 mst1^2 mst2^4-120 mg^5 msq^2 mst1^2 mst2^4-30 mg^3 msq^4 mst1^2 mst2^4-263 mg^5 mst1^4 mst2^4+60 mg^3 msq^2 mst1^4 mst2^4-30 mg msq^4 mst1^4 mst2^4+27 mg^3 mst1^6 mst2^4+60 mg msq^2 mst1^6 mst2^4+15 mg mst1^8 mst2^4+12 mg^7 mst2^6-30 mg^3 msq^4 mst2^6-62 mg^5 mst1^2 mst2^6+60 mg^3 msq^2 mst1^2 mst2^6+30 mg msq^4 mst1^2 mst2^6+104 mg^3 mst1^4 mst2^6-60 mg msq^2 mst1^4 mst2^6-38 mg mst1^6 mst2^6-5 mg^5 mst2^8-5 mg^3 mst1^2 mst2^8+10 mg mst1^4 mst2^8+3 mg^3 mst2^10-3 mg mst1^2 mst2^10) Xt)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)+(8 (128 mg^16-558 mg^14 mst1^2+60 mg^12 msq^2 mst1^2-90 mg^10 msq^4 mst1^2+426 mg^12 mst1^4+120 mg^10 msq^2 mst1^4+60 mg^8 msq^4 mst1^4+319 mg^10 mst1^6-180 mg^8 msq^2 mst1^6+30 mg^6 msq^4 mst1^6-400 mg^8 mst1^8+93 mg^6 mst1^10-846 mg^14 mst2^2-60 mg^12 msq^2 mst2^2+90 mg^10 msq^4 mst2^2+3662 mg^12 mst1^2 mst2^2-300 mg^10 msq^2 mst1^2 mst2^2+210 mg^8 msq^4 mst1^2 mst2^2-4533 mg^10 mst1^4 mst2^2-180 mg^8 msq^2 mst1^4 mst2^2-210 mg^6 msq^4 mst1^4 mst2^2+1857 mg^8 mst1^6 mst2^2+540 mg^6 msq^2 mst1^6 mst2^2-90 mg^4 msq^4 mst1^6 mst2^2-165 mg^6 mst1^8 mst2^2-7 mg^4 mst1^10 mst2^2+1164 mg^12 mst2^4+180 mg^10 msq^2 mst2^4-270 mg^8 msq^4 mst2^4-5053 mg^10 mst1^2 mst2^4+540 mg^8 msq^2 mst1^2 mst2^4-90 mg^6 msq^4 mst1^2 mst2^4+6223 mg^8 mst1^4 mst2^4-180 mg^6 msq^2 mst1^4 mst2^4+270 mg^4 msq^4 mst1^4 mst2^4-2840 mg^6 mst1^6 mst2^4-540 mg^4 msq^2 mst1^6 mst2^4+90 mg^2 msq^4 mst1^6 mst2^4+647 mg^4 mst1^8 mst2^4-93 mg^2 mst1^10 mst2^4-345 mg^10 mst2^6-180 mg^8 msq^2 mst2^6+270 mg^6 msq^4 mst2^6+1965 mg^8 mst1^2 mst2^6-420 mg^6 msq^2 mst1^2 mst2^6-90 mg^4 msq^4 mst1^2 mst2^6-2004 mg^6 mst1^4 mst2^6+420 mg^4 msq^2 mst1^4 mst2^6-150 mg^2 msq^4 mst1^4 mst2^6+412 mg^4 mst1^6 mst2^6+180 mg^2 msq^2 mst1^6 mst2^6-30 msq^4 mst1^6 mst2^6-91 mg^2 mst1^8 mst2^6+31 mst1^10 mst2^6-185 mg^8 mst2^8+60 mg^6 msq^2 mst2^8-90 mg^4 msq^4 mst2^8+41 mg^6 mst1^2 mst2^8+120 mg^4 msq^2 mst1^2 mst2^8+60 mg^2 msq^4 mst1^2 mst2^8-226 mg^4 mst1^4 mst2^8-180 mg^2 msq^2 mst1^4 mst2^8+30 msq^4 mst1^4 mst2^8+489 mg^2 mst1^6 mst2^8-111 mst1^8 mst2^8+87 mg^6 mst2^10+45 mg^4 mst1^2 mst2^10-135 mg^2 mst1^4 mst2^10+3 mst1^6 mst2^10-27 mg^4 mst2^12+18 mg^2 mst1^2 mst2^12+9 mst1^4 mst2^12) Xt^2)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(32 (2 mg^13-43 mg^11 mst1^2+60 mg^7 msq^4 mst1^2+243 mg^9 mst1^4-120 mg^7 msq^2 mst1^4-60 mg^5 msq^4 mst1^4-375 mg^7 mst1^6+120 mg^5 msq^2 mst1^6+141 mg^5 mst1^8-33 mg^11 mst2^2-60 mg^7 msq^4 mst2^2+403 mg^9 mst1^2 mst2^2+120 mg^7 msq^2 mst1^2 mst2^2-60 mg^5 msq^4 mst1^2 mst2^2-1217 mg^7 mst1^4 mst2^2+120 mg^5 msq^2 mst1^4 mst2^2+120 mg^3 msq^4 mst1^4 mst2^2+1351 mg^5 mst1^6 mst2^2-240 mg^3 msq^2 mst1^6 mst2^2-408 mg^3 mst1^8 mst2^2-40 mg^9 mst2^4+120 mg^5 msq^4 mst2^4-143 mg^7 mst1^2 mst2^4-240 mg^5 msq^2 mst1^2 mst2^4-60 mg^3 msq^4 mst1^2 mst2^4+869 mg^5 mst1^4 mst2^4+120 mg^3 msq^2 mst1^4 mst2^4-60 mg msq^4 mst1^4 mst2^4-1053 mg^3 mst1^6 mst2^4+120 mg msq^2 mst1^6 mst2^4+271 mg mst1^8 mst2^4+31 mg^7 mst2^6-60 mg^3 msq^4 mst2^6-185 mg^5 mst1^2 mst2^6+120 mg^3 msq^2 mst1^2 mst2^6+60 mg msq^4 mst1^2 mst2^6+93 mg^3 mst1^4 mst2^6-120 mg msq^2 mst1^4 mst2^6+93 mg mst1^6 mst2^6+30 mg^5 mst2^8+30 mg^3 mst1^2 mst2^8-60 mg mst1^4 mst2^8-18 mg^3 mst2^10+18 mg mst1^2 mst2^10) Xt^3)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)-(4 (124 mg^16 mst1^2-540 mg^14 mst1^4+60 mg^12 msq^2 mst1^4-90 mg^10 msq^4 mst1^4-214 mg^12 mst1^6+120 mg^10 msq^2 mst1^6+60 mg^8 msq^4 mst1^6+1627 mg^10 mst1^8-180 mg^8 msq^2 mst1^8+30 mg^6 msq^4 mst1^8-1148 mg^8 mst1^10+159 mg^6 mst1^12+1156 mg^16 mst2^2-5372 mg^14 mst1^2 mst2^2+9194 mg^12 mst1^4 mst2^2-180 mg^10 msq^2 mst1^4 mst2^2+270 mg^8 msq^4 mst1^4 mst2^2-3566 mg^10 mst1^6 mst2^2-360 mg^8 msq^2 mst1^6 mst2^2-180 mg^6 msq^4 mst1^6 mst2^2-4559 mg^8 mst1^8 mst2^2+540 mg^6 msq^2 mst1^8 mst2^2-90 mg^4 msq^4 mst1^8 mst2^2+3264 mg^6 mst1^10 mst2^2-141 mg^4 mst1^12 mst2^2-3032 mg^14 mst2^4-60 mg^12 msq^2 mst2^4+90 mg^10 msq^4 mst2^4+14826 mg^12 mst1^2 mst2^4-120 mg^10 msq^2 mst1^2 mst2^4-60 mg^8 msq^4 mst1^2 mst2^4-25992 mg^10 mst1^4 mst2^4+360 mg^8 msq^2 mst1^4 mst2^4-300 mg^6 msq^4 mst1^4 mst2^4+15344 mg^8 mst1^6 mst2^4+360 mg^6 msq^2 mst1^6 mst2^4+180 mg^4 msq^4 mst1^6 mst2^4+2717 mg^6 mst1^8 mst2^4-540 mg^4 msq^2 mst1^8 mst2^4+90 mg^2 msq^4 mst1^8 mst2^4-3856 mg^4 mst1^10 mst2^4+9 mg^2 mst1^12 mst2^4+2450 mg^12 mst2^6+180 mg^10 msq^2 mst2^6-270 mg^8 msq^4 mst2^6-13214 mg^10 mst1^2 mst2^6+360 mg^8 msq^2 mst1^2 mst2^6+180 mg^6 msq^4 mst1^2 mst2^6+24798 mg^8 mst1^4 mst2^6-600 mg^6 msq^2 mst1^4 mst2^6+180 mg^4 msq^4 mst1^4 mst2^6-17852 mg^6 mst1^6 mst2^6-120 mg^4 msq^2 mst1^6 mst2^6-60 mg^2 msq^4 mst1^6 mst2^6+2045 mg^4 mst1^8 mst2^6+180 mg^2 msq^2 mst1^8 mst2^6-30 msq^4 mst1^8 mst2^6+1792 mg^2 mst1^10 mst2^6-3 mst1^12 mst2^6-503 mg^10 mst2^8-180 mg^8 msq^2 mst2^8+270 mg^6 msq^4 mst2^8+3560 mg^8 mst1^2 mst2^8-360 mg^6 msq^2 mst1^2 mst2^8-180 mg^4 msq^4 mst1^2 mst2^8-7151 mg^6 mst1^4 mst2^8+540 mg^4 msq^2 mst1^4 mst2^8-90 mg^2 msq^4 mst1^4 mst2^8+5786 mg^4 mst1^6 mst2^8-1568 mg^2 mst1^8 mst2^8-148 mst1^10 mst2^8-155 mg^8 mst2^10+60 mg^6 msq^2 mst2^10-90 mg^4 msq^4 mst2^10+136 mg^6 mst1^2 mst2^10+120 mg^4 msq^2 mst1^2 mst2^10+60 mg^2 msq^4 mst1^2 mst2^10-321 mg^4 mst1^4 mst2^10-180 mg^2 msq^2 mst1^4 mst2^10+30 msq^4 mst1^4 mst2^10+490 mg^2 mst1^6 mst2^10-142 mst1^8 mst2^10+87 mg^6 mst2^12+18 mg^4 mst1^2 mst2^12-117 mg^2 mst1^4 mst2^12+12 mst1^6 mst2^12-27 mg^4 mst2^14+18 mg^2 mst1^2 mst2^14+9 mst1^4 mst2^14) Xt^4)/(3 (mg^2-mst1^2)^4 (mg^2-mst2^2)^3 (mst1^2-mst2^2)^4)-(32 (18 mg^11 mst1^4-30 mg^7 msq^4 mst1^4-149 mg^9 mst1^6+60 mg^7 msq^2 mst1^6+30 mg^5 msq^4 mst1^6+249 mg^7 mst1^8-60 mg^5 msq^2 mst1^8-102 mg^5 mst1^10-32 mg^11 mst1^2 mst2^2-182 mg^9 mst1^4 mst2^2+60 mg^5 msq^4 mst1^4 mst2^2+783 mg^7 mst1^6 mst2^2-120 mg^5 msq^2 mst1^6 mst2^2-60 mg^3 msq^4 mst1^6 mst2^2-868 mg^5 mst1^8 mst2^2+120 mg^3 msq^2 mst1^8 mst2^2+267 mg^3 mst1^10 mst2^2+14 mg^11 mst2^4+30 mg^7 msq^4 mst2^4-155 mg^9 mst1^2 mst2^4-60 mg^7 msq^2 mst1^2 mst2^4-30 mg^5 msq^4 mst1^2 mst2^4+719 mg^7 mst1^4 mst2^4+60 mg^5 msq^2 mst1^4 mst2^4-30 mg^3 msq^4 mst1^4 mst2^4-1297 mg^5 mst1^6 mst2^4+60 mg^3 msq^2 mst1^6 mst2^4+30 mg msq^4 mst1^6 mst2^4+894 mg^3 mst1^8 mst2^4-60 mg msq^2 mst1^8 mst2^4-175 mg mst1^10 mst2^4+6 mg^9 mst2^6-60 mg^5 msq^4 mst2^6+177 mg^7 mst1^2 mst2^6+120 mg^5 msq^2 mst1^2 mst2^6+60 mg^3 msq^4 mst1^2 mst2^6-621 mg^5 mst1^4 mst2^6-120 mg^3 msq^2 mst1^4 mst2^6+731 mg^3 mst1^6 mst2^6-261 mg mst1^8 mst2^6-8 mg^7 mst2^8+30 mg^3 msq^4 mst2^8+23 mg^5 mst1^2 mst2^8-60 mg^3 msq^2 mst1^2 mst2^8-30 mg msq^4 mst1^2 mst2^8+25 mg^3 mst1^4 mst2^8+60 mg msq^2 mst1^4 mst2^8-56 mg mst1^6 mst2^8-15 mg^5 mst2^10-6 mg^3 mst1^2 mst2^10+21 mg mst1^4 mst2^10+9 mg^3 mst2^12-9 mg mst1^2 mst2^12) Xt^5)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^5)+(1280 mg^2 (mst1^2+mst2^2) (mg^2 mst1^4+2 mg^2 mst1^2 mst2^2-3 mst1^4 mst2^2) Xt^6)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2) (-mst1^2+mst2^2)^5)) Log[mst1^2/MR^2]^2+(-12 (2 mg^2-2 msq^2) (-((40 msq^2 mst1^2)/(3 (-mg^2+mst1^2)^3))+(40 mst1^4)/(3 (-mg^2+mst1^2)^3)+(10 msq^2)/(-mg^2+mst1^2)^2-(10 mst1^2)/(-mg^2+mst1^2)^2-10/(3 (-mg^2+mst1^2))-(40 msq^2 mst2^2)/(3 (-mg^2+mst2^2)^3)+(40 mst2^4)/(3 (-mg^2+mst2^2)^3)+(10 msq^2)/(-mg^2+mst2^2)^2-(10 mst2^2)/(-mg^2+mst2^2)^2-10/(3 (-mg^2+mst2^2)))-(640 (mg^2-msq^2) (-2 mg^3 msq^2+mg^3 mst1^2+mg msq^2 mst1^2+mg^3 mst2^2+mg msq^2 mst2^2-2 mg mst1^2 mst2^2) Xt)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)-1/(-mst1^2+mst2^2) 24 (2 mg^2-2 msq^2) (-((40 msq^2 mst1^2)/(3 (-mg^2+mst1^2)^3))+(40 mst1^4)/(3 (-mg^2+mst1^2)^3)+(10 msq^2)/(-mg^2+mst1^2)^2-(10 mst1^2)/(-mg^2+mst1^2)^2-10/(3 (-mg^2+mst1^2))-(40 msq^2 mst2^2)/(3 (-mg^2+mst2^2)^3)+(40 mst2^4)/(3 (-mg^2+mst2^2)^3)+(10 msq^2)/(-mg^2+mst2^2)^2-(10 mst2^2)/(-mg^2+mst2^2)^2-10/(3 (-mg^2+mst2^2))) Xt^2-(1280 (mg^2-msq^2) (-2 mg^3 msq^2+mg^3 mst1^2+mg msq^2 mst1^2+mg^3 mst2^2+mg msq^2 mst2^2-2 mg mst1^2 mst2^2) Xt^3)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+1/(-mst1^2+mst2^2)^3 12 (2 mg^2-2 msq^2) (mst1^2+mst2^2) (-((40 msq^2 mst1^2)/(3 (-mg^2+mst1^2)^3))+(40 mst1^4)/(3 (-mg^2+mst1^2)^3)+(10 msq^2)/(-mg^2+mst1^2)^2-(10 mst1^2)/(-mg^2+mst1^2)^2-10/(3 (-mg^2+mst1^2))-(40 msq^2 mst2^2)/(3 (-mg^2+mst2^2)^3)+(40 mst2^4)/(3 (-mg^2+mst2^2)^3)+(10 msq^2)/(-mg^2+mst2^2)^2-(10 mst2^2)/(-mg^2+mst2^2)^2-10/(3 (-mg^2+mst2^2))) Xt^4+(640 (mg^2-msq^2) (mst1^2+mst2^2) (-2 mg^3 msq^2+mg^3 mst1^2+mg msq^2 mst1^2+mg^3 mst2^2+mg msq^2 mst2^2-2 mg mst1^2 mst2^2) Xt^5)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)) PolyLog[2,1-mg^2/msq^2]+((8 (128 mg^12-290 mg^10 mst1^2+98 mg^8 mst1^4+9 mg^6 mst1^6+20 mg^4 mst1^8-9 mg^2 mst1^10-350 mg^10 mst2^2+762 mg^8 mst1^2 mst2^2-251 mg^6 mst1^4 mst2^2-151 mg^4 mst1^6 mst2^2+41 mg^2 mst1^8 mst2^2-3 mst1^10 mst2^2+420 mg^8 mst2^4-821 mg^6 mst1^2 mst2^4+417 mg^4 mst1^4 mst2^4+41 mg^2 mst1^6 mst2^4-mst1^8 mst2^4-217 mg^6 mst2^6+323 mg^4 mst1^2 mst2^6-167 mg^2 mst1^4 mst2^6-19 mst1^6 mst2^6+31 mg^4 mst2^8-34 mg^2 mst1^2 mst2^8+23 mst1^4 mst2^8))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))-(64 (20 mg^9-37 mg^7 mst1^2+19 mg^5 mst1^4-5 mg^3 mst1^6+3 mg mst1^8-73 mg^7 mst2^2+135 mg^5 mst1^2 mst2^2-23 mg^3 mst1^4 mst2^2-7 mg mst1^6 mst2^2+24 mg^5 mst2^4-78 mg^3 mst1^2 mst2^4+22 mg mst1^4 mst2^4) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(16 (128 mg^12-294 mg^10 mst1^2+98 mg^8 mst1^4+9 mg^6 mst1^6+20 mg^4 mst1^8-9 mg^2 mst1^10-858 mg^10 mst2^2+1414 mg^8 mst1^2 mst2^2-251 mg^6 mst1^4 mst2^2-151 mg^4 mst1^6 mst2^2+41 mg^2 mst1^8 mst2^2-3 mst1^10 mst2^2+1304 mg^8 mst2^4-2049 mg^6 mst1^2 mst2^4+417 mg^4 mst1^4 mst2^4+41 mg^2 mst1^6 mst2^4-mst1^8 mst2^4-525 mg^6 mst2^6+903 mg^4 mst1^2 mst2^6-167 mg^2 mst1^4 mst2^6-19 mst1^6 mst2^6-37 mg^4 mst2^8-34 mg^2 mst1^2 mst2^8+23 mst1^4 mst2^8) Xt^2)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(64 (2 mg^7-37 mg^5 mst1^2-2 mg^3 mst1^4-6 mg mst1^6-33 mg^5 mst2^2+110 mg^3 mst1^2 mst2^2+20 mg mst1^4 mst2^2-6 mg^3 mst2^4-75 mg mst1^2 mst2^4+27 mg mst2^6) Xt^3)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)-(8 (124 mg^12 mst1^2-284 mg^10 mst1^4+90 mg^8 mst1^6+11 mg^6 mst1^8+20 mg^4 mst1^10-9 mg^2 mst1^12+1156 mg^12 mst2^2-3068 mg^10 mst1^2 mst2^2+2702 mg^8 mst1^4 mst2^2-474 mg^6 mst1^6 mst2^2-201 mg^4 mst1^8 mst2^2+32 mg^2 mst1^10 mst2^2-3 mst1^12 mst2^2-3048 mg^10 mst2^4+7406 mg^8 mst1^2 mst2^4-5516 mg^6 mst1^4 mst2^4+882 mg^4 mst1^6 mst2^4+184 mg^2 mst1^8 mst2^4-4 mst1^10 mst2^4+2602 mg^8 mst2^6-6126 mg^6 mst1^2 mst2^6+3984 mg^4 mst1^4 mst2^6-502 mg^2 mst1^6 mst2^6-54 mst1^8 mst2^6-695 mg^6 mst2^8+1718 mg^4 mst1^2 mst2^8-883 mg^2 mst1^4 mst2^8+4 mst1^6 mst2^8-3 mg^4 mst2^10-102 mg^2 mst1^2 mst2^10+57 mst1^4 mst2^10) Xt^4)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^4)-(64 (-18 mg^5 mst1^2-mg^3 mst1^4-3 mg mst1^6+14 mg^5 mst2^2+20 mg^3 mst1^2 mst2^2+4 mg mst1^4 mst2^2-11 mg^3 mst2^4+mg mst1^2 mst2^4-6 mg mst2^6) Xt^5)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)) PolyLog[2,1-mg^2/mst1^2]+((40 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(-mg^2+mst1^2)^3-(640 mg (-msq^2+mst1^2)^2 Xt)/((mg^2-mst1^2)^2 (mst1^2-mst2^2))+(160 (-msq^2+mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4) Xt^2)/((-mg^2+mst1^2)^3 (-mst1^2+mst2^2))+(1280 mg (-msq^2+mst1^2)^2 Xt^3)/((mg^2-mst1^2)^2 (mst1^2-mst2^2)^2)-(80 (-msq^2+mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4) (mst1^2+mst2^2) Xt^4)/((-mg^2+mst1^2)^3 (-mst1^2+mst2^2)^3)-(640 mg (-msq^2+mst1^2)^2 (mst1^2+mst2^2) Xt^5)/((mg^2-mst1^2)^2 (mst1^2-mst2^2)^4)) PolyLog[2,1-msq^2/mst1^2]+(-((32 mg^4 (mg^4-34 mg^2 mst2^2+17 mst2^4))/(3 (mg^2-mst2^2)^4))+(8 (-128 mg^12+346 mg^10 mst1^2-280 mg^8 mst1^4+13 mg^6 mst1^6+37 mg^4 mst1^8+294 mg^10 mst2^2-902 mg^8 mst1^2 mst2^2+1025 mg^6 mst1^4 mst2^2-391 mg^4 mst1^6 mst2^2+34 mg^2 mst1^8 mst2^2-98 mg^8 mst2^4+251 mg^6 mst1^2 mst2^4-417 mg^4 mst1^4 mst2^4+167 mg^2 mst1^6 mst2^4-23 mst1^8 mst2^4-9 mg^6 mst2^6+151 mg^4 mst1^2 mst2^6-41 mg^2 mst1^4 mst2^6+19 mst1^6 mst2^6-20 mg^4 mst2^8-41 mg^2 mst1^2 mst2^8+mst1^4 mst2^8+9 mg^2 mst2^10+3 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(-((64 (18 mg^5-37 mg^3 mst1^2+22 mg mst1^4+mg^3 mst2^2-7 mg mst1^2 mst2^2+3 mg mst2^4))/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2)))+(128 (mg^7-50 mg^5 mst2^2+17 mg^3 mst2^4))/(3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))) Xt-(16 (128 mg^12-346 mg^10 mst1^2+280 mg^8 mst1^4-13 mg^6 mst1^6-37 mg^4 mst1^8-806 mg^10 mst2^2+2438 mg^8 mst1^2 mst2^2-2561 mg^6 mst1^4 mst2^2+903 mg^4 mst1^6 mst2^2-34 mg^2 mst1^8 mst2^2+98 mg^8 mst2^4-251 mg^6 mst1^2 mst2^4+417 mg^4 mst1^4 mst2^4-167 mg^2 mst1^6 mst2^4+23 mst1^8 mst2^4+9 mg^6 mst2^6-151 mg^4 mst1^2 mst2^6+41 mg^2 mst1^4 mst2^6-19 mst1^6 mst2^6+20 mg^4 mst2^8+41 mg^2 mst1^2 mst2^8-mst1^4 mst2^8-9 mg^2 mst2^10-3 mst1^2 mst2^10) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)+((64 mg (2 mg^2-mst1^2-mst2^2) (mg^4-34 mg^2 mst2^2+17 mst2^4))/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)+(128 (18 mg^5-37 mg^3 mst1^2+22 mg mst1^4+mg^3 mst2^2-7 mg mst1^2 mst2^2+3 mg mst2^4))/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2)^2)) Xt^3+(8 (124 mg^12 mst1^2-332 mg^10 mst1^4+262 mg^8 mst1^6-3 mg^6 mst1^8-39 mg^4 mst1^10+1156 mg^12 mst2^2-4100 mg^10 mst1^2 mst2^2+5326 mg^8 mst1^4 mst2^2-2990 mg^6 mst1^6 mst2^2+526 mg^4 mst1^8 mst2^2+34 mg^2 mst1^10 mst2^2-1968 mg^10 mst2^4+6466 mg^8 mst1^2 mst2^4-7596 mg^6 mst1^4 mst2^4+3712 mg^4 mst1^6 mst2^4-543 mg^2 mst1^8 mst2^4-11 mst1^10 mst2^4+746 mg^8 mst2^6-2186 mg^6 mst1^2 mst2^6+2210 mg^4 mst1^4 mst2^6-774 mg^2 mst1^6 mst2^6+4 mst1^8 mst2^6-25 mg^6 mst2^8-29 mg^4 mst1^2 mst2^8-20 mg^2 mst1^4 mst2^8+14 mst1^6 mst2^8+20 mg^4 mst2^10+32 mg^2 mst1^2 mst2^10-4 mst1^4 mst2^10-9 mg^2 mst2^12-3 mst1^2 mst2^12) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)+(-((1024 mg mst2^2 (-2 mg^2+mst1^2+mst2^2))/(3 (mg^2-mst2^2) (-mst1^2+mst2^2)^4))-(64 (mst1^2+mst2^2) (18 mg^5-37 mg^3 mst1^2+22 mg mst1^4+mg^3 mst2^2-7 mg mst1^2 mst2^2+3 mg mst2^4))/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2)^4)) Xt^5) PolyLog[2,1-mg^2/mst2^2]+((40 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(-mg^2+mst2^2)^3-(640 mg (-msq^2+mst2^2)^2 Xt)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(80 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4) Xt^2)/((-mg^2+mst2^2)^3 (-mst1^2+mst2^2))-(1280 mg (-msq^2+mst2^2)^2 Xt^3)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)-(40 (mst1^2+mst2^2) (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4) Xt^4)/((-mg^2+mst2^2)^3 (-mst1^2+mst2^2)^3)+(640 mg (-msq^2+mst2^2)^2 (mst1^2+mst2^2) Xt^5)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)) PolyLog[2,1-msq^2/mst2^2]+(-((4 (24 mg^10 mst1^2-28 mg^8 mst1^4+68 mg^6 mst1^6-58 mg^4 mst1^8+18 mg^2 mst1^10-24 mg^10 mst2^2-220 mg^6 mst1^4 mst2^2+188 mg^4 mst1^6 mst2^2-70 mg^2 mst1^8 mst2^2+6 mst1^10 mst2^2+28 mg^8 mst2^4+220 mg^6 mst1^2 mst2^4-16 mg^2 mst1^6 mst2^4+8 mst1^8 mst2^4-68 mg^6 mst2^6-188 mg^4 mst1^2 mst2^6+16 mg^2 mst1^4 mst2^6+58 mg^4 mst2^8+70 mg^2 mst1^2 mst2^8-8 mst1^4 mst2^8-18 mg^2 mst2^10-6 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (-mg^2+mst2^2)^3))+(32 (-10 mg^5 mst1^2+10 mg^3 mst1^4-6 mg mst1^6+10 mg^5 mst2^2+8 mg mst1^4 mst2^2-10 mg^3 mst2^4-8 mg mst1^2 mst2^4+6 mg mst2^6) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)-(16 (12 mg^10-14 mg^8 mst1^2+34 mg^6 mst1^4-29 mg^4 mst1^6+9 mg^2 mst1^8-14 mg^8 mst2^2-76 mg^6 mst1^2 mst2^2+65 mg^4 mst1^4 mst2^2-26 mg^2 mst1^6 mst2^2+3 mst1^8 mst2^2+34 mg^6 mst2^4+65 mg^4 mst1^2 mst2^4-34 mg^2 mst1^4 mst2^4+7 mst1^6 mst2^4-29 mg^4 mst2^6-26 mg^2 mst1^2 mst2^6+7 mst1^4 mst2^6+9 mg^2 mst2^8+3 mst1^2 mst2^8) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3)+(128 (5 mg^5-5 mg^3 mst1^2+3 mg mst1^4-5 mg^3 mst2^2-mg mst1^2 mst2^2+3 mg mst2^4) Xt^3)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)+(8 (mst1^2+mst2^2) (12 mg^10-14 mg^8 mst1^2+34 mg^6 mst1^4-29 mg^4 mst1^6+9 mg^2 mst1^8-14 mg^8 mst2^2-76 mg^6 mst1^2 mst2^2+65 mg^4 mst1^4 mst2^2-26 mg^2 mst1^6 mst2^2+3 mst1^8 mst2^2+34 mg^6 mst2^4+65 mg^4 mst1^2 mst2^4-34 mg^2 mst1^4 mst2^4+7 mst1^6 mst2^4-29 mg^4 mst2^6-26 mg^2 mst1^2 mst2^6+7 mst1^4 mst2^6+9 mg^2 mst2^8+3 mst1^2 mst2^8) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)-(64 (mst1^2+mst2^2) (5 mg^5-5 mg^3 mst1^2+3 mg mst1^4-5 mg^3 mst2^2-mg mst1^2 mst2^2+3 mg mst2^4) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)) PolyLog[2,1-mst2^2/mst1^2])+Log[mst1^2/MR^2] ((4 (4 mg^12 mst1^4-140 mg^10 mst1^6+204 mg^8 mst1^8-68 mg^6 mst1^10+33 mg^10 mst1^4 mst2^2+360 mg^8 msq^2 mst1^4 mst2^2+583 mg^8 mst1^6 mst2^2-840 mg^6 msq^2 mst1^6 mst2^2-1259 mg^6 mst1^8 mst2^2+420 mg^4 msq^2 mst1^8 mst2^2+587 mg^4 mst1^10 mst2^2-180 mg^2 msq^2 mst1^10 mst2^2-4 mg^12 mst2^4+611 mg^10 mst1^2 mst2^4-360 mg^8 msq^2 mst1^2 mst2^4-2127 mg^8 mst1^4 mst2^4+540 mg^6 msq^2 mst1^4 mst2^4+1975 mg^6 mst1^6 mst2^4+283 mg^4 mst1^8 mst2^4+600 mg^2 msq^2 mst1^8 mst2^4-594 mg^2 mst1^10 mst2^4-60 msq^2 mst1^10 mst2^4+8 mg^10 mst2^6-704 mg^8 mst1^2 mst2^6+300 mg^6 msq^2 mst1^2 mst2^6+2217 mg^6 mst1^4 mst2^6-240 mg^4 msq^2 mst1^4 mst2^6-2359 mg^4 mst1^6 mst2^6-720 mg^2 msq^2 mst1^6 mst2^6+571 mg^2 mst1^8 mst2^6-60 msq^2 mst1^8 mst2^6+171 mst1^10 mst2^6-4 mg^8 mst2^8+207 mg^6 mst1^2 mst2^8-180 mg^4 msq^2 mst1^2 mst2^8-541 mg^4 mst1^4 mst2^8+300 mg^2 msq^2 mst1^4 mst2^8+505 mg^2 mst1^6 mst2^8+120 msq^2 mst1^6 mst2^8-183 mst1^8 mst2^8-18 mg^4 mst1^2 mst2^10+30 mg^2 mst1^4 mst2^10+12 mst1^6 mst2^10))/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^2 (mst1^2-mst2^2))+(64 (16 mg^7 mst1^2-16 mg^5 mst1^4+22 mg^7 mst2^2-112 mg^5 mst1^2 mst2^2+87 mg^3 mst1^4 mst2^2+30 mg msq^2 mst1^4 mst2^2-20 mg^5 mst2^4+30 mg^3 msq^2 mst2^4+100 mg^3 mst1^2 mst2^4-60 mg msq^2 mst1^2 mst2^4-74 mg mst1^4 mst2^4+3 mg^3 mst2^6-6 mg mst1^2 mst2^6) Xt)/(3 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2) (mst1^2-mst2^2))-(8 (-4 mg^10 mst1^2+136 mg^8 mst1^4-68 mg^6 mst1^6+4 mg^10 mst2^2-695 mg^8 mst1^2 mst2^2+360 mg^6 msq^2 mst1^2 mst2^2+326 mg^6 mst1^4 mst2^2-300 mg^4 msq^2 mst1^4 mst2^2-5 mg^4 mst1^6 mst2^2+180 mg^2 msq^2 mst1^6 mst2^2+18 mg^2 mst1^8 mst2^2-8 mg^8 mst2^4+886 mg^6 mst1^2 mst2^4-300 mg^4 msq^2 mst1^2 mst2^4-40 mg^4 mst1^4 mst2^4-240 mg^2 msq^2 mst1^4 mst2^4-308 mg^2 mst1^6 mst2^4+60 msq^2 mst1^6 mst2^4+6 mst1^8 mst2^4+4 mg^6 mst2^6-301 mg^4 mst1^2 mst2^6+180 mg^2 msq^2 mst1^2 mst2^6-212 mg^2 mst1^4 mst2^6+60 msq^2 mst1^4 mst2^6+237 mst1^6 mst2^6+18 mg^2 mst1^2 mst2^8+6 mst1^4 mst2^8) Xt^2)/(3 mst1^2 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+(-((128 (14 mg^3-30 mg msq^2-3 mg mst1^2-3 mg mst2^2))/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)))-(128 mg (3 mst1^2+mst2^2) (-11 mg^4+3 mg^2 mst1^2+3 mg^2 mst2^2+5 mst1^2 mst2^2))/(3 (mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^3)+(256 (-8 mg^5 mst1^2+8 mg^3 mst1^4+7 mg^5 mst2^2+26 mg^3 mst1^2 mst2^2-17 mg mst1^4 mst2^2))/(3 (mg^2-mst1^2)^2 mst2^2 (-mst1^2+mst2^2)^2)) Xt^3+(4 (4 mg^12 mst1^4-140 mg^10 mst1^6+204 mg^8 mst1^8-68 mg^6 mst1^10+176 mg^12 mst1^2 mst2^2-4793 mg^10 mst1^4 mst2^2-360 mg^8 msq^2 mst1^4 mst2^2+7283 mg^8 mst1^6 mst2^2+1020 mg^6 msq^2 mst1^6 mst2^2-1881 mg^6 mst1^8 mst2^2-360 mg^4 msq^2 mst1^8 mst2^2-975 mg^4 mst1^10 mst2^2+180 mg^2 msq^2 mst1^10 mst2^2-18 mg^2 mst1^12 mst2^2-4 mg^12 mst2^4+667 mg^10 mst1^2 mst2^4-360 mg^8 msq^2 mst1^2 mst2^4+10101 mg^8 mst1^4 mst2^4+600 mg^6 msq^2 mst1^4 mst2^4-17667 mg^6 mst1^6 mst2^4-1380 mg^4 msq^2 mst1^6 mst2^4+6073 mg^4 mst1^8 mst2^4-360 mg^2 msq^2 mst1^8 mst2^4+1508 mg^2 mst1^10 mst2^4+60 msq^2 mst1^10 mst2^4-6 mst1^12 mst2^4+8 mg^10 mst2^6-1222 mg^8 mst1^2 mst2^6+300 mg^6 msq^2 mst1^2 mst2^6-6913 mg^6 mst1^4 mst2^6+480 mg^4 msq^2 mst1^4 mst2^6+13435 mg^4 mst1^6 mst2^6+420 mg^2 msq^2 mst1^6 mst2^6-5659 mg^2 mst1^8 mst2^6+240 msq^2 mst1^8 mst2^6-417 mst1^10 mst2^6-4 mg^8 mst2^8+493 mg^6 mst1^2 mst2^8-180 mg^4 msq^2 mst1^2 mst2^8+1361 mg^4 mst1^4 mst2^8-240 mg^2 msq^2 mst1^4 mst2^8-2665 mg^2 mst1^6 mst2^8-60 msq^2 mst1^6 mst2^8+1167 mst1^8 mst2^8-18 mg^4 mst1^2 mst2^10-24 mg^2 mst1^4 mst2^10-6 mst1^6 mst2^10) Xt^4)/(3 mst1^2 (mg^2-mst1^2)^3 mst2^2 (mg^2-mst2^2)^2 (mst1^2-mst2^2)^3)-(64 (16 mg^7 mst1^2-16 mg^5 mst1^4+48 mg^7 mst2^2+172 mg^5 mst1^2 mst2^2-30 mg^3 msq^2 mst1^2 mst2^2-185 mg^3 mst1^4 mst2^2-30 mg msq^2 mst1^4 mst2^2+3 mg mst1^6 mst2^2-62 mg^5 mst2^4+30 mg^3 msq^2 mst2^4-166 mg^3 mst1^2 mst2^4+30 mg msq^2 mst1^2 mst2^4+184 mg mst1^4 mst2^4+3 mg^3 mst2^6+3 mg mst1^2 mst2^6) Xt^5)/(3 (mg^2-mst1^2)^2 mst2^2 (mg^2-mst2^2) (-mst1^2+mst2^2)^3)+((80 (18 mg^10+12 mg^8 msq^2+9 mg^6 msq^4-47 mg^8 mst1^2-51 mg^6 msq^2 mst1^2+3 mg^4 msq^4 mst1^2+43 mg^6 mst1^4+24 mg^4 msq^2 mst1^4-14 mg^4 mst1^6-9 mg^2 msq^2 mst1^6-35 mg^8 mst2^2-3 mg^6 msq^2 mst2^2-18 mg^4 msq^4 mst2^2+91 mg^6 mst1^2 mst2^2+39 mg^4 msq^2 mst1^2 mst2^2-6 mg^2 msq^4 mst1^2 mst2^2-83 mg^4 mst1^4 mst2^2+15 mg^2 msq^2 mst1^4 mst2^2+27 mg^2 mst1^6 mst2^2-3 msq^2 mst1^6 mst2^2+17 mg^6 mst2^4+3 mg^4 msq^2 mst2^4+9 mg^2 msq^4 mst2^4-44 mg^4 mst1^2 mst2^4-24 mg^2 msq^2 mst1^2 mst2^4+3 msq^4 mst1^2 mst2^4+40 mg^2 mst1^4 mst2^4-3 msq^2 mst1^4 mst2^4-13 mst1^6 mst2^4))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2)+(-((640 mg msq^2)/((mg^2-mst1^2) (mg^2-mst2^2)))-(320 mg mst1^2)/(3 (mg^2-mst1^2) (mst1^2-mst2^2))+(320 (-6 mg msq^4+mg^3 mst1^2+12 mg msq^2 mst1^2-mg mst1^4))/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+(640 mg)/(3 (-mst1^2+mst2^2))) Xt-(160 (18 mg^8+18 mg^6 msq^2-33 mg^6 mst1^2-15 mg^4 msq^2 mst1^2+15 mg^4 mst1^4+9 mg^2 msq^2 mst1^4-35 mg^6 mst2^2-15 mg^4 msq^2 mst2^2+64 mg^4 mst1^2 mst2^2-12 mg^2 msq^2 mst1^2 mst2^2-29 mg^2 mst1^4 mst2^2+3 msq^2 mst1^4 mst2^2+17 mg^4 mst2^4+9 mg^2 msq^2 mst2^4-31 mg^2 mst1^2 mst2^4+3 msq^2 mst1^2 mst2^4+14 mst1^4 mst2^4) Xt^2)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))+((1280 mg msq^2)/((mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2))-(640 mg (3 mst1^2+mst2^2))/(3 (-mst1^2+mst2^2)^3)) Xt^3+((160 (2 mg^4 msq^2-3 mg^2 msq^4+6 mg^2 msq^2 mst1^2-msq^4 mst1^2))/((mg^2-mst1^2)^3 (mst1^2-mst2^2)^2)+(80 (mst1^2+mst2^2) (11 mg^8+18 mg^6 msq^2-22 mg^6 mst1^2-15 mg^4 msq^2 mst1^2+11 mg^4 mst1^4+9 mg^2 msq^2 mst1^4-22 mg^6 mst2^2-15 mg^4 msq^2 mst2^2+44 mg^4 mst1^2 mst2^2-12 mg^2 msq^2 mst1^2 mst2^2-22 mg^2 mst1^4 mst2^2+3 msq^2 mst1^4 mst2^2+11 mg^4 mst2^4+9 mg^2 msq^2 mst2^4-22 mg^2 mst1^2 mst2^4+3 msq^2 mst1^2 mst2^4+11 mst1^4 mst2^4))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)+(80 (4 mg^8-3 mst1^4 mst2^2 (5 mst1^2+mst2^2)+mg^6 (7 mst1^2+3 mst2^2)-mg^4 (29 mst1^4+15 mst1^2 mst2^2+6 mst2^4)+mg^2 (16 mst1^6+31 mst1^4 mst2^2+7 mst1^2 mst2^4)))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2) (-mst1^2+mst2^2)^3)) Xt^4+(-((640 mg mst1^2)/(3 (mg^2-mst1^2) (mst1^2-mst2^2)^3))-(1280 (-mg msq^4+2 mg msq^2 mst1^2))/((mg^2-mst1^2)^2 (mst1^2-mst2^2)^3)-(640 mg msq^2 (mst1^2+mst2^2))/((mg^2-mst1^2) (mg^2-mst2^2) (-mst1^2+mst2^2)^3)) Xt^5) Log[msq^2/MR^2]+(-24 (-((5 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(6 (-mg^2+mst1^2)^3))-(5 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(6 (-mg^2+mst2^2)^3))-24 ((40 mg (-msq^2+mst1^2)^2)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+(40 mg (-msq^2+mst2^2)^2)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt+1/(-mst1^2+mst2^2) 48 (-((5 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(6 (-mg^2+mst1^2)^3))-(5 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(6 (-mg^2+mst2^2)^3)) Xt^2+(48 ((40 mg (-msq^2+mst1^2)^2)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+(40 mg (-msq^2+mst2^2)^2)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt^3)/(-mst1^2+mst2^2)-1/(-mst1^2+mst2^2)^3 24 (mst1^2+mst2^2) (-((5 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(6 (-mg^2+mst1^2)^3))-(5 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(6 (-mg^2+mst2^2)^3)) Xt^4-(24 (mst1^2+mst2^2) ((40 mg (-msq^2+mst1^2)^2)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2))+(40 mg (-msq^2+mst2^2)^2)/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt^5)/(-mst1^2+mst2^2)^3) Log[msq^2/MR^2]^2+(-12 (2 mg^2-2 msq^2) (-((40 msq^2 mst1^2)/(3 (-mg^2+mst1^2)^3))+(40 mst1^4)/(3 (-mg^2+mst1^2)^3)+(10 msq^2)/(-mg^2+mst1^2)^2-(10 mst1^2)/(-mg^2+mst1^2)^2-10/(3 (-mg^2+mst1^2))-(40 msq^2 mst2^2)/(3 (-mg^2+mst2^2)^3)+(40 mst2^4)/(3 (-mg^2+mst2^2)^3)+(10 msq^2)/(-mg^2+mst2^2)^2-(10 mst2^2)/(-mg^2+mst2^2)^2-10/(3 (-mg^2+mst2^2)))-(640 (mg^2-msq^2) (-2 mg^3 msq^2+mg^3 mst1^2+mg msq^2 mst1^2+mg^3 mst2^2+mg msq^2 mst2^2-2 mg mst1^2 mst2^2) Xt)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2)+1/(-mst1^2+mst2^2) 24 (2 mg^2-2 msq^2) (-((40 msq^2 mst1^2)/(3 (-mg^2+mst1^2)^3))+(40 mst1^4)/(3 (-mg^2+mst1^2)^3)+(10 msq^2)/(-mg^2+mst1^2)^2-(10 mst1^2)/(-mg^2+mst1^2)^2-10/(3 (-mg^2+mst1^2))-(40 msq^2 mst2^2)/(3 (-mg^2+mst2^2)^3)+(40 mst2^4)/(3 (-mg^2+mst2^2)^3)+(10 msq^2)/(-mg^2+mst2^2)^2-(10 mst2^2)/(-mg^2+mst2^2)^2-10/(3 (-mg^2+mst2^2))) Xt^2+(1280 (mg^2-msq^2) (-2 mg^3 msq^2+mg^3 mst1^2+mg msq^2 mst1^2+mg^3 mst2^2+mg msq^2 mst2^2-2 mg mst1^2 mst2^2) Xt^3)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))-1/(-mst1^2+mst2^2)^3 12 (2 mg^2-2 msq^2) (mst1^2+mst2^2) (-((40 msq^2 mst1^2)/(3 (-mg^2+mst1^2)^3))+(40 mst1^4)/(3 (-mg^2+mst1^2)^3)+(10 msq^2)/(-mg^2+mst1^2)^2-(10 mst1^2)/(-mg^2+mst1^2)^2-10/(3 (-mg^2+mst1^2))-(40 msq^2 mst2^2)/(3 (-mg^2+mst2^2)^3)+(40 mst2^4)/(3 (-mg^2+mst2^2)^3)+(10 msq^2)/(-mg^2+mst2^2)^2-(10 mst2^2)/(-mg^2+mst2^2)^2-10/(3 (-mg^2+mst2^2))) Xt^4-(640 (mg^2-msq^2) (mst1^2+mst2^2) (-2 mg^3 msq^2+mg^3 mst1^2+mg msq^2 mst1^2+mg^3 mst2^2+mg msq^2 mst2^2-2 mg mst1^2 mst2^2) Xt^5)/((mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^3)) PolyLog[2,1-mg^2/msq^2]+(-((32 mg^4 (mg^4-34 mg^2 mst1^2+17 mst1^4))/(3 (mg^2-mst1^2)^4))-(8 (-128 mg^12+294 mg^10 mst1^2-98 mg^8 mst1^4-9 mg^6 mst1^6-20 mg^4 mst1^8+9 mg^2 mst1^10+346 mg^10 mst2^2-902 mg^8 mst1^2 mst2^2+251 mg^6 mst1^4 mst2^2+151 mg^4 mst1^6 mst2^2-41 mg^2 mst1^8 mst2^2+3 mst1^10 mst2^2-280 mg^8 mst2^4+1025 mg^6 mst1^2 mst2^4-417 mg^4 mst1^4 mst2^4-41 mg^2 mst1^6 mst2^4+mst1^8 mst2^4+13 mg^6 mst2^6-391 mg^4 mst1^2 mst2^6+167 mg^2 mst1^4 mst2^6+19 mst1^6 mst2^6+37 mg^4 mst2^8+34 mg^2 mst1^2 mst2^8-23 mst1^4 mst2^8))/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (-mst1^2+mst2^2))+((128 (mg^7-50 mg^5 mst1^2+17 mg^3 mst1^4))/(3 (mg^2-mst1^2)^3 (mst1^2-mst2^2))-(64 (18 mg^5+mg^3 mst1^2+3 mg mst1^4-37 mg^3 mst2^2-7 mg mst1^2 mst2^2+22 mg mst2^4))/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2))) Xt-(16 (128 mg^12-806 mg^10 mst1^2+98 mg^8 mst1^4+9 mg^6 mst1^6+20 mg^4 mst1^8-9 mg^2 mst1^10-346 mg^10 mst2^2+2438 mg^8 mst1^2 mst2^2-251 mg^6 mst1^4 mst2^2-151 mg^4 mst1^6 mst2^2+41 mg^2 mst1^8 mst2^2-3 mst1^10 mst2^2+280 mg^8 mst2^4-2561 mg^6 mst1^2 mst2^4+417 mg^4 mst1^4 mst2^4+41 mg^2 mst1^6 mst2^4-mst1^8 mst2^4-13 mg^6 mst2^6+903 mg^4 mst1^2 mst2^6-167 mg^2 mst1^4 mst2^6-19 mst1^6 mst2^6-37 mg^4 mst2^8-34 mg^2 mst1^2 mst2^8+23 mst1^4 mst2^8) Xt^2)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (mst1^2-mst2^2)^2)+((64 mg (mg^4-34 mg^2 mst1^2+17 mst1^4) (-2 mg^2+mst1^2+mst2^2))/(3 (mg^2-mst1^2)^2 (-mst1^2+mst2^2)^3)+(128 (18 mg^5+mg^3 mst1^2+3 mg mst1^4-37 mg^3 mst2^2-7 mg mst1^2 mst2^2+22 mg mst2^4))/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)) Xt^3+(8 (1156 mg^12 mst1^2-1968 mg^10 mst1^4+746 mg^8 mst1^6-25 mg^6 mst1^8+20 mg^4 mst1^10-9 mg^2 mst1^12+124 mg^12 mst2^2-4100 mg^10 mst1^2 mst2^2+6466 mg^8 mst1^4 mst2^2-2186 mg^6 mst1^6 mst2^2-29 mg^4 mst1^8 mst2^2+32 mg^2 mst1^10 mst2^2-3 mst1^12 mst2^2-332 mg^10 mst2^4+5326 mg^8 mst1^2 mst2^4-7596 mg^6 mst1^4 mst2^4+2210 mg^4 mst1^6 mst2^4-20 mg^2 mst1^8 mst2^4-4 mst1^10 mst2^4+262 mg^8 mst2^6-2990 mg^6 mst1^2 mst2^6+3712 mg^4 mst1^4 mst2^6-774 mg^2 mst1^6 mst2^6+14 mst1^8 mst2^6-3 mg^6 mst2^8+526 mg^4 mst1^2 mst2^8-543 mg^2 mst1^4 mst2^8+4 mst1^6 mst2^8-39 mg^4 mst2^10+34 mg^2 mst1^2 mst2^10-11 mst1^4 mst2^10) Xt^4)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^3 (mst1^2-mst2^2)^4)+((1024 mg mst1^2 (2 mg^2-mst1^2-mst2^2))/(3 (mg^2-mst1^2) (mst1^2-mst2^2)^4)-(64 (mst1^2+mst2^2) (18 mg^5+mg^3 mst1^2+3 mg mst1^4-37 mg^3 mst2^2-7 mg mst1^2 mst2^2+22 mg mst2^4))/(3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)) Xt^5) PolyLog[2,1-mg^2/mst1^2]+((40 (-2 msq^2+2 mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4))/(-mg^2+mst1^2)^3-(640 mg (-msq^2+mst1^2)^2 Xt)/((mg^2-mst1^2)^2 (mst1^2-mst2^2))-(160 (-msq^2+mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4) Xt^2)/((-mg^2+mst1^2)^3 (-mst1^2+mst2^2))-(1280 mg (-msq^2+mst1^2)^2 Xt^3)/((mg^2-mst1^2)^2 (mst1^2-mst2^2)^2)+(80 (-msq^2+mst1^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst1^2+msq^2 mst1^2+mst1^4) (mst1^2+mst2^2) Xt^4)/((-mg^2+mst1^2)^3 (-mst1^2+mst2^2)^3)+(640 mg (-msq^2+mst1^2)^2 (mst1^2+mst2^2) Xt^5)/((mg^2-mst1^2)^2 (mst1^2-mst2^2)^4)) PolyLog[2,1-msq^2/mst1^2]+(-((8 (128 mg^12-350 mg^10 mst1^2+420 mg^8 mst1^4-217 mg^6 mst1^6+31 mg^4 mst1^8-290 mg^10 mst2^2+762 mg^8 mst1^2 mst2^2-821 mg^6 mst1^4 mst2^2+323 mg^4 mst1^6 mst2^2-34 mg^2 mst1^8 mst2^2+98 mg^8 mst2^4-251 mg^6 mst1^2 mst2^4+417 mg^4 mst1^4 mst2^4-167 mg^2 mst1^6 mst2^4+23 mst1^8 mst2^4+9 mg^6 mst2^6-151 mg^4 mst1^2 mst2^6+41 mg^2 mst1^4 mst2^6-19 mst1^6 mst2^6+20 mg^4 mst2^8+41 mg^2 mst1^2 mst2^8-mst1^4 mst2^8-9 mg^2 mst2^10-3 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)))-(64 (20 mg^9-73 mg^7 mst1^2+24 mg^5 mst1^4-37 mg^7 mst2^2+135 mg^5 mst1^2 mst2^2-78 mg^3 mst1^4 mst2^2+19 mg^5 mst2^4-23 mg^3 mst1^2 mst2^4+22 mg mst1^4 mst2^4-5 mg^3 mst2^6-7 mg mst1^2 mst2^6+3 mg mst2^8) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (mst1^2-mst2^2))+(16 (128 mg^12-858 mg^10 mst1^2+1304 mg^8 mst1^4-525 mg^6 mst1^6-37 mg^4 mst1^8-294 mg^10 mst2^2+1414 mg^8 mst1^2 mst2^2-2049 mg^6 mst1^4 mst2^2+903 mg^4 mst1^6 mst2^2-34 mg^2 mst1^8 mst2^2+98 mg^8 mst2^4-251 mg^6 mst1^2 mst2^4+417 mg^4 mst1^4 mst2^4-167 mg^2 mst1^6 mst2^4+23 mst1^8 mst2^4+9 mg^6 mst2^6-151 mg^4 mst1^2 mst2^6+41 mg^2 mst1^4 mst2^6-19 mst1^6 mst2^6+20 mg^4 mst2^8+41 mg^2 mst1^2 mst2^8-mst1^4 mst2^8-9 mg^2 mst2^10-3 mst1^2 mst2^10) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)-(64 (2 mg^7-33 mg^5 mst1^2-6 mg^3 mst1^4+27 mg mst1^6-37 mg^5 mst2^2+110 mg^3 mst1^2 mst2^2-75 mg mst1^4 mst2^2-2 mg^3 mst2^4+20 mg mst1^2 mst2^4-6 mg mst2^6) Xt^3)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2)^3)-(8 (1156 mg^12 mst1^2-3048 mg^10 mst1^4+2602 mg^8 mst1^6-695 mg^6 mst1^8-3 mg^4 mst1^10+124 mg^12 mst2^2-3068 mg^10 mst1^2 mst2^2+7406 mg^8 mst1^4 mst2^2-6126 mg^6 mst1^6 mst2^2+1718 mg^4 mst1^8 mst2^2-102 mg^2 mst1^10 mst2^2-284 mg^10 mst2^4+2702 mg^8 mst1^2 mst2^4-5516 mg^6 mst1^4 mst2^4+3984 mg^4 mst1^6 mst2^4-883 mg^2 mst1^8 mst2^4+57 mst1^10 mst2^4+90 mg^8 mst2^6-474 mg^6 mst1^2 mst2^6+882 mg^4 mst1^4 mst2^6-502 mg^2 mst1^6 mst2^6+4 mst1^8 mst2^6+11 mg^6 mst2^8-201 mg^4 mst1^2 mst2^8+184 mg^2 mst1^4 mst2^8-54 mst1^6 mst2^8+20 mg^4 mst2^10+32 mg^2 mst1^2 mst2^10-4 mst1^4 mst2^10-9 mg^2 mst2^12-3 mst1^2 mst2^12) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^2 (mst1^2-mst2^2)^4)-(64 (14 mg^5 mst1^2-11 mg^3 mst1^4-6 mg mst1^6-18 mg^5 mst2^2+20 mg^3 mst1^2 mst2^2+mg mst1^4 mst2^2-mg^3 mst2^4+4 mg mst1^2 mst2^4-3 mg mst2^6) Xt^5)/(3 (mg^2-mst1^2)^2 (mst1^2-mst2^2)^4)) PolyLog[2,1-mg^2/mst2^2]+((40 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4))/(-mg^2+mst2^2)^3-(640 mg (-msq^2+mst2^2)^2 Xt)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2))-(80 (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4) Xt^2)/((-mg^2+mst2^2)^3 (-mst1^2+mst2^2))+(1280 mg (-msq^2+mst2^2)^2 Xt^3)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)+(40 (mst1^2+mst2^2) (-2 msq^2+2 mst2^2) (-2 mg^4+3 mg^2 msq^2-3 mg^2 mst2^2+msq^2 mst2^2+mst2^4) Xt^4)/((-mg^2+mst2^2)^3 (-mst1^2+mst2^2)^3)-(640 mg (-msq^2+mst2^2)^2 (mst1^2+mst2^2) Xt^5)/((mg^2-mst2^2)^2 (-mst1^2+mst2^2)^4)) PolyLog[2,1-msq^2/mst2^2]+(-((4 (24 mg^10 mst1^2-28 mg^8 mst1^4+68 mg^6 mst1^6-58 mg^4 mst1^8+18 mg^2 mst1^10-24 mg^10 mst2^2-220 mg^6 mst1^4 mst2^2+188 mg^4 mst1^6 mst2^2-70 mg^2 mst1^8 mst2^2+6 mst1^10 mst2^2+28 mg^8 mst2^4+220 mg^6 mst1^2 mst2^4-16 mg^2 mst1^6 mst2^4+8 mst1^8 mst2^4-68 mg^6 mst2^6-188 mg^4 mst1^2 mst2^6+16 mg^2 mst1^4 mst2^6+58 mg^4 mst2^8+70 mg^2 mst1^2 mst2^8-8 mst1^4 mst2^8-18 mg^2 mst2^10-6 mst1^2 mst2^10))/(3 (mg^2-mst1^2)^3 (-mg^2+mst2^2)^3))+(32 (-10 mg^5 mst1^2+10 mg^3 mst1^4-6 mg mst1^6+10 mg^5 mst2^2+8 mg mst1^4 mst2^2-10 mg^3 mst2^4-8 mg mst1^2 mst2^4+6 mg mst2^6) Xt)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)+(16 (12 mg^10-14 mg^8 mst1^2+34 mg^6 mst1^4-29 mg^4 mst1^6+9 mg^2 mst1^8-14 mg^8 mst2^2-76 mg^6 mst1^2 mst2^2+65 mg^4 mst1^4 mst2^2-26 mg^2 mst1^6 mst2^2+3 mst1^8 mst2^2+34 mg^6 mst2^4+65 mg^4 mst1^2 mst2^4-34 mg^2 mst1^4 mst2^4+7 mst1^6 mst2^4-29 mg^4 mst2^6-26 mg^2 mst1^2 mst2^6+7 mst1^4 mst2^6+9 mg^2 mst2^8+3 mst1^2 mst2^8) Xt^2)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3)-(128 (5 mg^5-5 mg^3 mst1^2+3 mg mst1^4-5 mg^3 mst2^2-mg mst1^2 mst2^2+3 mg mst2^4) Xt^3)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2)-(8 (mst1^2+mst2^2) (12 mg^10-14 mg^8 mst1^2+34 mg^6 mst1^4-29 mg^4 mst1^6+9 mg^2 mst1^8-14 mg^8 mst2^2-76 mg^6 mst1^2 mst2^2+65 mg^4 mst1^4 mst2^2-26 mg^2 mst1^6 mst2^2+3 mst1^8 mst2^2+34 mg^6 mst2^4+65 mg^4 mst1^2 mst2^4-34 mg^2 mst1^4 mst2^4+7 mst1^6 mst2^4-29 mg^4 mst2^6-26 mg^2 mst1^2 mst2^6+7 mst1^4 mst2^6+9 mg^2 mst2^8+3 mst1^2 mst2^8) Xt^4)/(3 (mg^2-mst1^2)^3 (mg^2-mst2^2)^3 (-mst1^2+mst2^2)^2)+(64 (mst1^2+mst2^2) (5 mg^5-5 mg^3 mst1^2+3 mg mst1^4-5 mg^3 mst2^2-mg mst1^2 mst2^2+3 mg mst2^4) Xt^5)/(3 (mg^2-mst1^2)^2 (mg^2-mst2^2)^2 (-mst1^2+mst2^2)^2)) PolyLog[2,1-mst2^2/mst1^2])))
);



(* ::Chapter::RGBColor[0., 0., 1.]::Closed:: *)
(*---*)


(* ::Subchapter::Closed:: *)
(*Timer end*)


FinPoleMass = DateList[];
ElapsedTimePoleMass = DateDifference[IniPoleMass, FinPoleMass, {"Minute","Second"}];
Print["Evaluation time: ",ElapsedTimePoleMass];



(* ::Subchapter::Closed:: *)
(*Package end*)


(*
End[];
*)


EndPackage[];

